/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* ###--------------------------------------------------------------### */
/* file		: c_sr8k_8a.c						*/
/* date		: Nov 14 1995						*/
/* version	: v2.0							*/
/* author	: Pirouz Bazargan Sabet					*/
/* content	: C model of a circuit					*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include BEH_H
#include "vh_ltype.h"
#include "vh_lspec.h"
#include "vh_xspec.h"

/* ###--------------------------------------------------------------### */
/* function	: c_sr8k_8a						*/
/* description	: A 8 KByte static Ram with a chip select pin		*/
/* called func.	: vhx_getvalu						*/
/* ###--------------------------------------------------------------### */

void c_sr8k_8a (pt_lkdins)

struct lkdins *pt_lkdins;

  {
  char e_n     ;			/* chip select			*/
  char w_n     ;			/* write enable			*/
  char dat [ 8];			/* input-output 8-bit data	*/
  char adr [13];			/* 13-bit address		*/
  char vdd     ;			/* power supply			*/
  char vss     ;			/* power supply			*/


  char          *mem      ;
  unsigned char  mask     ;
  int            i        ;
  unsigned int   address  = 0;
  unsigned char  data     = 0;
  struct wrbsig *pt_wrbsig;

	/* ###------------------------------------------------------### */
	/*   get the memory space reserved for the current instance	*/
	/* ###------------------------------------------------------### */

  vhx_getvalu (pt_lkdins, &mem, 1024 * 8);

	/* ###------------------------------------------------------### */
	/*   read the current value of input signals			*/
	/*     - skip vdd and vss					*/
	/* ###------------------------------------------------------### */

  for (i=0 ; i<13 ; i++)
    adr [i] = (* ((pt_lkdins->REDLST + 2  + i)->VALU)) - '0';

  for (i=0 ; i<8 ; i++)
    dat [i] = (* ((pt_lkdins->REDLST + 15 + i)->VALU)) - '0';

  w_n = (* ((pt_lkdins->REDLST + 23 )->VALU)) - '0';
  e_n = (* ((pt_lkdins->REDLST + 24 )->VALU)) - '0';

	/* ###------------------------------------------------------### */
	/*   built the word address from separate address bits		*/
	/* ###------------------------------------------------------### */

  if (e_n == 0)
    {
    for (i=0 ; i<13 ; i++)
      {
      address  = address << 1 ;
      address += adr [12-i]   ;
      }

	/* ###------------------------------------------------------### */
	/*   built the data word from separate data bis			*/
	/* ###------------------------------------------------------### */

    for (i=0 ; i< 8 ; i++)
      {
      data    = data   << 1 ;
      data   += dat [7-i]   ;
      }

	/* ###------------------------------------------------------### */
	/*   read or write the data from the memory			*/
	/* ###------------------------------------------------------### */

    if (w_n == 0)
      mem [address] = data;
    else
      data = mem [address];

	/* ###------------------------------------------------------### */
	/*   split the data onto separet data bits			*/
	/* ###------------------------------------------------------### */

    mask = 0x80;
    for (i=0 ; i< 8 ; i++)
      {
      if ((data & mask) == 0)
        dat [7-i] = 0;
      else
        dat [7-i] = 1;

      mask = mask >> 1;
      }
    }

	/* ###------------------------------------------------------### */
	/*   write the result into the projected value of ouput signals	*/
	/* ###------------------------------------------------------### */

  pt_wrbsig = pt_lkdins->WRBSIG;

  for (i=0 ; i<8 ; i++)
    {
    if ((e_n == 0) && (w_n == 1))
      pt_wrbsig->PRJBVL->CNDVAL = '1';
    else
      pt_wrbsig->PRJBVL->CNDVAL = '0';
    pt_wrbsig->PRJBVL->DRVVAL = dat [i] + '0';
    VHX_BUSUPD = addchain (VHX_BUSUPD, pt_wrbsig->BUSSIG);

    pt_wrbsig = pt_wrbsig->NEXT;
    }

  }
