/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/* ###--------------------------------------------------------------### */
/* file		: vh_xspec.c						*/
/* date		: Jun 10 1996						*/
/* version	: v2.01							*/
/* author	: VUONG H.N.						*/
/* description	: This file contains functions relative to the execution*/
/*		  of a circuit :					*/
/*									*/
/*		  vhx_redpat , vhx_inireg , vhx_wrtpat  , vhx_muxbit  ,	*/
/*		  vhx_worbit , vhx_regbit , vhx_bddeval , vhx_savall  ,	*/
/*		  vhx_execute, vhx_chkerr , vhx_update  , vhx_initfig ,	*/
/*		  vhx_core   , vhx_gexeval, vhx_getvalu , vhx_gexlevel,	*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <time.h>
#include MUT_H
#include BEH_H
#include LOG_H
#include PAT_H
#include "vh_ltype.h"
#include "vh_xspec.h"

/* ###--------------------------------------------------------------### */
/* function	: traceX						*/
/* description	: Just a function to trace the content of the EXECUTE	*/
/* ###--------------------------------------------------------------### */

void traceX ()
  {
  unsigned int  level    ;
  char         *type     ;
  char         *ins_name ;
  char         *name     ;
  struct ptype *pt_ptype ;

  for (level=0 ; level<256 ; level++)
    {
    pt_ptype = VHX_EXECUTE [level];

    if (pt_ptype != NULL)
      printf ("------------- level : %u -----------\n", level);

    while (pt_ptype != NULL)
      {
      switch (pt_ptype->TYPE)
        {
        case 'I':
          type     = "instance\t";
          ins_name = "";
          name = ((struct lkdins *) pt_ptype->DATA)->INSNAME;
          break;
        case 'S':
          type     = "simple_signal";
          ins_name = ((struct wrssig *)pt_ptype->DATA)->LKDINS->INSNAME;
          name     = ((struct wrssig *)pt_ptype->DATA)->SIMSIG->NAMECHAIN->DATA;
          break;
        case 'B' :
          type     = "bussed_signal";
          ins_name = ((struct wrbsig *)pt_ptype->DATA)->LKDINS->INSNAME;
          name     = ((struct wrbsig *)pt_ptype->DATA)->BUSSIG->NAMECHAIN->DATA;
          break;
        case 'R' :
          type     = "internal_register";
          ins_name = ((struct wrireg *) pt_ptype->DATA)->LKDINS->INSNAME;
          name     = ((struct wrireg *) pt_ptype->DATA)->BEREG->NAME;
          break;
        case 'U' :
          type     = "internal_bussed_signal";
          ins_name = ((struct wribux *) pt_ptype->DATA)->LKDINS->INSNAME;
          name     = ((struct wribux *) pt_ptype->DATA)->BEBUX->NAME;
          break;
        case 'D' :
        case 'X' :
          type     = "internal_simple_signal";
          ins_name = ((struct wriaux *) pt_ptype->DATA)->LKDINS->INSNAME;
          name     = ((struct wriaux *) pt_ptype->DATA)->BEAUX->NAME;
          break;
        case 'M' :
          type     = "assert_instruction";
          ins_name = ((struct wrimsg *) pt_ptype->DATA)->LKDINS->INSNAME;
          name     = "";
          break;
        }

      printf ("%s\t, %s, %s\n", type, ins_name, name);
      pt_ptype = pt_ptype->NEXT;
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: getXcount						*/
/* description	: Just a function to count the number of events in a	*/
/*		  delta cycle ?						*/
/* ###--------------------------------------------------------------### */

unsigned int getXcount ()
  {
  unsigned int  level    ;
  unsigned int  count    = 0;
  struct ptype *pt_ptype ;

  for (level=0 ; level<256 ; level++)
    {
    pt_ptype = VHX_EXECUTE [level];
    while (pt_ptype != NULL)
      {
      count ++;
      pt_ptype = pt_ptype->NEXT;
      }
    }
  return (count);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_getvalu						*/
/* description	: for the first call the function makes memory		*/
/*		  allocation to save the value of static internal	*/
/*		  signals of a description using a C model. For further	*/
/*		  calls it simply search the already allocated space.	*/
/*		  The pointer of the space is passed as argument to the	*/
/*		  function. The function returns the size of the memory	*/
/*		  space. Each instance has its own memory space.	*/
/*									*/
/* called funct	: mbkalloc, addht, gethtitem, addhtitem,		*/
/* ###--------------------------------------------------------------### */

unsigned int vhx_getvalu (pt_lkdins, pt_ptmem, size)

struct lkdins  *pt_lkdins;
char          **pt_ptmem ;
unsigned int    size     ;

  {
  char      *pt_mem   = NULL;
  static ht *mem_tabl = NULL;
  static ht *siz_tabl = NULL;

  if (mem_tabl == NULL)
    {
    mem_tabl = addht (64);
    siz_tabl = addht (64);
    }

  pt_mem = (char *) gethtitem (mem_tabl, pt_lkdins);

  if (((int) pt_mem) == EMPTYHT)
    {
    pt_mem = mbkalloc (size);
    addhtitem (mem_tabl, pt_lkdins, pt_mem);
    addhtitem (siz_tabl, pt_lkdins, size  );
    }

  size      = (unsigned int) gethtitem (siz_tabl, pt_lkdins);
  *pt_ptmem = pt_mem;

  return (size);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_getsize						*/
/* description	: get the size of a description				*/
/* called func.	: vhx_getvalu						*/
/* ###--------------------------------------------------------------### */

unsigned int vhx_getsize (pt_lkdfig)

struct lkdfig *pt_lkdfig;

  {
  unsigned int   size      = 0;
  struct simsig *pt_simsig ;
  struct bussig *pt_bussig ;
  struct prjbvl *pt_prjbvl ;
  struct lkdins *pt_lkdins ;
  struct wriaux *pt_wriaux ;
  struct wribux *pt_wribux ;
  struct wrireg *pt_wrireg ;
  struct prjrvl *pt_prjrvl ;
  char          *pt_mem    ;

  if (pt_lkdfig->VALSIZ == 0)
    {
    pt_simsig = pt_lkdfig->SIMSIG;
    while (pt_simsig != NULL)
      {
      size      += 2;
      pt_simsig  = pt_simsig->NEXT;
      }

    pt_bussig = pt_lkdfig->BUSSIG;
    while (pt_bussig != NULL)
      {
      pt_prjbvl = pt_bussig->PRJBVL;
      while (pt_prjbvl != NULL)
        {
        size      += 2;
        pt_prjbvl  = pt_prjbvl->GLNEXT;
        }
      size ++;
      pt_bussig = pt_bussig->NEXT;
      }

    pt_lkdins = pt_lkdfig->LKDINS;
    while (pt_lkdins != NULL)
      {
      if (pt_lkdins->FLAG == VHL_CMODEL)
        size += vhx_getvalu (pt_lkdins, &pt_mem, 0);
      else
        {
        pt_wriaux = pt_lkdins->WRIAUX;
        while (pt_wriaux != NULL)
          {
          size      += 2;
          pt_wriaux  = pt_wriaux->NEXT;
          }

        pt_wribux = pt_lkdins->WRIBUX;
        while (pt_wribux != NULL)
          {
          pt_prjbvl = pt_wribux->PRJBVL;
          while (pt_prjbvl != NULL)
            {
            size      += 2;
            pt_prjbvl  = pt_prjbvl->NEXT;
            }
          size ++;
          pt_wribux = pt_wribux->NEXT;
          }

        pt_wrireg = pt_lkdins->WRIREG;
        while (pt_wrireg != NULL)
          {
          pt_prjrvl = pt_wrireg->PRJRVL;
          while (pt_prjrvl != NULL)
            {
            size      += 2;
            pt_prjrvl  = pt_prjrvl->NEXT;
            }
          size ++;
          pt_wrireg = pt_wrireg->NEXT;
          }
        }
      pt_lkdins = pt_lkdins->NEXT;
      }

    pt_lkdfig->VALSIZ = size;
    }

  return (pt_lkdfig->VALSIZ);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_execall						*/
/* description	: execute the whole description				*/
/* called func.	: addptype						*/
/* ###--------------------------------------------------------------### */

void vhx_execall (pt_lkdfig)

struct lkdfig *pt_lkdfig;

  {
  struct lkdins *pt_lkdins = pt_lkdfig->LKDINS;

  while (pt_lkdins != NULL)
    {
    VHX_EXECUTE [0] = addptype (VHX_EXECUTE [0], 'I', pt_lkdins);
    pt_lkdins->EVAL = '1';

    pt_lkdins = pt_lkdins->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_addexec						*/
/* description	: add an element to the execution list			*/
/* called func.	: addptype						*/
/* ###--------------------------------------------------------------### */

void vhx_addexec (dep_list)

struct ptype *dep_list;

  {
  int           add_flg = 0;
  unsigned char level   ;

  while (dep_list != NULL)
    {
    switch (dep_list->TYPE)
      {
      case 'I' :
        if (((struct lkdins *) dep_list->DATA)->EVAL == 0)
          {
          ((struct lkdins *) dep_list->DATA)->EVAL = 1;

          level   = 0;
          add_flg = 1;
          }
        break;

      case 'S' :
        if (((struct wrssig *) dep_list->DATA)->EVAL == 0)
          {
          ((struct wrssig *) dep_list->DATA)->EVAL = 1;

          level   = ((struct wrssig *) dep_list->DATA)->LEVEL;
          add_flg = 1;
          }
        break;

      case 'B' :
        if (((struct wrbsig *) dep_list->DATA)->EVAL == 0)
          {
          ((struct wrbsig *) dep_list->DATA)->EVAL = 1;

          level   = ((struct wrbsig *) dep_list->DATA)->LEVEL;
          add_flg = 1;
          }
        break;

      case 'R' :
        if (((struct wrireg *) dep_list->DATA)->EVAL == 0)
          {
          ((struct wrireg *) dep_list->DATA)->EVAL = 1;

          level   = 0;
          add_flg = 1;
          }
        break;

      case 'U' :
        if (((struct wribux *) dep_list->DATA)->EVAL == 0)
          {
          ((struct wribux *) dep_list->DATA)->EVAL = 1;

          level   = ((struct wribux *) dep_list->DATA)->LEVEL;
          add_flg = 1;
          }
        break;

      case 'D' :
      case 'X' :
        if (((struct wriaux *) dep_list->DATA)->EVAL == 0)
          {
          ((struct wriaux *) dep_list->DATA)->EVAL = 1;

          level   = ((struct wriaux *) dep_list->DATA)->LEVEL;
          add_flg = 1;
          }
        break;

      case 'M' :
        if (((struct wrimsg *) dep_list->DATA)->EVAL == 0)
          {
          ((struct wrimsg *) dep_list->DATA)->EVAL = 1;

          level   = ((struct wrimsg *) dep_list->DATA)->LEVEL;
          add_flg = 1;
          }
        break;
      }

    if (add_flg == 1)
      {
      if (level < VHX_LEVEL)
        VHX_LEVEL = level;

      VHX_EXECUTE [level] = addptype (VHX_EXECUTE [level], dep_list->TYPE,
                                      dep_list->DATA     );
      add_flg             = 0;
      }

    dep_list = dep_list->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_gexeval						*/
/* description	: computes a GEX's value depending on its node's value	*/
/* called func.	: vhu_toolbug						*/
/* ###--------------------------------------------------------------### */

char vhx_gexeval (pt_lkdins, gex)

struct lkdins *pt_lkdins;
struct magex  *gex      ;

  {
  int           indexgex;
  int           nb      ;
  struct chain *expr    ;

  if (gex->OPERAND != NULL)
    {
    expr = gex->OPERAND;

    switch (gex->TERM)
      {
      case OR:
        while (expr != NULL)
          {
          if (vhx_gexeval (pt_lkdins, (struct magex *) expr->DATA) == '1')
            break;
          expr = expr->NEXT;
          }

        if (expr != NULL)
          return ('1');
        else
          return ('0');
        break;

      case AND:
        while (expr != NULL)
          {
          if (vhx_gexeval (pt_lkdins, (struct magex *)expr->DATA) == '0')
            break;
          expr = expr->NEXT;
          }

        if (expr != NULL)
          return ('0');
        else
          return ('1');
        break;

      case XOR:
        nb = 0;
        while (expr != NULL)
          {
          if (vhx_gexeval (pt_lkdins, (struct magex *)expr->DATA) == '1')
            nb++;
          expr = expr->NEXT;
          }
        if (nb & 1)
          return ('1');
        else
          return ('0');
        break;

      case NOT:
        if (vhx_gexeval (pt_lkdins, (struct magex *)expr->DATA) == '0')
          return ('1');
        else
          return ('0');
        break;

      case NOR:
        while (expr != NULL)
          {
          if (vhx_gexeval (pt_lkdins, (struct magex *)expr->DATA) == '1')
            break;
          expr = expr->NEXT;
          }

        if (expr != NULL)
          return ('0');
        else
          return ('1');
        break;

      case NAND:
        while (expr != NULL)
          {
          if (vhx_gexeval (pt_lkdins, (struct magex *)expr->DATA) == '0')
            break;
          expr = expr->NEXT;
          }
        if (expr != NULL)
          return ('1');
        else
          return ('0');
        break;

      case NXOR:
        nb = 0;
        while (expr)
          {
          if (vhx_gexeval (pt_lkdins, (struct magex *)expr->DATA) == '1')
            nb++;
          expr = expr->NEXT;
          }
        if (nb & 1)
          return ('0');
        else
          return ('1');
        break;

      default:
        vhu_toolbug (1, "vhx_gexeval", NULL, 0);
      }
    }

  else
    {
    indexgex = gex->TERM;

    if (indexgex > 1)
      return (*((pt_lkdins->REDLST + indexgex - 2)->VALU));
    else
      {
      if (indexgex == 0)
        return ('0');
      else
        return ('1');
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_bddlevel						*/
/* description	: returns the maximum level of an expression's operands	*/
/* called func.	: vhx_bddlevel						*/
/* ###--------------------------------------------------------------### */

unsigned char vhx_bddlevel (pt_lkdins, pt_gex)

struct lkdins *pt_lkdins;
struct magex  *pt_gex   ;

  {
  return (0);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_gexlevel						*/
/* description	: returns the maximum level of an expression's operands	*/
/* called func.	: vhx_gexlevel						*/
/* ###--------------------------------------------------------------### */

unsigned char vhx_gexlevel (pt_lkdins, pt_gex)

struct lkdins *pt_lkdins;
struct magex  *pt_gex   ;

  {
  struct chain  *pt_operand ;
  int            sig_index  ;
  unsigned char  level      ;
  unsigned char  max_level  = 0;

  if (pt_gex->OPERAND != NULL)
    {
    pt_operand = pt_gex->OPERAND;
    while (pt_operand != NULL)
      {
      level = vhx_gexlevel (pt_lkdins, (struct magex *)pt_operand->DATA);

      if (max_level < level)
        max_level = level;

      pt_operand = pt_operand->NEXT;
      }
    }
  else
    {
    sig_index = pt_gex->TERM;

    if (sig_index > 1)
      max_level = *((pt_lkdins->REDLST + sig_index - 2)->VALU);
    else
      max_level = 0;
    }

  return (max_level);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_muxbit						*/
/* description	: resolution function for MUX_BIT typed signals		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char vhx_muxbit (pt_prjbvl)

struct prjbvl *pt_prjbvl;

  {
  char resval      = '1';
  int  active_driv = 0  ;

  while (pt_prjbvl != NULL)
    {
    if (pt_prjbvl->CNDVAL == '1')
      {
      resval = pt_prjbvl->DRVVAL;
      active_driv++;
      }
    pt_prjbvl = pt_prjbvl->GLNEXT;
    }

  if (active_driv > 1)
    resval = 'E';

  return (resval);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_worbit						*/
/* description	: resolution function for WOR_BIT typed signals		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char vhx_worbit (pt_prjbvl)

struct prjbvl *pt_prjbvl;

  {
  char resval      = '1';
  int  active_one  = 0  ;
  int  active_zero = 0  ;

  while (pt_prjbvl != NULL)
    {
    if (pt_prjbvl->CNDVAL == '1')
      {
      if ((resval = pt_prjbvl->DRVVAL) == '1')
        active_one  = 1;
      else
        active_zero = 1;
      }
    pt_prjbvl = pt_prjbvl->GLNEXT;
    }

  if ((active_one + active_zero) > 1)
    resval = 'E';

  return (resval);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_regbit						*/
/* description	: resolution function for REG_BIT typed signals		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char vhx_regbit (pt_prjrvl)

struct prjrvl *pt_prjrvl;

  {
  char resval      = 'Z';
  int  active_driv = 0  ;

  while (pt_prjrvl != NULL)
    {
    if (pt_prjrvl->CNDVAL == '1')
      {
      resval = pt_prjrvl->DRVVAL;
      active_driv++;
      }
    pt_prjrvl = pt_prjrvl->NEXT;
    }

  if (active_driv > 1)
    resval = 'E';

  return (resval);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_bddeval						*/
/* description	: computes a BDD's value depending on its node's value	*/
/* called func.	: vhu_toolbug						*/
/* ###--------------------------------------------------------------### */

char vhx_bddeval (pt_lkdins, pt_bdd)

struct lkdins *pt_lkdins;
pNode          pt_bdd   ;

  {
  struct redlst *pt_redlst = pt_lkdins->REDLST - 2;
  char           value     ;

  while ((pt_bdd != zero) && (pt_bdd != one))
    {
    value = *((pt_redlst + pt_bdd->index)->VALU) ;

    if (value == '0')
      pt_bdd = pt_bdd->low ;
    else
      {
      if (value == '1')
        pt_bdd = pt_bdd->high ;
      else
        vhu_toolbug (8, "vhx_bddeval", NULL, value);
      }
    }

  if (pt_bdd == zero)
    return ('0');
  else
    return ('1');
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_savall						*/
/* description	: save the state of the circuit on a binary file	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void vhx_savall (fp, pt_lkdfig)

FILE          *fp       ;
struct lkdfig *pt_lkdfig;

  {
  struct simsig *pt_simsig = pt_lkdfig->SIMSIG;
  struct bussig *pt_bussig = pt_lkdfig->BUSSIG;
  struct lkdins *pt_lkdins = pt_lkdfig->LKDINS;
  unsigned int   size      = 0                ;
  char          *pt_mem    = NULL             ;
  struct wrireg *pt_wrireg ;
  struct wriaux *pt_wriaux ;
  struct wribux *pt_wribux ;
  struct prjbvl *pt_prjbvl ;
  struct prjrvl *pt_prjrvl ;

	/* ###------------------------------------------------------### */
	/*    get the size of the circuit (number of bytes needed to	*/
	/* store the value of all signals)				*/
	/* ###------------------------------------------------------### */

  vhx_getsize (pt_lkdfig);

	/* ###------------------------------------------------------### */
	/*    write the size as a header in the save file (for further	*/	
	/* checks)							*/
	/* ###------------------------------------------------------### */

  fwrite (&(pt_lkdfig->VALSIZ), sizeof (unsigned int), 1, fp);

	/* ###------------------------------------------------------### */
	/*    save the value of simple and bussed signals:		*/
	/*      - a current and a projected value for simple signals	*/
	/*      - a current and a list of projected condition and value	*/
	/*        for bussed signals					*/
	/* ###------------------------------------------------------### */

  while (pt_simsig != NULL)
    {
    fputc (  pt_simsig->PRJVAL , fp);
    fputc (*(pt_simsig->CURVAL), fp);

    pt_simsig = pt_simsig->NEXT;
    }

  while (pt_bussig != NULL)
    {
    pt_prjbvl = pt_bussig->PRJBVL;
    while (pt_prjbvl != NULL)
      {
      fputc (pt_prjbvl->DRVVAL, fp);
      fputc (pt_prjbvl->CNDVAL, fp);

      pt_prjbvl = pt_prjbvl->GLNEXT;
      }
    fputc (*(pt_bussig->CURVAL), fp);
    pt_bussig = pt_bussig->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    for each instance;					*/
	/*      - if the instance uses a C model save the value of	*/
	/*        the static variables used by the model		*/
	/*      - for other instances sav the current and projected	*/
	/*        values of internal signals (simple and bussed		*/
	/*        internal signals and internal registers)		*/
	/* ###------------------------------------------------------### */

  while (pt_lkdins != NULL)
    {
    if (pt_lkdins->FLAG == VHL_CMODEL)
      {
      size = vhx_getvalu (pt_lkdins, &pt_mem, 0);
      fwrite (pt_mem, sizeof (char), size, fp);
      }
    else
      {
      pt_wrireg = pt_lkdins->WRIREG;
      pt_wriaux = pt_lkdins->WRIAUX;
      pt_wribux = pt_lkdins->WRIBUX;

      while (pt_wrireg != NULL)
        {
        pt_prjrvl = pt_wrireg->PRJRVL;
        while (pt_prjrvl != NULL)
          {
          fputc (pt_prjrvl->DRVVAL, fp);
          fputc (pt_prjrvl->CNDVAL, fp);

          pt_prjrvl = pt_prjrvl->NEXT;
          }
        fputc (*(pt_wrireg->CURVAL), fp);
        pt_wrireg = pt_wrireg->NEXT;
        }

      while (pt_wribux != NULL)
        {
        pt_prjbvl = pt_wribux->PRJBVL;
        while (pt_prjbvl != NULL)
          {
          fputc (pt_prjbvl->DRVVAL, fp);
          fputc (pt_prjbvl->CNDVAL, fp);

          pt_prjbvl = pt_prjbvl->NEXT;
          }
        fputc (*(pt_wribux->CURVAL), fp);
        pt_wribux = pt_wribux->NEXT;
        }

      while (pt_wriaux != NULL)
        {
        fputc (  pt_wriaux->PRJVAL , fp);
        fputc (*(pt_wriaux->CURVAL), fp);

        pt_wriaux = pt_wriaux->NEXT;
        }
      }

    pt_lkdins = pt_lkdins->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_inireg						*/
/* description	: forces new values into registers			*/
/* called func.	: beh_chktab, vhu_warning, addchain			*/
/* ###--------------------------------------------------------------### */

void vhx_inireg (dict, pt_paini)

struct beden     **dict    ;
struct paini      *pt_paini;

  {
  char          *sig_name ;
  char          *ins_name ;
  struct wrireg *pt_wrireg;

	/* ###------------------------------------------------------### */
	/*    For each paini structure, first find the pointer of the	*/
	/* register then, force its value				*/
	/* ###------------------------------------------------------### */

  while (pt_paini != NULL)
    {
    vhu_splitname (pt_paini->NAME, &ins_name, &sig_name);

    pt_wrireg = (struct wrireg *)
                beh_chktab (dict, sig_name, ins_name, VHL_PNTDFN);

    if (pt_wrireg == NULL)
      vhu_warning (1, pt_paini->NAME);
    else
      {
      *(pt_wrireg->CURVAL) = pt_paini->VALUE;
      vhx_addexec (pt_wrireg->DEPEND);
      }
    pt_paini = pt_paini->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_redpat						*/
/* description	: reads a simulation pattern from the pattern		*/
/*		  description structure and initialises values in the	*/
/*		  circuit						*/
/* called func.	: addchain						*/
/* ###--------------------------------------------------------------### */

void vhx_redpat (pt_paseq, pt_papat)

struct paseq      *pt_paseq;
struct papat      *pt_papat;

  {
  struct paiol  *pt_paiol ;
  struct bussig *pt_bussig;
  struct simsig *pt_simsig;
  struct paevt  *pt_paevt ;

	/* ###------------------------------------------------------### */
	/*    Set the projected value of inputs (ports of mode I or T)	*/
	/* ###------------------------------------------------------### */

  pt_paevt = pt_papat->PAEVT;
  while (pt_paevt != NULL)
    {
    pt_paiol = pt_paseq->PAIOL + pt_paevt->INDEX;
    if (pt_paiol->MODE == 'I')
      {
      VHX_SIGUPD = addchain (VHX_SIGUPD, pt_paiol->SIG);
      ((struct simsig *) pt_paiol->SIG)->PRJVAL = pt_paevt->USRVAL;
      }
    else
      {
      if (pt_paiol->MODE == 'T')
        {
        VHX_BUSUPD = addchain (VHX_BUSUPD, pt_paiol->SIG);
        if ((pt_paevt->USRVAL == '1') || (pt_paevt->USRVAL == '0'))
          {
          ((struct bussig *) pt_paiol->SIG)->PRJBVL->DRVVAL = pt_paevt->USRVAL;
          ((struct bussig *) pt_paiol->SIG)->PRJBVL->CNDVAL = '1';
          }
        else
          ((struct bussig *) pt_paiol->SIG)->PRJBVL->CNDVAL = '0';
        }
      }
    pt_paevt = pt_paevt->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_wrtpat						*/
/* description	: saves in the pattern description structure the result	*/
/*		  of the simulation cycle				*/
/* called func.	: vhu_toolbug, vhu_error, pat_addpaevt			*/
/* ###--------------------------------------------------------------### */

int vhx_wrtpat (pt_paseq, pt_papat)

struct paseq *pt_paseq;
struct papat *pt_papat;

  {
  int            i        ;
  int            err_flg  = 0;
  int            max_iol  ;
  int            max_err  = 0;
  struct paiol  *pt_paiol ;
  struct bussig *pt_bussig;
  struct simsig *pt_simsig;
  struct wrireg *pt_wrireg;
  struct paevt  *pt_paevt ;
  char           value    ;

  max_iol  = pt_paseq->IOLNBR;
  pt_paiol = pt_paseq->PAIOL;

  for (i=0 ; i<max_iol ; i++)
    {
	/* ###------------------------------------------------------### */
	/*    For each input-output ...					*/
	/* ###------------------------------------------------------### */

    if (pt_paiol->MODE != 'I')
      {
	/* ###------------------------------------------------------### */
	/*    Read the input-output's value from the description -	*/
	/* skip simple inputs						*/
	/* ###------------------------------------------------------### */

      switch (pt_paiol->MODE)
        {
        case 'X' :
        case 'T' :
        case 'Z' :
          pt_bussig =  (struct bussig *) pt_paiol->SIG;
          value     = *(pt_bussig->CURVAL);
          break;

        case 'B' :
        case 'O' :
        case 'W' :
          pt_simsig =  (struct simsig *) pt_paiol->SIG;
          value     = *(pt_simsig->CURVAL);
          break;

        case 'S' :
          value     = *( ((struct wriaux *) pt_paiol->SIG)->CURVAL);
          break;

        case 'U' :
          value     = *( ((struct wribux *) pt_paiol->SIG)->CURVAL);
          break;

        case 'R' :
          pt_wrireg =  (struct wrireg *) pt_paiol->SIG;
          value     = *(pt_wrireg->CURVAL);
          break;

        default :
          vhu_toolbug (7, "vhx_wrtpat", pt_paiol->NAME, pt_paiol->MODE);
          break;
        }

	/* ###------------------------------------------------------### */
	/*    Scan user predicted event for the current input-output	*/
	/* ###------------------------------------------------------### */

      pt_paevt = pt_papat->PAEVT;
      while (pt_paevt != NULL)
        {
        if (pt_paevt->INDEX == i)
          break;
        pt_paevt = pt_paevt->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*    If no user predicted event has been found and an event has*/
	/* been occured, add a new paevt structure to the pattern. Use	*/
	/* a wrong value (the not of the simulated value) as user	*/
	/* defined value						*/
	/* ###------------------------------------------------------### */

      if ((pt_paevt == NULL) && (pt_paiol->VALUE != value))
        {
        pt_paevt        = pat_addpaevt (pt_papat->PAEVT, i, '-');
        pt_papat->PAEVT = pt_paevt;
        if (value == '0')
          {
          pt_paevt->USRVAL = '+';
          pt_paevt->SIMVAL = '+';
          }
        }

	/* ###------------------------------------------------------### */
	/*    If there is an event related to the input-output, and if	*/
	/* the value predicted by the user and the simulated value do	*/
	/* not match print out an error message. In such a case add an	*/
	/* event on the same input-ouput with a wrong predicted value	*/
	/* to the next pattern (if it exists)				*/
	/* ###------------------------------------------------------### */

      if (pt_paevt != NULL)
        {
        switch (pt_paevt->USRVAL)
          {
          case '-':
          case '0':
            if (value == '1')
              {
              vhu_error (113, pt_paiol->NAME, NULL);
              max_err ++ ;
              err_flg = 1;
              pt_paevt->SIMVAL = '+';
              }
            break;
          case '*':
            if (value == '1')
              pt_paevt->SIMVAL = '+';
            else
              pt_paevt->SIMVAL = '-';
            break;
          case '+':
          case '1':
            if (value == '0')
              {
              vhu_error (113, pt_paiol->NAME, NULL);
              max_err ++ ;
              err_flg = 1;
              pt_paevt->SIMVAL = '-';
              }
            break;
          }

        if ((err_flg == 1) && (pt_papat->NEXT != NULL))
          {
          pt_paevt = pt_papat->NEXT->PAEVT;
          while (pt_paevt != NULL)
            {
            if (pt_paevt->INDEX == i)
              break;
            pt_paevt = pt_paevt->NEXT;
            }

          if (pt_paevt == NULL)
            {
            pt_paevt              = pat_addpaevt (pt_papat->NEXT->PAEVT,i,'-');
            pt_papat->NEXT->PAEVT = pt_paevt;
            if (value == '0')
              {
              pt_paevt->USRVAL = '+';
              pt_paevt->SIMVAL = '+';
              }
            }
          }
        }
      }
	/* ###------------------------------------------------------### */
	/*    Save the value in the paiol structure to make event	*/
	/* detections possible for the next pattern			*/
	/* ###------------------------------------------------------### */

    err_flg = 0;
    pt_paiol->VALUE = value;
    pt_paiol ++;
    }

	/* ###------------------------------------------------------### */
	/*    Set the simulation flag of the pattern			*/
	/* ###------------------------------------------------------### */

  pt_papat->SIMFLAG = 'S';
  return (max_err);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_update						*/
/* description	: updates the current value of signals computed in the	*/
/*		  last evaluation cycle					*/
/* called func.	: vhx_muxbit, vhx_worbit, vhx_regbit, addchain,		*/
/*		  freechain						*/
/* ###--------------------------------------------------------------### */

void vhx_update ()

  {
  struct chain  *cur_chain;
  struct wrireg *pt_wrireg;
  struct wriaux *pt_wriaux;
  struct wribux *pt_wribux;
  struct bussig *pt_bussig;
  struct simsig *pt_simsig;
  char           resval   ;

	/* ###------------------------------------------------------### */
	/*    Update simple signals' value				*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_SIGUPD;
  while (cur_chain != NULL)
    {
    pt_simsig = (struct simsig *) cur_chain->DATA;
    if (pt_simsig->PRJVAL != *(pt_simsig->CURVAL))
      {
      *(pt_simsig->CURVAL) = pt_simsig->PRJVAL;
      vhx_addexec (pt_simsig->DEPEND);
      }
    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_SIGUPD);
  VHX_SIGUPD = NULL;

	/* ###------------------------------------------------------### */
	/*    Update bused signals's value (also handles errors occured	*/
	/* when resolution functions are called)			*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_BUSUPD;
  while (cur_chain != NULL)
    {
    pt_bussig = (struct bussig *) cur_chain->DATA;
    switch (pt_bussig->TYPE)
      {
      case 'W' :
        resval = vhx_worbit (pt_bussig->PRJBVL);
        break;
      case 'M' :
        resval = vhx_muxbit (pt_bussig->PRJBVL);
        break;
      default :
        vhu_toolbug (22, "vhx_update", pt_bussig->NAMECHAIN->DATA,
                     pt_bussig->TYPE);
        break;
      }

    if (resval == 'E')
      {
      if (pt_bussig->ERRFLG == 0)
        {
        pt_bussig->ERRFLG = 1;
        VHX_BUSERR ++;
        }
      }
    else
      {
      if (pt_bussig->ERRFLG == 1)
        {
        pt_bussig->ERRFLG = 0;
        VHX_BUSERR --;
        }

      if ( *(pt_bussig->CURVAL) != resval )
        {
        *(pt_bussig->CURVAL) = resval;
        vhx_addexec (pt_bussig->DEPEND);
        }
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_BUSUPD);
  VHX_BUSUPD = NULL;


	/* ###------------------------------------------------------### */
	/*    Update internal bussed signals's value (also handles	*/
	/* errors occured when resolution functions are called)		*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_BUXUPD;
  while (cur_chain != NULL)
    {
    pt_wribux = (struct wribux *) cur_chain->DATA;
    switch (pt_wribux->BEBUX->TYPE)
      {
      case 'W' :
        resval = vhx_worbit (pt_wribux->PRJBVL);
        break;
      case 'M' :
        resval = vhx_muxbit (pt_wribux->PRJBVL);
        break;
      default :
        vhu_toolbug (22, "vhx_update", pt_wribux->BEBUX->NAME,
                     pt_wribux->BEBUX->TYPE);
        break;
      }

    if (resval == 'E')
      {
      if (pt_wribux->ERRFLG == 0)
        {
        pt_wribux->ERRFLG = 1;
        VHX_BUXERR ++;
        }
      }
    else
      {
      if (pt_wribux->ERRFLG == 1)
        {
        pt_wribux->ERRFLG = 0;
        VHX_BUXERR --;
        }
      if ( *(pt_wribux->CURVAL) != resval )
        {
        *(pt_wribux->CURVAL) = resval;
        vhx_addexec (pt_wribux->DEPEND);
        }
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_BUXUPD);
  VHX_BUXUPD = NULL;

	/* ###------------------------------------------------------### */
	/*    Updating registers (also handles errors occured when	*/
	/* resolution functions are called)				*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_REGUPD;
  while (cur_chain != NULL)
    {
    pt_wrireg = (struct wrireg *)cur_chain->DATA;
    resval = vhx_regbit (pt_wrireg->PRJRVL);

    if (resval != 'Z')
      {
      if (resval == 'E')
        {
        if (pt_wrireg->ERRFLG == 0)
          {
          pt_wrireg->ERRFLG = 1;
          VHX_REGERR ++;
          }
        }
      else
        {
        if (pt_wrireg->ERRFLG == 1)
          {
          pt_wrireg->ERRFLG = 0;
          VHX_REGERR --;
          }
        if ( *(pt_wrireg->CURVAL) != resval )
          {
          *(pt_wrireg->CURVAL) = resval;
          vhx_addexec (pt_wrireg->DEPEND);
          }
        }
      }
    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_REGUPD);
  VHX_REGUPD = NULL;


	/* ###------------------------------------------------------### */
	/*    Updating simple internal signals				*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_AUXUPD;
  while (cur_chain != NULL)
    {
    pt_wriaux = (struct wriaux *)cur_chain->DATA;
    if (pt_wriaux->PRJVAL != *(pt_wriaux->CURVAL))
      {
      *(pt_wriaux->CURVAL) = pt_wriaux->PRJVAL;
      vhx_addexec (pt_wriaux->DEPEND);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_AUXUPD);
  VHX_AUXUPD = NULL;

  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_execute						*/
/* description	: executes all instances of the circuit (execution of a	*/
/*		  delta cycle)						*/
/* called func.	: vhx_gexeval, vhx_bddeval, addchain, freeptype		*/
/* ###--------------------------------------------------------------### */

int vhx_execute (flag)

char flag;			/* execute on bdd (1) or gex (0)	*/

  {
  struct ptype   *exec_list   ;
  struct lkdins  *pt_lkdins   ;
  struct beout   *pt_beout    ;
  struct bebus   *pt_bebus    ;
  struct bebux   *pt_bebux    ;
  struct beaux   *pt_beaux    ;
  struct bereg   *pt_bereg    ;
  struct bemsg   *pt_bemsg    ;
  struct wrssig  *pt_wrssig   ;
  struct wrbsig  *pt_wrbsig   ;
  struct wrireg  *pt_wrireg   ;
  struct wriaux  *pt_wriaux   ;
  struct wribux  *pt_wribux   ;
  struct wrimsg  *pt_wrimsg   ;
  struct binode  *pt_binode   ;
  struct prjbvl  *pt_prjbvl   ;
  struct prjrvl  *pt_prjrvl   ;
  void          (* pt_func) ();
  unsigned int    level       ;
  int             ret_flg     = 1;
  char            resval      ;

	/* ###------------------------------------------------------### */
	/*    find the level to be executed				*/
	/* ###------------------------------------------------------### */

  for (level=VHX_LEVEL ; level<256; level++)
    {
    if ((exec_list = VHX_EXECUTE [level]) != NULL)
      {
      VHX_LEVEL = level;
      ret_flg   = 0;
      break;
      }
    }

	/* ###------------------------------------------------------### */
	/*    execute all the signals that have that level		*/
	/* ###------------------------------------------------------### */

  while (exec_list != NULL)
    {
    switch (exec_list->TYPE)
      {
      case 'I' :

	/* ###------------------------------------------------------### */
	/*    for each lkdins that is to be executed ...		*/
	/* ###------------------------------------------------------### */

        pt_lkdins = (struct lkdins *) exec_list->DATA;

	/* ###------------------------------------------------------### */
	/*    for instances that have C models execute the C function	*/
	/* ###------------------------------------------------------### */

        if (pt_lkdins->FLAG == VHL_CMODEL)
          {
          pt_func = (void (*) ()) pt_lkdins->BEFIG;
          (void) (* pt_func) (pt_lkdins);
          }

	/* ###------------------------------------------------------### */
	/*    for instances that have VHDL models ...			*/
	/* ###------------------------------------------------------### */

        else
          {
	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRSSIGs			*/
	/* ###------------------------------------------------------### */

          pt_wrssig = pt_lkdins->WRSSIG;
          while (pt_wrssig != NULL)
            {
            pt_beout  = pt_wrssig->BEOUT;

            if (flag == 1)
              resval = vhx_bddeval (pt_lkdins, pt_beout->NODE);
            else
              resval = vhx_gexeval (pt_lkdins, pt_beout->NODE);

            pt_wrssig->SIMSIG->PRJVAL = resval;
            VHX_SIGUPD = addchain (VHX_SIGUPD, pt_wrssig->SIMSIG);
            pt_wrssig = pt_wrssig->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRIAUXs			*/
	/* ###------------------------------------------------------### */

          pt_wriaux = pt_lkdins->WRIAUX;
          while (pt_wriaux != NULL)
            {
            pt_beaux  = pt_wriaux->BEAUX;

            if (flag == 1)
              resval = vhx_bddeval (pt_lkdins, pt_beaux->NODE);
            else
              resval = vhx_gexeval (pt_lkdins, pt_beaux->NODE);

            pt_wriaux->PRJVAL = resval;
            VHX_AUXUPD = addchain (VHX_AUXUPD, pt_wriaux);
            pt_wriaux = pt_wriaux->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRBSIGs			*/
	/* ###------------------------------------------------------### */

          pt_wrbsig = pt_lkdins->WRBSIG;
          while (pt_wrbsig != NULL)
            {
            pt_bebus  = pt_wrbsig->BEBUS;
            pt_prjbvl = pt_wrbsig->PRJBVL;
            pt_binode = pt_bebus->BINODE;
            while (pt_binode != NULL)
              {
              if (flag == 1)
                {
                if ((pt_prjbvl->CNDVAL = vhx_bddeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
                  pt_prjbvl->DRVVAL = vhx_bddeval (pt_lkdins, pt_binode->VALNODE);
                }
              else
                {
                if ((pt_prjbvl->CNDVAL = vhx_gexeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
                  pt_prjbvl->DRVVAL = vhx_gexeval (pt_lkdins, pt_binode->VALNODE);
                }

              pt_prjbvl = pt_prjbvl->NEXT;
              pt_binode = pt_binode->NEXT;
              }
            VHX_BUSUPD = addchain (VHX_BUSUPD, pt_wrbsig->BUSSIG);
            pt_wrbsig  = pt_wrbsig->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRIREGs			*/
	/* ###------------------------------------------------------### */

          pt_wrireg = pt_lkdins->WRIREG;
          while (pt_wrireg != NULL)
            {
            pt_bereg  = pt_wrireg->BEREG ;
            pt_prjrvl = pt_wrireg->PRJRVL;
            pt_binode = pt_bereg->BINODE ;
            while (pt_binode != NULL)
              {
              if (flag == 1)
                {
                if ((pt_prjrvl->CNDVAL = vhx_bddeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
                  pt_prjrvl->DRVVAL = vhx_bddeval (pt_lkdins, pt_binode->VALNODE);
                }
              else
                {
                if ((pt_prjrvl->CNDVAL = vhx_gexeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
                  pt_prjrvl->DRVVAL = vhx_gexeval (pt_lkdins, pt_binode->VALNODE);
                }

              pt_prjrvl = pt_prjrvl->NEXT;
              pt_binode = pt_binode->NEXT;
              }
            VHX_REGUPD = addchain (VHX_REGUPD, pt_wrireg);
            pt_wrireg  = pt_wrireg->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Compute projected value of WRIBUXs			*/
	/* ###------------------------------------------------------### */

          pt_wribux = pt_lkdins->WRIBUX;
          while (pt_wribux != NULL)
            {
            pt_bebux  = pt_wribux->BEBUX ;
            pt_prjbvl = pt_wribux->PRJBVL;
            pt_binode = pt_bebux->BINODE ;
            while (pt_binode != NULL)
              {
              if (flag == 1)
                {
                if ((pt_prjbvl->CNDVAL = vhx_bddeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
                  pt_prjbvl->DRVVAL = vhx_bddeval (pt_lkdins, pt_binode->VALNODE);
                }
              else
                {
                if ((pt_prjbvl->CNDVAL = vhx_gexeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
                  pt_prjbvl->DRVVAL = vhx_gexeval (pt_lkdins, pt_binode->VALNODE);
                }

              pt_prjbvl = pt_prjbvl->NEXT;
              pt_binode = pt_binode->NEXT;
              }
            VHX_BUXUPD = addchain (VHX_BUXUPD, pt_wribux);
            pt_wribux  = pt_wribux->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Compute BEMSG's condition					*/
	/* ###------------------------------------------------------### */

          pt_wrimsg = pt_lkdins->WRIMSG;
          while (pt_wrimsg != NULL)
            {
            pt_bemsg = pt_wrimsg->BEMSG;

            if (flag == 1)
              resval = vhx_bddeval (pt_lkdins, pt_bemsg->NODE);
            else
              resval = vhx_gexeval (pt_lkdins, pt_bemsg->NODE);

            if (resval != '1')
              {
              if (pt_wrimsg->ERRFLG == 0)
                {
                pt_wrimsg->ERRFLG = 1;
                VHX_MSGERR ++;
                }
              }
            else
              {
              if (pt_wrimsg->ERRFLG == 1)
                {
                pt_wrimsg->ERRFLG = 0;
                VHX_MSGERR --;
                }
              }

            pt_wrimsg = pt_wrimsg->NEXT;
            }
          }
        pt_lkdins->EVAL = 0;
        break;

	/* ###------------------------------------------------------### */
	/*    for each simple output port that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'S' :
        pt_wrssig = (struct wrssig *) exec_list->DATA;

        pt_lkdins = pt_wrssig->LKDINS;
        pt_beout  = pt_wrssig->BEOUT ;

        if (flag == 1)
          resval = vhx_bddeval (pt_lkdins, pt_beout->NODE);
        else
          resval = vhx_gexeval (pt_lkdins, pt_beout->NODE);

        pt_wrssig->SIMSIG->PRJVAL = resval;
        pt_wrssig->EVAL           = 0;
        VHX_SIGUPD                = addchain (VHX_SIGUPD, pt_wrssig->SIMSIG);
        break;

	/* ###------------------------------------------------------### */
	/*    for each bussed output port that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'B' :
        pt_wrbsig = (struct wrbsig *) exec_list->DATA;

        pt_lkdins = pt_wrbsig->LKDINS;
        pt_bebus  = pt_wrbsig->BEBUS ;
        pt_prjbvl = pt_wrbsig->PRJBVL;
        pt_binode = pt_bebus->BINODE ;
        while (pt_binode != NULL)
          {
          if (flag == 1)
            {
            if ((pt_prjbvl->CNDVAL = vhx_bddeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
              pt_prjbvl->DRVVAL = vhx_bddeval (pt_lkdins, pt_binode->VALNODE);
            }
          else
            {
            if ((pt_prjbvl->CNDVAL = vhx_gexeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
              pt_prjbvl->DRVVAL = vhx_gexeval (pt_lkdins, pt_binode->VALNODE);
            }

          pt_prjbvl = pt_prjbvl->NEXT;
          pt_binode = pt_binode->NEXT;
          }
        pt_wrbsig->EVAL = 0;
        VHX_BUSUPD      = addchain (VHX_BUSUPD, pt_wrbsig->BUSSIG);
        break;

	/* ###------------------------------------------------------### */
	/*    for each internal register that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'R' :
        pt_wrireg = (struct wrireg *) exec_list->DATA;

        pt_lkdins = pt_wrireg->LKDINS;
        pt_bereg  = pt_wrireg->BEREG ;
        pt_prjrvl = pt_wrireg->PRJRVL;
        pt_binode = pt_bereg->BINODE ;
        while (pt_binode != NULL)
          {
          if (flag == 1)
            {
            if ((pt_prjrvl->CNDVAL = vhx_bddeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
              pt_prjrvl->DRVVAL = vhx_bddeval (pt_lkdins, pt_binode->VALNODE);
            }
          else
            {
            if ((pt_prjrvl->CNDVAL = vhx_gexeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
              pt_prjrvl->DRVVAL = vhx_gexeval (pt_lkdins, pt_binode->VALNODE);
            }

          pt_prjrvl = pt_prjrvl->NEXT;
          pt_binode = pt_binode->NEXT;
          }
        pt_wrireg->EVAL   = 0;
        VHX_REGUPD        = addchain (VHX_REGUPD, pt_wrireg);
        break;

	/* ###------------------------------------------------------### */
	/*    for each bussed internal signal that is to be executed ...*/
	/* ###------------------------------------------------------### */

      case 'U' :
        pt_wribux = (struct wribux *) exec_list->DATA;

        pt_lkdins = pt_wribux->LKDINS;
        pt_bebux  = pt_wribux->BEBUX ;
        pt_prjbvl = pt_wribux->PRJBVL;
        pt_binode = pt_bebux->BINODE ;
        while (pt_binode != NULL)
          {
          if (flag == 1)
            {
            if ((pt_prjbvl->CNDVAL = vhx_bddeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
              pt_prjbvl->DRVVAL = vhx_bddeval (pt_lkdins, pt_binode->VALNODE);
            }
          else
            {
            if ((pt_prjbvl->CNDVAL = vhx_gexeval (pt_lkdins, pt_binode->CNDNODE)) == '1')
              pt_prjbvl->DRVVAL = vhx_gexeval (pt_lkdins, pt_binode->VALNODE);
            }

          pt_prjbvl = pt_prjbvl->NEXT;
          pt_binode = pt_binode->NEXT;
          }

        pt_wribux->EVAL   = 0;
        VHX_BUXUPD        = addchain (VHX_BUXUPD, pt_wribux);
        break;

	/* ###------------------------------------------------------### */
	/*    for each simple output port that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'D' :
      case 'X' :
        pt_wriaux = (struct wriaux *) exec_list->DATA;

        pt_lkdins = pt_wriaux->LKDINS;
        pt_beaux  = pt_wriaux->BEAUX ;

        if (flag == 1)
          resval = vhx_bddeval (pt_lkdins, pt_beaux->NODE);
        else
          resval = vhx_gexeval (pt_lkdins, pt_beaux->NODE);

        pt_wriaux->PRJVAL = resval;
        pt_wriaux->EVAL   = 0;
        VHX_AUXUPD        = addchain (VHX_AUXUPD, pt_wriaux);
        break;

	/* ###------------------------------------------------------### */
	/*    for each simple output port that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'M' :
        pt_wrimsg = (struct wrimsg *) exec_list->DATA;

        pt_lkdins = pt_wrimsg->LKDINS;
        pt_bemsg  = pt_wrimsg->BEMSG ;

        if (flag == 1)
          resval = vhx_bddeval (pt_lkdins, pt_bemsg->NODE);
        else
          resval = vhx_gexeval (pt_lkdins, pt_bemsg->NODE);

        if (resval != '1')
          {
          if (pt_wrimsg->ERRFLG == 0)
            {
            pt_wrimsg->ERRFLG = 1;
            VHX_MSGERR ++;
            }
          }
        else
          {
          if (pt_wrimsg->ERRFLG == 1)
            {
            pt_wrimsg->ERRFLG = 0;
            VHX_MSGERR --;
            }
          }

        pt_wrimsg->EVAL = 0;
        break;
      }

    exec_list = exec_list ->NEXT;
    }

  if (ret_flg == 0)
    {
    freeptype (VHX_EXECUTE [level]);
    VHX_EXECUTE [level] = NULL;
    }

  return (ret_flg);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_chkerr						*/
/* description	: verifies that no errors remains at the end of the last*/
/*		  delta cycle						*/
/* called func.	: vhu_error						*/
/* ###--------------------------------------------------------------### */

int vhx_chkerr ()

  {
  short          errflg       = 0;
  struct lkdins *pt_lkdins    ;
  struct bussig *pt_bussig    ;
  struct wribux *pt_wribux    ;
  struct wrireg *pt_wrireg    ;
  struct wrimsg *pt_wrimsg    ;
  struct bemsg  *pt_bemsg     ;
  char          *level        ;
  char          *label        ;

  if (VHX_MSGERR != 0)
    {
    pt_lkdins = VHL_HEDLKF->LKDINS;
    while (pt_lkdins != NULL)
      {
      pt_wrimsg = pt_lkdins->WRIMSG;
      while (pt_wrimsg != NULL)
        {
        if (pt_wrimsg->ERRFLG == 1)
          {
          VHX_MSGERR --;
          pt_wrimsg->ERRFLG = 0;

          pt_bemsg = pt_wrimsg->BEMSG;
          if (pt_bemsg->LEVEL == 'W')
            level = "WARNING";
          else
            {
            level  = "ERROR";
            errflg = 1;
            }

          if (pt_bemsg->LABEL != NULL)
            label = pt_bemsg->LABEL;
          else
            label = "";

          fprintf (stderr, "%s: assert violation in %s: %s: %s\n",
                   level , pt_lkdins->INSNAME, label, pt_bemsg->MESSAGE);
          }
        pt_wrimsg = pt_wrimsg->NEXT;
        }
      pt_lkdins = pt_lkdins->NEXT;
      }
    }

  if (VHX_BUSERR != 0)
    {
    pt_bussig = VHL_HEDLKF->BUSSIG;
    while (pt_bussig != NULL)
      {
      if (pt_bussig->ERRFLG == 1)
        errflg = vhu_error (112, pt_bussig->NAMECHAIN->DATA, NULL);
      pt_bussig = pt_bussig->NEXT;
      }
    }

  if (VHX_BUXERR != 0)
    {
    pt_lkdins = VHL_HEDLKF->LKDINS;
    while (pt_lkdins != NULL)
      {
      pt_wribux = pt_lkdins->WRIBUX;
      while (pt_wribux != NULL)
        {
        if (pt_wribux->ERRFLG == 1)
          errflg = vhu_error (121, pt_lkdins->INSNAME, pt_wribux->BEBUX->NAME);
        pt_wribux = pt_wribux->NEXT;
        }
      pt_lkdins = pt_lkdins->NEXT;
      }
    }

  if (VHX_REGERR != 0)
    {
    pt_lkdins = VHL_HEDLKF->LKDINS;
    while (pt_lkdins != NULL)
      {
      pt_wrireg = pt_lkdins->WRIREG;
      while (pt_wrireg != NULL)
        {
        if (pt_wrireg->ERRFLG == 1)
          errflg = vhu_error (121, pt_lkdins->INSNAME, pt_wrireg->BEREG->NAME);
        pt_wrireg = pt_wrireg->NEXT;
        }
      pt_lkdins = pt_lkdins->NEXT;
      }
    }

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_initfig						*/
/* description	: initialize a description (reading values from a save	*/
/*		  file)							*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char vhx_initfig (pt_lkdfig, fp)

struct lkdfig *pt_lkdfig;
FILE          *fp       ;

  {
  struct simsig *pt_simsig = pt_lkdfig->SIMSIG;
  struct bussig *pt_bussig = pt_lkdfig->BUSSIG;
  struct lkdins *pt_lkdins = pt_lkdfig->LKDINS;
  struct wrireg *pt_wrireg ;
  struct wriaux *pt_wriaux ;
  struct wribux *pt_wribux ;
  struct prjbvl *pt_prjbvl ;
  struct prjrvl *pt_prjrvl ;
  unsigned int   size      ;
  char          *pt_mem    ;
  char           errflg    = 0;

	/* ###------------------------------------------------------### */
	/*    get the size of the circuit (number of bytes needed to	*/
	/* store the value of all signals)				*/
	/* ###------------------------------------------------------### */

  vhx_getsize (pt_lkdfig);

	/* ###------------------------------------------------------### */
	/*    read the size stored in the initialization file. An error	*/
	/* is detected if the size read from the file differs from the	*/
	/* the size of the circuit.					*/
	/* ###------------------------------------------------------### */

  fread (&size, sizeof (unsigned int), 1, fp);
  if (size != pt_lkdfig->VALSIZ)
    {
    errflg = vhu_error (115, NULL, NULL);
    }
  else
    {
	/* ###------------------------------------------------------### */
	/*    if sizes are the same, initialize the current and		*/
	/* projected values of simple and bussed signals		*/
	/* ###------------------------------------------------------### */

    while (pt_simsig != NULL)
      {
      pt_simsig->PRJVAL    = fgetc (fp);
      *(pt_simsig->CURVAL) = fgetc (fp);
      pt_simsig            = pt_simsig->NEXT;
      }

    while (pt_bussig != NULL)
      {
      pt_prjbvl = pt_bussig->PRJBVL;
      while (pt_prjbvl != NULL)
        {
        pt_prjbvl->DRVVAL = fgetc (fp);
        pt_prjbvl->CNDVAL = fgetc (fp);
        pt_prjbvl         = pt_prjbvl->GLNEXT;
        }
      *(pt_bussig->CURVAL) = fgetc (fp);
      pt_bussig = pt_bussig->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    for each instance:					*/	
	/*      - if the instance uses a C model get the pointer of the	*/
	/*        area used to store the static variables of the	*/
	/*        instance. Initialize the area with values read from	*/
	/*        the file.						*/
	/*      - for other instances initialize the current and	*/
	/*        projected values of intsrenal signals (simple and	*/
	/*        bussed signals and internal registers) with values	*/
	/*        read from the initialization file.			*/
	/* ###------------------------------------------------------### */

    while (pt_lkdins != NULL)
      {
      if (pt_lkdins->FLAG == VHL_CMODEL)
        {
        size = vhx_getvalu (pt_lkdins, &pt_mem, 0);
        fread (pt_mem, sizeof (char), size, fp);
        }
      else
        {
        pt_wrireg = pt_lkdins->WRIREG;
        pt_wriaux = pt_lkdins->WRIAUX;
        pt_wribux = pt_lkdins->WRIBUX;

        while (pt_wrireg != NULL)
          {
          pt_prjrvl = pt_wrireg->PRJRVL;
          while (pt_prjrvl != NULL)
            {
            pt_prjrvl->DRVVAL = fgetc (fp);
            pt_prjrvl->CNDVAL = fgetc (fp);
            pt_prjrvl         = pt_prjrvl->NEXT;
            }
          *(pt_wrireg->CURVAL) = fgetc (fp);
          pt_wrireg = pt_wrireg->NEXT;
          }

        while (pt_wribux != NULL)
          {
          pt_prjbvl = pt_wribux->PRJBVL;
          while (pt_prjbvl != NULL)
            {
            pt_prjbvl->DRVVAL = fgetc (fp);
            pt_prjbvl->CNDVAL = fgetc (fp);
            pt_prjbvl         = pt_prjbvl->NEXT;
            }
          *(pt_wribux->CURVAL) = fgetc (fp);
          pt_wribux = pt_wribux->NEXT;
          }

        while (pt_wriaux != NULL)
          {
          pt_wriaux->PRJVAL    = fgetc (fp);
          *(pt_wriaux->CURVAL) = fgetc (fp);
          pt_wriaux            = pt_wriaux->NEXT;
          }
        }

      pt_lkdins = pt_lkdins->NEXT;
      }
    }
  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_initval						*/
/* description	: initialize a description (set all values to the same	*/
/*		  value)						*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void vhx_initval (pt_lkdfig, value)

struct lkdfig *pt_lkdfig;
unsigned char  value    ;

  {
  struct simsig *pt_simsig = pt_lkdfig->SIMSIG;
  struct bussig *pt_bussig = pt_lkdfig->BUSSIG;
  struct lkdins *pt_lkdins = pt_lkdfig->LKDINS;
  struct wrireg *pt_wrireg ;
  struct wriaux *pt_wriaux ;
  struct wribux *pt_wribux ;

  while (pt_simsig != NULL)
    {
    *(pt_simsig->CURVAL) = value;
    pt_simsig = pt_simsig->NEXT;
    }

  while (pt_bussig != NULL)
    {
    *(pt_bussig->CURVAL) = value;
    pt_bussig = pt_bussig->NEXT;
    }

  while (pt_lkdins != NULL)
    {
    pt_wrireg = pt_lkdins->WRIREG;
    pt_wriaux = pt_lkdins->WRIAUX;
    pt_wribux = pt_lkdins->WRIBUX;

    while (pt_wrireg != NULL)
      {
      *(pt_wrireg->CURVAL) = value;
      pt_wrireg = pt_wrireg->NEXT;
      }

    while (pt_wribux != NULL)
      {
      *(pt_wribux->CURVAL) = value;
      pt_wribux = pt_wribux->NEXT;
      }

    while (pt_wriaux != NULL)
      {
      *(pt_wriaux->CURVAL) = value;
      pt_wriaux = pt_wriaux->NEXT;
      }
    pt_lkdins = pt_lkdins->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_core						*/
/* description	: create a core file of the lkdfig			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void vhx_core (fp, pt_lkdfig)

FILE          *fp       ;
struct lkdfig *pt_lkdfig;

  {
  time_t         clock    ;
  struct lkdins *pt_lkdins;
  struct simsig *pt_simsig;
  struct bussig *pt_bussig;
  struct wrireg *pt_wrireg;
  struct wriaux *pt_wriaux;
  struct wribux *pt_wribux;

	/* ###------------------------------------------------------### */
	/*   print a header for the core file				*/
	/* ###------------------------------------------------------### */

  time (&clock);

  fprintf (fp, "Core file generated from %s\n", pt_lkdfig->NAME);
  fprintf (fp, "\t\tdate : %s \n\n", ctime (&clock));

	/* ###------------------------------------------------------### */
	/*   print simple signals' current value			*/
	/* ###------------------------------------------------------### */

  fprintf (fp, "\t------- simple signals --------\n");
  pt_simsig = pt_lkdfig->SIMSIG;
  while (pt_simsig != NULL)
    {
    if (pt_simsig->NAMECHAIN != NULL)
      {
      fprintf (fp, "%s = %c\n", (char *) (pt_simsig->NAMECHAIN->DATA),
                                        *(pt_simsig->CURVAL));
      }
    pt_simsig = pt_simsig->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*   print bussed signals' current value			*/
	/* ###------------------------------------------------------### */

  fprintf (fp, "\t------- buses --------\n");
  pt_bussig = pt_lkdfig->BUSSIG;
  while (pt_bussig != NULL)
    {
    if (pt_bussig->NAMECHAIN != NULL)
      {
      fprintf (fp, "%s = %c\n", (char *) (pt_bussig->NAMECHAIN->DATA),
                                        *(pt_bussig->CURVAL));
      }
    pt_bussig = pt_bussig->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*   for each instance :					*/
	/*     - print instance's name					*/
	/*     - internal registers' current value			*/
	/*     - internal buses' current value				*/
	/*     - internal simple signals' current value			*/
	/* ###------------------------------------------------------### */

  pt_lkdins = pt_lkdfig->LKDINS;
  while (pt_lkdins != NULL)
    {
    fprintf (fp, "\t======= instance %s =======\n\n", pt_lkdins->INSNAME);

    fprintf (fp, "\t------- internal registers --------\n");
    pt_wrireg = pt_lkdins->WRIREG;
    while (pt_wrireg != NULL)
      {
      fprintf (fp, "%s.%s = %c\n",  pt_wrireg->LKDINS->INSNAME,
                                    pt_wrireg->BEREG->NAME,
                                  *(pt_wrireg->CURVAL));
      pt_wrireg = pt_wrireg->NEXT;
      }

    fprintf (fp, "\t------- internal simple signals -------\n");
    pt_wriaux = pt_lkdins->WRIAUX;
    while (pt_wriaux != NULL)
      {
      fprintf (fp, "%s.%s = %c\n",  pt_wriaux->LKDINS->INSNAME,
                                    pt_wriaux->BEAUX->NAME,
                                  *(pt_wriaux->CURVAL));
      pt_wriaux = pt_wriaux->NEXT;
      }

    fprintf (fp, "\t------- internal bus --------\n");
    pt_wribux = pt_lkdins->WRIBUX;
    while (pt_wribux != NULL)
      {
      fprintf (fp, "%s.%s = %c\n",  pt_wribux->LKDINS->INSNAME,
                                    pt_wribux->BEBUX->NAME,
                                  *(pt_wribux->CURVAL));
      pt_wribux = pt_wribux->NEXT;
      }
    pt_lkdins = pt_lkdins->NEXT;
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_getlevel						*/
/* description	: save the level of each signal in the corresponding	*/
/*		  field							*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void vhx_getlevel (pt_lkdfig)

struct lkdfig *pt_lkdfig;

  {
  unsigned char  level    ;
  struct lkdins *pt_lkdins;
  struct wrssig *pt_wrssig;
  struct wrbsig *pt_wrbsig;
  struct wriaux *pt_wriaux;
  struct wribux *pt_wribux;

  pt_lkdins = pt_lkdfig->LKDINS;
  while (pt_lkdins != NULL)
    {
    if (pt_lkdins->FLAG != VHL_CMODEL)
      {
      pt_wrssig = pt_lkdins->WRSSIG;
      while (pt_wrssig != NULL)
        {
        pt_wrssig->LEVEL = *(pt_wrssig->SIMSIG->CURVAL);
        pt_wrssig        =   pt_wrssig->NEXT;
        }

      pt_wrbsig = pt_lkdins->WRBSIG;
      while (pt_wrbsig != NULL)
        {
        pt_wrbsig->LEVEL = *(pt_wrbsig->BUSSIG->CURVAL);
        pt_wrbsig        =   pt_wrbsig->NEXT;
        }

      pt_wriaux = pt_lkdins->WRIAUX;
      while (pt_wriaux != NULL)
        {
        pt_wriaux->LEVEL = *(pt_wriaux->CURVAL);
        pt_wriaux        =   pt_wriaux->NEXT;
        }

      pt_wribux = pt_lkdins->WRIBUX;
      while (pt_wribux != NULL)
        {
        pt_wribux->LEVEL = *(pt_wribux->CURVAL);
        pt_wribux        =   pt_wribux->NEXT;
        }

      }
    pt_lkdins = pt_lkdins->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_execlevel						*/
/* description	: evalue the level of a signal or an instance		*/
/* called func.	: vhx_gexlevel, vhx_bddlevel, addchain, freeptype	*/
/* ###--------------------------------------------------------------### */

int vhx_execlevel (flag)

char flag;			/* execute on bdd (1) or gex (0)	*/

  {
  struct ptype   *exec_list   ;
  struct lkdins  *pt_lkdins   ;
  struct beout   *pt_beout    ;
  struct bebus   *pt_bebus    ;
  struct bebux   *pt_bebux    ;
  struct beaux   *pt_beaux    ;
  struct bemsg   *pt_bemsg    ;
  struct wrssig  *pt_wrssig   ;
  struct wrbsig  *pt_wrbsig   ;
  struct wriaux  *pt_wriaux   ;
  struct wribux  *pt_wribux   ;
  struct wrimsg  *pt_wrimsg   ;
  struct binode  *pt_binode   ;
  struct prjbvl  *pt_prjbvl   ;
  unsigned int    level       ;
  int             ret_flg     = 1;
  unsigned char   reslevel    ;

	/* ###------------------------------------------------------### */
	/*    find the level to be executed				*/
	/* ###------------------------------------------------------### */

  for (level=VHX_LEVEL ; level<256; level++)
    {
    if ((exec_list = VHX_EXECUTE [level]) != NULL)
      {
      VHX_LEVEL = level;
      ret_flg   = 0;
      break;
      }
    }

	/* ###------------------------------------------------------### */
	/*    compute the level of all signals that have that level	*/
	/* ###------------------------------------------------------### */

  while (exec_list != NULL)
    {
    switch (exec_list->TYPE)
      {
      case 'I' :

	/* ###------------------------------------------------------### */
	/*    for each lkdins that is to be executed ...		*/
	/* ###------------------------------------------------------### */

        pt_lkdins = (struct lkdins *) exec_list->DATA;

	/* ###------------------------------------------------------### */
	/*    for instances that have C models the level is zero.	*/
	/*    for instances that have VHDL models ...			*/
	/* ###------------------------------------------------------### */

        if (pt_lkdins->FLAG != VHL_CMODEL)
          {

	/* ###------------------------------------------------------### */
	/*    Compute projected level of simple outputs			*/
	/* ###------------------------------------------------------### */

          pt_wrssig = pt_lkdins->WRSSIG;
          while (pt_wrssig != NULL)
            {
            pt_beout  = pt_wrssig->BEOUT;

            if (flag == 1)
              reslevel = vhx_bddlevel (pt_lkdins, pt_beout->NODE);
	    else
              reslevel = vhx_gexlevel (pt_lkdins, pt_beout->NODE);

	    if (reslevel == 255)
              pt_wrssig->SIMSIG->PRJVAL = 255;
            else
              pt_wrssig->SIMSIG->PRJVAL = reslevel + 1;
	     		
            VHX_SIGUPD = addchain (VHX_SIGUPD, pt_wrssig->SIMSIG);
            pt_wrssig  = pt_wrssig->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Compute projected level of internal simple signals	*/
	/* ###------------------------------------------------------### */

          pt_wriaux = pt_lkdins->WRIAUX;
          while (pt_wriaux != NULL)
            {
            pt_beaux = pt_wriaux->BEAUX;

            if (pt_wriaux->FLAG != 'D')
              {
              if (flag == 1)
                reslevel = vhx_bddlevel (pt_lkdins, pt_beaux->NODE);
              else
                reslevel = vhx_gexlevel (pt_lkdins, pt_beaux->NODE);

       	      if (reslevel == 255)
                pt_wriaux->PRJVAL = 255;
              else
                pt_wriaux->PRJVAL = reslevel + 1;

              VHX_AUXUPD = addchain (VHX_AUXUPD, pt_wriaux);
              }
            pt_wriaux = pt_wriaux->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Computes projected level of bussed output ports		*/
	/* ###------------------------------------------------------### */

          pt_wrbsig = pt_lkdins->WRBSIG;
          while (pt_wrbsig != NULL)
            {
            pt_bebus  = pt_wrbsig->BEBUS;
            pt_prjbvl = pt_wrbsig->PRJBVL;
            pt_binode = pt_bebus->BINODE;
            while (pt_binode != NULL)
              {
              if (flag == 1)
                {
                pt_prjbvl->CNDVAL = vhx_bddlevel (pt_lkdins, pt_binode->CNDNODE);
                pt_prjbvl->DRVVAL = vhx_bddlevel (pt_lkdins, pt_binode->VALNODE);
                }
              else
                {
                pt_prjbvl->CNDVAL = vhx_gexlevel (pt_lkdins, pt_binode->CNDNODE); 
                pt_prjbvl->DRVVAL = vhx_gexlevel (pt_lkdins, pt_binode->VALNODE);
                }

              pt_prjbvl = pt_prjbvl->NEXT;
              pt_binode = pt_binode->NEXT;
              }
            VHX_BUSUPD = addchain (VHX_BUSUPD, pt_wrbsig->BUSSIG);
            pt_wrbsig  = pt_wrbsig->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Compute projected value of WRIBUXs			*/
	/* ###------------------------------------------------------### */

          pt_wribux = pt_lkdins->WRIBUX;
          while (pt_wribux != NULL)
            {
            pt_bebux  = pt_wribux->BEBUX ;
            pt_prjbvl = pt_wribux->PRJBVL;
            pt_binode = pt_bebux->BINODE ;
            while (pt_binode != NULL)
              {
              if (flag == 1)
                {
                pt_prjbvl->CNDVAL = vhx_bddlevel (pt_lkdins, pt_binode->CNDNODE);
                pt_prjbvl->DRVVAL = vhx_bddlevel (pt_lkdins, pt_binode->VALNODE);
                }
              else
                {
                pt_prjbvl->CNDVAL = vhx_gexlevel (pt_lkdins, pt_binode->CNDNODE);
                pt_prjbvl->DRVVAL = vhx_gexlevel (pt_lkdins, pt_binode->VALNODE);
                }

              pt_prjbvl = pt_prjbvl->NEXT;
              pt_binode = pt_binode->NEXT;
              }
            VHX_BUXUPD = addchain (VHX_BUXUPD, pt_wribux);
            pt_wribux  = pt_wribux->NEXT;
            }

	/* ###------------------------------------------------------### */
	/*    Compute assert instruction's level			*/
	/* ###------------------------------------------------------### */

          pt_wrimsg = pt_lkdins->WRIMSG;
          while (pt_wrimsg != NULL)
            {
            pt_bemsg = pt_wrimsg->BEMSG;

            if (flag == 1)
              reslevel = vhx_bddlevel (pt_lkdins, pt_bemsg->NODE);
            else
              reslevel = vhx_gexlevel (pt_lkdins, pt_bemsg->NODE);

	    if (reslevel == 255)
              pt_wrimsg->LEVEL = 255;
	    else
              pt_wrimsg->LEVEL = reslevel + 1;

            pt_wrimsg = pt_wrimsg->NEXT;
            }
          }
        pt_lkdins->EVAL = 0;
        break;

	/* ###------------------------------------------------------### */
	/*    for each simple output port that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'S' :
        pt_wrssig = (struct wrssig *) exec_list->DATA;

        pt_lkdins = pt_wrssig->LKDINS;
        pt_beout  = pt_wrssig->BEOUT ;

        if (flag == 1)
          reslevel = vhx_bddlevel (pt_lkdins, pt_beout->NODE);
        else
          reslevel = vhx_gexlevel (pt_lkdins, pt_beout->NODE);

	if (reslevel == 255)
          pt_wrssig->SIMSIG->PRJVAL = 255;
        else
	  pt_wrssig->SIMSIG->PRJVAL = reslevel + 1 ;

        pt_wrssig->EVAL = 0;
        VHX_SIGUPD      = addchain (VHX_SIGUPD, pt_wrssig->SIMSIG);

        break;

	/* ###------------------------------------------------------### */
	/*    for each bussed output port that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'B' :
        pt_wrbsig = (struct wrbsig *) exec_list->DATA;

        pt_lkdins = pt_wrbsig->LKDINS;
        pt_bebus  = pt_wrbsig->BEBUS ;
        pt_prjbvl = pt_wrbsig->PRJBVL;
        pt_binode = pt_bebus->BINODE ;

        while (pt_binode != NULL)
          {
          if (flag == 1)
            {
            pt_prjbvl->CNDVAL = vhx_bddlevel (pt_lkdins, pt_binode->CNDNODE);
            pt_prjbvl->DRVVAL = vhx_bddlevel (pt_lkdins, pt_binode->VALNODE);
            }
          else
            {
            pt_prjbvl->CNDVAL = vhx_gexlevel (pt_lkdins, pt_binode->CNDNODE);
            pt_prjbvl->DRVVAL = vhx_gexlevel (pt_lkdins, pt_binode->VALNODE);
            }

          pt_prjbvl = pt_prjbvl->NEXT;
          pt_binode = pt_binode->NEXT;
          }

        pt_wrbsig->EVAL = 0;
        VHX_BUSUPD      = addchain (VHX_BUSUPD, pt_wrbsig->BUSSIG);

        break;

	/* ###------------------------------------------------------### */
	/*    for each bussed internal signal that is to be executed ...*/
	/* ###------------------------------------------------------### */

      case 'U' :
        pt_wribux = (struct wribux *) exec_list->DATA;

        pt_lkdins = pt_wribux->LKDINS;
        pt_bebux  = pt_wribux->BEBUX ;
        pt_prjbvl = pt_wribux->PRJBVL;
        pt_binode = pt_bebux->BINODE ;

        while (pt_binode != NULL)
          {
          if (flag == 1)
            {
            pt_prjbvl->CNDVAL = vhx_bddlevel (pt_lkdins, pt_binode->CNDNODE);
            pt_prjbvl->DRVVAL = vhx_bddlevel (pt_lkdins, pt_binode->VALNODE);
            }
          else
            {
            pt_prjbvl->CNDVAL = vhx_gexlevel (pt_lkdins, pt_binode->CNDNODE);
            pt_prjbvl->DRVVAL = vhx_gexlevel (pt_lkdins, pt_binode->VALNODE);
            }

          pt_prjbvl = pt_prjbvl->NEXT;
          pt_binode = pt_binode->NEXT;
          }

        pt_wribux->EVAL = 0;
        VHX_BUXUPD      = addchain (VHX_BUXUPD, pt_wribux);

        break;

	/* ###------------------------------------------------------### */
	/*    compute the level of simple internal signals		*/
	/* ###------------------------------------------------------### */

      case 'X' :
        pt_wriaux = (struct wriaux *) exec_list->DATA;

        pt_lkdins = pt_wriaux->LKDINS;
        pt_beaux  = pt_wriaux->BEAUX ;

        if (flag == 1)
          reslevel = vhx_bddlevel (pt_lkdins, pt_beaux->NODE);
        else
          reslevel = vhx_gexlevel (pt_lkdins, pt_beaux->NODE);

	if (reslevel == 255)
          pt_wriaux->PRJVAL = 255;
	else
          pt_wriaux->PRJVAL = reslevel + 1;

        pt_wriaux->EVAL = 0;
        VHX_AUXUPD      = addchain (VHX_AUXUPD, pt_wriaux);

        break;

	/* ###------------------------------------------------------### */
	/*    the level of internal delayed signals is always zero	*/
	/* ###------------------------------------------------------### */

      case 'D' :
        ((struct wriaux *) exec_list->DATA)->EVAL = 0;
        break;

	/* ###------------------------------------------------------### */
	/*    the level of registers is always zero			*/
	/* ###------------------------------------------------------### */

      case 'R' :
        ((struct wrireg *) exec_list->DATA)->EVAL = 0;
        break;

	/* ###------------------------------------------------------### */
	/*    for each simple output port that is to be executed ...	*/
	/* ###------------------------------------------------------### */

      case 'M' :
        pt_wrimsg = (struct wrimsg *) exec_list->DATA;

        pt_lkdins = pt_wrimsg->LKDINS;
        pt_bemsg  = pt_wrimsg->BEMSG ;

        if (flag == 1)
          reslevel = vhx_bddlevel (pt_lkdins, pt_bemsg->NODE);
	else
          reslevel = vhx_gexlevel (pt_lkdins, pt_bemsg->NODE);
	  
	if (reslevel == 255)
          pt_wrimsg->LEVEL = 255;
        else
	  pt_wrimsg->LEVEL = reslevel + 1;

        pt_wrimsg->EVAL = 0;

        break;
      }

    exec_list = exec_list ->NEXT;
    }

  if (ret_flg == 0)
    {
    freeptype (VHX_EXECUTE [level]);
    VHX_EXECUTE [level] = NULL;
    }

  return (ret_flg);
  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_uplevel						*/
/* description	: updates the current level of signals			*/
/* called func.	: vhx_addexec, freechain				*/
/* ###--------------------------------------------------------------### */

void vhx_uplevel ()

  {
  struct chain  *cur_chain;
  struct wriaux *pt_wriaux;
  struct wribux *pt_wribux;
  struct bussig *pt_bussig;
  struct simsig *pt_simsig;
  struct prjbvl *pt_prjbvl;
  unsigned char  level    ;

	/* ###------------------------------------------------------### */
	/*    Update simple signals' level				*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_SIGUPD;
  while (cur_chain != NULL)
    {
    pt_simsig = (struct simsig *) cur_chain->DATA;
    if (pt_simsig->PRJVAL != *(pt_simsig->CURVAL))
      {
      *(pt_simsig->CURVAL) = pt_simsig->PRJVAL;
      vhx_addexec (pt_simsig->DEPEND);
      }
    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_SIGUPD);
  VHX_SIGUPD = NULL;

	/* ###------------------------------------------------------### */
	/*    Update bused signals's level				*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_BUSUPD;
  while (cur_chain != NULL)
    {
    pt_bussig = (struct bussig *) cur_chain->DATA;
    pt_prjbvl = pt_bussig->PRJBVL;
    while (pt_prjbvl != NULL)
      {
      if (pt_prjbvl->CNDVAL > level)
        level = pt_prjbvl->CNDVAL;

      if (pt_prjbvl->DRVVAL > level)
        level = pt_prjbvl->DRVVAL;

      pt_prjbvl = pt_prjbvl->GLNEXT;
      }

    if (level != 255)
      level ++;

    if (*(pt_bussig->CURVAL) != level)
      {
      *(pt_bussig->CURVAL) = level;
      vhx_addexec (pt_bussig->DEPEND);
      }
    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_BUSUPD);
  VHX_BUSUPD = NULL;

	/* ###------------------------------------------------------### */
	/*    Update internal bussed signals' level			*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_BUXUPD;
  while (cur_chain != NULL)
    {
    pt_wribux = (struct wribux *) cur_chain->DATA;
    pt_prjbvl = pt_wribux->PRJBVL;
    while (pt_prjbvl != NULL)
      {
      if (pt_prjbvl->CNDVAL > level)
        level = pt_prjbvl->CNDVAL;

      if (pt_prjbvl->DRVVAL > level)
        level = pt_prjbvl->DRVVAL;

      pt_prjbvl = pt_prjbvl->GLNEXT;
      }

    if (level != 255)
      level ++;

    if (*(pt_wribux->CURVAL) != level)
      {
      *(pt_wribux->CURVAL) = level;
      vhx_addexec (pt_wribux->DEPEND);
      }
    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_BUXUPD);
  VHX_BUXUPD = NULL;

	/* ###------------------------------------------------------### */
	/*    Updating simple internal signals' level			*/
	/* ###------------------------------------------------------### */

  cur_chain = VHX_AUXUPD;
  while (cur_chain != NULL)
    {
    pt_wriaux = (struct wriaux *)cur_chain->DATA;
    if (pt_wriaux->PRJVAL != *(pt_wriaux->CURVAL))
      {
      *(pt_wriaux->CURVAL) = pt_wriaux->PRJVAL;
      vhx_addexec (pt_wriaux->DEPEND);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (VHX_AUXUPD);
  VHX_AUXUPD = NULL;

  }

/* ###--------------------------------------------------------------### */
/* function	: vhx_levelize						*/
/* description	: levelize the description				*/
/* called func.	: vhx_execlevel, vhx_execall, vhx_uplevel, vhx_getlevel	*/
/* ###--------------------------------------------------------------### */

void vhx_levelize (pt_lkdfig, flag)

struct lkdfig *pt_lkdfig;
char           flag     ;	/* execute on bdd (1) or gex (0)	*/

  {
  vhx_execall (pt_lkdfig);
  
  while (vhx_execlevel (flag) == 0)
    vhx_uplevel ();

  vhx_getlevel (pt_lkdfig);
  }
