/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/* ###--------------------------------------------------------------### */
/* file		: dlx_defs.h						*/
/* date		: Nov 14 1996						*/
/* version	: v0.5							*/
/* author	: Pirouz BAZARGAN SABET					*/
/* description	: DLX assembler - defines (operation codes)		*/
/* ###--------------------------------------------------------------### */

#define DLX_PLSDFN	0		/* add				*/
#define DLX_MNSDFN	1		/* subtract			*/
#define DLX_SHLDFN	2		/* shift left			*/
#define DLX_SHRDFN	3		/* shift right			*/
#define DLX_ORDFN	4		/* logical or			*/
#define DLX_ANDDFN	5		/* logical and			*/
#define DLX_XORDFN	6		/* logical exclusive or		*/
#define DLX_MULDFN	7		/* multiply			*/
#define DLX_DIVDFN	8		/* divide			*/

#define DLX_MIPSDFN	2		/* Mips         architecture	*/
#define DLX_DLXPDFN	1		/* Dlx pipeline implementation	*/
#define DLX_DLXMDFN	0		/* Dlx micro-programmed implem.	*/

	/* ### ---------------------------------------------------- ### */
	/* instruction set table:					*/
	/*    Opcods in   lower case         are MIPS instructions	*/
	/*    Opcods with upper case initial are DLX  specific		*/
	/*    Opcods in   upper case         are application specific	*/
	/*								*/
	/*								*/
	/* primary opcod (31 downto 26):				*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 |speci|bcond|  j  | jal | beq | bne |blez |bgtz |	*/
	/*    1 |addi |addui|slti |sltui|andi | ori |xori | lui |	*/
	/*    2 |cop0 |  +  |  +  |  +  |     |Trap |     |     |	*/
	/*    3 |     |     |     |     |     |     |     |     |	*/
	/*    4 | lb  | lh  |  +  | lw  | lbu | lhu |  +  |SWAP |	*/
	/*    5 | sb  | sh  |  +  | sw  |     |     |  +  |     |	*/
	/*    6 |  +  |  +  |  +  |  +  |Seqi |Snei |Slei |Sgti |	*/
	/*    7 |  +  |  +  |  +  |  +  |Sgei |Sgeui|Sleui|Sgtui|	*/
	/*								*/
	/*								*/
	/* special opcod extension (5 downto 0):			*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 | sll |     | srl | sra |sllv |     |srlv |srav |	*/
	/*    1 | jr  |jalr |     |     |sysca|break|     |SLEEP|	*/
	/*    2 |  +  |  +  |  +  |  +  |     |     |     |     |	*/
	/*    3 |  +  |  +  |  +  |  +  |     |     |     |     |	*/
	/*    4 | add |addu | sub |subu | and | or  | xor | nor |	*/
	/*    5 |     |     | slt |sltu |     |     |     |     |	*/
	/*    6 |     |     |     |     | Seq | Sne | Sle | Sgt |	*/
	/*    7 |     |     |     |     | Sge |Sgeu |Sleu |Sgtu |	*/
	/*								*/
	/*								*/
	/* bcond opcod extension (20 downto 16):			*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 |bltz |bgez |     |     |     |     |     |     |	*/
	/*    1 |     |     |     |     |     |     |     |     |	*/
	/*    2 |bltza|bgeza|     |     |     |     |     |     |	*/
	/*    3 |     |     |     |     |     |     |     |     |	*/
	/*								*/
	/*								*/
	/* cop0 opcod extension (22, 21, 16, 25, 24, 23):		*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 | mf  | mt  |  +  |  +  | c0  | c0  | c0  | c0  |	*/
	/*    1 | mf  | mt  |  +  |  +  | c0  | c0  | c0  | c0  |	*/
	/*    2 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*    3 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*    4 |  +  |  +  |     |     | c0  | c0  | c0  | c0  |	*/
	/*    5 |  +  |  +  |     |     | c0  | c0  | c0  | c0  |	*/
	/*    6 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*    7 |     |     |     |     | c0  | c0  | c0  | c0  |	*/
	/*								*/
	/*								*/
	/* c0 cop0 extension extension (4 downto 0):			*/
	/*      |  0     1     2     3     4     5     6     7		*/
	/*    --+-----+-----+-----+-----+-----+-----+-----+-----+	*/
	/*    0 |  +  |  +  |  +  |     |     |     |  +  |     |	*/
	/*    1 |  +  |     |     |     |     |     |     |     |	*/
	/*    2 |rstfe|     |     |     |     |     |     |     |	*/
	/*    3 |Retfe|     |     |     |     |     |     |     |	*/
	/*								*/
	/* ### ---------------------------------------------------- ### */

#define _ADD		0x00000020	/* Mips Dlxp Dlxm architectures	*/
#define _ADDI		0x20000000	/* Mips Dlxp Dlxm architectures	*/
#define _ADDU		0x00000021	/* Mips Dlxp      architectures	*/
#define _ADDUI		0x24000000	/* Mips Dlxp      architectures	*/
#define _AND		0x00000024	/* Mips Dlxp Dlxm architectures	*/
#define _ANDI		0x30000000	/* Mips Dlxp Dlxm architectures	*/
#define _BEQ		0x10000000	/* Mips           architecture	*/
#define _BEQZ		0x10000000	/*      Dlxp Dlxm architectures	*/
#define _BGEZ		0x04010000	/* Mips           architecture	*/
#define _BGEZAL		0x04110000	/* Mips           architecture	*/
#define _BGTZ		0x1C000000	/* Mips           architecture	*/
#define _BLEZ		0x18000000	/* Mips           architecture	*/
#define _BLTZ		0x04000000	/* Mips           architecture	*/
#define _BLTZAL		0x04100000	/* Mips           architecture	*/
#define _BNE		0x14000000	/* Mips           architecture	*/
#define _BNEZ		0x14000000	/*      Dlxp Dlxm architectures	*/
#define _BREAK		0x0000000D	/* Mips           architecture	*/
#define _J		0x08000000	/* Mips Dlxp Dlxm architectures	*/
#define _JAL		0x0C000000	/* Mips Dlxp Dlxm architectures	*/
#define _JALR		0x00000009	/* Mips           architecture	*/
#define _JALR31		0x0000f809	/*      Dlxp Dlxm architectures	*/
#define _JR		0x00000008	/* Mips Dlxp Dlxm architectures	*/
#define _LB		0x80000000	/* Mips Dlxp      architectures	*/
#define _LBU		0x90000000	/* Mips Dlxp Dlxm architectures	*/
#define _LH		0x84000000	/* Mips Dlxp      architectures	*/
#define _LHU		0x94000000	/* Mips Dlxp      architectures	*/
#define _LUI		0x3C000000	/* Mips Dlxp Dlxm architectures	*/
#define _LW		0x8C000000	/* Mips Dlxp Dlxm architectures	*/
#define _MOVI2S		0x40800000	/* Mips Dlxp Dlxm architectures	*/
#define _MOVS2I		0x40000000	/* Mips Dlxp Dlxm architectures	*/
#define _NOR		0x00000027	/* Mips           architecture	*/
#define _OR		0x00000025	/* Mips Dlxp Dlxm architectures	*/
#define _ORI		0x34000000	/* Mips Dlxp Dlxm architectures	*/
#define _RETFE		0x42000018	/*      Dlxp Dlxm architectures	*/
#define _RSTFE		0x42000010	/* Mips           architecture	*/
#define _SB		0xA0000000	/* Mips Dlxp Dlxm architectures	*/
#define _SEQ		0x00000034	/*      Dlxp Dlxm architectures	*/
#define _SEQI		0xD0000000	/*      Dlxp Dlxm architectures	*/
#define _SNE		0x00000035	/*      Dlxp Dlxm architectures	*/
#define _SNEI		0xD4000000	/*      Dlxp Dlxm architectures	*/
#define _SGE		0x0000003C	/*      Dlxp Dlxm architectures	*/
#define _SGEI		0xF0000000	/*      Dlxp Dlxm architectures	*/
#define _SGEU		0x0000003D	/*      Dlxp      architecture	*/
#define _SGEUI		0xF4000000	/*      Dlxp      architecture	*/
#define _SGT		0x00000037	/*      Dlxp Dlxm architectures	*/
#define _SGTI		0xDC000000	/*      Dlxp Dlxm architectures	*/
#define _SGTU		0x0000003F	/*      Dlxp      architecture	*/
#define _SGTUI		0xF3000000	/*      Dlxp      architecture	*/
#define _SH		0xA4000000	/* Mips Dlxp      architectures	*/
#define _SLE		0x00000036	/*      Dlxp Dlxm architectures	*/
#define _SLEEP		0x0000000F	/* Mips Dlxp      architectures	*/
#define _SLEI		0xD8000000	/*      Dlxp Dlxm architectures	*/
#define _SLEU		0x0000003E	/*      Dlxp      architecture	*/
#define _SLEUI		0xF8000000	/*      Dlxp      architecture	*/
#define _SLL		0x00000000	/* Mips Dlxp Dlxm architectures	*/
#define _SLLV		0x00000004	/* Mips Dlxp Dlxm architectures	*/
#define _SLT		0x0000002A	/* Mips Dlxp Dlxm architectures	*/
#define _SLTI		0x28000000	/* Mips Dlxp Dlxm architectures	*/
#define _SLTU		0x0000002B	/* Mips Dlxp      architectures	*/
#define _SLTUI		0x2C000000	/* Mips Dlxp      architectures	*/
#define _SRA		0x00000003	/* Mips Dlxp Dlxm architectures	*/
#define _SRAV		0x00000007	/* Mips Dlxp Dlxm architectures	*/
#define _SRL		0x00000002	/* Mips Dlxp Dlxm architectures	*/
#define _SRLV		0x00000006	/* Mips Dlxp Dlxm architectures	*/
#define _SUB		0x00000022	/* Mips Dlxp Dlxm architectures	*/
#define _SUBI		0x60000000	/*      Dlxp Dlxm architectures	*/
#define _SUBU		0x00000023	/* Mips Dlxp      architectures	*/
#define _SUBUI		0x64000000	/*      Dlxp      architecture	*/
#define _SW		0xAC000000	/* Mips Dlxp Dlxm architectures	*/
#define _SWAP		0x9C000000	/* Mips Dlxp      architectures	*/
#define _SYSCALL	0x0000000C	/* Mips           architecture	*/
#define _TRAP		0x54000000	/*      Dlxp Dlxm architectures	*/
#define _XOR		0x00000026	/* Mips Dlxp Dlxm architectures	*/
#define _XORI		0x38000000	/* Mips Dlxp Dlxm architectures	*/

#define _NOP		0x00000001	/* Mips Dlxp Dlxm macro		*/
#define _LOADI		0x00000002	/* Mips Dlxp Dlxm macro		*/

#define _IAR		0x00000001
#define _TNR		0x00000002
#define _SR		0x00000003
#define _SVS		0x00000004
