/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_makvarlist.c					*/
/* date		: Mar 20 1995						*/
/* version	: v109							*/
/* authors	: DUNOYER Julien.					*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include CST_H
#include BEH_H
#include BHL_H

/* ###--------------------------------------------------------------### */
/* function	: beh_abl2varlist					*/
/* description	: create recursively variable list from ABL.		*/
/* called func.	: searchInputCct, beh_abl2varlist , beh_error, 		*/
/* 		: cst_AddSetElement, cst_CreateUnionSet ,		*/
/* ###--------------------------------------------------------------### */

static chain_list *beh_abl2varlist (ptr_befig, expr)

struct befig *ptr_befig;	/* pointer on  behavioural figure	*/
chain_list   *expr     ;	/* pointer abl expression		*/

  {
  short         oper  ;
  chain_list   *set_pnt  = NULL;
  pCircuit      ptr_circuit = ptr_befig->CIRCUI;
  int           err_flg     = 0;
  int           index;
  char         *atom;
  char          name[256];

	/* ###------------------------------------------------------### */
	/*    check that the expression exists				*/
	/* ###------------------------------------------------------### */

  if (expr == NULL)
    err_flg = beh_error (2, NULL);
  else
    {
    if (ATOM(expr))
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is a terminal (atom)			*/
	/* ###------------------------------------------------------### */

      atom = (char *) VALUE_ATOM (expr);

	/* ###------------------------------------------------------### */
	/*    if the terminal is a primary signal create an element	*/
	/* with its index and return					*/
	/* ###------------------------------------------------------### */

      if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
        set_pnt = cst_AddSetElement (set_pnt, index);
      }
    else
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is not a terminal (is a complex expr.)	*/
	/* ###------------------------------------------------------### */

      if ((oper = OPER(expr)) != STABLE)
       {
       while ((expr = CDR(expr)) != NULL)
         {
         set_pnt =cst_CreateUnionSet
                      (set_pnt, beh_abl2varlist (ptr_befig, CAR(expr)));
         }
       }
      else
       {
	expr = CDR(expr);
	if (!ATOM(expr))
	  err_flg = beh_error (4, NULL);
        else
          {
	  atom = (char *)VALUE_ATOM (CAR(expr));

          if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
            {
            set_pnt = cst_AddSetElement (set_pnt, index);

	    sprintf (name, "%s'delayed", atom);
	    atom = namealloc (name);

            if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
              {
              set_pnt = cst_AddSetElement (set_pnt, index);
              }
            else
	      err_flg = beh_error (3, atom);
            }
          else
	    err_flg = beh_error (3, atom);
          }
        }
      }
    }

  if (err_flg != 0)
    ptr_befig->ERRFLG = 1;

  return (set_pnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_mvlsttot						*/
/* description	: make list of variables for all primary signals	*/
/*		  simple and bussed outputs,				*/
/*		  simple and bussed internal signals, registers.	*/
/* called func.	: beh_abl2varlist					*/
/* ###--------------------------------------------------------------### */

static void beh_mvlsttot (ptr_befig)

struct befig *ptr_befig;		/* pointer on current BEFIG	*/

  {
  struct beaux  *ptr_beaux;
  struct beaux  *ptr_bedly;
  struct bemsg  *ptr_bemsg;
  struct beout  *ptr_beout;
  struct bebus  *ptr_bebus;
  struct bebux  *ptr_bebux;
  struct bereg  *ptr_bereg;
  struct biabl  *ptr_biabl;
  struct binode *ptr_binode;

	/* ###------------------------------------------------------### */
	/*    make a var list for each simple internal signal		*/
	/* ###------------------------------------------------------### */

  ptr_beaux = ptr_befig->BEAUX;

  while (ptr_beaux != NULL)
    {
    if (ptr_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(ptr_beaux->NODE))->VARLIST = 
                        beh_abl2varlist (ptr_befig, ptr_beaux->ABL);
      }
    else
      {
      ptr_beaux->NODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_beaux->ABL);
      }
    ptr_beaux = ptr_beaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each delayed internal signal		*/
	/* ###------------------------------------------------------### */

  ptr_bedly = ptr_befig->BEDLY;
  while (ptr_bedly != NULL)
    {
    if (ptr_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(ptr_bedly->NODE))->VARLIST = 
                        beh_abl2varlist (ptr_befig, ptr_bedly->ABL);
      }
    else
      {
      ptr_bedly->NODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_bedly->ABL);
      }
    ptr_bedly = ptr_bedly->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each assertion			*/
	/* ###------------------------------------------------------### */

  ptr_bemsg = ptr_befig->BEMSG;
  while (ptr_bemsg != NULL)
    {
    if (ptr_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(ptr_bemsg->NODE))->VARLIST = 
                        beh_abl2varlist (ptr_befig, ptr_bemsg->ABL);
      }
    else
      {
      ptr_bemsg->NODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_bemsg->ABL);
      }

    ptr_bemsg = ptr_bemsg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each simple output port		*/
	/* ###------------------------------------------------------### */

  ptr_beout = ptr_befig->BEOUT;
  while (ptr_beout != NULL)
    {
    if (ptr_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(ptr_beout->NODE))->VARLIST = 
                        beh_abl2varlist (ptr_befig, ptr_beout->ABL);
      }
    else
      {
      ptr_beout->NODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_beout->ABL);
      }

    ptr_beout = ptr_beout->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each bussed output port		*/
	/* ###------------------------------------------------------### */

  ptr_bebus = ptr_befig->BEBUS;
  while (ptr_bebus != NULL)
    {
    ptr_biabl  = ptr_bebus->BIABL;
    ptr_binode = ptr_bebus->BINODE;

    while (ptr_biabl != NULL)
      {
      if (ptr_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(ptr_binode->CNDNODE))->VARLIST = 
                              beh_abl2varlist (ptr_befig, ptr_biabl->CNDABL);
        ((bequad_list *)(ptr_binode->VALNODE))->VARLIST = 
                              beh_abl2varlist (ptr_befig, ptr_biabl->VALABL);
        }
      else
        {
        ptr_binode->CNDNODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_biabl->CNDABL);
        ptr_binode->VALNODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_biabl->VALABL);
        }

      ptr_biabl = ptr_biabl->NEXT;
      ptr_binode = ptr_binode->NEXT;
      }

    ptr_bebus = ptr_bebus->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each bussed internal signal		*/
	/* ###------------------------------------------------------### */

  ptr_bebux = ptr_befig->BEBUX;
  while (ptr_bebux != NULL)
    {
    ptr_biabl  = ptr_bebux->BIABL;
    ptr_binode = ptr_bebux->BINODE;

    while (ptr_biabl != NULL)
      {
      if (ptr_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(ptr_binode->CNDNODE))->VARLIST = 
                              beh_abl2varlist (ptr_befig, ptr_biabl->CNDABL);
        ((bequad_list *)(ptr_binode->VALNODE))->VARLIST = 
                              beh_abl2varlist (ptr_befig, ptr_biabl->VALABL);
        }
      else
        {
        ptr_binode->CNDNODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_biabl->CNDABL);
        ptr_binode->VALNODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_biabl->VALABL);
        }

      ptr_biabl  = ptr_biabl->NEXT;
      ptr_binode = ptr_binode->NEXT;
      }
    ptr_bebux = ptr_bebux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each internal register		*/
	/* ###------------------------------------------------------### */

  ptr_bereg = ptr_befig->BEREG;
  while (ptr_bereg != NULL)
    {
    ptr_biabl  = ptr_bereg->BIABL;
    ptr_binode = ptr_bereg->BINODE;

    while (ptr_biabl != NULL)
      {
      if (ptr_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(ptr_binode->CNDNODE))->VARLIST = 
                              beh_abl2varlist (ptr_befig, ptr_biabl->CNDABL);
        ((bequad_list *)(ptr_binode->VALNODE))->VARLIST = 
                              beh_abl2varlist (ptr_befig, ptr_biabl->VALABL);
        }
      else
        {
        ptr_binode->CNDNODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_biabl->CNDABL);
        ptr_binode->VALNODE = (struct node *)beh_abl2varlist (ptr_befig, ptr_biabl->VALABL);
        }

      ptr_biabl  = ptr_biabl->NEXT;
      ptr_binode = ptr_binode->NEXT;
      }
    ptr_bereg = ptr_bereg->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_makvarlist					*/
/* description	: make var list for each signal of a befig structure 	*/
/* called func.	: initializeCct, addInputCct, beh_mvlsttot		*/
/* ###--------------------------------------------------------------### */

void beh_makvarlist (ptr_befig)

struct befig *ptr_befig;

  {
  berin_list *ptr_rin;
  beout_list *ptr_out;
  bebus_list *ptr_bus;
  int         count_rin = 0;
  int         count_out = 0;
  int         err_flg   = 0;

  if (!(ptr_befig->TYPE & BEH_NODE_VARLIST));
    {
    if (ptr_befig->CIRCUI == NULL)
      {
        /* ###------------------------------------------------------### */
        /*    count inputs 						*/
        /* ###------------------------------------------------------### */

      ptr_rin = ptr_befig->BERIN;
      while (ptr_rin != NULL)
        {
        ptr_rin = ptr_rin->NEXT;
        count_rin++;
        }

        /* ###------------------------------------------------------### */
        /*    count outputs 						*/
        /* ###------------------------------------------------------### */

      ptr_out = ptr_befig->BEOUT;
      while (ptr_out != NULL)
        {
        ptr_out = ptr_out->NEXT;
        count_out++;
        }

      ptr_bus = ptr_befig->BEBUS;
      while (ptr_bus != NULL)
         {
         ptr_bus = ptr_bus->NEXT;
         count_out++;
         }

        /* ###------------------------------------------------------### */
        /*    initialization						*/
        /* ###------------------------------------------------------### */

      ptr_befig->CIRCUI = initializeCct (ptr_befig->NAME, count_rin*2, count_out);

        /* ###------------------------------------------------------### */
        /*    Define an index for each primary signal			*/
        /* ###------------------------------------------------------### */

      ptr_rin = ptr_befig->BERIN;
      while (ptr_rin != NULL)
        {
        addInputCct (ptr_befig->CIRCUI, ptr_rin->NAME);
        ptr_rin = ptr_rin->NEXT;
        }
      }
        /* ###------------------------------------------------------### */
        /*    Change ABLs to VARLISTS					*/
        /* ###------------------------------------------------------### */

    beh_mvlsttot (ptr_befig);

    ptr_befig->TYPE |= BEH_NODE_VARLIST;

    if (err_flg != 0)
      ptr_befig->ERRFLG = 1;
    }

  }
