/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/* 10/26/95 Cone Netlist Structure functions: cone.c                          */

#include"cnsall.h"

static void				*addconepage();
static cone_list		*getfreecone();

/* define number of cones per page                                            */
#define CONESPERPAGE 100

static chain_list *CONEPAGES;		/* chain list of cone pages                  */
static int PAGECONEINDEX = 0;		/* cone index in current page                */
static cone_list *FREECONELIST;	/* list of free cones in current page        */
static cone_list *FREEDCONELIST;	/* list of freed cones in current page       */

/*============================================================================*
 | function initconemem();                                                    |
 | initialize cone memory structure                                           |
 *============================================================================*/
void				*initconemem()

{
cone_list		*ptcone;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns initconemem\n");

ptcone = (cone_list *)mbkalloc(CONESPERPAGE * (sizeof(cone_list)));
CONEPAGES = addchain((chain_list *)NULL, (void *)ptcone);
FREEDCONELIST = NULL;
PAGECONEINDEX = 1;
FREECONELIST = ptcone;
}

/*============================================================================*
 | function addconepage();                                                    |
 | add a new cone page                                                        |
 *============================================================================*/
static void		*addconepage()

{
cone_list		*ptcone;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns addconepage\n");

ptcone = (cone_list *)mbkalloc(CONESPERPAGE * (sizeof(cone_list)));
CONEPAGES = addchain(CONEPAGES, (void *)ptcone);
PAGECONEINDEX = 1;
FREECONELIST = ptcone;
}

/*============================================================================*
 | function getfreecone();                                                    |
 | get a new cone                                                             |
 *============================================================================*/
static cone_list	*getfreecone()

{
cone_list			*ptcone;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns getfreecone in\n");

if (FREEDCONELIST != NULL) {
	ptcone = FREEDCONELIST;
	FREEDCONELIST = FREEDCONELIST->NEXT;
	ptcone->NEXT = NULL;
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns getfreecone out\n");
	return(ptcone);
	}
else if (PAGECONEINDEX < CONESPERPAGE) {
	PAGECONEINDEX++;
	FREECONELIST++;
	FREECONELIST->NEXT = NULL;
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns getfreecone out\n");
	return(FREECONELIST);
	}
else {
	addconepage();
	FREECONELIST->NEXT = NULL;
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns getfreecone out\n");
	return(FREECONELIST);
	}
}

/*============================================================================*
 | function freecone();                                                       |
 | free a cone                                                                |
 *============================================================================*/
void				*freecone(ptcone)

cone_list		*ptcone;

{
edge_list		*pteelist;
branch_list		*ptbrlist;
chain_list		*ptchain;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns freecone in\n");

if (ptcone == NULL) {
	(void)fflush(stdout);
	(void)fprintf(stderr, "*** cns error ***");
	(void)fprintf(stderr, " freecone() impossible: NULL pointer !\n");
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns freecone out\n");
	exit(-1);
	}

if ((pteelist = ptcone->INCONE) != NULL) freeeelist(pteelist);
if ((pteelist = ptcone->OUTCONE) != NULL) freeeelist(pteelist);
if ((ptbrlist = ptcone->BREXT) != NULL) freebrlist(ptbrlist);
if ((ptbrlist = ptcone->BRVDD) != NULL) freebrlist(ptbrlist);
if ((ptbrlist = ptcone->BRGND) != NULL) freebrlist(ptbrlist);
if ((ptbrlist = ptcone->BRVSS) != NULL) freebrlist(ptbrlist);
if ((ptchain = ptcone->CELLS) != NULL) freechain(ptchain);
ptcone->NEXT = FREEDCONELIST;
FREEDCONELIST = ptcone;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns freecone out\n");
}

/*============================================================================*
 | function freecelist();                                                     |
 | free a cone list                                                           |
 *============================================================================*/
void				*freecelist(ptcelist)

cone_list		*ptcelist;

{
cone_list		*ptcone;
edge_list		*pteelist;
branch_list		*ptbrlist;
chain_list		*ptchain;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns freecelist in\n");

if (ptcelist == NULL) {
	(void)fflush(stdout);
	(void)fprintf(stderr, "*** cns error ***");
	(void)fprintf(stderr, " freecelist() impossible: NULL pointer !\n");
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns freecelist out\n");
	exit(-1);
	}

for (ptcone = ptcelist; ptcone->NEXT != NULL; ptcone = ptcone->NEXT) {
	if ((pteelist = ptcone->INCONE) != NULL) freeeelist(pteelist);
	if ((pteelist = ptcone->OUTCONE) != NULL) freeeelist(pteelist);
	if ((ptbrlist = ptcone->BREXT) != NULL) freebrlist(ptbrlist);
	if ((ptbrlist = ptcone->BRVDD) != NULL) freebrlist(ptbrlist);
	if ((ptbrlist = ptcone->BRGND) != NULL) freebrlist(ptbrlist);
	if ((ptbrlist = ptcone->BRVSS) != NULL) freebrlist(ptbrlist);
	if ((ptchain = ptcone->CELLS) != NULL) freechain(ptchain);
	}
if ((pteelist = ptcone->INCONE) != NULL) freeeelist(pteelist);
if ((pteelist = ptcone->OUTCONE) != NULL) freeeelist(pteelist);
if ((ptbrlist = ptcone->BREXT) != NULL) freebrlist(ptbrlist);
if ((ptbrlist = ptcone->BRVDD) != NULL) freebrlist(ptbrlist);
if ((ptbrlist = ptcone->BRGND) != NULL) freebrlist(ptbrlist);
if ((ptbrlist = ptcone->BRVSS) != NULL) freebrlist(ptbrlist);
if ((ptchain = ptcone->CELLS) != NULL) freechain(ptchain);
ptcone->NEXT = FREEDCONELIST;
FREEDCONELIST = ptcelist;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns freecelist out\n");
}

/*============================================================================*
 | function addcone();                                                        |
 | add an element to a cone list                                              |
 *============================================================================*/
cone_list		*addcone(ptconehead, index, name, type, tectype, xmax, xmin, ymax, ymin, ptincone, ptoutcone, ptbrext, ptbrvdd, ptbrgnd, ptbrvss, ptchain)

cone_list		*ptconehead;
char				*name;
long				 index;
long				 type;
long				 tectype;
long				 xmax,xmin,ymax,ymin;
edge_list		*ptincone;
edge_list		*ptoutcone;
branch_list		*ptbrext;
branch_list		*ptbrvdd;
branch_list		*ptbrgnd;
branch_list		*ptbrvss;
chain_list		*ptchain;

{
cone_list		*ptcone;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns addcone in\n");

if (index == 0) {
	(void)fflush(stdout);
	(void)fprintf(stderr, "*** cns error ***");
	(void)fprintf(stderr, " addcone() impossible: index 0 may not be employed\n");
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns addcone out\n");
	exit(-1);
	}

if (CNS_DEBUG_LEVEL >= 2) {
	for (ptcone = ptconehead; ptcone != NULL; ptcone = ptcone->NEXT) {
		if (ptcone->INDEX == index) {
			(void)fflush(stdout);
			(void)fprintf(stderr, "*** cns error ***");
			(void)fprintf(stderr, " addcone() impossible: cone of index %ld already exists\n", ptcone->INDEX);
			if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns addcone out\n");
			exit(-1);
			}
		}
	}

ptcone = getfreecone();
ptcone->NEXT = ptconehead;
ptcone->INDEX = index;
ptcone->NAME = namealloc(name);
ptcone->TYPE = type;
ptcone->TECTYPE = tectype;
ptcone->XM = xmax;
ptcone->Xm = xmin;
ptcone->YM = ymax;
ptcone->Ym = ymin;
ptcone->INCONE = ptincone;
ptcone->OUTCONE = ptoutcone;
ptcone->BREXT = ptbrext;
ptcone->BRVDD = ptbrvdd;
ptcone->BRGND = ptbrgnd;
ptcone->BRVSS = ptbrvss;
ptcone->CELLS = ptchain;
ptcone->USER = NULL;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns addcone out\n");
return(ptcone);
}

/*============================================================================*
 | function delcone();                                                        |
 | delete an element from a cone list and free that element                   |
 *============================================================================*/
cone_list		*delcone(ptconehead, ptcone2del)

cone_list		*ptconehead;
cone_list		*ptcone2del;

{
cone_list		*ptcone;
cone_list		*ptcone2sav;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns delcone in\n");

if ((ptconehead == NULL) || (ptcone2del == NULL)) {
	(void)fflush(stdout);
	(void)fprintf(stderr, "*** cns error ***");
	(void)fprintf(stderr, " delcone() impossible: NULL pointer !\n");
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns delcone out\n");
	exit(-1);
	}

if (ptcone2del == ptconehead) {
	ptcone = ptconehead->NEXT;
	ptconehead->NEXT = NULL;
	freecone(ptconehead);
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns delcone out\n");
	return(ptcone);
	}
else {
	for (ptcone = ptconehead; ptcone; ptcone = ptcone->NEXT) {
		if (ptcone == ptcone2del) {
			break;
			}
		ptcone2sav = ptcone;
		}
	if (ptcone != NULL) {
		ptcone2sav->NEXT = ptcone->NEXT;
		ptcone2del->NEXT = NULL;
		freecone(ptcone2del);
		}
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns delcone out\n");
	return(ptconehead);
	}
}

/*============================================================================*
 | function appendcone();                                                     |
 | chain two cone lists. WARNING: no consistency check                        |
 *============================================================================*/
cone_list		*appendcone(ptcone1, ptcone2)

cone_list		*ptcone1;
cone_list		*ptcone2;

{
cone_list		*ptcone;
 
if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns appendcone\n");

if (ptcone1 == NULL) {
	return(ptcone2);
	}
else {
	for (ptcone = ptcone1; ptcone->NEXT !=  NULL; ptcone = ptcone->NEXT);
	ptcone->NEXT = ptcone2;
	return(ptcone1);
	}
}

/*============================================================================*
 | function getcone();                                                        |
 | returns a cone of a given index and, or name                               |
 *============================================================================*/
cone_list		*getcone(ptcnsfig, index, name)

cnsfig_list		*ptcnsfig;
long				 index;
char				*name;

{
cone_list		*ptcone;
short				 foundindex = 0;
short				 foundname = 0;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns getcone\n");

if ((ptcnsfig == NULL) || (index == 0) && (name == NULL)) {
	(void)fflush(stdout);
	(void)fprintf(stderr, "*** cns warning ***");
	(void)fprintf(stderr, " getcone() impossible: NULL pointer(s) !\n");
	}

if ((index != 0) && (name != NULL)) {
	for (ptcone = ptcnsfig->CONE; ptcone != NULL; ptcone = ptcone->NEXT) {
		if (ptcone->INDEX == index) foundindex = 1;
		if (strcmp(ptcone->NAME, name) == 0) foundname = 1;
		if (foundindex && foundname) return(ptcone);
		else if (foundindex) {
			(void)fflush(stdout);
			(void)fprintf(stderr, "*** cns warning ***");
			(void)fprintf(stderr, " getcone() conflict: found index but name does not match !\n");
			return(ptcone);
			}
		else if (foundname) {
			(void)fflush(stdout);
			(void)fprintf(stderr, "*** cns warning ***");
			(void)fprintf(stderr, " getcone() conflict: found name but index does not match !\n");
			return(ptcone);
			}
		}
	}
else if (index != 0) {
	for (ptcone = ptcnsfig->CONE; ptcone != NULL; ptcone = ptcone->NEXT) {
		if (ptcone->INDEX == index) return(ptcone);
		}
	}
else if (name != NULL) {
	for (ptcone = ptcnsfig->CONE; ptcone != NULL; ptcone = ptcone->NEXT) {
		if (strcmp(ptcone->NAME, name) == 0) return(ptcone);
		}
	}

return((cone_list *)NULL);
}

/*============================================================================*
 | function viewcone();                                                       |
 | display the content of a given cone structure                              |
 *============================================================================*/
void				 viewcone(ptcone, depth)

cone_list		*ptcone;
int				 depth;

{
char				*margin = (char *)mbkalloc(80);
int				 k = 1;
edge_list		*ptincone;
edge_list		*ptoutcone;
branch_list		*ptbranch;
ptype_list		*ptptype;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns viewcone in\n");

if (ptcone == NULL) {
	(void)fflush(stdout);
	(void)fprintf(stderr, "*** cns error ***");
	(void)fprintf(stderr, " viewcone() impossible: NULL pointer !\n");
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns viewcone out\n");
	exit(-1);
	}

if (depth < 0) {
	depth = CNS_CE_VIEW;
	}

margin = strcpy(margin, " ");
while (k <= CNS_VIEW_LEVEL) {
	margin = strcat(margin, "   ");
	k++;
	}

CNS_VIEW_LEVEL++;
printf("%scone->NEXT    = 0x%x\n", margin, ptcone->NEXT);
printf("%scone->INDEX   = %ld\n", margin, ptcone->INDEX);
printf("%scone->NAME    = %s\n", margin, ptcone->NAME);
printf("%scone->TYPE    = 0x%x\n", margin, ptcone->TYPE);
printf("%scone->TECTYPE = 0x%x\n", margin, ptcone->TECTYPE);
printf("%scone->XM      = 0x%x\n", margin, ptcone->XM);
printf("%scone->Xm      = 0x%x\n", margin, ptcone->Xm);
printf("%scone->YM      = 0x%x\n", margin, ptcone->YM);
printf("%scone->ym      = 0x%x\n", margin, ptcone->Ym);
printf("%scone->INCONE  = 0x%x\n", margin, ptcone->INCONE);
if ((depth >= 1) && (ptcone->INCONE != NULL)) {
	for (ptincone = ptcone->INCONE; ptincone != NULL; ptincone = ptincone->NEXT) {
		viewincone(ptincone, depth-1);
		}
	}
printf("%scone->OUTCONE = 0x%x\n", margin, ptcone->OUTCONE);
if ((depth >= 1) && (ptcone->OUTCONE != NULL)) {
	for (ptoutcone = ptcone->OUTCONE; ptoutcone != NULL; ptoutcone = ptoutcone->NEXT) {
		viewoutcone(ptoutcone, depth-1);
		}
	}
printf("%scone->BREXT   = 0x%x\n", margin, ptcone->BREXT);
if ((depth >= 1) && (ptcone->BREXT != NULL)) {
	for (ptbranch = ptcone->BRVDD; ptbranch != NULL; ptbranch = ptbranch->NEXT) {
		viewbranch(ptbranch, depth-1);
		}
	}
printf("%scone->BRVDD   = 0x%x\n", margin, ptcone->BRVDD);
if ((depth >= 1) && (ptcone->BRVDD != NULL)) {
	for (ptbranch = ptcone->BRVDD; ptbranch != NULL; ptbranch = ptbranch->NEXT) {
		viewbranch(ptbranch, depth-1);
		}
	}
printf("%scone->BRGND   = 0x%x\n", margin, ptcone->BRGND);
if ((depth >= 1) && (ptcone->BRGND != NULL)) {
	for (ptbranch = ptcone->BRGND; ptbranch != NULL; ptbranch = ptbranch->NEXT) {
		viewbranch(ptbranch, depth-1);
		}
	}
printf("%scone->BRVSS   = 0x%x\n", margin, ptcone->BRVSS);
if ((depth >= 1) && (ptcone->BRVSS != NULL)) {
	for (ptbranch = ptcone->BRVSS; ptbranch != NULL; ptbranch = ptbranch->NEXT) {
		viewbranch(ptbranch, depth-1);
		}
	}
printf("%scone->CELLS   = 0x%x\n", margin, ptcone->CELLS);
printf("%scone->USER    = 0x%x\n", margin, ptcone->USER);
if ((depth >= 1) && (ptcone->USER != NULL)) {
	for (ptptype = ptcone->USER; ptptype != NULL; ptptype = ptptype->NEXT) {
		viewcnsptype(ptptype);
		}
	}
else {
	printf("\n");
	}
CNS_VIEW_LEVEL--;
mbkfree((void *)margin);

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns viewcone out\n");
}

/*============================================================================*
 | function coneparabrs();                                                    |
 | search cone for parallel branches and eventually update it's user field    |
 | and the types of it's branches                                             |
 *============================================================================*/
void				 coneparabrs(ptcone)

cone_list		*ptcone;

{
chain_list		*ptchain0;
chain_list		*ptchain1;
chain_list		*ptchain2;
chain_list		*ptchain3;
chain_list		*ptchain;
ptype_list		*ptptype;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns coneparabrs in\n");

ptchain0 = parabrs(ptcone->BREXT);
ptchain1 = parabrs(ptcone->BRVDD);
ptchain2 = parabrs(ptcone->BRGND);
ptchain3 = parabrs(ptcone->BRVSS);

if (ptchain2 != NULL) {
	for (ptchain = ptchain2; ptchain->NEXT != NULL; ptchain = ptchain->NEXT);
	ptchain->NEXT = ptchain3;
	}

if (ptchain1 != NULL) {
	for (ptchain = ptchain1; ptchain->NEXT != NULL; ptchain = ptchain->NEXT);
	if (ptchain2 != NULL) ptchain->NEXT = ptchain2;
	else ptchain->NEXT = ptchain3;
	}

if (ptchain0 != NULL) {
	for (ptchain = ptchain0; ptchain->NEXT != NULL; ptchain = ptchain->NEXT);
	if (ptchain1 != NULL) ptchain->NEXT = ptchain1;
	else if (ptchain2 != NULL) ptchain->NEXT = ptchain2;
	else ptchain->NEXT = ptchain3;
	}
else if (ptchain1 != NULL) ptchain0 = ptchain1;
else if (ptchain2 != NULL) ptchain0 = ptchain2;
else ptchain0 = ptchain3;

if (ptchain0 != NULL) {
	if ((ptptype = getptype(ptcone->USER, (long)CNS_PARALLEL)) != NULL) {
		ptptype->DATA = (void *)ptchain0;
		}
	else {
		ptcone->USER = addptype(ptcone->USER, (long)CNS_PARALLEL, (void *)ptchain0);
		}
	}
else if (getptype(ptcone->USER, (long)CNS_PARALLEL) != NULL) {
	ptcone->USER = delptype(ptcone->USER, (long)CNS_PARALLEL);
	}

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns coneparabrs out\n");
}

/*============================================================================*
 | function cnbrlotrsnum();                                                   |
 | returns the maximum number of series transistor found in the input cone    |
 *============================================================================*/
short				 cnbrlotrsnum(ptcone)

cone_list		*ptcone;

{
short				 lotrsnum1;
short				 lotrsnum2;
short				 lotrsnum3;
short				 lotrsnum4;
short				 lotrsmax1;
short				 lotrsmax2;
short				 lotrsmax;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns cnbrlotrsnum in\n");

if (ptcone == NULL) {
	(void)fflush(stdout);
	(void)fprintf(stderr, "*** cns error ***");
	(void)fprintf(stderr, " cnbrlotrsnum() impossible: NULL pointer !\n");
	if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns cnbrlotrsnum out\n");
	exit(-1);
	}

lotrsnum1 = brlotrsnum(ptcone->BREXT);
lotrsnum2 = brlotrsnum(ptcone->BRVDD);
lotrsnum3 = brlotrsnum(ptcone->BRGND);
lotrsnum4 = brlotrsnum(ptcone->BRVSS);

if (lotrsnum1 > lotrsnum2) lotrsmax1 = lotrsnum1;
else lotrsmax1= lotrsnum2;

if (lotrsnum3 > lotrsnum4) lotrsmax2 = lotrsnum3;
else lotrsmax2= lotrsnum4;

if (lotrsmax1 > lotrsmax2) lotrsmax = lotrsmax1;
else lotrsmax= lotrsmax2;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns cnbrlotrsnum out\n");

return(lotrsmax);
}

/*============================================================================*
 | function locon2cone();                                                     |
 | add a connector to a cone's CNS_LOCON USER field if it isn't already in    |
 | the list (unicity check)                                                   |
 *============================================================================*/
void				 locon2cone(ptlocon, ptcone)

locon_list		*ptlocon;
cone_list		*ptcone;

{
ptype_list		*ptptype;
chain_list		*ptchain;
short				 found = 0;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns locon2cone\n");

if ((ptlocon != NULL) && (ptcone != NULL)) {
	if ((ptptype = getptype(ptcone->USER, (long)CNS_LOCON)) != NULL) {
		for (ptchain = (chain_list *)ptptype->DATA; ptchain != NULL; ptchain = ptchain->NEXT) {
			if ((locon_list *)ptchain->DATA == ptlocon) {
				found = 1;
				break;
				}
			}
		if (!found) {
			ptptype->DATA = addchain((chain_list *)ptptype->DATA, (void *)ptlocon);
			}
		}
	else {
		ptcone->USER = addptype(ptcone->USER, CNS_LOCON, (void *)addchain((chain_list *)NULL, (void *)ptlocon));
		}
	}
}

/*============================================================================*
 | function lotrs2cone();                                                     |
 | add a transistor to a cone's CNS_LOTRS USER field if it isn't already in   |
 | the list (unicity check)                                                   |
 *============================================================================*/
void				 lotrs2cone(ptlotrs, ptcone)
 
lotrs_list		*ptlotrs;
cone_list		*ptcone;

{
ptype_list		*ptptype;
chain_list		*ptchain;
short				 found =0;

if (CNS_TRACE_MODE >= CNS_TRACE) (void)printf("TRA_cns lotrs2cone\n");

if ((ptlotrs != NULL) && (ptcone != NULL)) {
	if ((ptptype = getptype(ptcone->USER, (long)CNS_LOTRS)) != NULL) {
		for (ptchain = (chain_list *)ptptype->DATA; ptchain != NULL; ptchain = ptchain->NEXT) {
			if ((lotrs_list *)ptchain->DATA == ptlotrs) {
				found = 1;
				break;
				}
			}
		if (!found) {
			ptptype->DATA = (void *)addchain((chain_list *)ptptype->DATA, (void *)ptlotrs);
			}
		}
	else {
		ptcone->USER = addptype(ptcone->USER, CNS_LOTRS, (void *)addchain((chain_list *)NULL, (void *)ptlotrs));
		}
	}
}

