/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :    placer_create_con.c

    Date        :   30-06-94
    Version     :   1

    Auteur      :   BEAURIN oliver

    Description :   Create connectors structure for placer

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <strings.h>

#include MUT_H
#include MLO_H
#include MPH_H
#include APR_H

#include GFP_H

#include "dpr.h"
#include "placer_structure.h"
#include "placer_utils.h"
#include "placer_decoup.h"

PlaceConList     *LeadPlaceConList;

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define myabs(x) ( ((x)<0) ? (-(x)) : (x) )

/*---------------------------------------------------------\
                     create_con_list_for_routed_bloc
\---------------------------------------------------------*/
void create_con_list_for_routed_bloc( ptcol )
column *ptcol;
{
    phfig_list *ptphfig;
    phcon_list *ptphcon;
    phref_list *ptphref;
    char       *con_name;
    slice      *ptslice;
    char       *ref_ref;

    if( ptcol->TYPE != COL_ROUTED )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (create_con_list_for_routed_bloc), i can't continue...\n" );
        exit( 1 );
    }

    for( ptslice = ptcol->SLICES;
         ( ptslice ) && ( ptslice->MODEL_NAME == NULL );
         ptslice = ptslice->NEXT);

    if( ptslice == NULL )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (create_con_list_for_routed_bloc), i can't continue...\n" );
        exit( 1 );
    }

    ptphfig = getphfig( ptslice->MODEL_NAME, 'P' );

    for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
    {
        if( isvdd( ptphcon->NAME ) || isvss( ptphcon->NAME ) )
        {
            continue;
        }

        if( ( ptphcon->ORIENT == NORTH )
          ||( ptphcon->ORIENT == SOUTH ) )
        {
            ptcol->CON_NS = add_con_NS( ptcol->CON_NS,
                                        ptphcon->NAME,
                                        ptphcon->ORIENT,
                                        ptphcon->XCON - ptphfig->XAB1,
                                        ptphcon->WIDTH );
        }
        else
        {
            ptcol->CON_EW = add_con_EW( ptcol->CON_EW,
                                        ptphcon->NAME,
                                        ( ptphcon->YCON - ptphfig->YAB1 ) / SCALE_X );
        }
    }

    ref_ref = namealloc( "ref_ref" );
    for( ptphref = ptphfig->PHREF; ptphref; ptphref = ptphref->NEXT )
    {
        if( ( strncmp( ptphref->NAME, "nwell", 5 ) == 0 )
          ||( ref_ref == ptphref->FIGNAME ) )
        {
            continue;
        }

        con_name = get_con_name_from_ref_name( ptphref->NAME );

        ptcol->CON_EW = add_con_EW( ptcol->CON_EW,
                                    con_name,
                                    ( ptphref->YREF - ptphfig->YAB1 ) / SCALE_X );
    }

} /* end of create_con_list_for_routed_bloc */

/*---------------------------------------------------------\
                     create_con_list_for_column
\---------------------------------------------------------*/
void create_con_list_for_column( ptcol )
column *ptcol;
{
    slice      *ptslice;
    slice      *ptmaxslice;
    slice      *ptminslice;
    phfig_list *ptphfig;
    phcon_list *ptphcon;
    lofig_list *ptlofig;
    loins_list *ptloins;
    phref_list *ptphref;
    char       *con_name;
    char       *parent_conname;

    if( ptcol->TYPE != COL_COLUMN )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (create_con_list_for_column), i can't continue...\n" );
        exit( 1 );
    }

    ptlofig = getlofig( ptcol->MODEL_NAME, 'A' );
    lofigchain( ptlofig );

    ptmaxslice = ptcol->SLICES;
    ptminslice = ptcol->SLICES;
    for( ptslice=ptcol->SLICES; ptslice; ptslice = ptslice->NEXT )
    {
        if( ( ptslice->INDEX > ptmaxslice->INDEX )
          &&( ptslice->MODEL_NAME != NULL ) )
        {
            ptmaxslice = ptslice;
        }
        if( ptslice->INDEX < ptminslice->INDEX )
        {
            ptminslice = ptslice;
        }
    }

    ptphfig = getphfig( ptmaxslice->MODEL_NAME, 'P' );

    for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
    {
        if( isvdd( ptphcon->NAME ) || isvss( ptphcon->NAME ) )
        {
            continue;
        }

        if( ptphcon->ORIENT == NORTH )
        {
            parent_conname = getonemorelevelconname( ptphcon->NAME, ptlofig, ptmaxslice->INSTANCE_NAME );
            if( parent_conname == NULL )
            {
                continue;
            }
            ptcol->CON_NS = add_con_NS( ptcol->CON_NS,
                                        parent_conname,
                                        ptphcon->ORIENT,
                                        ptphcon->XCON - ptphfig->XAB1,
                                        ptphcon->WIDTH );
        }
    }

    ptphfig = getphfig( ptminslice->MODEL_NAME, 'P' );

    for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
    {
        if( isvdd( ptphcon->NAME ) || isvss( ptphcon->NAME ) )
        {
            continue;
        }

        if( ptphcon->ORIENT == SOUTH )
        {
            parent_conname = getonemorelevelconname( ptphcon->NAME, ptlofig, ptminslice->INSTANCE_NAME );
            if( parent_conname == NULL )
            {
                continue;
            }
            ptcol->CON_NS = add_con_NS( ptcol->CON_NS,
                                        parent_conname,
                                        ptphcon->ORIENT,
                                        ptphcon->XCON - ptphfig->XAB1,
                                        ptphcon->WIDTH );
        }
    }

    for( ptloins = ptlofig->LOINS; ptloins; ptloins = ptloins->NEXT )
    {
        ptphfig = getphfig( ptloins->FIGNAME, 'P' );

        for( ptphref = ptphfig->PHREF; ptphref; ptphref = ptphref->NEXT )
        {
            if( strncmp( ptphref->NAME, "nwell", 5 ) == 0 )
            {
                continue;
            }

            con_name = get_con_name_from_ref_name( ptphref->NAME );

            parent_conname = getonemorelevelconname( con_name, ptlofig, ptloins->INSNAME );
            if( parent_conname == NULL )
            {
                continue;
            }
            ptcol->CON_EW = add_con_EW( ptcol->CON_EW,
                                        parent_conname,
                                        ( ptphref->YREF - ptphfig->YAB1 ) / SCALE_X );
        }
    }

} /* end of create_con_list_for_column */

/*---------------------------------------------------------\
                     create_con_list_for_placed_bloc
\---------------------------------------------------------*/
void create_con_list_for_placed_bloc( ptcol )
column *ptcol;
{
    phfig_list *ptmainphfig;
    phfig_list *ptphfig;
    lofig_list *ptlofig;
    loins_list *ptloins;
    phins_list *ptphins;
    phref_list *ptphref;
    phcon_list *ptphcon;
    char       *con_name;
    char       *parent_conname;
    long        minY;
    long        maxY;

    if( ( ptcol->TYPE != COL_PLACED ) && ( ptcol->TYPE != COL_GLUE ) )
    {
        fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (create_con_list_for_placed_bloc 1), i can't continue...\n" );
        exit( 1 );
    }

    ptmainphfig = getphfig( ptcol->MODEL_NAME, 'A' );
    ptlofig     = getlofig( ptcol->MODEL_NAME, 'A' );
    lofigchain( ptlofig );

    maxY = ptmainphfig->YAB1;
    minY = ptmainphfig->YAB2;
    for( ptphins = ptmainphfig->PHINS; ptphins; ptphins=ptphins->NEXT )
    {
        ptphfig = getphfig( ptphins->FIGNAME, 'P' );

        if( ptphins->YINS < minY )
            minY = ptphins->YINS;
        if( ( ptphins->YINS + ptphfig->YAB2 - ptphfig->YAB1 ) > maxY )
            maxY = ptphins->YINS + ptphfig->YAB2 - ptphfig->YAB1;
    }   

    for( ptphins = ptmainphfig->PHINS; ptphins; ptphins=ptphins->NEXT )
    {
        ptphfig = getphfig( ptphins->FIGNAME, 'P' );

        if( ptphins->YINS == minY )
        {
            for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
            {
                if( isvdd( ptphcon->NAME ) || isvss( ptphcon->NAME ) )
                {
                    continue;
                }

                if( ptphcon->ORIENT == SOUTH )
                {
                    parent_conname = getonemorelevelconname( ptphcon->NAME, ptlofig, ptphins->INSNAME );
                    if( parent_conname == NULL )
                    {
                        continue;
                    }
                    ptcol->CON_NS = add_con_NS( ptcol->CON_NS,
                                                parent_conname,
                                                ptphcon->ORIENT,
                                                ptphcon->XCON - ptphfig->XAB1,
                                                ptphcon->WIDTH );
                }
            }
        }

        if( ( ptphins->YINS + ptphfig->YAB2 - ptphfig->YAB1 ) == maxY )
        {
            for( ptphcon = ptphfig->PHCON; ptphcon; ptphcon = ptphcon->NEXT )
            {
                if( isvdd( ptphcon->NAME ) || isvss( ptphcon->NAME ) )
                {
                    continue;
                }

                if( ptphcon->ORIENT == NORTH )
                {
                    parent_conname = getonemorelevelconname( ptphcon->NAME, ptlofig, ptphins->INSNAME );
                    if( parent_conname == NULL )
                    {
                        continue;
                    }
                    ptcol->CON_NS = add_con_NS( ptcol->CON_NS,
                                                parent_conname,
                                                ptphcon->ORIENT,
                                                ptphcon->XCON - ptphfig->XAB1,
                                                ptphcon->WIDTH );
                }
            }
        }
    }   

    for( ptloins = ptlofig->LOINS; ptloins; ptloins = ptloins->NEXT )
    {
        ptphfig = getphfig( ptloins->FIGNAME, 'A' );

        for( ptphref = ptphfig->PHREF; ptphref; ptphref = ptphref->NEXT )
        {
            if( strncmp( ptphref->NAME, "nwell", 5 ) == 0 )
            {
                continue;
            }

            con_name = get_con_name_from_ref_name( ptphref->NAME );

            parent_conname = getonemorelevelconname( con_name, ptlofig, ptloins->INSNAME );
            if( parent_conname == NULL )
            {
                continue;
            }
            for( ptphins = ptmainphfig->PHINS; ptphins; ptphins=ptphins->NEXT )
            {
                if( ptphins->INSNAME == ptloins->INSNAME )
                {
                    break;
                }
            }
            if( ptphins == NULL )
            {
                fprintf( stderr, "*** dpp bug *** : i am sorry but you find a bug (create_con_list_for_placed_bloc 2), i can't continue...\n" );
                exit( 1 );
            }

            ptcol->CON_EW = add_con_EW( ptcol->CON_EW,
                                        parent_conname,
                                        ( ptphins->YINS - ptmainphfig->YAB1
                                        + ptphref->YREF - ptphfig->YAB1 ) / SCALE_X );
        }
    }

} /* end of create_con_list_for_placed_bloc */
/*---------------------------------------------------------\
                     already_inplacecon 
\---------------------------------------------------------*/
int already_inplacecon ( ptplacecon, name )
PlaceConList *ptplacecon;
char *name;
{
    CaracConList *ptcon;

    for( ptcon = ptplacecon->NORTH_CON; ptcon; ptcon = ptcon->NEXT )
    {
        if( ptcon->NAME == name )
        {
            return TRUE;
        }
    }
    for( ptcon = ptplacecon->SOUTH_CON; ptcon; ptcon = ptcon->NEXT )
    {
        if( ptcon->NAME == name )
        {
            return TRUE;
        }
    }
    for( ptcon = ptplacecon->EAST_CON; ptcon; ptcon = ptcon->NEXT )
    {
        if( ptcon->NAME == name )
        {
            return TRUE;
        }
    }
    for( ptcon = ptplacecon->WEST_CON; ptcon; ptcon = ptcon->NEXT )
    {
        if( ptcon->NAME == name )
        {
            return TRUE;
        }
    }

    return FALSE;
} /* end of already_inplacecon  */
/*---------------------------------------------------------\
                     create_con_for_glue_placing
\---------------------------------------------------------*/
PlaceConList *create_con_for_glue_placing( ptplacement, ptmaincolumn )
placement  *ptplacement;
column     *ptmaincolumn;
{
    PlaceConList *ptplacecon;
    CaracConList *ptonecon;
	locon_list   *ptlocon;
	locon_list   *ptloconi;
	chain_list   *ptchain;
    column       *ptcolumn;
    con_EW       *pt_conEW;
	long          distance;
	long          newdist;
	long          y;
    long          nbslices;
    long          slicei;
    ptype_list   *ptype;
	ulocon_place_list *locon_place;

    extern FILE *yyin;
    static first_time = TRUE;

    if( !first_time )
    {
        yyrestart( yyin );
    }

    if( yyin = mbkfopen( ptmaincolumn->LOINS->FIGNAME, "scr", READ_TEXT ) )
    {
        if( verbose_mode == VERBOSE_ALL )
        {
            fprintf( stdout,
                     "     -> Loading %s.scr.\n",
                     ptmaincolumn->LOINS->FIGNAME);
        }
        yyparse( );
        fclose( yyin );
        return LeadPlaceConList;
    }

    nbslices = ptplacement->NB_SLICES;

    ptplacecon = mbkalloc( sizeof( PlaceConList ) );
    ptplacecon->SOUTH_CON = NULL;
    ptplacecon->NORTH_CON = NULL;
    ptplacecon->EAST_CON  = NULL;
    ptplacecon->WEST_CON  = NULL;
    ptplacecon->VER_FEED  = 0;
    ptplacecon->HOR_FEED  = 0;

	for( ptlocon = ptmaincolumn->LOINS->LOCON;
		 ptlocon;
		 ptlocon = ptlocon->NEXT)
	{
        if( isvdd( ptlocon->NAME ) || isvss( ptlocon->NAME ) )
        {
            continue;
        }

		distance = 0L;
        ptchain = (chain_list *)( getptype( ptlocon->SIG->USER, (long)LOFIGCHAIN)->DATA );

		while( ptchain )
		{
            ptloconi = (locon_list *)( ptchain->DATA );
            ptchain  = ptchain->NEXT;

            if( isvdd( ptloconi->NAME ) || isvss( ptloconi->NAME ) )
            {
                continue;
            }

            if( ptloconi->TYPE == EXTERNAL )
            {
                ptype = getptype( ptloconi->USER, (long)C_GfpUT_UserLoConPlace );
                if( ptype != NULL )
                {
                    locon_place = (ulocon_place_list *)( ptype->DATA );

                    if( ( locon_place->SIDE == C_UtFDP_SideTypeNorth )
                      ||( locon_place->SIDE == C_UtFDP_SideTypeSouth ) )
                    {

                        if( verbose_mode == VERBOSE_ALL )
                        {
                            fprintf( stdout,
                                     "     -> %s place to the %s side.\n",
                                      ptlocon->NAME,
                                      ( locon_place->SIDE == C_UtFDP_SideTypeNorth ) ? "north" : "south" );
                        }

                        ptonecon = mbkalloc( sizeof( CaracConList ) );
                        ptonecon->NEXT = ( locon_place->SIDE == C_UtFDP_SideTypeNorth )
                                         ? ptplacecon->NORTH_CON
                                         : ptplacecon->SOUTH_CON;
                        ptonecon->NAME = ptlocon->NAME;
                        ptonecon->USER = NULL;

                        if( locon_place->SIDE == C_UtFDP_SideTypeNorth )
                        {
                            ptplacecon->NORTH_CON = ptonecon;
                        }
                        else
                        {
                            ptplacecon->SOUTH_CON = ptonecon;
                        }
                    }
                    else
                    {
                        if( ( locon_place->SLICE >= ptplacement->NB_SLICES )
                          ||( locon_place->SLICE < 0 ) )
                        {
/*                             slicei = -1L; */
                            continue;
                        }
                        else
                        {
                            slicei = locon_place->SLICE;
                        }
                        if( verbose_mode == VERBOSE_ALL )
                        {
                            fprintf( stdout,
                                     "     -> %s place to the %s side on slice %ld (-1 is unknown).\n",
                                     ptlocon->NAME,
                                     ( locon_place->SIDE == C_UtFDP_SideTypeEast ) ? "east" : "west",
                                     slicei );
                        }

                        ptonecon = mbkalloc( sizeof( CaracConList ) );
                        ptonecon->NEXT = ( locon_place->SIDE == C_UtFDP_SideTypeEast )
                                         ? ptplacecon->EAST_CON
                                         : ptplacecon->WEST_CON;
                        ptonecon->NAME = ptlocon->NAME;
                        ptonecon->USER = slicei;

                        if( locon_place->SIDE == C_UtFDP_SideTypeEast )
                        {
                            ptplacecon->EAST_CON = ptonecon;
                        }
                        else
                        {
                            ptplacecon->WEST_CON = ptonecon;
                        }
                    }
                }
            }

		}
	}
	for( ptlocon = ptmaincolumn->LOINS->LOCON;
		 ptlocon;
		 ptlocon = ptlocon->NEXT)
	{
        if( isvdd( ptlocon->NAME ) || isvss( ptlocon->NAME ) )
        {
            continue;
        }

        if( already_inplacecon( ptplacecon, ptlocon->NAME ) )
        {
            continue;
        }

		distance = 0L;
        ptchain = (chain_list *)( getptype( ptlocon->SIG->USER, (long)LOFIGCHAIN)->DATA );

		while( ptchain )
		{
            ptloconi = (locon_list *)( ptchain->DATA );
            ptchain  = ptchain->NEXT;

            if( isvdd( ptloconi->NAME ) || isvss( ptloconi->NAME ) )
            {
                continue;
            }

            if( ptloconi->TYPE == INTERNAL )
            ptcolumn = (column *)getptype( ( (loins_list *)ptloconi->ROOT)->USER, (long)COLUMN_PTYPE )->DATA;
			if( ptcolumn == ptmaincolumn )
			{
				continue;
			}

			newdist = ptcolumn->INDEX - ptmaincolumn->INDEX;

			if( ( myabs(newdist) < myabs(distance) ) || ( distance == 0 ) )
			{
                for( pt_conEW = ptcolumn->CON_EW; pt_conEW; pt_conEW=pt_conEW->NEXT )
                {
                    if( pt_conEW->NAME == ptloconi->NAME )
                    {
				        y = YtoSlice( ptplacement, ptcolumn, pt_conEW );
                        distance = newdist;

                        if( verbose_mode == VERBOSE_ALL )
                        {
                            fprintf( stdout,
                                     "     -> %s connected with %s on slice %ld on %s side.\n",
                                      ptlocon->NAME, ptcolumn->INSTANCE_NAME, y, ( distance > 0 ) ? "east" : "west" );
                        }

                        ptonecon = mbkalloc( sizeof( CaracConList ) );
                        ptonecon->NEXT = ( distance > 0 ) ? ptplacecon->EAST_CON : ptplacecon->WEST_CON;
                        ptonecon->NAME = ptlocon->NAME;
                        ptonecon->USER = y;

                        if( distance > 0 )
                        {
                            ptplacecon->EAST_CON = ptonecon;
                        }
                        else
                        {
                            ptplacecon->WEST_CON = ptonecon;
                        }
                    }
                }
			}
        }
    }
    return( ptplacecon );
} /* end of create_con_for_glue_placing */

