/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/********************************************************************
*                                                                   *
* Laboratoire MASI CAO-VLSI, UPMC, Copyright 1991 1992 1993 1994    *
*                                                                   *
* Software support Email : cao-vlsi@masi.ibp.fr                     *
*                                                                   *
* Authors : Zouina AKTOUF &                                         *
*           El Arabi RHOMARI &                                      *
*           Jose MARTINS DOS SANTOS                                 *
*                                                                   *
* Supervision & Modifications : Lotfi BEN AMMAR                     *
*                                                                   *
* Restructuration : Olivier BEAURIN                                 *
*                                                                   *
********************************************************************/
/*********************************************************************************
*                      Fichier: router_util.c                                    *
*                                                                                *
*       version : 2.2  date    : 24/09/91                                        *
*                                                                                *
*       version : 3.1  date    : 27/03/94                                        *
*                                                                                *
*********************************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include MLO_H
#include MPH_H

#include "router_dpr.h"

/*---------------------------------------------------------\
                     manorme
\---------------------------------------------------------*/
char manorme( segtype )
char segtype;

{
    if( ( segtype == RIGHT ) || ( segtype == LEFT ) )
    {
        return( 'H' );
    }
    else if( ( segtype == UP ) || ( segtype == DOWN ) )
    {
       return('V');
    }

    fprintf( stderr, "Error: Segment type %c not valid (MBK problem)!!\n", segtype );
    exit( 1 );
}

/*---------------------------------------------------------\
                     chain_pred

 fonction chain_pred() : retourne l'element dans le chainage
 qui est juste avant l'element ptchain en donnee.
\---------------------------------------------------------*/
chain_list     *chain_pred (ptchead, ptchain)
chain_list     *ptchead;
chain_list     *ptchain;

{
    chain_list     *ptccour = NULL;

    for (ptccour = ptchead; ptccour; ptccour = ptccour->NEXT)
        if (ptccour->NEXT == ptchain)
            return ptccour;

    return NULL;
}

/*---------------------------------------------------------\
                     addchannel
\---------------------------------------------------------*/
 chain_list *addchannel(ptchead,ptsouth,ptnorth,ptwest,pteast)
 chain_list    *ptchead;
 ConnectorList *ptsouth;
 ConnectorList *ptnorth;
 ConnectorList *ptwest;
 ConnectorList *pteast;

 {
 CHANNEL *ptnewchan;

 ptnewchan = (CHANNEL *)mbkalloc(sizeof(CHANNEL));
 ptnewchan->SOUTHCON = ptsouth ;
 ptnewchan->NORTHCON = ptnorth ;
 ptnewchan->WESTCON  = ptwest ;
 ptnewchan->EASTCON  = pteast ;

 return addchain(ptchead,(void *)ptnewchan);

 }

/**************************************************************************************
*existsig
* recherche d'un signal dans une liste chaine de signaux; s'il existe retourne
* le signale dans la liste sinon NULL .
**************************************************************************************/

 chain_list  *existsig(ptcsig_list,ptsig)
 chain_list  *ptcsig_list;
 SIGNAL      *ptsig;
 {
 chain_list *ptcsig=NULL;
  
 for (ptcsig=ptcsig_list;ptcsig;ptcsig=ptcsig->NEXT)
     if (ptsig==(SIGNAL *)ptcsig->DATA)
        return(ptcsig);
 return(NULL);
 }


/************************************************************************ 
*		fonction	Warning()				*
************************************************************************/

 void Warning(index,ps1,ps2)
 int  index;
 char *ps1;
 char *ps2;
 {
 fprintf(stderr,"Warning DPR message( %ld ) : ",index);
 switch (index)
    {
    case 1  : fprintf(stderr,"Superposition d'instances (selon X ou Y) entre\
                            %s et %s\n",ps1,ps2);break;
    case 10 : fprintf(stderr," : %s \n",ps1);affcon(stderr,(CONNECTOR *)ps2);break;
    default : fprintf(stderr," : %s , %s .\n",ps1,ps2);break;
    }
 }



/************************************************************************ 
*		fonction	Error()				*
************************************************************************/

 void Error(index,ps,ptinfo)
 int  index;
 char *ps;
 char *ptinfo;
 {
 fprintf(stderr,"Error DPR message(%ld) : ",index);fprintf(stderr," %s . \n",ps);
 switch (index)
    {
    case 10 :;
    case 11 :;
    case 12 :;
    case 13 :;
    case 14 : affcon(stderr,(CONNECTOR *)ptinfo);
    default : break; 
    }
 exit(1);
 }

/**************************************************************
**  strrch()
**************************************************************/
 long strrch(ptstr,car)
 char *ptstr;
 char car;
 {
 long i;

 for (i=strlen(ptstr);i>0;i--)
     if(ptstr[i]==car) return(i);
 return 0 ;
 }

/******************************************************************************
**         Bibliotheque de fonctions pour la gestion des orientations        **
**         en fonction du champ TRANSF des instances.                        **
**                                                                           **
******************************************************************************/

/******************************************************************************
**     Fonction delta_x :  Calcul de la largeur de l'instance en tenant      **
**                        compte de sa transformation.                       **
******************************************************************************/

 long delta_x(xab1,yab1,xab2,yab2,x,y,TRANSF)
 long xab1,yab1,xab2,yab2,x,y;
 int TRANSF;
 {
 long delta;

 switch (TRANSF)
    {
    case (int) NOSYM : delta = x - xab1 ; break;
    case (int) SYM_X : delta = -x + xab2 ; break ;
    case (int) SYM_Y : delta = x - xab1 ; break;
    case (int) SYMXY : delta = -x + xab2 ; break ;
    case (int) ROT_P : delta = yab2 - y ; break;
    case (int) ROT_M : delta = y - yab1 ; break;
    case (int) SY_RM : delta = yab2 - y ; break;
    case (int) SY_RP : delta = y - yab1 ; break;
    };
 return(delta);
 }


/******************************************************************************
**     Fonction delta_y :  Calcul de la hauteur de l'instance en tenant      **
**                        compte de sa transformation.                       **
******************************************************************************/

 long delta_y(xab1,yab1,xab2,yab2,x,y,TRANSF)
 long xab1,yab1,xab2,yab2,x,y;
 int TRANSF;
 {
 long delta;

 switch (TRANSF)
    {
    case (int) NOSYM : delta = y - yab1 ; break;
    case (int) SYM_X : delta = y - yab1 ; break;
    case (int) SYM_Y : delta = yab2 - y ; break;
    case (int) SYMXY : delta = yab2 - y ; break;
    case (int) ROT_P : delta = x - xab1 ; break;
    case (int) ROT_M : delta = xab2 - x ; break;
    case (int) SY_RM : delta = xab2 - x ; break;
    case (int) SY_RP : delta = x - xab1 ; break;
    };
 return(delta);
 }

/******************************************************************************
**     Fonction orient_r :  Calcul de l'orientation reelle d'une reference   **
**      de type RIGHT en tenant compte de la transformation de l'instance.   **
******************************************************************************/
 char orient_r(TRANSF)
 int TRANSF;
 {
 switch (TRANSF)
    {
    case (int) NOSYM : ;
    case (int) SYM_Y : return (EAST);
    case (int) SYM_X : ;
    case (int) SYMXY : return (WEST);
    case (int) ROT_P : ;
    case (int) SY_RP : return (NORTH);
    case (int) ROT_M : ;
    case (int) SY_RM : return (SOUTH);
    };
 }

/******************************************************************************
**     Fonction orient_l :  Calcul de l'orientation reelle d'une reference   **
**      de type LEFT en tenant compte de la transformation de l'instance.    **
******************************************************************************/
 char orient_l(TRANSF)
 int TRANSF;
 {
 switch (TRANSF)
    {
    case (int) NOSYM : ;
    case (int) SYM_Y : return (WEST);
    case (int) SYM_X : ;
    case (int) SYMXY : return (EAST);
    case (int) ROT_P : ;
    case (int) SY_RP : return (SOUTH);
    case (int) ROT_M : ;
    case (int) SY_RM : return (NORTH);
    };
 }

/******************************************************************************
**  Fonction orient_rl :  Calcul de l'orientation reelle d'une reference     **
**  de type RIGHT/LEFT en tenant compte de la transformation de l'instance   **
******************************************************************************/
 char orient_rl(TRANSF)
 int TRANSF;
 {
 switch (TRANSF)
    {
    case (int) NOSYM : ;
    case (int) SYM_Y : ; 
    case (int) SYM_X : ;
    case (int) SYMXY : return ('H');
    case (int) ROT_P : ;
    case (int) SY_RP : ;
    case (int) ROT_M : ;
    case (int) SY_RM : return ('V');
    };
 }

/***************************************************************************
**  Fonction orient :  Calcul de l'orientation reelle d'une reference     **
**   de type quelconque parmis NORTH/SOUTH/EAST/WEST en tenant compte     **
**   de la transformation de l'instance.                                  **
***************************************************************************/
 char Orient(ORIENT,TRANSF)
 char ORIENT;
 char TRANSF;
 {
 if (ORIENT == NORTH)
    switch (TRANSF)
       {
       case (int) NOSYM : return (NORTH);
       case (int) SYM_X : return (NORTH); 
       case (int) SYM_Y : return (SOUTH);
       case (int) SYMXY : return (SOUTH);
       case (int) ROT_P : return (WEST);
       case (int) SY_RP : return (EAST);
       case (int) ROT_M : return (EAST);
       case (int) SY_RM : return (WEST);
       };

 if (ORIENT == SOUTH)
    switch (TRANSF)
       {
       case (int) NOSYM : return (SOUTH);
       case (int) SYM_X : return (SOUTH);
       case (int) SYM_Y : return (NORTH);
       case (int) SYMXY : return (NORTH);
       case (int) ROT_P : return (EAST);
       case (int) SY_RP : return (WEST);
       case (int) ROT_M : return (WEST);
       case (int) SY_RM : return (EAST);
       };

 if (ORIENT == EAST)
    switch (TRANSF)
       {
       case (int) NOSYM : return (EAST);
       case (int) SYM_X : return (WEST);
       case (int) SYM_Y : return (EAST);
       case (int) SYMXY : return (WEST);
       case (int) ROT_P : return (NORTH);
       case (int) SY_RP : return (NORTH);
       case (int) ROT_M : return (SOUTH);
       case (int) SY_RM : return (SOUTH);
       };

 if (ORIENT == WEST)
    switch(TRANSF)
       {
       case (int) NOSYM : return (WEST);
       case (int) SYM_X : return (EAST); 
       case (int) SYM_Y : return (WEST);
       case (int) SYMXY : return (EAST);
       case (int) ROT_P : return (SOUTH);
       case (int) SY_RP : return (SOUTH);
       case (int) ROT_M : return (NORTH);
       case (int) SY_RM : return (NORTH);
       };

 Error(20,"Orientation non standard.","");
 }

/******************************************************************************
**     Fonction deltab_y :  Calcul de la hauteur de l'instance en tenant     **
**                        compte de sa transformation.                       **
******************************************************************************/
 long deltab_y(xab1,yab1,xab2,yab2,TRANSF)
 long xab1,yab1,xab2,yab2;
 int TRANSF;
 {
 long delta;

 switch (TRANSF)
    {
    case (int) NOSYM : ;
    case (int) SYM_X : ;
    case (int) SYM_Y : ;
    case (int) SYMXY : delta = yab2 - yab1 ; break;
    case (int) ROT_P : ;
    case (int) ROT_M : ;
    case (int) SY_RM : ;
    case (int) SY_RP : delta = xab2 - xab1 ; break;
    };
 return(delta);
 }

/******************************************************************************
**     Fonction deltab_x :  Calcul de la largeur de l'instance en tenant     **
**                        compte de sa transformation.                       **
******************************************************************************/
 long deltab_x(xab1,yab1,xab2,yab2,TRANSF)
 long xab1,yab1,xab2,yab2;
 int TRANSF;
 {
 long delta;

 switch (TRANSF)
    {
    case (int) NOSYM : ;
    case (int) SYM_X : ;
    case (int) SYM_Y : ;
    case (int) SYMXY : delta = xab2 - xab1 ; break;
    case (int) ROT_P : ;
    case (int) ROT_M : ;
    case (int) SY_RM : ;
    case (int) SY_RP : delta = yab2 - yab1 ; break;
    };
 return(delta);
 }

/*---------------------------------------------------------\
                     addconchannel
\---------------------------------------------------------*/
 ConnectorList *addconchannel(ptconsig_CH,indexsig,ycon)
 ConnectorList *ptconsig_CH ;
 long           indexsig;
 long           ycon;

 {
 ConnectorList *ptnewcon;

 ptnewcon = (ConnectorList *)mbkalloc(sizeof(ConnectorList));
 ptnewcon->NextCon = ptconsig_CH ;
 ptnewcon->ConName = indexsig ;
 ptnewcon->Mark = ycon ;

 return ptnewcon ;
 }

/*---------------------------------------------------------\
                     insertcon_inchan
\---------------------------------------------------------*/
 ConnectorList *insertcon_inchan(ptchannel,orient,index,y)
 CHANNEL *ptchannel;
 char     orient;
 long     index;
 long     y;
 {
 
 ConnectorList *ptcon_ch=NULL;
 ConnectorList *ptcon_save=NULL;

 switch(orient)
    {
    case NORTH : ptcon_ch = ptchannel->NORTHCON;break;
    case SOUTH : ptcon_ch = ptchannel->SOUTHCON;break;
    case WEST  : ptcon_ch = ptchannel->EASTCON ;break;
    case EAST  : ptcon_ch = ptchannel->WESTCON ;break;
    }
 for (ptcon_save=NULL;ptcon_ch;ptcon_ch=ptcon_ch->NextCon)
     if (ptcon_ch->Mark > y)
        break;
     else
        ptcon_save = ptcon_ch;
 
 if (ptcon_save)
    {
    if ((orient==WEST)||(orient==EAST))
       {
       if (ptcon_save->Mark!=y)
          {
          ptcon_save->NextCon = addconchannel(ptcon_save->NextCon,index,y);
          return ptcon_save->NextCon;
          }
       else
          if (ptcon_save->ConName!=index)
             {
             fflush( stdout );
             fprintf(stderr,"*** dpr bugs *** : Different signals on the same Y in channel(signal %ld and %ld).\n",
                     ptcon_save->ConName, index);
             exit(1);
             }
          else
             return NULL;
       }
    else
       {
       ptcon_save->NextCon = addconchannel(ptcon_save->NextCon,index,y);
       return ptcon_save->NextCon;
       }
    }
 else
    switch(orient)
      {
      case NORTH : ptchannel->NORTHCON = addconchannel(ptchannel->NORTHCON,index,y);
                   return ptchannel->NORTHCON;
      case SOUTH : ptchannel->SOUTHCON = addconchannel(ptchannel->SOUTHCON,index,y);
                   return ptchannel->SOUTHCON;
      case WEST  : ptchannel->EASTCON = addconchannel(ptchannel->EASTCON,index,y);
                   return ptchannel->EASTCON;
      case EAST  : ptchannel->WESTCON = addconchannel(ptchannel->WESTCON,index,y);
                   return ptchannel->WESTCON;
      }
 }

/********************************************************************
**        fonction addnwell()                                      **
**     rajoute un nwell a une liste pointe par pthead              **
********************************************************************/
 chain_list *addnwell(pthead,x,y,width,ptccell,ptccol)
 chain_list *pthead;
 long        x,y;
 long        width;
 chain_list *ptccell;
 chain_list *ptccol;

 {
 NWELLREF *ptnwell=NULL;

 ptnwell        = (NWELLREF *)mbkalloc(sizeof(NWELLREF));
 ptnwell->X     = x;
 ptnwell->Y     = y;
 ptnwell->WIDTH = width;
 ptnwell->INS   = ptccell;
 ptnwell->COL   = ptccol;

 return(addchain(pthead,(void *)ptnwell));

 }

/********************************************************************
**        fonction addpos()                                        **
**     rajoute une possibilite a un connecteur                     **
********************************************************************/
 chain_list *addpos(pthead,x,y,layer,orient,type,track,width,ptcon)
 chain_list *pthead;
 long        x,y;
 char        layer;
 char        orient;
 char        type;
 long        track;
 long        width;
 CONNECTOR  *ptcon;

 {
 POSSIBILITY *ptpos=NULL;

 ptpos = (POSSIBILITY *)mbkalloc(sizeof(POSSIBILITY));
 ptpos->X = x;
 ptpos->Y = y;
 ptpos->LAYER = layer;
 ptpos->ACCES = orient;
 ptpos->TYPE  = type;
 ptpos->TRACK = track;
 ptpos->WIDTH = width;
 ptpos->CON   = ptcon;
 ptpos->USER  = NULL;

 return(addchain(pthead,(void *)ptpos));

 }

/*********************************************************************************
*     addtrack                                                                   *
* insertion d'une piste dans la liste des pistes en tete de la structure.        *
*********************************************************************************/
 TRACK  *addtrack(pthead,y)
 TRACK  *pthead;
 long    y;

 {
 TRACK  *pttrack=NULL;
  
 pttrack = (TRACK *)mbkalloc(sizeof(TRACK));
 pttrack->NEXT = pthead ;
 pttrack->Y = y;
 
 return pttrack ;
 }

