/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_phcma.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_support.h"
#include  "strecher_supply.h"
#include  "strecher_phcma.h"

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define  MAXTPLAYER    1000

#define   NO_ELT       2
#define  ELT_SEG       3
#define  ELT_VIA       4

/*---------------------------------------------------------\
                     locals
\---------------------------------------------------------*/
static char     TabPileLayer[MAXTPLAYER];
static long     sp_push_layer, sp_pop_layer;
static phseg_list *Lptr_SEG;
static phvia_list *Lptr_VIA;
static phref_list *Lptr_REF;
static phfig_list *ptfig_EXTRACT;
static supply_list *ptfig_SUPPLY;

/*---------------------------------------------------------\
                     debug_cma
\---------------------------------------------------------*/
void            debug_cma(a_ptfig)
phfig_list     *a_ptfig;
{
    ptype_list     *ppcma;
    phcma_list     *pcma;

    if(!getenv("DEBUG_STRECHER"))
        return;

    ppcma = get_ls_data(a_ptfig, C_SUPP_UCMA, FALSE);
    if (ppcma == (ptype_list *) NULL)
        return;

    fprintf(stdout, "Affichage des CMA de %s\n",a_ptfig->NAME);
    pcma = (phcma_list *) ppcma->DATA;
    while (pcma != (phcma_list *) NULL)
    {
        fprintf(stdout, "--> CMA name : %s\n",pcma->NAME);
        pcma=pcma->NEXT;
    }
}

/*---------------------------------------------------------\
                     layervia
\---------------------------------------------------------*/
static long     layervia(via, layer)
phvia_list     *via;
long            layer;
{
    long            vlayer[2];

    vlayer[1] = ALU1;
    switch (via->TYPE)
    {
    case CONT_POLY:
        vlayer[0] = POLY;
        break;
    case CONT_VIA:
        vlayer[0] = ALU2;
        break;
    case CONT_DIF_N:
        vlayer[0] = NDIF;
        break;
    case CONT_DIF_P:
        vlayer[0] = PDIF;
        break;
    case CONT_VIA2:
        vlayer[0] = ALU3;
        vlayer[1] = ALU2;
        break;
    case CONT_BODY_N:;
    case CONT_BODY_P:;
    case C_X_N:;
    case C_X_P:;
    default:
        fflush( stdout );
        fprintf( stderr, "*** strecher error *** : via's layer not authorized.\n" );
        exit( 1 );
    }

    return (vlayer[(layer != vlayer[1]) ? 1 : 0]);
}

/*---------------------------------------------------------\
                     reset_seg
\---------------------------------------------------------*/
static void     reset_seg()
{
    Lptr_SEG = ptfig_EXTRACT->PHSEG;
}


/*---------------------------------------------------------\
                     islayerseg
\---------------------------------------------------------*/
static long     islayerseg(pseg, LAYER)
phseg_list     *pseg;
char            LAYER;
{

    switch (pseg->LAYER)
    {
    case ALU2:
        return (LAYER == ALU2);
    case ALU1:
        return (LAYER == ALU1);
    case NDIF:
        return (LAYER == NDIF);
    case PDIF:
        return (LAYER == PDIF);
    case POLY:;
    case NTRANS:;
    case PTRANS:
        return ((LAYER == POLY)
                || (LAYER == PTRANS)
                || (LAYER == NTRANS));
    default:
        return (FALSE);
    }
}

/*---------------------------------------------------------\
                     next_seg
\---------------------------------------------------------*/
static phseg_list *next_seg(LAYER)
char            LAYER;
{
    phseg_list     *cseg;

    while (Lptr_SEG != (phseg_list *) NULL)
    {
        cseg = Lptr_SEG;
        Lptr_SEG = cseg->NEXT;

        if ((cseg->USER == (ptype_list *) NULL) && (islayerseg(cseg, LAYER)))
            return (cseg);
    }

    return ((phseg_list *) NULL);
}


/*---------------------------------------------------------\
                     reset_via
\---------------------------------------------------------*/
static void     reset_via()
{
    Lptr_VIA = ptfig_EXTRACT->PHVIA;
}

/*---------------------------------------------------------\
                     islayervia
\---------------------------------------------------------*/
static long     islayervia(pvia, LAYER)
phvia_list     *pvia;
char            LAYER;
{

    switch (pvia->TYPE)
    {
    case CONT_POLY:
        return ((LAYER == ALU1) || (LAYER == POLY)
                || (LAYER == PTRANS)
                || (LAYER == NTRANS));
    case CONT_VIA:
        return ((LAYER == ALU1) || (LAYER == ALU2));
    case CONT_DIF_N:
        return ((LAYER == ALU1) || (LAYER == NDIF));
    case CONT_DIF_P:
        return ((LAYER == ALU1) || (LAYER == PDIF));
    case CONT_BODY_N:
        return ((LAYER == ALU1) || (LAYER == NWELL));
    case CONT_BODY_P:
        return ((LAYER == ALU1) || (LAYER == PWELL));
    case CONT_VIA2:
        return ((LAYER == ALU2) || (LAYER == ALU3));
    case C_X_N:
        return (FALSE);
    case C_X_P:
        return (FALSE);
    }
    return FALSE;
}

/*---------------------------------------------------------\
                     next_via
\---------------------------------------------------------*/
static phvia_list *next_via(LAYER)
char            LAYER;
{
    phvia_list     *pvia;

    while (Lptr_VIA != (phvia_list *) NULL)
    {
        pvia = Lptr_VIA;
        Lptr_VIA = pvia->NEXT;

        if ((pvia->USER == (ptype_list *) NULL) && (islayervia(pvia, LAYER)))
            return (pvia);
    }

    return ((phvia_list *) NULL);
}

/*---------------------------------------------------------\
                     reset_pile_layer
\---------------------------------------------------------*/
static void     reset_pile_layer()
{
    sp_push_layer = sp_pop_layer = 0;
}

/*---------------------------------------------------------\
                     push_layer
\---------------------------------------------------------*/
static void     push_layer(LAYER)
char            LAYER;
{

    TabPileLayer[sp_push_layer] = LAYER;
    sp_push_layer = (sp_push_layer + 1) % MAXTPLAYER;
} /* end of push_layer */

/*---------------------------------------------------------\
                     pop_layer
\---------------------------------------------------------*/
static char     pop_layer()
{
    char            LAYER;


    LAYER = TabPileLayer[sp_pop_layer];
    sp_pop_layer = (sp_pop_layer + 1) % MAXTPLAYER;

    return (LAYER);
}

/*---------------------------------------------------------\
                     reset_ref
\---------------------------------------------------------*/
static void     reset_ref()
{
    Lptr_REF = ptfig_EXTRACT->PHREF;
} /* end of pop_layer */

/*---------------------------------------------------------\
                     next_ref
\---------------------------------------------------------*/
static phref_list *next_ref()
{
    phref_list     *pref;

    while (Lptr_REF != (phref_list *) NULL)
    {
        pref = Lptr_REF;
        Lptr_REF = pref->NEXT;

        if ((pref->USER == (ptype_list *) NULL) && (pref->FIGNAME == namealloc("ref_con")))
            return (pref);
    }

    return ((phref_list *) NULL);
}


/*---------------------------------------------------------\
                     check_name
\---------------------------------------------------------*/
static void     check_name(ptcma, pref)
phcma_list     *ptcma;
phref_list     *pref;
{
    ptype_list     *lpref;

    long            index;
    char           *pname;

    pname = getrefconname(pref->NAME, &index);

    if ((ptcma->TYPE == C_PHCMA_CMA) && (strcmp(pname, ptcma->NAME) != 0))
    {
        fflush( stdout );
        fprintf( stderr,
                 "*** strecher error *** : reference name(%s) and connector name(%s) incompatible.\n",
                 pname,ptcma->NAME);
        exit( 1 );
    }

    for (lpref = ptcma->PPHREF; lpref != (ptype_list *) NULL; lpref = lpref->NEXT)
    {
        if ((((phref_list *) lpref->DATA)->NAME == pref->NAME) && (((phref_list *) lpref->DATA) != pref))
        {
            fflush( stdout );
            fprintf( stderr,
                     "*** strecher error *** : reference name(%s) and connector name(%s) incompatible.\n",
                     pname,ptcma->NAME);
            exit( 1 );
        }
    }
}

/*---------------------------------------------------------\
                     addreftocma
\---------------------------------------------------------*/
static void     addreftocma(ptcma, ptref)
phcma_list     *ptcma;
phref_list     *ptref;
{
#define  access_ref  ( (phref_list*)(pphref->DATA) )

    ptype_list     *pphref;
    supply_list    *pslice;
    long            index, mult;
    char           *pname, new_name[MBK_SZ_NAME];

    pname = getrefconname(ptref->NAME, &index);

    pslice = getnsupply(ptfig_SUPPLY, ptref->YREF);

    mult = 0;
    for (pphref = ptcma->PPHREF; pphref != (ptype_list *) NULL;
         pphref = pphref->NEXT)
        if (access_ref->YREF == ptref->YREF)
            mult += 1;

    sprintf(new_name, "%s_%ld", pname,
            index_calc(pslice, mult, ptref->YREF));

    /* Allocation du nouveau nom. */
    ptref->NAME = namealloc(new_name);

    ptcma->TYPE = C_PHCMA_CMA;
    ptcma->NBREF += 1;
    ptcma->PPHREF = addptype(ptcma->PPHREF, PTYPE_LREF, (void *) ptref);

    ptref->USER = (ptype_list *) ptcma;

#undef  access_ref
}

/*---------------------------------------------------------\
                     joinreftocma
\---------------------------------------------------------*/
static void     joinreftocma(ptcma, ptref)
phcma_list     *ptcma;
phref_list     *ptref;
{
    long            index;
    char           *pname;

    pname = getrefconname(ptref->NAME, &index);

    ptcma->NAME = namealloc(pname);
    addreftocma(ptcma, ptref);
}

/*---------------------------------------------------------\
                     extract_cma
\---------------------------------------------------------*/
static void     extract_cma(ptcma, amorce_SEG)
phcma_list     *ptcma;
phseg_list     *amorce_SEG;
{
    long            element, Xr1, Yr1, Xr2, Yr2;
    char            LAYER;
    phseg_list     *Ltete_SEG, *Lqueue_SEG, *pseg;
    phvia_list     *Ltete_VIA, *Lqueue_VIA, *pvia, amorce_VIA;
    phref_list     *pref;

    element = ELT_SEG;
    Ltete_SEG = Lqueue_SEG = amorce_SEG;
    Ltete_VIA = Lqueue_VIA = &amorce_VIA;
    amorce_SEG->USER = (ptype_list *) ptcma;
    amorce_VIA.USER = (ptype_list *) ptcma;
    reset_pile_layer();

    Xr1 = amorce_SEG->X1;
    Yr1 = amorce_SEG->Y1;
    Xr2 = amorce_SEG->X2;
    Yr2 = amorce_SEG->Y2;
    LAYER = amorce_SEG->LAYER;

    do
    {
        reset_seg();
        reset_via();
        reset_ref();

        while ((pseg = next_seg(LAYER)) != (phseg_list *) NULL)
        {
            if (isconnect(pseg->X1, pseg->Y1, pseg->X2, pseg->Y2,
                          Xr1, Yr1, Xr2, Yr2))
            {
                Lqueue_SEG->USER = (ptype_list *) pseg;
                Lqueue_SEG = pseg;
                Lqueue_SEG->USER = (ptype_list *) ptcma;
            }
        }

        while ((pvia = next_via(LAYER)) != (phvia_list *) NULL)
        {
            if (isconnect(pvia->XVIA, pvia->YVIA,
                          pvia->XVIA, pvia->YVIA,
                          Xr1, Yr1, Xr2, Yr2))
            {
                Lqueue_VIA->USER = (ptype_list *) pvia;
                Lqueue_VIA = pvia;
                Lqueue_VIA->USER = (ptype_list *) ptcma;
                push_layer(layervia(pvia, LAYER));
            }
        }

        if (LAYER == ALU1)
            while ((pref = next_ref()) != (phref_list *) NULL)
            {
                if (isconnect(pref->XREF, pref->YREF,
                              pref->XREF, pref->YREF,
                              Xr1, Yr1, Xr2, Yr2))
                {
                    if (ptcma->TYPE == C_PHCMA_CMA)
                    {
                        addreftocma(ptcma, pref);
                        check_name(ptcma, pref);
                    }
                    else
                    {
                        joinreftocma(ptcma, pref);
                    }
                }
            }

        if ((Ltete_SEG->USER != (ptype_list *) ptcma) || (Ltete_VIA->USER != (ptype_list *) ptcma))
            element = (Ltete_SEG->USER != (ptype_list *) ptcma) ? ELT_SEG : ELT_VIA;
        else
            element = NO_ELT;

        if (element == ELT_SEG)
        {
            pseg = Ltete_SEG;
            Ltete_SEG = (phseg_list *) pseg->USER;
            pseg->USER = (ptype_list *) ptcma;

            Xr1 = Ltete_SEG->X1;
            Yr1 = Ltete_SEG->Y1;
            Xr2 = Ltete_SEG->X2;
            Yr2 = Ltete_SEG->Y2;
            LAYER = Ltete_SEG->LAYER;
        }

        if (element == ELT_VIA)
        {
            pvia = Ltete_VIA;
            Ltete_VIA = (phvia_list *) pvia->USER;
            pvia->USER = (ptype_list *) ptcma;

            Xr1 = Xr2 = Ltete_VIA->XVIA;
            Yr1 = Yr2 = Ltete_VIA->YVIA;
            LAYER = pop_layer();
        }
    }
    while (element != NO_ELT);
}

/*---------------------------------------------------------\
                     addphcma
\---------------------------------------------------------*/
static phcma_list *addphcma(pthead)
phcma_list     *pthead;
{
    static long     id = 0;
    char            default_name[MBK_SZ_NAME];
    phcma_list     *pcma;


    pcma = (phcma_list *) mbkalloc(sizeof(phcma_list));

    /* Nom par defaut du CMA. ( unique a cause du compteur "id" ) */
    sprintf(default_name, "default_%ld", id++);

    pcma->NAME = namealloc(default_name);
    pcma->TYPE = C_PHCMA_EQUI;
    pcma->LINK = (phcma_link_list *) NULL;
    pcma->PPHREF = (ptype_list *) NULL;
    pcma->NBREF = 0;
    pcma->NEXT = pthead;

    return (pcma);
}

/*---------------------------------------------------------\
                     extract_ls_cma
\---------------------------------------------------------*/
void            extract_ls_cma(a_ptfig_EXTRACT, a_ptfig_supply)
phfig_list     *a_ptfig_EXTRACT;
supply_list    *a_ptfig_supply;
{
#define  access_ls_cma  ( (phcma_list*)(ptuser->DATA) )

    long            flag;
    phref_list     *pref;
    phseg_list     *pseg;
    ptype_list     *ptuser;

    ptfig_EXTRACT = a_ptfig_EXTRACT;
    ptfig_SUPPLY = a_ptfig_supply;

    ptuser = get_ls_data(ptfig_EXTRACT, C_SUPP_UCMA, F_SUPP_UCREAT);

    reset_ref();

    while ((pref = next_ref()) != (phref_list *) NULL)
    {
        reset_seg();
        flag = FALSE;

        while ((pseg = next_seg(ALU1)) != (phseg_list *) NULL)
        {
            if (isconnect(pref->XREF, pref->YREF,
                          pref->XREF, pref->YREF,
                          pseg->X1, pseg->Y1,
                          pseg->X2, pseg->Y2))
            {
                access_ls_cma = addphcma(access_ls_cma);
                joinreftocma(access_ls_cma, pref);

                extract_cma(access_ls_cma, pseg);

                reset_ref();
                flag = TRUE;

                break;
            }
        }

        if (flag == FALSE)
        {
            fflush( stdout );
            fprintf( stderr, "*** strecher error *** : ref_con not in ALU1 segment in %s.\n",a_ptfig_EXTRACT->NAME);
            exit( 1 );
        }
    }
#undef  access_ls_cma
}

/*---------------------------------------------------------\
                     del_ls_cmalink
\---------------------------------------------------------*/
static void     del_ls_cmalink(ptlink)
phcma_link_list *ptlink;
{
    phcma_link_list *plink, *pDEL;

    plink = ptlink;
    while (plink != (phcma_link_list *) NULL)
    {
        pDEL = plink;
        plink = plink->NEXT;

        mbkfree(pDEL);
    }
} /* end of del_ls_cmalink */

/*---------------------------------------------------------\
                     del_ls_cma
\---------------------------------------------------------*/
void            del_ls_cma(a_ptfig)
phcma_list     *a_ptfig;
{
    ptype_list     *ppcma;
    phcma_list     *pcma, *pDEL;


    ppcma = get_ls_data(a_ptfig, C_SUPP_UCMA, FALSE);
    if (ppcma == (ptype_list *) NULL)
        return;

    pcma = (phcma_list *) ppcma->DATA;
    while (pcma != (phcma_list *) NULL)
    {
        pDEL = pcma;
        pcma = pcma->NEXT;

        del_ls_cmalink(pDEL->LINK);
        mbkfree(pDEL);
    }

    del_ls_data(a_ptfig, C_SUPP_UCMA, FALSE);
} /* end of del_ls_cma */


/*---------------------------------------------------------\
                     joinlinktocma
\---------------------------------------------------------*/
static void     joinlinktocma(ptcma, ptlink, ptfig, xflat, yflat, atransf)
phcma_list     *ptcma;
phcma_list     *ptlink;
phfig_list     *ptfig;
long            xflat;
long            yflat;
char            atransf;
{
    phcma_link_list *pNEW, *pcmalink;

    /* Verification de la non-existence du lien. */
    for (pcmalink = ptcma->LINK; pcmalink != (phcma_link_list *) NULL;
         pcmalink = pcmalink->NEXT)
        if ((pcmalink->PHFIG == ptfig)
            && (pcmalink->PHCMA == ptlink)
            && (pcmalink->XFLAT == xflat)
            && (pcmalink->YFLAT == yflat)
            && (pcmalink->ATRANSF == atransf))
            return;

    /* Le lien est bien nouveau : addition de l'element. */
    pNEW = (phcma_link_list *) mbkalloc(sizeof(phcma_link_list));

    pNEW->PHFIG = ptfig;
    pNEW->PHCMA = ptlink;
    pNEW->XFLAT = xflat;
    pNEW->YFLAT = yflat;
    pNEW->ATRANSF = atransf;
    pNEW->NEXT = ptcma->LINK;

    ptcma->LINK = pNEW;
} /* end of joinlinktocma */

/*---------------------------------------------------------\
                     dup_ls_cma
\---------------------------------------------------------*/
void           *dup_ls_cma(a_ptfig_src, a_ptfig_dest)
phfig_list     *a_ptfig_src;
phfig_list     *a_ptfig_dest;
{
    phcma_link_list *plink;
    phcma_list     *pcma, *pcma_SRC, *pcma_DUP;
    phseg_list     *pseg;
    phvia_list     *pvia;
    phref_list     *pref;
    ptype_list     *ptuser_SRC, *ptuser_DUP;


    /* Controle de l'existence des structures d'accueil. */
    if ((ptuser_SRC = get_ls_data(a_ptfig_src, C_SUPP_UCMA, FALSE)) != (ptype_list *) NULL)
        pcma_SRC = (phcma_list *) ptuser_SRC->DATA;
    else
        return ((void *) NULL);

    if ((ptuser_DUP = get_ls_data(a_ptfig_dest, C_SUPP_UCMA, FALSE))
                != (ptype_list *) NULL)
        if (ptuser_DUP->DATA != (void *) NULL)
        {
            del_ls_cma(a_ptfig_dest);
            del_ls_data(a_ptfig_dest, C_SUPP_UCMA, FALSE);
        }

    pcma_DUP = (phcma_list *) NULL;

    for (pcma = pcma_SRC; pcma != (phcma_list *) NULL;
         pcma = pcma->NEXT)
    {
        pcma_DUP = addphcma(pcma_DUP);

        /* Transfert des segments appartenant au CMA. */
        for (pseg = a_ptfig_dest->PHSEG;
             pseg != (phseg_list *) NULL;
             pseg = pseg->NEXT)
            if ((phcma_list *) (pseg->USER) == pcma)
                pseg->USER = (ptype_list *) pcma_DUP;

        /* Transfert des VIA et contacts du CMA. */
        for (pvia = a_ptfig_dest->PHVIA;
             pvia != (phvia_list *) NULL;
             pvia = pvia->NEXT)
            if ((phcma_list *) (pvia->USER) == pcma)
                pvia->USER = (ptype_list *) pcma_DUP;

        /* Transfert des references du CMA. */
        for (pref = a_ptfig_dest->PHREF;
             pref != (phref_list *) NULL;
             pref = pref->NEXT)
            if ((phcma_list *) (pref->USER) == pcma)
            {
                pref->USER = (ptype_list *) pcma_DUP;

                if (pcma_DUP->NAME == (char *) NULL)
                    joinreftocma(pcma_DUP, pref);
            }

        /* Transfert des LINKs. */
        for (plink = pcma_DUP->LINK; plink != (phcma_link_list *) NULL;
             plink = plink->NEXT)
            joinlinktocma(pcma_DUP, plink->PHCMA,
                          plink->PHFIG,
                          plink->XFLAT,
                          plink->YFLAT,
                          plink->ATRANSF);
    }


    return ((void *) pcma_DUP);
} /* end of dup_ls_cma */

/*---------------------------------------------------------\
                     extract_cmalink
\---------------------------------------------------------*/
static phcma_list *extract_cmalink(ptfig, xref, yref)
phfig_list     *ptfig;
long            xref;
long            yref;
{
#define  access_ls_cma  ( (phcma_list*)(ptuser->DATA) )

    ptype_list     *ptuser;
    phseg_list     *pseg;

    /* Recuperation ou creation de la liste des CMAs. */
    ptuser = get_ls_data(ptfig, C_SUPP_UCMA, F_SUPP_UCREAT);

    /* Variable interne au module servant en cas d'extraction. */
    ptfig_EXTRACT = ptfig;

    /* Recherche d'un segment ALU1 en contact au point (xref,yref). */
    for (pseg = ptfig->PHSEG; pseg != (phseg_list *) NULL;
         pseg = pseg->NEXT)
        if ((pseg->LAYER != ALU1)
            || (!isconnect(pseg->X1, pseg->Y1,
                           pseg->X2, pseg->Y2,
                           xref, yref,
                           xref, yref)))
            /* Segment non connecte. */
            continue;
        else
        {
            /* Si le CMA du segment a deja ete extrait, on retourne le pointeur sur ce dernier ( champs USER ) */
            if (pseg->USER != (ptype_list *) NULL)
            {
                return ((phcma_list *) pseg->USER);
            }
            else
            {
                /* Exctraction de l'equipotentielle associe au segment ( il n'existe pas deja ). */
                access_ls_cma = addphcma(access_ls_cma);
                extract_cma(access_ls_cma, pseg);

                return (access_ls_cma);
            }
        }
#undef  access_ls_cma
    return NULL;
} /* end of extract_cmalink */

/*---------------------------------------------------------\
                     link_ls_cma
\---------------------------------------------------------*/
long            link_ls_cma(ptfig_father, ptfig_son, xflat, yflat, atransf)
phfig_list     *ptfig_father;
phfig_list     *ptfig_son;
long            xflat;
long            yflat;
char            atransf;
{
#define  access_ref  ( (phref_list*)(ppref->DATA) )

    ptype_list     *ptuser_father, *ppref;
    phcma_list     *pcma, *plink;
    long            NB_reflink;


    NB_reflink = 0;

    /* Recuperation des equipotentielles du pere. */
    if ((ptuser_father = get_ls_data(ptfig_father, C_SUPP_UCMA, FALSE))
                == (ptype_list *) NULL)
        return (0);

    /* Boucle sur tous les CMA. */
    for (pcma = (phcma_list *) ptuser_father->DATA;
         pcma != (phcma_list *) NULL;
         pcma = pcma->NEXT)
    {
        /* Boucle sur la liste des references du CMA pour associer a chaque "ref_con" du pere le "ref_con" du fils, et par deduction, le CMA du fils. */
        for (ppref = pcma->PPHREF; ppref != (ptype_list *) NULL;
             ppref = ppref->NEXT)
        {
            /* Reference du pere dans l'AB du fils. */
            if ((access_ref->XREF > ptfig_son->XAB2)
                || (access_ref->XREF < ptfig_son->XAB1)
                || (access_ref->YREF > ptfig_son->YAB2)
                || (access_ref->YREF < ptfig_son->YAB1))
            {
                continue;
            }

            /* Tentative d'e0xtraction du CMA du fils associe au point ( XREF, YREF ). */
            if ((plink = extract_cmalink(ptfig_son, access_ref->XREF, access_ref->YREF))
                        == (phcma_list *) NULL)
            {
                fflush( stdout );
                fprintf( stderr, "*** strecher warning *** : equi. not found in %s.\n", ptfig_son->NAME );
            }
            else
            {
                joinlinktocma(pcma, plink, ptfig_son, xflat, yflat, atransf);
                NB_reflink += 1;
            }
        }
    }
    return (NB_reflink);

#undef  access_ref
} /* end of link_ls_cma */

/*---------------------------------------------------------\
                     extract_ls_cmaref
\---------------------------------------------------------*/
void            extract_ls_cmaref(ptfig, a_ptfig_supply)
phfig_list     *ptfig;
supply_list    *a_ptfig_supply;
{
#define  access_ls_cma  ( (phcma_list*)(ptuser->DATA) )

    ptype_list     *ptuser;
    phref_list     *pref;
    phseg_list      pseudo_seg;
    phcma_list     *pcma;
    char           *p_ref_ref, *p_namecma;
    long            join, index;

    /* Initialisation generale du pseudo-segment. */
    pseudo_seg.NEXT = (phseg_list *) NULL;
    pseudo_seg.USER = (ptype_list *) NULL;
    pseudo_seg.LAYER = ALU1;
    pseudo_seg.NAME = (char *) NULL;
    pseudo_seg.TYPE = UP;
    pseudo_seg.WIDTH = 0;

    /* Variable d'extraction interne au module. */
    ptfig_EXTRACT = ptfig;
    ptfig_SUPPLY = a_ptfig_supply;

    /* Recuperation ou creation de la liste des CMAs. */
    ptuser = get_ls_data(ptfig, C_SUPP_UCMA, F_SUPP_UCREAT);

    p_ref_ref = namealloc("ref_ref");

    /* Examuns de tous les "ref_con" de la figure. */
    for (pref = ptfig->PHREF; pref != (phref_list *) NULL;
         pref = pref->NEXT)
    {
        /* Si la reference n'est pas un "ref_con" ou qu'elle appartient deja a un CMA, on passe. */
        if (pref->FIGNAME == p_ref_ref)
            continue;
        if (pref->USER != (ptype_list *) NULL)
            pref->USER = NULL;

        /* Extraction du nom du CMA a partir de celui de la ref. */
        p_namecma = namealloc(getrefconname(pref->NAME, &index));

        join = FALSE;
        /* Si la reference appartient a un CMA deja existant, on lui ajoute. */
        for (pcma = access_ls_cma; pcma != (phcma_list *) NULL;
             pcma = pcma->NEXT)
            if (p_namecma == pcma->NAME)
            {
                join = TRUE;
                joinreftocma(pcma, pref);
                break;
            }

        if (join == TRUE)
            continue;

        /* La reference n'appartient a aucun CMA existant, on cree un nouveau CMA et on lui associe cette reference. */
        access_ls_cma = addphcma(access_ls_cma);
        joinreftocma(access_ls_cma, pref);

        pseudo_seg.X1 = pref->XREF;
        pseudo_seg.Y1 = pref->YREF;
        pseudo_seg.X2 = pref->XREF;
        pseudo_seg.Y2 = pref->YREF;

        extract_cma(access_ls_cma, &pseudo_seg);
    }
#undef  access_ls_cma
}
