/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_support.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  <stdlib.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_support.h"
#include  "strecher_vrdrules.h"

/*---------------------------------------------------------\
                     defines
\---------------------------------------------------------*/
#define  VRT           5
#define  HRZ           6
#define  HRZ_HRZ      (HRZ*2+HRZ)
#define  HRZ_VRT      (HRZ*2+VRT)
#define  VRT_VRT      (VRT*2+VRT)
#define  VRT_HRZ      (VRT*2+HRZ)

/*---------------------------------------------------------\
                     globals
\---------------------------------------------------------*/
long            SX_WIDTH_VSS = 1000, SX_WIDTH_VDD = 1000, SX_WIDTH_TALU1 = 1000, SX_WIDTH_TALU2 = 1000, SX_DIST_ALU1 = 1000, SX_DIST_ALU2 = 1000, SX_SIDE_VIA1 = 1000, SX_SIDE_VIA2 = 1000, SX_SIDE_DIFF = 1000, SX_PITCH_TALU2 = 1000, SX_PITCH_TALU1 = 1000, SX_XTRANS = 1000;

/*---------------------------------------------------------\
                     locals
\---------------------------------------------------------*/
static void    *(*TabFcUserDup[MAX_USER_TDUPL]) ();
static long     TabTpUserDup[MAX_USER_TDUPL], NB_user_dup = 0;
static phfig_list *ptfig_SRC = (phfig_list *) NULL, *ptfig_DEST = (phfig_list *) NULL;

/*---------------------------------------------------------\
                     set_vrd_rules
\---------------------------------------------------------*/
void            set_vrd_rules()
{
    SX_WIDTH_VSS = mc_lambdatosx(WIDTH_VSS);
    SX_WIDTH_VDD = mc_lambdatosx(WIDTH_VDD);
    SX_WIDTH_TALU1 = mc_lambdatosx(WIDTH_TALU1);
    SX_WIDTH_TALU2 = mc_lambdatosx(WIDTH_TALU2);
    SX_DIST_ALU1 = mc_lambdatosx(DIST_ALU1);
    SX_DIST_ALU2 = mc_lambdatosx(DIST_ALU2);
    SX_SIDE_VIA1 = mc_lambdatosx(SIDE_VIA1);
    SX_SIDE_VIA2 = mc_lambdatosx(SIDE_VIA2);
    SX_SIDE_DIFF = mc_lambdatosx(SIDE_DIFF);

    SX_PITCH_TALU2 = mc_lambdatosx(PITCH_TALU2);
    SX_PITCH_TALU1 = mc_lambdatosx(PITCH_TALU2);
    SX_XTRANS = mc_lambdatosx(XTRANS_MIN);
}

/*---------------------------------------------------------\
                     isconnect
\---------------------------------------------------------*/
long isconnect(XA1, YA1, XA2, YA2, XB1, YB1, XB2, YB2)
long XA1, YA1, XA2, YA2, XB1, YB1, XB2, YB2;
{
    long adir, bdir;

    adir = (YA1 == YA2) ? HRZ : VRT;
    bdir = (YB1 == YB2) ? HRZ : VRT;

    switch (adir * 2 + bdir)
    {
    case HRZ_HRZ:
        return ((YA1 == YB1)
                && (((XA1 >= XB1) && (XA1 <= XB2))
                    || ((XA2 >= XB1) && (XA2 <= XB2))));
    case HRZ_VRT:
        return ((XB1 >= XA1) && (XB1 <= XA2)
                && (YA1 >= YB1) && (YA1 <= YB2));
    case VRT_VRT:
        return ((XA1 == XB1)
                && (((YA1 >= YB1) && (YA1 <= YB2))
                    || ((YA2 >= YB1) && (YA2 <= YB2))));
    case VRT_HRZ:
        return ((XA1 >= XB1) && (XA1 <= XB2)
                && (YB1 >= YA1) && (YB1 <= YA2));
    }
    return 0L;
} /* end of set_vrd_rules */

/*---------------------------------------------------------\
                     getrefconname
\---------------------------------------------------------*/
char           *getrefconname(ptname, p_index)
char           *ptname;
long           *p_index;
{
    static char     name[MBK_SZ_NAME];
    char           *pcar, *psepar;

    strcpy(name, ptname);
    for (psepar = (char *) NULL, pcar = name; *pcar != (char) 0; pcar++)
        if (*pcar == '_')
            psepar = pcar;

    if (psepar == (char *) NULL)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher error *** : reference name incorrect %s.\n", ptname );
        exit( 1 );
    }

    *psepar = (char) 0;
    *p_index = -1;
    sscanf(psepar + 1, "%ld", p_index);

    if (*p_index == -1)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher error *** : reference name incorrect %s.\n", ptname );
        exit( 1 );
    }

    return (name);
}


/*---------------------------------------------------------\
                     addoptype
\---------------------------------------------------------*/
ptype_list     *addoptype(pthead, otype, ptdata)
ptype_list     *pthead;
long            otype;
void           *ptdata;
{
    ptype_list     *pHEAD, **pPREV;


    pHEAD = pthead;
    for (pPREV = &pHEAD; (*pPREV) != (ptype_list *) NULL;
         pPREV = &(*pPREV)->NEXT)
        if ((*pPREV)->TYPE >= otype)
        {
            *pPREV = addptype(*pPREV, otype, ptdata);

            return (pHEAD);
        }
    (*pPREV) = addptype(*pPREV, otype, ptdata);

    return (pHEAD);
} /* end of addoptype */

/*---------------------------------------------------------\
                     get_ls_data
\---------------------------------------------------------*/
ptype_list     *get_ls_data(ptfig, a_type, a_flags)
phfig_list     *ptfig;
long            a_type;
long            a_flags;
{
    ptype_list     *pt;

    /* Recherche de l'element. */
    if ((pt = getptype(ptfig->USER, a_type)) != (ptype_list *) NULL)
    {
        return (pt);
    }

    /* Element inexistant : creation si demande. */
    if (a_flags & F_SUPP_UCREAT)
    {
        ptfig->USER = addptype(ptfig->USER, a_type, (void *) NULL);

        return (ptfig->USER);
    }
    else
        return ((ptype_list *) NULL);
}


/*---------------------------------------------------------\
                     del_ls_data
\---------------------------------------------------------*/
ptype_list     *del_ls_data(ptfig, a_type, a_flags)
phfig_list     *ptfig;
long            a_type;
long            a_flags;
{
    /* Destruction de l'element. */
    return (ptfig->USER = delptype(ptfig->USER, a_type));
} /* end of del_ls_data */

/*---------------------------------------------------------\
                     set_user_dup
\---------------------------------------------------------*/
long            set_user_dup(a_type, funct)
long            a_type;
void           *(*funct) ();
{

    if (NB_user_dup < MAX_USER_TDUPL)
    {
        TabFcUserDup[NB_user_dup] = funct;
        TabTpUserDup[NB_user_dup++] = a_type;
    }

    return (NB_user_dup);
}

/*---------------------------------------------------------\
                     dupuser
\---------------------------------------------------------*/
void            dupuser(a_ptfig_src, a_ptfig_dest)
phfig_list     *a_ptfig_src, *a_ptfig_dest;
{
    ptype_list     *ptuser;
    long            i;

    for (ptuser = a_ptfig_src->USER; ptuser != (ptype_list *) NULL; ptuser = ptuser->NEXT)
        for (i = 0; i < NB_user_dup; i++)
            if (ptuser->TYPE == TabTpUserDup[i])
                a_ptfig_dest->USER = addptype(a_ptfig_dest->USER,
                                              ptuser->TYPE,
                                              (TabFcUserDup[i]) (a_ptfig_src,
                                                                 a_ptfig_dest));
}

/*---------------------------------------------------------\
                     dup_phcon
\---------------------------------------------------------*/
static void     dup_phcon()
{
    phcon_list     *pcon, *newpcon;

    for (pcon = ptfig_SRC->PHCON; pcon != (phcon_list *) NULL; pcon = pcon->NEXT)
    {
        newpcon = addphcon(ptfig_DEST, pcon->ORIENT,
                           pcon->NAME,
                           pcon->XCON,
                           pcon->YCON,
                           pcon->LAYER,
                           pcon->WIDTH);
/*         newpcon->USER = pcon->USER; */
        newpcon->USER = NULL;
    }
}

/*---------------------------------------------------------\
                     dup_phvia
\---------------------------------------------------------*/
static void     dup_phvia()
{
    phvia_list     *pvia, *newpvia;

    for (pvia = ptfig_SRC->PHVIA; pvia != (phvia_list *) NULL; pvia = pvia->NEXT)
    {
        newpvia = addphvia(ptfig_DEST, pvia->TYPE,
                           pvia->XVIA,
                           pvia->YVIA);
/*         newpvia->USER = pvia->USER; */
        newpvia->USER = NULL;
    }
}

/*---------------------------------------------------------\
                     dup_phref
\---------------------------------------------------------*/
static void     dup_phref()
{
    phref_list     *pref, *newpref;

    for (pref = ptfig_SRC->PHREF; pref != (phref_list *) NULL; pref = pref->NEXT)
    {
        newpref = addphref(ptfig_DEST, pref->FIGNAME,
                           pref->NAME,
                           pref->XREF,
                           pref->YREF);
/*         newpref->USER = pref->USER; */
        newpref->USER = NULL;
    }
}

/*---------------------------------------------------------\
                     dup_phins
\---------------------------------------------------------*/
static void     dup_phins()
{
    phins_list     *pins, *newpins;

    for (pins = ptfig_SRC->PHINS; pins != (phins_list *) NULL; pins = pins->NEXT)
    {
        newpins = addphins(ptfig_DEST, pins->FIGNAME,
                           pins->INSNAME,
                           pins->TRANSF,
                           pins->XINS,
                           pins->YINS);
/*         newpins->USER = pins->USER; */
        newpins->USER = NULL;
    }
}

/*---------------------------------------------------------\
                     dup_phseg
\---------------------------------------------------------*/
static void     dup_phseg()
{
    phseg_list     *pseg, *newpseg;

    for (pseg = ptfig_SRC->PHSEG; pseg != (phseg_list *) NULL; pseg = pseg->NEXT)
    {
        newpseg = addphseg(ptfig_DEST, pseg->LAYER,
                           pseg->WIDTH,
                           pseg->X1,
                           pseg->Y1,
                           pseg->X2,
                           pseg->Y2,
                           pseg->NAME);
/*         newpseg->USER = pseg->USER; */
        newpseg->USER = NULL;
    }
}

/*---------------------------------------------------------\
                     dup_ab
\---------------------------------------------------------*/
static void     dup_ab()
{
    defab(ptfig_DEST, ptfig_SRC->XAB1,
          ptfig_SRC->YAB1,
          ptfig_SRC->XAB2,
          ptfig_SRC->YAB2);
}

/*---------------------------------------------------------\
                     dupphfig
\---------------------------------------------------------*/
void            dupphfig(a_ptfig_SRC, a_ptfig_DEST)
phfig_list     *a_ptfig_SRC, *a_ptfig_DEST;
{

    if (a_ptfig_SRC == a_ptfig_DEST)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher bugs *** : Unable to dupplicate same figure, i'am sorry.\n" );
        exit( 1 );
    }
    ptfig_SRC = a_ptfig_SRC;
    ptfig_DEST = a_ptfig_DEST;

    dup_ab();
    dup_phcon();
    dup_phseg();
    dup_phvia();
    dup_phref();
    dup_phins();
}
