/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_upper.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  <string.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_intv.h"
#include  "strecher_support.h"
#include  "strecher_allowed.h"
#include  "strecher_supply.h"
#include  "strecher_phcut.h"
#include  "strecher_dtdupl.h"
#include  "strecher_phcma.h"
#include  "strecher_hierarch.h"
#include  "strecher_upper.h"
#include  "strecher_phtalu.h"

/*---------------------------------------------------------\
                     make_work
\---------------------------------------------------------*/
intv_list      *make_work(ptwork, flags, must_cross)
dtdupl_list    *ptwork;
long            flags;
int             must_cross;
{
    if (flags & F_UPPER_PHCUT)
    {
        /* Chargement hierarchique de la figure depuis le disque. Creation des gabarits de coupe. */
        get_hr_phfig(ptwork);
        extract_hr_allowed(ptwork, must_cross);
        return (intv_list *) get_ls_data(ptwork->PHFIG, C_SUPP_UALLOWED, FALSE)->DATA;
    }

    /* Detail des operations :
    I   - Streching de la liste des duplicatas.
    II  - Extraction des CMA. ( sur les cellules feuilles )
    III - Extraction des CMA du bloc. ( deux niveaux hierarchiques )
    IV  - Creation des gabarits de transparences.
    V   - Retaillage hierarchique des gabarits de transparence du bloc.
    VI  - Addition des transparences aux cellules feuilles.
    VII - Exportation des CMA depuis les cellules feuilles jusque dans l'interface du bloc. */

    if (flags & F_UPPER_PHDUP)
    {
        alloc_hr_dtdupl(ptwork);

        strech_hr_dtdupl(ptwork, must_cross);
        extract_hr_cma(ptwork);
        alloc_hr_talu(ptwork);
        modify_ln_talu();
        modify_hr_talu(ptwork);
        add_hr_through(ptwork);
        clean_ln_through();

        /* Ecriture sur disque du bloc et des modeles duplique. */
        save_ln_dtdupl();
    }
    return (intv_list *) NULL;
}

/*---------------------------------------------------------\
                     global_init
\---------------------------------------------------------*/
void            global_init()
{
    set_vrd_rules();

    set_user_dup(C_SUPP_UCMA, dup_ls_cma);
    set_user_dup(C_SUPP_USUPPLY, dup_ls_supply);
    set_user_dup(C_SUPP_ULASTPOS, dupflatpos);
    set_user_dup(C_SUPP_UMODLPOS, dupmodlpos);

}
