/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*------------------------------------------------------------\
|                                                             |
| Tool    :                    DRUC                           |
|                                                             |
| File    :                  DRUCGRAAL.C                      |
|                                                             |
| Authors :                Renaud Patrick                     |
|                                                             |
| Date    :                   11/01/94                        |
|                                                             |
\------------------------------------------------------------*/

/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include MPH_H
#include MUT_H
#include RDS_H
#include RUT_H
#include RTL_H
#include RWI_H
#include RFM_H
#include RPR_H

#include "drucutil.h"
#include "vmctools.h"
#include "vmcerror.h"
#include "druchier.h"
#include "vmcasmld.h"
#include "vmcaract.h"
#include "defdefin.h"
#include "vmcrelat.h"
#include "vmcunify.h"
#include "drucmin.h"

/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

extern rdsfig_list            *DrucFigureRdsErreur;
extern DrucTypeStructRdsLayer  DrucStructRdsLayer[];

/*------------------------------------------------------------\
|                                                             |
|                          Fonctions                          |
|                                                             |
\------------------------------------------------------------*/

/*------------------------------------------------------------\
|                                                             |
|                         drucgeterror                        |
|                                                             |
\------------------------------------------------------------*/

char *drucgeterror( Rule 
                  )
      long  Rule;

{

  return ( DrucGetRuleComment ( Rule
                              )
         );
}

/*------------------------------------------------------------\
|                                                             |
|                             drucrdsfig                      |
|                                                             |
\------------------------------------------------------------*/
rdsfig_list *drucrdsfig ( DrucFigureRds,
                          DrucFigureRdsErrorName
                        )
    rdsfig_list *DrucFigureRds;
    char        *DrucFigureRdsErrorName;

{

rdswindow                 *DrucRdsWindow;
static DrucTypeStructInstruction *DrucTabInstruction = ( DrucTypeStructInstruction *)NULL;
int                        LayerCourant;
int                        IndexLayer;

  DrucStatInit          ( DRUC_FLATTEN,
                          DRUC_SILENT,
                          DRUC_SILENT,
                          DRUC_NO_ERROR_FILE
                        );

  DrucFigureRdsErreur    = addrdsfig ( DrucFigureRdsErrorName,
                                       MACRO_PLUS_TAILLE
                                     );

/*
** Modified by L.Jacomme the 14/10/95
*/
  if ( DrucTabInstruction == ( DrucTypeStructInstruction * )NULL )
  {
    DrucInitInstruction ( & DrucTabInstruction
                        );
  }
/*
** End
*/

  DrucRdsWindow          = buildrdswindow ( DrucFigureRds
                                          );

  for ( IndexLayer = 0;
        IndexLayer < RDS_MAX_LAYER -1;
        IndexLayer ++
      )
  {
    LayerCourant = DrucStructRdsLayer [ IndexLayer ].RDS_INDEX ;

    if ( LayerCourant != DRUC_UNDEFINED_LAYER
       )
    {
      if ( DrucFigureRds->LAYERTAB [ LayerCourant ] != ( rdsrec_list * ) NULL
         )
      {
        DrucComputeEquiLayer ( DrucFigureRds ,
                               LayerCourant
                             );
      }
    }
  }

  destroyrdswindow ( DrucFigureRds,
                     DrucRdsWindow
                   );
/*\
 * le parametre DrucRdsWindow etait ooublie dans l'appel de DrucUnify
 * thnaks to Xavier
\*/
  DrucUnify        ( DrucFigureRds,
                     DrucRdsWindow
                   );

  DrucRdsWindow    = buildrdswindow ( DrucFigureRds
                                    );
  for ( IndexLayer = 0;
        IndexLayer < RDS_MAX_LAYER -1;
        IndexLayer ++
      )
  {
    LayerCourant = DrucStructRdsLayer [ IndexLayer ].RDS_INDEX ;

    if ( LayerCourant != DRUC_UNDEFINED_LAYER
       )
    {
      if ( DrucFigureRds->LAYERTAB [ LayerCourant ] != ( rdsrec_list * ) NULL
         )
      {
        DrucComputeEquiLayer ( DrucFigureRds ,
                               LayerCourant
                             );
      }
    }
  }

  DrucMin          ( DrucFigureRds,
                     DrucRdsWindow,
                     DrucTabInstruction ,
                     DrucNombreInstructions
                   );
  destroyrdswindow ( DrucFigureRds,
                     DrucRdsWindow
                   );
  delrdsfig        ( DrucFigureRds->NAME
                   );
  return           ( DrucFigureRdsErreur
                   );
}


