/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: ScLibGN.c,v 1.10 1997/10/30 13:27:56 fitpath Exp fitpath $        |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :        cao-vlsi@masi.ibp.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module    :        "ScLibGN.c"                                    | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# include  <stdio.h>
# include  <string.h>
# include  <time.h>

# include  <genlib.h>

# include  "GenericGN.h"
# include  "SharedGN.h"


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                        Internal Constants                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

# define       IC_GfpGN_CellNBInv            4
# define       IC_GfpGN_CellNBBuff           3
# define       IC_GfpGN_CellNBNand2          2
# define       IC_GfpGN_CellNBNand3          2
# define       IC_GfpGN_CellNBAnd2           2
# define       IC_GfpGN_CellNBAnd3           2
# define       IC_GfpGN_CellNBNor2           2
# define       IC_GfpGN_CellNBNor3           2
# define       IC_GfpGN_CellNBOr2            2
# define       IC_GfpGN_CellNBOr3            2
# define       IC_GfpGN_CellNBXnor2          1
# define       IC_GfpGN_CellNBXor2           1

/* ***** debut non_modifie ***** */
# define       IC_GfpGN_CellNBNand2mask      3
# define       IC_GfpGN_CellNBNor2mask       3
# define       IC_GfpGN_CellNBXnor2mask      3
# define       IC_GfpGN_CellNBConst          2
# define       IC_GfpGN_CellNBNbuse          2
# define       IC_GfpGN_CellNBBuse           2
# define       IC_GfpGN_CellNBNmux2cs        2
# define       IC_GfpGN_CellNBMux2cs         2
# define       IC_GfpGN_CellNBMux3cd         2
# define       IC_GfpGN_CellNBMux3cs         2
# define       IC_GfpGN_CellNBMux4cs         2
# define       IC_GfpGN_CellNBMux5cs         2
# define       IC_GfpGN_CellNBMux6cs         2
# define       IC_GfpGN_CellNBAdd2s         (7+1)
# define       IC_GfpGN_CellNBNpl1           2
# define       IC_GfpGN_CellNBPl1            2
# define       IC_GfpGN_CellNBPmemnw         3
# define       IC_GfpGN_CellNBPmemnwq        3
# define       IC_GfpGN_CellNBPmem           3
# define       IC_GfpGN_CellNBPmemr          3
# define       IC_GfpGN_CellNBPmemq          3
# define       IC_GfpGN_CellNBPmemt          8
# define       IC_GfpGN_CellNBPmemrt         8
# define       IC_GfpGN_CellNBPmemtq         8
/* ***** fin non_modifie ***** */


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |               Internal Structures and Variables                    | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/* Name of the leaf cells belonging to all generators. */
static char *TablLeafCell[] =
    {
/* drive = 1 ,       2 ,        3,      4 ,      5 ,      6 */
    /* Boolean generators leaf cells. */
         "n1",    "np1",  "undef",  "ndrv",
      "undef",     "p1",  "undef",
        "na2",   "na2p",
        "na3",   "na3p",
         "a2",    "a2p",
         "a3",    "a3p",
        "no2",   "nop2",
        "no3",   "nop3",
         "o2",    "op2",
         "o3",    "op3",
       "nxr2",
        "xr2",
    /* ***** debut non_modifie ***** */
    /* Mask generators leaf cells. */
      "bna2x",   "na2x", "n1na2x",
      "bno2x", "n1no2x",   "no2x",
     "bnxr2x", "n1xr2x",  "nxr2x",
    /* Buse generators leaf cells. */
       "bnt1",    "nt1",
        "bt1",     "t1",
    /* Multiplexer generators leaf cells. */
    "nmx2sb0", "nmx2sc",
     "mx2sb0",  "mx2sc",
     "mx3db0",  "mx3dc",
     "mx3sb0",  "mx3sc",
     "mx4sb0",  "mx4sc",
     "mx5sb0",  "mx5sc",
     "mx6sb0",  "mx6sc",
    /* Adder generators leaf cells. */
   "nobuffer",   "bcsa",
       "csag",   "csa0",   "csap",  "csai", "csap0", "csai0",
    /* Register generators leaf cells. */
      "pl1b0", "npl1a0",
      "pl1b0",  "pl1a0",
      "pmsb0", "pdffb0",  "pmsa0",
      "pmsb0", "pdffb0",  "pmsa1",
      "pmsb2", "pdffb2",  "pmsa0",
      "pmsb2", "pdffb2",  "pmsa1",
      "pmsba", "pdffba",  "pmsa9",
      "pmsb6", "pdffb6",
      "pmsg4",  "pmsh4",  "pmsp4", "pmsi4", "pmsq4", "pmsj4",
      "pmsb6", "pdffb6",
      "pmsg5",  "pmsh5",  "pmsp5", "pmsi5", "pmsq5", "pmsj5",
      "pmsbe", "pdffbe",
      "pmsgd",  "pmshd",  "pmspd", "pmsid", "pmsqd", "pmsjd",
    /* Constant generators leaf cells. */
       "zero",    "one"
    /* ***** fin non_modifie ***** */
    };


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                  Internal Functions Declarations                   | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

static char *FunctionModelData __FP(( long aSlice ));
static char *FunctionModelBuff __FP(( long aSlice ));


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                     Functions Definitions                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ScGN_FunctionArgs()"                                | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Get all the arguments from the command line, according to the function.|%
 * %| Also performs checking to ensure correctness and coherency of these    |%
 * %| parameters.                                                            |%
 * %|                                                                        |%
 */

extern void  ScGN_FunctionArgs( ArgList )
  va_list  ArgList;
{
  long  ArgFlags;

  switch(GV.aFunction) {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
      ArgFlags =  IF_GfpGN_WidthOp | IF_GfpGN_Slice | IF_GfpGN_Drive; break;
    /* ***** debut non_modifie ***** */
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
      ArgFlags = IF_GfpGN_WidthOp | IF_GfpGN_Slice | IF_GfpGN_Mask; break;
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      ArgFlags = IF_GfpGN_WidthOp | IF_GfpGN_Slice; break;
    /* ***** fin non_modifie ***** */
    default:
      ArgFlags = 0L;
  } /* End of switch : select arguments to be read. */
  GV.aWidthDP = va_arg( ArgList, long );
  /* Get and check the data-path wide. */
  if (   (GV.aWidthDP < C_GfpGN_LimitsMinWidthDP)
      || (GV.aWidthDP > C_GfpGN_LimitsMaxWidthDP)) {
    GfpUT_PrintError( GV.ModuleId, IE_GfpGN_WidthDP,
                      GV.aWidthDP, C_GfpGN_LimitsMinWidthDP - 1,
                                   C_GfpGN_LimitsMaxWidthDP + 1 );
  }
  /* Get and check the bus wide. */
  if (ArgFlags & IF_GfpGN_WidthOp) {
    GV.aWidthOp = va_arg(ArgList, long);
    if (   (GV.aWidthOp < 1)
        || (GV.aWidthOp > GV.aWidthDP))
      GfpUT_PrintError(
        GV.ModuleId, IE_GfpGN_WIDTH, ShGN_FunctionStr(), GV.aWidthOp);
  }
  else
    GV.aWidthOp = GV.aWidthDP;
  /* Get and check the LSB starting slice. */
  if (ArgFlags & IF_GfpGN_Slice) {
    GV.aSlice = va_arg(ArgList, long);
    if (   (GV.aSlice             < 0)
        || (GV.aSlice+GV.aWidthOp > GV.aWidthDP))
      GfpUT_PrintError(
        GV.ModuleId, IE_GfpGN_SLICE, ShGN_FunctionStr(), GV.aSlice);
  }
  else
    GV.aSlice = 0L;
  /* Get and check the output drive capacity. */
  if (ArgFlags & IF_GfpGN_Drive) {
    GV.aDrive = va_arg(ArgList, long);
    if (   (GV.aDrive < 1)
        || (GV.aDrive > ShGN_FunctionDrive()))
      GfpUT_PrintError(
        GV.ModuleId, IE_GfpGN_DRIVE, ShGN_FunctionStr(), GV.aDrive);
  }
  else
    GV.aDrive = 1L;
  /* ***** debut non_modifie ***** */
  /* Get the Mask (extended long pointer). */
  if (ArgFlags & IF_GfpGN_Mask)
    GV.aXlMask = va_arg(ArgList, long*);
  else
    GV.aXlMask = (long*)NULL;
  /* Get the bus index convention. */
  GV.aMSB0    = va_arg(ArgList, long);
  /* Get all the views asked. */
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewVHDL    : 0L;
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewNetlist : 0L;
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewLayout  : 0L;
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewIcon    : 0L;
  GV.aPower   = va_arg(ArgList, long);
  /* Check the operator's width. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionUnknow:
      break;
    case C_GfpGN_FunctionNul:
      if (   (GV.aWidthOp <  4)
          || (GV.aWidthOp > 32)) {
		GfpUT_PrintError(
          GV.ModuleId, IE_GfpGN_WidthOp,
          ShGN_FunctionStr(), GV.aWidthOp, 4 - 1, 32 + 1 );
      }
      if (GV.aSlice % 2) {
		GfpUT_PrintError(
          GV.ModuleId, IE_GfpGN_SliceEven, ShGN_FunctionStr(), GV.aSlice );
      }
      break;
  } /* End of switch : operator's width. */
  /* ***** fin non_modifie ***** */
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ScGN_FunctionCells()"                               | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Return the offset of the first operator leaf cell, in the              |%
 * %| "TablLeafCell" table.                                                  |%
 * %|                                                                        |%
 */

extern long  ScGN_FunctionCells()
{
  long  LeafCellIndex;

  LeafCellIndex = 0;
  switch(GV.aFunction) {
    /* ***** debut non_modifie ***** */
    case C_GfpGN_FunctionNul:       LeafCellIndex += IC_GfpGN_CellNBConst;
    case C_GfpGN_FunctionConst:     LeafCellIndex += IC_GfpGN_CellNBPmemrt;
    case C_GfpGN_FunctionPdffrt:    LeafCellIndex += IC_GfpGN_CellNBPmemt;
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdfft:     LeafCellIndex += IC_GfpGN_CellNBPmemtq;
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:    LeafCellIndex += IC_GfpGN_CellNBPmemr;
    case C_GfpGN_FunctionPdffr:     LeafCellIndex += IC_GfpGN_CellNBPmem;
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPdff:      LeafCellIndex += IC_GfpGN_CellNBPmemq;
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffq:     LeafCellIndex += IC_GfpGN_CellNBPmemnw;
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPdffnw:    LeafCellIndex += IC_GfpGN_CellNBPmemnwq;
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPdffnwq:   LeafCellIndex += IC_GfpGN_CellNBPl1;
    case C_GfpGN_FunctionPl1:       LeafCellIndex += IC_GfpGN_CellNBNpl1;
    case C_GfpGN_FunctionNpl1:      LeafCellIndex += IC_GfpGN_CellNBAdd2s;
    case C_GfpGN_FunctionAdd2s:     LeafCellIndex += IC_GfpGN_CellNBMux6cs;
    case C_GfpGN_FunctionMux6cs:    LeafCellIndex += IC_GfpGN_CellNBMux5cs;
    case C_GfpGN_FunctionMux5cs:    LeafCellIndex += IC_GfpGN_CellNBMux4cs;
    case C_GfpGN_FunctionMux4cs:    LeafCellIndex += IC_GfpGN_CellNBMux3cs;
    case C_GfpGN_FunctionMux3cs:    LeafCellIndex += IC_GfpGN_CellNBMux3cd;
    case C_GfpGN_FunctionMux3cd:    LeafCellIndex += IC_GfpGN_CellNBMux2cs;
    case C_GfpGN_FunctionMux2cs:    LeafCellIndex += IC_GfpGN_CellNBNmux2cs;
    case C_GfpGN_FunctionNmux2cs:   LeafCellIndex += IC_GfpGN_CellNBBuse;
    case C_GfpGN_FunctionBuse:      LeafCellIndex += IC_GfpGN_CellNBNbuse;
    case C_GfpGN_FunctionNbuse:     LeafCellIndex += IC_GfpGN_CellNBXnor2mask;
    case C_GfpGN_FunctionXnor2mask: LeafCellIndex += IC_GfpGN_CellNBNor2mask;
    case C_GfpGN_FunctionNor2mask:  LeafCellIndex += IC_GfpGN_CellNBNand2mask;
    case C_GfpGN_FunctionNand2mask: LeafCellIndex += IC_GfpGN_CellNBXnor2;
    /* ***** fin non_modifie ***** */
    case C_GfpGN_FunctionXor2:      LeafCellIndex += IC_GfpGN_CellNBXor2;
    case C_GfpGN_FunctionXnor2:     LeafCellIndex += IC_GfpGN_CellNBOr3;
    case C_GfpGN_FunctionOr3:       LeafCellIndex += IC_GfpGN_CellNBOr2;
    case C_GfpGN_FunctionOr2:       LeafCellIndex += IC_GfpGN_CellNBNor3;
    case C_GfpGN_FunctionNor3:      LeafCellIndex += IC_GfpGN_CellNBNor2;
    case C_GfpGN_FunctionNor2:      LeafCellIndex += IC_GfpGN_CellNBAnd3;
    case C_GfpGN_FunctionAnd3:      LeafCellIndex += IC_GfpGN_CellNBAnd2;
    case C_GfpGN_FunctionAnd2:      LeafCellIndex += IC_GfpGN_CellNBNand3;
    case C_GfpGN_FunctionNand3:     LeafCellIndex += IC_GfpGN_CellNBNand2;
    case C_GfpGN_FunctionNand2:     LeafCellIndex += IC_GfpGN_CellNBBuff;
    case C_GfpGN_FunctionBuff:      LeafCellIndex += IC_GfpGN_CellNBInv;
    case C_GfpGN_FunctionInv:     ;
  }
  return( GV.LeafCellIndex = LeafCellIndex );
}


/*  Function  :  'FunctionModelData'                                        */
/* ------------------------------------------------------------------------ */
/*  Compute the model name of a data leaf cell, according to his slice index.
 */

static char *FunctionModelData( aSlice )
  long  aSlice;
{
  char *LeafCellModelName;


  /* Select the model name. */
  switch( GV.aFunction )
  {
/* ***** debut non_modifie ***** */
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;
/* ***** fin non_modifie ***** */

    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + GV.aDrive - 1 ];
      break;

/* ***** debut non_modifie ***** */
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      LeafCellModelName = TablLeafCell[
                            GV.LeafCellIndex + 1 + GfpUT_XlGetBit(
                                                     GV.aXlMask,
                                                     aSlice - GV.aSlice ) ];
      break;

    case C_GfpGN_FunctionConst:

      LeafCellModelName = TablLeafCell[
                            GV.LeafCellIndex + GfpUT_XlGetBit(
                                                 GV.aXlMask,
                                                 aSlice - GV.aSlice ) ];
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 2 ];
      break;

    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      if (aSlice == GV.aSlice)
        LeafCellModelName = TablLeafCell[
                              GV.LeafCellIndex + 6
                              + (GV.aWidthDP - aSlice) % 2 ];
      else
      {
        if (aSlice == GV.aSlice + GV.aWidthOp - 1)
          LeafCellModelName = TablLeafCell[
                                GV.LeafCellIndex + 2
                                + (GV.aWidthDP - aSlice) % 2 ];
        else
          LeafCellModelName = TablLeafCell[
                                GV.LeafCellIndex+4
                                + (GV.aWidthDP - aSlice) % 2 ];
      }
      break;
/* ***** fin non_modifie ***** */

    default:

      return( (char*)NULL );
  }

  /* undefined leaf cell name due to unknown drive value */
  if (!strcmp(LeafCellModelName,"undef"))
    return( (char*)NULL );

  return( LeafCellModelName );
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ScGN_HaveBuff()"                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern long  ScGN_HaveBuff()
{
  switch(GV.aFunction) {
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      return( TRUE );
      break;
    default:
      break;
  }
  return( FALSE );
}


/* ***** debut non_modifie ***** */
/*  Function  :  'FunctionModelBuff'                                        */
/* ------------------------------------------------------------------------ */
/*  Compute the model name of a buffer leaf cell, according to his slice
 *  index.
 */

static char *FunctionModelBuff( aSlice )
  long  aSlice;
{
  char *LeafCellModelName;


  /* Select the buffer model name.
   * (NULL if no need of a buffer)
   */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex ];
      break;

    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    default:

      LeafCellModelName = (char*)NULL;
      break;
  }

  return( LeafCellModelName );
}
/* ***** fin non_modifie ***** */


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ScGN_CreateNetlist()"                               | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Generation of the Netlist view, using GenLib.                          |%
 * %|                                                                        |%
 */

extern void  ScGN_CreateNetlist()
{
  /* A completer, en s'inspirant de `FpGN_CreateNetlist',
   * dans le module `FpLibGN.c'.
   */
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ScGN_CreateNetlistComp()"                           | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Generation of the comparator Netlist view, using GENLIB. This function |%
 * %| is separated from "CreateNetlist" because the don't have intrinsics    |%
 * %| regularities.                                                          |%
 * %|                                                                        |%
 */

extern void  ScGN_CreateNetlistComp()
{
  /* A completer, en s'inspirant de `FpGN_CreateNetlistComp',
   * dans le module `FpLibGN.c'.
   */
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ScGN_CreateLayout()"                                | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Generation of the Layout view, using GenLib.                           |%
 * %|                                                                        |%
 */

extern void  ScGN_CreateLayout()
{
  /* A completer, en s'inspirant de `FpGN_CreateLayout',
   * dans le module `FpLibGN.c'.
   */
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "ScGN_CreateLayoutComp()"                            | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Generation of the Layout view for heterogeneous operators,             |%
 * %| using GENLIB.                                                          |%
 * %|                                                                        |%
 */

extern void  ScGN_CreateLayoutComp()
{
  /* A completer, en s'inspirant de `FpGN_CreateLayoutComp',
   * dans le module `FpLibGN.c'.
   */
}
