/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtFileDPR.c,v 1.10 1997/10/30 13:27:59 fitpath Exp fitpath $ */
/*                                                                */
/*          Alliance CAD System 3.2b                               */
/*             FpGen Package 1.10                                 */
/*                                                                */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                       */
/*                                                                */
/*  Author      :          Jean-Paul CHAPUT                       */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                       */
/* ************************************************************** */
/*  Module      :      "UtFileDPR.c"                              */


/* ************************** Includes ************************** */


#include  <stdio.h>
#include  <string.h>

#include   MUT_H
#include   MLO_H

#include  "GenericGN.h"


/* ********************* Internal constants ********************* */


#define        IE_UtFDP_AddLoFigABDup             3
#define        IE_UtFDP_AddLoFigPowerDup          4
#define        IE_UtFDP_AddLoFigGlueDup           5
#define        IE_UtFDP_ParserMissToken           6
#define        IE_UtFDP_ParserExtraToken          7
#define        IE_UtFDP_ParserFunctionUnknow      8
#define        IE_UtFDP_ParserNumberUnknow        9
#define        IE_UtFDP_ParserCoherWide          10
#define        IE_UtFDP_ParserFieldUnknow        11
#define        IE_UtFDP_ParserLineLength         12

#define        IW_UtFDP_ParserOpenFile            7
#define        IW_UtFDP_ParserDefABDup            8

#define        IC_UtFDP_ParserLineLength      16384  /* 16 Ko           */
#define        IC_UtFDP_ParserFunctionUnknow      2
#define        IC_UtFDP_ParserFunctionLoCon       3
#define        IC_UtFDP_ParserFunctionDefAB       4
#define        IC_UtFDP_ParserFunctionPower       5
#define        IC_UtFDP_ParserFunctionKeep        6
#define        IC_UtFDP_ParserFunctionGlue        7
#define        IC_UtFDP_MinPowerNumb              1
#define        IC_UtFDP_MinPowerWide             10
#define        IC_UtFDP_MinGlueIter             100

#define        IS_UtFDP_ParserFunctionLoCon     "DP_LOCON"
#define        IS_UtFDP_ParserFunctionDefAB     "DP_DEFAB"
#define        IS_UtFDP_ParserFunctionPower     "DP_POWER"
#define        IS_UtFDP_ParserFunctionKeep      "DP_KEEP"
#define        IS_UtFDP_ParserFunctionGlue      "DP_GLUE"
#define        IS_UtFDP_ParserSideNorth         "NORTH"
#define        IS_UtFDP_ParserSideSouth         "SOUTH"
#define        IS_UtFDP_ParserSideEast          "EAST"
#define        IS_UtFDP_ParserSideWest          "WEST"
#define        IS_UtFDP_ParserStrDefault        "DEFAULT"


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "UtFileDPR.c" module identifier. */
	FILE *ParserFile;        /* File descriptor of the parsed file.    */
	long  ParserCurrLine;    /* Number of the current line.            */
	char  ParserLineBuffer[IC_UtFDP_ParserLineLength];
							 /* Buffer of holding the current line.    */
	} GlobVar;


/* ********************* Internal Variables ********************* */


/* ******************** Interface Functions ********************* */


/* extern         void  UtFDP_Init(void);
 * extern         void  UtFDP_AddLoConPlace(locon_list *aLoCon,
 *                                                char  aSide ,
 *                                                long  aSlice,
 *                                                long  aTrack);
 * extern         void  UtFDP_DefLoConPlace(locon_list *aLoCon,
 *                                                char  aSide ,
 *                                                long  aSlice,
 *                                                long  aTrack);
 * extern         void  UtFDP_FreeLoConPlace(locon_list *aLoCon);
 * extern         void  UtFDP_AddLoFigAB(lofig_list *aLoFig,
 *                                             long  aDY1,
 *                                             long  aDY2);
 * extern         void  UtFDP_DelLoFigAB(lofig_list *aLoFig);
 * extern         void  UtFDP_AddLoFigPower(lofig_list *aLoFig,
 *                                                long  aNumb,
 *                                                long  aWide);
 * extern         void  UtFDP_DelLoFigPower(lofig_list *aLoFig);
 * extern         void  UtFDP_AddLoFigGlue(lofig_list *aLoFig,
 *                                               long  aIter,
 *                                               long  aWide,
 *                                               long  aCPC);
 * extern  ulofig_glue *UtFDP_GetLoFigGlue(lofig_list *aLoFig);
 * extern         void  UtFDP_DelLoFigGlue(lofig_list *aLoFig);
 * extern         void  UtFDP_SetLoFigFlags(lofig_list *aLoFig, long  aFlags);
 * extern ulofig_flags *UtFDP_GetLoFigFlags(lofig_list *aLoFig);
 * extern         void  UtFDP_DelLoFigFlags(lofig_list *aLoFig);
 * extern         void  UtFDP_AddLoFigPower(lofig_list *aLoIns,
 *                                                long  aSide,
 *                                                long  aWide);
 * extern         void  UtFDP_DelLoFigPower(lofig_list *aLoFig);
 * extern         void  UtFDP_AddLoInsPower(lofig_list *aLoIns, long  aSide);
 * extern         void  UtFDP_DelLoInsPower(loins_list *aLoIns);
 * extern         void  UtFDP_LoadPlacement(char *aFigName);
 */


/* ********************* Internal Functions ********************* */


#ifndef  ANSI_PROTO
static long  FGetLine();
static long  StrToFunction();
static long  StrToSide();
static long  StrToTrack();
#else
static long  FGetLine(void);
static long  StrToFunction(char *aFunction);
static long  StrToSide(char *aSide);
static long  StrToTrack(char *aStr, long *aPtTrack);
#endif


#define  mc_Max(a,b) (((a) > (b)) ? (a) : (b))


/* ******************** Functions Definitions ******************* */


/*  Function  :  'UtFDP_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtFDP_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtFileDPR", F_GfpUT_ModuleAdd);

	/* Errors belonging to additions functions.
	 */
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_AddLoFigABDup,
	"[%s]:Figure \"%s\" already have is abutment box set.");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_AddLoFigPowerDup,
	"[%s]:Figure \"%s\" already have is vertical power supplies set.");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_AddLoFigGlueDup,
	"[%s]:Figure \"%s\" already have is glue parameters sets.");

	/* Errors belonging to parser/driver functions.
	 */
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_ParserLineLength,
	"[%s, line %ld]:Line too long.");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_ParserMissToken,
	"[%s, line %ld]:Missing token \'%s\'.");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_ParserExtraToken,
	"[%s, line %ld]:Unreacheable token \"%s\".");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_ParserFunctionUnknow,
	"[%s, line %ld]:Invalid function \"%s\".");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_ParserNumberUnknow,
	"[%s, line %ld]:Field \'%s\':Non-numerical value (\"%s\").");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_ParserCoherWide,
"[%s, line %ld]:For terminal \"%s\", bus/slice wide discrepency (%ld<>%ld).");
	GfpUT_AddError(GlobVar.ModuleId, IE_UtFDP_ParserFieldUnknow,
	"[%s, line %ld]:Field \'%s\':Invalid value \"%s\".");

	/* Warnings belonging to parser/driver functions.
	 */
	GfpUT_AddWarng(GlobVar.ModuleId, IW_UtFDP_ParserOpenFile,
	"[%s]:Unable to open file \"%s.dpr\" (ignored).");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_UtFDP_ParserDefABDup,
	"[%s, line %ld]:Duplicate \"DP_DEFAB\" (ignored).");
}


/*  Function  :  'UtFDP_AddLoConPlace'.                          */
/* -------------------------------------------------------------- */
/*  Add to the logical connector pointed by "aLoCon" the informa-
 *  tions needed to place his associated physical connector.
 *      All informations about the physical connector(s) are
 *  chained together. The head of list is pointed by one 'ptype'
 *  structure of the USER field.
 */

extern void  UtFDP_AddLoConPlace(aLoCon, aSide, aSlice, aTrack)
	locon_list *aLoCon;  /*               Logical connector. */
		  char  aSide;   /* Side  of the physical connector. */
		  long  aSlice;  /* Slice of the physical connector. */
		  long  aTrack;  /* Track of the physical connector. */
{
		   ptype_list *PtUser;
	ulocon_place_list *PtNew;
	ulocon_place_list *PtPlace;


	PtUser = GfpUT_GetUser(&(aLoCon->USER), C_GfpUT_UserLoConPlace,
											F_GfpUT_UserCreat     );

	/* Attempt to merge the new placement request whith an older
	 * one. The placement constrainst have to be compatibles.
	 */
	for(PtPlace  = (ulocon_place_list*)(PtUser->DATA);
		PtPlace != (ulocon_place_list*)NULL;
		PtPlace  =            PtPlace->NEXT)
	{
		/* Check side coherency. */
		if (     (        aSide != PtPlace->SIDE          )
			||(  (        aSide != C_UtFDP_SideTypeDefault)
			   &&(PtPlace->SIDE != C_UtFDP_SideTypeDefault)) )
			continue;

		/* Check slice coherency. */
		if (     (        aSlice != PtPlace->SLICE          )
			||(  (        aSlice != C_UtFDP_SliceTypeDefault)
			   &&(PtPlace->SLICE != C_UtFDP_SliceTypeDefault)) )
			continue;

		/* Check track coherency. */
		if (     (        aTrack != PtPlace->TRACK          )
			||(  (        aTrack != C_UtFDP_TrackTypeDefault)
			   &&(PtPlace->TRACK != C_UtFDP_TrackTypeDefault)) )
			continue;

		/* Merge the two requests. */
		PtPlace->SIDE  =  (aSide  != C_UtFDP_SideTypeDefault)
						 ? aSide    :   PtPlace->SIDE;
		PtPlace->SLICE =  (aSlice != C_UtFDP_SliceTypeDefault)
						 ? aSlice   :   PtPlace->SLICE;
		PtPlace->TRACK =  (aTrack != C_UtFDP_TrackTypeDefault)
						 ? aTrack   :   PtPlace->TRACK;

		return;
	}

	/* Create, fill and add to the list the new request. */
	PtNew = (ulocon_place_list*)mbkalloc(sizeof(ulocon_place_list));

	 PtNew->SIDE  = aSide ;
	 PtNew->SLICE = aSlice;
	 PtNew->TRACK = aTrack;
	 PtNew->NEXT  = (ulocon_place_list*)PtUser->DATA;
	PtUser->DATA  =              (void*)PtNew;
}


/*  Function  :  'UtFDP_DefLoConPlace'.                          */
/* -------------------------------------------------------------- */

extern void  UtFDP_DefLoConPlace(aLoCon, aSide, aSlice, aTrack)
	locon_list *aLoCon;  /*               Logical connector. */
		  char  aSide;   /* Side  of the physical connector. */
		  long  aSlice;  /* Slice of the physical connector. */
		  long  aTrack;  /* Track of the physical connector. */
{
		   ptype_list *PtUser;
	ulocon_place_list *PtPlace;


	PtUser = GfpUT_GetUser(&(aLoCon->USER), C_GfpUT_UserLoConPlace,
											F_GfpUT_UserCreat     );

	/* Attempt to merge the new placement request whith an older
	 * one. The placement constrainst have to be compatibles.
	 */
	for(PtPlace  = (ulocon_place_list*)(PtUser->DATA);
		PtPlace != (ulocon_place_list*)NULL;
		PtPlace  =            PtPlace->NEXT)
	{
		/* Set the default side, if allowed. */
		if (PtPlace->SIDE == C_UtFDP_SideTypeDefault)
			PtPlace->SIDE = aSide;

		/* Set the default slice, if allowed. */
		if (PtPlace->SLICE == C_UtFDP_SliceTypeDefault)
			PtPlace->SLICE = aSlice;

		/* Set the default track, if allowed. */
		if (PtPlace->TRACK == C_UtFDP_TrackTypeDefault)
			PtPlace->TRACK = aTrack;
	}
}


/*  Function  :  'UtFDP_FreeLoConPlace'.                         */
/* -------------------------------------------------------------- */

extern void  UtFDP_FreeLoConPlace(aLoCon)
	locon_list *aLoCon;
{
		   ptype_list *PtUser;
	ulocon_place_list *PtNext;
	ulocon_place_list *PtPlace;


	PtUser = GfpUT_GetUser(&(aLoCon->USER), C_GfpUT_UserLoConPlace,
											F_GfpUT_UserGet       );

	if (PtUser != (ptype_list*)NULL)
	{
		for(PtPlace  = (ulocon_place_list*)(PtUser->DATA);
			PtPlace != (ulocon_place_list*)NULL;)
		{ PtNext=PtPlace->NEXT; mbkfree((void*)PtPlace); PtPlace=PtNext; }

		GfpUT_DelUser(&(aLoCon->USER), C_GfpUT_UserLoConPlace,
									   F_GfpUT_UserNoFlags   );
	}
}


/*  Function  :  'UtFDP_AddLoFigAB'.                             */
/* -------------------------------------------------------------- */
/*  Add to a logical figure the informations regarding the
 *  abutment-box ("AB") of the associated physical figure.
 */

extern void  UtFDP_AddLoFigAB(aLoFig, aDY1, aDY2)
	lofig_list *aLoFig;  /* Logical figure.                       */
		  long  aDY1;    /* Value to add to the bottom of the AB. */
		  long  aDY2;    /* Value to add to the top    of the AB. */
{
	ptype_list *PtUser;
	lofig_ab   *PtNew;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigAB,
											F_GfpUT_UserCreat  );

	if (PtUser->DATA != (void*)NULL)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_UtFDP_AddLoFigABDup,
									"AddLoFigAB",
									aLoFig->NAME);

	PtNew = (lofig_ab*)mbkalloc(sizeof(lofig_ab));

	PtNew->DY1 = aDY1;
	PtNew->DY2 = aDY2;

	PtUser->DATA = (void*)PtNew;
}


/*  Function  :  'UtFDP_DelLoFigAB'.                             */
/* -------------------------------------------------------------- */

extern void  UtFDP_DelLoFigAB(aLoFig)
	lofig_list *aLoFig;
{
	ptype_list *PtUser;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigAB,
											F_GfpUT_UserGet    );

	if (PtUser != (ptype_list*)NULL)
	{
		if (PtUser->DATA != (void*)NULL)
			mbkfree((void*)PtUser->DATA);

		GfpUT_DelUser(&(aLoFig->USER), C_GfpUT_UserLoFigAB,
									   F_GfpUT_UserNoFlags);
	}
}


/*  Function  :  'UtFDP_AddLoFigPower'.                          */
/* -------------------------------------------------------------- */
/*  Add to a logical figure the informations regarding the
 *  number of vertical power supplies refreshment.
 */

extern void  UtFDP_AddLoFigPower(aLoFig, aNumb, aWide)
	lofig_list *aLoFig;  /* Logical figure.                    */
		  long  aNumb;   /* Number of vertical power supplies. */
		  long  aWide;   /* Wide of the power segment.         */
{
	 ptype_list  *PtUser;
	ulofig_power *PtNew;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigPower,
											F_GfpUT_UserCreat     );

	if (PtUser->DATA != (void*)NULL)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_UtFDP_AddLoFigPowerDup,
									"AddLoFigPower",
									   aLoFig->NAME);

	PtNew = (ulofig_power*)mbkalloc(sizeof(ulofig_power));

	 PtNew->NUMB =  mc_Max(aNumb,IC_UtFDP_MinPowerNumb);
	 PtNew->WIDE =  mc_Max(aWide,IC_UtFDP_MinPowerWide);
	PtUser->DATA =  (void*)PtNew;
}


/*  Function  :  'UtFDP_DelLoFigPower'.                           */
/* -------------------------------------------------------------- */

extern void  UtFDP_DelLoFigPower(aLoFig)
	lofig_list *aLoFig;
{
	ptype_list *PtUser;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigPower,
											F_GfpUT_UserGet       );

	if (PtUser != (ptype_list*)NULL)
	{
		if (PtUser->DATA != (void*)NULL)
			mbkfree((void*)PtUser->DATA);

		GfpUT_DelUser(&(aLoFig->USER), C_GfpUT_UserLoFigPower,
									   F_GfpUT_UserNoFlags   );
	}
}


/*  Function  :  'UtFDP_AddLoFigGlue'.                            */
/* -------------------------------------------------------------- */

extern void  UtFDP_AddLoFigGlue(aLoFig, aIter, aWide, aCPC)
	lofig_list *aLoFig;
		  long  aCPC;
		  long  aIter;
		  long  aWide;
{
	 ptype_list *PtUser;
	ulofig_glue *PtNew;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigGlue,
											F_GfpUT_UserCreat    );

	if (PtUser->DATA != (void*)NULL)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_UtFDP_AddLoFigGlueDup,
								   "AddLoFigGlue",
									 aLoFig->NAME);

	PtNew = (ulofig_glue*)mbkalloc(sizeof(ulofig_glue));

	 PtNew->ITER =  mc_Max(aIter,IC_UtFDP_MinGlueIter);
	 PtNew->WIDE =         aWide;
	 PtNew->CPC  =         aCPC;
	PtUser->DATA =  (void*)PtNew;
}


/*  Function  :  'UtFDP_GetLoFigGlue'.                            */
/* -------------------------------------------------------------- */

extern ulofig_glue *UtFDP_GetLoFigGlue(aLoFig)
	lofig_list *aLoFig;
{
	ptype_list *PtUser;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigGlue,
											F_GfpUT_UserGet      );

	if (PtUser != (ptype_list*)NULL)
		return((ulofig_glue*)(PtUser->DATA));
	else
		return((ulofig_glue*)NULL);
}


/*  Function  :  'UtFDP_DelLoFigGlue'.                            */
/* -------------------------------------------------------------- */

extern void  UtFDP_DelLoFigGlue(aLoFig)
	lofig_list *aLoFig;
{
	ptype_list *PtUser;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigGlue,
											F_GfpUT_UserGet      );

	if (PtUser != (ptype_list*)NULL)
	{
		if (PtUser->DATA != (void*)NULL)
			mbkfree((void*)PtUser->DATA);

		GfpUT_DelUser(&(aLoFig->USER), C_GfpUT_UserLoFigGlue,
									   F_GfpUT_UserNoFlags  );
	}
}


/*  Function  :  'UtFDP_AddLoInsPower'.                          */
/* -------------------------------------------------------------- */
/*  Add to a logical instance informations about the power
 *  refreshment that could be set on it sides.
 */

extern void  UtFDP_AddLoInsPower(aLoIns, aSide, aWide)
	loins_list *aLoIns;  /* Logical instance.                      */
		  char  aSide;   /* Side of the instance to set the power. */
		  long  aWide;   /* Wide of the power segment.             */
{
	 ptype_list  *PtUser;
	uloins_power *PtNew;


	PtUser = GfpUT_GetUser(&(aLoIns->USER), C_GfpUT_UserLoInsPower,
											F_GfpUT_UserCreat     );

	if (PtUser->DATA == (void*)NULL)
	{
		PtNew        = (uloins_power*)mbkalloc(sizeof(uloins_power));
		PtUser->DATA = (void*)PtNew;
	}
	else
		PtNew = (uloins_power*)(PtUser->DATA);

	PtNew->SIDE |=         aSide;
	PtNew->WIDE  =  mc_Max(aWide,IC_UtFDP_MinPowerWide);
}


/*  Function  :  'UtFDP_DelLoInsPower'.                          */
/* -------------------------------------------------------------- */

extern void  UtFDP_DelLoInsPower(aLoIns)
	loins_list *aLoIns;
{
	ptype_list *PtUser;


	PtUser = GfpUT_GetUser(&(aLoIns->USER), C_GfpUT_UserLoInsPower,
											F_GfpUT_UserGet       );

	if (PtUser != (ptype_list*)NULL)
	{
		if (PtUser->DATA != (void*)NULL)
			mbkfree((void*)PtUser->DATA);

		GfpUT_DelUser(&(aLoIns->USER), C_GfpUT_UserLoInsPower,
									   F_GfpUT_UserNoFlags   );
	}
}


/*  Function  :  'UtFDP_SetLoFigFlags'.                           */
/* -------------------------------------------------------------- */

extern void  UtFDP_SetLoFigFlags(aLoFig, aFlags)
	lofig_list *aLoFig;
		  long  aFlags;
{
	 ptype_list *PtUser;
	ulofig_flags *PtNew;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigFlags,
											F_GfpUT_UserCreat     );

	if (PtUser->DATA == (void*)NULL)
	{
		PtNew        = (ulofig_flags*)mbkalloc(sizeof(ulofig_flags));
		PtNew->FLAGS = 0L;
	}

	 PtNew->FLAGS |= aFlags;
	PtUser->DATA   = (void*)PtNew;
}


/*  Function  :  'UtFDP_GetLoFigFlags'.                           */
/* -------------------------------------------------------------- */

extern ulofig_flags *UtFDP_GetLoFigFlags(aLoFig)
	lofig_list *aLoFig;
{
	ptype_list *PtUser;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigFlags,
											F_GfpUT_UserGet       );

	if (PtUser != (ptype_list*)NULL)
		return((ulofig_flags*)(PtUser->DATA));
	else
		return((ulofig_flags*)NULL);
}


/*  Function  :  'UtFDP_DelLoFigFlags'.                           */
/* -------------------------------------------------------------- */

extern void  UtFDP_DelLoFigFlags(aLoFig)
	lofig_list *aLoFig;
{
	ptype_list *PtUser;


	PtUser = GfpUT_GetUser(&(aLoFig->USER), C_GfpUT_UserLoFigFlags,
											F_GfpUT_UserGet       );

	if (PtUser != (ptype_list*)NULL)
	{
		if (PtUser->DATA != (void*)NULL)
			mbkfree((void*)PtUser->DATA);

		GfpUT_DelUser(&(aLoFig->USER), C_GfpUT_UserLoFigFlags,
									   F_GfpUT_UserNoFlags   );
	}
}


/*  Function  :  'FGetLine'.                                      */
/* -------------------------------------------------------------- */
/*  This function read one line from parser the input file.
 *      The line is the concatenation of all contiguous lines
 *  terminated by a backslash ('\'). The white spaces (space and
 *  tabulation) are translated to exactly one space character.
 *  This is done to simplify the token recognition.
 */

static long  FGetLine()
{
	void *RetV;
	long  IndexOutpt = 0L;
	long   AntiSlash = TRUE;
	long  MergeSpace;
	long  IndexClean;
	char  BufferClean[IC_UtFDP_ParserLineLength];
	char  BufferInput[IC_UtFDP_ParserLineLength];
	char *PtChar;


	while(    AntiSlash
		  && (RetV=fgets(BufferInput,
						 IC_UtFDP_ParserLineLength-2,
								   GlobVar.ParserFile)))
	{
		 AntiSlash = FALSE;
		MergeSpace = TRUE;
		IndexClean = 0L;

		/* Number of the current input line. */
		GlobVar.ParserCurrLine += 1;

		/* Process the input line. Merge white space, skip comments
		 * and detect anti-slash for splitted lines.
		 */
		for(PtChar=BufferInput; *PtChar!=(char)0; PtChar++)
		{
			/* Concatenate all contiguous white space in one single
			 * space caracter.
			 */
			if (isspace(*PtChar))
			{
				if (!MergeSpace)
					BufferClean[IndexClean++] = ' ';

				while(isspace(*(PtChar+1))) PtChar++;

				continue;
			}
			else
				MergeSpace = FALSE;

			/* Skip comments. */
			if (*PtChar ==  '#') break;

			/* Detect anti-slash for splitted lines. */
			if (*PtChar == '\\') { AntiSlash=TRUE; break; }

			BufferClean[IndexClean++] = *PtChar;
		}

		if (IndexClean == 0L)
		{
			/* Check for empty input lines. */
			if (IndexOutpt == 0L) AntiSlash = TRUE;
		}
		else
		{
			/* Add a space separator between merged lines. */
			if (BufferClean[IndexClean-1] != ' ')
				BufferClean[IndexClean++]  = ' ';
		}

		/* Concatenate the clean buffer with the output buffer. */
		if (IndexOutpt+IndexClean >= IC_UtFDP_ParserLineLength)
			GfpUT_PrintError(GlobVar.ModuleId,
								 IE_UtFDP_ParserLineLength,
											"LoadPlacement",
									 GlobVar.ParserCurrLine);
		else
		{
			strncpy(GlobVar.ParserLineBuffer      +IndexOutpt,
									  BufferClean, IndexClean);

			IndexOutpt += IndexClean;
			GlobVar.ParserLineBuffer[IndexOutpt] = (char)0;
		}
	}

	return(  (      (RetV==(void*)NULL) ? FALSE : TRUE)
		   ||((IndexOutpt==0L)          ? FALSE : TRUE));
}


/*  Function  :  'StrToFunction'.                                 */
/* -------------------------------------------------------------- */
/*  Try to convert the string "aFunction" (a token) in a
 *  numerical value.
 */

static long  StrToFunction(aFunction)
	char *aFunction;  /* Token to convert in a function. */
{
	if (!strcmp(aFunction,IS_UtFDP_ParserFunctionLoCon))
		return(IC_UtFDP_ParserFunctionLoCon);

	if (!strcmp(aFunction,IS_UtFDP_ParserFunctionDefAB))
		return(IC_UtFDP_ParserFunctionDefAB);

	if (!strcmp(aFunction,IS_UtFDP_ParserFunctionPower))
		return(IC_UtFDP_ParserFunctionPower);

	if (!strcmp(aFunction,IS_UtFDP_ParserFunctionKeep))
		return(IC_UtFDP_ParserFunctionKeep);

	if (!strcmp(aFunction,IS_UtFDP_ParserFunctionGlue))
		return(IC_UtFDP_ParserFunctionGlue);

	return(IC_UtFDP_ParserFunctionUnknow);
}


/*  Function  :  'StrToSide'.                                     */
/* -------------------------------------------------------------- */
/*  Try to convert the string "aSide" (a token) in a side value.
 */

static long  StrToSide(aSide)
	char *aSide;  /* Token to convert in a side. */
{
	if (!strcmp(aSide, IS_UtFDP_ParserSideNorth))
		return(C_UtFDP_SideTypeNorth);

	if (!strcmp(aSide, IS_UtFDP_ParserSideSouth))
		return(C_UtFDP_SideTypeSouth);

	if (!strcmp(aSide, IS_UtFDP_ParserSideEast))
		return(C_UtFDP_SideTypeEast);

	if (!strcmp(aSide, IS_UtFDP_ParserSideWest))
		return(C_UtFDP_SideTypeWest);

	if (!strcmp(aSide, IS_UtFDP_ParserStrDefault))
		return(C_UtFDP_SideTypeDefault);

	return(C_UtFDP_SideTypeUnknow);
}


/*  Function  :  'StrToTrack'.                                    */
/* -------------------------------------------------------------- */
/*  Try to convert the string "aStr" (a token) in a numerical
 *  track value, in decimal base.
 */

static long  StrToTrack(aStr, aPtTrack)
	char *aStr;      /* String to convert in a number. */
	long *aPtTrack;  /* Pointer to converted value.    */
{
	/* Check for the default string. */
	if (!strcmp(aStr, IS_UtFDP_ParserStrDefault))
	{
		*aPtTrack = C_UtFDP_TrackTypeDefault;
			 return(C_UtFDP_TrackTypeDefault);
	}

	/* Check if the is a legal decimal number. */
	if (!GfpUT_IsDecimalNumber(aStr))
		return(C_UtFDP_TrackTypeUnknow);

	/* Convert the string. */
	*aPtTrack = atol(aStr);

	return(C_UtFDP_TrackTypeNumber);
}


/*  Function  :  'UtFDP_LoadPlacement'.                          */
/* -------------------------------------------------------------- */
/*  This function load (parse) the file <aFigName>.dpr that con-
 *  tain the placement informations.
 *      This informations are added to the logical figure, via
 *  some special structures. Notice that this specials structures
 *  are stored in the USER fields.
 */

extern void  UtFDP_LoadPlacement(aFigName)
	char *aFigName;  /* Name of the logical figure. */
{
#define  mc_NextToken(aPtStart,aPtNext,aToken)                 \
			 if (((aPtNext)=strchr((aPtStart),' ')) == NULL)   \
				 GfpUT_PrintError(GlobVar.ModuleId,            \
								  IE_UtFDP_ParserMissToken,   \
								  "LoadPlacement"          ,   \
								  GlobVar.ParserCurrLine   ,   \
								  aToken                   );  \
			 *((aPtNext)++) = (char)0;

#define  mc_LastToken(aPtStart,aToken)                        \
			 if (strchr((aPtStart),(char)0) == NULL)          \
				 GfpUT_PrintError(GlobVar.ModuleId,           \
								  IE_UtFDP_ParserMissToken,  \
								  "LoadPlacement"          ,  \
								  GlobVar.ParserCurrLine   ,  \
								  aToken                   );


	lofig_list *PtLoFig;
	lofig_list *PtLoFigGlue;
	lofig_list *PtLoFigKeep;
	loins_list *PtLoIns;
	locon_list *PtLoCon;
	ptype_list *PtUser;
		  char *PtToken1, *PtToken2;
		  char *PowIns;
		  char *KeepName;
		  char  SigRoot[256],
				SigName[256];
		  long  SigType,
				SigOrig,
				SigExtr,
				SigWide,
				SigInc,
				Sig,
				Side,
				SliceType,
				SliceOrig,
				SliceExtr,
				SliceWide,
				SliceInc,
				Slice,
				Track;
		  long  DY1,
				DY2;
		  long  PowNum,
				PowWide;
		  long  GlueIter,
				GlueWide,
				GlueCPC;
		  long  i;


	/* Get logical and physical figures. */
	PtLoFig = getlofig(aFigName, 'A');

	/* Opening the file holding user placement informations. */
	if ((GlobVar.ParserFile=mbkfopen(PtLoFig->NAME,
										     "dpr",
										 READ_TEXT)) == (FILE*)NULL)
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_UtFDP_ParserOpenFile,
						 "LoadPlacement"         ,
						 PtLoFig->NAME           );

		return;
	}
	else
		GlobVar.ParserCurrLine = 0;


	while(FGetLine())
	{
		/* Read the fisrt line token: FUNCTION. */
		mc_NextToken(GlobVar.ParserLineBuffer,
									 PtToken1,
								   "FUNCTION");

		/* Recognition of the FUNCTION token. */
		switch(StrToFunction(GlobVar.ParserLineBuffer))
		{
			case IC_UtFDP_ParserFunctionLoCon:
				/* Attempt to recognize a line whith the following
				 * syntax:
				 *
				 * "^DP_LOCON<spc>ConName<spc>Side<spc>Slice<spc>Track$"
				 *
				 * In which fields have the following syntax:
				 *     - ConName : pin|pin_vect[S0]|vect[S0:S1]
				 *     - Side    : NORTH|SOUTH|EAST|WEST|DEFAULT
				 *     - Slice   : S0|S0:S1|DEFAULT
				 *     - Track   : T|DEFAULT
				 */

				/* Get and check the "ConName" token. */
				mc_NextToken(PtToken1, PtToken2, "NAME");

				SigType = GfpUT_StrSigDecod(PtToken1, SigRoot,
													 &SigOrig,
													 &SigExtr);

				if (SigType == C_GfpUT_SigTypeUnknow)
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserFieldUnknow,
												 "LoadPlacement",
										  GlobVar.ParserCurrLine,
														  "NAME",
														PtToken1);

				/* Get and check the "Side" token. */
				mc_NextToken(PtToken2, PtToken1, "SIDE");

				Side = StrToSide(PtToken2);

				if (Side == C_UtFDP_SideTypeUnknow)
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserFieldUnknow,
												 "LoadPlacement",
										  GlobVar.ParserCurrLine,
														  "SIDE",
													    PtToken2);

				/* Get and check the "Slice" token. */
				mc_NextToken(PtToken1, PtToken2, "SLICE");

				if (strcmp(PtToken1, IS_UtFDP_ParserStrDefault))
					SliceType = GfpUT_StrIndexDecod( PtToken1 ,
													&SliceOrig,
													&SliceExtr);
				else
				{
					SliceOrig = SliceExtr = 0L;
					SliceType = C_GfpUT_IdxTypeDefault;
				}

				if (SliceType == C_GfpUT_IdxTypeUnknow)
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserFieldUnknow,
												 "LoadPlacement",
										  GlobVar.ParserCurrLine,
														 "SLICE",
														PtToken1);

				/* Get and check the "Track" token. */
				mc_NextToken(PtToken2, PtToken1, "TRACK");

				switch(StrToTrack(PtToken2,&Track))
				{
					case C_UtFDP_TrackTypeNumber:
					case C_UtFDP_TrackTypeDefault:
						break;
					case C_UtFDP_TrackTypeUnknow:
					default:
						GfpUT_PrintError(GlobVar.ModuleId,
										 IE_UtFDP_ParserFieldUnknow,
													 "LoadPlacement",
											  GlobVar.ParserCurrLine,
															 "TRACK",
														    PtToken2);
						break;
				}

				/* Check if there is nothing after the last field. */
				if (*PtToken1 != (char)0)
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserExtraToken,
												"LoadPlacement",
										 GlobVar.ParserCurrLine,
													   PtToken1);

				/* Check coherency between signal and slice
				 * indexations, they must have the same wide.
				 */
				  SigWide =   SigExtr-  SigOrig;
				SliceWide = SliceExtr-SliceOrig;

				if (  SigWide < 0)   SigWide = -   SigWide;
				if (SliceWide < 0) SliceWide = - SliceWide;

				if (  (SliceType != C_GfpUT_IdxTypeDefault)
					&&(SliceWide != SigWide))
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserCoherWide,
											   "LoadPlacement",
										GlobVar.ParserCurrLine,
													 SigRoot  ,
													 SigWide+1,
												   SliceWide+1);

				/* Addition of the physical placement informations
				 * to the logical connector.
				 */
				  SigInc = (  SigOrig <=   SigExtr) ? 1 : -1;
				SliceInc = (SliceOrig <= SliceExtr) ? 1 : -1;
				for(Sig =SigOrig, Slice =SliceOrig; Sig!=SigExtr+SigInc;
					Sig+=SigInc , Slice+=SliceInc                      )
				{
					/* Build the signal name, depending on the
					 * signal type (scalar or bus/vector).
					 */
					switch(SigType)
					{
						case C_GfpUT_SigTypeBus:
						case C_GfpUT_SigTypeBusPin:
							sprintf(SigName, "%s %ld", SigRoot, Sig);
							break;
						case C_GfpUT_SigTypeScalar:
						default:
							strcpy(SigName, SigRoot);
							break;
					}

					/* Look for the associated logical connector. */
					PtLoCon = getlocon(PtLoFig, SigName);

					/* Add the physicals informations. */
					UtFDP_AddLoConPlace(PtLoCon,
										 Side   ,
 (SliceType == C_GfpUT_IdxTypeDefault) ?
		     C_UtFDP_SliceTypeDefault  : Slice  ,
										 Track  );
				}

				break;
			case IC_UtFDP_ParserFunctionDefAB:
				/* Attempt to recognize a line whith the following
				 * syntax:
				 *
				 *          "^DP_DEFAB<spc>dy1<spc>dy2$"
				 *
				 *     Where "DY1" and "DY2" are numerical fields.
				 */

				/* Get and check the "DY1" token. */
				mc_NextToken(PtToken1, PtToken2, "DY1");

				if (!GfpUT_IsDecimalNumber(PtToken1))
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserNumberUnknow,
												  "LoadPlacement",
										   GlobVar.ParserCurrLine,
															"DY1",
														 PtToken1);
				DY1 = atol(PtToken1);

				/* Get and check the "DY2" token. */
				mc_NextToken(PtToken2, PtToken1, "DY2");

				if (!GfpUT_IsDecimalNumber(PtToken2))
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserNumberUnknow,
												  "LoadPlacement",
										   GlobVar.ParserCurrLine,
															"DY2",
														 PtToken2);
				DY2 = atol(PtToken2);

				/* Check if there is nothing after the last field. */
				if (*PtToken1 != (char)0)
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserExtraToken,
												"LoadPlacement",
										 GlobVar.ParserCurrLine,
													   PtToken1);

				/* Append to the lofig USER field the informations
				 * about the Abutment-Box resizing.
				 *     If there is more than one "DP_DEFAB" in the
				 * file, the first is taken.
				 */
				PtUser = GfpUT_GetUser(&(PtLoFig->USER),
									   C_GfpUT_UserLoFigAB,
									   F_GfpUT_UserCreat  );

				if (PtUser->DATA != (void*)NULL)
					GfpUT_PrintWarng(GlobVar.ModuleId,
									 IW_UtFDP_ParserDefABDup,
											  "LoadPlacement",
									   GlobVar.ParserCurrLine);
				else
					UtFDP_AddLoFigAB(PtLoFig, DY1, DY2);

				break;
			case IC_UtFDP_ParserFunctionPower:
				/* Get and analys the first token. */
				mc_NextToken(PtToken1, PtToken2, "Number|InstName");

				if (GfpUT_IsDecimalNumber(PtToken1))
				{
					PowNum = atol(PtToken1);

					mc_NextToken(PtToken2, PtToken1, "PowWide");

					if (!GfpUT_IsDecimalNumber(PtToken2))
						GfpUT_PrintError(GlobVar.ModuleId,
										 IE_UtFDP_ParserNumberUnknow,
													  "LoadPlacement",
											   GlobVar.ParserCurrLine,
															"PowWide",
															 PtToken2);
					PowWide = atol(PtToken2);

					UtFDP_AddLoFigPower(PtLoFig, PowNum, PowWide);
				}
				else
				{
					PtLoIns = getloins(PtLoFig, PtToken1);
					
					/* Get and check the "SIDE" token. */
					mc_NextToken(PtToken2, PtToken1, "SIDE");
				
					Side = StrToSide(PtToken2);

					switch(Side)
					{
						case C_UtFDP_SideTypeEast:
						case C_UtFDP_SideTypeWest:
						case C_UtFDP_SideTypeDefault:
							break;
						case C_UtFDP_SideTypeNorth:
						case C_UtFDP_SideTypeSouth:
						case C_UtFDP_SideTypeUnknow:
						default:
							GfpUT_PrintError(GlobVar.ModuleId,
											 IE_UtFDP_ParserFieldUnknow,
														 "LoadPlacement",
												  GlobVar.ParserCurrLine,
																  "SIDE",
															    PtToken2);
					}

					if (Side == C_UtFDP_SideTypeDefault)
						Side  = C_UtFDP_SideTypeEast;

					mc_NextToken(PtToken1, PtToken2, "PowWide");

					if (!GfpUT_IsDecimalNumber(PtToken1))
						GfpUT_PrintError(GlobVar.ModuleId,
										 IE_UtFDP_ParserNumberUnknow,
													  "LoadPlacement",
											   GlobVar.ParserCurrLine,
															"PowWide",
															 PtToken1);
					PowWide = atol(PtToken1);

					UtFDP_AddLoInsPower(PtLoIns, Side, PowWide);

					PtToken1 = PtToken2;
				}

				/* Check if there is nothing after the last field. */
				if (*PtToken1 != (char)0)
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserExtraToken,
												"LoadPlacement",
										 GlobVar.ParserCurrLine,
													   PtToken1);
				break;
			case IC_UtFDP_ParserFunctionKeep:
				/* Loop on the instance name list. */
				for(i=0, KeepName=PtToken1; TRUE; i++)
				{
					if (*KeepName == (char)0) break;

					mc_NextToken(KeepName, PtToken2, "KEEPNAME");

					PtLoFigKeep = getlofig(KeepName, 'P');

					UtFDP_SetLoFigFlags(PtLoFigKeep, F_UtFDP_FlagLoFigKeep);

					KeepName = PtToken2;
				}
				break;
			case IC_UtFDP_ParserFunctionGlue:
				mc_NextToken(PtToken1, PtToken2, "FigName");

				PtLoFigGlue = getlofig(PtToken1, 'A');

				mc_NextToken(PtToken2, PtToken1, "ITER");

				if (!GfpUT_IsDecimalNumber(PtToken2))
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserNumberUnknow,
												  "LoadPlacement",
										   GlobVar.ParserCurrLine,
														   "ITER",
														 PtToken2);
				GlueIter = atol(PtToken2);

				mc_NextToken(PtToken1, PtToken2, "WIDTH");

				if (!GfpUT_IsDecimalNumber(PtToken1))
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserNumberUnknow,
												  "LoadPlacement",
										   GlobVar.ParserCurrLine,
														"WIDTH",
														 PtToken1);
				GlueWide = atol(PtToken1);

				mc_NextToken(PtToken2, PtToken1, "CPC");

				if (!GfpUT_IsDecimalNumber(PtToken2))
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserNumberUnknow,
												  "LoadPlacement",
										   GlobVar.ParserCurrLine,
														"CPC",
														 PtToken2);
				GlueCPC = atol(PtToken2);

				UtFDP_AddLoFigGlue(PtLoFigGlue, GlueIter, GlueWide, GlueCPC);

				/* Check if there is nothing after the last field. */
				if (*PtToken1 != (char)0)
					GfpUT_PrintError(GlobVar.ModuleId,
									 IE_UtFDP_ParserExtraToken,
												"LoadPlacement",
										 GlobVar.ParserCurrLine,
													   PtToken1);
				break;
			default:
				GfpUT_PrintError(GlobVar.ModuleId,
								 IE_UtFDP_ParserFunctionUnknow,
												"LoadPlacement",
										 GlobVar.ParserCurrLine,
										 GlobVar.ParserLineBuffer);
		}
	}


	/* Close the user placement file. */
	fclose(GlobVar.ParserFile);


#undef  mc_NextToken
#undef  mc_LastToken
}
