/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtFlat.c,v 1.10 1997/10/30 13:28:00 fitpath Exp fitpath $ */
/*                                                                */
/*          Alliance CAD System 3.2b                               */
/*             FpGen Package 1.10                                 */
/*                                                                */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                       */
/*                                                                */
/*  Author      :          Jean-Paul CHAPUT                       */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                       */
/* ************************************************************** */
/*  Module      :      "UtFlat.c"                                 */


/* ************************** Includes ************************** */


#include   MUT_H
#include   MLO_H
#include   MPH_H

#include  "GenericGN.h"


/* ********************* Internal constants ********************* */


#define   IE_PhStackOverflow      2
#define   IE_PhStackUnderflow     3

#define   IC_MaxPhStack           256
#define   IC_TransfNOSYM          0
#define   IC_TransfSYM_X          1
#define   IC_TransfSYM_Y          2
#define   IC_TransfSYMXY         (IC_TransfSYM_X | IC_TransfSYM_Y)


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "FpGenPAD" module identifier. */
	} GlobVar;


/* Structure holding an element of the physical stack.
 */
typedef struct {
	phfig_list *PhFigFather; /* The current model (phfig). */
	phins_list *PhInsSon;    /* The next instance (phins). */
	} TypePhState;


/* This structure is used to perform a model coordinates flattening.
 *     It allow multiple flatten whitout reseting the model coordinates
 * to their initial values. When a new coordinate flatten is atempted,
 * the differential transformation from the previous flatten is compu-
 * ted and then applied.
 *     To reset the coordinates of the model to their original values,
 * just ask the flatten to (XMODL, YMODL, NOSYM).
 *     This structure is appended to the USER ptype list.
 */
typedef struct {
	long  XMODL;    /*   Initial "XAB1" of the model.               */
	long  YMODL;    /*   Initial "YAB1" of the model.               */
	long  XFLAT;    /* Flattened "XAB1" of the model.               */
	long  YFLAT;    /* Flattened "XAB1" of the model.               */
	long  XAB1;     /*  Value of "XAB1" before flattening.          */
	long  YAB1;     /*  Value of "YAB1" before flattening.          */
	long  XAB2;     /*  Value of "XAB2" before flattening.          */
	long  YAB2;     /*  Value of "YAB2" before flattening.          */
	char  ATRANSF;  /* New absolute transformation of the model.    */
	char  RTRANSF;  /* Transformation relative to the last ATRANSF. */
	} uphfig_xyflat;


/* ********************* Internal Variables ********************* */


static        long     PhStackPoint;
static TypePhState     PhStackState[IC_MaxPhStack];


/* ******************** Interface Functions ********************* */


/* extern          void  UtFlt_Init();
 * extern          long  UtFlt_XYFlatPhFig();
 * extern          long  UtFlt_XYUnFlatPhFig();
 * extern          void  UtFlt_HrXYUnFlatPhFig();
 */


/* ********************* Internal Functions ********************* */


static          char  FlatOrient();
static          char  FlatTransf();
static uphfig_xyflat *AllocPhFigXYFlat();
static          long  XYFlatAB();
static          long  XYFlatIns();
static          long  XYFlatSeg();
static          long  XYFlatVia();
static          long  XYFlatRef();
static          long  XYFlatCon();

static          void  ResetPhStack();
static          void  PushPhState();
static          void  PopPhState();


/* ******************** Functions Definitions ******************* */


/*  Function  :  'UtFlt_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtFlt_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtFlat", F_GfpUT_ModuleAdd);

	/* Errors associated to the hierarchical analysis functions.
	 */
	GfpUT_AddError(GlobVar.ModuleId, IE_PhStackOverflow,
	"[%s]:Stack overflow, more than %ld levels.\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_PhStackUnderflow,
	"[%s]:Stack underflow, less than 0 levels.\n");
}


/*  Function  :  'FlatOrient'.                                    */
/* -------------------------------------------------------------- */
/*  Apply the transformation "aTransf" to the (connector) orien-
 *  tation "aOrient".
 */

static char  FlatOrient(aOrient, aTransf)
	char  aOrient;  /* Orientation to be flattened. */
	char  aTransf;  /* Transformation to apply.     */
{
	/* Transformation : X --> -X,
	 * invert EAST and WEST connectors.
	 */
	if ((aTransf == SYM_X) || (aTransf == SYMXY))
		switch(aOrient)
		{
			case EAST:  return(WEST);
			case WEST:  return(EAST);
			case NORTH:
			case SOUTH: break;
		}

	/* Transformation : Y --> -Y,
	 * invert SOUTH and NORTH connectors.
	 */
	if ((aTransf == SYM_Y) || (aTransf == SYMXY))
		switch(aOrient)
		{
			case NORTH: return(SOUTH);
			case SOUTH: return(NORTH);
			case EAST:
			case WEST:  break;
		}

	/* Transformation : NOSYM. */
	return(aOrient);
}


/*  Function  :  'FlatTransf'.                                    */
/* -------------------------------------------------------------- */
/*  Compose the two transformations "aTransfA" and "aTransfB".
 *      The allowed transformations is a subset of the MBK trans-
 *  formations. All rotations are excluded.
 */

static char  FlatTransf(aTransfA, aTransfB)
	char  aTransfA;
	char  aTransfB;
{
	char  TransfCompose;


	/* Convert the "aTransfA" transformation. */
	switch(aTransfA)
	{
		case NOSYM: TransfCompose = IC_TransfNOSYM; break;
		case SYM_X: TransfCompose = IC_TransfSYM_X; break;
		case SYM_Y: TransfCompose = IC_TransfSYM_Y; break;
		case SYMXY: TransfCompose = IC_TransfSYMXY; break;
	}

	/* Convert the "aTransfB" transformation and merge it
	 * with the "aTransfA" transformation.
	 */
	switch(aTransfB) 
	{
		case NOSYM: break;
		case SYM_X: TransfCompose ^= IC_TransfSYM_X; break;
		case SYM_Y: TransfCompose ^= IC_TransfSYM_Y; break;
		case SYMXY: TransfCompose ^= IC_TransfSYMXY; break;
	}

	/* Return the computed transformation. */
	switch(TransfCompose)
	{
		case IC_TransfNOSYM: return(NOSYM);
		case IC_TransfSYM_X: return(SYM_X);
		case IC_TransfSYM_Y: return(SYM_Y);
		case IC_TransfSYMXY: return(SYMXY);
	}
}


/*  Function  :  'AllocPhFigXYFlat'.                              */
/* -------------------------------------------------------------- */
/*  Allocate in memory a new "uphfig_xyflat" structure and fill
 *  it in order to flatten the physical figure "aPhFig".
 *      Once the the model has been flattened the fields XMODL,
 *  and YMODL must not be changed.
 */

static uphfig_xyflat *AllocPhFigXYFlat(aPhFig)
	phfig_list *aPhFig;
{
	uphfig_xyflat *PtNew;


	PtNew = (uphfig_xyflat*)mbkalloc(sizeof(uphfig_xyflat));

	PtNew->XMODL   = aPhFig->XAB1;
	PtNew->YMODL   = aPhFig->YAB1;
	PtNew->XAB1    = aPhFig->XAB1;
	PtNew->YAB1    = aPhFig->YAB1;
	PtNew->XAB2    = aPhFig->XAB2;
	PtNew->YAB2    = aPhFig->YAB2;
	PtNew->ATRANSF = NOSYM;
	PtNew->RTRANSF = NOSYM;

	return(PtNew);
}


/*  Function  :  'XYFlatAB'.                                      */
/* -------------------------------------------------------------- */
/*  Flatten the Abutment-Box of the figure "aPhFig". The flatten
 *  informations are kept in the "uphfig_xyflat" structure.
 */

static long  XYFlatAB(aPhFig)
	phfig_list *aPhFig;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	ptype_list *PtUser;
		  long  SwapVal;


	/* Attempt to recover the structure that holds the
	 * flatten informations. If not found, return a null value.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigXYFlat,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(0L);

	/* Flatten the bottom left corner of the AB. */
	xyflat(       &aPhFig->XAB1, &aPhFig->YAB1,
				   aPhFig->XAB1,  aPhFig->YAB1,
		   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
			access_XYFlat->XAB1, access_XYFlat->YAB1,
			access_XYFlat->XAB2, access_XYFlat->YAB2, 
								 access_XYFlat->RTRANSF);

	/* Flatten the top right corner of the AB. */
	xyflat(       &aPhFig->XAB2, &aPhFig->YAB2,
				   aPhFig->XAB2,  aPhFig->YAB2,
		   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
			access_XYFlat->XAB1, access_XYFlat->YAB1,
			access_XYFlat->XAB2, access_XYFlat->YAB2, 
								 access_XYFlat->RTRANSF);

	/* Restore the X coordinate order of the AB.
	 * (we must have XAB1 <= XAB2 in the MBK data base)
	 */
	if (  (access_XYFlat->RTRANSF == SYM_X) 
		||(access_XYFlat->RTRANSF == SYMXY))
	{
			 SwapVal = aPhFig->XAB1;
		aPhFig->XAB1 = aPhFig->XAB2;
		aPhFig->XAB2 = SwapVal;
	}

	/* Restore the Y coordinate order of the AB.
	 * (we must have YAB1 <= YAB2 in the MBK data base)
	 */
	if (  (access_XYFlat->RTRANSF == SYM_Y)
		||(access_XYFlat->RTRANSF == SYMXY) )
	{
			 SwapVal = aPhFig->YAB1;
		aPhFig->YAB1 = aPhFig->YAB2;
		aPhFig->YAB2 = SwapVal;
	}

	return(1L);


#undef    access_XYFlat
}


/*  Function  :  'XYFlatIns'.                                     */
/* -------------------------------------------------------------- */
/*  Flatten all the instances of the figure "aPhFig". The flatten
 *  informations are kept in the "uphfig_xyflat" structure.
 */

static long  XYFlatIns(aPhFig)
	phfig_list *aPhFig;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	phfig_list *PtInsFig;
	phins_list *PtIns;
	ptype_list *PtUser;


	/* Attempt to recover the structure that holds the
	 * flatten informations. If not found, return a null value.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigXYFlat,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(0L);


	/* Loop on figure instance list. */
	for(PtIns  = aPhFig->PHINS;
		PtIns != (phins_list*)NULL;
		PtIns  =       PtIns->NEXT)
	{
		PtInsFig = getphfig( PtIns->FIGNAME, 'A');

		/* Flatten the instance coordinates XINS and YINS. */
		xyflat(        &PtIns->XINS, &PtIns->YINS,
					    PtIns->XINS,  PtIns->YINS,
			   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
				access_XYFlat->XAB1, access_XYFlat->YAB1,
				access_XYFlat->XAB2, access_XYFlat->YAB2,
									 access_XYFlat->RTRANSF);

		/* Compute the transformation of the instance. Compose
		 * the figure and the instance transformation.
		 */
		PtIns->TRANSF = FlatTransf(access_XYFlat->RTRANSF,
												 PtIns->TRANSF );

		/* As (XINS, YINS) must correspond to the bottom left corner
		 * of the model of the instance, it must be corrected after
		 * the flatten operation.
		 */
		switch(access_XYFlat->RTRANSF)
		{
			case NOSYM: break;
			case SYM_X:
				/* Substract the wide of the instance AB. */
				PtIns->XINS -= PtInsFig->XAB2 - PtInsFig->XAB1;
				break;
			case SYM_Y:
				/* Substract the height of the instance AB. */
				PtIns->YINS -= PtInsFig->YAB2 - PtInsFig->YAB1;
				break;
			case SYMXY:
				/* Substract the width and the height of the
				 * instance AB.
				 */
				PtIns->XINS -= PtInsFig->XAB2 - PtInsFig->XAB1;
				PtIns->YINS -= PtInsFig->YAB2 - PtInsFig->YAB1;
				break;
			default: return(0L);
		  }
	}

	return(1L);


#undef    access_XYFlat
}


/*  Function  :  'XYFlatSeg'.                                     */
/* -------------------------------------------------------------- */
/*  Flatten all the segments of the figure "aPhFig". The flatten
 *  informations are kept in the "uphfig_xyflat" structure.
 */

static long  XYFlatSeg(aPhFig)
	phfig_list *aPhFig;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	phseg_list *PtSeg;
	ptype_list *PtUser;
		  long  SwapVal;


	/* Attempt to recover the structure that holds the
	 * flatten informations. If not found, return a null value.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigXYFlat,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(0L);

	/* Loop on figure segment list. */
	for(PtSeg  = aPhFig->PHSEG;
		PtSeg != (phseg_list*)NULL;
		PtSeg  =       PtSeg->NEXT)
	{
		/* Flatten of (X1,Y1). */
		xyflat(          &PtSeg->X1, &PtSeg->Y1,
						  PtSeg->X1,  PtSeg->Y1,
			   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
				access_XYFlat->XAB1, access_XYFlat->YAB1,
				access_XYFlat->XAB2, access_XYFlat->YAB2,
									 access_XYFlat->RTRANSF);

		/* Flatten of (X2,Y2). */
		xyflat(          &PtSeg->X2, &PtSeg->Y2,
						  PtSeg->X2,  PtSeg->Y2,
			   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
				access_XYFlat->XAB1, access_XYFlat->YAB1,
				access_XYFlat->XAB2, access_XYFlat->YAB2,
									 access_XYFlat->RTRANSF);

		/* Restore the rigth ordering on X coordinates. */
		if (  (access_XYFlat->RTRANSF == SYM_X)
			||(access_XYFlat->RTRANSF == SYMXY))
		{
			  SwapVal = PtSeg->X1;
			PtSeg->X1 = PtSeg->X2;
			PtSeg->X2 = SwapVal;
		}

		/* Restore the rigth ordering on Y coordinates. */
		if (  (access_XYFlat->RTRANSF == SYM_Y)
			||(access_XYFlat->RTRANSF == SYMXY))
		{
			  SwapVal = PtSeg->Y1;
			PtSeg->Y1 = PtSeg->Y2;
			PtSeg->Y2 = SwapVal;
		}
	}

	return(1L);


#undef    access_XYFlat
}


/*  Function  :  'XYFlatVia'.                                     */
/* -------------------------------------------------------------- */
/*  Flatten all the vias of the figure "aPhFig". The flatten
 *  informations are kept in the "uphfig_xyflat" structure.
 */

static long  XYFlatVia(aPhFig)
	phfig_list *aPhFig;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	phvia_list *PtVia;
	ptype_list *PtUser;


	/* Attempt to recover the structure that holds the
	 * flatten informations. If not found, return a null value.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigXYFlat,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(0L);

	/* Loop on figure VIA list. */
	for(PtVia  = aPhFig->PHVIA;
		PtVia != (phvia_list*)NULL;
		PtVia  =       PtVia->NEXT)
		/* Flatten of (XVIA,YVIA). */
		xyflat(        &PtVia->XVIA, &PtVia->YVIA,
						PtVia->XVIA,  PtVia->YVIA,
			   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
				access_XYFlat->XAB1, access_XYFlat->YAB1,
				access_XYFlat->XAB2, access_XYFlat->YAB2,
									 access_XYFlat->RTRANSF);

	return(1L);


#undef    access_XYFlat
}


/*  Function  :  'XYFlatRef'.                                     */
/* -------------------------------------------------------------- */
/*  Flatten all the references of the figure "aPhFig". The flatten
 *  informations are kept in the "uphfig_xyflat" structure.
 */

static long  XYFlatRef(aPhFig)
	phfig_list *aPhFig;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	phref_list *PtRef;
	ptype_list *PtUser;


	/* Attempt to recover the structure that holds the
	 * flatten informations. If not found, return a null value.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigXYFlat,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(0L);

	/* Loop on figure reference list. */
	for(PtRef  = aPhFig->PHREF;
		PtRef != (phref_list*)NULL;
		PtRef  =       PtRef->NEXT)
		/* Flatten of (XREF,YREF). */
		xyflat(        &PtRef->XREF, &PtRef->YREF,
						PtRef->XREF,  PtRef->YREF,
			   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
				access_XYFlat->XAB1, access_XYFlat->YAB1,
				access_XYFlat->XAB2, access_XYFlat->YAB2,
									 access_XYFlat->RTRANSF);

	return(1L);


#undef    access_XYFlat
}


/*  Function  :  'XYFlatCon'.                                     */
/* -------------------------------------------------------------- */
/*  Flatten all the connectors of the figure "aPhFig". The flatten
 *  informations are kept in the "uphfig_xyflat" structure.
 */

static long  XYFlatCon(aPhFig)
	phfig_list *aPhFig;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	phcon_list *PtCon;
	ptype_list *PtUser;


	/* Attempt to recover the structure that holds the
	 * flatten informations. If not found, return a null value.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigXYFlat,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(0L);

	/* Loop on figure connector list. */
	for(PtCon  = aPhFig->PHCON;
		PtCon != (phcon_list*)NULL;
		PtCon  =       PtCon->NEXT)
	{
		/* Flatten of (XCON,YCON). */
		xyflat(        &PtCon->XCON, &PtCon->YCON,
						PtCon->XCON,  PtCon->YCON,
			   access_XYFlat->XFLAT, access_XYFlat->YFLAT,
				access_XYFlat->XAB1, access_XYFlat->YAB1,
				access_XYFlat->XAB2, access_XYFlat->YAB2,
									 access_XYFlat->RTRANSF);

		/* Compute the new orientation of the connector. */
		PtCon->ORIENT = FlatOrient(        PtCon->ORIENT ,
										 access_XYFlat->RTRANSF);
	}

	return(1L);


#undef    access_XYFlat
}


/*  Function  :  'UtFlt_XYFlatPhFig'.                             */
/* -------------------------------------------------------------- */
/*  Flatten all the coordinates of a physical figure. The flatten
 *  informations are given to sub-function by the structure
 *  'uphfig_xyflat'.
 *      If necessary, this structure is created by the function.
 *      The structure is not removed at the end of the function,
 *  to remove it, please use 'UtFlt_XYUnFlatPhFig'.
 */

extern long  UtFlt_XYFlatPhFig(aPhFig, aXFlat, aYFlat, aTransf)
	phfig_list *aPhFig;
		  long  aXFlat;
		  long  aYFlat;
		  char  aTransf;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	ptype_list *PtUser;
		  long  RetVal = 1L;


	/* Attempt to get the flatten structure in the USER
	 * ptype_list, and create it if necessary.
	 */
	PtUser = GfpUT_GetUser(&(aPhFig->USER),
						   C_GfpUT_UserPhFigXYFlat,
						   F_GfpUT_UserCreat      );

	/* Allocate the flatten structure, if needed. */
	if (access_XYFlat == (uphfig_xyflat*)NULL)
		PtUser->DATA = (void*)AllocPhFigXYFlat(aPhFig);

	access_XYFlat->XFLAT   = aXFlat;
	access_XYFlat->YFLAT   = aYFlat;
	access_XYFlat->XAB1    = aPhFig->XAB1;
	access_XYFlat->YAB1    = aPhFig->YAB1;
	access_XYFlat->XAB2    = aPhFig->XAB2;
	access_XYFlat->YAB2    = aPhFig->YAB2;
	access_XYFlat->RTRANSF = FlatTransf(access_XYFlat->ATRANSF, aTransf);
	access_XYFlat->ATRANSF = aTransf;


	/* Flatten all the objects inside the figure. */
	RetVal &= XYFlatAB( aPhFig);
	RetVal &= XYFlatIns(aPhFig);
	RetVal &= XYFlatSeg(aPhFig);
	RetVal &= XYFlatVia(aPhFig);
	RetVal &= XYFlatRef(aPhFig);
	RetVal &= XYFlatCon(aPhFig);

	return(RetVal);


#undef    access_XYFlat
}


/*  Function  :  'UtFlt_XYUnFlatPhFig'.                           */
/* -------------------------------------------------------------- */

extern long  UtFlt_XYUnFlatPhFig(aPhFig)
	phfig_list *aPhFig;
{
#define    access_XYFlat    ((uphfig_xyflat*)PtUser->DATA)

	ptype_list *PtUser;
		  long  RetVal;


	/* Attempt to recover the structure that holds the
	 * flatten informations. If not found, return a null value.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigXYFlat,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(1L);

	/* Unflattening the figure is equivalent to flatten it to the
	 * coordinate (XMODL,YMODL) whith the transformation NOSYM.
	 */
	RetVal = UtFlt_XYFlatPhFig(aPhFig,
							   access_XYFlat->XMODL,
							   access_XYFlat->YMODL,
											  NOSYM);

	/* Removing the 'uphfig_xyflat' structure. */
	mbkfree((void*)access_XYFlat);

	GfpUT_DelUser(&(aPhFig->USER), C_GfpUT_UserPhFigXYFlat,
								   F_GfpUT_UserNoFlags    );

	return(RetVal);

#undef    access_XYFlat
}


/*  Function  : 'ResetPhStack'                                    */
/* -------------------------------------------------------------- */
/*  Reset the physical state by setting his pointer to zero.
 */

static void  ResetPhStack()
{ PhStackPoint = 0L; }


/*  Function  : 'PushPhState'                                     */
/* -------------------------------------------------------------- */
/*  Push a phgical state to the (physical) stack.
 */

static void  PushPhState(PhFigFather, PhInsSon)
	phfig_list *PhFigFather; /* The current physical model. */
	phins_list *PhInsSon;    /* The next physical instance. */
{
	/* Test stack overflow. */
	if (PhStackPoint == IC_MaxPhStack)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_PhStackOverflow,
							  "PushPhState",
							   PhStackPoint);

	PhStackState[PhStackPoint].PhFigFather = PhFigFather;
	PhStackState[PhStackPoint].PhInsSon    = PhInsSon;

	PhStackPoint++;
}


/*  Function  : 'PopPhState'                                      */
/* -------------------------------------------------------------- */
/*  Pop a physical state from the (physical) stack.
 */

static void  PopPhState(PtPhFigFather, PtPhInsSon)
	phfig_list  **PtPhFigFather; /* Pointer to the current physical model. */
	phins_list  **PtPhInsSon;    /* Pointer to the next physical instance. */
{
	/* Test stack underflow. */
	if (PhStackPoint == 0)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_PhStackUnderflow,
									  "PopPhState");

	PhStackPoint--;

	*PtPhFigFather = PhStackState[PhStackPoint].PhFigFather;
	*PtPhInsSon    = PhStackState[PhStackPoint].PhInsSon;
}


/*  Function  :  'UtFlt_HrXYUnFlatPhFig'.                         */
/* -------------------------------------------------------------- */
/*  Hierarchical extention of the function 'UtFlt_XYUnFlatPhFig'.
 *  Unflatten the complete hierachy of the model "aPhFig".
 */

extern void  UtFlt_HrXYUnFlatPhFig(aPhFig)
	phfig_list *aPhFig;
{
	phfig_list *PhFigFather;
	phfig_list *PhFigSon;
	phins_list *PhInsSon;


	PhFigFather = aPhFig;
	PhInsSon    =  PhFigFather->PHINS;


	/* Reset physical stack and push the stop value. */
	ResetPhStack();
	PushPhState((phfig_list*)NULL, (phins_list*)NULL );

	/* Recursion main loop. */
	while(PhFigFather != (phfig_list*)NULL)
	{
	    /* If all instances have been processed, return to
		 * the upper levels.
		 */
		if (PhInsSon == (phins_list*)NULL)
		{
			PopPhState(&PhFigFather, &PhInsSon);
			continue;
		}

		/* Get the model of the instance. */
		PhFigSon = getphfig(PhInsSon->FIGNAME, 'A');

		/* Flaten the coordinate of the model. */
		UtFlt_XYUnFlatPhFig(PhFigSon);

		/* Go down the instances of the current son model
		 * or process the next instance of the father model.
		 */
		if (PhFigSon->PHINS != (phins_list*)NULL)
		{
			PushPhState(PhFigFather, PhInsSon->NEXT);

			PhFigFather = PhFigSon;
			PhInsSon    = PhFigSon->PHINS;
		}
		else
			PhInsSon = PhInsSon->NEXT;
	}
}
