/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: UtHierarchy.c,v 1.10 1997/10/30 13:28:01 fitpath Exp fitpath $    |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, ASIM, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :            cao-vlsi@lip6.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module    :  "UtHierarchy.c"                                      | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# include   MUT_H
# include   MLO_H
# include   MPH_H

# include  "GenericGN.h"


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                        Internal Constants                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

# define        IW_HiddenLevel                2
# define        IW_ColumnEmpty                3
# define        IW_ColumnMixed                4
# define        IW_ColumnUnexpectedLeafCell   5
# define        IW_FlattenLeafCell            6
# define        IW_FlattenColumn              7
# define        IW_TerminalLocation           8

# define        IE_LoStackOverflow            2
# define        IE_LoStackUnderflow           3
# define        IE_OperatorEmpty              4
# define        IE_OperatorMixed              5

# define        IC_MaxLoStack                 256
# define        IC_TypeModelUnknow            0
# define        IC_TypeModelDefault           1
# define        IC_TypeModelGenerator         2
# define        IC_TypeModelDplib             3
# define        IC_TypeModelFplib             4

# define        IS_TypeModelFplib             "fp"
# define        IS_TypeModelDplib             "dp"
# define        IS_TypeModelGenerator         "c"


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |             External Structures and Variables                  | |%
 *   %| \----------------------------------------------------------------/ |%
 *   %|                                                                    |%
 *   %| All those variables must be declared `extern' in "GenericGN.h".    |%
 */

       UtHir_V_t  UtHir_V;


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |               Internal Structures and Variables                    | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/* Structure holding together all internal module global variables. */
static struct {
	long  ModuleId;          /* "UtHierarchy" module identifier. */
	} GlobVar;

/* Structure holding an element of the logical stack. */
typedef struct {
	lofig_list *LoFigFather; /* The current model (lofig). */
	loins_list *LoInsSon;    /* The next instance (loins). */
	} TypeState;

static       long     LoStackPoint;
static  TypeState     LoStackState[IC_MaxLoStack];


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                       Macros Definitions                       | |%
 *   %| \----------------------------------------------------------------/ |%
 */

# define  mc_IsHSeg(seg)   (((seg)->TYPE == RIGTH) || ((seg)->TYPE == LEFT))
# define  mc_IsVSeg(seg)   (((seg)->TYPE == UP   ) || ((seg)->TYPE == DOWN))
# define  mc_InCatalog(s)  (incatalog(s) ? TRUE : FALSE)


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                  Internal Functions Declarations                   | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

static  int  mprintf         __FP(( char *aFormat, ... ));
static void  ResetLoStack    __FP(( void ));
static void  PushLoState
               __FP(( lofig_list *LoFigFather, loins_list *LoInsSon ));
static void  PopLoState
               __FP(( lofig_list **PtLoFigFather, loins_list **PtLoInsSon ));
static long  IsPhantom       __FP(( char *file_name ));
static long  TypeModelName   __FP(( char *ModelName ));
static long  IsModelLeafCell __FP(( lofig_list *PtLoFig ));
static long  IsModelColumn   __FP(( lofig_list *PtLoFig ));
static long  IsModelCluster  __FP(( lofig_list *PtLoFig ));
static long  HaveStruct      __FP(( char *aModelName ));
static long  HaveLayout      __FP(( char *aModelName ));
static long  IsRoutedLayout  __FP(( phfig_list *PtPhFig ));
static long  IsModelUS       __FP((  char *aName  ));
static long  IsModelCL       __FP((  char *aName  ));
static long  IsModelBK       __FP((  char *aName  ));
static long  IsOperatorUser  __FP(( lofig_list *PtLoFig ));


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                     Functions Definitions                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "mprintf()"                                          | |%
 * %| \--------------------------------------------------------------------/ |%
 */

static int  mprintf __FPV(( char *aFormat, ... ))
{
          va_list  ArgList;
  __KR_C(    char *aFormat; )
              int  RetVal;

  __ANSI_C( va_start(ArgList, aFormat); )
  __KR_C( va_start(ArgList); )
  __KR_C( aFormat = va_arg(ArgList, char*); )

  if (UtHir_V.Verbose)
    RetVal = vfprintf(stderr, aFormat, ArgList);
  va_end( ArgList );

  return( RetVal );
}


/*  Function  :  'UtHir_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtHir_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtHierarchy", F_GfpUT_ModuleAdd);
    UtHir_V.Verbose  = FALSE;

	/* Errors associated to the hierarchical analysis functions.
	 */
	GfpUT_AddError(GlobVar.ModuleId, IE_LoStackOverflow,
	"[%s]:Stack overflow, more than %ld levels.\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_LoStackUnderflow,
	"[%s]:Stack underflow, less than 0 levels.\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_OperatorEmpty,
	"User operator \"%s\" has no leaf cell.\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_OperatorMixed,
	"All instances of user operator \"%s\" are not of the same type.\n");

	/* Warnings associated to the hierarchical analysis functions.
	 */
	GfpUT_AddWarng(GlobVar.ModuleId, IW_HiddenLevel,
	"[%s]:Hidden level detected in leaf cell model \"%s\".");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_ColumnEmpty,
	"[%s]:Column model \"%s\" has no instance.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_ColumnMixed,
"[%s]:Instance \"%s\" of model \"%s\" in column \"%s\" is not a leaf cell.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_ColumnUnexpectedLeafCell,
	"[%s]:Column expected as place of leaf cell model \"%s\".");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_FlattenLeafCell,
	"[%s]:Unable to flatten data-path model \"%s\" (present in CATAL).");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_FlattenColumn,
	"[%s]:For model \"%s\":A column cannot be flattened.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_TerminalLocation,
	"Terminal \"%s\"(%ld,%ld) of model \"%s\" is not on the A-Box.");
}


/*  Function  : 'ResetLoStack'                                    */
/* -------------------------------------------------------------- */
/*  Reset the logical state by setting his pointer to zero.
 */

static void  ResetLoStack()
{ LoStackPoint = 0L; }


/*  Function  : 'PushLoState'                                     */
/* -------------------------------------------------------------- */
/*  Push a logical state to the (logical) stack.
 */

static void  PushLoState(LoFigFather, LoInsSon)
	lofig_list *LoFigFather; /* The current logical model. */
	loins_list *LoInsSon;    /* The next logical instance. */
{
	/* Test stack overflow. */
	if (LoStackPoint == IC_MaxLoStack)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_LoStackOverflow,
										  "PushLoState",
										   LoStackPoint);

	LoStackState[LoStackPoint].LoFigFather = LoFigFather;
	LoStackState[LoStackPoint].LoInsSon    = LoInsSon;

	LoStackPoint++;
}


/*  Function  : 'PopLoState'                                      */
/* -------------------------------------------------------------- */
/*  Pop a logical state from the (logical) stack.
 */

static void  PopLoState(PtLoFigFather, PtLoInsSon)
	lofig_list  **PtLoFigFather; /* Pointer to the current logical model. */
	loins_list  **PtLoInsSon;    /* Pointer to the next logical instance. */
{
	/* Test stack underflow. */
	if (LoStackPoint == 0)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_LoStackUnderflow,
											"PopLoState");

	LoStackPoint--;

	*PtLoFigFather = LoStackState[LoStackPoint].LoFigFather;
	*PtLoInsSon    = LoStackState[LoStackPoint].LoInsSon;
}


/*  Function  :  'IsPhantom'                                      */
/* -------------------------------------------------------------- */
/*  Attempt to find on disk the file named "file_name".vst, using
 *  the MBK search path "MBK_CATA_LIB".
 */

static long  IsPhantom(file_name)
	char *file_name;
{ 
	FILE *StreamPhantom;

	StreamPhantom = mbkfopen(file_name, IN_LO, READ_TEXT);

	if (StreamPhantom == NULL)
		return(1);
	else
	{
	    fclose(StreamPhantom);
		return(0);
	}
}



/*  Function  :  'TypeModelName'                                  */
/* -------------------------------------------------------------- */
/*  Return the type of model, according to is name :              */
/*              -   '*_c' : IC_TypeModelGenerator.            */
/*              -  '*_fp' : IC_TypeModelFplib.                */
/*              -  '*_dp' : IC_TypeModelDplib.                */
/*              -  others : IC_TypeModelDefault.              */

static long  TypeModelName(ModelName)
	char *ModelName; /* Model name to analys. */
{
	char *PtSuffix = (char)NULL;
	char *PtUnderScore;

	/* Search the last '_' in the name.*/
	for( PtUnderScore  =  ModelName; 
		*PtUnderScore != (char)NULL; PtUnderScore++)
		if (*PtUnderScore == '_') PtSuffix = PtUnderScore+1;

	/* If no underscore found then return the default type. */
	if (PtSuffix == (char*)NULL) return( IC_TypeModelDefault);


	/* **** Try to match the suffix of a leaf cell **** */

	/* Suffix of (big) generators such as 'rsa', 'rfg' and 'bsg'. */
	if (!strcmp(PtSuffix,IS_TypeModelGenerator))
		return(          IC_TypeModelGenerator);

	/* Suffix of dedicated generator leaf cell. */
	if (!strcmp(PtSuffix,IS_TypeModelFplib))
		return(          IC_TypeModelFplib);

	/* Suffix of user and SL leaf cell. */
	if (!strcmp(PtSuffix,IS_TypeModelDplib))
		return(          IC_TypeModelDplib);

	/* Default. */
	return( IC_TypeModelDefault);


/*	return( (incatalog(ModelName) ? IC_TypeModelFplib
 *								  : IC_TypeModelDefault) );
 */
}


/*  Function  :  'IsModelLeafCell'                                */
/* -------------------------------------------------------------- */
/*  Check if a model is a leaf cell.
 */

static long  IsModelLeafCell(PtLoFig)
	lofig_list *PtLoFig;
{
	/* Check the name and if the LOINS list is empty. */
	switch(TypeModelName(PtLoFig->NAME))
	{
		case IC_TypeModelGenerator:
			return(1);
		case IC_TypeModelFplib:
		case IC_TypeModelDplib:
			/* Detect instances under leaf cells. */
			if (PtLoFig->LOINS != (loins_list*)NULL)
				GfpUT_PrintWarng(GlobVar.ModuleId,
								 IW_HiddenLevel,
											"IsModelLeafCell",
											    PtLoFig->NAME);
			return( 1);
		case IC_TypeModelDefault:
		default:
			return( 0);
	}
}


/*  Function  :  'IsModelColumn'                                  */
/* -------------------------------------------------------------- */
/*  Check if a model is a column.
 */

static long  IsModelColumn(PtLoFig)
	lofig_list *PtLoFig;
{
	lofig_list *LoFigLeaf;
	loins_list *LoInsLeaf;
	      long  LoFigMode;
	      long  TypeModel;
	      long  ModelLeafFlag;
	      long  ConsistencyFlag;

	/* Check the name of the column. */
	if ((TypeModel=TypeModelName(PtLoFig->NAME))
		                  != IC_TypeModelDefault)
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_ColumnUnexpectedLeafCell,
												   "IsModelColumn",
													 PtLoFig->NAME);

	/* Check if there is some instances. */
	if (PtLoFig->LOINS == (loins_list*)NULL)
	{
	    if (TypeModel == IC_TypeModelDefault)
			GfpUT_PrintWarng(GlobVar.ModuleId,
							 IW_ColumnEmpty,
										  "IsModelColumn",
											PtLoFig->NAME);

		return( 0);
	}

	/* Check if all the instance models are leaf cells. */
	for(LoInsLeaf=PtLoFig->LOINS; LoInsLeaf != (  loins_list*)NULL;
		                          LoInsLeaf  =    LoInsLeaf->NEXT )


	{
	    /* Get the model of the instance. */
		LoFigMode = (IsPhantom(LoInsLeaf->FIGNAME)) ? 'P' : 'A';
	    LoFigLeaf = getlofig(LoInsLeaf->FIGNAME,
							 LoFigMode         );

		/* Check if the model is a leaf cell. */
		ModelLeafFlag = IsModelLeafCell(LoFigLeaf);

		/* Check if there is only one type of models. */
		if (LoInsLeaf == PtLoFig->LOINS)
			ConsistencyFlag = ModelLeafFlag;
		else
			if (ConsistencyFlag != ModelLeafFlag)
			{
				GfpUT_PrintWarng(GlobVar.ModuleId,
								 IW_ColumnMixed,
											  "IsModelColumn",
										   LoInsLeaf->INSNAME,
										   LoInsLeaf->FIGNAME,
											 PtLoFig->NAME   );
				return( 0);
			}
	}

	/* If "ConsistencyFlag" is set to one, this is a column. */
	return(ConsistencyFlag);
}


/*  Function  :  'IsModelCluster'                                 */
/* -------------------------------------------------------------- */
/*  Check if a model is a cluster, by searching an associated
 *  layout file.
 */

static long  IsModelCluster(PtLoFig)
	lofig_list *PtLoFig;
{
	FILE *StreamLayout;

	StreamLayout = mbkfopen( PtLoFig->NAME, IN_PH, READ_TEXT);

	if (StreamLayout == NULL)
		return( 0);
	else
	{
	    fclose( StreamLayout);
		return( 1);
	}
}


/*  Function  :  'HaveStruct'                                     */
/* -------------------------------------------------------------- */
/*  Attempt to find on the disk a structural description of the
 *  model "aModelName", this is needed when "IN_LO" is to "vst".
 */

static long  HaveStruct(aModelName)
	char *aModelName;
{ 
	FILE *StreamStruct;

	StreamStruct = mbkfopen(aModelName, IN_LO, READ_TEXT);

	if (StreamStruct == NULL)
		return(1);
	else
	{
	    fclose(StreamStruct);
		return(0);
	}
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "HaveLayout()"                                       | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Checks if there is a physical layout associated to the logical figure  |%
 * %| given by "PtLoFig". If the layout is found, the function return '1',   |%
 * %| '0' else.                                                              |%
 */

static long  HaveLayout( aModelName )
  char *aModelName;
{
  FILE *StreamLayout;

  StreamLayout = mbkfopen(aModelName, IN_PH, READ_TEXT);
  if (StreamLayout == NULL) {
    mprintf( "HaveLayout: can't open file \"%s\".\n", aModelName );
    return(0);
  }
  fclose(StreamLayout);
  return(1);
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "IsRoutedLayout()"                                   | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| This function checks if a layout is just a placement or if             |%
 * %| it was already routed.                                                 |%
 * %|   The choice is take while looking if all terminals are                |%
 * %| connected to segments or instance connectors.                          |%
 */

static long  IsRoutedLayout(aPhFig)
  phfig_list *aPhFig;
{
# define  access_SegAB  ((phseg_list*)(PPtSegAB->DATA))
# define  access_InsAB  ((phins_list*)(PPtInsAB->DATA))
# define  mc_IsOnHSeg(x,y,seg)  (   ((y) == (seg)->Y1)  \
                                 && ((x) >= (seg)->X1)  \
                                 && ((x) <= (seg)->X2))
# define  mc_IsOnVSeg(x,y,seg)  (   ((x) == (seg)->X1)  \
                                 && ((y) >= (seg)->Y1)  \
                                 && ((y) <= (seg)->Y2))

  phfig_list *PtPhFig;
  phins_list *PtPhIns;
  phseg_list *PtPhSeg;
  phcon_list *PtPhCon, *PtPhInsCon;
  chain_list *SegAB, *PPtSegAB, *PPtInsAB,
             *SegAlu1AB = (chain_list*)NULL,
             *SegAlu2AB = (chain_list*)NULL,
             *SegAlu3AB = (chain_list*)NULL,
                 *InsAB = (chain_list*)NULL;
        long  XAB1 = aPhFig->XAB1,
              XAB2 = aPhFig->XAB2,
              YAB1 = aPhFig->YAB1,
              YAB2 = aPhFig->YAB2;
        long  IsRoutedLayout = TRUE,
              IsConnected;

  /* Check if there is physical connectors. */
  if (aPhFig->PHCON == (phcon_list*)NULL) return(FALSE);
  /* Create the list of all segments that touch the abutment-box.
   * One list for each routing layer (today: ALU1, ALU2 and ALU3).
   * This increase the search speed.
   */
  for(PtPhSeg  =  aPhFig->PHSEG;
      PtPhSeg != (phseg_list*)NULL;
      PtPhSeg  =  PtPhSeg->NEXT) {
    switch(PtPhSeg->TYPE) {
      case UP:
      case DOWN:
        if (             (PtPhSeg->X1 <  XAB1)
            ||           (PtPhSeg->X1 >  XAB2)
            || (!(   (   (PtPhSeg->Y2 >= YAB1)
                      && (PtPhSeg->Y1 <= YAB1))
                  || (   (PtPhSeg->Y2 >= YAB2)
                      && (PtPhSeg->Y1 <= YAB2))) )) continue;
        break;
      case LEFT:
      case RIGHT:
        if (             (PtPhSeg->Y1 <  YAB1)
            ||           (PtPhSeg->Y1 >  YAB2)
            || (!(   (   (PtPhSeg->X2 >= XAB1)
                      && (PtPhSeg->X1 <= XAB1))
                  || (   (PtPhSeg->X2 >= XAB2)
                      && (PtPhSeg->X1 <= XAB2))) )) continue;
        break;
    }
    switch(PtPhSeg->LAYER) {
      case ALU1: SegAlu1AB = addchain(SegAlu1AB, (void*)PtPhSeg); break;
      case ALU2: SegAlu2AB = addchain(SegAlu2AB, (void*)PtPhSeg); break;
      case ALU3: SegAlu2AB = addchain(SegAlu2AB, (void*)PtPhSeg); break;
    }
  }
  /* Create the list of all instances that touch the abutment-box,
   * in order to increase the search speed.
   */
  for(PtPhIns  =  aPhFig->PHINS;
      PtPhIns != (phins_list*)NULL;
      PtPhIns  =  PtPhIns->NEXT) {
    PtPhFig = getphfig(PtPhIns->FIGNAME, 'P');
    UtFlt_XYFlatPhFig(PtPhFig, PtPhIns->XINS,
                   PtPhIns->YINS,
                   PtPhIns->TRANSF);
    if (   (PtPhFig->XAB1 == XAB1)
        || (PtPhFig->XAB1 == XAB2)
        || (PtPhFig->XAB2 == XAB1)
        || (PtPhFig->XAB2 == XAB2)
        || (PtPhFig->YAB1 == YAB1)
        || (PtPhFig->YAB1 == YAB2)
        || (PtPhFig->YAB2 == YAB1)
        || (PtPhFig->YAB2 == YAB2))
      InsAB = addchain(InsAB, (void*)PtPhIns);
  }
  /* Attempt to match each father connector with a segment extremity or
   * a connector of an embedded instance. If at least one terminal is
   * neither connected to a segment nor a terminal of an instance,
   * the function return that the father is NOT routed.
   */
  for( PtPhCon  = aPhFig->PHCON;
      (PtPhCon != (phcon_list*)NULL) && IsRoutedLayout;
       PtPhCon  = PtPhCon->NEXT) {
    if (   (PtPhCon->XCON != XAB1)
        && (PtPhCon->XCON != XAB2)
        && (PtPhCon->YCON != YAB1)
        && (PtPhCon->YCON != YAB2)) {
      GfpUT_PrintWarng( GlobVar.ModuleId, IW_TerminalLocation,
                        PtPhCon->NAME,
                        PtPhCon->XCON / SCALE_X,
                        PtPhCon->YCON / SCALE_X,
                         aPhFig->NAME );
      continue;
    }

    IsConnected = FALSE;
    SegAB = (chain_list*)NULL;
    switch(PtPhCon->LAYER) {
      case ALU1: SegAB = SegAlu1AB; break;
      case ALU2: SegAB = SegAlu2AB; break;
      case ALU3: SegAB = SegAlu2AB; break;
    }
    /* Look for a segment with an extremity matching connector
     * coordinates.
     */
    for( PPtSegAB  = SegAB;
        (PPtSegAB != (chain_list*)NULL) && !IsConnected;
         PPtSegAB  = PPtSegAB->NEXT)
      switch(access_SegAB->TYPE) {
        case UP:
        case DOWN:
          if (mc_IsOnVSeg(PtPhCon->XCON, PtPhCon->YCON, access_SegAB))
            IsConnected = TRUE;
          break;
        case LEFT:
        case RIGHT:
          if (mc_IsOnHSeg(PtPhCon->XCON, PtPhCon->YCON, access_SegAB))
            IsConnected = TRUE;
          break;
      }
    if (IsConnected) {
      mprintf(
        "IsRoutedLayout(): Matched \"%s\"(%ld,%ld) <---> Segment.\n",
        PtPhCon->NAME, PtPhCon->XCON / SCALE_X,
                       PtPhCon->YCON / SCALE_X );
      continue;
    }
    /* Look for a connector of same type and same place in the
     * instance below, if any.
     */
    for( PPtInsAB  = InsAB;
        (PPtInsAB != (chain_list*)NULL) && !IsConnected;
         PPtInsAB  = PPtInsAB->NEXT) {
      PtPhFig = getphfig(access_InsAB->FIGNAME, 'P');
      UtFlt_XYFlatPhFig(PtPhFig, access_InsAB->XINS,
                                 access_InsAB->YINS,
                                 access_InsAB->TRANSF);
      /* Test if the terminal is inside the instance abutment-box. */
      if (   (PtPhFig->XAB1 <= PtPhCon->XCON)
          && (PtPhFig->XAB2 >= PtPhCon->XCON)
          && (PtPhFig->YAB1 <= PtPhCon->YCON)
          && (PtPhFig->YAB2 >= PtPhCon->YCON))
        for(PtPhInsCon  = PtPhFig->PHCON;
            PtPhInsCon != (phcon_list*)NULL;
            PtPhInsCon  = PtPhInsCon->NEXT)
          if (   (PtPhInsCon->XCON   == PtPhCon->XCON  )
              && (PtPhInsCon->YCON   == PtPhCon->YCON  )
              && (PtPhInsCon->WIDTH  == PtPhCon->WIDTH )
              && (PtPhInsCon->ORIENT == PtPhCon->ORIENT)
              && (PtPhInsCon->LAYER  == PtPhCon->LAYER )) {
            IsConnected = TRUE;
            mprintf(
              "IsRoutedLayout(): Matched \"%s\"(%ld,%ld) <---> \"%s\"<%s>.\n",
              PtPhCon->NAME, PtPhCon->XCON / SCALE_X,
                             PtPhCon->YCON / SCALE_X,
              access_InsAB->INSNAME, access_InsAB->FIGNAME );
            break;
          }
      mprintf(
        "IsRoutedLayout():   \"%s\"(%ld,%ld) <-X-> \"%s\"<%s>.\n",
        PtPhCon->NAME, PtPhCon->XCON / SCALE_X,
                       PtPhCon->YCON / SCALE_X,
        access_InsAB->INSNAME, access_InsAB->FIGNAME );
    }
    if (!IsConnected) {
      IsRoutedLayout = FALSE;
      mprintf(
        "IsRoutedLayout(): Unmatched \"%s\"(%ld,%ld).\n",
        PtPhCon->NAME, PtPhCon->XCON / SCALE_X,
                       PtPhCon->YCON / SCALE_X );
    }
  }
  /* Unflatten all the models of the instances. */
  for(PtPhIns  =  aPhFig->PHINS;
      PtPhIns != (phins_list*)NULL;
      PtPhIns  =  PtPhIns->NEXT) {
    PtPhFig = getphfig(PtPhIns->FIGNAME, 'P');
    UtFlt_XYUnFlatPhFig(PtPhFig);
  }

  freechain(SegAlu1AB);
  freechain(SegAlu2AB);
  freechain(SegAlu3AB);
  freechain(    InsAB);

  return(IsRoutedLayout);

# undef  access_SegAB
# undef  access_InsAB
# undef  mc_IsOnHSeg
# undef  mc_IsOnVSeg
}


/*  Function  :  'IsModelUS'                                                */
/* ======================================================================== */
/*  Attempt to recognize a User column model name.                          */

static long  IsModelUS( aName )
  char *aName;
{
  char *Ext, *Pt;


  for( Pt = aName, Ext = "" ; *Pt != (char)0 ; Pt++ )
    if ( *Pt == '_' ) Ext = Pt + 1;

  return( ( !strcmp( Ext, "us" ) ) ? TRUE : FALSE );
}


/*  Function  :  'IsModelCL'                                                */
/* ======================================================================== */
/*  Attempt to recognize a FpGen column model name.                         */

static long  IsModelCL( aName )
  char *aName;
{
   int  NB;
  char  Root[   256 ];
  char  Params[ 256 ];
  char  Index[  256 ];
  char  Ext[    256 ];


  NB = sscanf( aName, "%[a-z0-9]_%[0-9x]%[lm]_%s", Root, Params, Index, Ext );

  return( ( ( NB == 4 ) && ( !strcmp( Ext, "cl" ) ) ) ? TRUE : FALSE );
}


/*  Function  :  'IsModelBK'                                                */
/* ======================================================================== */
/*  Attempt to recognize a FpGen block model name.                          */

static long  IsModelBK( aName )
  char *aName;
{
   int  NB;
  char  Root[   256 ];
  char  Params[ 256 ];
  char  Index[  256 ];
  char  Ext[    256 ];


  NB = sscanf( aName, "%[a-z0-9]_%[0-9x]%[lm]_%s", Root, Params, Index, Ext );

  return( ( ( NB == 4 ) && ( !strcmp( Ext, "bk" ) ) ) ? TRUE : FALSE );
}


/*  Function  :  'IsOperatorUser'                                           */
/* ======================================================================== */
/*  Checks if the logical figure "PtLoFig" is made with FpGen columns or
 *  whith DPLIB leaf cells.
 *      In the last case, the model is assumed to be a user defined column,
 *  using the DPLIB library or a synthetized bloc.
 *      If some models are FpGen columns and some others not, an error
 *  occurs.
 *
 *  NOTICE: In release 1.0 this analysis was based on the CATAL, but we
 *          we decide to put also in this file the FpGen operators.
 */

static long  IsOperatorUser(PtLoFig)
  lofig_list *PtLoFig;
{
  loins_list *PtLoIns;
        long  FlagModelType;

  if ( PtLoFig->LOINS == (loins_list *)NULL )
    GfpUT_PrintError( GlobVar.ModuleId, IE_OperatorEmpty, PtLoFig->NAME );

  /* Sets "FlagNodelType" whith the type of the first instance,
   * then checks if all others are of the same type.
   */
  FlagModelType =    IsModelCL( PtLoFig->LOINS->FIGNAME )
                  || IsModelBK( PtLoFig->LOINS->FIGNAME )
                  || IsModelUS( PtLoFig->LOINS->FIGNAME );

  for( PtLoIns  = PtLoFig->LOINS->NEXT ;
       PtLoIns !=   (loins_list *)NULL ;
       PtLoIns  =        PtLoIns->NEXT )
    if ( FlagModelType != (    IsModelCL( PtLoIns->FIGNAME )
                            || IsModelBK( PtLoIns->FIGNAME )
                            || IsModelUS( PtLoIns->FIGNAME ) ) )
      GfpUT_PrintError( GlobVar.ModuleId,
                        IE_OperatorMixed,
                        PtLoFig->NAME );

  return( FlagModelType ? FALSE : TRUE );
}


/*  Function  :  'UtHir_GetModelType'                             */
/* -------------------------------------------------------------- */
/*  This function return the type of the model given as argument.
 *      This is one of the four following values:
 *              - C_UtHir_TypeOperatorRouted
 *              - C_UtHir_TypeOperatorPlaced
 *              - C_UtHir_TypeOperatorUser
 *              - C_UtHir_TypeOperatorSubDP
 *      For one given figure, the analysis is performed only one
 */

extern long  UtHir_GetModelType(aModelName)
	char *aModelName;
{
	lofig_list *PtLoFig;
		  char  ModelType;

	/* Check file type on the disk. */
	ModelType = (HaveStruct(aModelName)) ? 'P' : 'A';

	/* Get the logical figure whith the appropriate mode. */
	PtLoFig = getlofig(aModelName, ModelType);

	return(UtHir_GetOperatorType(PtLoFig));
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "UtHir_GetOperatorType()"                            | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| This function return the type of the logical figure given as argument. |%
 * %|   This is one of the four following values :                           |%
 * %|     - C_UtHir_TypeOperatorRouted                                       |%
 * %|     - C_UtHir_TypeOperatorPlaced                                       |%
 * %|     - C_UtHir_TypeOperatorUser                                         |%
 * %|     - C_UtHir_TypeOperatorSubDP                                        |%
 * %|   For one given figure, the analysis is performed only one time.       |%
 * %| The result is stored in the USER list.                                 |%
 */

extern long  UtHir_GetOperatorType( aLoFig )
  lofig_list *aLoFig;
{
  ulofig_flags *PtLoFigFlags;
    ptype_list *PtUser;
          long *PtNew;

  PtUser = GfpUT_GetUser(
             &(aLoFig->USER), C_GfpUT_UserTypeOperator, F_GfpUT_UserCreat );
  /* Look if the figure have been already checked.
   * If so, return the previously computed type.
   */
  if (PtUser->DATA == (void *)NULL) {
    PtNew        = (long *)mbkalloc( sizeof( long ) );
    PtUser->DATA = (void *)PtNew;
  } else {
    return( *(long *)( PtUser->DATA ) );
  }
  /* Recognition of routed or placed blocks. */
  if (HaveLayout(aLoFig->NAME)) {
    if (IsRoutedLayout(getphfig(aLoFig->NAME, 'A'))) {
      return( *PtNew = C_UtHir_TypeOperatorRouted );
    } else
      if ((PtLoFigFlags = UtFDP_GetLoFigFlags(aLoFig))
                                    != (ulofig_flags *)NULL) {
        if (PtLoFigFlags->FLAGS & F_UtFDP_FlagLoFigKeep)
          return(*PtNew = C_UtHir_TypeOperatorPlaced);
      }
  }
  /* Recognition of user defined blocks. */
  if (IsModelUS(aLoFig->NAME))
    return(*PtNew = C_UtHir_TypeOperatorUser);
  else
    return(*PtNew = C_UtHir_TypeOperatorSubDP);
}


#ifdef  UNDEFINED
/*  Function  :  'UtHir_HierarchCheck'                            */
/* -------------------------------------------------------------- */
/*  Check the hierarchical consistency of a data-path.
 */

extern void  UtHir_HierarchCheck(ptlofig)
	lofig_list *ptlofig;
{
	lofig_list *LoFigFather;
	lofig_list *LoFigSon;
	loins_list *LoInsSon;
	      char  LoFigSonMode;

	LoFigFather = ptlofig;
	LoInsSon    =   LoFigFather->LOINS;

	/* Reset logical stack and push the stop value. */
	ResetLoStack();
	PushLoState((lofig_list*)NULL, (loins_list*)NULL );

	/* Perform the first check of the reccursion.      */

	/* If a cluster is detect, do not attempt to check
	   the hierarchy. */
	if (IsModelCluster(LoFigFather)) return;

	/* (if a column is detect, terminate the function) */
	if (IsModelColumn(LoFigFather)) return;

	/* Recursion main loop. */
	while(LoFigFather != (lofig_list*)NULL)
	{
	    /* If all instances have been checked, return to
		   the upper levels. */
		if ( LoInsSon == (loins_list*)NULL )
		{
			PopLoState(&LoFigFather, &LoInsSon);
			continue;
		}

		/* Check file type on the disk. */
		LoFigSonMode = (IsPhantom(LoInsSon->FIGNAME)) ? 'P' : 'A';

		/* Get the model of the instance. */
		LoFigSon = getlofig( LoInsSon->FIGNAME, LoFigSonMode);

		/* Analisys of instance model type. */
		if (   !IsModelCluster(LoFigSon)
			&& !IsModelColumn( LoFigSon)
			&& (LoFigSon->LOINS != (loins_list*)NULL))
		{
			PushLoState(LoFigFather, LoInsSon->NEXT);

			LoFigFather = LoFigSon;
			LoInsSon    = LoFigSon->LOINS;
		}
		else
			LoInsSon = LoInsSon->NEXT;
		}
}
#endif


/*  Function  :  'UtHir_HierarchCheck'                            */
/* -------------------------------------------------------------- */
/*  Check the hierarchical consistency of a data-path.
 */

extern void  UtHir_HierarchCheck(aLoFig)
	lofig_list *aLoFig;
{
	lofig_list *LoFigFather;
	lofig_list *LoFigSon;
	loins_list *LoInsSon;
	      char  LoFigSonMode;

	LoFigFather = aLoFig;
	LoInsSon    = LoFigFather->LOINS;

	/* Reset logical stack and push the stop value. */
	ResetLoStack();
	PushLoState((lofig_list*)NULL, (loins_list*)NULL );

	/* Perform the first check of the recursion.
	 * We perform a full hierarchical check only if
	 * the first hierarchical level is a "SubDP".
	 */
	switch(UtHir_GetOperatorType(aLoFig))
	{
		case C_UtHir_TypeOperatorUser:   return;
		case C_UtHir_TypeOperatorRouted: ;
		case C_UtHir_TypeOperatorPlaced: ;
		case C_UtHir_TypeOperatorSubDP:  ;
	}

	/* Recursion main loop. */
	while(LoFigFather != (lofig_list*)NULL)
	{
	    /* If all instances have been checked, return to
		   the upper levels. */
		if ( LoInsSon == (loins_list*)NULL )
		{
			PopLoState(&LoFigFather, &LoInsSon);
			continue;
		}

		/* Check file type on the disk. */
		LoFigSonMode = (HaveStruct(LoInsSon->FIGNAME)) ? 'P' : 'A';

		/* Get the model of the instance. */
		LoFigSon = getlofig(LoInsSon->FIGNAME, LoFigSonMode);

		/* Analysis of instance model type. */
		if (   (UtHir_GetOperatorType(LoFigSon) == C_UtHir_TypeOperatorSubDP)
			&& (LoFigSon->LOINS != (loins_list*)NULL))
		{
			PushLoState(LoFigFather, LoInsSon->NEXT);

			LoFigFather = LoFigSon;
			LoInsSon    = LoFigSon->LOINS;
		}
		else
			LoInsSon = LoInsSon->NEXT;
		}
}


/*  Function  :  'UtHir_HierarchFlattenLo'                        */
/* -------------------------------------------------------------- */
/*  Normalize a data-path hierarchy from 'n' levels down to 2.
 *      The order of the instances is significant for the initial
 *  placement step of DPR. The flattening function 'flattenlofig'
 *  add the new instances in end of the LOINS list. So, we must
 *  re-order this list after each flattenning operation.
 *      The MBK separator is temporarily set to '_'. It is neces-
 *  sary for hiding the flattened levels to the placer (the
 *  placer recognize only two levels of hierarchy, with the def-
 *  fault MBK separator '.').
 */

extern void  UtHir_HierarchFlattenLo(aLoFig)
	lofig_list *aLoFig;
{
	lofig_list     *LoFigFather;
	lofig_list     *LoFigSon;
	loins_list **PPtLoInsSon;
	loins_list **PPtLoInsLast;
	loins_list   *PtLoInsTmp;
	      char      LoFigSonMode;
		  char      OldSepar;

	LoFigFather = aLoFig;

	/* For the sub-data-path flattens, the separator to use is '_'.
	 * This constraints belong to the DPR placer.
	 */
	OldSepar = SEPAR; SEPAR = '_';

	/* if a leaf cell is detect, terminate the function. */
	if (IsModelLeafCell(LoFigFather)) 
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_FlattenLeafCell,
						 "HierarchFlatten",
						 LoFigFather->NAME);
		return;
	}

	/* Perform the first check of the recursion.
	 * We perform a full hierarchical flatten only if
	 * the first hierarchical level is a "SubDP".
	 */
	switch(UtHir_GetOperatorType(aLoFig))
	{
		case C_UtHir_TypeOperatorUser:   return;
		case C_UtHir_TypeOperatorRouted: ;
		case C_UtHir_TypeOperatorPlaced: ;
		case C_UtHir_TypeOperatorSubDP:  ;
	}


	/* Process the complete list of LOINS. */
	for( PPtLoInsSon  = &LoFigFather->LOINS;
		*PPtLoInsSon !=   (loins_list*)NULL;
		 PPtLoInsSon  = &((*PPtLoInsSon)->NEXT))
	{
		/* Looking for the last instance. */
		for( PPtLoInsLast  = &LoFigFather->LOINS;
			*PPtLoInsLast !=   (loins_list*)NULL;
			 PPtLoInsLast  = &((*PPtLoInsLast)->NEXT));

		/* Check file type on the disk. */
		LoFigSonMode = (HaveStruct((*PPtLoInsSon)->FIGNAME)) ? 'P' : 'A';

		/* Get the model of the instance. */
		LoFigSon = getlofig((*PPtLoInsSon)->FIGNAME, LoFigSonMode);

		/* Flatten all instance that are recognized
		 * as sub-data-path.
		 */
		if (    (UtHir_GetOperatorType(LoFigSon) == C_UtHir_TypeOperatorSubDP)
			 && ((loins_list*)NULL != LoFigSon->LOINS))
		{
			if ((*PPtLoInsSon)->NEXT == *PPtLoInsLast)
			{
				/* The flattened instance is itself the head of list,
				 * thus no reordering is necessary.
				 */
				flattenlofig(LoFigFather, (*PPtLoInsSon)->INSNAME, YES);
			}
			else
			{
				flattenlofig(LoFigFather, (*PPtLoInsSon)->INSNAME, YES);

				/* Searching the end of the newly inserted instances.
				 * (done by retrieving the previous loins list head )
				 */
				for(PtLoInsTmp        = *PPtLoInsLast    ;
					PtLoInsTmp->NEXT != (loins_list*)NULL;
					PtLoInsTmp        =  PtLoInsTmp->NEXT);

				/* Moving the new instances at the rigth place.
				 * (by swapping some next pointer of the list) 
				 */
				  PtLoInsTmp->NEXT  = *PPtLoInsSon;
				*PPtLoInsSon        = *PPtLoInsLast;
				*PPtLoInsLast       = (loins_list*)NULL;
			}
		}
	}

	/* Restoring the previous value of the MBK separator. */
	SEPAR = OldSepar;
}


/*  Function  :  'UtHir_rFlattenCatalPhFig'                       */
/* -------------------------------------------------------------- */

extern void  UtHir_rFlattenCatalPhFig(aPhFig)
	phfig_list *aPhFig;
{
	phfig_list     *PhFigFather;
	phfig_list     *PhFigSon;
	phins_list **PPtPhInsSon;
	phins_list   *PtPhInsFirst;
	phins_list   *PtPhInsTmp;
		  char      OldSepar;

	PhFigFather = aPhFig;

	/* For the sub-data-path flattens, the separator to use is '_'.
	 * This constraints bePhng to the DPR placer.
	 */
	OldSepar = SEPAR; /* SEPAR = '_'; */

	/* if a leaf cell is detect, terminate the function. */
	if (   IsModelCL( PhFigFather->NAME )
        || IsModelBK( PhFigFather->NAME )
        || IsModelUS( PhFigFather->NAME ) ) 
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_FlattenLeafCell,
						 "rFlattenCatalPhFig",
						 PhFigFather->NAME);
		return;
	}

	/* Perform the first check of the recursion.
	 * We perform a full hierarchical flatten only if
	 * the first hierarchical level is not routed.
	 */
	if (IsRoutedLayout(PhFigFather)) return;


	/* Process the complete list of PHINS. */
	for( PPtPhInsSon  = &PhFigFather->PHINS;
		*PPtPhInsSon !=   (phins_list*)NULL;
		 PPtPhInsSon  = &((*PPtPhInsSon)->NEXT))
	{
		/* Get the model of the instance. */
		PhFigSon = getphfig((*PPtPhInsSon)->FIGNAME, 'A');

		/* Flatten all instance that are recognized
		 * as sub-data-path.
		 */
		if (    (     !IsRoutedLayout(PhFigSon)      )
			 && ((phins_list*)NULL != PhFigSon->PHINS))
		{
			if ((*PPtPhInsSon) == PhFigFather->PHINS)
			{
				/* The flattened instance is itself the head of list,
				 * thus no reordering is necessary.
				 */
				flattenphfig(PhFigFather, (*PPtPhInsSon)->INSNAME, YES);
			}
			else
			{
				/* Storing the old head of PHINS list. */
				PtPhInsFirst = PhFigFather->PHINS;

				flattenphfig(PhFigFather, (*PPtPhInsSon)->INSNAME, YES);

				/* Searching the end of the newly inserted instances.
				 * (done by retrieving the previous phins list head )
				 */
				for(PtPhInsTmp        =    PhFigFather->PHINS;
					PtPhInsTmp->NEXT !=  PtPhInsFirst;
					PtPhInsTmp        =  PtPhInsTmp->NEXT);

				/* Moving the new instances at the rigth place.
				 * (by swapping some next pointer of the list) 
				 */
				  PtPhInsTmp->NEXT     = *PPtPhInsSon;
				*PPtPhInsSon           =     PhFigFather->PHINS;
				    PhFigFather->PHINS =   PtPhInsFirst;
			}
		}
	}

	/* Restoring the previous value of the MBK separator. */
	/* SEPAR = OldSepar; */
}


/*  Function  :  'UtHir_rFlattenCatalLoFig'                       */
/* -------------------------------------------------------------- */

extern void  UtHir_rFlattenCatalLoFig(aLoFig)
	lofig_list *aLoFig;
{
	lofig_list     *LoFigFather;
	lofig_list     *LoFigSon;
	loins_list **PPtLoInsSon;
	loins_list **PPtLoInsLast;
	loins_list   *PtLoInsTmp;
		  char      LoFigSonMode;
		  char      OldSepar;

	LoFigFather = aLoFig;

	/* For the sub-data-path flattens, the separator to use is '_'.
	 * This constraints beLong to the DPR placer.
	 */
	OldSepar = SEPAR; /* SEPAR = '_'; */

	/* if a leaf cell is detect, terminate the function. */
	if (   IsModelCL( LoFigFather->NAME ) 
        || IsModelBK( LoFigFather->NAME ) 
        || IsModelUS( LoFigFather->NAME ) ) 
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_FlattenLeafCell,
						 "rFlattenCatalLoFig",
						 LoFigFather->NAME);
		return;
	}

	/* Perform the first check of the recursion.
	 * We perform a full hierarchical flatten only if
	 * the first hierarchical level is not routed.
	 */
	if (IsRoutedLayout(getphfig(LoFigFather->NAME,'A'))) return;


	/* Process the complete list of LOINS. */
	for( PPtLoInsSon  = &LoFigFather->LOINS;
		*PPtLoInsSon !=   (loins_list*)NULL;
		 PPtLoInsSon  = &((*PPtLoInsSon)->NEXT))
	{
		/* Looking for the last instance. */
		for( PPtLoInsLast  = &LoFigFather->LOINS;
			*PPtLoInsLast !=   (loins_list*)NULL;
			 PPtLoInsLast  = &((*PPtLoInsLast)->NEXT));

		/* Check file type on the disk. */
		LoFigSonMode = (HaveStruct((*PPtLoInsSon)->FIGNAME)) ? 'P' : 'A';

		/* Get the model of the instance. */
		LoFigSon = getlofig((*PPtLoInsSon)->FIGNAME, LoFigSonMode);

		/* Flatten all instance that are recognized
		 * as sub-data-path.
		 */
		if (    (!IsRoutedLayout(getphfig(LoFigSon->NAME,'A')))
			 && (    (loins_list*)NULL != LoFigSon->LOINS     ))
		{
			if ((*PPtLoInsSon)->NEXT == *PPtLoInsLast)
			{
				/* The flattened instance is itself the head of list,
				 * thus no reordering is necessary.
				 */
				flattenlofig(LoFigFather, (*PPtLoInsSon)->INSNAME, YES);
			}
			else
			{
				flattenlofig(LoFigFather, (*PPtLoInsSon)->INSNAME, YES);

				/* Searching the end of the newly inserted instances.
				 * (done by retrieving the previous loins list head )
				 */
				for(PtLoInsTmp        = *PPtLoInsLast    ;
					PtLoInsTmp->NEXT != (loins_list*)NULL;
					PtLoInsTmp        =  PtLoInsTmp->NEXT);

				/* Moving the new instances at the rigth place.
				 * (by swapping some next pointer of the list) 
				 */
				  PtLoInsTmp->NEXT  = *PPtLoInsSon;
				*PPtLoInsSon        = *PPtLoInsLast;
				*PPtLoInsLast       = (loins_list*)NULL;
			}
		}
	}

	/* Restoring the previous value of the MBK separator. */
	/* SEPAR = OldSepar; */
}


#ifdef  UNDEFINED
/*  Function  :  'UtHir_HierarchFlatten'                          */
/* -------------------------------------------------------------- */
/*  Normalize a data-path hierarchy from 'n' levels down to 2.
 *      The order of the instances is significant for the initial
 *  placement step of DPR. The flattening function 'flattenlofig'
 *  add the new instances in end of the LOINS list. So, we must
 *  re-order this list after each flattenning operation.
 *      The MBK separator is temporarily set to '_'. It is neces-
 *  sary for hiding the flattened levels to the placer (the
 *  placer recognize only two levels of hierarchy, with the def-
 *  fault MBK separator '.').
 */

extern void  UtHir_HierarchFlatten(aLoFig)
	lofig_list *aLoFig;
{
	lofig_list     *LoFigFather;
	lofig_list     *LoFigSon;
	loins_list **PPtLoInsSon;
	loins_list **PPtLoInsLast;
	loins_list   *PtLoInsTmp;
		  char      LoFigSonMode;
		  char      OldSepar;

	LoFigFather = aLoFig;

	/* For the sub-data-path flattens, the separator to use is '_'. */
	/* This constraints belong to the DPR placer.                   */
	OldSepar = SEPAR; SEPAR = '_';

	/* if a leaf cell is detect, terminate the function. */
	if (incatalog(LoFigFather)) 
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_FlattenLeafCell,
						 "HierarchFlatten",
						 LoFigFather->NAME);
		return;
	}

	/* if a column is detect, terminate the function. */
	if (IsModelColumn(LoFigFather))
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_FlattenColumn,
									  "HierarchFlatten",
									  LoFigFather->NAME);
		return;
	}


	/* Process the complete list of LOINS. */
	for( PPtLoInsSon  = &LoFigFather->LOINS;
		*PPtLoInsSon !=   (loins_list*)NULL;
		 PPtLoInsSon  = &((*PPtLoInsSon)->NEXT))
	{
		/* Looking for the last instance. */
		for( PPtLoInsLast  = &LoFigFather->LOINS;
			*PPtLoInsLast !=   (loins_list*)NULL;
			 PPtLoInsLast  = &((*PPtLoInsLast)->NEXT));

	    /* Check file type on the disk. */
		LoFigSonMode = (IsPhantom((*PPtLoInsSon)->FIGNAME)) ? 'P' : 'A';

		/* Get the model of the instance. */
		LoFigSon = getlofig((*PPtLoInsSon)->FIGNAME, LoFigSonMode);

		/* Flatten all instance that are not recognized */
		/* as columns or clusters.                      */
		if (      !IsModelCluster(LoFigSon)
			 &&   !IsModelColumn( LoFigSon)
			 && ((loins_list*)NULL != LoFigSon->LOINS))
		{
			if ((*PPtLoInsSon)->NEXT == *PPtLoInsLast)
			{
				/* The flattened instance is itself the head of list, */
				/* thus no reordering is necessary.                   */
				flattenlofig(LoFigFather, (*PPtLoInsSon)->INSNAME, YES);
			}
			else
			{
				flattenlofig(LoFigFather, (*PPtLoInsSon)->INSNAME, YES);

				/* Searching the end of the newly inserted instances. */
				/* (done by retrieving the previous loins list head ) */
				for(PtLoInsTmp        = *PPtLoInsLast    ;
					PtLoInsTmp->NEXT != (loins_list*)NULL;
					PtLoInsTmp        =  PtLoInsTmp->NEXT);

				/* Moving the new instances at the rigth place. */
				/* (by swapping some next pointer of the list)  */
				  PtLoInsTmp->NEXT  = *PPtLoInsSon;
				*PPtLoInsSon        = *PPtLoInsLast;
				*PPtLoInsLast       = (loins_list*)NULL;
			}
		}
	}

	/* Restoring the previous value of the MBK separator. */
	SEPAR = OldSepar;
}
#endif


/*  Function  :  'UtHir_LoadLoFig'                                */
/* -------------------------------------------------------------- */

extern lofig_list *UtHir_LoadLoFig(aModelName)
	char *aModelName;
{
	lofig_list *PtLoFig;
	      char    LoFigMode;

	/* Check file type on the disk. */
	LoFigMode = (HaveStruct(aModelName)) ? 'P' : 'A';

	/* Get the model of the instance. */
	PtLoFig = getlofig(aModelName, LoFigMode);

	/* Check the specific data-path hierarchy consistency. */
	UtHir_HierarchCheck(PtLoFig);

	return(PtLoFig);
}
