/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtMBKLo.c,v 1.10 1997/10/30 13:28:02 fitpath Exp fitpath $ */
/*                                                                */
/*          Alliance CAD System 3.2b                               */
/*             FpGen Package 1.10                                 */
/*                                                                */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                       */
/*                                                                */
/*  Author      :          Jean-Paul CHAPUT                       */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                       */
/* ************************************************************** */
/*  Module      :      "UtMBKLo.c"                                */


/* ************************** Includes ************************** */


#include  <ctype.h>
#include  <string.h>

#include   MUT_H
#include   MLO_H

#include  "GenericGN.h"


/* ********************* Internal constants ********************* */


#define   IW_MBKSigNameMissOpenBracket   2
#define   IW_MBKSigNameMissClosBracket   3
#define   IW_MBKSigNameDuplOpenBracket   4
#define   IW_MBKSigNameDuplClosBracket   5
#define   IW_MBKSigNameInvalidChar       6
#define   IW_MBKSigNameMissRoot          7
#define   IW_MBKIndexMissOrig            8
#define   IW_MBKIndexMissExtr            9
#define   IW_MBKIndexSpaceOrig          10
#define   IW_MBKIndexSpaceExtr          11
#define   IW_MBKIndexInvalidChar        12

#define   IE_MBKSigNameSpace             2
#define   IE_MBKSigNameOpenBracket       3
#define   IE_MBKSigNameCloseBracket      4

#define   IC_StringBufferSize         2048
#define   IS_TypeConImplicit            "ic"


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "UtMBKLo.c" module identifier. */
	} GlobVar;


/* ********************* Internal Variables ********************* */


/* ********************* External Variables ********************* */


       char  DP_SEPAR = '_';
       char  DP_IC[C_SUPP_MaxLenString] = "_ic";


/* ******************** Interface Functions ********************* */


/* extern       char *GfpUT_ConcatName(char *aName1, char *aName2);
 * extern       char *GfpUT_CheckName(char *pin_name);
 * extern       long  GfpUT_IsDecimalNumber(char *aStr);
 * extern       long  GfpUT_StrIndexDecod(char *aStr,
                                          long *aPtOrig,
                                          long *aPtExtr);
 * extern       long  GfpUT_StrSigDecod(char *aStr,
                                        char *aRoot,
                                        char *aPtOrig,
                                        char *aPtExtr);
 * extern       char *GfpUT_DecodMBKSigName(char *aMBKSigName,
                                            long *aPtSigIndex);
 * extern       char *GfpUT_MakeImplicitConName(char *aInstName,
                                                char *aConName ,
                                                long  aIndex   );
 * extern       long  GfpUT_TypeConName(char *aConName);
 * extern locon_list *GfpUT_AddLoConEnd(locon_list *aLoFig,
                                              char *aName ,
                                        losig_list *aLoSig,
                                              char  aDir  );
 * extern locon_list  Dut_GetLoConBySig(lofig_list *aLoFig,
 *                                      losig_list *aLoSig);
 */


/* ********************* Internal Functions ********************* */


/* ******************** Functions Definitions ******************* */


/*  Function  :  'UtXlg_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtMLo_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtMBKLo", F_GfpUT_ModuleAdd);

	/* Errors associated to the 'GfpUT_CheckName' function.
	 */
	GfpUT_AddError(GlobVar.ModuleId, IE_MBKSigNameSpace,
	"[%s]:Invalid signal name, space detect:\"%s\".\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_MBKSigNameOpenBracket,
	"[%s]:Invalid signal name, [ unmatched:\"%s\".\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_MBKSigNameCloseBracket,
	"[%s]:invalid signal name, ] unmatched:\"%s\".\n");

	/* Warnings assoited to the signal and index analysis.
	 */
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKSigNameMissOpenBracket,
	"[%s]:In signal name \"%s\", missing open bracket.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKSigNameMissClosBracket,
	"[%s]:In signal name \"%s\", missing close bracket.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKSigNameDuplOpenBracket,
	"[%s]:In signal name \"%s\", duplicate open bracket.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKSigNameDuplClosBracket,
	"[%s]:In signal name \"%s\", duplicate close bracket.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKSigNameInvalidChar,
	"[%s]:In signal root name \"%s\", invalid character \'%c\'.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKSigNameMissRoot,
	"[%s]:In signal name \"%s\", missing root name.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKIndexMissOrig,
	"[%s]:In index string \"%s\", missing origin value.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKIndexMissExtr,
	"[%s]:In index string \"%s\", missing extremity value.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKIndexSpaceOrig,
	"[%s]:In index string \"%s\", space detect in origin member.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKIndexSpaceExtr,
	"[%s]:In index string \"%s\", space detect in extremity member.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_MBKIndexInvalidChar,
	"[%s]:In index string \"%s\", invalid character \'%c\'.");
}


/*  Function  : 'GfpUT_ConcatName'                                */
/* -------------------------------------------------------------- */
/*  Concatenate two string using the "DP_SEPAR" variable. This
 *  function is dedicated to the making of the implicit connector
 *  names. In addition, it perform a "namealloc".
 */

extern char *GfpUT_ConcatName(aName1,aName2)
	char *aName1;
	char *aName2;
{
	char  StringBuffer[IC_StringBufferSize];

	sprintf(         StringBuffer, "%s%c%s", aName1, DP_SEPAR, aName2);
	return(namealloc(StringBuffer));
}


/*  Fonction  : 'GfpUT_CheckName'                                 */
/* -------------------------------------------------------------- */

extern char *GfpUT_CheckName(pin_name)
	char *pin_name;
{
	static char  StringBuffer[IC_StringBufferSize];

		   char *pt;
		   long   OpenBracket = 0;
		   long  CloseBracket = 0;

	/* Recopie de la chaine dans le tampon.       */
	/* (on ne touche pas aux chaines utilisateur) */
	strcpy( StringBuffer, pin_name);

	for( pt=StringBuffer; *pt!=(char)0; pt++)
		{
		/* Verification : recherche des espaces. */
		if (*pt == ' ')
			GfpUT_PrintError(GlobVar.ModuleId,
							 IE_MBKSigNameSpace,
										  "CheckName",
											 pin_name);

		/* Elimination des crochets ouvrants. */
		if (*pt == '[')
			{
			/* Verification des crochets ouvrants. */
			if (OpenBracket > 1)
				GfpUT_PrintError(GlobVar.ModuleId,
								 IE_MBKSigNameOpenBracket,
													"CheckName",
													   pin_name);
			else
				OpenBracket += 1;

			*pt = ' ';
			}

		/* Elimination des crochets fermants. */
		if (*pt == ']')
			{
			/* Verification des crochets ouvrants. */
			if (   (CloseBracket >               1)
				|| ( OpenBracket != CloseBracket+1))
				GfpUT_PrintError(GlobVar.ModuleId,
								 IE_MBKSigNameCloseBracket,
													 "CheckName",
														pin_name);
			else
				CloseBracket += 1;

			*pt = (char)0;
			}
		}

	/* Check if all brackets are correctly closed. */
	if (OpenBracket != CloseBracket)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_MBKSigNameCloseBracket,
											 "CheckName",
												pin_name);

	/* Renvoi de la chaine 'nettoyee'. */
	return( StringBuffer);
}


/*  Function  :  'GfpUT_IsDecimalNumber'                          */
/* -------------------------------------------------------------- */

extern long  GfpUT_IsDecimalNumber(aStr)
	char *aStr;
{
	long  DigitNB = 0L;
	char *PtChar  = aStr;


	if ((*PtChar == '+') || (*PtChar == '-')) PtChar++;

	for(DigitNB=0; *PtChar!=(char)0; PtChar++,DigitNB++)
		if (!isdigit(*PtChar)) return(FALSE);

	return((DigitNB) ? TRUE : FALSE);
}


/*  Function  :  'GfpUT_StrIndexDecod'                            */
/* -------------------------------------------------------------- */

extern long  GfpUT_StrIndexDecod(aStr, aPtOrig, aPtExtr)
	char *aStr;
	long *aPtOrig;
	long *aPtExtr;
{
	long  IdxType   = C_GfpUT_IdxTypeScalar;
	char *PtChar;
	char *PtStrOrig = (char*)NULL;
	char *PtStrExtr = (char*)NULL;
	char  TmpString[C_SUPP_MaxLenString];


	strcpy(TmpString, aStr);

	
	for(PtChar=TmpString; *PtChar!=(char)0; PtChar++)
	{
		if (isspace(*PtChar)) continue;
		if (isdigit(*PtChar))
		{
			if (PtStrOrig == (char*)NULL)
			{
				if (IdxType == C_GfpUT_IdxTypeBus)
				{
					GfpUT_PrintWarng(GlobVar.ModuleId,
									 IW_MBKIndexMissOrig,
											   "StrIndexDecod",
														  aStr);
					return(C_GfpUT_IdxTypeUnknow);
				}

				PtStrOrig = PtChar; 
				while(isdigit(*(PtChar+1))) PtChar++;
				continue;
			}
			else
			{
				if (IdxType != C_GfpUT_IdxTypeBus)
				{
					GfpUT_PrintWarng(GlobVar.ModuleId,
									 IW_MBKIndexSpaceOrig,
											    "StrIndexDecod",
														   aStr);
					return(C_GfpUT_IdxTypeUnknow);
				}

				if (PtStrExtr == (char*)NULL)
				{
					PtStrExtr = PtChar; 
					while(isdigit(*(PtChar+1))) PtChar++;
					continue;
				}
				else
				{
					GfpUT_PrintWarng(GlobVar.ModuleId,
									 IW_MBKIndexSpaceExtr,
											    "StrIndexDecod",
														   aStr);
					return(C_GfpUT_IdxTypeUnknow);
				}
			}
		}
		if (*PtChar == ':')
			{ IdxType = C_GfpUT_IdxTypeBus; continue; }

		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_MBKIndexInvalidChar,
									  "StrIndexDecod",
												 aStr,
											  *PtChar);
		return(C_GfpUT_IdxTypeUnknow);
	}


	if (PtStrOrig == (char*)NULL)
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_MBKIndexMissOrig,
								   "StrIndexDecod",
											  aStr);
		return(C_GfpUT_IdxTypeUnknow);
	}

	if (PtStrExtr == (char*)NULL)
	{
		if (IdxType == C_GfpUT_IdxTypeBus)
		{
			GfpUT_PrintWarng(GlobVar.ModuleId,
							 IW_MBKIndexMissExtr,
									   "StrIndexDecod",
												  aStr);
			return(C_GfpUT_IdxTypeUnknow);
		}
		else
			PtStrExtr = PtStrOrig;
	}

	*aPtOrig = atol(PtStrOrig);
	*aPtExtr = atol(PtStrExtr);

	return(IdxType);
}


/*  Function  :  'GfpUT_StrSigDecod'                              */
/* -------------------------------------------------------------- */

extern long  GfpUT_StrSigDecod(aStr, aRoot, aPtOrig, aPtExtr)
	char *aStr;
	char *aRoot;
	long *aPtOrig;
	long *aPtExtr;
{
	long  SigType       = C_GfpUT_SigTypeScalar;
	char *PtChar;
	char *PtOpenBracket = (char*)NULL;
	char *PtClosBracket = (char*)NULL;
	char  TmpString[C_SUPP_MaxLenString];


	strcpy(TmpString, aStr);

	for(PtChar=TmpString; *PtChar!=(char)0; PtChar++)
	{
		if  (*PtChar == '[')
		{
			if (PtOpenBracket != (char*)NULL)
			{
				GfpUT_PrintWarng(GlobVar.ModuleId,
								 IW_MBKSigNameDuplOpenBracket,
								 					  "StrSigDecod",
															   aStr);
				return(C_GfpUT_SigTypeUnknow);
			}

			PtOpenBracket = PtChar+1;
			*PtChar = (char)0;
			continue;
		}
		if  (*PtChar == ']')
		{
			if (PtOpenBracket == (char*)NULL)
			{
				GfpUT_PrintWarng(GlobVar.ModuleId,
								 IW_MBKSigNameMissOpenBracket,
								 					  "StrSigDecod",
															   aStr);
				return(C_GfpUT_SigTypeUnknow);
			}

			if (PtClosBracket != (char*)NULL)
			{
				GfpUT_PrintWarng(GlobVar.ModuleId,
								 IW_MBKSigNameDuplClosBracket,
								 					  "StrSigDecod",
															   aStr);
				return(C_GfpUT_SigTypeUnknow);
			}

			PtClosBracket = PtChar;
			*PtChar = (char)0;
			continue;
		}

		if (   (PtOpenBracket == (char*)NULL)
			&& (      *PtChar != '_'        )
			&& (     !isalnum(*PtChar)      ))
		{
			GfpUT_PrintWarng(GlobVar.ModuleId,
							 IW_MBKSigNameInvalidChar,
							 				  "StrSigDecod",
													   aStr,
													*PtChar);
			return(C_GfpUT_SigTypeUnknow);
		}
	}


	if (TmpString[0] == (char)0)
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_MBKSigNameMissRoot,
						 			   "StrSigDecod",
												aStr);
		return(C_GfpUT_SigTypeUnknow);
	}
	else
		strcpy(aRoot, TmpString);


	if (PtOpenBracket != (char*)NULL)
	{
		if (PtClosBracket == (char*)NULL)
		{
			GfpUT_PrintWarng(GlobVar.ModuleId,
							 IW_MBKSigNameMissClosBracket,
							 					  "StrSigDecod",
														   aStr);
			return(C_GfpUT_SigTypeUnknow);
		}

		switch(GfpUT_StrIndexDecod(PtOpenBracket,aPtOrig,aPtExtr))
		{
			case C_GfpUT_IdxTypeBus:    return(C_GfpUT_SigTypeBus   );
			case C_GfpUT_IdxTypeScalar: return(C_GfpUT_SigTypeBusPin);
			case C_GfpUT_IdxTypeUnknow:
			default:                    return(C_GfpUT_SigTypeUnknow);
		}
	}
	else
	{
		*aPtOrig = *aPtExtr = 0L;
		return(C_GfpUT_SigTypeScalar);
	}
}


/*  Function  : 'GfpUT_DecodMBKSigName'                           */
/* -------------------------------------------------------------- */

extern char *GfpUT_DecodMBKSigName(aMBKSigName,
								   aPtSigIndex)
	char *aMBKSigName;
	long *aPtSigIndex;
{
	char *PtChar;
	char *PtIndexSepar;
	static char  StringBuffer[IC_StringBufferSize];


	strcpy( StringBuffer, aMBKSigName);

	/* Search the last space in the signal name. */
	PtIndexSepar = (char*)NULL;
	for(PtChar = StringBuffer; *PtChar != (char)0; PtChar++)
		if (*PtChar == ' ') PtIndexSepar = PtChar;

	/* Read, if detected, the bus index. */
	if (PtIndexSepar != (char*)NULL)
	{
	     *PtIndexSepar = (char)0;
		*aPtSigIndex   = atol(PtIndexSepar+1);
	}
	else
		*aPtSigIndex = C_GfpUT_ImplicitConNoIndex;

	return( StringBuffer);
}


/*  Fonction  :  'GfpUT_MakeImplicitConName'                      */
/* -------------------------------------------------------------- */
/*      Generation automatique et normalisee des noms de connec-
    teurs implicites. 
	    Les noms sont generes en concatenant dans l'ordre :
		    - le nom d'instance "instname".
			- le separateur MBK : MBK_SEPAR.
			- le nom specifique du connecteur : "conname".
			- l'extention reservee aux connecteurs : DP_IC.
	*/

extern char *GfpUT_MakeImplicitConName(aInstName, aConName, aIndex)
	char *aInstName;
	char *aConName;
	long  aIndex;
{
	char  StringBuffer[IC_StringBufferSize];

	if (aIndex == C_GfpUT_ImplicitConNoIndex)
		sprintf(StringBuffer, "%s%s"    , aConName, DP_IC);
	else
		sprintf(StringBuffer, "%s%s %ld", aConName, DP_IC, aIndex);

	return(concatname(aInstName,StringBuffer));			
}


/*  Function  :  'GfpUT_TypeConName'                              */
/* -------------------------------------------------------------- */
/*  Return the type of connector, according to is name :          */
/*              -  '*_ic' : C_GfpUT_TypeConImplicit.               */
/*              - '*vdd*' : C_GfpUT_TypeConVdd.                    */
/*              - '*vss*' : C_GfpUT_TypeConVss.                    */
/*              - others  : C_GfpUT_TypeConNormal.                 */

extern long GfpUT_TypeConName(aConName)
	char *aConName; /* Connector name to analys. */
{
	char *PtSuffix = (char)NULL;
	char *PtUnderScore;

	/* First recognize power supplies. */
	if (isvdd(aConName)) return(C_GfpUT_TypeConVdd);
	if (isvss(aConName)) return(C_GfpUT_TypeConVss);

	/* Search the last '_' in the name. */
	for( PtUnderScore  =  aConName; 
		*PtUnderScore != (char)NULL; PtUnderScore++)
		if (*PtUnderScore == '_') PtSuffix = PtUnderScore+1;

	/* If no underscore found then return the default type. */
	if (PtSuffix == (char*)NULL) return(C_GfpUT_TypeConNormal);

	/* Try to match the suffix of an implicit connector. */
	if (!strcmp(PtSuffix,IS_TypeConImplicit))
		return(           C_GfpUT_TypeConImplicit);

	/* Default. */
	return(C_GfpUT_TypeConNormal);
}


/*  Function  :  'GfpUT_AddLoConEnd'                              */
/* -------------------------------------------------------------- */
/*  This function works exactly like the classical MBK "addlocon"
 *  exept that the new connector is added at the end of the list.
 */

extern locon_list *GfpUT_AddLoConEnd(aLoFig,
									 aName ,
									 aLoSig,
									 aDir  )
	lofig_list *aLoFig;  /* Figure in which the connector should be added. */
		  char *aName;   /* Name of the connector.                         */
	losig_list *aLoSig;  /* Signal linked to the connector.                */
		  char  aDir;    /* Connector direction type.                      */
{
	locon_list **PPtTmpCon;
	locon_list     *OldConHead;

	/* Get the old head of the connector list. */
	OldConHead = aLoFig->LOCON;

	/* Add the new connector using the MBK function. */
	addlocon(aLoFig, aName, aLoSig, aDir);

	if (OldConHead != (locon_list*)NULL)
	{
		/* Search the address of the last NEXT pointer of the list. */
		for( PPtTmpCon  = &(aLoFig->LOCON);
			*PPtTmpCon != (locon_list*)NULL;
			 PPtTmpCon  = &((*PPtTmpCon)->NEXT));

		/* Put the first CON at the end of the list. */
		   *PPtTmpCon        = aLoFig->LOCON;
		  (*PPtTmpCon)->NEXT = (locon_list*)NULL;
		aLoFig->LOCON        = OldConHead;
	}

	/* Return the new head of the list. */
	/* (unchanged the most of the time) */
	return(aLoFig->LOCON);

}


/*  Function  :  'Dut_GetLoConBySig'                              */
/* -------------------------------------------------------------- */
/*  Search the logical connector associated to the signal
 *  "aLoSig" in the logical figure "aLoFig", if any.
 *    The function will return a NULL pointer if the search fail.
 */

extern locon_list *Dut_GetLoConBySig(aLoFig,aLoSig)
	lofig_list *aLoFig;
	losig_list *aLoSig;
{
	locon_list *TmpLoCon;

	for(TmpLoCon  =     aLoFig->LOCON;
		TmpLoCon != (locon_list*)NULL;
		TmpLoCon  =    TmpLoCon->NEXT)
		if (TmpLoCon->SIG == aLoSig)
			return(TmpLoCon);

	return((locon_list*)NULL);
}


