/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: UtMBKPh.c,v 1.10 1997/10/30 13:28:03 fitpath Exp fitpath $        |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                   | |%
 * %| |     F p G e n   P a c k a g e   1.10                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author    :            Jean-Paul CHAPUT                           | |%
 * %| |  E-mail    :        cao-vlsi@masi.ibp.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Header    :        "UtMBKPh.c"                                    | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */

# include  <stdio.h>
# include  <stdlib.h>
# include  <string.h>

# include  "GenericGN.h"
# include  "mbkgen.h"


/* ********************* Internal constants ********************* */


#define    IE_WorkPhFigMissing    3
#define    IE_BadNewName          4
#define    IE_DupNwellRef         5
#define    IE_MissNwellRef        6


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "UtProto.c" module identifier. */
	} GlobVar;


/* ********************* Internal Variables ********************* */


/* ******************** Interface Functions ********************* */


/* extern void  UtMPh_Init(void);
 * extern void  UtMPh_CopyUpCMA(char *aCMAName,
 *                              char *aInsName,
 *                              char *aNewName);
 * extern void  UtMPh_CopyUpHoDPThru(char *aInsName);
 * extern void  UtMPh_CopyUpNwell(char *aInsName);
 */


/* ********************* Internal Functions ********************* */


/* ******************** Functions Definitions ******************* */


/*  Function  :  'UtMPh_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtMPh_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtMBKPh", F_GfpUT_ModuleAdd);


	GfpUT_AddError(GlobVar.ModuleId, IE_WorkPhFigMissing,
	"[%s]:Missing \"DEF_PHFIG\".\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_BadNewName,
	"[%s]:Bad new reference name \"%s\".\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_DupNwellRef,
	"[%s]:Duplicate NWELL Reference in model \"%s\".\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_MissNwellRef,
	"[%s]:Missing NWELL Reference in model \"%s\".\n");
}


/*  Function  :  'UtMPh_CopyUpCMA()'.                             */
/* -------------------------------------------------------------- */

extern void  UtMPh_CopyUpCMA(aCMAName, aInsName, aNewName)
	char *aCMAName;
	char *aInsName;
	char *aNewName;
{
	phins_list *PtPhIns;
	phfig_list *PtPhFig;
	phref_list *PtPhRef;
		  char  RefName[256],
				NewName[256];
		  char *PtChar, *PtIndex;
		  long  XRef, YRef;


	if (WORK_PHFIG == (phfig_list*)NULL)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_WorkPhFigMissing,
								 "CopyUpCMA");

	if (is_bus(aNewName))
		GfpUT_PrintError(GlobVar.ModuleId,
						 	IE_BadNewName,
							  "CopyUpCMA",
								 aNewName);

	PtPhIns = getphins(WORK_PHFIG, aInsName);
	PtPhFig = getphfig(PtPhIns->FIGNAME, 'A');

	for(PtPhRef  = PtPhFig->PHREF;
		PtPhRef != (phref_list*)NULL;
		PtPhRef  = PtPhRef->NEXT)
	{
		strcpy(RefName, PtPhRef->NAME);
		for(PtChar=RefName; *PtChar!=(char)0; *PtChar++)
			if (*PtChar == '_') PtIndex = PtChar;
		*PtIndex = (char)0;

		if ((PtIndex == (char*)NULL) || strcmp(aCMAName,RefName)) continue;

		sprintf(NewName, "%s_%s", GfpUT_CheckName(aNewName), PtIndex+1);
		xyflat(         &XRef,         &YRef,
				PtPhRef->XREF, PtPhRef->YREF,
				PtPhIns->XINS, PtPhIns->YINS,
				PtPhFig->XAB1, PtPhFig->YAB1,
				PtPhFig->XAB2, PtPhFig->YAB2,
				PtPhIns->TRANSF);

		addphref(WORK_PHFIG,
				 PtPhRef->FIGNAME,
				 NewName,
				 XRef, YRef);
	}
}


/*  Function  :  'UtMPh_CopyUpHoDPThru()'.                        */
/* -------------------------------------------------------------- */

extern void  UtMPh_CopyUpHoDPThru(aInsName)
	char *aInsName;
{
	phins_list *PtPhIns;
	phfig_list *PtPhFig;
	phseg_list *PtPhSeg;
		  long  X1, Y1, X2, Y2, XRight;


	if (WORK_PHFIG == (phfig_list*)NULL)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_WorkPhFigMissing,
							  "CopyUpcwThhu");

	PtPhIns = getphins(WORK_PHFIG, aInsName);
	PtPhFig = getphfig(PtPhIns->FIGNAME, 'A');

	for(PtPhSeg  = PtPhFig->PHSEG;
		PtPhSeg != (phseg_list*)NULL;
		PtPhSeg  = PtPhSeg->NEXT)
	{
		if (   (PtPhSeg->LAYER != TALU1)
			&& (PtPhSeg->LAYER != TALU2)
			&& (PtPhSeg->LAYER != TALU3)) continue;

		if (   (PtPhSeg->TYPE !=  LEFT)
			&& (PtPhSeg->TYPE != RIGHT)) continue;

		xyflat(         &X1  ,         &Y1,
				PtPhSeg->X1  , PtPhSeg->Y1,
				PtPhIns->XINS, PtPhIns->YINS,
				PtPhFig->XAB1, PtPhFig->YAB1,
				PtPhFig->XAB2, PtPhFig->YAB2,
				PtPhIns->TRANSF);
		xyflat(         &X2  ,         &Y2,
				PtPhSeg->X2  , PtPhSeg->Y2,
				PtPhIns->XINS, PtPhIns->YINS,
				PtPhFig->XAB1, PtPhFig->YAB1,
				PtPhFig->XAB2, PtPhFig->YAB2,
				PtPhIns->TRANSF);

		/* We assume that, in a data-path operator, instances are never
		 * horizontally swapped. This ensure that the right extremity of
		 * the feed through is also to the right side of the column.
		 * So we can do our test in the instances coordinates for extend
		 * the segment to the right of the AB.
		 */
		XRight = (PtPhSeg->X2 == PtPhFig->XAB2) ? WORK_PHFIG->XAB2 : X2;

		addphseg(WORK_PHFIG,
				 PtPhSeg->LAYER,
				 PtPhSeg->WIDTH,
				 X1, Y1, XRight, Y2, (char*)NULL);
	}
}


/*  Function  :  'UtMPh_CopyUpNwell()'.                           */
/* -------------------------------------------------------------- */

extern void  UtMPh_CopyUpNwell(aInsName)
	char *aInsName;
{
	phins_list *PtPhIns;
	phfig_list *PtPhFig;
	phref_list *PtPhRef,
			   *PtNwellRef = (phref_list*)NULL;
		  long  XRef, YRef, WideOp;


	if (WORK_PHFIG == (phfig_list*)NULL)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_WorkPhFigMissing,
							"CopyUpNwellRef");

	PtPhIns = getphins(WORK_PHFIG, aInsName);
	PtPhFig = getphfig(PtPhIns->FIGNAME, 'A');

	WideOp =  PtPhFig->XAB2 - PtPhFig->XAB1;
	
	for(PtPhRef  = PtPhFig->PHREF;
		PtPhRef != (phref_list*)NULL;
		PtPhRef  = PtPhRef->NEXT)
	{
		if (strncmp(PtPhRef->NAME,"nwell_",6)) continue;

		PtNwellRef = PtPhRef;

		xyflat(         &XRef, &YRef,
			 PtNwellRef->XREF, PtNwellRef->YREF,
				PtPhIns->XINS, PtPhIns->YINS,
				PtPhFig->XAB1, PtPhFig->YAB1,
				PtPhFig->XAB2, PtPhFig->YAB2,
				PtPhIns->TRANSF);
	
		addphref(WORK_PHFIG,
				 PtNwellRef->FIGNAME,
				 PtNwellRef->NAME,
				 XRef, YRef);
	
		if (WORK_PHFIG->XAB2-WORK_PHFIG->XAB1 > WideOp)
			addphseg(WORK_PHFIG,
					 NWELL,
					 atol(PtNwellRef->NAME+6)*SCALE_X,
					 WORK_PHFIG->XAB1+WideOp, YRef,
					 WORK_PHFIG->XAB2       , YRef,
					 (char*)NULL);
	}
	if (PtNwellRef == (phref_list*)NULL)
		GfpUT_PrintError(GlobVar.ModuleId,
						  IE_MissNwellRef,
							"CopyUpNwell",
							PtPhFig->NAME);

}
	
