/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtPowerBar.c,v 1.10 1997/10/30 13:28:03 fitpath Exp fitpath $ */
/*             FpGen Package 1.10                                 */
/*          Alliance CAD System 3.2b                               */
/*             FpGen Package 1.10                                 */
/*                                                                */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                       */
/*                                                                */
/*  Author      :          Jean-Paul CHAPUT                       */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                       */
/* ************************************************************** */
/*  Module      :      "UtPowerBar.c"                             */


/* ************************** Includes ************************** */


#include   MUT_H
#include   MPH_H

#include  "GenericGN.h"


/* ********************* Internal constants ********************* */


#define   IE_PhStackOverflow            2
#define   IE_PhStackUnderflow           3

#define   IC_MaxPhStack                 256


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "UtPowerBar.c" module identifier. */
	} GlobVar;

/* Structure holding an element of the logical stack.
 */
typedef struct {
	phfig_list *PhFigFather; /* The current model (phfig). */
	phins_list *PhInsSon;    /* The next instance (phins). */
	} TypeState;


/* ********************* Internal Variables ********************* */


static       long     PhStackPoint;
static  TypeState     PhStackState[IC_MaxPhStack];


/* ******************** Interface Functions ********************* */


/* extern uphfig_powbar_list *GfpUT_AddOPowBar(phfig_list *aPhFig,
													 long  aYPow,
													 char  aType);
 * extern uphfig_powbar_list *GfpUT_GetPowBar(phfig_list *aPhFig, long  aY);
 * extern uphfig_powbar_list *GfpUT_GetPowBarBottom(phfig_list *aPhFig);
 * extern uphfig_powbar_list *GfpUT_GetPowBarTop(phfig_list *aPhFig);
 * extern               long  GfpUT_GetSliceByPowBar(phfig_list *aPhFig,
														   long aY);
 * extern               void  GfpUT_FreePowBar(phfig_list *aPhFig);
 * extern               void  GfpUT_HrExtractPowBar(phfig_list *aPhFig);
 */


/* ********************* Internal Functions ********************* */


#ifndef  ANSI_PROTO

static void  ResetPhStack();
static void  PushPhState();
static void  PopPhState();

#else

static void  ResetPhStack(void);
static void  PushPhState(phfig_list *PhFigFather, Phins_List *PhInsSon);
static void  PopPhState(phfig_list **PtPhFigFather, Phins_List **PtPhInsSon);

#endif


/* ******************** Functions Definitions ******************* */


/*  Function  :  'UtPwB_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtPwB_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtPowerBar", F_GfpUT_ModuleAdd);


	/* Errors associated to the hierarchical analysis functions.
	 */
	GfpUT_AddError(GlobVar.ModuleId, IE_PhStackOverflow,
	"[%s]:Stack overflow, more than %ld levels.\n");
	GfpUT_AddError(GlobVar.ModuleId, IE_PhStackUnderflow,
	"[%s]:Stack underflow, less than 0 levels.\n");
}


/*  Function  : 'ResetPhStack'                                    */
/* -------------------------------------------------------------- */
/*  Reset the logical state by setting his pointer to zero.
 */

static void  ResetPhStack()
{ PhStackPoint = 0L; }


/*  Function  : 'PushPhState'                                     */
/* -------------------------------------------------------------- */
/*  Push a logical state to the (logical) stack.
 */

static void  PushPhState(PhFigFather, PhInsSon)
	phfig_list *PhFigFather; /* The current logical model. */
	phins_list *PhInsSon;    /* The next logical instance. */
{
	/* Test stack overflow. */
	if (PhStackPoint == IC_MaxPhStack)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_PhStackOverflow,
										  "PushPhState",
										   PhStackPoint);

	PhStackState[PhStackPoint].PhFigFather = PhFigFather;
	PhStackState[PhStackPoint].PhInsSon    = PhInsSon;

	PhStackPoint++;
}


/*  Function  : 'PopPhState'                                      */
/* -------------------------------------------------------------- */
/*  Pop a logical state from the (logical) stack.
 */

static void  PopPhState(PtPhFigFather, PtPhInsSon)
	phfig_list  **PtPhFigFather; /* Pointer to the current logical model. */
	phins_list  **PtPhInsSon;    /* Pointer to the next logical instance. */
{
	/* Test stack underflow. */
	if (PhStackPoint == 0)
		GfpUT_PrintError(GlobVar.ModuleId,
						 IE_PhStackUnderflow,
											"PopPhState");

	PhStackPoint--;

	*PtPhFigFather = PhStackState[PhStackPoint].PhFigFather;
	*PtPhInsSon    = PhStackState[PhStackPoint].PhInsSon;
}


/*  Function  :  'GfpUT_AddOPowBar'                               */
/* -------------------------------------------------------------- */
/*  Add a new element to the "aPhFig" supply power bar ordered
 *  list. Elements are sorted in the increasing order of their
 *  Y coordinates. Two elements whith the same Y coordinates are
 *  merged.
 *      The returned value is a pointer to the newly created
 *  element in the list.
 */

extern uphfig_powbar_list *GfpUT_AddOPowBar(aPhFig, aYPow, aType)
	phfig_list *aPhFig;
		  long  aYPow;
		  char  aType;
{
			ptype_list   *PtUser;
	uphfig_powbar_list **PPtPow;
	uphfig_powbar_list   *PtNew;


	/* Attempt to get the 'powerbar' structure in the USER
	 * ptype_list, and create it if necessary.
	 */
	PtUser = GfpUT_GetUser(&(aPhFig->USER),
						   C_GfpUT_UserPhFigPowBar,
						   F_GfpUT_UserCreat      );

	for( PPtPow  = (uphfig_powbar_list**)&(PtUser->DATA);
		*PPtPow != (uphfig_powbar_list*)NULL;
		 PPtPow  =        &((*PPtPow)->NEXT))
	{
		if ((*PPtPow)->YPOW == aYPow) return(*PPtPow);
		/* In the future, check the TYPE cohenrency. */
		if ((*PPtPow)->YPOW >  aYPow) break;
	}

	PtNew = (uphfig_powbar_list*)mbkalloc(sizeof(uphfig_powbar_list));

	PtNew->YPOW =  aYPow;
	PtNew->TYPE =  aType;
	PtNew->NEXT = *PPtPow;
		*PPtPow =   PtNew;

	return(PtNew);
}


/*  Function  :  'GfpUT_GetPowBar'                                */
/* -------------------------------------------------------------- */
/*  Look for the supply power bar nearest of the coordinate given
 *  by "aY".
 *      Return a pointer the element founded and NULL otherwise.
 */

extern uphfig_powbar_list *GfpUT_GetPowBar(aPhFig, aY)
	phfig_list *aPhFig;
		  long  aY;
{
			ptype_list *PtUser;
	uphfig_powbar_list *PtPow;


	/* Attempt to get the 'powerbar' structure in the USER
	 * ptype_list.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigPowBar,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return((uphfig_powbar_list*)NULL);
	else
		PtPow = (uphfig_powbar_list*)(PtUser->DATA);

	/* Check if there is less than two elements in the list. */
	if (  (PtPow       == (uphfig_powbar_list*)NULL)
		||(PtPow->NEXT == (uphfig_powbar_list*)NULL))
		return((uphfig_powbar_list*)NULL);

	for(; PtPow->NEXT != (uphfig_powbar_list*)NULL;
		  PtPow        =               PtPow->NEXT)
		if (  (      PtPow->YPOW <= aY)
			&&(PtPow->NEXT->YPOW >  aY))
			return(PtPow);

	return((uphfig_powbar_list*)NULL);
}


/*  Function  :  'GfpUT_GetPowBarBottom'                          */
/* -------------------------------------------------------------- */

extern uphfig_powbar_list *GfpUT_GetPowBarBottom(aPhFig)
	phfig_list *aPhFig;
{
	ptype_list *PtUser;


	/* Attempt to get the 'powerbar' structure in the USER
	 * ptype_list.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigPowBar,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return((uphfig_powbar_list*)NULL);
	else
		return((uphfig_powbar_list*)(PtUser->DATA));
}


/*  Function  :  'GfpUT_GetPowBarTop'                             */
/* -------------------------------------------------------------- */

extern uphfig_powbar_list *GfpUT_GetPowBarTop(aPhFig)
	phfig_list *aPhFig;
{
			ptype_list *PtUser;
	uphfig_powbar_list *PtPow;


	/* Attempt to get the 'powerbar' structure in the USER
	 * ptype_list.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigPowBar,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return((uphfig_powbar_list*)NULL);
	else
		PtPow = (uphfig_powbar_list*)(PtUser->DATA);

	if (PtPow != (uphfig_powbar_list*)NULL)
		for(; PtPow->NEXT != (uphfig_powbar_list*)NULL;
			  PtPow        =               PtPow->NEXT);

	return(PtPow);
}


/*  Function  :  'GfpUT_GetSliceByPowBar'                         */
/* -------------------------------------------------------------- */

extern long  GfpUT_GetSliceByPowBar(aPhFig, aY)
	phfig_list *aPhFig;
		  long  aY;
{
	uphfig_powbar_list *PtPow;
			ptype_list *PtUser;
				  long  SliceNB;


	/* Attempt to get the 'powerbar' structure in the USER
	 * ptype_list.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigPowBar,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return(0);
	else
		PtPow = (uphfig_powbar_list*)(PtUser->DATA);

	SliceNB = -1;
	for(; PtPow != (uphfig_powbar_list*)NULL;
		  PtPow  =               PtPow->NEXT, SliceNB++)
		if (PtPow->YPOW > aY) break;

	return((SliceNB == -1) ? 0 : SliceNB);
}


/*  Function  :  'GfpUT_FreePowBar'                               */
/* -------------------------------------------------------------- */
/*  Free the complete 'powerbar' list and delete his associated
 *  pointer in the USER PhFig list.
 */

extern void  GfpUT_FreePowBar(aPhFig)
	phfig_list *aPhFig;
{
			ptype_list *PtUser;
	uphfig_powbar_list *PtPow;
	uphfig_powbar_list *PtDel;


	/* Attempt to get the 'powerbar' structure in the USER
	 * ptype_list.
	 */
	if ((PtUser=GfpUT_GetUser(&(aPhFig->USER),
							  C_GfpUT_UserPhFigPowBar,
							  F_GfpUT_UserGet        ))
								   == (ptype_list*)NULL)
		return;

	/* Deleting all the elements of the list. */
	for(PtPow  = (uphfig_powbar_list*)(PtUser->DATA);
		PtPow != (uphfig_powbar_list*)NULL;
		)
	{
		PtDel = PtPow;
		PtPow = PtPow->NEXT;

		mbkfree((void*)PtDel);
	}

	/* Deleting the head of list in the USER list. */
	GfpUT_DelUser(&(aPhFig->USER), C_GfpUT_UserPhFigPowBar,
								   F_GfpUT_UserNoFlags    );
}


/*  Function  :  'GfpUT_HrExtractPowBar'.                         */
/* -------------------------------------------------------------- */
/*  Hierarchical extraction of all the Y coordinates of the power
 *  supply bars. they are stored in a 'powbar' list.
 *      After this functions the coordinates of all model compo-
 *  may have change. In order to reset them, use the function
 *  'GfpUT_HrXYUnFlatPhFig'.
 */

extern void  GfpUT_HrExtractPowBar(aPhFig)
	phfig_list *aPhFig;
{
	phfig_list *PhFigFather;
	phfig_list *PhFigSon;
	phins_list *PhInsSon;
	phcon_list *PtCon;


	PhFigFather = aPhFig;
	PhInsSon    =  PhFigFather->PHINS;


	/* Look for the power supplies connectors
	 * of the  top level model.
	 */
	for(PtCon  = PhFigFather->PHCON;
		PtCon != (phcon_list*)NULL;
		PtCon  =       PtCon->NEXT)
	{
		if (  (PtCon->LAYER  !=  ALU2)
			||(PtCon->ORIENT == NORTH)
			||(PtCon->ORIENT == SOUTH))
			continue;

		if (isvdd(PtCon->NAME))
			GfpUT_AddOPowBar(aPhFig, PtCon->YCON, C_GfpUT_PowBarTypeVdd);

		if (isvss(PtCon->NAME))
			GfpUT_AddOPowBar(aPhFig, PtCon->YCON, C_GfpUT_PowBarTypeVss);
	}


	/* Reset physical stack and push the stop value. */
	ResetPhStack();
	PushPhState((phfig_list*)NULL, (phins_list*)NULL );

	/* Recursion main loop. */
	while(PhFigFather != (phfig_list*)NULL)
	{
	    /* If all instances have been processed, return to
		 * the upper levels.
		 */
		if (PhInsSon == (phins_list*)NULL)
		{
			PopPhState(&PhFigFather, &PhInsSon);
			continue;
		}

		/* Get the model of the instance. */
		PhFigSon = getphfig(PhInsSon->FIGNAME, 'A');

		/* Flaten the coordinate of the model. */
		UtFlt_XYFlatPhFig(PhFigSon, PhInsSon->XINS,
									PhInsSon->YINS,
									PhInsSon->TRANSF);

		/* Look for the power supplies connectors. */
		for(PtCon  = PhFigSon->PHCON;
			PtCon != (phcon_list*)NULL;
			PtCon  =       PtCon->NEXT)
		{
			if (  (PtCon->LAYER  !=  ALU2)
				||(PtCon->ORIENT == NORTH)
				||(PtCon->ORIENT == SOUTH))
				continue;

			if (isvdd(PtCon->NAME))
				GfpUT_AddOPowBar(aPhFig, PtCon->YCON, C_GfpUT_PowBarTypeVdd);

			if (isvss(PtCon->NAME))
				GfpUT_AddOPowBar(aPhFig, PtCon->YCON, C_GfpUT_PowBarTypeVss);
		}

		/* Go down the instances of the current son model
		 * or process the next instance of the father model.
		 */
		if (PhFigSon->PHINS != (phins_list*)NULL)
		{
			PushPhState(PhFigFather, PhInsSon->NEXT);

			PhFigFather = PhFigSon;
			PhInsSon    = PhFigSon->PHINS;
		}
		else
			PhInsSon = PhInsSon->NEXT;
	}
}
