/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtTechno.c,v 1.10 1997/10/30 13:28:04 fitpath Exp fitpath $ */
/*                                                                          */
/*          Alliance CAD System 3.2b                                         */
/*             FpGen Package 1.10                                           */
/*                                                                          */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                                 */
/*                                                                          */
/*  Author      :          Jean-Paul CHAPUT                                 */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                                 */
/* ************************************************************************ */
/*  Module      :      "UtTechno.c"                                         */


/* ******************************* Includes ******************************* */


#include  <string.h>

#include  "GenericGN.h"

/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |             External Structures and Variables                  | |%
 *   %| \----------------------------------------------------------------/ |%
 *   %|                                                                    |%
 *   %| All those variables must be declared `extern' in "GenericGN.h".    |%
 */

       UtTec_V_t  UtTec_V;


/* ************************** Internal constants ************************** */


#define    IW_UtTec_TypeUnknow    2


/* ********************* Internal Types And Structures ******************** */


/* Structure holding together all internal module global variables.
 */
static struct {
  long  ModuleId;          /* "UtTechno.c" module identifier. */
  } GV;


/* ************************** Internal Variables ************************** */


/* List of library suffixes. */
static char  TablLibSuffix[2] = {
                                 'g',  /* AsGa3 library. */
                                 'p'   /* CMOS2 library. */
                                };


/* ************************* Interface Functions ************************** */


/* extern void  UtTec_Init( char *aPath );
 * extern long  UtTec_Family( void );
 */


/* ************************** Internal Functions ************************** */


/* ************************* Functions Definitions ************************ */


/*  Function  :  'UtTec_Init()'.                                            */
/* ------------------------------------------------------------------------ */
/*  Perform all initialisations neededs to use the functions of the utilies
 *  module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtTec_Init( aPath )
  char *aPath;
{
  static long  FirstCall = TRUE;

         char *PtFile, *Pt;


  if (FirstCall == FALSE) return;
    FirstCall  = FALSE;

  /* Attribution of a module identifier. */
  GV.ModuleId = GfpUT_GetModuleId( "UtTechno", F_GfpUT_ModuleAdd );


  GfpUT_AddWarng( GV.ModuleId, IW_UtTec_TypeUnknow,
  "[%s]:Unknow technology file \"%s\".");


  /* Default techno. */
  UtTec_V.Type = C_UtTec_TypeCMOS2;

  if (aPath != (char*)NULL)
  {
    UtTec_V.Type = C_UtTec_TypeUnknow;

    /* Remove the leading path of the techno name. */
    for(PtFile = Pt = aPath ; *Pt != (char)0 ;
        (*Pt == '/') ? (PtFile = ++Pt) : Pt++ );
  
    /* Attempt to recognize a valid techno. */
    if (! strncmp( PtFile, "hcmos", 5 )) UtTec_V.Type = C_UtTec_TypeCMOS2;
    if (! strncmp( PtFile,  "cmos", 4 )) UtTec_V.Type = C_UtTec_TypeCMOS2;
    if (! strncmp( PtFile,  "prol", 4 )) UtTec_V.Type = C_UtTec_TypeCMOS2;
    if (! strncmp( PtFile,  "ecpd", 4 )) UtTec_V.Type = C_UtTec_TypeCMOS2;
    if (! strncmp( PtFile,  "asga", 4 )) UtTec_V.Type = C_UtTec_TypeAsGa3;
  
    if (UtTec_V.Type == C_UtTec_TypeUnknow)
      GfpUT_PrintWarng( GV.ModuleId, IW_UtTec_TypeUnknow, "Init", aPath );
  }


  switch( UtTec_V.Type )
  {
    case C_UtTec_TypeAsGa3:

      UtTec_V.PowerLayer    = ALU3;
      UtTec_V.PowerWideVDD  = 25;    /* lambdas units. */
      UtTec_V.PowerWideVSS  = 50;    /* lambdas units. */
      UtTec_V.SliceHeight   = 62;    /* lambdas units. */
      UtTec_V.ThruLayer     = TALU2;
      UtTec_V.ThruPitch     = 5;     /* lambdas units. */
      UtTec_V.ThruWide      = 2;     /* lambdas units. */
      UtTec_V.ThruNB        = 11;
      UtTec_V.ThruTop       = 6;     /* lambdas units. */
      UtTec_V.ThruBot       = 6;     /* lambdas units. */
      UtTec_V.LibrarySuffix = TablLibSuffix[0];
      break;

    case C_UtTec_TypeCMOS2:
    case C_UtTec_TypeUnknow:
    default:

      UtTec_V.PowerLayer    = ALU2;
      UtTec_V.PowerWideVDD  = 6;     /* lambdas units. */
      UtTec_V.PowerWideVSS  = 8;     /* lambdas units. */
      UtTec_V.SliceHeight   = 60;    /* lambdas units. */
      UtTec_V.ThruLayer     = TALU2;
      UtTec_V.ThruPitch     = 5;     /* lambdas units. */
      UtTec_V.ThruWide      = 2;     /* lambdas units. */
      UtTec_V.ThruNB        = 10;
      UtTec_V.ThruTop       = 7;     /* lambdas units. */
      UtTec_V.ThruBot       = 8;     /* lambdas units. */
      UtTec_V.LibrarySuffix = TablLibSuffix[1];
      break;
  }
}


/*  Function  :  'UtTec_Family()'.                                          */
/* ------------------------------------------------------------------------ */

extern long  UtTec_Family()
{
  switch( UtTec_V.Type )
  {
    case C_UtTec_TypeAsGa3:

      return( C_UtTec_FamilyAsGa );

    case C_UtTec_TypeCMOS2:
    case C_UtTec_TypeUnknow:
    default:

      return( C_UtTec_FamilyCMOS );
  }
}


