/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtUser.c,v 1.10 1997/10/30 13:28:04 fitpath Exp fitpath $ */
/*                                                                */
/*          Alliance CAD System 3.2b                               */
/*             FpGen Package 1.10                                 */
/*                                                                */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                       */
/*                                                                */
/*  Author      :          Jean-Paul CHAPUT                       */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                       */
/* ************************************************************** */
/*  Module      :      "UtUser.c"                                */


/* ************************** Includes ************************** */


#include   MUT_H
#include   MLO_H
#include   MPH_H

#include  "GenericGN.h"


/* ********************* Internal constants ********************* */


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "UtUser.c" module identifier. */
	} GlobVar;


/* ********************* Internal Variables ********************* */


static void *(*TabFcDupUser[C_GfpUT_UserMaxTdupl])();
static long    TabTpDupUser[C_GfpUT_UserMaxTdupl];
static long         DupUserNB = 0;


/* ******************** Interface Functions ********************* */


/* extern ptype_list *GfpUT_GetUser(ptype_list **aPPtHead,
 *                                        long   aType   ,
 *                                        long   aFlags  );
 * extern ptype_list *GfpUT_DelUser(ptype_list **aPPtHead,
 *                                        long   aType   ,
 *                                        long   aFlags  );
 * extern       long  GfpUT_SetDupUser(long    aType   ,
 *                                     void *(*aFunc)());
 * extern       void  GfpUT_DupUser(ptype_list **aPPtUserSrc ,
 *                                  ptype_list **aPPtUserDest);
 */


/* ********************* Internal Functions ********************* */


/* ******************** Functions Definitions ******************* */


/*  Function  :  'UtUsr_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtUsr_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtUser", F_GfpUT_ModuleAdd);
}


/*  Function  :  'GfpUT_GetUser'                                  */
/* -------------------------------------------------------------- */
/*  Search the first object of type "aType" in the 'ptype_list'
 *  pointed by "aPPtHead". The function return a pointer to the 
 *  object found and a NULL value otherwise.
 *      If the flag 'F_GfpUT_UserCreat' in the argument "aFlags"
 *  is set, and the object is not found then it is created.
 */

extern ptype_list *GfpUT_GetUser(aPPtHead, aType, aFlags)
	ptype_list **aPPtHead;  /* Pointer to the list head pointer. */
		  long   aType;     /* Type of data to search.           */
		  long   aFlags;    /* Behavior of the search.           */
{
	ptype_list *PtUser;


	/* Search the element of type "aType" in USER ptype list. */
	if ((PtUser=getptype(*aPPtHead,aType)) != (ptype_list*)NULL)
		return(PtUser);

	/* If not found, create it if asked. */ 
	if (aFlags & F_GfpUT_UserCreat) 
	{
		*aPPtHead = addptype(*aPPtHead, aType, (void*)NULL);

		return(*aPPtHead);
	}
	else
		return((ptype_list*)NULL);
}


/*  Function  :  'GfpUT_DelUser'                                  */
/* -------------------------------------------------------------- */
/*  Delete the first object of type "aType" in the list pointed
 *  by "aPPtHead".
 */

extern ptype_list *GfpUT_DelUser(aPPtHead, aType, aFlags)
	ptype_list **aPPtHead;  /* Pointer to the list head pointer. */
		  long   aType;     /* Type of data to search.           */
		  long   aFlags;    /* Behavior of the search (unused).  */
{
    /* Deletion of the first object that match "aType". */
    return(*aPPtHead = delptype(*aPPtHead,aType));
}


/*  Function  :  'GfpUT_SetDupUser'                               */
/* -------------------------------------------------------------- */
/*  Set the specific duplicate function to use for the duplica-
 *  of the type "aType" object. The pointer to the dplicate
 *  function is stored in an internal table : "TablFcDupUser"
 *  and "TablTpDupUser".
 *      The duplicate function is called by the 'GfpUT_DupUser'
 *  function.
 *      The number of duplicate function is limited by the cons-
 *  tant 'C_GfpUT_UserMaxTdupl'. The function return the number
 *  of duplicate function already stored in the internal table.
 */

extern long  GfpUT_SetDupUser(aType, aFunc)
	long    aType;
	void *(*aFunc)();
{
	if (DupUserNB < C_GfpUT_UserMaxTdupl) 
	{
		TabFcDupUser[DupUserNB  ] = aFunc;
		TabTpDupUser[DupUserNB++] = aType;
	}

	return(DupUserNB);
}


/*  Function  :  'GfpUT_DupUser'                                  */
/* -------------------------------------------------------------- */
/*  Duplicate a user 'ptype_list' and all object pointed by the
 *  list using the object duplicate functions. Such duplication
 *  functions must be set by a previous call to "GfpUT_SetDupUser".
 */

extern void  GfpUT_DupUser(aPPtUserSrc,aPPtUserDup)
	ptype_list **aPPtUserSrc;  /* Pointer to the source ptype list.    */
	ptype_list **aPPtUserDup;  /* Pointer to the duplicate ptype list. */
{
	ptype_list *PtUser;
		  long  i;

	for(PtUser  = *aPPtUserSrc     ;
		PtUser != (ptype_list*)NULL;
		PtUser  =      PtUser->NEXT)
		for(i=0; i<DupUserNB; i++)
			if (PtUser->TYPE == TabTpDupUser[i])
				*aPPtUserDup = addptype(*aPPtUserDup,
										PtUser->TYPE,
					  (TabFcDupUser[i])(aPPtUserSrc,
										aPPtUserDup));
}
