/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: UtXlong.c,v 1.10 1997/10/30 13:28:04 fitpath Exp fitpath $ */
/*                                                                */
/*          Alliance CAD System 3.2b                               */
/*             FpGen Package 1.10                                 */
/*                                                                */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                       */
/*                                                                */
/*  Author      :          Jean-Paul CHAPUT                       */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                       */
/* ************************************************************** */
/*  Module      :      "UtXlong.c"                                */


/* ************************** Includes ************************** */


#include  <string.h>

#include  "GenericGN.h"


/* ********************* Internal constants ********************* */


#define       IW_XlStrNullPtr                      2
#define       IW_XlStrOverflow                     3
#define       IW_XlInvalidDigit                    4
#define       IW_XlIllegalDigit                    5

#define       IC_MaxLenString                        16384 /* 16 Ko */


/* **************** Internal Types And Structures *************** */


/* Structure holding together all internal module global variables.
 */
static struct {
	long  ModuleId;          /* "FpGenPAD" module identifier. */
	char  StringBuffer[IC_MaxLenString];
	} GlobVar;


/* ********************* Internal Variables ********************* */


/* ********************* External Variables ********************* */


       long  DP_LGSIZE;  /* Number of long used for a eXtended Long. */


/* ******************** Interface Functions ********************* */


/* extern void  UtXlg_Init(void);
 * extern long *GfpUT_XlAlloc(void);
 * extern void  GfpUT_XlSetLong(long *aPtXlong, long  aLong, long  aValue);
 * extern void  GfpUT_XlFree(long *aPtXlong);
 * extern long  GfpUT_XlGetBit(long *aPtXlong, long  aBit);
 * extern long  GfpUT_XlStrCheck(char *aStringNumber);
 * extern void  GfpUT_XlStrToXl(char *aStringNumber, long *aPtXlong);
 * extern char *GfpUT_XlToStr(long *aPtXlong   ,
 *                            long  aConstWidth,
 *                            long  aBaseValue );
 * extern char *GfpUT_XlToStrVHDL(long *aPtXlong, long  aConstWidth);
 */

/* ********************* Internal Functions ********************* */


/* Convert a long integer to an alphanumerical digit. */
#define  DutMC_LToDigit(vdigit)                  \
	( ((vdigit) > 9) ? ((char)((vdigit)-10)+'a') \
	                 : ((char)((vdigit)   )+'0') )

static long  DigitToL();


/* ******************** Functions Definitions ******************* */


/*  Function  :  'UtXlg_Init()'.                                  */
/* -------------------------------------------------------------- */
/*  Perform all initialisations neededs to use the functions of
 *  the utilies module :
 *       I - Load internal errors table.
 *      II - Load internal warnings table.
 */

extern void  UtXlg_Init()
{
	static long  FirstCall    = TRUE;

	if (FirstCall == FALSE) return;
		FirstCall  = FALSE;

	/* Attribution of a module identifier. */
	GlobVar.ModuleId = GfpUT_GetModuleId("UtXlong", F_GfpUT_ModuleAdd);


    /* Detect the size of a long integer.
	 */
    DP_LGSIZE = sizeof( long);

	/* Warnings associated to the extended long ('Xl' ) functions.
	 */
	GfpUT_AddWarng(GlobVar.ModuleId, IW_XlStrNullPtr,
	"[%s]:(null) pointer passed as string argument.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_XlStrOverflow,
	"[%s]:Constant string overflow:\"%s\".");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_XlInvalidDigit,
	"[%s]:Invalid digit \'%c\'.");
	GfpUT_AddWarng(GlobVar.ModuleId, IW_XlIllegalDigit,
	":[%s]:Illegal digit \'%c\' (string %s:base %ld).");
}


/*  Function  : 'DigitToL'                                        */
/* -------------------------------------------------------------- */

static long  DigitToL(cdigit)
	char  cdigit;
{
	if ((cdigit >= '0') && (cdigit <= '9')) return( (long)(cdigit-'0'   ));
	if ((cdigit >= 'a') && (cdigit <= 'f')) return( (long)(cdigit-'a'+10));
	if ((cdigit >= 'A') && (cdigit <= 'F')) return( (long)(cdigit-'A'+10));

	/* Unrecognized digit case. */
	GfpUT_PrintWarng(GlobVar.ModuleId,
					 IW_XlInvalidDigit,
								  "DigitToL",
									  cdigit);

	return( 0);
}


/*  Function  : 'GfpUT_XlAlloc'                                   */
/* -------------------------------------------------------------- */
/*  Allocate an extended long integer.
 */

extern long *GfpUT_XlAlloc()
{
	long *Xlong;
	long  Offset;

	Xlong = (long*)mbkalloc(sizeof(long)*C_GfpUT_XLSIZE);
	for(Offset=0; Offset < C_GfpUT_XLSIZE; Offset++)
		GfpUT_XlSetLong(Xlong, Offset, 0L);

	return(Xlong);
}


/*  Function  : 'GfpUT_XlFree'                                    */
/* -------------------------------------------------------------- */
/*  Free memory allocated to an extended long integer.
 */

extern void  GfpUT_XlFree(aPtXlong)
	long *aPtXlong;
{ mbkfree( (void*)aPtXlong ); }


/*  Function  : 'GfpUT_XlSetLong'                                 */
/* -------------------------------------------------------------- */
/*  Set one long inside an extended long.
 */

extern void  GfpUT_XlSetLong(aPtXlong,aLong,aValue)
	long *aPtXlong;
	long  aLong;
	long  aValue;
{
	/* Check if the long is whithin the correct range. */
	if (  (aLong >= C_GfpUT_XLSIZE)
		||(aLong <  0)           )
		return;

	aPtXlong[aLong] = aValue;
}


/*  Function  : 'GfpUT_XlGetBit'                                  */
/* -------------------------------------------------------------- */
/*  Return the value of the specified bit of an extended long.
 */

extern long  GfpUT_XlGetBit(aPtXlong,aBit)
	long *aPtXlong;
	long  aBit;
{
	/* Check if the bit is whitin the valid range. */
	if (aBit > C_GfpUT_XLSIZE*DP_LGSIZE*8) return( 0L);

	return( ( aPtXlong[aBit/(DP_LGSIZE*8)]
			 &(1L  << (aBit%(DP_LGSIZE*8))) ) ? 1L : 0L);
}


/*  Function  : 'GfpUT_XlStrCheck'                                */
/* -------------------------------------------------------------- */

extern long  GfpUT_XlStrCheck(aStringNumber)
	char *aStringNumber;
{
	long  BaseValue;
	long  BaseDigitMax;
	long  StringIndex;
	long  StringStart;
	long  CheckFlag;

	/* Detect NULL pointer passed as argument. */
	if (aStringNumber == (char*)NULL)
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_XlStrNullPtr,
								  "XlStrCheck");

		return( 0);
	}

	/* Recognize the base number, depending on the first two */
	/* digits :                                              */
	/*         -  "0nnnn" : Octal.                           */
	/*         - "0Bnnnn" : Binary.                          */
	/*         - "0Xnnnn" : Hexadecimal.                     */
	if (   (aStringNumber[0] != (char)0 )
		 &&(aStringNumber[1] != (char)0 )
		 &&(aStringNumber[0] ==      '0'))
		switch( aStringNumber[1] )
		{
			case 'b':
		    case 'B': 
			      BaseValue = C_GfpUT_BaseBinary;
				StringStart = 2;
				break;
			case 'x':
		    case 'X':
				  BaseValue = C_GfpUT_BaseHexadecimal;
				StringStart = 2;
				break;
			default:
				  BaseValue = C_GfpUT_BaseOctal;
				StringStart = 1;
		}
	else
	{
		  BaseValue = C_GfpUT_BaseHexadecimal;
		StringStart = 0;
	}

	/* Set the of the maximum digit value. */
	BaseDigitMax = 1L << BaseValue;

	/* Get the end of the string. */
	StringIndex = strlen(aStringNumber)-1;

	/* Check the digits of the string number (from rigth to left). */
	for(CheckFlag=TRUE; StringIndex >= StringStart; StringIndex--)
		if (DigitToL(aStringNumber[StringIndex]) >= BaseDigitMax)
		{
			GfpUT_PrintWarng(GlobVar.ModuleId,
							 IW_XlIllegalDigit,
										"XlStrCheck",
						  aStringNumber[StringIndex],
						  aStringNumber             ,
						  BaseDigitMax              );

			CheckFlag = FALSE;
		}

	/* Compute the number of long neededs to encode the      */
	/* string and check if it is less than an extended long. */
	if (   (strlen(aStringNumber)-StringStart)*BaseValue
		 > (C_GfpUT_XLSIZE*DP_LGSIZE*8)                  )
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_XlStrOverflow,
								   "XlStrCheck",
								  aStringNumber);

	return( CheckFlag);
}


/*  Function  : 'GfpUT_XlStrtToXl'                                */
/* -------------------------------------------------------------- */

extern void  GfpUT_XlStrToXl(aStringNumber,aPtXlong)
	char *aStringNumber;
	long *aPtXlong;
{
	long  BaseValue;
	long  StringIndex;
	long  NumberWidth;
	long  DigitValue;

	/* Detect NULL pointer passed as argument. */
	if (aStringNumber == (char*)NULL)
	{
		GfpUT_PrintWarng(GlobVar.ModuleId,
						 IW_XlStrNullPtr,
								   "XlStrToXl");
		return;
	}

	/* Recognize the base number, depending on the first two */
	/* digits :                                              */
	/*         -  "0nnnn" : Octal.                           */
	/*         - "0Bnnnn" : Binary.                          */
	/*         - "0Xnnnn" : Hexadecimal.                     */
	if (   (aStringNumber[0] != (char)0 )
		 &&(aStringNumber[1] != (char)0 )
		 &&(aStringNumber[0] ==      '0'))
		switch( aStringNumber[1] )
		{
			case 'b':
		    case 'B': 
			      BaseValue = C_GfpUT_BaseBinary;
				StringIndex = 2;
				break;
			case 'x':
		    case 'X':
				  BaseValue = C_GfpUT_BaseHexadecimal;
				StringIndex = 2;
				break;
			default:
				  BaseValue = C_GfpUT_BaseOctal;
				StringIndex = 1;
		}
	else
	{
		  BaseValue = C_GfpUT_BaseHexadecimal;
		StringIndex = 0;
	}


	/* Read the string digit by digit from left to right. */
	for( NumberWidth = 0;   (aStringNumber[StringIndex] != (char)0)
		                  &&(NumberWidth < DP_LGSIZE*8*C_GfpUT_XLSIZE);
		 StringIndex += 1,
		 NumberWidth += BaseValue)
	{
		/* Convert the digit. */
		DigitValue = DigitToL(aStringNumber[StringIndex]);

		/* Shift the sum then add the digit value. */
		aPtXlong[NumberWidth/(DP_LGSIZE*8)] <<=  BaseValue;
		aPtXlong[NumberWidth/(DP_LGSIZE*8)]  |= DigitValue;
	}
}


/*  Function  : 'GfpUT_XlToStr'                                   */
/* -------------------------------------------------------------- */
/*  Convert an extended long into an hexadecimal constant string.
 */

extern char *GfpUT_XlToStr(aPtXlong,aConstWidth,aBaseValue)
	long *aPtXlong;
	long  aConstWidth;
	long  aBaseValue;
{
	long  StringIndex;
	long  NumberIndex;

	for( NumberIndex  = aConstWidth + 4
		             - (  (aConstWidth%C_GfpUT_BaseHexadecimal)
					    ? (aConstWidth%C_GfpUT_BaseHexadecimal) : 4),
		 StringIndex  = 0;
		 NumberIndex >  0;
		 NumberIndex -= C_GfpUT_BaseHexadecimal,
		 StringIndex += 1)
		GlobVar.StringBuffer[StringIndex] = DutMC_LToDigit(
			(    aPtXlong[(NumberIndex-4)/(DP_LGSIZE*8)]
			  >>         ((NumberIndex-4)%(DP_LGSIZE*8)) ) & 15L);

	GlobVar.StringBuffer[StringIndex] = (char)0;

	return( GlobVar.StringBuffer);
}


/*  Function  : 'GfpUT_XlToStrVHDL'                               */
/* -------------------------------------------------------------- */
/*  Convert an extended long into a VHDL constant string.
 */

extern char *GfpUT_XlToStrVHDL(aPtXlong,aConstWidth)
	long *aPtXlong;
	long  aConstWidth;
{
	long  StringIndex;
	long  NumberIndex;
	long    FlagBinary;

	  FlagBinary = FALSE;
	StringIndex  = 0;
	NumberIndex  = (aConstWidth > 128) ? 128 : aConstWidth;

	/* Convert the head of the constant, no more than 3 bits. */
	if (NumberIndex%C_GfpUT_BaseHexadecimal)
	{
	    strcpy( GlobVar.StringBuffer+StringIndex, "B\"");

		for( StringIndex                         += 2;
			 NumberIndex%C_GfpUT_BaseHexadecimal >= C_GfpUT_BaseBinary;
			 NumberIndex                         -= C_GfpUT_BaseBinary,
			 StringIndex                         += 1)
		    GlobVar.StringBuffer[StringIndex] = DutMC_LToDigit(
				(    aPtXlong[NumberIndex/(DP_LGSIZE*8)]
				 >>          (NumberIndex%(DP_LGSIZE*8)) ) & 1L);

		FlagBinary = TRUE;
	}
		
	/* Convert the first digits into an hexadecimal constant. */
	if (NumberIndex >= C_GfpUT_BaseHexadecimal)
	{
	    if (FlagBinary)
		{
	        strcpy( GlobVar.StringBuffer+StringIndex, "\" & ");
			StringIndex += 4;
		}

	    strcpy( GlobVar.StringBuffer+StringIndex, "X\"");

		for( StringIndex += 2;
			 NumberIndex >= C_GfpUT_BaseHexadecimal;
			 NumberIndex -= C_GfpUT_BaseHexadecimal,
			 StringIndex += 1)
			GlobVar.StringBuffer[StringIndex] = DutMC_LToDigit(
			   (    aPtXlong[(NumberIndex-4)/(DP_LGSIZE*8)]
				 >>         ((NumberIndex-4)%(DP_LGSIZE*8)) ) & 15L);
	}

	/* Write the end of the constant. */
	strcpy( GlobVar.StringBuffer+StringIndex, "\"");

	/* Return the generated string. */
	return( GlobVar.StringBuffer);
}
