/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*------------------------------------------------------------\
|                                                             |
| Tool    :                   GRAAL                           |
|                                                             |
| File    :                  graal.c                          |
|                                                             |
| Author  :                Jacomme Ludovic                    |
|                                                             |
| Date    :                  28.03.95                         |
|                                                             |
\------------------------------------------------------------*/

# ident "@(#)GRAAL ver 1.10 January 1995"

/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>
# include <Xm/Xm.h>
# include MUT_H
# include MPH_H
# include RDS_H
# include RPR_H
# include RUT_H
# include GSB_H
# include GRM_H
# include GMX_H

# include "graal.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                           present                           |
|                                                             |
\------------------------------------------------------------*/

void GraalPresent()
{
  alliancebanner( "Graal", GRAAL_VERSION,
                  "Graphic layout editor", "1995",
                  ALLIANCE_VERSION );
}

/*------------------------------------------------------------\
|                                                             |
|                            main                             |
|                                                             |
\------------------------------------------------------------*/

int main( argc, argv )

    int   argc;
    char *argv[];
{
  char *FileName   = (char *)NULL;
  char  WindowInfo = 0;
  char  Debug      = 0;
  int   Index;
 
  GraalPresent();

  mbkenv();
  rdsenv();

  for ( Index = 1;
        Index < argc;
        Index++ )
  {
    if ( ! strcmp( argv[ Index ], "-l" ) )
    {
      if ( ( Index + 1 ) < argc )
      {
        Index    = Index + 1;
        FileName = argv[ Index ];
      }
    }
    else
    if ( ! strcmp( argv[ Index ], "-w" ) )
    {
      WindowInfo = 1;
    }
    else 
    if ( ! strcmp( argv[ Index ], "-debug" ) )
    {
      Debug = 1;
    }
  }

  if ( Debug ) rdsdebug();

  rdsbegin();

  Graalenv();
  loadrdsparam();
  GraalLoadParameters();

  GraalInitializeRessources( &argc, argv );
  GraalInitializeWindow();

  if ( WindowInfo )
  {
    fprintf( stdout, "\nGraal window\n" );
    fprintf( stdout, "\n  XMIN : %d"  , GraalWindowXmin / RDS_LAMBDA );
    fprintf( stdout, "\n  XMAX : %d"  , GraalWindowXmax / RDS_LAMBDA );
    fprintf( stdout, "\n  YMIN : %d"  , GraalWindowYmin / RDS_LAMBDA );
    fprintf( stdout, "\n  YMAX : %d"  , GraalWindowYmax / RDS_LAMBDA );
    fprintf( stdout, "\n  SIDE : %d"  , GraalWindowSide / RDS_LAMBDA );
    fprintf( stdout, "\n  DX   : %d"  , GraalWindowDx                );
    fprintf( stdout, "\n  DY   : %d"  , GraalWindowDy                );
    fprintf( stdout, "\n  SIZE : %d\n", GraalWindowSize              );
  }

  GraalInitializeCutCxTable();
  GraalInitializeErrorMessage( Debug );

  if ( FileName != (char *)NULL )
  {
    GraalFigureMbk = Graalgetphfig( FileName, 'A' );
  }

  XtAppMainLoop( GraalApplication ); 

  rdsend();

  return( 0 );
}
