/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*************************************************************
*                  SOURCES                                   *
*                                                            *
* Changed by:    Greg                                        *
* Date:          05/07/95                                    *
* New file:      -                                           *
* Comments:                                                  *
*   Fonctions permettant d'ajouter un nouveau reseau dans    *
*   la figure logique existante.                             *
*                                                            *
*                                                            *
*************************************************************/

# include MUT_H
# include RDS_H
# include MLO_H
# include RCN_H
# include MPH_H
# include RPR_H
# include "extract.h"
# include "lynxrcn.h"
# include <stdio.h>
# include <stdlib.h>
# include <sys/timeb.h>
# include "addlynxrcn.h"
# include "Windows95.h"
# include "inter.h"
# include "mbkrds.h"
# include "chklynxrcn.h"
# include "buildrcn.h"
# include "via.h"

# include <math.h>

/*************************************************************
* La variable rcnlynxdebug doit etre fixee a 0 pour NE PAS   *
* avoir d'affichage sur la sortie STDERR.                    *
*                                                            *
*                                                            *
*************************************************************/

short rcnlynxdebug=0;
short rcnlynxchk=0;

long GrgNodeINDEX=1L;

/*************************************************************
*                                                            *
* Fonctions utilisees en interne par le programme.           *
*                                                            *
*                                                            *
*************************************************************/

void GrgConPh2ConLo();
void GrgDualConLoPh();
void GrgFatalError();
void GrgChrono();
void GrgGroupVia();

/*************************************************************
* SOURCES                                                    *
*                                                            *
*                                                            *
*                                                            *
*************************************************************/

void generate_RCN(fig)
lofig_list *fig;

/*************************************************************

En entree, les rds doivent avoir dans leur champ user une ptype
normalement constituee. Les connecteurs logiques DOIVENT etre
relies aux connecteurs physiques grace a une liste chainee
accessible par un ptype du champ USER des locons. Les valeurs
que doit prendre l'identificateur de la ptype est definie
dans le fichier 'lynxrcn.h'.
A un connecteur physique, il n'y aura qu'un seul connecteur
logique attache, tandis qu'a un connecteur logique, il peut y
avoir plusieurs connecteurs physiques (ex des instances).
Pour les transistors, il n'y a qu'un connecteur physique par
connecteur logique.

*************************************************************/

{
  losig_list    *ScanSignal;    

  fflush(stdout);
  fflush(stderr);
  
  if(rcnlynxdebug)
  {
    fflush(stderr);
    fprintf(stderr,"\n\n--------- DEBUT DE LA FONCTION generate_RCN -----------\n");
    fflush(stderr);
  }

  printf("\t---> Net extract is running...\n"); fflush(stdout);

  if(rcnlynxdebug)
  {
    fprintf(stderr,"Figure d'entree\n");
    Grgdisplay_datain(fig);
    fprintf(stderr,"____________________________________________\n");
    fflush(stderr);
  }

  if(rcnlynxchk)
  {
    if(rcnlynxdebug)
      fprintf(stderr,
	      "Verification de la conformite des structures d'entree...\n");
    GrgChkInConFEqui(fig);
  }

  /*
  Ajoute pour chaque rectangle representant un connecteur un pointeur vers
  le connecteur logique correspondant. Ce pointeur est accessible par
  la valeur RDSLOCON
  */

  GrgDualConLoPh(fig);

  /*
  Permet de retrouver a partir d'un rectangle constituant un via les deux
  autres rectangles constituant ce via. Le pointeur est accessible par la
  valeur GRGVIA
  */
  GrgGroupVia(fig);

  if(rcnlynxchk)
  {
    if(rcnlynxdebug)
      fprintf(stderr,"Verification de la conformite des structures de travail...\n");
    GrgChkOutConVia(fig,rcnlynxdebug);
  }
  
  if(rcnlynxdebug)
  {
    fprintf(stderr,"Figure apres liaisons connecteur physique-logique\n");
    Grgdisplay_datain(fig);
    fprintf(stderr,"____________________________________________\n");
    fflush(stderr);
  }

  for(ScanSignal=fig->LOSIG; ScanSignal; ScanSignal=ScanSignal->NEXT)
  {
    /* Remise a un du compteur de noeud pour chaque Signal */

    GrgNodeINDEX=1L;

    buildrcn(fig,ScanSignal);

/*
    if(rcnlynxchk)
    {
      if(!rcn_check(getlorcnet(ScanSignal),0))
        printf("RCN error on signal %ld.\n",ScanSignal->INDEX);
    }
*/
  }


  if(rcnlynxdebug)
  {
    fprintf(stderr,"-------- SORTIE DE LA FONCTION generate_RCN -----------\n");
  }

  fflush(stderr);
  fflush(stdout);

}

void GrgDualConLoPh(fig)
lofig_list      *fig;
/*************************************************************
* Creer la relation connecteur physique vers connecteur      *
* logique a partir de la relation connecteur logique vers    *
* connecteur logique existante.                              *
*                                                            *
*************************************************************/
{
  loins_list    *ScanLoIns;     /* Pour parcourir les instances logiques                */
  lotrs_list    *ScanTrs;       /* Pour parcourir les transistors logiques              */

  if(rcnlynxdebug)
  {
    fprintf(stderr,"\nPlace dans le champ USER des rectangles connecteurs\n");
    fprintf(stderr,"le connecteur logique lui correspondant...\n");
  }

  if(rcnlynxdebug)
    fprintf(stderr,"  Traitement des connecteurs externes de la figure...\n");
  GrgConPh2ConLo(fig->LOCON,RDSLOCON1);
  
  if(rcnlynxdebug)
    fprintf(stderr,"  Traitement des connecteurs des instances...\n");
  for(ScanLoIns=fig->LOINS;ScanLoIns;ScanLoIns=ScanLoIns->NEXT)
  {
    GrgConPh2ConLo(ScanLoIns->LOCON,RDSLOCON2);
  }

  if(rcnlynxdebug)
    fprintf(stderr,"  Traitement des connecteurs des transistors...\n");
  for(ScanTrs=fig->LOTRS;ScanTrs;ScanTrs=ScanTrs->NEXT)
  {
    GrgConPh2ConLo(ScanTrs->DRAIN,RDSLOCON3);
    GrgConPh2ConLo(ScanTrs->GRID,RDSLOCON3);
    GrgConPh2ConLo(ScanTrs->SOURCE,RDSLOCON3);
  }
}





void GrgConPh2ConLo(connector,identificateur)
locon_list      *connector;     /*Structure classique des connecteurs chaines */
long             identificateur;

/*************************************************************
Cette fonction ajoute un ptype dans le champ USER des rectangles
connecteurs qui pointe sur le connecteur logique lui
correspondant. Les parametres de cette fonction sont:
-Une locon_list des connecteurs possedant un ptype repere par
 'identificateur' pointant sur une liste chainee contenant les
 adresses des rdsrec_list connecteur.
-Un identificateur defini dans 'lynxrcn.h'
*************************************************************/

{
  locon_list *ScanLoCon;
  chain_list *ScanChain;        /* liste chainee des rectangles connecteur */
  rdsrec_list *Rec_a;

  for(ScanLoCon=connector;ScanLoCon;ScanLoCon=ScanLoCon->NEXT)
  {
    for(ScanChain=((chain_list*)(getptype(ScanLoCon->USER,identificateur)->DATA));
	ScanChain;
	ScanChain=ScanChain->NEXT)
    {
      Rec_a=((rdsrec_list*)(ScanChain->DATA));
      Rec_a->USER=addptype(Rec_a->USER,RDSLOCON,ScanLoCon);
    }
  }
}



void GrgFatalError(code)
int code;
{
  fprintf(stderr,"\n\n********************************************\n");
  fprintf(stderr,"*                                          *\n");
  fprintf(stderr,"*    ERREUR DANS LES STRUCTURES DE DONNEE  *\n");
  fprintf(stderr,"*                                          *\n");
  fprintf(stderr,"********************************************\n\n");

  fprintf(stderr,"Code d'erreur: %d\n",code);
  exit(code);
}



IsRdsRealVia(r)
rdsrec_list *r;
{
  
  if(IsRdsVia(r))
  {
    if(GetLynxAttribute(r)!= C_X_N && GetLynxAttribute(r)!=C_X_P)
      return(1);
  } 
  return(0);
}

void GrgGroupVia(fig)
lofig_list      *fig;
/*************************************************************
* Ajoute un ptype dans le champ user des vias permettant de  *
* retrouver instantanement les deux autres couches du via.   *
*                                                            *
*                                                            *
*************************************************************/
{
  chain_list    *Vias=NULL;     /* Contient la chaine des vias de la figure (+rapide)   */
  chain_list    *FirstVias;     /* Pour effacer la liste chainee des Vias               */
  losig_list    *ScanLoSig;     /* Pour parcourir la liste des equipotentielles         */
  rdsrec_list   *ScanPhRec;     /* Pour la meme chose                                   */
  long           CX1,CY1;       /* Evite de recalculer le centre de chaque via          */
  long           CX2,CY2;       /* Evite de recalculer le centre de chaque via          */
  char           EndEqui;       /* Indique la fin du traitement de l'equi               */

  chain_list    *HeadGrpVia;	/* Tete de liste des nouveau GrgPhVia_list */
  chain_list    *ScanGrpVia;	/* Pour pacourir cette liste */
  rdsrec_list   *Rec;		/* Contient le rectangle du via */
  GrgPhVia_list *NewGrpVia;	/* Nouvel element */
  ptype_list	*ptl;

  /* Donnees sur le fenetrage */
  long           mx,MX,my,MY;   /* Contient la taille mini/maxi du circuit */
  long           nbvia;         /* Contient le nombre de via du signal */
  long           WinSizeX,WinSizeY;	/* Taille des fenetres */
  long           WinNbX,WinNbY;		/* Nombre de fenetre */
  unsigned       N;		/* Nombre de fenetre sur 1 cote */
  chain_list   **WinVia;	/* Tete de liste chainee des Via element de la fenetre */
  long           WinX,WinY;	/* Emplacement de la fenetre */
  long           i;		
  
  
/*************************************************************
* Principe: creer une liste chainee des nouvelles structures *
* crees, puis supprime le chainage a la fin.                 *
*                                                            *
*                                                            *
*************************************************************/

  if(rcnlynxdebug)
    fprintf(stderr,"  Traitement des vias...\n");
  for(ScanLoSig=fig->LOSIG; ScanLoSig; ScanLoSig=ScanLoSig->NEXT)
  {

    /* Cree une liste chainee des vias de l'equipotentielle traitee */

    EndEqui=0;
    Vias=NULL;
    nbvia=0L;
    ScanPhRec=(rdsrec_list*)(getptype(ScanLoSig->USER,RCN_FEQUI)->DATA); 
    while(!EndEqui)
    {
      if(IsRdsRealVia(ScanPhRec))
      {
	Vias=addchain(Vias,ScanPhRec);
        
        nbvia++;
        CX1= ScanPhRec->X + ((ScanPhRec->DX)>>1);
        CY1= ScanPhRec->Y + ((ScanPhRec->DY)>>1);
      
        if(nbvia>1)
        { 
          if(ScanPhRec->X<mx) mx=ScanPhRec->X;
          if(ScanPhRec->Y<my) my=ScanPhRec->Y;
          if(ScanPhRec->X+ScanPhRec->DX>MX) MX=ScanPhRec->X+ScanPhRec->DX;
          if(ScanPhRec->Y+ScanPhRec->DY>MY) MY=ScanPhRec->Y+ScanPhRec->DY;
        }
        else
        {
          mx=MX=CX1;
          my=MY=CY1;
        }
      }
      if(IsLynxEndEqui(ScanPhRec))
	EndEqui=1;
      else
	ScanPhRec=LYNX_EQUI(ScanPhRec);
    }

    /* Traite les vias de l'equi */
    if(nbvia)
    {
      FirstVias=Vias;

      /* Creation du fenetrage du chip */
      N=(unsigned)(sqrt((double)nbvia));
      WinSizeX=(MX-mx)/N;
      WinSizeY=(MY-my)/N;
      WinNbX=(MX-mx)/WinSizeX+1;
      WinNbY=(MY-my)/WinSizeY+1;
  
      WinVia=(chain_list**)mbkalloc(sizeof(chain_list)*WinNbX*WinNbY);
      for(i=0;i<WinNbX*WinNbY;i++)
        WinVia[i]=NULL;

      /* Place les vias dans la bonne fenetre */
      for(Vias=FirstVias; Vias; Vias=Vias->NEXT)
      {
        if(rcnlynxdebug)
          fprintf(stderr,"  %6ld %6ld %6ld x %6ld",((rdsrec_list*)(Vias->DATA))->X,((rdsrec_list*)(Vias->DATA))->Y,((rdsrec_list*)(Vias->DATA))->DX,((rdsrec_list*)(Vias->DATA))->DY);

        CX1=((rdsrec_list*)(Vias->DATA))->X +
            (((rdsrec_list*)(Vias->DATA))->DX>>1);
        CY1=((rdsrec_list*)(Vias->DATA))->Y +
            (((rdsrec_list*)(Vias->DATA))->DY>>1);
    
        WinX=(CX1-mx)/WinSizeX;
        WinY=(CY1-my)/WinSizeY;

        WinVia[WinX+WinY*WinNbX]=addchain(WinVia[WinX+WinY*WinNbX],Vias->DATA);

        if(rcnlynxdebug)
          fprintf(stderr,"   Centre %ld %ld, fenetre %u\n",CX1,CY1,WinX+WinY*WinNbX);
      }

      freechain(FirstVias);

      /* Parcours la liste de chaque fenetre et regroupe les vias au
       sein de chaqune d'elle */

      for(i=0;i<WinNbX*WinNbY;i++)
      {
        HeadGrpVia=NULL;
        for(Vias=WinVia[i];Vias;Vias=Vias->NEXT)
        {
          CX1=((rdsrec_list*)(Vias->DATA))->X +
              (((rdsrec_list*)(Vias->DATA))->DX>>1);
          CY1=((rdsrec_list*)(Vias->DATA))->Y +
              (((rdsrec_list*)(Vias->DATA))->DY>>1);
          Rec=(rdsrec_list*)(Vias->DATA);
          for(ScanGrpVia=HeadGrpVia;ScanGrpVia;ScanGrpVia=ScanGrpVia->NEXT)
          {
            CX2=((GrgPhVia_list*)(ScanGrpVia->DATA))->X;
            CY2=((GrgPhVia_list*)(ScanGrpVia->DATA))->Y;
        
            if(CX1==CX2 && CY1==CY2)
              break;
          }

          if(ScanGrpVia)
          {
            NewGrpVia=(GrgPhVia_list*)(ScanGrpVia->DATA);

            /* Verification que sur les trois couches du vias les
               couches sont effectivement differents. Il peut y avoir
               deux vias superposes */

            if( NewGrpVia->LAYER2 == NULL)
            {
              if( GetRdsLayer(NewGrpVia->LAYER1) == GetRdsLayer(Rec) )
                Rec=NULL;

              NewGrpVia->LAYER2=Rec;
            }
            else
            {
              /* dans le cas de vias superposes, il se peut que l'on
                 rentre deux fois la meme structure dans la ptype, ce
                 qui provoque une boucle infinie. On verifie donc si
                 le layer 3 est bien a la valeur NULL, signifiant que
                 cette structure n'a jamais ete rentree dans la liste */

              if( ! NewGrpVia->LAYER3)
              {
                if( GetRdsLayer(NewGrpVia->LAYER1) == GetRdsLayer(Rec) ||
                    GetRdsLayer(NewGrpVia->LAYER2) == GetRdsLayer(Rec) )
                  Rec=NULL;

                NewGrpVia->LAYER3=Rec;
  
                if(Rec)
                {
                  if(ptl=getptype(ScanLoSig->USER,GRGVIA))
                  {
                    NewGrpVia->NEXT=ptl->DATA;
                    ptl->DATA=NewGrpVia;
                  }
                  else
                  {
                    NewGrpVia->NEXT=NULL;
                    ScanLoSig->USER=addptype(ScanLoSig->USER,GRGVIA,
                                             NewGrpVia);
                  }
                }
              }
            }
          }
          else
          {
            NewGrpVia=(GrgPhVia_list*)mbkalloc(sizeof(GrgPhVia_list));
            NewGrpVia->LAYER1=Rec;
            NewGrpVia->LAYER2=NULL;
            NewGrpVia->LAYER3=NULL; 
            NewGrpVia->X=CX1;
            NewGrpVia->Y=CY1;
            HeadGrpVia=addchain(HeadGrpVia,NewGrpVia);
          }
        }
        freechain(WinVia[i]);
        freechain(HeadGrpVia);
      }
      mbkfree(WinVia);
    }
  }
  if(rcnlynxdebug)
    fprintf(stderr,"Fin du traitement des vias\n");
}
