/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Construction du reseau RC correspondant aux rectangles de 
   la figure */


# include MUT_H
# include RDS_H
# include MLO_H
# include RCN_H
# include MPH_H
# include RPR_H
# include "znc.h"
# include "Windows95.h"
# include "buildrcn.h"
# include "lynxrcn.h"
# include "extract.h"
# include "mbkrds.h"
# include "inter.h"
# include "via.h"

/* A mettre en commentaire pour ne pas avoir de message de debuggage */
/*
# define BUILDRCNDEBUG
*/

/*
#define BUILDRCNCTL
*/

int RectangleValide();
int RectangleDiff();
void traiteinterparra();
void traiteinterperpen();
void addzncvia();
void znc2rcn();
chain_list* sortchain();
void sortloconnode();
void sortbullelocon();

#define min(a,b) ((a>b)?b:a)
#define max(a,b) ((a>b)?a:b)

/******************************************************************************/



void buildrcn(fig,scansignal)
lofig_list	*fig;
losig_list	*scansignal;
{
  GrgWindow	*WindowsInfo;
  rdsrec_list	*scanrec,*irec;
  int		 EndEqui;
  lorcnet_list	*ptrcnet;
  WinZnc	*zncwininfo;
  chain_list	*inter_list,*parra,*perpen,*scanchain;
  ptype_list	*ptlvia;
  chain_list	*connect;
  long		 layer,node;

#ifdef BUILDRCNCTL
fprintf(stderr,"Traitement du signal %ld [%s]\n",
               scansignal->INDEX,
               getsigname(scansignal)?getsigname(scansignal):"no name");
#endif

#ifdef BUILDRCNDEBUG
fprintf(stderr,"Traitement du signal %ld [%s]\n",
               scansignal->INDEX,
               getsigname(scansignal)?getsigname(scansignal):"no name");
#endif

#ifdef BUILDRCNDEBUG
fprintf(stderr,"Construction des fenetrages\n");
#endif
#ifdef BUILDRCNCTL
fprintf(stderr,"Construction du fenetrage...                                 ");
#endif

  WindowsInfo=GrgCreatWindows(scansignal);

  zncwininfo=creatwinznc( WindowsInfo->mx,
                          WindowsInfo->my,
                          WindowsInfo->Mx,
                          WindowsInfo->My,
                          WindowsInfo->Px,
                          WindowsInfo->Py,
                          WindowsInfo->nx,
                          WindowsInfo->ny  );

  scanrec=(rdsrec_list*)(getptype(scansignal->USER,RCN_FEQUI)->DATA);

  EndEqui=0;
  connect=NULL;

  ptrcnet=addlorcnet(scansignal);
 

#ifdef BUILDRCNCTL
fprintf(stderr,"\rTraitement de l'equipotentielle...                         ");
#endif

  while(!EndEqui)
  {
    if(RectangleValide(scanrec))
    {
#ifdef BUILDRCNDEBUG
fprintf(stderr,"\n\n******** Rectangle trait :\n");
dbg_affrec(scanrec);
#endif
      inter_list=GrgGetInterWindows(scanrec,WindowsInfo);
      parra=NULL;
      perpen=NULL;

      for(scanchain=inter_list;scanchain;scanchain=scanchain->NEXT)
      {
        irec=(rdsrec_list*)(scanchain->DATA);

        if(RectangleValide(irec))
        {
          if(Intersect(scanrec,irec)==PERPEN)
            perpen=addchain(perpen,scanchain->DATA);
          else
            parra=addchain(parra,scanchain->DATA);
        }
      }

      /* Traite les extremits du rectangle */

      if(GETDIRECTION(scanrec)==HORIZ_DIR) /* horizontal */
      {
        addznc(scanrec->X,scanrec->Y,scanrec->X,scanrec->Y+scanrec->DY,
               GetRdsLayer(scanrec),zncwininfo);
        addznc(scanrec->X+scanrec->DX,scanrec->Y,
               scanrec->X+scanrec->DX,scanrec->Y+scanrec->DY,
               GetRdsLayer(scanrec),zncwininfo);
      }
      else
      {
        addznc(scanrec->X,scanrec->Y,scanrec->X+scanrec->DX,scanrec->Y,
               GetRdsLayer(scanrec),zncwininfo);
        addznc(scanrec->X,scanrec->Y+scanrec->DY,
               scanrec->X+scanrec->DX,scanrec->Y+scanrec->DY,
               GetRdsLayer(scanrec),zncwininfo);
      }

      /* Traite les rectangles en intersection perpendiculaire */

      traiteinterperpen(scanrec,perpen,zncwininfo);

      /* Traite les rectangles en intersection parralelles */

      traiteinterparra(scanrec,parra,zncwininfo);


      if(perpen) freechain(perpen);
      if(parra) freechain(parra);
      if(inter_list) freechain(inter_list);
    }
    /* On ajoute les CxP et les CxN qui servent dans les transistors
       coud\'es. Ils sont fait en GATE, que l'on traite pour l'extraction
       comme du POLY. Ces morceaux seront toujours isoles a ce stade,
       on a juste qu'a ajouter une ZNC le representant */

    if(  IsRdsVia(scanrec) &&
            (GetLynxAttribute(scanrec)==C_X_N ||
             GetLynxAttribute(scanrec)==C_X_P    ))

      addznc(scanrec->X,scanrec->Y,
             scanrec->X+scanrec->DX,scanrec->Y+scanrec->DY,
             RDS_POLY,zncwininfo);

    if(getptype(scanrec->USER,RDSLOCON))
      connect=addchain(connect,scanrec);
    else /* plus clean : on traite les connecteurs plus loin */
      if(RectangleDiff(scanrec))
      {
        addznc(scanrec->X,scanrec->Y,
               scanrec->X+scanrec->DX,scanrec->Y+scanrec->DY,
               GetRdsLayer(scanrec),zncwininfo);
      }

    if(IsLynxEndEqui(scanrec))
      EndEqui=1;
    else
      scanrec=LYNX_EQUI(scanrec);
  }

#ifdef BUILDRCNDEBUG
fprintf(stderr,"ZNC apres le traitement des rectangles normaux\n");
dbg_affznc(zncwininfo);
#endif

#ifdef BUILDRCNCTL
fprintf(stderr,"\rAjout des vias...                                          ");
#endif
  /* Traitement des connecteurs */
  /* Ajoute les vias dans les znc */
  if(ptlvia=getptype(scansignal->USER,GRGVIA))
    addzncvia((GrgPhVia_list*)(ptlvia->DATA),zncwininfo,WindowsInfo);


#ifdef BUILDRCNDEBUG
fprintf(stderr,"ZNC apres le traitement des vias\n");
dbg_affznc(zncwininfo);
#endif

#ifdef BUILDRCNCTL
fprintf(stderr,"\rAjout des connecteurs...                                   ");
#endif

  for(scanchain=connect;scanchain;scanchain=scanchain->NEXT)
  {

    scanrec=(rdsrec_list*)(scanchain->DATA);
    layer=GetRdsLayer(scanrec);


    if(layer==RDS_GATE)
    {
      addznc(scanrec->X,scanrec->Y,scanrec->X+scanrec->DX,
             scanrec->Y+scanrec->DY,RDS_POLY,zncwininfo);
    }
    else
    { 
      addznc(scanrec->X,scanrec->Y,scanrec->X+scanrec->DX,
             scanrec->Y+scanrec->DY,layer,zncwininfo);
    }
  }

#ifdef BUILDRCNDEBUG
fprintf(stderr,"ZNC apres le traitement des connecteurs\n");
dbg_affznc(zncwininfo);
#endif


#ifdef BUILDRCNCTL
fprintf(stderr,"\rCreation des lonodes...                                    ");
#endif

  makezncnode(zncwininfo,ptrcnet);

#ifdef BUILDRCNCTL
fprintf(stderr,"\rConstruction de la table...                                ");
#endif

  buildtable(ptrcnet);

#ifdef BUILDRCNCTL
fprintf(stderr,"\rConstruction des Wires...                                  ");
#endif

  znc2rcn(zncwininfo,scansignal);

#ifdef BUILDRCNCTL
fprintf(stderr,"\rAjout des lonodes dans les locons...                       ");
#endif

  linklonodelocon(connect,ptrcnet,zncwininfo);

#ifdef BUILDRCNCTL
fprintf(stderr,"\rLiberation de la table...                                  ");
#endif

  freetable(ptrcnet);

#ifdef BUILDRCNCTL
fprintf(stderr,"\rTrie des connecteurs d'instance...                         ");
#endif

  sortloconnode(connect,zncwininfo);

  if(connect)
    freechain(connect);



#ifdef BUILDRCNDEBUG
fprintf(stderr,"Liberation du fenetrage\n");
#endif

#ifdef BUILDRCNCTL
fprintf(stderr,"\rLiberation des fenetrages...                               ");
#endif


  GrgClearWindows(WindowsInfo,scansignal);
  freewinznc(zncwininfo);

#ifdef BUILDRCNCTL
fprintf(stderr,"\rOk.                                                      \n");
#endif



#ifdef BUILDRCNDEBUG
fprintf(stderr,"fin du traitement des rectangles du signal %ld\n\n\n",
               scansignal->INDEX);
#endif
}

int RectangleValide(R)
rdsrec_list	*R;
{
  if(    !IsRdsVia(R)
      && !IsRdsConnector(R)
      && (    GetRdsLayer(R)==RDS_ALU1
           || GetRdsLayer(R)==RDS_ALU2
           || GetRdsLayer(R)==RDS_ALU3
           || GetRdsLayer(R)==RDS_ALU4
           || GetRdsLayer(R)==RDS_POLY )    )
  {
    return(1);
  }
  return(0);
}

int RectangleDiff(R)
rdsrec_list	*R;
{
  if( GetRdsLayer(R)==RDS_PDIF || 
      GetRdsLayer(R)==RDS_NDIF   )
    return(1);
  return(0);
}

void traiteinterperpen(scanrec,perpen,zncwininfo)
rdsrec_list	*scanrec;
chain_list	*perpen;
WinZnc		*zncwininfo;
{
  chain_list	*scanchain;
  rdsrec_list	*irec;
  
  for(scanchain=perpen;scanchain;scanchain=scanchain->NEXT)
  {
    irec=(rdsrec_list*)(scanchain->DATA);

    if(RectangleValide(irec))
    {
#ifdef BUILDRCNDEBUG
fprintf(stderr,"Intersection perpendiculaire : \n");
dbg_affrec(irec);
#endif

      if(GETDIRECTION(scanrec)==HORIZ_DIR) /* horizontal */
        addznc(irec->X,scanrec->Y,irec->X+irec->DX,scanrec->Y+scanrec->DY,
             GetRdsLayer(scanrec),zncwininfo,NULL,NULL);
      else /* vertical */
        addznc(scanrec->X,irec->Y,scanrec->X+scanrec->DX,irec->Y+irec->DY,
             GetRdsLayer(scanrec),zncwininfo,NULL,NULL);
    }          
  }
}

void traiteinterparra(scanrec,parra,zncwininfo)
rdsrec_list	*scanrec;
chain_list	*parra;
WinZnc		*zncwininfo;
{
  chain_list	*scanchain,*scanchain2;
  rdsrec_list	*irec,*irec2;
  long		 mx,my,Mx,My;

  for(scanchain=parra;scanchain;scanchain=scanchain->NEXT)
  {
    irec=(rdsrec_list*)(scanchain->DATA);

    if(GETDIRECTION(scanrec)==HORIZ_DIR)
    {
      if(irec->X>=scanrec->X)
      {
        my=min(irec->Y,scanrec->Y);
        My=max(irec->Y+irec->DY,scanrec->Y+scanrec->DY);
        for(scanchain2=parra;scanchain2;scanchain2=scanchain2->NEXT)
        {
          irec2=(rdsrec_list*)(scanchain2->DATA);
          if(irec2->X<=irec->X && irec2->X+irec2->DX>=irec->X)
          {
            my=min(my,irec2->Y);
            My=max(My,irec2->Y+irec2->DY);
          }
        }
        addznc(irec->X,my,irec->X,My,GetRdsLayer(scanrec),zncwininfo,NULL,NULL);
      }

      if(irec->X+irec->DX<=scanrec->X+scanrec->DX)
      {
        my=min(irec->Y,scanrec->Y);
        My=max(irec->Y+irec->DY,scanrec->Y+scanrec->DY);
        for(scanchain2=parra;scanchain2;scanchain2=scanchain2->NEXT)
        {
          irec2=(rdsrec_list*)(scanchain2->DATA);
          if( irec2->X<=irec->X+irec->DX &&
              irec2->X+irec2->DX>=irec->X+irec->DX )
          {
            my=min(my,irec2->Y);
            My=max(My,irec2->Y+irec2->DY);
          }
        }
        addznc(irec->X+irec->DX,my,irec->X+irec->DX,My,
               GetRdsLayer(scanrec),zncwininfo,NULL,NULL);
      }
    }
    else
    {
      if(irec->Y>=scanrec->Y)
      {
        mx=min(irec->X,scanrec->X);
        Mx=max(irec->X+irec->DX,scanrec->X+scanrec->DX);
        for(scanchain2=parra;scanchain2;scanchain2=scanchain2->NEXT)
        {
          irec2=(rdsrec_list*)(scanchain2->DATA);
          if(irec2->Y<=irec->Y && irec2->Y+irec2->DY>=irec->Y)
          {
            mx=min(mx,irec2->X);
            Mx=max(Mx,irec2->X+irec2->DX);
          }
        }
        addznc(mx,irec->Y,Mx,irec->Y,GetRdsLayer(scanrec),zncwininfo,NULL,NULL);
      }

      if(irec->Y+irec->DY<=scanrec->Y+scanrec->DY)
      {
        mx=min(irec->X,scanrec->X);
        Mx=max(irec->X+irec->DX,scanrec->X+scanrec->DX);
        for(scanchain2=parra;scanchain2;scanchain2=scanchain2->NEXT)
        {
          irec2=(rdsrec_list*)(scanchain2->DATA);
          if( irec2->Y<=irec->Y+irec->DY &&
              irec2->Y+irec2->DY>=irec->Y+irec->DY )
          {
            mx=min(mx,irec2->X);
            Mx=max(Mx,irec2->X+irec2->DX);
          }
        }
        addznc(mx,irec->Y+irec->DY,Mx,irec->Y+irec->DY,
               GetRdsLayer(scanrec),zncwininfo,NULL,NULL);
      }
    }
  }
}

void addzncvia(headvia,winznc,winrec)
GrgPhVia_list	*headvia;
WinZnc		*winznc;
GrgWindow	*winrec;
{
  GrgPhVia_list	*scanvia;
  chain_list	*intervia1,*intervia2;
  chain_list	*scaninter;
  long		 mx1,my1,Mx1,My1,layer1;
  long		 mx2,my2,Mx2,My2,layer2;
  rdsrec_list	*rec;

#ifdef BUILDRCNDEBUG
fprintf(stderr,"Traitement des vias\n");
#endif

  for(scanvia=headvia;scanvia;scanvia=scanvia->NEXT)
  {
    intervia1=NULL;
    intervia2=NULL;

    if( GetRdsLayer(scanvia->LAYER1) == RDS_CONT ||
        GetRdsLayer(scanvia->LAYER1) == RDS_VIA1 ||
        GetRdsLayer(scanvia->LAYER1) == RDS_VIA2 ||
        GetRdsLayer(scanvia->LAYER1) == RDS_VIA3   )
    {
      intervia1=GrgGetInterWindows(scanvia->LAYER2,winrec);
      intervia2=GrgGetInterWindows(scanvia->LAYER3,winrec);
      mx1=scanvia->LAYER2->X;
      my1=scanvia->LAYER2->Y;
      Mx1=scanvia->LAYER2->X+scanvia->LAYER2->DX;
      My1=scanvia->LAYER2->Y+scanvia->LAYER2->DY;
      layer1=GetRdsLayer(scanvia->LAYER2);
      mx2=scanvia->LAYER3->X;
      my2=scanvia->LAYER3->Y;
      Mx2=scanvia->LAYER3->X+scanvia->LAYER3->DX;
      My2=scanvia->LAYER3->Y+scanvia->LAYER3->DY;
      layer2=GetRdsLayer(scanvia->LAYER3);
    }

    if( GetRdsLayer(scanvia->LAYER2) == RDS_CONT ||
        GetRdsLayer(scanvia->LAYER2) == RDS_VIA1 ||
        GetRdsLayer(scanvia->LAYER2) == RDS_VIA2 ||
        GetRdsLayer(scanvia->LAYER2) == RDS_VIA3   )
    {
      intervia1=GrgGetInterWindows(scanvia->LAYER1,winrec);
      intervia2=GrgGetInterWindows(scanvia->LAYER3,winrec);
      mx1=scanvia->LAYER1->X;
      my1=scanvia->LAYER1->Y;
      Mx1=scanvia->LAYER1->X+scanvia->LAYER1->DX;
      My1=scanvia->LAYER1->Y+scanvia->LAYER1->DY;
      layer1=GetRdsLayer(scanvia->LAYER1);
      mx2=scanvia->LAYER3->X;
      my2=scanvia->LAYER3->Y;
      Mx2=scanvia->LAYER3->X+scanvia->LAYER3->DX;
      My2=scanvia->LAYER3->Y+scanvia->LAYER3->DY;
      layer2=GetRdsLayer(scanvia->LAYER3);
    }

    if( GetRdsLayer(scanvia->LAYER3) == RDS_CONT ||
        GetRdsLayer(scanvia->LAYER3) == RDS_VIA1 ||
        GetRdsLayer(scanvia->LAYER3) == RDS_VIA2 ||
        GetRdsLayer(scanvia->LAYER3) == RDS_VIA3   )
    {
      intervia1=GrgGetInterWindows(scanvia->LAYER1,winrec);
      intervia2=GrgGetInterWindows(scanvia->LAYER2,winrec);
      mx1=scanvia->LAYER1->X;
      my1=scanvia->LAYER1->Y;
      Mx1=scanvia->LAYER1->X+scanvia->LAYER1->DX;
      My1=scanvia->LAYER1->Y+scanvia->LAYER1->DY;
      layer1=GetRdsLayer(scanvia->LAYER1);
      mx2=scanvia->LAYER2->X;
      my2=scanvia->LAYER2->Y;
      Mx2=scanvia->LAYER2->X+scanvia->LAYER2->DX;
      My2=scanvia->LAYER2->Y+scanvia->LAYER2->DY;
      layer2=GetRdsLayer(scanvia->LAYER2);
    }

/*
      fprintf(stderr,"Layer 1 : %s (%ld,%ld)-(%ld,%ld)\n",
                      RDS_LAYER_NAME[GetRdsLayer(scanvia->LAYER1)],
                      scanvia->LAYER1->X,
                      scanvia->LAYER1->Y,
                      scanvia->LAYER1->X+scanvia->LAYER1->DX,
                      scanvia->LAYER1->Y+scanvia->LAYER1->DY);
      fprintf(stderr,"Layer 2 : %s (%ld,%ld)-(%ld,%ld)\n",
                      RDS_LAYER_NAME[GetRdsLayer(scanvia->LAYER2)],
                      scanvia->LAYER2->X,
                      scanvia->LAYER2->Y,
                      scanvia->LAYER2->X+scanvia->LAYER2->DX,
                      scanvia->LAYER2->Y+scanvia->LAYER2->DY);
      fprintf(stderr,"Layer 3 : %s (%ld,%ld)-(%ld,%ld)\n",
                      RDS_LAYER_NAME[GetRdsLayer(scanvia->LAYER3)],
                      scanvia->LAYER3->X,
                      scanvia->LAYER3->Y,
                      scanvia->LAYER3->X+scanvia->LAYER3->DX,
                      scanvia->LAYER3->Y+scanvia->LAYER3->DY);
*/
    if(intervia1)
    {
      for(scaninter=intervia1;scaninter;scaninter=scaninter->NEXT)
      {
        rec=(rdsrec_list*)(scaninter->DATA);
        if(RectangleValide(rec))
        {
          if(GETDIRECTION(rec)==HORIZ_DIR)
          {
            my1=min(my1,rec->Y);
            My1=max(My1,rec->Y+rec->DY);
          }
          else
          {
            mx1=min(mx1,rec->X);
            Mx1=max(Mx1,rec->X+rec->DX);
          }
        }
      }
      addznc(mx1,my1,Mx1,My1,layer1,winznc);
      freechain(intervia1);
    }
    else
      addznc(mx1,my1,Mx1,My1,layer1,winznc);


    if(intervia2)
    {
      for(scaninter=intervia2;scaninter;scaninter=scaninter->NEXT)
      {
        rec=(rdsrec_list*)(scaninter->DATA);
        if(RectangleValide(rec))
        {
          if(GETDIRECTION(rec)==HORIZ_DIR)
          {
            my2=min(my2,rec->Y);
            My2=max(My2,rec->Y+rec->DY);
          }
          else
          {
            mx2=min(mx2,rec->X);
            Mx2=max(Mx2,rec->X+rec->DX);
          }
        }
      }
      addznc(mx2,my2,Mx2,My2,layer2,winznc);
      freechain(intervia2);
    }
    else
      addznc(mx2,my2,Mx2,My2,layer2,winznc);
  }
}

void znc2rcn(win,sig)
WinZnc		*win;
losig_list	*sig;
{
  rdsrec_list	*scanrec,*r;
  short		 EndEqui;
  znc		 elem,*noeud1,*noeud2;
  chain_list	*interznc,*scaninter;
  chain_list	*scanchain,*scanchain1,*scanchain2;
  long		 n1,n2;
  long		 N1,N2;
  lorcnet_list	*ptrcnet;
  long		 rcntype;
  float		 rcnresi,rcncapa;
  long		 x,y,dx,dy,mx1,my1,mx2,my2;
  ptype_list	*ptl;
  GrgPhVia_list	*scanvia;
  lowire_list	*w;

#ifdef BUILDRCNDEBUG
  fprintf(stderr,"\n\nConstruction des wires \n\n");
#endif

  ptrcnet=getlorcnet(sig);

/* On place dans les intersections les rectangles qui les composent */

  scanrec=(rdsrec_list*)(getptype(sig->USER,RCN_FEQUI)->DATA);
  EndEqui=0;
  while(!EndEqui)
  {
    if(RectangleValide(scanrec))
    {
      elem.x1=scanrec->X;
      elem.y1=scanrec->Y;
      elem.x2=scanrec->X+scanrec->DX;
      elem.y2=scanrec->Y+scanrec->DY;
      elem.layer=GetRdsLayer(scanrec);
      elem.NEXT=NULL;

      interznc=getwininterznc(&elem,win,1);
      for(scaninter=interznc;scaninter;scaninter=scaninter->NEXT)
      {
        noeud1=(znc*)(scaninter->DATA);
        for(scanchain=noeud1->rec;scanchain;scanchain=scanchain->NEXT)
          if(scanchain->DATA==scanrec)
            break;
        if(!scanchain)
          noeud1->rec=addchain(noeud1->rec,scanrec);
      }
    }

    if(IsLynxEndEqui(scanrec))
      EndEqui=1;
    else
      scanrec=LYNX_EQUI(scanrec);
  }
  
  scanrec=(rdsrec_list*)(getptype(sig->USER,RCN_FEQUI)->DATA);
  EndEqui=0;

  while(!EndEqui)
  {
    if(RectangleValide(scanrec))
    {
      elem.x1=scanrec->X;
      elem.y1=scanrec->Y;
      elem.x2=scanrec->X+scanrec->DX;
      elem.y2=scanrec->Y+scanrec->DY;
      elem.layer=GetRdsLayer(scanrec);
      elem.NEXT=NULL;

      interznc=getwininterznc(&elem,win,1);


      if(interznc)
      {
#ifdef   BUILDRCNDEBUG
fprintf(stderr,"Traite   le rectangle (%ld,%ld)-(%ld,%ld)\n",
          scanrec->X,scanrec->Y,scanrec->X+scanrec->DX,scanrec->Y+scanrec->DY);
#endif

        if(GETDIRECTION(scanrec)==HORIZ_DIR)
          interznc=sortchain(interznc,'X');
        else
          interznc=sortchain(interznc,'Y');

        n1=0l;
        n2=0l;

        for(scaninter=interznc;scaninter;scaninter=scaninter->NEXT)
        {
          noeud1=(znc*)(scaninter->DATA);
          n1=noeud1->node;
    
          if(n1 && n2 && getlowire(ptrcnet,n1,n2)==NULL)
          {
            /* Place les rectangles communs au noeud1 et au noeud2 dans une
               chain_list */

            mx1=scanrec->X;
            mx2=mx1+scanrec->DX;
            my1=scanrec->Y;
            my2=my1+scanrec->DY;
            for(scanchain1=noeud1->rec;scanchain1;scanchain1=scanchain1->NEXT)
            {
              for(scanchain2=noeud2->rec;scanchain2;scanchain2=scanchain2->NEXT)
                if(scanchain1->DATA==scanchain2->DATA)
                {
                  r=(rdsrec_list*)(scanchain1->DATA);
                  mx1=min(mx1,r->X);
                  mx2=max(mx2,r->X+r->DX);
                  my1=min(my1,r->Y);
                  my2=max(my2,r->Y+r->DY);
                }
            }

            if(GETDIRECTION(scanrec)==HORIZ_DIR)
            {
              x=(noeud2->x2-noeud2->x1)/2 + noeud2->x1;
              dx=(noeud1->x2-noeud1->x1)/2 + noeud1->x1 - x;
              y=scanrec->Y+ scanrec->DY/2;
              dy=my2-my1;
              rcnresi=dx/dy*GET_LYNX_RESISTOR_LAYER(GetRdsLayer(scanrec));
              rcncapa=2*dx*GET_LYNX_PERIMETER_CAPA_LAYER(GetRdsLayer(scanrec))+
                      dx*dy*GET_LYNX_SURFACE_CAPA_LAYER(GetRdsLayer(scanrec));
            }
            else
            {
              x=scanrec->X+scanrec->DX/2;
              dx=mx2-mx1;
              y=(noeud2->y2-noeud2->y1)/2 + noeud2->y1;
              dy=(noeud1->y2-noeud1->y1)/2 + noeud1->y1 -y;
   
              rcnresi=dy/dx*GET_LYNX_RESISTOR_LAYER(GetRdsLayer(scanrec));
              rcncapa=2*dy*GET_LYNX_PERIMETER_CAPA_LAYER(GetRdsLayer(scanrec))+
                      dx*dy*GET_LYNX_SURFACE_CAPA_LAYER(GetRdsLayer(scanrec));
            }

            rcntype=RCN_WIRE_UNKNOW;

            if( GetRdsLayer(scanrec)==RDS_ALU1 ) rcntype=RCN_WIRE_ALU1;
            if( GetRdsLayer(scanrec)==RDS_ALU2 ) rcntype=RCN_WIRE_ALU2;
            if( GetRdsLayer(scanrec)==RDS_POLY ) rcntype=RCN_WIRE_POLY;

          
#ifdef   BUILDRCNDEBUG
fprintf(stderr,"Ajout   d'un wire en (%ld,%ld) de dimensions (%ld,%ld)\n",
                 x,y,dx,dy);
#endif

            if(dx>dy)
            {
              if( noeud1->x1 < noeud2->x1 )
              {
                N1=n1;
                N2=n2;
              }
              else
              {
                N1=n2;
                N2=n1;
              }
            }
            else
            {
              if(noeud1->y1 < noeud2->y1)
              {
                N1=n1;
                N2=n2;
              }
              else
              {
                N1=n2;
                N2=n1;
              }
            } 
            addlowire(ptrcnet,rcntype,rcnresi,rcncapa,x,y,dx,dy,N1,N2);
          }
          n2=n1;
          noeud2=noeud1;
        }

        freechain(interznc);
      }
    }

    if(IsLynxEndEqui(scanrec))
      EndEqui=1;
    else
      scanrec=LYNX_EQUI(scanrec);
  }

  /* On s'occupe maintenant des vias */
  if(ptl=getptype(sig->USER,GRGVIA))
  {
    for(scanvia=(GrgPhVia_list*)(ptl->DATA);scanvia;scanvia=scanvia->NEXT)
    {
      if( GetRdsLayer(scanvia->LAYER1) == RDS_CONT ||
          GetRdsLayer(scanvia->LAYER1) == RDS_VIA1 ||
          GetRdsLayer(scanvia->LAYER1) == RDS_VIA2 ||
          GetRdsLayer(scanvia->LAYER1) == RDS_VIA3   )
      {
        elem.x1=scanvia->LAYER2->X;
        elem.y1=scanvia->LAYER2->Y;
        elem.x2=scanvia->LAYER2->DX+elem.x1;
        elem.y2=scanvia->LAYER2->DY+elem.y1;
        elem.layer=GetRdsLayer(scanvia->LAYER2);
        scanchain1=getwininterznc(&elem,win,1);
        n1=((znc*)(scanchain1->DATA))->node;
        freechain(scanchain1);

        elem.x1=scanvia->LAYER3->X;
        elem.y1=scanvia->LAYER3->Y;
        elem.x2=scanvia->LAYER3->DX+elem.x1;
        elem.y2=scanvia->LAYER3->DY+elem.y1;
        elem.layer=GetRdsLayer(scanvia->LAYER3);
        scanchain1=getwininterznc(&elem,win,1);
        n2=((znc*)(scanchain1->DATA))->node;
        rcnresi = GET_LYNX_RESISTOR_LAYER(GetRdsLayer(scanvia->LAYER1));
        freechain(scanchain1);
      }

      if( GetRdsLayer(scanvia->LAYER2) == RDS_CONT ||
          GetRdsLayer(scanvia->LAYER2) == RDS_VIA1 ||
          GetRdsLayer(scanvia->LAYER2) == RDS_VIA2 ||
          GetRdsLayer(scanvia->LAYER2) == RDS_VIA3   )
      {
        elem.x1=scanvia->LAYER1->X;
        elem.y1=scanvia->LAYER1->Y;
        elem.x2=scanvia->LAYER1->DX+elem.x1;
        elem.y2=scanvia->LAYER1->DY+elem.y1;
        elem.layer=GetRdsLayer(scanvia->LAYER1);
        scanchain1=getwininterznc(&elem,win,1);
        n1=((znc*)(scanchain1->DATA))->node;
        freechain(scanchain1);

        elem.x1=scanvia->LAYER3->X;
        elem.y1=scanvia->LAYER3->Y;
        elem.x2=scanvia->LAYER3->DX+elem.x1;
        elem.y2=scanvia->LAYER3->DY+elem.y1;
        elem.layer=GetRdsLayer(scanvia->LAYER3);
        scanchain1=getwininterznc(&elem,win,1);
        n2=((znc*)(scanchain1->DATA))->node;
        rcnresi = GET_LYNX_RESISTOR_LAYER(GetRdsLayer(scanvia->LAYER2));
        freechain(scanchain1);
      }

      if( GetRdsLayer(scanvia->LAYER3) == RDS_CONT ||
          GetRdsLayer(scanvia->LAYER3) == RDS_VIA1 ||
          GetRdsLayer(scanvia->LAYER3) == RDS_VIA2 ||
          GetRdsLayer(scanvia->LAYER3) == RDS_VIA3   )
      {
        elem.x1=scanvia->LAYER2->X;
        elem.y1=scanvia->LAYER2->Y;
        elem.x2=scanvia->LAYER2->DX+elem.x1;
        elem.y2=scanvia->LAYER2->DY+elem.y1;
        elem.layer=GetRdsLayer(scanvia->LAYER2);
        scanchain1=getwininterznc(&elem,win,1);
        n1=((znc*)(scanchain1->DATA))->node;
        freechain(scanchain1);

        elem.x1=scanvia->LAYER1->X;
        elem.y1=scanvia->LAYER1->Y;
        elem.x2=scanvia->LAYER1->DX+elem.x1;
        elem.y2=scanvia->LAYER1->DY+elem.y1;
        elem.layer=GetRdsLayer(scanvia->LAYER1);
        scanchain1=getwininterznc(&elem,win,1);
        n2=((znc*)(scanchain1->DATA))->node;
        rcnresi = GET_LYNX_RESISTOR_LAYER(GetRdsLayer(scanvia->LAYER3));
        freechain(scanchain1);
      }


#ifdef BUILDRCNDEBUG
fprintf(stderr,"Via entre les noeuds %ld et %ld\n",n1,n2);
#endif
      if((w=getlowire(ptrcnet,n1,n2))==NULL)
      {
         addlowire(ptrcnet,RCN_WIRE_CONT_VIA,rcnresi,0.0,
                   (elem.x2-elem.x1)/2+elem.x1,
                   (elem.y2-elem.y1)/2+elem.y1,0,0,n1,n2);
      }
      else
      {
         addpararesiwire(w,rcnresi);
      }
    }
  }
}

chain_list* sortchain(entree,champ)
chain_list	*entree;
char		 champ;
{
  chain_list	*scanchain,*sortie,*last,*prevmax,*eff;
  znc		*maxi,*z;

  /* tri a bulle de la liste des znc */

  /* On veut faire un tri par ordre croissant en utilisant la fonction
     chain_list qui ajoute les elements a l'envers : on va donc faire
     un tri par ordre decroissant */

  if(!entree)
  {
    fprintf(stderr,"Trie de chaine (null). \n");
    exit(-1);
  }

  if(entree->NEXT==NULL)
    return(entree);

  sortie=NULL;

  while(entree)
  {
    /* il faut debuter avec qqchose */
    maxi=(znc*)entree->DATA;

    /* recherche de l'element le plus grand dans la liste */
    last=NULL;
    prevmax=NULL;

    for(scanchain=entree;scanchain;scanchain=scanchain->NEXT)
    {
      z=(znc*)(scanchain->DATA);
 
      if(champ=='X' && z->x1 > maxi->x1 ||
         champ=='Y' && z->y1 > maxi->y1   )
      { 
        maxi = z;
        prevmax=last;
      }
      last=scanchain;
    }

    /* maitenant qu'on l'a trouve, on le met dans la chainlist de sortie */
    sortie=addchain(sortie,maxi);

    /* efface l'element de la liste chainee d'entree */
    if(prevmax)
    {
      eff=prevmax->NEXT;
      prevmax->NEXT=prevmax->NEXT->NEXT;
    }
    else
    {
      eff=entree;
      entree=entree->NEXT;
    }
    if(!eff)
    {
      fprintf(stderr,"eff=null\n");  
      exit(-1);
    }
    eff->NEXT=NULL;
    freechain(eff);
  }

#ifdef BUILDRCNDEBUG
{
  chain_list	*scan;
  for(scan=sortie;scan;scan=scan->NEXT)
    fprintf(stderr,"znc : (%ld,%ld)-(%ld,%ld) [%ld] lonode :%ld\n",
      ((znc*)(scan->DATA))->x1,
      ((znc*)(scan->DATA))->y1,
      ((znc*)(scan->DATA))->x2,
      ((znc*)(scan->DATA))->y2,
      ((znc*)(scan->DATA))->layer,
      ((znc*)(scan->DATA))->node);
}
#endif
  return(sortie);
}

void sortloconnode(connect,winznc)
chain_list	*connect;
WinZnc		*winznc;
{

  chain_list	*conlofig;  
  chain_list	*scanchain;
  rdsrec_list	*rec;
  ptype_list	*ptl;
  locon_list	*logical_connector;

  struct ssort
  {
    struct ssort	*NEXT;
    locon_list		*connector;
    chain_list		*conlofig;
  } *sort,*scansort;


  /* Recupere la liste des connecteurs de la figure */
  sort=NULL;
  for(scanchain=connect;scanchain;scanchain=scanchain->NEXT)
  {
    rec=(rdsrec_list*)(scanchain->DATA);
    if(ptl=getptype(rec->USER,RDSLOCON))
    {
      logical_connector=(locon_list*)(ptl->DATA);
      
      if(logical_connector->TYPE=='I' ||
         logical_connector->TYPE=='E'    )
      {

        for(scansort=sort;scansort;scansort=scansort->NEXT)
        {
          if(scansort->connector == logical_connector)
            break;
        }

        if(scansort)
        {
          scansort->conlofig=addchain(scansort->conlofig,rec);
        }
        else
        {
          scansort=mbkalloc(sizeof(struct ssort));
          scansort->NEXT=sort;
          scansort->connector=logical_connector;
          scansort->conlofig=addchain(NULL,rec);

          sort=scansort;
        }
      }
    }
    else
    {
      fprintf(stderr,"**** ERREUR DANS LA FONCTION SORTLOCONNODE !!!!\n");
      fprintf(stderr,"Un rectangle qui n'est pas un connecteur a ete classe\n");
      fprintf(stderr,"dans une liste de connecteurs.\n\n");
      exit(1);
    }
  }

  for(scansort=sort;scansort;scansort=scansort->NEXT)
  {
    if(! scansort->conlofig) 
    {
      fprintf(stderr,"Connecteur-> %s\n",scansort->connector->NAME);
      exit(1);
    }
    sortbullelocon(scansort->connector,scansort->conlofig);
  }
}

void sortbullelocon(logical_connector,conlofig)
locon_list	*logical_connector;
chain_list	*conlofig;
{
  chain_list	*scanchain,*trie;
  rdsrec_list	*rec,*maxi;
  chain_list	*prevmax,*last;
  chain_list	*eff;
  ptype_list	*ptl;
  num_list	*node;

  /* Dans la chain_list conlofig on a la liste des rectangles connecteurs 
     de la figure */

  /* on effectue un trie a bulle sur le champ Y */  

  if(conlofig->NEXT!=NULL)
  {
    trie=NULL;
    while(conlofig)
    {
      maxi=(rdsrec_list*)conlofig->DATA;
      last=NULL;
      prevmax=NULL;

      for(scanchain=conlofig;scanchain;scanchain=scanchain->NEXT)
      {
        rec=(rdsrec_list*)(scanchain->DATA);
        if( rec->Y < maxi->Y )
        {
          maxi=rec;
          prevmax=last;
        }
        else
        {
          if(rec->Y == maxi->Y)
            if(rec->X < maxi->X)
            {
              maxi=rec;
              prevmax=last;
            }
        }
        
        last=scanchain;
      }

      trie=addchain(trie,maxi);

      if(prevmax)
      {
        eff=prevmax->NEXT;
        prevmax->NEXT=prevmax->NEXT->NEXT;
      }
      else
      {
        eff=conlofig;
        conlofig=conlofig->NEXT;
      }
      eff->NEXT=NULL;
      freechain(eff);
    }
  }
  else
    trie=conlofig;

  node=NULL;
  for(scanchain=trie;scanchain;scanchain=scanchain->NEXT)
  {
    rec=((rdsrec_list*)(scanchain->DATA));
    ptl=getptype(rec->USER,TMPLONODE);
    if(!ptl)
    {
      fprintf(stderr,"Un connecteur (%s) n'est relie a aucun lonode. %c\n",
                     logical_connector->NAME,
                     logical_connector->TYPE);
      exit(1);
    }
    node=addnum(node,((long)(ptl->DATA)));
  }

  /* on a la liste des connecteurs triee dans la chain_list trie */
  
  ptl=getptype(logical_connector->USER,RCN_LONODE);
  if(!ptl)
  {
    fprintf(stderr,"Pas de RCN_LONODE pour le connecteur logique %s\n",
                   logical_connector->NAME);
    exit(1);
  }
  freenum(ptl->DATA);
  ptl->DATA=node;
  freechain(trie);
}
