/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/************************************************************************
*       MBK : Parser alliance logic                                     *
*                                                                       *
*       version : 5.00	                                                *
*       date : 10/05/95                                                 *
************************************************************************/

#include <ctype.h>
#include <string.h>
#include "generic_target.h"
#include MUT_H
#include MLO_H
#include MLU_H
#include RCN_H

#define TAILLE_CHAINE BUFSIZE


/********************************************************
* addnumlocon(locon*,num_list*)                         *
* Cree un lien entre le connecteur et les noeud de la   *
* num_list en utilisant les fonctions prevues a cet     *
* usage.                                                *
********************************************************/

void addnumlocon(ptlocon,headnum)
locon_list	*ptlocon;
num_list	*headnum;
{
  num_list	*scannum;
  losig_list	*signal;
  lorcnet_list	*rcview;
  num_list	*invnum;

  signal=ptlocon->SIG;
  if(!signal)
  {
    fprintf(stderr,"\n*** rcn error *** loadlofig (al) No signal found on connector %s\n",ptlocon->NAME);
    EXIT(1);
  }

  rcview=getlorcnet(signal);
  if(!rcview)
    rcview=addlorcnet(signal);

  /* reverse num_list */
  invnum=NULL;
  for(scannum=headnum;scannum;scannum=scannum->NEXT)
  {
    invnum=addnum(invnum,scannum->DATA);
  }
  freenum(headnum);

  for(scannum=invnum;scannum;scannum=scannum->NEXT)
  {
    setloconnode(rcview,scannum->DATA,ptlocon);
  }
  freenum(invnum);
}

/************************************************************************
*                  alc_typecode()                                       *
************************************************************************/
char alc_typedecode(type)
char * type;
{
        switch (type[0])
        {
                case 'E' :
                        return EXTERNAL;
                default :
                case 'I' :
                        return INTERNAL;
        }
}


/************************************************************************
*                  rcn_typewire()                                       *
************************************************************************/

long  rcn_typewire(type)
char *type;
{
	if(strcmp(type,"X")==0)  return RCN_WIRE_UNKNOW;
 	if(strcmp(type,"PY")==0) return RCN_WIRE_POLY;
 	if(strcmp(type,"A1")==0) return RCN_WIRE_ALU1;
 	if(strcmp(type,"A2")==0) return RCN_WIRE_ALU2;
 	if(strcmp(type,"CY")==0) return RCN_WIRE_CONT_POLY;
 	if(strcmp(type,"CN")==0) return RCN_WIRE_CONT_DIF_N;
 	if(strcmp(type,"CP")==0) return RCN_WIRE_CONT_DIF_P;
 	if(strcmp(type,"CV")==0) return RCN_WIRE_CONT_VIA;
 	if(strcmp(type,"CW")==0) return RCN_WIRE_CONT_VIA2;
(void)fprintf(stderr,"\n*** rcn error *** loadlofig (al) invalid wire type : [%s]\n",type);
return RCN_WIRE_UNKNOW;
}

/************************************************************************
*                  alcloadlofig5()                                      *
************************************************************************/

void alcloadlofig5(pfig,fname,mode,in)
lofig_list * pfig;
char * fname;
char mode;
FILE * in;
{
	ptype_list * ptype;
	chain_list * pchains;
	chain_list * pchainaux;
	chain_list * pchainnode;
	chain_list * pchainnodex;
	num_list * pnum;
	num_list * pnumaux;
	lofig_list * ptfigmodele;
	lofig_list * tetemodele = NULL;
	loins_list * pins;
	losig_list * psig;
	losig_list * psigd;
	losig_list * psigg;
	losig_list * psigs;
	lotrs_list * ptrs;
	locon_list * pcon;
	lorcnet_list * prcnet;
	lowire_list * pwire;
	lonode_list * pnode;
	char ligne[5*TAILLE_CHAINE];
	char nomtmp[TAILLE_CHAINE], * paux, *ptemp, inst_avant, sig_avant;
	char * type, decodedtype, * name, insname[TAILLE_CHAINE], figname[TAILLE_CHAINE], nomodele;
	long i, index, l, w, x, y, ps, pd, xs, xd, equig, equid, equis, wx, wy, wdx, wdy;
	long indexd, indexg, indexs, n1, n2, typew;
	float capa;
	float resi;
/* to check signal name unicity */
	ht * h;

        pchainnode=NULL;
/* Analyse de la ligne d'entete H */
	(void)fgets(ligne,5*TAILLE_CHAINE,in);
	if (ligne[0]!='H')
	{
		(void)fprintf(stderr,"\n*** mbk error *** line 2 unexpected, header needed\n");
		EXIT(1);
	}
	if ((paux=(char *)strchr(ligne,',')) == NULL)
	{
		(void)fprintf(stderr,"\n*** mbk error *** syntax error line : 2\n");
		EXIT(1);
	}
	*paux='\0';
/* controle du nom */
	if (strcmp(ligne+2,fname)!=0)
	{
		(void)fprintf(stderr,"\n*** mbk error *** invalid name, line : 2\n");
		EXIT(1);
	}
	paux++;
/*controle du type de fichier */
	if (*paux!='L')
	{
		(void)fprintf(stderr,"\n*** mbk error *** bad file type\n");
		EXIT(1);
	}
/* boucle de lecture */
	inst_avant  = nomodele = sig_avant = (char)0;
	h = addht (100); /* create Hash Table for signal name unicity */
	for(i=3;;i++)
	{
		if (fgets(ligne,5*TAILLE_CHAINE,in)==NULL)
		{
			(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
			EXIT(1);
		}
		if (strncmp(ligne,"EOF",3)==0)
			break;
		switch(ligne[0])
		{
			case 'C' :
				if ((mode == 'C') && (!inst_avant))
					continue;
				name = ligne + 2;
				if ((paux=(char *)strchr(ligne+2,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				type = ptemp;
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				decodedtype = alc_typedecode(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
				index = atol(ptemp);
				pnum = NULL;
				}
				else
				{
				*paux='\0';
				index = atol(ptemp);
				ptemp=paux+1;
				pnum = NULL;
				while((paux=(char *)strchr(ptemp,',')) != NULL)
				   {
				    *paux='\0';
				    pnum = addnum(pnum,atol(ptemp));
				    ptemp=paux+1;
				   }
				pnum = addnum(pnum,atol(ptemp));
				pnum = (num_list *)reverse((chain_list *)pnum);
				}
				if (inst_avant)
				{
					if (decodedtype != INTERNAL)
					{
						(void)fprintf(stderr,"\n*** mbk Warning *** loadlofig (al) bad connector direction line : %d\n",i);
					}
					psig = givelosig(ptfigmodele,index);
					psig->TYPE = 'E';
					if (nomodele)
						(void)addlocon(ptfigmodele,name,psig,alc_directiondecode(type));
					pchains = addchain(pchains,(void *)givelosig(pfig,index));
					if(pnum != NULL) pchainnode = addchain(pchainnode,(void *)pnum);
					else pchainnode = addchain(pchainnode,NULL);
				}
				else
				{
					if (decodedtype != EXTERNAL)
					{
						(void)fprintf(stderr,"\n*** mbk Warning *** loadlofig (al) bad connector direction line : %d\n",i);
					}
					psig = givelosig(pfig,index);
					psig->TYPE = 'E';
					pcon = addlocon(pfig,name,psig,alc_directiondecode(type));
					if (pnum != NULL) 
					{
                                         /* 
					 pcon->USER = addptype(pcon->USER,RCN_LONODE,pnum);
                                         Modified by Greg 12/06/96 */
                                         addnumlocon(pcon,pnum);
					}
				}
				sig_avant = 0;
				break;
			case 'T' :
				if (mode == 'P')
					goto end_lopars;
				if (inst_avant)
				{
					pins = addloins(pfig,insname,ptfigmodele,pchains);
					pchainnodex = pchainnode;
					pchainaux = pchains;
					for(pcon=pins->LOCON; pcon; pcon=pcon->NEXT)
					{
						pnum = (num_list *)pchainnodex->DATA;
						if (pnum != NULL) 
						{
                                                 /*
					 	 pcon->USER = addptype(pcon->USER,RCN_LONODE,pnum);
                                                 Modified by Greg 12/06/96 */
                                                 addnumlocon(pcon,pnum);
						}
						pchainnodex = pchainnodex->NEXT;
						pchainaux = pchainaux->NEXT;
					}
					freechain(pchains);
					freechain(pchainnode);
                                        /* Modified by Greg 22/08/96 */
                                        pchainnode=NULL;
                                        /* End modified */
				}
				inst_avant = 0;
				pchains = NULL;
				if ((paux=(char *)strchr(ligne+2,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				ptemp=paux+1;
				type = ligne+2;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				l = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				w = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				equid = atol(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				equig = atol(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				equis = atol(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				xs = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				xd = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				ps = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				pd = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				x = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
				  indexd = 0L;
				  indexg = 0L;
				  indexs = 0L;
				  y = (long)(atof(ptemp) * (float)SCALE_X);
				}
				else
				{
				  *paux='\0';
				  y = (long)(atof(ptemp) * (float)SCALE_X);
				  ptemp=paux+1;
				  if ((paux=(char *)strchr(ptemp,',')) == NULL)
				  {
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				  }
				  *paux='\0';
				  indexd = atol(ptemp);
				  ptemp=paux+1;
				  if ((paux=(char *)strchr(ptemp,',')) == NULL)
				  {
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				  }
				  *paux='\0';
				  indexg = atol(ptemp);
				  ptemp=paux+1;
				  indexs = atol(ptemp);
				}
				psigd = givelosig(pfig,equid);
				psigg = givelosig(pfig,equig);
				psigs = givelosig(pfig,equis);
				ptrs = addlotrs(pfig, (*type == 'N') ? TRANSN : TRANSP, x, y, w, l, ps, pd, xs, xd, psigg, psigs, psigd);
				if (indexd != 0L) 
				{
                                 /*
				 ptrs->DRAIN->USER = addptype(ptrs->DRAIN->USER,RCN_LONODE,addnum(NULL,indexd)) ;
                                 Modified by Greg 12/06/96 */
                                 addnumlocon(ptrs->DRAIN,addnum(NULL,indexd));
				}
				if (indexg != 0L) 
				{
                                 /*
				 ptrs->GRID->USER = addptype(ptrs->GRID->USER,RCN_LONODE,addnum(NULL,indexg)) ;
                                 Modified by Greg 12/06/96 */
                                 addnumlocon(ptrs->GRID,addnum(NULL,indexg));
				}
				if (indexs != 0L) 
				{
                                 /*
				 ptrs->SOURCE->USER = addptype(ptrs->SOURCE->USER,RCN_LONODE,addnum(NULL,indexs)) ;
                                 Modified by Greg 12/06/96 */
                                 addnumlocon(ptrs->SOURCE,addnum(NULL,indexs));
				}
				sig_avant = 0;
				break;
			case 'I' :
				if (mode == 'P')
					goto end_lopars;
				if (inst_avant)
				{
					pins = addloins(pfig,insname,ptfigmodele,pchains);
					pchainnodex = pchainnode;
					pchainaux = pchains;
					for(pcon=pins->LOCON; pcon; pcon=pcon->NEXT)
					{
						pnum = (num_list *)pchainnodex->DATA;
						if (pnum != NULL) 
						{
                                                 /*
						 pcon->USER = addptype(pcon->USER,RCN_LONODE,pnum);
                                                 Modified by Greg 12/06/96 */
                                                 addnumlocon(pcon,pnum);
						}
						pchainnodex = pchainnodex->NEXT;
						pchainaux = pchainaux->NEXT;
					}
					freechain(pchains);
					freechain(pchainnode);
                                        /* Modified by Greg 22/08/96 */
                                        pchainnode=NULL;
                                        /* End modified */
				}
				pchains = NULL;
				pnum = NULL;
				inst_avant = 1;
				if ((paux=(char *)strchr(ligne+2,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				(void)strcpy(figname,ligne+2);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,'\n')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				(void)strcpy(insname,ptemp);
				if ((ptfigmodele = getlomodel(tetemodele,figname)) == NULL)
				{
					nomodele = 1;
					ptfigmodele = tetemodele = addlomodel(tetemodele,figname);
				}
				else
					nomodele = 0;
				sig_avant = 0;
				break;
			case 'S' :
				sig_avant = 1;
				if (mode == 'P')
					goto end_lopars;
				if (inst_avant)
				{
					pins = addloins(pfig,insname,ptfigmodele,pchains);
					pchainnodex = pchainnode;
					pchainaux = pchains;
					for(pcon=pins->LOCON; pcon; pcon=pcon->NEXT)
					{
						pnum = (num_list *)pchainnodex->DATA;
						if (pnum != NULL) 
						{
                                                 /*
						 pcon->USER = addptype(pcon->USER,RCN_LONODE,pnum);
                                                 Modified by Greg 12/06/96 */
                                                 addnumlocon(pcon,pnum);
						}
						pchainnodex = pchainnodex->NEXT;
						pchainaux = pchainaux->NEXT;
					}
					freechain(pchains);
					freechain(pchainnode);
                                        /* Modified by Greg 22/08/96 */
                                        pchainnode=NULL;
                                        /* End modified */
				}
				inst_avant = 0;
				pchains = NULL;
				if ((paux=(char *)strchr(ligne+2,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				index = atol(ligne+2);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				decodedtype = alc_typedecode(ptemp);
				psig = givelosig(pfig,index);
				psig->TYPE = decodedtype;
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					capa = atof(ptemp);
					addcapa(psig,capa);
					(void)sprintf(nomtmp,"mbk_sig%d", index);
					if (sethtitem(h, namealloc (nomtmp), 0))
					{
						for (paux = nomtmp; *paux != '\0'; paux++);
						*(paux + 11) = '\0';
						for (x = (long)'a'; x <= (long)'z'; x++)
						for (y = (long)'a'; y <= (long)'z'; y++)
						for (l = (long)'a'; l <= (long)'z'; l++)
						for (w = (long)'a'; w <= (long)'z'; w++)
						for (ps = (long)'a'; ps <= (long)'z'; ps++)
						for (pd = (long)'a'; pd <= (long)'z'; pd++)
						for (xs = (long)'a'; xs <= (long)'z'; xs++)
						for (xd = (long)'a'; xd <= (long)'z'; xd++)
						for (equig = (long)'a'; equig <= (long)'z'; equig++)
						for (equid = (long)'a'; equid <= (long)'z'; equid++)
						for (equis = (long)'a'; equis <= (long)'z'; equis++)
						{
							*paux = (char)x;
							*(paux + 1) = (char)y;
							*(paux + 2) = (char)l;
							*(paux + 3) = (char)w;
							*(paux + 4) = (char)ps;
							*(paux + 5) = (char)pd;
							*(paux + 6) = (char)xs;
							*(paux + 7) = (char)xd;
							*(paux + 8) = (char)equig;
							*(paux + 9) = (char)equid;
							*(paux + 10) = (char)equis;
							
							if (!sethtitem(h, namealloc(nomtmp), 0))
							{
								goto break_11_for;
								
							}
						}
					}
break_11_for:					
					psig->NAMECHAIN = addchain(psig->NAMECHAIN,(void *)namealloc(nomtmp));
					break; /* case 'S' : */
				}
				*paux='\0';
				capa = atof(ptemp);
				addcapa(psig,capa);
				/* restore NAMECHAIN */
				x = 1;
				ptemp=paux+1;
				while ((paux=(char *)strchr(ptemp,',')) != NULL)
				{
					*paux='\0';
					if ((*ptemp != '\0') && (strcmp(ptemp,"*") != 0))
					{
						x++;
						psig->NAMECHAIN = addchain(psig->NAMECHAIN,(void *)namealloc(ptemp));
					}
					ptemp=paux+1;
				}
				if ((paux=(char *)strchr(ptemp,'\n')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				if ((*ptemp != '\0') && (strcmp(ptemp,"*") != 0))
				{
					x++;
					psig->NAMECHAIN = addchain(psig->NAMECHAIN,(void *)namealloc(ptemp));
				}
				/* set only one valid name for the signal */
				while (x > 0)
				{
					paux = getsigname(psig);
					if (sethtitem(h, paux, 0)) /* check unicity */
					{ /* name is not unique */
						for (pchains = psig->NAMECHAIN; pchains; pchains = pchains->NEXT)
							if (paux == ((char *)(pchains->DATA)))
								break;
						if (pchains)
							psig->NAMECHAIN = delchain (psig->NAMECHAIN, pchains);
						pchains = NULL;
						x --;
					}
					else
						break;
				}
				freechain (psig->NAMECHAIN);
				psig->NAMECHAIN = NULL;

				if (x <= 0)
				{ /* no unique name is found in the list, generate one */
					(void)sprintf(nomtmp,"mbk_sig%d", index);
					if (sethtitem(h, namealloc(nomtmp), 0))
					{
						for (paux = nomtmp; *paux != '\0'; paux++);
						*(paux + 11) = '\0';
						for (x = (long)'a'; x <= (long)'z'; x++)
						for (y = (long)'a'; y <= (long)'z'; y++)
						for (l = (long)'a'; l <= (long)'z'; l++)
						for (w = (long)'a'; w <= (long)'z'; w++)
						for (ps = (long)'a'; ps <= (long)'z'; ps++)
						for (pd = (long)'a'; pd <= (long)'z'; pd++)
						for (xs = (long)'a'; xs <= (long)'z'; xs++)
						for (xd = (long)'a'; xd <= (long)'z'; xd++)
						for (equig = (long)'a'; equig <= (long)'z'; equig++)
						for (equid = (long)'a'; equid <= (long)'z'; equid++)
						for (equis = (long)'a'; equis <= (long)'z'; equis++)
						{
							*paux = (char)x;
							*(paux + 1) = (char)y;
							*(paux + 2) = (char)l;
							*(paux + 3) = (char)w;
							*(paux + 4) = (char)ps;
							*(paux + 5) = (char)pd;
							*(paux + 6) = (char)xs;
							*(paux + 7) = (char)xd;
							*(paux + 8) = (char)equig;
							*(paux + 9) = (char)equid;
							*(paux + 10) = (char)equis;
							
							if (!sethtitem(h, namealloc(nomtmp), 0))
							{
								goto break_11_for2;
								
							}
						}
					}
break_11_for2:
					paux = nomtmp;
				}
				psig->NAMECHAIN = addchain(psig->NAMECHAIN,(void *)namealloc(paux));
				break;
			case 'W' :
				if (sig_avant == 0)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				if((ptype = getptype(psig->USER,RCN_RCNET)) != NULL) prcnet = (lorcnet_list *)ptype->DATA ;
				else prcnet = addlorcnet(psig);
				if ((paux=(char *)strchr(ligne+2,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				ptemp=paux+1;
				n1 = atol(ligne+2);
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				n2 = atol(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				typew = rcn_typewire(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				resi = atof(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				capa = atof(ptemp);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				wx = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				wy = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				if ((paux=(char *)strchr(ptemp,',')) == NULL)
				{
					(void)fprintf(stderr,"\n*** mbk error *** syntax error line : %d\n",i);
					EXIT(1);
				}
				*paux='\0';
				wdx = (long)(atof(ptemp) * (float)SCALE_X);
				ptemp=paux+1;
				wdy = (long)(atof(ptemp) * (float)SCALE_X);
				(void)addlowire(prcnet, typew, resi, capa, wx, wy, wdx, wdy, n1, n2);
				break;
			default :
				(void)fprintf(stderr,"\n*** mbk error *** line %d unexpected\n",i);
				EXIT(1);
				
		}
	}
end_lopars :
	if (inst_avant)
	{
		pins = addloins(pfig,insname,ptfigmodele,pchains);
		pchainnodex = pchainnode;
		pchainaux = pchains;
		for(pcon=pins->LOCON; pcon; pcon=pcon->NEXT)
		{
			pnum = (num_list *)pchainnodex->DATA;
			if (pnum != NULL) 
			{
                         /*
                         pcon->USER = addptype(pcon->USER,RCN_LONODE,pnum);
                         Modified by Greg 12/06/96 */
                         addnumlocon(pcon,pnum);
			}
			pchainnodex = pchainnodex->NEXT;
			pchainaux = pchainaux->NEXT;
		}
		freechain(pchains);
		freechain(pchainnode);
                /* Modified by Greg 22/08/96 */
                pchainnode=NULL;
                /* End modified */
	}
	delht(h); /* delete the Hash Table */
	freelomodel(tetemodele);
}
