/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include "msl_headers.h"

/*****************************************************************************/
/*                                                                           */
/*                                  FORMATER_LIGNE                           */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Met la ligne en minuscule, et supprime tous les blancs et tabs inutiles   */
/*                                                                           */
/* ENTREE : la ligne a traiter                                               */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void 
formater_ligne(ligne)
    char            ligne[];
{
    char            bid[MAXLIG + 1];
    register int    lg, cpt_lig, i, j;

    cpt_lig = 0;
    strcpy(bid, ligne);
    lg = strlen(bid) - 1;        /* - 1, car on ne veut pas le '\n'*/

    /* enleve les blancs de debut */
    for (i = 0; (i < lg) && (isspace(bid[i])); ++i);

    for (j = i; j < lg; ++j) {
        if (isspace(bid[j])) {        /* si on a trouve un espace */
            for (; (j < lg) && isspace(bid[j]); ++j);
            --j;
            /* 1 seul espce enttre deux mots */
            ligne[cpt_lig++] = ' ';
        }
        else if (isalpha(bid[j]))        /* si c'est une lettre */
            /* mise en minuscule */
            ligne[cpt_lig++] = tolower(bid[j]);
        else
            ligne[cpt_lig++] = bid[j];
    }
    ligne[cpt_lig] = '\0';

    /* on enleve les blancs de la fin de ligne */
    for (i = strlen(ligne) - 1; (i > 0) && (isspace(ligne[i])); --i)
        ligne[i] = '\0';

}

/*****************************************************************************/
/*                                                                           */
/*                               ENLEVE_BLANCS_EGAL                          */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Enleve les blancs autour des signes '='                                   */
/*                                                                           */
/* ENTREE : la ligne a traiter                                               */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void 
enleve_blancs_egal(ligne)
    char            ligne[];
{
    char            bid[MAXLIG + 1];
    register int    lg, cpt_lig, i;

    cpt_lig = 0;
    strcpy(bid, ligne);
    lg = strlen(bid);

    for (i = 0; i < lg; ++i) {
        /* on a trouve la chaine ' = ' */
        if (!strncmp(&bid[i], " = ", 3)) {
            ligne[cpt_lig++] = '=';
            i += 2;
        }
        else
            /* on a trouve la chaine '= ' */
        if (!strncmp(&bid[i], "= ", 2)) {
            ligne[cpt_lig++] = '=';
            i++;
        }
        else
            /* on a trouve la chaine ' =' */
        if (!strncmp(&bid[i], " =", 2)) {
            ligne[cpt_lig++] = '=';
            i++;
        }
        else {                        /* on n'a pas trouve de chaine avec un egal et espace */
            ligne[cpt_lig++] = bid[i];
        }
    }
    ligne[cpt_lig] = '\0';
}

/*****************************************************************************/
/*                                                                           */
/*                                 TROUVER_PLUS                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne suivante contient un plus                           */
/* Si c'est le cas, il faut la concatener a celle deja lue                   */
/*                                                                           */
/* ENTREE : la ligne deja lue                                                */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void 
trouver_plus(ligne)
    char            ligne[];
{
    char            car;

    while ((car = fgetc(infile)) == '+') {
        concat_ligne(ligne);
        ++msl_line;
    }

    if (car != EOF)
        fseek(infile, -1, 1);

}

/*****************************************************************************/
/*                                                                           */
/*                                 CONCAT_LIGNE                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Concatene la ligne lue a la ligne existante                               */
/*                                                                           */
/* ENTREE : la ligne deja lue                                                */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void 
concat_ligne(ligne)
    char            ligne[];
{
    char            bid[MAXLIG + 1];

    fgets(bid, MAXLIG, infile);
    formater_ligne(bid);
    strcat(ligne, " ");
    strcat(ligne, bid);
}

/*****************************************************************************/
/*                                                                           */
/*                                 DICO_CADENCE                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante est de la forme "* net <num> = <nom>"      */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne de type cadence                             */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
dico_cadence(ligne)
    char            ligne[];
{
    char           *ptr, *mot;
    char            bid[MAXLIG + 1];

    strcpy(bid, ligne);

    ptr = bid;
    mot = strtok(ptr, " ");        /* on a le caractere '*' */
    if (mot == (char *) 0)
        return (0);
    if (*mot != '*')
        return (0);
    ptr = (char *) 0;
    mot = strtok(ptr, " ");        /* on a le 'net' */
    if (mot == (char *) 0)
        return (0);
    if (strcmp(mot, "net"))
        return (0);
    mot = strtok(ptr, " ");        /* on a le nombre */
    if (mot == (char *) 0)
        return (0);
    if (!is_num(mot))
        return (0);
    mot = strtok(ptr, " ");        /* on a le '=' */
    if (mot == (char *) 0)
        return (0);
    if (*mot != '=')
        return (0);
    mot = strtok(ptr, " ");        /* on a le nom */
    if (mot == (char *) 0)
        return (0);
    if (strtok(ptr, " ") != (char *) 0)        /* syntaxe mauvaise */
        return (0);
    return (1);
}

/*****************************************************************************/
/*                                                                           */
/*                                 DICO_DAX                                  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante est '************* nodes *************     */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne de type dax                                 */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
dico_dax(ligne)
    char            ligne[];
{
    return (!strcmp(ligne, DICO_DAX));
}

/*****************************************************************************/
/*                                                                           */
/*                                 DICO_VTI                                  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante est '*  and synonyms'                      */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne de type vti                                 */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
dico_vti(ligne)
    char            ligne[];
{
    return (!strcmp(ligne, DICO_VTI));
}

/*****************************************************************************/
/*                                                                           */
/*                                 DICO_ELDO                                 */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante commence par '.connect'                    */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne de type eldo                                */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
dico_eldo(ligne)
    char            ligne[];
{
    register int    num, i, lg;
    char            nom[MAXNAME + 1];
    char           *ptr, *mot;
    char            bid[MAXLIG + 1];

    strcpy(bid, ligne);
    ptr = bid;
    mot = strtok(ptr, " ");        /* on passe le .connect */
    if (mot == (char *) 0)
        return (0);
    if (strcmp(mot, ".connect"))
        return (0);
    ptr = (char *) 0;

    mot = strtok(ptr, " ");        /* on a soit le nom, soit le num */
    if (mot == (char *) 0)
        return (0);

    if (isalpha(*mot)) {        /* cas du nom, syntaxe .connect nom num */
        mot = strtok(ptr, " ");        /* recuperation du num */
        if ((mot == (char *) 0) || (!isdigit(*mot)))
            return (0);

    }
    else {                        /* cas du nom, syntaxe .connect num nom */
        mot = strtok(ptr, " ");        /* recuperation du nom */
        if ((mot == (char *) 0) || (!isalpha(*mot)))
            return (0);
    }

    return (1);
}

/*****************************************************************************/
/*                                                                           */
/*                                TRAITE_DICO_DAX                            */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Lecture de tous les noeuds et remplissage du dictionnaire                 */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
int 
traite_dico_dax(ligne)
    char            ligne[];
{
    char            bid[MAXLIG + 1], nom[MAXNAME + 1];
    register int    pas_fin_dico = 1, lg, num, i;

    while ((pas_fin_dico) && (fgets(bid, MAXLIG, infile) != (char *) 0)) {
        formater_ligne(bid);
        pas_fin_dico = strncmp(bid, FIN_DICO_DAX, strlen(FIN_DICO_DAX));
        if (pas_fin_dico) {
            lg = strlen(bid);
            num = atoi(bid + 2);
            if (num < 0)
                /*----------------------------------------------*/
                /* on se deplace jusqu'a l'espace, ou la fin de */
                /* la ligne                                     */
                /*----------------------------------------------*/
                mslError(MAUVAIS_CARDINAL);
            for (i = 3; (bid[i] != ' ') && (i < lg); ++i);
            if (i >= lg)
                mslError(MAUVAIS_DICO);

            if (strlen(bid + i + 1) > MAXNAME)
                mslError(NOM_TROP_GRAND);

            /* on commence apres l'espace */
            strcpy(nom, bid + i + 1);
            verif_nom(nom);
            maj_dic(nom, num);

        }
        ++msl_line;
    }
    if (pas_fin_dico)
        mslError(DICO_DAX_PAS_TERMINE);
}

/*****************************************************************************/
/*                                                                           */
/*                                TRAITE_DICO_CADENCE                        */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Lecture de tous les noeuds et remplissage du dictionnaire                 */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
int 
traite_dico_cadence(ligne)
    char            ligne[];
{
    register int    num, i, lg;
    char            nom[MAXNAME + 1];


    lg = strlen(ligne);
    num = atoi(ligne + 6);


    if (num < 0)
        mslError(MAUVAIS_CARDINAL);

    /* on se deplace jusqu'au '=' ou la fin de la ligne */
    for (i = 8; (ligne[i] != '=') && (i < lg); ++i);
    if (i >= lg)
        mslError(MAUVAIS_DICO);

    if (strlen(ligne + i + 2) > MAXNAME)
        mslError(NOM_TROP_GRAND);

    strcpy(nom, ligne + i + 2);        /* on commence apres le '=' et l'espace */
    verif_nom(nom);
    maj_dic(nom, num);
}

/*****************************************************************************/
/*                                                                           */
/*                                TRAITE_DICO_VTI                            */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Lecture de tous les noeuds et remplissage du dictionnaire                 */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
int 
traite_dico_vti(ligne)
    char            ligne[];
{
    char            bid[MAXLIG + 1], nom[MAXNAME + 1];
    register int    pas_fin_dico = 1, lg, num, i;

    fgets(bid, MAXLIG, infile);        /* lecture bidon de la premiere ligne */
    while ((pas_fin_dico) && (fgets(bid, MAXLIG, infile) != (char *) 0)) {
        formater_ligne(bid);
        pas_fin_dico = strcmp(bid, FIN_DICO_VTI);
        if (pas_fin_dico) {
            lg = strlen(bid);
            num = atoi(bid + 2);
            if (num < 0)
                /*-------------------------------------------*/
                /* on se deplace jusqu'a l'espace, ou la fin */
                /* de la ligne                               */
                /*-------------------------------------------*/
                mslError(MAUVAIS_CARDINAL);
            for (i = 3; (bid[i] != ' ') && (i < lg); ++i);
            if (i >= lg)
                mslError(MAUVAIS_DICO);

            if (strlen(bid + i + 1) > MAXNAME)
                mslError(NOM_TROP_GRAND);

            /* on commence apres l'espace */
            strcpy(nom, bid + i + 1);
            verif_nom(nom);
            maj_dic(nom, num);
        }
        ++msl_line;
    }
    if (pas_fin_dico)
        mslError(DICO_VTI_PAS_TERMINE);
    ++msl_line;
}

/*****************************************************************************/
/*                                                                           */
/*                                TRAITE_DICO_ELDO                           */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Lecture de tous les noeuds et remplissage du dictionnaire                 */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
int 
traite_dico_eldo(ligne)
    char            ligne[];
{
    register int    num, i, lg;
    char            nom[MAXNAME + 1];
    char           *ptr, *mot;


    ptr = ligne;
    mot = strtok(ptr, " ");        /* on passe le .connect */
    ptr = (char *) 0;

    mot = strtok(ptr, " ");        /* on a soit le nom, soit le num */

    if (isalpha(*mot)) {        /* cas du nom, syntaxe .connect nom num */
        if (strlen(mot) > MAXNAME)
            mslError(NOM_TROP_GRAND);
        strcpy(nom, mot);
        mot = strtok(ptr, " ");        /* recuperation du num */
        num = atoi(mot);
        if (num < 0)
            mslError(MAUVAIS_CARDINAL);

    }
    else {                        /* cas du nom, syntaxe .connect num nom */
        num = atoi(mot);
        if (num < 0)
            mslError(MAUVAIS_CARDINAL);
        mot = strtok(ptr, " ");        /* recuperation du nom */
        if (strlen(mot) > MAXNAME)
            mslError(NOM_TROP_GRAND);
        strcpy(nom, mot);
    }

    verif_nom(nom);
    maj_dic(nom, num);
}

/*****************************************************************************/
/*                                                                           */
/*                                 DEBUT_SUBCKT                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante commence par '.subckt'                     */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne commencant par '.subckt'                    */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
debut_subckt(ligne)
    char            ligne[];
{
    return (!strncmp(ligne, DEB_SUBCKT, strlen(DEB_SUBCKT)));
}

/*****************************************************************************/
/*                                                                           */
/*                              TRAITE_DEBUT_SUBCKT                          */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Traitement du debut de subckt                                             */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
traite_debut_subckt(ptlofig, ligne)
    lofig_list     *ptlofig;
    char            ligne[];
{
    char           *mot;
    register int    num;
    struct locon   *ptaux_locon;
    lofig_list     *ptf;


    mot = strtok(ligne, " ");        /* on passe le mot '.subckt' */
    mot = strtok(NULL, " ");        /* recuperation du nom du subckt */
    if (mot == (char *) 0)
        mslError(MANQUE_NOM_SUBCKT);
    if (!isalpha(mot[0]))
        mslError(MAUVAIS_NOM_SUBCKT_MBK);

    strcpy(msl_subckt, mot);        /* sauvegarde du nom */

    /* figure mere */
    flag_dico = msl_father = (!strcmp(msl_subckt, msl_figname));

    /* figure inconnue */
    if ((ptf = spi_findlofig(namealloc(msl_subckt))) == NULL) {
        ptmaindic = ptdic;
        ptdic = NULL;
        msl_ptfig = addlofig(msl_subckt);
        msl_figloaded = 1;
    }
    else if (!strcmp(msl_figname, msl_subckt)) {        /* figure pere */
        if (msl_ptfig != ptlofig) {
            for (ptf = msl_ptfig; ptf->NEXT != ptlofig; ptf = ptf->NEXT);
            ptf->NEXT = ptlofig->NEXT;
            ptlofig->NEXT = msl_ptfig;
            msl_ptfig = ptlofig;
            HEAD_LOFIG = ptlofig;
        }
        msl_figloaded = 1;
    }
    else {                        /* figure existante et non pere */
        ptmaindic = ptdic;
        ptdic = NULL;
        dellofig(msl_ptfig->NAME);
        addlofig(msl_subckt);
        mslWarning(FIG_CANCELLED);
        fprintf(stderr, "Name figure is %s\n", msl_subckt);
        msl_figloaded = 1;
    }

    while ((mot = strtok(NULL, " ")) != (char *) 0) {
        if (isdigit(*mot) && (atoi(mot) == 0))
            mslError(NOEUD_ZERO);
        spi_addlocname(msl_ptfig, namealloc(mot), ptdic);
    }
}

/*****************************************************************************/
/*                      fonction main_gdt()                                  */
/* dit si la ligne contient le "* start main cell %s": c'est le debut de la  */
/* description de la figure mere.                                            */
/*****************************************************************************/
short 
main_gdt(ptfig, ligne)
    lofig_list     *ptfig;
    char            ligne[];

{
    char            name[100];
    short           flag = 0;

    flag = sscanf(ligne, "* start main cell %s", name);

/* si c'est la figure principale, c'est le nom de la figure mere */
    if (flag == 1) {
        ptfig->NAME = namealloc(name);
        ptfig->NEXT = msl_ptfig;
        msl_ptfig = ptfig;
        msl_figloaded = 1;
    }

    return (flag);
}


/*****************************************************************************/
/*                                  dico_gdt()                               */
/* traite le dictionnaire format GDT dans lequel les correspondances         */
/* nom-signal sont a l'interieur du subckt. Cette fonction dit seulement si  */
/* la ligne est de ce format ou non.                                         */
/*****************************************************************************/
short 
dico_gdt(ligne)
    char            ligne[];

{
    char            aux[50], nom[100], index[50];
    long            ind;

    ind = sscanf(ligne, "*.%s %s %s", aux, nom, index);

    if (ind == 3) {
        if ((strcmp(aux, "in") == 0) || (strcmp(aux, "out") == 0) ||
            (strcmp(aux, "vdd") == 0) || (strcmp(aux, "gnd") == 0))
            if ((ind = atoi(index)) != 0) {
                losig_list     *losig = givelosig(msl_ptfig, ind);
                locon_list     *locon;
                char           *name = namealloc(spi_checkname(nom));

                losig->NAMECHAIN = addchain(losig->NAMECHAIN, name);

                /*------------------------------------------------------------*/
                /* si la variable spi_main_gdt est positionnee, c'est que les */
                /* .in, .out, etc... sont les connecteurs. Alors il faut non  */
                /* pas les chercher dans la liste des connecteurs, mais creer */
                /* les connecteurs.                                           */
                /*------------------------------------------------------------*/
                if (spi_main_gdt == 1) {
                    for (locon = msl_ptfig->LOCON;
                         (locon != NULL) && (locon->SIG != losig);
                         locon = locon->NEXT);

                    if (locon == NULL) {
                        msl_ptfig->LOCON = addlocon(msl_ptfig, name, losig, UNKNOWN);
                        losig->TYPE = 'E';
                    }
                }
                /* sinon recherche du nom de connecteur */
                else {
                    sprintf(aux, "mbk.%s", index);

                    for (locon = msl_ptfig->LOCON;
                         (locon != NULL) && (strcmp(locon->NAME, aux) != 0);
                         locon = locon->NEXT);

                    if (locon != NULL)
                        locon->NAME = name;
                }

                return (1);
            }
    }

    return (0);
}


/*****************************************************************************/
/*                                                                           */
/*                                   FIN_SUBCKT                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante commence par '.ends'                       */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne commencant par '.ends'                      */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
fin_subckt(ligne)
    char            ligne[];
{
    return (!strncmp(ligne, FIN_SUBCKT, strlen(FIN_SUBCKT)));
}

/*****************************************************************************/
/*                                                                           */
/*                              TRAITE_FIN_SUBCKT                            */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Traitement de la fin de subckt                                            */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
traite_fin_subckt(ligne)
    char            ligne[];
{
    struct locon   *ptaux_locon;

    if (strcmp(ligne + 6, msl_subckt))
        mslError(MAUVAIS_NOM_SUBCKT);

    for (ptaux_locon = msl_ptfig->LOCON;
         ptaux_locon; ptaux_locon = ptaux_locon->NEXT)
        ptaux_locon->SIG->TYPE = 'E';

    traite_dicos(msl_ptfig);
    spi_free_triplet_list(ptdic);
    ptdic = ptmaindic;
    spi_clear_hash();
    spi_index = 2;
}

/*****************************************************************************/
/*                                                                           */
/*                                   TRANSISTOR                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante commence par 'm'                           */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne commencant par 'm'                          */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
transistor(ligne)
    char            ligne[];
{
/*        return( ligne[0] == 'm' );*/
/* MODIF PAYAM POUR LECTURE DES TRANSISTORS ASGA */
    return ((ligne[0] == 'm') || (ligne[0] == 'j'));
}

/*****************************************************************************/
/*                                                                           */
/*                              TRAITE_TRANSISTOR                            */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Traitement du transistor                                                  */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
traite_transistor(ligne)
    char            ligne[];
{
    losig_list     *drain, *grid, *source;
    lotrs_list     *pttrans;
    char            l[MAXNAME + 1], w[MAXNAME + 1], ps[MAXNAME + 1], pd[MAXNAME + 1],
                    as[MAXNAME + 1], ad[MAXNAME + 1], type_trans,
                   *mot, *name;
    char           *transname;
    long            wr = 0, xs = 0, xd = 0;
    int             flw, fas, fad, fps, fpd;


    flw = fas = fad = fps = fpd = 0;


    mot = strtok(ligne, " ");        /* on passe le nom du transistor */
    if (mot == NULL)
        mslError(MAUVAISE_DEF_TRANS);
    transname = namealloc(mot);
    mot = strtok(NULL, " ");        /* 1 er identificateur = DRAIN */
    if (mot == NULL)
        mslError(MAUVAISE_DEF_TRANS_DRAIN);
    drain = spi_addsigname(mot);
    mot = strtok(NULL, " ");        /* 2 eme identificateur = GRID */
    if (mot == NULL)
        mslError(MAUVAISE_DEF_TRANS_GRID);
    grid = spi_addsigname(mot);
    mot = strtok(NULL, " ");        /* 3 eme identificateur = SOURCE */
    if (mot == NULL)
        mslError(MAUVAISE_DEF_TRANS_SOURCE);
    source = spi_addsigname(mot);
    mot = strtok(NULL, " ");        /* 4 eme identificateur = BULK */
    if (mot == NULL)
        mslError(MAUVAISE_DEF_TRANS_BULK);
    mot = strtok(NULL, " ");        /* 5 eme identificateur = TYPE */
    if (mot == NULL)
        mslError(MAUVAISE_DEF_TRANS_TYPE);
    name = namealloc(mot);

    if (!strcmp(name, modn))
        type_trans = TRANSN;
    else if (!strcmp(name, modp))
        type_trans = TRANSP;
    else
        mslError(MAUVAIS_TRANS_TYPE);

    /* parcours de param_list + geo_list */
    while ((mot = strtok(NULL, " ")) != NULL) {
        verif_unite(mot);
        switch (*mot) {
        case 'w':
            if (flw)
                mslError(MAUVAISE_DEF_TRANS_WL);
            if ((mot + 2) == NULL)
                mslError(MAUVAISE_DEF_TRANS_WL);
            strcpy(w, mot + 2);
            mot = strtok(NULL, " ");
            if ((mot == NULL) ||
                (*mot != 'l'))
                mslError(MAUVAISE_DEF_TRANS_WL);
            if ((mot + 2) == NULL)
                mslError(MAUVAISE_DEF_TRANS_WL);
            strcpy(l, mot + 2);
            flw = 1;
            break;
        case 'l':
            if (flw)
                mslError(MAUVAISE_DEF_TRANS_WL);
            if ((mot + 2) == NULL)
                mslError(MAUVAISE_DEF_TRANS_WL);
            strcpy(l, mot + 2);
            mot = strtok(NULL, " ");
            if ((mot == NULL) ||
                (*mot != 'w'))
                mslError(MAUVAISE_DEF_TRANS_WL);
            if ((mot + 2) == NULL)
                mslError(MAUVAISE_DEF_TRANS_WL);
            strcpy(w, mot + 2);
            flw = 1;
            break;
        case 'a':
            if (mot[1] == 's') {
                if (fas)
                    mslError(MAUVAISE_DEF_TRANS_ASD);
                if ((mot + 3) == NULL)
                    mslError(MAUVAISE_DEF_TRANS_ASD);
                strcpy(as, mot + 3);
                fas = 1;
            }
            else if (mot[1] == 'd') {
                if (fad)
                    mslError(MAUVAISE_DEF_TRANS_ASD);
                if ((mot + 3) == NULL)
                    mslError(MAUVAISE_DEF_TRANS_ASD);
                strcpy(ad, mot + 3);
                fad = 1;
            }
            else
                mslError(MAUVAISE_DEF_TRANS_ASD);
            break;
        case 'p':
            if (mot[1] == 's') {
                if (fps)
                    mslError(MAUVAISE_DEF_TRANS_PSD);
                if ((mot + 3) == NULL)
                    mslError(MAUVAISE_DEF_TRANS_PSD);
                strcpy(ps, mot + 3);
                fps = 1;
            }
            else if (mot[1] == 'd') {
                if (fpd)
                    mslError(MAUVAISE_DEF_TRANS_PSD);
                if ((mot + 3) == NULL)
                    mslError(MAUVAISE_DEF_TRANS_PSD);
                strcpy(pd, mot + 3);
                fpd = 1;
            }
            else
                mslError(MAUVAISE_DEF_TRANS_PSD);
            break;
        case 'n':
            if (mot[1] != 'r' ||
                (mot[2] != 's' && mot[2] != 'd')
                )
                mslError(MAUVAISE_DEF_TRANS_NRSD);
            break;
        default:
            mslError(MAUVAISE_DEF_TRANS);
        }
    }
    wr = spi_givelong(w);
    if (wr) {
        xs = spi_givearea(as, wr);
        xd = spi_givearea(ad, wr);
    }
    pttrans = addlotrs(msl_ptfig, type_trans, 0, 0, (unsigned short) spi_givelong(w), (unsigned short) spi_givelong(l), (unsigned short) spi_givelong(ps), (unsigned short) spi_givelong(pd), (unsigned short) xs, (unsigned short) xd, grid, source, drain);
    pttrans->USER = addptype(pttrans->USER, MSL_TRANSNAME, transname);
}

/*****************************************************************************/
/*                                                                           */
/*                                   CAPACITE                                */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante commence par 'c'                           */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne commencant par 'c'                          */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
capacite(ligne)
    char            ligne[];
{
    return (ligne[0] == 'c');
}

/*****************************************************************************/
/*                                                                           */
/*                              TRAITE_CAPACITE                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Traitement de la capacite                                                 */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
traite_capacite(ligne)
    char            ligne[];
{
    char           *ptr, *mot;
    double          capa;
    losig_list     *ptsigx, *ptsigy;


    ptr = ligne;
    mot = strtok(ptr, " ");        /* on passe le nom de la capacite */
    if (mot == (char *) 0)
        mslError(MAUVAISE_DEF_CAPA);
    ptr = (char *) 0;
    mot = strtok(ptr, " ");        /* 1 er identificateur */
    if (mot == (char *) 0)
        mslError(MAUVAISE_DEF_CAPA);
    /*ptsigx = givelosig(msl_ptfig, (long)spi_giveindex(mot));*/
    ptsigx = spi_addsigname(mot);
    mot = strtok(ptr, " ");        /* 2 eme identificateur */
    if (mot == (char *) 0)
        mslError(MAUVAISE_DEF_CAPA);
    /*ptsigy = givelosig(msl_ptfig, (long)spi_giveindex(mot));*/
    ptsigy = spi_addsigname(mot);
    mot = strtok(ptr, " ");        /* la valeur */
    if (mot == (char *) 0)
        mslError(MAUVAISE_DEF_CAPA);
    verif_unite(mot);
    capa = spi_givefloat_capa(mot);
    spi_addcapa(ptsigx, ptsigy, capa);
}

/*****************************************************************************/
/*                                                                           */
/*                                   INSTANCE                                */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Recherche si la ligne courante commence par 'x'                           */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : 1 si c'est une ligne commencant par 'x'                          */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
instance(ligne)
    char            ligne[];
{
    return (ligne[0] == 'x');
}

/*****************************************************************************/
/*                                                                           */
/*                              TRAITE_INSTANCE                              */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Traitement de l' instance                                                 */
/*                                                                           */
/* ENTREE : la ligne lue                                                     */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
traite_instance(ligne)
    char            ligne[];
{
    char            nom[MAXNAME + 1], inst_nom[MAXNAME + 1], *inst_name,
                   *name, *ptr, *mot;
    chain_list     *ptchain = NULL;


    ptr = ligne;
    mot = strtok(ptr, " ");        /* le nom de l'instance */

    if (mot == (char *) 0)
        mslError(MAUVAISE_DEF_INSTANCE);

    if (strlen(mot) > MAXNAME)
        mslError(NOM_TROP_GRAND);

    strcpy(nom, mot + 1);
    verif_nom(nom);
    name = namealloc(nom);
    ptr = (char *) 0;

    /* parcours de l'ident_list */
    while ((mot = strtok(ptr, " ")) != (char *) 0) {
        losig_list     *ptinstsig;

        if (strlen(mot) > MAXNAME)
            mslError(NOM_TROP_GRAND);
        /*ptinstsig = givelosig(msl_ptfig, (long)spi_giveindex(mot));*/
        ptinstsig = spi_addsigname(mot);
        ptchain = addchain(ptchain, ptinstsig);
        strcpy(inst_nom, mot);
        inst_name = namealloc(inst_nom);
    }
    verif_nom(inst_nom);
    dellosig(msl_ptfig, spi_giveindex(inst_name));
    ptchain = delchain(ptchain, ptchain);
    addloins(msl_ptfig, name, getlofig(inst_name, 'A'), ptchain);
    freechain(ptchain);
    ptchain = NULL;
}

/*****************************************************************************/
/*                                                                           */
/*                              VERIF_NOM                                    */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Regarde si la premier caractere est une lettre                            */
/*                                                                           */
/* ENTREE : le mot a traiter                                                 */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
verif_nom(nom)
    char            nom[];
{
    char           *name_clean;

    name_clean = cleaname(nom);
    if (!isalpha(*name_clean))
        mslError(MAUVAIS_NOM_MBK);
}

/*****************************************************************************/
/*                                                                           */
/*                              VERIF_UNITE                                  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Regarde si la premiere lettre de l'unite est : m, u, n, p, ou f           */
/*                                                                           */
/* ENTREE : le mot a traiter                                                 */
/* RETOUR : aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
verif_unite(mot)
    char           *mot;
{
    char           *pt;

    pt = mot + strlen(mot) - 1;        /* pt pointe sur le dernier caractere */

    if (isdigit(*pt))                /* pas d'unite */
        return;

    /* on recule tant qu'il y a une lettre */
    while ((isalpha(*pt)) && (pt != mot))
        --pt;

    /* si il n'y avait pas de chiffre */
    if ((pt == mot) && (!isdigit(*pt)))
        mslError(MAUVAISE_VALEUR);

    ++pt;                        /* on se repositionne sur la premiere lettre */
    if ((*pt != 'm') && (*pt != 'u') && (*pt != 'p') && (*pt != 'f'))
        mslError(MAUVAISE_UNITE);
}

/*****************************************************************************/
/*                                                                           */
/*                                    IS_NUM                                 */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Regarde si le mot passe est un nombre                                     */
/*                                                                           */
/* ENTREE : le mot a traiter                                                 */
/* RETOUR : 1 si le mot est pas un nombre                                    */
/*          0 sinon                                                          */
/*                                                                           */
/*****************************************************************************/
int 
is_num(mot)
    char           *mot;
{
    while (*mot != '\0') {
        if (!isdigit(*mot))
            return (0);
        ++mot;
    }
    return (1);
}

/*****************************************************************************/
/*                                                                           */
/*                                     MAJ_DIC                               */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Met a jour la liste chainee du dictionnaire                               */
/*                                                                           */
/* ENTREE : Le nom et le numero du connecteur                                */
/* RETOUR : Aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
maj_dic(nom, num)
    char            nom[];
    int             num;
{
    spi_triplet_list *ptaux_dic = ptdic;

    ptdic = spi_create_triplet(nom, (num - MAXSIGINDEX));
    ptdic->NEXT = ptaux_dic;
}

/*****************************************************************************/
/*                                                                           */
/*                                   TRAITE_DICOS                            */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/* Parcours tous les connecteurs et signaux de la figure mere,               */
/* si le nom de signal commence par mbk, recuperation du numero qui suit     */
/* recherche dans le dico du bon nom                                         */
/*                                                                           */
/* ENTREE : Aucune                                                           */
/* RETOUR : Aucun                                                            */
/*                                                                           */
/*****************************************************************************/
void 
traite_dicos(ptfig)
    lofig_list     *ptfig;
{
    locon_list     *ptaux_locon;
    losig_list     *ptaux_losig;
    register int    num;
    char           *str;

    /* parcours de tous les connecteurs */


    for (ptaux_locon = ptfig->LOCON; ptaux_locon;
         ptaux_locon = ptaux_locon->NEXT) {
        if (!strncmp(ptaux_locon->NAME, "mbk", 3)) {        /* mbk trouve */
            num = atoi(ptaux_locon->NAME + 4);
            str = spi_getdicname(num, ptdic);
            if (str)
                ptaux_locon->NAME = str;
        }
        else {
            num = spi_getdicindex(ptaux_locon->NAME, ptdic);
            if (num != 0 && num < MAXSIGINDEX)
                ptaux_locon->SIG->INDEX = num;
        }
    }
    /* parcours des signaux */

    for (ptaux_losig = ptfig->LOSIG; ptaux_losig != NULL;
         ptaux_losig = ptaux_losig->NEXT) {
        str = spi_getdicname(ptaux_losig->INDEX, ptdic);
        if (str)
            ptaux_losig->NAMECHAIN = addchain(ptaux_losig->NAMECHAIN, (void *) str);
    }

}
