/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/************************************************************************************/

/*                                                                                  */
/*            Parser SPICE -> MBK : TRICHES Laurent le 20/08/92                     */
/*                                                                                  */
/************************************************************************************/

#include "msl_headers.h"

FILE           *infile;
char           *msl_figname;
char            msl_subckt[MAXLIG + 1];
char            modn[20] = "tn";
char            modp[20] = "tp";
int             msl_line;
int             flag_dico;
int             msl_father = 0;
int             msl_figloaded = 0;
int             spi_index = 2;
int             spi_main_gdt = 0;

int             MAXSIGINDEX = 1000000;

lofig_list     *msl_ptfig = NULL;

spi_triplet_list *ptdic = NULL;
spi_triplet_list *ptmaindic = NULL;

static int (*ext_handler)();
static int (*cmp_handler)();

void 
spiceloadlofig(ptgif, name, mode)
    lofig_list     *ptgif;
    char           *name;
    char            mode;
{
    char            tmp_name[80];
    char            tmp_mv[80];
    char            file_name[80];
    char            ligne[MAXLIG + 1];
    char           *str;
    short           dicogdt = 0;
    register int    dans_subckt = 0, go_dico = 0;
    register int    eldo, cadence, vti, dax;
    losig_list     *ptsig;

/* position de l'environnement */
    spigetenv();

    if (!(infile = mbkfopen(name, "spi", READ_TEXT))) {
        fprintf(stderr, "*** mbk error *** can't open file %s\n", name);
        EXIT(-1);
    }

    msl_figname = name;
    spi_init_hash();
    msl_ptfig = ptgif;

    msl_line = 1;
    eldo = cadence = vti = dax = flag_dico = 0;

    while (fgets(ligne, MAXLIG, infile) != (char *) 0) {
        formater_ligne(ligne);
        trouver_plus(ligne);


        if (dico_cadence(ligne)) {
            go_dico = flag_dico;
            if (dax | vti | eldo)
                mslError(DEUX_DICOS);
            traite_dico_cadence(ligne);
            cadence = 1;
        }
        else if (dico_dax(ligne)) {
            go_dico = flag_dico;
            if (cadence | vti | eldo)
                mslError(DEUX_DICOS);
            traite_dico_dax(ligne);
            dax = 1;
        }
        else if (dico_vti(ligne)) {
            go_dico = flag_dico;
            if (dax | cadence | eldo)
                mslError(DEUX_DICOS);
            traite_dico_vti(ligne);
            vti = 1;
        }
        else if (dico_eldo(ligne)) {
            go_dico = flag_dico;
            if (dax | vti | cadence)
                mslError(DEUX_DICOS);
            traite_dico_eldo(ligne);
            eldo = 1;
        }
        else
            /*---------------------------------------------------*/
            /* particularite du format GDT: le dico est a LA FIN */
            /* de chaque subckt.                                 */
            /*---------------------------------------------------*/
        if (dico_gdt(ligne)) {
            dicogdt = 1;
        }
        else if (debut_subckt(ligne)) {
            if (spi_main_gdt == 1) {
                spi_main_gdt = 0;
                spi_clear_hash();
                spi_index = 2;
            }

            if (dans_subckt)
                mslError(MANQUE_FIN_SUBCKT);
            traite_debut_subckt(ptgif, ligne);
            dans_subckt = 1;
        }
        else if (transistor(ligne)) {
            if ((dans_subckt) || (spi_main_gdt)) {
                if (msl_figloaded) {
                    enleve_blancs_egal(ligne);
                    traite_transistor(ligne);
                }
            }
            else
                mslError(TRANSISTOR_EXT_SUBCKT);
        }
        else if (capacite(ligne)) {
            if (dans_subckt || spi_main_gdt)
                if (msl_figloaded)
                    traite_capacite(ligne);
                else
                    mslError(CAPACITE_EXT_SUBCKT);
        }
        else if (instance(ligne)) {
            if (dans_subckt || spi_main_gdt)
                if (msl_figloaded)
                    traite_instance(ligne);
                else
                    mslError(INSTANCE_EXT_SUBCKT);
        }
        else if (fin_subckt(ligne)) {
            if (!dans_subckt)
                mslError(MANQUE_DEBUT_SUBCKT);
            traite_fin_subckt(ligne);
            dans_subckt = 0;
        }
        else if (!dans_subckt) {
            if (main_gdt(ptgif, ligne) != 0) {
                spi_main_gdt = 1;
            }
        }
        else if (ext_handler != NULL) {
            ext_handler(ligne);
        }
        ++msl_line;
    }

    fclose(infile);

    spi_clear_hash();
    spi_index = 2;

    /* parcours du dico pour les signaux */

    if (ptdic != NULL)
        traite_dicos(ptgif);

    if (cmp_handler != NULL)
        cmp_handler();
}

void
mslAddExtension(mslExtHandler)
    int (*mslExtHandler)();
{
    ext_handler = mslExtHandler;
}

void
mslRmvExtension()
{
    ext_handler = NULL;
}

void
mslAddCompletion(mslCmpHandler)
    int (*mslCmpHandler)();
{
    cmp_handler = mslCmpHandler;
}

void
mslRmvCompletion()
{
    cmp_handler = NULL;
}

