/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include "msl_headers.h"

spi_triplet_list *hashtable[HASH_MAX];
spi_triplet_list *HEAD_TRIPLET = NULL;
chain_list       *ALLOC_TRIPLET = NULL;

/* lecture des variables d'environnement: AJOUT PAYAM du 17/07/94*/
/* cette fonction a ete enlevee de spi_prs.c (loadspicefig()) ou */
/* un truc comme ca. Le parser et le driver y font appel.        */
/* POUR LES VARIABLES D'ENVIRONNEMENT MBK_SPI_TN ET MBK_SPI_TP   */
/* SI C'EST DE LA FORME "ASGA TOTO", ALORS LE DRIVER CRACHE DE   */
/* L'ASGA (DEVICE J) AVEC COMME NOM DE MODEL TOTO. SINON, IL     */
/* CRACHE DU MOS (DEVICE M) COMME NOM DE MODEL TOTO. LA          */
/* VERIFICATION D'HOMOGENEITE ENTRE P ET N N'EST PAS FAITE ET    */
/* LE P QUI L'EMPORTE: SI SPI_TN EST A "TOTO" ALORS QUE SPI_TP   */
/* EST A "ASGA TITI", ALORS LE DRIVER CRACHE DE L'ASGA AVEC      */
/* MODELES TOTO ET TITI. MAIS L'INVERSE FAIT CRACHER DU MOS, AVEC*/
/* TOUJOURS LES MODELES TOTO ET TITI.                            */
int 
spigetenv()
{
    char           *str = getenv("MBK_SPI_TN");

    SPITRANS = SPIMOS;

    if (str != NULL) {
        /* determination si ASGA ou MOS */
        int             i = 0;

        str = namealloc(str);        /* juste pour avoir des caracteres en minuscule. */
        /* tolower() a mon avis n'est pas portable. Et   */
        /* je sais qu'elle a deja ete utilisee dans cette*/
        /* librairie!!!                                  */

        for (; (i != strlen(str)) && (str[i] != ' ') && (str[i] != '\t'); i++);
        if (i == strlen(str))
            strcpy(modn, str);
        else {
            if (!strncmp(str, "asga", 4))
                SPITRANS = SPIASGA;

            for (; (i != strlen(str)) && ((str[i] == ' ') || (str[i] == '\t'));
                 i++);

            if (i != strlen(str))
                strcpy(modn, str + i);
        }
    }

    str = getenv("MBK_SPI_TP");
    if (str != NULL) {
        /* determination si ASGA ou MOS */
        int             i = 0;

        str = namealloc(str);        /* juste pour avoir des caracteres en minuscule. */
        /* tolower() a mon avis n'est pas portable.      */

        for (; (i != strlen(str)) && (str[i] != ' ') && (str[i] != '\t'); i++);
        if (i == strlen(str))
            strcpy(modp, str);
        else {
            if (!strncmp(str, "asga", 4))
                SPITRANS = SPIASGA;

            for (; (i != strlen(str)) && ((str[i] == ' ') || (str[i] == '\t'));
                 i++);

            if (i != strlen(str))
                strcpy(modp, str + i);
        }
    }

    str = getenv("MBK_SPI_MAXSIGINDEX");
    if (str != NULL)
        MAXSIGINDEX = atoi(str);
}

/* emulation de la fonction mathematique pow(x,y) */

double 
spi_modpow(x, y)
    double          x;
    int             y;
{
    int             i;
    double          val;

    val = x;
    if (y > 0) {
        for (i = 1; i < y; i++) {
            val *= x;
        }
        return (val);
    }
    else {
        y = -y;
        for (i = 1; i < y; i++) {
            val *= x;
        }
        val = 1.0 / val;
        return (val);
    }
}


/*la declaration de la fonction de spi_givefloat_capa */

double 
spi_givefloat_capa(s)
    char           *s;
{
    double          x = 0.0;
    char            ch[10];

    strcpy(ch, "?");
    if (s != (char *) NULL) {
        sscanf(s, "%lf %s", &x, ch);
        if (strncmp(ch, "m", 1) == 0 || strncmp(ch, "M", 1) == 0)
            return (spi_modpow(10.0, 9) * x);
        else if (strncmp(ch, "u", 1) == 0 || strncmp(ch, "U", 1) == 0)
            return (spi_modpow(10.0, 6) * x);
        else if (strncmp(ch, "n", 1) == 0 || strncmp(ch, "N", 1) == 0)
            return (spi_modpow(10.0, 3) * x);
        else if (strncmp(ch, "p", 1) == 0 || strncmp(ch, "P", 1) == 0)
            return (x);
        else if (strncmp(ch, "f", 1) == 0 || strncmp(ch, "F", 1) == 0)
            return (spi_modpow(10.0, -3) * x);
        else if (strncmp(ch, "?", 1) == 0)
            return (spi_modpow(10.0, 12) * x);
    }
}

/*la declaration de la fonction de spi_givelong */

long 
spi_givelong(s)
    char           *s;
{
    double          x = 0.0;
    char            ch[10];

    strcpy(ch, "?");
    if (s != (char *) NULL) {
        sscanf(s, "%lf %s", &x, ch);
        if (strncmp(ch, "m", 1) == 0 || strncmp(ch, "M", 1) == 0)
            return ((long) (1000.0 * x * SCALE_X));
        else if (strncmp(ch, "u", 1) == 0 || strncmp(ch, "U", 1) == 0)
            return ((long) (x * SCALE_X));
        else if (strncmp(ch, "n", 1) == 0 || strncmp(ch, "N", 1) == 0)
            return ((long) (0.0010 * x * SCALE_X));
        else if (strncmp(ch, "p", 1) == 0 || strncmp(ch, "P", 1) == 0)
            return ((long) (0.0000010 * x * SCALE_X));
        else if (strncmp(ch, "f", 1) == 0 || strncmp(ch, "F", 1) == 0)
            return ((long) (0.0000000010 * x * SCALE_X));
        else if (strncmp(ch, "?", 1) == 0)
            return ((long) (1000000.0 * x * SCALE_X));
    }
}

/*la declaration de la fonction de spi_givelarea */

long 
spi_givearea(s, wr)
    char           *s;
    long            wr;
{
    double          x = 0.0;
    char            ch[10];

    strcpy(ch, "?");
    if (s != (char *) NULL && wr != 0) {
        sscanf(s, "%lf %s", &x, ch);
        if (strncmp(ch, "m", 1) == 0 || strncmp(ch, "M", 1) == 0)
            return ((long) ((x * SCALE_X * SCALE_X) * (1000000000.0 / wr)));
        else if (strncmp(ch, "u", 1) == 0 || strncmp(ch, "U", 1) == 0)
            return ((long) ((x * SCALE_X * SCALE_X) * (1000000.0 / wr)));
        else if (strncmp(ch, "n", 1) == 0 || strncmp(ch, "N", 1) == 0)
            return ((long) ((x * SCALE_X * SCALE_X) * (1000.0 / wr)));
        else if (strncmp(ch, "p", 1) == 0 || strncmp(ch, "P", 1) == 0)
            return ((long) ((x * SCALE_X * SCALE_X) / wr));
        else if (strncmp(ch, "f", 1) == 0 || strncmp(ch, "F", 1) == 0)
            return ((long) (0.0010 * (x * SCALE_X * SCALE_X) / wr));
        else if (strncmp(ch, "?", 1) == 0)
            return ((long) ((x * SCALE_X * SCALE_X) * (spi_modpow(10.0, 12) / wr)));
    }
}

/*allocation */

spi_triplet_list *
spi_create_triplet(name, index)
    char           *name;
    int             index;
{
    spi_triplet_list *pt = NULL;
    register int i;

    if (HEAD_TRIPLET == NULL) {
        pt = (spi_triplet_list *)mbkalloc(BUFSIZE*sizeof(spi_triplet_list));
        ALLOC_TRIPLET = addchain(ALLOC_TRIPLET, pt);
        HEAD_TRIPLET = pt;
        for (i = 1; i < BUFSIZE; i++) {
             pt->NEXT = pt + 1;
             pt++;
        }
        pt->NEXT = NULL;
    }
    pt = HEAD_TRIPLET;
    HEAD_TRIPLET = pt->NEXT;

    name = namealloc(spi_checkname(name));
    pt->NEXT = NULL;
    pt->NAME = name;
    pt->INDEX = index + MAXSIGINDEX;
    return (pt);
}

void
spi_free_triplet_list(pthead)
    spi_triplet_list *pthead;
{
    spi_triplet_list *pt;

    if (pthead == NULL) return;
    for (pt = pthead; pt->NEXT; pt = pt->NEXT);
    pt->NEXT = HEAD_TRIPLET;
    HEAD_TRIPLET = pthead;
}

void
destroy_triplets()
{
    chain_list      *pt;

    for (pt = ALLOC_TRIPLET; pt; pt = pt->NEXT) {
        mbkfree(pt->DATA);
    }
    freechain(ALLOC_TRIPLET);
    ALLOC_TRIPLET = NULL;
}

/*la fonction de la gestion de la table de hachage */

int 
spi_getcode(s)
    char           *s;
{
    int             code = 0;

    while (*s) {
        code = code << 1 ^ *s;
        s++;
    }
    if (code < 0)
        code = -code;
    code = (code % HASH_MAX);
    return (code);
}

spi_triplet_list *
spi_append2(pt1, pt2)
    spi_triplet_list *pt1, *pt2;
{
    if (pt2 == NULL)
        fprintf(stderr, " **********no memory space********** \n");
    if (pt1 == NULL)
        pt1 = pt2;
    else {
        pt2->NEXT = pt1;
        pt1 = pt2;
    }
    return (pt1);
}

/* initialisation de la table de hashage */

void
spi_init_hash()
{
    int             i;

    for (i = 0; i < HASH_MAX; i++)
        hashtable[i] = NULL;
}

/* ajout d'un element dans la table */

spi_triplet_list *
spi_addtriplet(s, i, n)
    char           *s;
    int             i, n;
{
    spi_triplet_list *pt_aux = (spi_triplet_list *) NULL;

    pt_aux = spi_append2(hashtable[i], spi_create_triplet(s, n));
    return (pt_aux);
}

/* la fonction scan() */

spi_triplet_list *
spi_scan_list2(s, i)
    char           *s;
    int             i;
{
    spi_triplet_list *pt_aux = (spi_triplet_list *) NULL;        /* pointeur signal */

    for (pt_aux = hashtable[i]; pt_aux != NULL; pt_aux = pt_aux->NEXT) {
        downstr(s, s);
        if (!strcmp(pt_aux->NAME, s))
            return (pt_aux);
    }
    return ((spi_triplet_list *) NULL);
}

/* fonction de (long)spi_giveindex */

int 
spi_giveindex(s)
    char           *s;
{
    int             i = 0;
    spi_triplet_list *pt = (spi_triplet_list *) NULL;

    s = spi_checkname(s);
    i = spi_getcode(s);
    if (hashtable[i] == NULL) {
        hashtable[i] = spi_addtriplet(s, i, spi_index);

        if (hashtable[i] != NULL) {
            spi_index++;
            return (hashtable[i]->INDEX);
        }
        else
            return (-1);
    }
    else if ((pt = spi_scan_list2(s, i)) != NULL)
        return (pt->INDEX);
    else {
        hashtable[i] = spi_addtriplet(s, i, spi_index);

        if (hashtable[i] != NULL) {
            spi_index++;
            return (hashtable[i]->INDEX);
        }
        else
            return (-1);
    }
}

/* nettoyage de la table de hashage */

void 
spi_clear_hash()
{
    int             i;
    spi_triplet_list *pt_aux = NULL, *pt_aux1 = NULL;

    for (i = 0; i < HASH_MAX; i++) {
        if (hashtable[i] != NULL)
            for (pt_aux = hashtable[i]; pt_aux != NULL; pt_aux = pt_aux1) {
                pt_aux1 = pt_aux->NEXT;
                hashtable[i] = pt_aux1;
                pt_aux->NEXT = HEAD_TRIPLET;
                HEAD_TRIPLET = pt_aux;
            }
    }
}

lofig_list     *
spi_findlofig(name)
    char           *name;
{
    lofig_list     *pt = (lofig_list *) NULL;

    /* scan figure list */
    for (pt = HEAD_LOFIG; pt; pt = pt->NEXT)
        if (!strcmp(pt->NAME, name))
            return pt;
    return (lofig_list *) NULL;
}


/*******************************************************************************
* checkname : check for busses with single index for parsers
*
*******************************************************************************/
char           *
spi_checkname(name)
    char           *name;
{
    char           *s, *t;

    s = t = name;
    while (*t) {
        if (*t == '[')
            *t = ' ';
        else if (*t == ']')
            if (*(++t) == '\0')        /* ok, it's finished */
                goto end;
            else if (*t == '[')        /* multiple array */
                continue;
        *s++ = *t++;
    }
  end:
    *s = '\0';
    return name;
}

/*******************************************************************************
* function busname for drivers
*
*******************************************************************************/
char           *
spi_busname(name)
    char           *name;
{
    char            buffer[255], *s, *t;
    char            one = 1;

    s = name;
    t = buffer;
    while (*s) {
        if (*s == ' ')
            if (one) {
                *t++ = '[';
                s++;
                one = 0;
            }
            else {
                *t++ = ']';
                *t++ = '[';
                s++;
            }
        *t++ = *s++;
    }
    if (!one)
        *t++ = ']';
    *t = '\0';
    return namealloc(buffer);
}


int 
spi_lit_string(pfic, repere)
    FILE           *pfic;
    char           *repere;
{
    char            chaine[LINESIZE];        /* recherche d'un repere     */

    if (!feof(pfic))
        do {
            fscanf(pfic, "%s", &chaine[0]);
        } while ((!feof(pfic)) && (strcmp(chaine, repere)));
    return (strcmp(chaine, repere) == 0);
}

int 
spi_lit_line(pfic, repere)
    FILE           *pfic;
    char           *repere;
{
    char            chaine[LINESIZE];        /* recherche d'un repere     */

    if (!feof(pfic))
        do {
            fgets(chaine, strlen(repere) + 1, pfic);
        } while ((!feof(pfic)) && (strcmp(chaine, repere)));
    return (strcmp(chaine, repere) == 0);
}

char           *
cleaname(str)
    char           *str;
{
    char           *name;
    int             i;

    name = namealloc(str);
    if (str[0] == '/') {
        for (i = 1; i <= strlen(str) + 1; i++)
            name[i - 1] = str[i];
    }
    if (name[strlen(name) - 1] == '!')
        name[strlen(name) - 1] = '\0';
    return (name);
}

void 
spi_viewtrip(pt)
    spi_triplet_list *pt;
{
    spi_triplet_list *pt_aux = (spi_triplet_list *) NULL;

    for (pt_aux = pt; pt_aux != NULL; pt_aux = pt_aux->NEXT)
        fprintf(stdout, "* net %d = %s\n", pt_aux->INDEX, pt_aux->NAME);
}

char           *
spi_getdicname(index, ptdic)
    int             index;
    spi_triplet_list *ptdic;
{
    spi_triplet_list *pt_aux = (spi_triplet_list *) NULL;

    for (pt_aux = ptdic; pt_aux != NULL; pt_aux = pt_aux->NEXT) {
        if (pt_aux->INDEX == index)
            return (pt_aux->NAME);
    }
    return ((char *) NULL);
}

int 
spi_getdicindex(str, ptdic)
    char           *str;
    spi_triplet_list *ptdic;
{
    spi_triplet_list *pt_aux = (spi_triplet_list *) NULL;

    if (str) {
        for (pt_aux = ptdic; pt_aux != NULL; pt_aux = pt_aux->NEXT) {
            if (!strcmp(pt_aux->NAME, str))
                return (pt_aux->INDEX);
        }
    }
    return (0);
}


losig_list     *
spi_addsigname(signame)
    char           *signame;
{
    losig_list     *ptsig;
    char           *str;


    if (isalpha(signame[0])) {
        ptsig = givelosig(msl_ptfig, (long) spi_giveindex(signame));
        if (ptsig->NAMECHAIN == NULL) {
            ptsig->NAMECHAIN = addchain(ptsig->NAMECHAIN, namealloc(signame));
        }
    }
    else {
        if (atoi(signame) > MAXSIGINDEX) {
            fprintf(stderr, "*** parser spice error *****\n");
            fprintf(stderr, "line %d : index too big, you must change MAXSIGINDEX\n");
            exit(-1);
        }
        ptsig = givelosig(msl_ptfig, atoi(signame));
        /* si aucun nom n'est donne on n'en donne pas */
    }
    return (ptsig);
}


losig_list     *
spi_givelosig(ptfig, index)
    lofig_list     *ptfig;
    long            index;
{
    losig_list     *ptsig = (losig_list *) NULL;

    for (ptsig = ptfig->LOSIG; ptsig != NULL; ptsig = ptsig->NEXT)
        if (ptsig->INDEX == index)
            return ptsig;
    return addlosig(ptfig, index, (chain_list *) NULL, 'I', 0.0);
}

void 
spi_addcapa(ptsigx, ptsigy, capa)
    losig_list     *ptsigx, *ptsigy;
    double          capa;
{
    if (ptsigx != (losig_list *) NULL)
        addcapa(ptsigx, capa);
    if (ptsigy != (losig_list *) NULL)
        addcapa(ptsigy, capa);
}

void 
spi_addlocname(ptfig, locname, ptdic)
    lofig_list     *ptfig;
    spi_triplet_list *ptdic;
    char           *locname;
{
    char           *str;

    if (isalpha(locname[0])) {
        locon_list     *locon = addlocon(ptfig, locname, givelosig(ptfig, (long) spi_giveindex(locname)), UNKNOWN);

        /* ajout PAYAM du 20/06/93 pour ne pas perdre les noms des connecteurs */
        locon->SIG->NAMECHAIN = addchain(locon->SIG->NAMECHAIN, locon->NAME);
    }
    else {
        if (atoi(locname) > MAXSIGINDEX) {
            fprintf(stderr, "*** parser spice error *****\n");
            fprintf(stderr, "line %d : index too big, you must change MAXSIGINDEX\n");
            exit(-1);
        }
        addlocon(ptfig, concatname("mbk", locname), givelosig(ptfig, atoi(locname)), UNKNOWN);

    }
}


void 
spi_drive_subckt(lofig)
    lofig_list     *lofig;
{
    locon_list     *ptlocon = (locon_list *) NULL;
    lotrs_list     *ptlotrs = (lotrs_list *) NULL;
    losig_list     *ptlosig = (losig_list *) NULL;
    losig_list     *pt_aux = (losig_list *) NULL;
    loins_list     *ptloins = (loins_list *) NULL;
    int             countt, countc;
    char            drived_cell();

    int             vdd = 1, vss = 0, loci = 0, flagvdd = 0, flagvss = 0,
                    maxindex = 0;

    if (drived_cell(already_driven, lofig->NAME))
        return;

    already_driven = addchain(already_driven, namealloc(lofig->NAME));

    fprintf(outfile, "\n.SUBCKT %s ", lofig->NAME);
    lofig->LOCON = (locon_list *) reverse((chain_list *) lofig->LOCON);
    for (ptlocon = lofig->LOCON; ptlocon != NULL; ptlocon = ptlocon->NEXT) {
        if (isvdd(ptlocon->NAME)) {
            vdd = ptlocon->SIG->INDEX;
            flagvdd = 1;
        }
        if (isvss(ptlocon->NAME)) {
            vss = ptlocon->SIG->INDEX;
            flagvss = 1;
        }

        fprintf(outfile, "%d ", ptlocon->SIG->INDEX);
        loci++;

        /*-----------------------------------------------------*/
        /* correction PAYAM du 03/06/1993: avant: si loci = 20 */
        /* on ne faisait que rajouter '+' en oubliant en oubli-*/
        /* le nom du 20-eme connecteur.                        */
        /*-----------------------------------------------------*/
        if ((loci % 20) == 0)
            fprintf(outfile, "\n+");
    }
    /* Recherche des signaux vdd et vss */
    for (pt_aux = lofig->LOSIG; pt_aux != NULL; pt_aux = pt_aux->NEXT) {
        if (pt_aux->INDEX > maxindex)
            maxindex = pt_aux->INDEX;
        if (isvdd(getsigname(pt_aux)))
            if (flagvdd != 1) {
                vdd = pt_aux->INDEX;
                flagvdd = 1;
            }
        if (isvss(getsigname(pt_aux)))
            if (flagvss != 1) {
                vss = pt_aux->INDEX;
                flagvss = 1;
            }
    }
    if (flagvdd != 1) {
        vdd = maxindex + 1;
        fprintf(stdout, "From parser SPICE\n");
        fprintf(stdout, " WARNING : No VDD Connector or Signal in %s\n", lofig->NAME);
    }
    if (flagvss != 1) {
        vss = maxindex + 2;
        fprintf(stdout, "From parser SPICE\n");
        fprintf(stdout, " WARNING : No VSS Connector or Signal in %s\n", lofig->NAME);
    }
    lofig->LOCON = (locon_list *) reverse((chain_list *) lofig->LOCON);
    fprintf(outfile, "\n");
    lofig->LOTRS = (lotrs_list *) reverse((chain_list *) lofig->LOTRS);
    for (ptlotrs = lofig->LOTRS, countt = 1; ptlotrs != NULL; ptlotrs = ptlotrs->NEXT) {
        if (SPITRANS == SPIASGA)
            fprintf(outfile, "J%d ", countt++);
        else
            fprintf(outfile, "M%d ", countt++);

        if (ptlotrs->DRAIN->SIG != NULL)
            fprintf(outfile, "%d ", (ptlotrs->DRAIN->SIG->INDEX));
        else {
            fprintf(stderr, "******* from driver spice ********\n");
            fprintf(stderr, "signal with no index in figure %s\n", lofig->NAME);
            exit(1);
        }

        if (ptlotrs->GRID->SIG != NULL)
            fprintf(outfile, "%d ", (ptlotrs->GRID->SIG->INDEX));
        else {
            fprintf(stderr, "******* from driver spice ********\n");
            fprintf(stderr, "signal with no index in figure %s\n", lofig->NAME);
            exit(1);
        }

        if (ptlotrs->SOURCE->SIG != NULL)
            fprintf(outfile, "%d ", (ptlotrs->SOURCE->SIG->INDEX));
        else {
            fprintf(stderr, "******* from driver spice ********\n");
            fprintf(stderr, "signal with no index in figure %s\n", lofig->NAME);
            exit(1);
        }



        fprintf(outfile, "%d %s L=%gU W=%gU",
                ptlotrs->TYPE == TRANSN ? vss : vdd,
                ptlotrs->TYPE == TRANSN ? modn : modp,
                (float) (ptlotrs->LENGTH / (float) SCALE_X),
                (float) (ptlotrs->WIDTH / (float) SCALE_X));

        /* geometrie du transistor: seulement pour MOS */
        if (SPITRANS != SPIASGA) {
            /* ecriture de la geometrie du transistor: surface source */
            if (ptlotrs->XS != 0)
                fprintf(outfile, " AS=%gP",
                        (float) (ptlotrs->XS * ptlotrs->WIDTH) / (float) (SCALE_X * SCALE_X));

            /* ecriture de la geometrie du transistor: surface drain */
            if (ptlotrs->XD != 0)
                fprintf(outfile, " AD=%gP",
                        (float) (ptlotrs->XD * ptlotrs->WIDTH) / (float) (SCALE_X * SCALE_X));

            /* ecriture de la geometrie du transistor: perimetre source */
            if (ptlotrs->PS != 0)
                fprintf(outfile, " PS=%gU", (float) (ptlotrs->PS / (float) SCALE_X));

            /* ecriture de la geometrie du transistor: perimetre drain */
            if (ptlotrs->PD != 0)
                fprintf(outfile, " PD=%gU", (float) (ptlotrs->PD / (float) SCALE_X));
        }

        fprintf(outfile, "\n");
    }
    lofig->LOTRS = (lotrs_list *) reverse((chain_list *) lofig->LOTRS);
    fprintf(outfile, "\n");
    lofig->LOSIG = (losig_list *) reverse((chain_list *) lofig->LOSIG);
    for (ptlosig = lofig->LOSIG, countc = 1; ptlosig != NULL; ptlosig = ptlosig->NEXT) {
        /* Laurent TRICHES le 21/08/92 */
        if ((ptlosig->CAPA != 0.0000) && (ptlosig->INDEX != vss)) {
            fprintf(outfile, "C%d %d  %d  %.4fPF\n", countc++, (ptlosig->INDEX), vss, (ptlosig->CAPA));
        }
    }
    lofig->LOSIG = (losig_list *) reverse((chain_list *) lofig->LOSIG);
    fprintf(outfile, "\n");
    lofig->LOINS = (loins_list *) reverse((chain_list *) lofig->LOINS);
    for (ptloins = lofig->LOINS; ptloins != NULL; ptloins = ptloins->NEXT) {
        ptloins->LOCON = (locon_list *) reverse((chain_list *) ptloins->LOCON);
        /* TRICHES laurent rajout le 21/08/92 */
        if (ptloins->INSNAME[0] != 'x')
            fprintf(outfile, "X%s ", spi_busname(ptloins->INSNAME));
        else
            fprintf(outfile, "%s ", spi_busname(ptloins->INSNAME));
        for (ptlocon = ptloins->LOCON; ptlocon != NULL; ptlocon = ptlocon->NEXT) {
            /*fprintf(outfile," %s",spi_busname(ptlocon->NAME));*/
            fprintf(outfile, " %d", ptlocon->SIG->INDEX);
        }
        fprintf(outfile, " %s", ptloins->FIGNAME);
        fprintf(outfile, "\n");
    }
    fprintf(outfile, ".ENDS %s\n", lofig->NAME);
    lofig->LOINS = (loins_list *) reverse((chain_list *) lofig->LOINS);
}

/*------------------------------------------------------------------------------*/
/*      Driving of basic cells (leaves) available in figure "ptlofig".
 */
/*------------------------------------------------------------------------------*/

void 
drive_basic_cells(ptlofig)
    lofig_list     *ptlofig;
{
    chain_list     *ptchain = (chain_list *) NULL;
    lofig_list     *ptlfig = (lofig_list *) NULL;
    void            spi_drive_subckt();

    for (ptchain = ptlofig->MODELCHAIN; ptchain; ptchain = ptchain->NEXT) {
        ptlfig = getlofig((char *) ptchain->DATA, 'A');
        if (ptlfig->MODELCHAIN == (chain_list *) NULL)
            spi_drive_subckt(ptlfig);
    }
}


 /*------------------------------------------------------------------------------*/
/*              Returns 0 if a cell has not already been written.
 */
/*------------------------------------------------------------------------------*/

char 
drived_cell(ptchain, cell_name)
    chain_list     *ptchain;
    char           *cell_name;
{
    chain_list     *chainp = (chain_list *) NULL;

    for (chainp = ptchain; chainp; chainp = chainp->NEXT)
        if ((char *) chainp->DATA == cell_name)
            return 1;

    return 0;
}
