/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*******************************************************************************
* rcn     : constantes, externs, and data-structures                           *
*                                                                              *
* version : 1.02                                                               *
* date    : 09/05/96                                                           *
*******************************************************************************/
#ifndef RCN
#define RCN

/* network ptype */
#define RCN_RCNET             ((long) 0x10000000)

/* wire ptype */
#define RCN_WIRE_UNKNOW       ((long) 0x00000001)
#define RCN_WIRE_POLY         ((long) 0x00000002)
#define RCN_WIRE_ALU1         ((long) 0x00000003)
#define RCN_WIRE_ALU2         ((long) 0x00000004)
#define RCN_WIRE_CONT_POLY    ((long) 0x00000005)
#define RCN_WIRE_CONT_DIF_N   ((long) 0x00000006)
#define RCN_WIRE_CONT_DIF_P   ((long) 0x00000007)
#define RCN_WIRE_CONT_VIA     ((long) 0x00000008)
#define RCN_WIRE_CONT_VIA2    ((long) 0x00000009)
#define RCN_WIRE_CAPA         ((long) 0x00000010)
#define RCN_WIRE_RESI         ((long) 0x00000020)

/* node type for locon ptype */
#define RCN_LONODE            ((long) 0x10000000)

/* node type for lonode ptype */
#define RCN_LOCON             ((long) 0x10000000)

/* flags for lowire */

#define RCN_FLAG_MASK_SYS	((long) 0x0000000F)

#define	RCN_FLAG_CROSS		((long) 0x00000001)
#define RCN_FLAG_ONE		((long) 0x00000002)
#define RCN_FLAG_LOCON		((long) 0x00000004)
#define RCN_FLAG_TWO		((long) 0x00000008)

#define RCN_FLAG_MASK_USER	((long) 0xFFFFFFF0)

#define RCN_FLAG_USER_00	((long) 0x00000010)
#define RCN_FLAG_USER_01	((long) 0x00000020)
#define RCN_FLAG_USER_02	((long) 0x00000040)
#define RCN_FLAG_USER_03	((long) 0x00000080)
#define RCN_FLAG_USER_04	((long) 0x00000100)
#define RCN_FLAG_USER_05	((long) 0x00000200)
#define RCN_FLAG_USER_06	((long) 0x00000400)
#define RCN_FLAG_USER_07	((long) 0x00000800)
#define RCN_FLAG_USER_08	((long) 0x00001000)
#define RCN_FLAG_USER_09	((long) 0x00002000)
#define RCN_FLAG_USER_10	((long) 0x00004000)
#define RCN_FLAG_USER_11	((long) 0x00008000)
#define RCN_FLAG_USER_12	((long) 0x00010000)
#define RCN_FLAG_USER_13	((long) 0x00020000)
#define RCN_FLAG_USER_14	((long) 0x00040000)
#define RCN_FLAG_USER_15	((long) 0x00080000)
#define RCN_FLAG_USER_16	((long) 0x00100000)
#define RCN_FLAG_USER_17	((long) 0x00200000)
#define RCN_FLAG_USER_18	((long) 0x00400000)
#define RCN_FLAG_USER_19	((long) 0x00800000)
#define RCN_FLAG_USER_20	((long) 0x01000000)
#define RCN_FLAG_USER_21	((long) 0x02000000)
#define RCN_FLAG_USER_22	((long) 0x04000000)
#define RCN_FLAG_USER_23	((long) 0x08000000)
#define RCN_FLAG_USER_24	((long) 0x10000000)
#define RCN_FLAG_USER_25	((long) 0x20000000)
#define RCN_FLAG_USER_26	((long) 0x40000000)
#define RCN_FLAG_USER_27	((long) 0x80000000)

#define RCN_SETFLAG(m,b)	(m=(m)|(b))
#define RCN_CLEARFLAG(m,b)	(m=(m)&(~b))
#define RCN_GETFLAG(m,b)	((m)&(b))

#define RCN_VALIDNODE(n)	((n->FLAG)&(RCN_FLAG_MASK_SYS))

/*******************************************************************************
* interconnect network  structures                                             *
*******************************************************************************/

typedef struct lorcnet                        /* logical rc network           */
{
struct lorcnet    *NEXT;                      /* next lonrcet                 */
struct lowire     *WIRE;                      /* wire list                    */
struct loblocnode *NODE;                      /* node list                    */
long               NBNODE;                    /* node list                    */
long               INDEX;                     /* node list                    */
chain_list        *LOCON;                     /* locon list                   */
struct ptype      *USER;                      /* Application specific         */
}
lorcnet_list;

typedef struct lowire                         /* logical wire                 */
{
struct lowire		*NEXT;		/* next wire                          */
long			 NODE1;		/* interconnect network node1         */
long			 NODE2;		/* interconnect network node2         */
long			 TYPE;		/* wire type                          */
float			 RESI;		/* wire resistance                    */
float			 CAPA;		/* wire capacitance                   */
long			 X;		/* x position                         */
long			 Y;		/* y position                         */
long			 DX;		/* x rectangle wide                   */
long			 DY;		/* y rectangle length                 */
struct ptype		*USER;		/* Application specific               */
}
lowire_list;



typedef struct lonode			/* logical node                       */
{
struct chain		*WIRELIST;	/* wire list                          */
long			 FLAG;		/* Flags                              */
long			 INDEX;		/* node index                         */
struct ptype		*USER;		/* wire list                          */
}
lonode_list;



typedef struct lotabnode                      /* logical node table           */
{
struct lotabnode  *NEXT;
struct lonode      NODE[BUFSIZE];             /* table of node                */
}
lotabnode_list;

typedef struct loblocnode                     /* logical tabnode bloc         */
{
struct loblocnode *NEXT;
struct lotabnode  *TABNODE[BUFSIZE];          /* bloc of tabnode              */
}
loblocnode_list;


/*******************************************************************************
* externals for rcn view                                                       *
*******************************************************************************/

extern lorcnet_list	*addlorcnet();
extern lorcnet_list     *givelorcnet();
extern int		 freelorcnet();
extern lowire_list	*addlowire();
extern long		 addlonode();
extern int		 dellowire();
extern lorcnet_list	*getlorcnet();
extern lowire_list	*getlowire();
extern lonode_list	*getlonode();
extern void		 setloconnode();
extern void		 delloconnode();
extern void		 addcapawire();
extern void		 setcapawire();
extern void		 addresiwire();
extern void		 setresiwire();
extern void		 addpararesiwire();
extern num_list*	 getloconnode();
extern void		 delnumloconnode();
extern void		 reverseloconnode();
extern void		 buildtable();
extern void		 freetable();
extern void		 setallwireflag();
extern void		 clearallwireflag();
extern int		 chkloop();
#endif /* !RCN */
