/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*******************************************************************************
* rcn    : acces functions to logical rc network structures                    *
*                                                                              *
* version : 1.02                                                               *
* date   : 07/06/96                                                            *
*******************************************************************************/

#include MUT_H
#include MLO_H
#include RCN_H
#include "rcn_lo.h"

lorcnet_list	*HEAD_RCNET_FREE	= NULL;
unsigned int	 NBRCNET;
lowire_list	*HEAD_WIRE_FREE		= NULL;
unsigned int	 NBWIRE;
loblocnode_list *HEAD_BLOCNODE_FREE	= NULL;
unsigned int	 NBBLOCNODE;
lotabnode_list	*HEAD_TABNODE_FREE	= NULL;
unsigned int	 NBTABNODE;

/*******************************************************************************
* function addlorcnet()                                                        *
*******************************************************************************/
lorcnet_list *addlorcnet(ptsig)
losig_list *ptsig;
{
lorcnet_list *ptrcnet;
ptype_list *ptptype;
int i ;

  ptrcnet=heaprcnet();
  if(ptsig)
  {
    if( getptype(ptsig->USER,RCN_RCNET) )
    {
      (void)fflush(stdout);
      (void)fprintf(stderr, "*** rcn error ***\n");
      (void)fprintf(stderr, "addlorcnet impossible : ");
      (void)fprintf(stderr, "rc network already exists on signal %ld",ptsig->INDEX);
      exit(-1);
    }

    ptrcnet->INDEX = ptsig->INDEX;
    ptsig->USER = (ptype_list *)addptype(ptsig->USER,RCN_RCNET,ptrcnet);
  }
  else
  {
    ptrcnet->INDEX=0l;
  }

  return ptrcnet;
}

/*******************************************************************************
* function givelorcnet()                                                       *
*******************************************************************************/
lorcnet_list *givelorcnet(ptsig)
losig_list *ptsig;
{
  lorcnet_list *pt;
 
  pt=getlorcnet(ptsig);
  if(!pt)
    pt=addlorcnet(ptsig);
  return(pt);
}

/*******************************************************************************
* function addlowire()                                                         *
*******************************************************************************/
lowire_list *addlowire(ptrcnet, type, resi, capa, x, y, dx, dy, n1, n2)
lorcnet_list *ptrcnet;
long type;
float resi;
float capa;
long x;
long y;
long dx;
long dy;
long n1;
long n2;
{
lowire_list *ptwire;
lonode_list *ptnode1;
lonode_list *ptnode2;

  ptwire       = heaplowire();

  ptwire->TYPE = type;
  ptwire->RESI = resi;
  ptwire->CAPA = capa;
  ptwire->X    = x;
  ptwire->Y    = y;
  ptwire->DX   = dx;
  ptwire->DY   = dy;
  ptwire->NODE1 = n1;
  ptwire->NODE2 = n2;
  ptwire->USER = NULL;
  ptwire->NEXT = ptrcnet->WIRE;
  ptrcnet->WIRE = ptwire;

  if(ptrcnet->NODE != NULL)
  {
    ptnode1 = givelonode(ptrcnet, n1);
    ptnode2 = givelonode(ptrcnet, n2);
    ptnode1->WIRELIST = addchain(ptnode1->WIRELIST,ptwire);
    ptnode2->WIRELIST = addchain(ptnode2->WIRELIST,ptwire);

    switch( RCN_GETFLAG(ptnode1->FLAG,RCN_FLAG_ONE|RCN_FLAG_CROSS|RCN_FLAG_TWO))
    {
      case RCN_FLAG_ONE:
        RCN_CLEARFLAG(ptnode1->FLAG,RCN_FLAG_ONE);
        RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_TWO);
        break;
      case RCN_FLAG_TWO:
        RCN_CLEARFLAG(ptnode1->FLAG,RCN_FLAG_TWO);
        RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_CROSS);
      case RCN_FLAG_CROSS:
        break;
      default:
        RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_ONE);
    }

    switch( RCN_GETFLAG(ptnode2->FLAG,RCN_FLAG_ONE|RCN_FLAG_CROSS|RCN_FLAG_TWO))
    {
      case RCN_FLAG_ONE:
        RCN_CLEARFLAG(ptnode2->FLAG,RCN_FLAG_ONE);
        RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_TWO);
        break;
      case RCN_FLAG_TWO:
        RCN_CLEARFLAG(ptnode2->FLAG,RCN_FLAG_TWO);
        RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_CROSS);
        break;
      case RCN_FLAG_CROSS:
        break;
      default:
        RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_ONE);
    }
  }


  if(n1>=ptrcnet->NBNODE) 
    ptrcnet->NBNODE=n1+1;
  if(n2>=ptrcnet->NBNODE) 
    ptrcnet->NBNODE=n2+1;

  if (TRACE_MODE == 'Y')
    (void)fprintf(stdout, "--- rcn --- addlowire : between node %ld and node %ld on signal %ld\n",
	         ptwire->NODE1, ptwire->NODE2, ptrcnet->INDEX);

  return ptwire;
}

/*******************************************************************************
* function addlonode()                                                         *
*******************************************************************************/
long addlonode(ptrcnet, ptlocon)
lorcnet_list *ptrcnet;
locon_list *ptlocon;
{
  long index ;


  index = ptrcnet->NBNODE ;
  ptrcnet->NBNODE++ ;
  if(ptrcnet->NODE != NULL) givelonode(ptrcnet, index);
  if(ptlocon != NULL) setloconnode(ptrcnet,index, ptlocon);

  if (TRACE_MODE == 'Y')
  (void)fprintf(stdout, "--- rcn --- addlonode	  : on signal %ld\n",
                ptrcnet->INDEX);
  return index;
}

/*******************************************************************************
* function freelorcnet()                                                       *
*******************************************************************************/
int freelorcnet(ptrcnet,signal)
lorcnet_list	*ptrcnet;
losig_list	*signal;
{
  lowire_list		*scanwire;
  lowire_list		*prevwire;
  ptype_list		*pt;
  chain_list		*scancon;
  locon_list		*ptcon;
  num_list		*nodecon;
  num_list		*scannum;


  /* Efface les liens avec les connecteurs */

  /* On cre la liste des noeuds dont on veut effacer un lien avec un
     connecteur */

  nodecon=NULL;
  for(scancon=ptrcnet->LOCON;scancon;scancon=scancon->NEXT)
  {
    ptcon=(locon_list*)(scancon->DATA);
    if(pt=getptype(ptcon->USER,RCN_LONODE))
    {
      freenum(pt->DATA);
      ptcon->USER=delptype(ptcon->USER,RCN_LONODE);
    }
  }

  /* efface les wires */

  prevwire=NULL;
  for(scanwire=ptrcnet->WIRE;scanwire;scanwire=scanwire->NEXT)
  {
    if(prevwire)
      freelowire(prevwire);
    prevwire=scanwire;
  }

  if(prevwire)
    freelowire(prevwire);

  /* efface les noeuds si ils existent */

  if(ptrcnet->NODE)
    freetable(ptrcnet);

  /* Efface la liste des connecteurs */

  if(ptrcnet->LOCON)
    freechain(ptrcnet->LOCON);

  /* Finallement, on efface le reseau lui meme */

  freercnet(ptrcnet);
  if(signal)
    signal->USER=delptype(signal->USER,RCN_RCNET);
}

/*******************************************************************************
* function dellowire()                                                         *
*******************************************************************************/
int dellowire(ptrcnet, node1, node2)
lorcnet_list *ptrcnet;
long node1;
long node2;
{
lowire_list	*ptwire;
lonode_list	*ptnode1;
lonode_list	*ptnode2;
lowire_list	*ptsav;
chain_list	*ptchain;

  for (ptwire = ptrcnet->WIRE; ptwire; ptwire = ptwire->NEXT)
  {
    if((ptwire->NODE1 == node1 && ptwire->NODE2 == node2)
         ||(ptwire->NODE1 == node2 && ptwire->NODE2 == node1))
      break;
    ptsav = (void *)ptwire;
  }
  if (ptwire == NULL)   
    return 0;
  else 
  {
    if (ptwire == ptrcnet->WIRE)  
      ptrcnet->WIRE = ptwire->NEXT;
    else
      ptsav->NEXT = ptwire->NEXT;
  }

  freelowire(ptwire);

  if(ptrcnet->NODE != NULL)
  {
    ptnode1 = getlonode(ptrcnet, node1);
    ptnode2 = getlonode(ptrcnet, node2);

    for(ptchain = ptnode1->WIRELIST;ptchain != NULL;ptchain = ptchain->NEXT)
      if(ptchain->DATA == (void *)ptwire) break;
    ptnode1->WIRELIST = delchain(ptnode1->WIRELIST,ptchain);

    for(ptchain = ptnode2->WIRELIST;ptchain != NULL;ptchain = ptchain->NEXT)
      if(ptchain->DATA == (void *)ptwire) break;
        ptnode2->WIRELIST = delchain(ptnode2->WIRELIST,ptchain);

    switch(RCN_GETFLAG(ptnode1->FLAG,RCN_FLAG_CROSS|RCN_FLAG_ONE|RCN_FLAG_TWO))
    {
      case RCN_FLAG_CROSS:
        if(ptnode1->WIRELIST->NEXT->NEXT==NULL)
        {
          RCN_CLEARFLAG(ptnode1->FLAG,RCN_FLAG_CROSS);
          RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_TWO);
        }
        break;
      case RCN_FLAG_TWO:
        RCN_CLEARFLAG(ptnode1->FLAG,RCN_FLAG_TWO);
        RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_ONE);
        break;
      case RCN_FLAG_ONE:
        RCN_CLEARFLAG(ptnode1->FLAG,RCN_FLAG_ONE);
        break;
      default:
        fprintf(stderr,"burp... (dellowire)\n");
        exit(-1);
    }

    switch(RCN_GETFLAG(ptnode2->FLAG,RCN_FLAG_CROSS|RCN_FLAG_ONE|RCN_FLAG_TWO))
    {
      case RCN_FLAG_CROSS:
        if(ptnode2->WIRELIST->NEXT->NEXT==NULL)
        {
          RCN_CLEARFLAG(ptnode2->FLAG,RCN_FLAG_CROSS);
          RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_TWO);
        }
        break;
      case RCN_FLAG_TWO:
        RCN_CLEARFLAG(ptnode2->FLAG,RCN_FLAG_TWO);
        RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_ONE);
        break;
      case RCN_FLAG_ONE:
        RCN_CLEARFLAG(ptnode2->FLAG,RCN_FLAG_ONE);
        break;
      default:
        fprintf(stderr,"burp... (dellowire)\n");
        exit(-1);
    }
  }
  
  if (TRACE_MODE == 'Y')
    (void)fprintf(stdout, "--- rcn --- dellowire  : between node %ld and node %ld on signal %ld\n",
                           node1, node2, ptrcnet->INDEX);

  return 1;
}

/*******************************************************************************
* function getlorcnet()                                                        *
*******************************************************************************/
lorcnet_list *getlorcnet(ptsig)
losig_list  *ptsig;
{
lorcnet_list  *ptrcnet;
ptype_list  *ptptype;

	ptptype = (ptype_list  *)getptype(ptsig->USER, RCN_RCNET);
	if (ptptype != NULL) {
	   ptrcnet = (lorcnet_list *)ptptype->DATA;
	   return ptrcnet;
	}
  return(NULL);
}

/*******************************************************************************
* function getlowire()                                                         *
*******************************************************************************/
lowire_list *getlowire(ptrcnet, node1, node2)
lorcnet_list  *ptrcnet;
long  node1;
long  node2;
{
lowire_list  *ptwire;
lonode_list  *ptnode;
chain_list   *ptchain;
	if(ptrcnet->NODE == NULL)
        {
 	  for (ptwire = ptrcnet->WIRE; ptwire; ptwire = ptwire->NEXT)
	    if ((ptwire->NODE1 == node1 && ptwire->NODE2 == node2)
	       ||(ptwire->NODE1 == node2 && ptwire->NODE2 == node1))
            {
	       return ptwire;
            }

          return(NULL);
        }

	ptnode = givelonode(ptrcnet,node1);
	for(ptchain=ptnode->WIRELIST;ptchain!=NULL;ptchain=ptchain->NEXT)
        {
	    ptwire = (lowire_list*)ptchain->DATA;
	    if ((ptwire->NODE1 == node1 && ptwire->NODE2 == node2)
	        ||(ptwire->NODE1 == node2 && ptwire->NODE2 == node1))
	       return ptwire;
	}

        return(NULL);
}

/*******************************************************************************
* function getlonode()                                                         *
*******************************************************************************/
lonode_list *getlonode(ptrcnet, index)
lorcnet_list  *ptrcnet;
long  index;
{
  lonode_list  *ptnode;
  loblocnode_list  *ptblocnode;
  long million;
  long millier;
  long unit;
  int i;
  if( (ptblocnode=ptrcnet->NODE)==NULL)
  {
    fflush(stdout);
    fprintf(stderr,"*** rcn error ***\n");
    fprintf(stderr,"You must build the node table before use the function\n");
    fprintf(stderr,"getlonode.\n");
    EXIT(1);
  }
  million = index / (BUFSIZE * BUFSIZE);
  millier = (index % (BUFSIZE * BUFSIZE)) / BUFSIZE;
  unit    = (index % (BUFSIZE * BUFSIZE)) % BUFSIZE;

  for(i=0;i<million;i++)
    ptblocnode=ptblocnode->NEXT;

  return(ptblocnode->TABNODE[millier]->NODE+unit);
}
 
/*******************************************************************************
* function givelonode()                                                        *
*******************************************************************************/
lonode_list *givelonode(ptrcnet, index)
lorcnet_list  *ptrcnet;
long  index;
{
  loblocnode_list  *ptblocnode;
  loblocnode_list  *ptblocnodesav;
  lotabnode_list  *pttabnode;
  lonode_list  *ptnode;
  long million;
  long millier;
  long unit;
  long i;

  million = index / (BUFSIZE * BUFSIZE);
  millier = (index % (BUFSIZE * BUFSIZE)) / BUFSIZE;
  unit    = (index % (BUFSIZE * BUFSIZE)) % BUFSIZE;

  if(ptrcnet->NODE == NULL)
    ptrcnet->NODE=heaploblocnode();

  ptblocnode = ptrcnet->NODE;
  ptblocnodesav = ptblocnode;

  for(i = 0 ; i < million ; i++)
  {
    if(ptblocnode == NULL)
    {
      ptblocnode = heaploblocnode();
      ptblocnodesav->NEXT = ptblocnode;
    }
    ptblocnodesav = ptblocnode;
    ptblocnode = ptblocnode->NEXT;
  }

  if(ptblocnode->TABNODE[millier] == NULL)
  {
    ptblocnode->TABNODE[millier] = heaplotabnode( (million*BUFSIZE+millier)*BUFSIZE);
  }

  return(ptblocnode->TABNODE[millier]->NODE+unit);
}
 
/*******************************************************************************
* function setloconnode()                                                      *
*******************************************************************************/

void setloconnode(ptrcnet, node, ptlocon)
lorcnet_list *ptrcnet;
long node;
locon_list *ptlocon;
{ 
  ptype_list		*ptptype;
  lonode_list		*ptnode;
  chain_list		*chain;

  if(ptrcnet->NBNODE<=node)
    ptrcnet->NBNODE=node+1;

  if(ptrcnet->NODE != NULL)
  {
    /* Le noeud qu'on ajoute peut ne pas avoir ete mis dans la table */
    /* ptnode = getlonode(ptrcnet,node); */

    ptnode=givelonode(ptrcnet,node);
    if(!RCN_GETFLAG(ptnode->FLAG,RCN_FLAG_LOCON))
    {
      RCN_SETFLAG(ptnode->FLAG,RCN_FLAG_LOCON);
      ptnode->USER = addptype(ptnode->USER,RCN_LOCON,addchain(NULL,ptlocon));
    }
    else
    {
      ptptype = getptype(ptnode->USER,RCN_LOCON);
      for(chain = (chain_list *)ptptype->DATA; chain; chain = chain->NEXT)
        if(ptlocon == (locon_list *)chain->DATA)
          break;

      if(chain == NULL) 
        ptptype->DATA = addchain(ptptype->DATA,ptlocon);
    }
  }

  ptptype = getptype(ptlocon->USER,RCN_LONODE);

  if(ptptype == NULL)
  {
    ptlocon->USER = addptype(ptlocon->USER,RCN_LONODE,addnum(NULL,node));
  }
  else
  {
    num_list *chain;
    for(chain = (num_list *)ptptype->DATA ; chain ; chain = chain->NEXT)
      if(node == chain->DATA)
        break;
    if(chain == NULL)
      ptptype->DATA = addnum(ptptype->DATA,node);
  }

  for(chain=ptrcnet->LOCON;chain;chain=chain->NEXT)
  {
    if(chain->DATA == ptlocon)
      break;
  }

  if(!chain)
  {
    ptrcnet->LOCON=addchain(ptrcnet->LOCON,ptlocon);
  }
}

/*******************************************************************************
* function delloconnode()                                                      *
*******************************************************************************/

void delloconnode(ptrcnet,index)
lorcnet_list	*ptrcnet;
long		 index;
{
  chain_list	*scanchain;
  chain_list	*erasecon;
  locon_list	*ptlocon;
  num_list	*scannum,*prev,*elem;
  lonode_list	*ptlonode;
  ptype_list	*pt,*pt2;

  erasecon=NULL;

  if(!ptrcnet->NODE)
  {
    for(scanchain=ptrcnet->LOCON;scanchain;scanchain=scanchain->NEXT)
    {
      ptlocon=(locon_list*)(scanchain->DATA);
      pt=getptype(ptlocon->USER,RCN_LONODE);
      prev=NULL;
      for(scannum=(num_list*)(pt->DATA);scannum;scannum=scannum->NEXT)
      {
        if(scannum->DATA==index)
          break;
        prev=scannum;
      }

      if(scannum) /* On a trouve un lonode sur ce connecteur */
      {
        if(!prev)
        {  
          if(scannum->NEXT)
            pt->DATA=scannum->NEXT;
          else
          {
            ptlocon->USER=delptype(ptlocon->USER,RCN_LONODE);
            erasecon=addchain(erasecon,ptlocon);
          }
        }
        else
        {
          prev->NEXT=scannum->NEXT;
        }

        scannum->NEXT=NULL;
        freenum(scannum);
      }
    }
  }
  else
  {
    ptlonode=getlonode(ptrcnet,index);

    if(RCN_GETFLAG(ptlonode->FLAG,RCN_FLAG_LOCON))
    {
      pt=getptype(ptlonode->USER,RCN_LOCON);
      for(scanchain=(chain_list*)(pt->DATA);scanchain;scanchain=scanchain->NEXT)
      {
        ptlocon=(locon_list*)(scanchain->DATA);
        pt2=getptype(ptlocon->USER,RCN_LONODE);
        prev=NULL;
        for(scannum=(num_list*)(pt2->DATA);scannum;scannum=scannum->NEXT)
        {
          if(scannum->DATA==index)
            break;
          prev=scannum;
        }

        if(!prev)
        {
          if(scannum->NEXT)
            pt2->DATA=scannum->NEXT;
          else
          {
            ptlocon->USER=delptype(ptlocon->USER,RCN_LONODE);
            erasecon=addchain(erasecon,ptlocon);
          }
        }
        else
        {
          prev->NEXT=scannum->NEXT;
        }

        scannum->NEXT=NULL;
        freenum(scannum);
      }
      freechain(pt->DATA);
      ptlonode->USER=delptype(ptlonode->USER,RCN_LOCON);
      RCN_CLEARFLAG(ptlonode->FLAG,RCN_FLAG_LOCON);
    }
  }

  if(erasecon)
  {
    chain_list	*new;
    chain_list	*scanerase;

    new=NULL;

    for(scanchain=ptrcnet->LOCON;scanchain;scanchain=scanchain->NEXT)
    {
      for(scanerase=erasecon;scanerase;scanerase=scanerase->NEXT)
      {
        if(scanerase->DATA==scanchain->DATA) 
          /* on a trouve un element a effacer*/
          break;
      }
      if(!scanerase)
        new=addchain(new,scanchain->DATA); 
    }

    freechain(erasecon);
    freechain(ptrcnet->LOCON);
    ptrcnet->LOCON=new;
  }
}

/*******************************************************************************
* function addcapawire()                                                       *
*******************************************************************************/

void addcapawire(ptwire, capa)
lowire_list *ptwire;
float capa;
{
	ptwire->CAPA += capa;
}

/*******************************************************************************
* function setcapawire()                                                       *
*******************************************************************************/

void setcapawire(ptwire, capa)
lowire_list *ptwire;
float capa;
{
	ptwire->CAPA = capa;
}

/*******************************************************************************
* function addresiwire()                                                       *
*******************************************************************************/
void addresiwire(ptwire, resi)
lowire_list *ptwire;
float resi;
{
	ptwire->RESI += resi;
}

/*******************************************************************************
* function setresiwire()                                                       *
*******************************************************************************/
void setresiwire(ptwire, resi)
lowire_list *ptwire;
float resi;
{
	ptwire->RESI = resi;
}

/*============================================================================*/

/*******************************************************************************
* function addpararesiwire()                                                   *
*******************************************************************************/
void addpararesiwire(ptwire, resi)
lowire_list *ptwire;
float resi;
{
        if(ptwire->RESI==0.0 || resi==0.0)
	  ptwire->RESI = 0.0;
        else
          ptwire->RESI = 1 / ( 1/ptwire->RESI + 1/resi );
}

/*******************************************************************************
* function heaploblocnode()                                                    *
*******************************************************************************/
loblocnode_list* heaploblocnode()
{
  loblocnode_list *ptblocnode;
  int i;

  if (HEAD_BLOCNODE_FREE == NULL)
  {
    HEAD_BLOCNODE_FREE = (loblocnode_list*)mbkalloc(sizeof(loblocnode_list));
    HEAD_BLOCNODE_FREE->NEXT = NULL;
  }
  ptblocnode = HEAD_BLOCNODE_FREE;
  HEAD_BLOCNODE_FREE=HEAD_BLOCNODE_FREE->NEXT;

  /* Clean the new loblocnode */
  ptblocnode->NEXT = NULL;
  for(i=0;i<BUFSIZE;i++)
    ptblocnode->TABNODE[i]=NULL;

  NBBLOCNODE++;
  return(ptblocnode);
}

/*******************************************************************************
* function heaplotabnode()                                                     *
*******************************************************************************/
lotabnode_list* heaplotabnode(baseindex)
long baseindex;
{
  lotabnode_list *pttabnode;
  int i;

  if (HEAD_TABNODE_FREE == NULL)
  {
    HEAD_TABNODE_FREE = (lotabnode_list*)mbkalloc(sizeof(lotabnode_list));
    HEAD_TABNODE_FREE->NEXT = NULL;
  }
  pttabnode = HEAD_TABNODE_FREE;
  HEAD_TABNODE_FREE=HEAD_TABNODE_FREE->NEXT;

  /* Clean the new lotabnode */
  pttabnode->NEXT = NULL;
  for(i=0;i<BUFSIZE;i++)
  {
    pttabnode->NODE[i].WIRELIST=NULL;
    pttabnode->NODE[i].USER=NULL;
    pttabnode->NODE[i].INDEX=baseindex+i;
    pttabnode->NODE[i].FLAG=0;
  }

  NBTABNODE++;
  return(pttabnode);
}

/*******************************************************************************
* lowire_list* heaplowire(void)						       *
*******************************************************************************/


lowire_list* heaplowire()
{
  lowire_list *ptwire;
  int i;

  if(HEAD_WIRE_FREE==NULL)
  {
    HEAD_WIRE_FREE=(lowire_list*)mbkalloc(BUFSIZE*sizeof(lowire_list));
    ptwire=HEAD_WIRE_FREE;
    for(i=1;i<BUFSIZE;i++)
    {
      ptwire->NEXT=ptwire+1;
      ptwire++;
    }
    ptwire->NEXT=NULL;
  }

  ptwire=HEAD_WIRE_FREE;
  HEAD_WIRE_FREE=HEAD_WIRE_FREE->NEXT;

  /* Clean the new wire */
  ptwire->NEXT		= NULL;
  ptwire->NODE1		= 0L;
  ptwire->NODE2		= 0L;
  ptwire->TYPE		= RCN_WIRE_UNKNOW;
  ptwire->RESI		= 0.0;
  ptwire->CAPA		= 0.0;
  ptwire->X		= 0l;
  ptwire->Y		= 0l;
  ptwire->DX		= 0l;
  ptwire->DY		= 0l;
  ptwire->USER		= NULL;

  NBWIRE++;
  return(ptwire);
}

/*******************************************************************************
* lowire_list* heaprcnet(void)						       *
*******************************************************************************/

lorcnet_list* heaprcnet()
{
  lorcnet_list *ptrcnet;
  int i;

  if(HEAD_RCNET_FREE==NULL)
  {
    HEAD_RCNET_FREE=(lorcnet_list*)mbkalloc(BUFSIZE*sizeof(lorcnet_list));
    ptrcnet=HEAD_RCNET_FREE;
    for(i=1;i<BUFSIZE;i++)
    {
      ptrcnet->NEXT=ptrcnet+1;
      ptrcnet++;
    }
    ptrcnet->NEXT=NULL;
  }

  ptrcnet=HEAD_RCNET_FREE;
  HEAD_RCNET_FREE=HEAD_RCNET_FREE->NEXT;

  /* Clean the new rcnet figure */
  ptrcnet->NEXT		= NULL;
  ptrcnet->WIRE		= NULL;
  ptrcnet->NODE		= NULL;
  ptrcnet->NBNODE	= 1l;
  ptrcnet->INDEX	= 0l;
  ptrcnet->LOCON	= NULL;
  ptrcnet->USER		= NULL;

  NBRCNET++;
  return(ptrcnet);
}

/*******************************************************************************
* void freercnet(lorcnet_list*)						       *
*******************************************************************************/

void freercnet(pt)
lorcnet_list	*pt;
{
  pt->NEXT=HEAD_RCNET_FREE;
  HEAD_RCNET_FREE=pt;

  NBRCNET--;
}

/*******************************************************************************
* void freelowire(lowire_list*)						       *
*******************************************************************************/

void freelowire(pt)
lowire_list	*pt;
{
  pt->NEXT=HEAD_WIRE_FREE;
  HEAD_WIRE_FREE=pt;

  NBWIRE--;
}

/*******************************************************************************
* void freeloblocnode(loblocnode_list*)					       *
*******************************************************************************/

void freeloblocnode(pt)
loblocnode_list	*pt;
{
  pt->NEXT=HEAD_BLOCNODE_FREE;
  HEAD_BLOCNODE_FREE=pt;

  NBBLOCNODE--;
}

/*******************************************************************************
* void freelotabnode(lotabnode_list*)					       *
*******************************************************************************/

void freelotabnode(pt)
lotabnode_list	*pt;
{
  pt->NEXT=HEAD_TABNODE_FREE;
  HEAD_TABNODE_FREE=pt;

  NBTABNODE--;
}

/*******************************************************************************
* void freetable(rcnet_list*)						       *
*******************************************************************************/

void freetable(ptrcnet)
lorcnet_list	*ptrcnet;
{
  loblocnode_list	*scanloblocnode;
  loblocnode_list	*prevloblocnode;
  int			 indexloblocnode;
  int			 indexlotabnode;
  ptype_list		*pt;


  if(ptrcnet->NODE==NULL)
  {
    fprintf(stderr,"\n*** rcn warning ***\n");
    fprintf(stderr,"You can use the function freetable only when de node\n");
    fprintf(stderr,"table exist.\n");
    return;
  }
  prevloblocnode=NULL;

  for(scanloblocnode=ptrcnet->NODE;
      scanloblocnode;
      scanloblocnode=scanloblocnode->NEXT)
  {
    if(prevloblocnode)
      freeloblocnode(prevloblocnode);

    for(indexloblocnode=0;indexloblocnode<BUFSIZE;indexloblocnode++)
    {
      if( scanloblocnode->TABNODE[indexloblocnode] )
      {
        for(indexlotabnode=0;indexlotabnode<BUFSIZE;indexlotabnode++)
        {
          if( 
         scanloblocnode->TABNODE[indexloblocnode]->NODE[indexlotabnode].WIRELIST
            )
          {
            freechain(
         scanloblocnode->TABNODE[indexloblocnode]->NODE[indexlotabnode].WIRELIST
                     );
          }
          if(
             scanloblocnode->TABNODE[indexloblocnode]->NODE[indexlotabnode].USER
                     );
          {
            if(RCN_GETFLAG(
            scanloblocnode->TABNODE[indexloblocnode]->NODE[indexlotabnode].FLAG,
                       RCN_FLAG_LOCON))
            {
              pt=getptype(
            scanloblocnode->TABNODE[indexloblocnode]->NODE[indexlotabnode].USER,
                           RCN_LOCON);
/* debug*/
              if(!pt)
              {
                printf("Plouf!\n");
                exit(0);
              }
              freechain(pt->DATA);
            }

            freeptype(
             scanloblocnode->TABNODE[indexloblocnode]->NODE[indexlotabnode].USER
                     );
          }
        }
        freelotabnode(scanloblocnode->TABNODE[indexloblocnode]);
      }
    }
    prevloblocnode=scanloblocnode;
  }
  if(prevloblocnode)
    freeloblocnode(prevloblocnode);

  ptrcnet->NODE=NULL;
}

/*******************************************************************************
* void buildtable(ptrcnet)						       *
*******************************************************************************/

void buildtable(ptrcnet)
lorcnet_list	*ptrcnet;
{
  lowire_list		*ptwire;
  lonode_list		*ptnode1;
  lonode_list		*ptnode2;
  chain_list		*scancon;
  long			 i,max;
  locon_list		*ptcon;
  ptype_list		*ptlonode;
  num_list		*scannum;
  lonode_list		*ptnode;
  ptype_list		*ptlocon;

  /* Cree les table pour tous les noeuds jusqu'a nbnode */
    givelonode(ptrcnet,ptrcnet->NBNODE);

  for(ptwire=ptrcnet->WIRE;ptwire;ptwire=ptwire->NEXT)
  {
    ptnode1=givelonode(ptrcnet,ptwire->NODE1);
    ptnode2=givelonode(ptrcnet,ptwire->NODE2);
    ptnode1->WIRELIST=addchain(ptnode1->WIRELIST,ptwire);
    ptnode2->WIRELIST=addchain(ptnode2->WIRELIST,ptwire);

    switch( RCN_GETFLAG(ptnode1->FLAG,RCN_FLAG_ONE|RCN_FLAG_CROSS|RCN_FLAG_TWO))
    {
      case RCN_FLAG_ONE:
        RCN_CLEARFLAG(ptnode1->FLAG,RCN_FLAG_ONE);
        RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_TWO);
        break;
      case RCN_FLAG_TWO:
        RCN_CLEARFLAG(ptnode1->FLAG,RCN_FLAG_TWO);
        RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_CROSS);
        break;
      case RCN_FLAG_CROSS:
        break;
      default:
        RCN_SETFLAG(ptnode1->FLAG,RCN_FLAG_ONE);
    }

    switch( RCN_GETFLAG(ptnode2->FLAG,RCN_FLAG_ONE|RCN_FLAG_CROSS|RCN_FLAG_TWO))
    {
      case RCN_FLAG_ONE:
        RCN_CLEARFLAG(ptnode2->FLAG,RCN_FLAG_ONE);
        RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_TWO);
        break;
      case RCN_FLAG_TWO:
        RCN_CLEARFLAG(ptnode2->FLAG,RCN_FLAG_TWO);
        RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_CROSS);
        break;
      case RCN_FLAG_CROSS:
        break;
      default:
        RCN_SETFLAG(ptnode2->FLAG,RCN_FLAG_ONE);
    }
  }  

  /* Logicals connectors of figure */
  for(scancon=ptrcnet->LOCON;scancon;scancon=scancon->NEXT)
  {
    ptcon=(locon_list*)(scancon->DATA);

    if(ptlonode=getptype(ptcon->USER,RCN_LONODE))
    {
      for(scannum=(num_list*)(ptlonode->DATA);scannum;scannum=scannum->NEXT)
      {
        ptnode=givelonode(ptrcnet,scannum->DATA);
        if(RCN_GETFLAG(ptnode->FLAG,RCN_FLAG_LOCON))
        {
          ptlocon=getptype(ptnode->USER,RCN_LOCON);
          (chain_list*)(ptlocon->DATA)=
                                   addchain((chain_list*)(ptlocon->DATA),ptcon);
        }
        else
        {
          RCN_SETFLAG(ptnode->FLAG,RCN_FLAG_LOCON);
          ptnode->USER=addptype(ptnode->USER,RCN_LOCON,addchain(NULL,ptcon));
        }
      }
    }
    else
    {
      fprintf(stderr,"Houmpfa\n");
      exit(-1);
    }
  }
}

/*******************************************************************************
*  getloconnode(ptlocon)                                                       *
*******************************************************************************/

num_list* getloconnode(ptlocon)
locon_list	*ptlocon;
{
  ptype_list	*pt;

  pt=getptype(ptlocon->USER,RCN_LONODE);
  if(pt)
    return((num_list*)(pt->DATA));

  return(NULL);
}

/*******************************************************************************
*  delnumloconnode(ptlocon)                                                    *
*******************************************************************************/

void delnumloconnode(ptlocon)
locon_list	*ptlocon;
{
  ptype_list	*pt;

  pt=getptype(ptlocon->USER,RCN_LONODE);
  if(pt)
  {
    freenum((num_list*)(pt->DATA));
    ptlocon->USER=delptype(ptlocon->USER,RCN_LONODE);
  }
  else
    fprintf(stderr,
   "\n *** rcn warning *** you can't delete a numlist if it doesn't exist.!\n");
}

void reverseloconnode(ptlocon)
locon_list	*ptlocon;
{
  ptype_list	*pt;
  num_list	*headnum,*invnum,*scannum;

  pt=getptype(ptlocon->USER,RCN_LONODE);

  if(pt)
  {
    headnum=(num_list*)(pt->DATA);
    invnum=NULL;

    for(scannum=headnum;scannum;scannum=scannum->NEXT)
      invnum=addnum(invnum,scannum->DATA);

    freenum(headnum);
    pt->DATA=invnum;
  }
}

int chkloopmain(ptrcnet,flag,org,from)
lorcnet_list	*ptrcnet;
long		 flag;
long		 org;
lowire_list	*from;
{
  lonode_list	*ptnode;
  long		 current_node;
  int		 r;
  chain_list	*scanchain;
  lowire_list	*current_wire;

  ptnode=getlonode(ptrcnet,org);
  if(RCN_GETFLAG(ptnode->FLAG,flag))
    return(2);
  RCN_SETFLAG(ptnode->FLAG,flag);

  for(scanchain=ptnode->WIRELIST;scanchain;scanchain=scanchain->NEXT)
  {
    current_wire=(lowire_list*)(scanchain->DATA);
    if(current_wire!=from)
    {
      current_node=(current_wire->NODE1==org ? current_wire->NODE2 :
                                               current_wire->NODE1 );
      ptnode=getlonode(ptrcnet,current_node);

      while(RCN_GETFLAG(ptnode->FLAG,RCN_FLAG_TWO))
      {
        if(RCN_GETFLAG(ptnode->FLAG,flag))
          return(2);
        RCN_SETFLAG(ptnode->FLAG,flag);
        current_wire=(ptnode->WIRELIST->DATA==current_wire ?
                      ptnode->WIRELIST->NEXT->DATA :
                      ptnode->WIRELIST->DATA                  );
        current_node=(current_wire->NODE1==current_node ? current_wire->NODE2 :
                                                 current_wire->NODE1 );
        ptnode=getlonode(ptrcnet,current_node);
      }

      if( RCN_GETFLAG(ptnode->FLAG,RCN_FLAG_ONE))
      {
        if(RCN_GETFLAG(ptnode->FLAG,flag))
          return(2);
        RCN_SETFLAG(ptnode->FLAG,flag);
      }

      if( RCN_GETFLAG(ptnode->FLAG,RCN_FLAG_CROSS))
      {
        r=chkloopmain(ptrcnet,flag,current_node,current_wire);
        if(r)
          return(r);
      }
    }
  }
  return(0);
}

int chkloop(ptrcnet,flag)
lorcnet_list	*ptrcnet;
long		 flag;
{
  int		 r;
  long		 i;
  int		 nbpaq;
  lonode_list	*ptnode;
  
  if(!ptrcnet->NODE)
    return(1);

  clearallwireflag(ptrcnet,flag);

  nbpaq=0; 
  for(i=1;i<ptrcnet->NBNODE;i++)
  {
    ptnode=getlonode(ptrcnet,i);
    if(!RCN_GETFLAG(ptnode->FLAG,flag) && RCN_VALIDNODE(ptnode))
    {
      nbpaq++;
      r=chkloopmain(ptrcnet,flag,i,NULL);

      if(r) break;
    }
  }

  clearallwireflag(ptrcnet,flag);

  return(r);
}

void setallwireflag(ptrcnet,flag)
lorcnet_list	*ptrcnet;
long		 flag;
{
  long			 i;
  long			 j;
  loblocnode_list	*ptblocnode;

  if(!ptrcnet->NODE)
    return;

  for(ptblocnode=ptrcnet->NODE;ptblocnode;ptblocnode=ptblocnode->NEXT)
  {
    for(i=0;i<BUFSIZE;i++)
    {
      if( *(ptblocnode->TABNODE+i) )
      {
        for(j=0;j<BUFSIZE;j++)
        {
          RCN_SETFLAG( ((*(ptblocnode->TABNODE+i))->NODE+j)->FLAG,flag);
        }
      }
    }
  }
}

void clearallwireflag(ptrcnet,flag)
lorcnet_list	*ptrcnet;
long		 flag;
{
  long			 i;
  long			 j;
  loblocnode_list	*ptblocnode;

  if(!ptrcnet->NODE)
    return;

  for(ptblocnode=ptrcnet->NODE;ptblocnode;ptblocnode=ptblocnode->NEXT)
  {
    for(i=0;i<BUFSIZE;i++)
    {
      if( *(ptblocnode->TABNODE+i) )
      {
        for(j=0;j<BUFSIZE;j++)
        {
          RCN_CLEARFLAG( ((*(ptblocnode->TABNODE+i))->NODE+j)->FLAG,flag);
        }
      }
    }
  }
}

void disprcnstat()
{
  int sizercnet,sizewire,sizeblocnode,sizetabnode;

  sizercnet=sizeof(lorcnet_list)*NBRCNET;
  sizewire=sizeof(lowire_list)*NBWIRE;
  sizeblocnode=sizeof(loblocnode_list)*NBBLOCNODE;
  sizetabnode=sizeof(lotabnode_list)*NBTABNODE;

  printf("\n");
  printf("******** Taille mmoire des vue RCN *********\n\n\n");
  printf("Nombre d'RCNET.......... %6u = %8u octets\n",NBRCNET,sizercnet);
  printf("Nombre d'RCNWIRE........ %6u = %8u octets\n",NBWIRE,sizewire);
  printf("Nombre d'RCNTABNODE..... %6u = %8u octets\n",NBTABNODE,sizetabnode);
  printf("Nombre d'RCNBLOCNODE.... %6u = %8u octets\n",NBBLOCNODE,sizeblocnode);
  printf("                                  --------\n");
  printf("                                 %9u octets\n\n",sizercnet+sizewire+
                                                      sizetabnode+sizeblocnode);

}

