/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
# include <stdio.h>

# include MPH_H
# include MUT_H
# include RDS_H
# include RUT_H
# include RTL_H
# include RFM_H
# include RPR_H
# include RWI_H

void Syntax( Name )

  char *Name;
{
  fprintf( stdout, "%s [options] in out\n", Name );
  fprintf( stdout, " -r : real figure\n" );
  fprintf( stdout, " -f : flatten figure\n" );
  fprintf( stdout, " -v : display figure \n" );
  fprintf( stdout, " -s : save figure \n" );

  exit( 1 );
}

int main( argc, argv )

    int   argc;
    char *argv[];
{
  rdsfig_list    *FigureRds;
  phfig_list     *FigureMbk;

  char           *InputName;
  char           *OutputName;
  char            FlagSymbolic;
  char            FlagFlatten;
  char            FlagVerbose;
  char            FlagSave;

  int             Number;

  if ( argc < 2 )
  {
    Syntax( argv[ 0 ] );
  }

  mbkenv();
  rdsenv();

  if ( argc < 2 )
  {
    Syntax( argv[ 0 ] );
  }

  FlagSymbolic = 1;
  FlagFlatten  = 0;
  FlagVerbose  = 0;
  FlagSave     = 0;

  InputName  = (char *)NULL;
  OutputName = (char *)NULL;

  for ( Number  = 1;
        Number  < argc;
        Number++ )
  {
    if ( ! strcmp( argv[ Number ], "-r" )) FlagSymbolic = 0;
    else
    if ( ! strcmp( argv[ Number ], "-f" )) FlagFlatten = 1;
    else
    if ( ! strcmp( argv[ Number ], "-v" )) FlagVerbose = 1;
    else
    if ( ! strcmp( argv[ Number ], "-s" )) FlagSave = 1;
    else
    if ( InputName == NULL ) InputName = argv[ Number ];
    else
    if ( OutputName == NULL ) OutputName = argv[ Number ];
  }

  if ( InputName == NULL )
  {
    Syntax( argv[ 0 ] );
  }

  if ( OutputName == NULL )
  {
    OutputName = InputName;
  }

  rdsbegin();
  loadrdsparam();

  if ( FlagSymbolic )
  {
    fprintf( stdout, "Load figure MBK\n" );
    FigureMbk = getphfig( InputName, 'A' );

    if ( FlagFlatten )
    {
      fprintf( stdout, "Flatten figure MBK\n" );
      rflattenphfig( FigureMbk, YES, NO );
    }

    fprintf( stdout, "Translate figure MBK -> RDS\n" );
    FigureRds = figmbkrds( FigureMbk, RDS_EQUI_SIZE, 0 );

    fprintf( stdout, "Translate model MBK -> RDS\n" );
    modelmbkrds( FigureRds, 0 );
  }
  else
  {
    fprintf( stdout, "Load figure RDS\n" );
    FigureRds = getrdsfig( InputName, 'A', RDS_EQUI_SIZE  );

    if ( FlagFlatten )
    {
      fprintf( stdout, "Flatten figure RDS\n" );
      rflattenrdsfig( FigureRds, RDS_YES, RDS_NO );
    }
  }

  if ( FlagVerbose )
  {
    fprintf( stdout, "Display figure RDS\n" );
    viewrdsfig( FigureRds );
  }

  if ( FlagSave )
  {
    FigureRds->NAME = namealloc( OutputName );
    fprintf( stdout, "Save figure RDS\n" );
    saverdsfig( FigureRds );
  }

  rdsend();
  return( 0 );
 }
