/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rdsdel.c                        |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>

# include MUT_H
# include RDS_H

# include "rdsdel.h"
# include "rdserror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Rds Del Functions                    |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Rds Del Figure                       |
|                                                             |
\------------------------------------------------------------*/

int delrdsfig( Name )

   char *Name;
{
  rdsfig_list  *Figure;
  rdsfig_list **Previous;
  rdsins_list  *Instance;
  rdsrec_list  *Rectangle;
  void         *Save;
  char          Layer;

  Name     = namealloc( Name );
  Previous = &HEAD_RDSFIG;

  for ( Figure  = HEAD_RDSFIG;
        Figure != (rdsfig_list *)NULL;
        Figure  = Figure->NEXT )
  {
    if ( Figure->NAME == Name ) break;

    Previous = &Figure->NEXT;
  }

  if ( Figure == (rdsfig_list *)NULL ) 
  
    return( 0 );

  *Previous = Figure->NEXT;

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    Rectangle  = Figure->LAYERTAB[ Layer ];

    while ( Rectangle != (rdsrec_list *)NULL )
    {
      Save = (void *)(Rectangle->NEXT);

      freerdsrec( Rectangle, Figure->SIZE );

      Rectangle = (rdsrec_list *)Save;
    }
  }

  Instance = Figure->INSTANCE;

  while ( Instance != (rdsins_list *)NULL )
  {
    for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
    {
      Rectangle  = Instance->LAYERTAB[ Layer ];
  
      while ( Rectangle != (rdsrec_list *)NULL )
      {
        Save = (void *)(Rectangle->NEXT);
  
        freerdsrec( Rectangle, Figure->SIZE );
  
        Rectangle = (rdsrec_list *)Save;
      }
    }

    Save = (void *)(Instance->NEXT);

    freerdsins( Instance );

    Instance = (rdsins_list *)Save;
  }
   
  freerdsfig( Figure );

  return( 1 );
}

/*------------------------------------------------------------\
|                                                             |
|                        Rds Del Instance                     |
|                                                             |
\------------------------------------------------------------*/

int delrdsins( Figure, Name )

   rdsfig_list *Figure;
   char        *Name;
{
  rdsins_list **Previous;
  rdsins_list  *Instance;
  rdsrec_list  *Rectangle;
  void         *Save;
  char          Layer;

  Name     = namealloc( Name );
  Previous = &Figure->INSTANCE;

  for ( Instance  = Figure->INSTANCE;
        Instance != (rdsins_list *)NULL;
        Instance  = Instance->NEXT )
  {
    if ( Instance->INSNAME == Name ) break;

    Previous = &Instance->NEXT;
  }

  if ( Instance == (rdsins_list *)NULL ) 
  
    return( 0 );

  *Previous = Instance->NEXT;

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    Rectangle  = Instance->LAYERTAB[ Layer ];

    while ( Rectangle != (rdsrec_list *)NULL )
    {
      Save = (void *)(Rectangle->NEXT);

      freerdsrec( Rectangle, Figure->SIZE );

      Rectangle = (rdsrec_list *)Save;
    }
  }

  freerdsins( Instance );

  return( 1 );
}
 
/*------------------------------------------------------------\
|                                                             |
|                   Rds Del Rectangle In Instance             |
|                                                             |
\------------------------------------------------------------*/

int delrdsinsrec( Instance, Rectangle )

   rdsins_list *Instance;
   rdsrec_list *Rectangle;
{
  rdsrec_list **Previous;
  rdsrec_list  *Scan;

  Previous = &Instance->LAYERTAB[ GetRdsLayer( Rectangle ) ];

  for ( Scan  = Instance->LAYERTAB[ GetRdsLayer( Rectangle ) ];
        Scan != (rdsrec_list *)NULL;
        Scan  = Scan->NEXT )
  {
    if ( Scan == Rectangle ) break;

    Previous = &Scan->NEXT;
  }

  if ( Scan == (rdsrec_list *)NULL ) 
  
    return( 0 );

  *Previous = Rectangle->NEXT;

  freerdsrec( Rectangle, Instance->SIZE );

  return( 1 );
}
 
/*------------------------------------------------------------\
|                                                             |
|                   Rds Del Rectangle In Figure               |
|                                                             |
\------------------------------------------------------------*/

int delrdsfigrec( Figure, Rectangle )

   rdsfig_list *Figure;
   rdsrec_list *Rectangle;
{
  rdsrec_list **Previous;
  rdsrec_list  *Scan;

  Previous = &Figure->LAYERTAB[ GetRdsLayer( Rectangle ) ];

  for ( Scan  = Figure->LAYERTAB[ GetRdsLayer( Rectangle ) ];
        Scan != (rdsrec_list *)NULL;
        Scan  = Scan->NEXT )
  {
    if ( Scan == Rectangle ) break;

    Previous = &Scan->NEXT;
  }

  if ( Scan == (rdsrec_list *)NULL ) 
  
    return( 0 );

  *Previous = Rectangle->NEXT;

  freerdsrec( Rectangle, Figure->SIZE );

  return( 1 );
}

