/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rdsview.c                       |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>

# include MUT_H
# include MPH_H
# include RDS_H

# include "rdsview.h"
# include "rdserror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         View Functions                      |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                       View Rds Rectangle                    |
|                                                             |
\------------------------------------------------------------*/

void viewrdsrec( Rectangle )

     rdsrec_list *Rectangle;
{
  char *LayerName;

  LayerName = RDS_LAYER_NAME[ GetRdsLayer( Rectangle ) ];

  fprintf( stdout, "\n\t--> Rectangle" );

  if ( Rectangle->NAME != (char *)NULL )
  {
    fprintf( stdout, "\n\t\tNAME  : %s" , Rectangle->NAME  );
  }
  else
  {   
    fprintf( stdout, "\n\t\tNAME  : NO NAME" );
  }

  fprintf( stdout, "\n\t\tX     : %ld" , Rectangle->X     );
  fprintf( stdout, "\n\t\tY     : %ld" , Rectangle->Y     );
  fprintf( stdout, "\n\t\tDX    : %ld" , Rectangle->DX    );
  fprintf( stdout, "\n\t\tDY    : %ld" , Rectangle->DY    );
  fprintf( stdout, "\n\t\tFLAGS : %l0X", Rectangle->FLAGS );
  fprintf( stdout, "\n\t\tLAYER : %s"  , LayerName        );
}

/*------------------------------------------------------------\
|                                                             |
|                       View Rds Instance                     |
|                                                             |
\------------------------------------------------------------*/

void viewrdsins( Instance )

     rdsins_list *Instance;
{
  char         Layer;
  rdsrec_list *Rectangle;
  char        *TransfName;

  TransfName = RDS_TRANSF_NAME[ Instance->TRANSF ];

  fprintf( stdout, "\n--> Instance" );

  fprintf( stdout, "\n\tINSNAME : %s" , Instance->INSNAME );
  fprintf( stdout, "\n\tFIGNAME : %s" , Instance->FIGNAME );
  fprintf( stdout, "\n\tX       : %ld", Instance->X       );
  fprintf( stdout, "\n\tY       : %ld", Instance->Y       );
  fprintf( stdout, "\n\tTRANSF  : %s" , TransfName        );
  fprintf( stdout, "\n\tSIZE    : %d" , Instance->SIZE    );
  fprintf( stdout, "\n\tFLAGS   : %x" , Instance->FLAGS   );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    for ( Rectangle  = Instance->LAYERTAB[ Layer ];
          Rectangle != (rdsrec_list *)NULL;
          Rectangle  = Rectangle->NEXT )
    {
      viewrdsrec( Rectangle );
    }
  }
}

/*------------------------------------------------------------\
|                                                             |
|                       View Rds Figure                       |
|                                                             |
\------------------------------------------------------------*/

void viewrdsfig( Figure )

     rdsfig_list *Figure;
{
  char         Layer;
  rdsins_list *Instance;
  rdsrec_list *Rectangle;

  fprintf( stdout, "\n--> Figure" );

  fprintf( stdout, "\n\tNAME  : %s", Figure->NAME  );
  fprintf( stdout, "\n\tMODE  : %c", Figure->MODE  );
  fprintf( stdout, "\n\tSIZE  : %d", Figure->SIZE  );
  fprintf( stdout, "\n\tFLAGS : %x", Figure->FLAGS );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    for ( Rectangle  = Figure->LAYERTAB[ Layer ];
          Rectangle != (rdsrec_list *)NULL;
          Rectangle  = Rectangle->NEXT )
    {
      viewrdsrec( Rectangle );
    }
  }

  for ( Instance  = Figure->INSTANCE;
        Instance != (rdsins_list *)NULL;
        Instance  = Instance->NEXT )
  {
    viewrdsins( Instance );
  }
}
