/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                  rprerror.c                       |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>

# include MUT_H
# include MPH_H
# include RDS_H
# include RUT_H
# include RPR_H
# include RTL_H

# include "rprerror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/

void rpr_error( Error, Text, Data, File, Line )

  char  Error;
  char *Text;
  long  Data;
  char *File;
  long  Line;
{
  char Name[ 128 ];

  strcpy( Name, File );
  Name[ strlen( File ) - 1 ] = '\0';

  fprintf( stderr, "rpr:%s%d ", Name, Line );

  switch( Error )
  {
    case RPR_DEFINE_MISSING : 
  
      fprintf( stderr, "Some defines missing\n");
      break;
  
    case RPR_TABLE_MISSING  : 
   
      fprintf( stderr, "Some tables missing\n");
      break;
  
    case RPR_LINE_EXPECTED  :
   
      fprintf( stderr, "%s expected line %lu\n", 
      Text, Data );
      break;
  
    case RPR_UNKNOWN_DEFINE : 
  
      fprintf( stderr, "Unknown define %s line %lu\n",
      Text, Data );
      break;
  
    case RPR_NOT_DEFINED    : 
  
      fprintf( stderr, "%s not defined line %lu\n",
      Text, Data );
      break;
  
    case RPR_MULTIPLE_GRID  : 
   
      fprintf( stderr, "%s not multiple of physical grid line %lu\n",
      Text, Data );
      break;
  
    case RPR_TOO_SMAL       :
  
      fprintf( stderr, "%s too smal line %lu\n",
      Text, Data );
      break;
  
    case RPR_MISSING_VALUE  :
  
      fprintf( stderr, "Missing value line %lu\n", Data );
      break;
  
    case RPR_MISSING_NAME   :
  
      fprintf( stderr, "Missing name of %s line %lu\n",
      Text, Data );
      break;
  
    case RPR_ILLEGAL_FLOAT  :
  
      fprintf( stderr, "Illegal floating point number %s line %lu\n",
      Text, Data );
      break;
  
    case RPR_UNEXPECTED_LINE :
  
      fprintf( stderr, "%s unexpected line %lu\n",
      Text, Data );
      break;
  
    case RPR_UNEXPECTED_EOF  :
  
      fprintf( stderr, "Unexpected end of file, missing definitions\n");
      break;
  
    case RPR_TOO_MANY_WORDS  :
  
      fprintf( stderr, "Too many words %s unexpected line %lu\n",
      Text, Data );
      break;
  
    case RPR_MISSING_TABLE  :
  
      fprintf( stderr, "%s Missing value in %s table line %lu\n",
      Text, Data );
      break;
  
    case RPR_OPEN_FILE      :
  
      fprintf( stderr, "Technological file can't be opened\n" );
      break;
  
    case RPR_MISSING_POST   :
  
      fprintf( stderr, "Missing %s value in postreat table line %lu\n",
      Text , Data );
      break;
  
    case RPR_NEGATIVE_VALUE :
  
      fprintf( stderr, "%s has to be positive value line %lu\n",
      Text , Data );
      break;
  
    case RPR_SYNTAX_ERROR   :
  
      fprintf( stderr, "Syntax Error %s at line %lu\n",
      Text , Data );
      break;

    default :

      fprintf( stderr, "unknown internal error %d !\n", Error );
  }

  EXIT( 1 );
}
