/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rprparse.h                      |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/

# ifndef RPR_PARSE_H
# define RPR_PARSE_H

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/

# define RPR_MAX_RDS_UNIT      100000L
# define RPR_EPSILON           ((double)0.001)
# define RPR_MAX_BUFFER        512
# define RPR_MAX_KEYWORD       78

# define RPR_SEPARATORS_STRING " \t\n"
# define RPR_COMMENT_CHAR      '#'

/*------------------------------------------------------------\
|                                                             |
|                        Keyword Constants                    |
|                                                             |
\------------------------------------------------------------*/

# define DEFINE_KEYWORD          "DEFINE"
# define TABLE_KEYWORD           "TABLE" 
# define END_TABLE_KEYWORD       "END"
# define END_RECORD_KEYWORD       NULL
# define LAMBDA_KEYWORD          "LAMBDA"
# define PHYSICAL_GRID_KEYWORD   "PHYSICAL_GRID"
# define SEGMENT_KEYWORD         "MBK_TO_RDS_SEGMENT"
# define VIA_KEYWORD             "MBK_TO_RDS_VIA"
# define CONNECTOR_KEYWORD       "MBK_TO_RDS_CONNECTOR"
# define REFERENCE_KEYWORD       "MBK_TO_RDS_REFERENCE"
# define CIF_LAYER_KEYWORD       "CIF_LAYER"
# define GDS_LAYER_KEYWORD       "GDS_LAYER" 
# define S2R_POST_TREAT_KEYWORD  "S2R_POST_TREAT"
# define S2R_OVERSIZE_KEYWORD    "S2R_OVERSIZE_DENOTCH"
# define S2R_RING_WIDTH_KEYWORD  "S2R_BLOC_RING_WIDTH"
# define S2R_LAYER_WIDTH_KEYWORD "S2R_MINIMUM_LAYER_WIDTH"
# define LYNX_GRAPH_KEYWORD      "LYNX_GRAPH"
# define LYNX_CAPA_KEYWORD       "LYNX_CAPA"
# define LYNX_RESISTOR_KEYWORD   "LYNX_RESISTOR"
# define LYNX_TRANSISTOR_KEYWORD "LYNX_TRANSISTOR"
# define LYNX_DIFFUSION_KEYWORD  "LYNX_DIFFUSION"

/*------------------------------------------------------------\
|                                                             |
|                            Mask                             |
|                                                             |
\------------------------------------------------------------*/

# define RPR_LAMBDA_MASK            0x00001
# define RPR_PHYSICAL_GRID_MASK     0x00002
# define RPR_SEGMENT_MASK           0x00004
# define RPR_VIA_MASK               0x00008
# define RPR_CONNECTOR_MASK         0x00010
# define RPR_REFERENCE_MASK         0x00020
# define RPR_CIF_LAYER_MASK         0x00040
# define RPR_GDS_LAYER_MASK         0x00080
# define RPR_S2R_POST_TREAT_MASK    0x00100
# define RPR_S2R_OVERSIZE_MASK      0x00200
# define RPR_S2R_RING_WIDTH_MASK    0x00400
# define RPR_S2R_LAYER_WIDTH_MASK   0x00800
# define RPR_LYNX_GRAPH_MASK        0x01000
# define RPR_LYNX_CAPA_MASK         0x02000
# define RPR_LYNX_TRANSISTOR_MASK   0x04000
# define RPR_LYNX_RESISTOR_MASK     0x08000
# define RPR_LYNX_DIFFUSION_MASK    0x18000
                                         
# define RPR_ALL_DEFINED_MASK       0x1FFFF

/*------------------------------------------------------------\
|                                                             |
|                           Macros                            |
|                                                             |
\------------------------------------------------------------*/

# define RprEqual(a,b)                                       \
                                                             \
     ((((a)-RPR_EPSILON) <= (b)) && (((b)) <= (a)+RPR_EPSILON))

/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/

  typedef struct keyword
  {
    char *NAME;
    char  VALUE;

  } keyword;

/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/

# endif
