/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rprview.c                       |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <math.h>
# include <ctype.h>
# include <stdlib.h>
# include <string.h>

# include MUT_H
# include MPH_H
# include RDS_H
# include RUT_H
# include RPR_H
# include RTL_H

# include "rprerror.h"
# include "rprview.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Rds view Parameters                    |
|                                                             |
\------------------------------------------------------------*/

void viewrdsparam()

{
  char  Layer;
  char  Index;

  fprintf( stdout, "\n  DEFINE PHYSICAL_GRID %d", RDS_PHYSICAL_GRID );
  fprintf( stdout, "\n  DEFINE LAMBDA %d"       , RDS_LAMBDA );
  fprintf( stdout, "\n# DEFINE RDS_UNIT %d"     , RDS_UNIT );

  fprintf( stdout, "\n\nTABLE MBK_TO_RDS_SEGMENT\n" );

  for ( Layer = 0; Layer < MBK_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", MBK_LAYER_NAME[ Layer ] );

    Index = 0;

    while ( GET_SEGMENT_LAYER( Layer, Index ) != RDS_SEGMENT_EMPTY )
    {
      fprintf( stdout, "  %s %s %2d %2d %2d %s", 
               RDS_LAYER_NAME[ GET_SEGMENT_LAYER( Layer, Index ) ],
               RDS_TRANS_NAME[ GET_SEGMENT_TRANS( Layer, Index ) ],
               GET_SEGMENT_DLR   ( Layer, Index ),
               GET_SEGMENT_DWR   ( Layer, Index ),
               GET_SEGMENT_OFFSET( Layer, Index ),
               RDS_USE_NAME[ GET_SEGMENT_USE( Layer, Index ) ] );

      Index = Index + 1;
    }
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE MBK_TO_RDS_CONNECTOR\n" );

  for ( Layer = 0; Layer < MBK_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", MBK_LAYER_NAME[ Layer ] );

    if ( GET_CONNECTOR_LAYER( Layer ) != RDS_CONNECTOR_EMPTY )
    {
      fprintf( stdout, " %s %2d %2d", 
               RDS_LAYER_NAME[ GET_CONNECTOR_LAYER( Layer ) ],
               GET_CONNECTOR_DER  ( Layer ),
               GET_CONNECTOR_DWR  ( Layer ) );
    }
  } 

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE MBK_TO_RDS_REFERENCE\n" );

  for ( Layer = 0; Layer < MBK_MAX_REFERENCE; Layer++ )
  {
    if ( Layer == 0 ) fprintf( stdout, "\nREF_REF" );
    else              fprintf( stdout, "\nREF_CON" );

    Index = 0;

    while ( GET_REFERENCE_LAYER( Layer, Index ) != RDS_REFERENCE_EMPTY )
    {
      fprintf( stdout, " %s %2d", 
               RDS_LAYER_NAME[ GET_REFERENCE_LAYER( Layer, Index ) ],
               GET_REFERENCE_SIZE( Layer, Index ) );

      Index = Index + 1;
    }
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE MBK_TO_RDS_VIA\n" );

  for ( Layer = 0; Layer < MBK_MAX_VIA; Layer++ )
  {
    Index = 0;

    fprintf( stdout, "\n%s", MBK_VIA_NAME[ Layer ] );

    while ( GET_VIA_LAYER( Layer, Index ) != RDS_VIA_EMPTY )
    {
      fprintf( stdout, "  %s %2d %s", 
               RDS_LAYER_NAME[ GET_VIA_LAYER( Layer, Index ) ],
               GET_VIA_SIZE( Layer, Index ),
               RDS_USE_NAME[ GET_VIA_USE( Layer, Index ) ] );

      Index = Index + 1;
    }
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE CIF_LAYER\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    if ( GET_CIF_LAYER ( Layer ) != RDS_CIF_LAYER_EMPTY )
    {
      fprintf( stdout, " %s", GET_CIF_LAYER ( Layer ) );
    }
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE GDS_LAYER\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    if ( GET_GDS_LAYER( Layer ) != RDS_GDS_LAYER_EMPTY )
    {
      fprintf( stdout, " %d", GET_GDS_LAYER ( Layer ) );
    }
  }
 
  fprintf( stdout, "\n\nEND" );
 
  fprintf( stdout, "\n\nTABLE S2R_OVERSIZE_DENOTCH\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    if ( GET_S2R_OVERSIZE( Layer ) != RDS_S2R_OVERSIZE_EMPTY )
    {
      fprintf( stdout, " %ld", GET_S2R_OVERSIZE ( Layer ) );
    }
  }
 
  fprintf( stdout, "\n\nEND" );
 
  fprintf( stdout, "\n\nTABLE S2R_BLOC_RING_WIDTH\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    if ( GET_S2R_RING_WIDTH( Layer ) != RDS_S2R_RING_WIDTH_EMPTY )
    {
      fprintf( stdout, " %ld", GET_S2R_RING_WIDTH ( Layer ) );
    }
  }
 
  fprintf( stdout, "\n\nEND" );
 
  fprintf( stdout, "\n\nTABLE S2R_MINIMUM_LAYER_WIDTH\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    if ( GET_S2R_LAYER_WIDTH( Layer ) != RDS_S2R_LAYER_WIDTH_EMPTY )
    {
      fprintf( stdout, " %ld", GET_S2R_LAYER_WIDTH ( Layer ) );
    }
  }
 
  fprintf( stdout, "\n\nEND" );
 
  fprintf( stdout, "\n\nTABLE S2R_POST_TREAT\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    if ( GET_S2R_POST_Y_OR_N( Layer ) != RDS_S2R_POST_TREAT_EMPTY )
    {
      fprintf( stdout, " %ld", GET_S2R_POST_Y_OR_N ( Layer ) );

      if ( GET_S2R_POST_COMPLEMENTARY( Layer ) != RDS_S2R_POST_TREAT_EMPTY )
      {
        fprintf( stdout, " %s", 
                 RDS_LAYER_NAME[ GET_S2R_POST_COMPLEMENTARY( Layer ) ] );
      }
    }
  }
 
  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE LYNX_GRAPH\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    Index = 0;

    while ( GET_LYNX_GRAPH_LAYER( Layer, Index ) != RDS_LYNX_GRAPH_EMPTY )
    {
      fprintf( stdout, "  %s ", 
               RDS_LAYER_NAME[ GET_LYNX_GRAPH_LAYER( Layer, Index ) ] );

      Index = Index + 1;
    }
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE LYNX_DIFFUSION\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    Index = 0;

    while ( GET_LYNX_DIFFUSION_LAYER( Layer, Index ) != RDS_LYNX_DIFFUSION_EMPTY )
    {
      fprintf( stdout, "  %s ",
               RDS_LAYER_NAME[ GET_LYNX_DIFFUSION_LAYER( Layer, Index ) ] );
      fprintf( stdout, "  %d ",
               GET_LYNX_DIFFUSION_Y_OR_N( Layer, Index ) );

      Index = Index + 1;
    }
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE LYNX_CAPA\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    fprintf( stdout, "  %g ", GET_LYNX_SURFACE_CAPA_LAYER( Layer ) * RDS_UNIT2  );
    fprintf( stdout, "  %g ", GET_LYNX_PERIMETER_CAPA_LAYER( Layer ) * RDS_UNIT );
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE LYNX_RESISTOR\n" );

  for ( Layer = 0; Layer < RDS_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", RDS_LAYER_NAME[ Layer ] );

    fprintf( stdout, "  %g ", GET_LYNX_RESISTOR_LAYER( Layer ) );
  }

  fprintf( stdout, "\n\nEND" );

  fprintf( stdout, "\n\nTABLE LYNX_TRANSISTOR\n" );

  for ( Layer = 0; Layer < MBK_MAX_LAYER; Layer++ )
  {
    fprintf( stdout, "\n%s", MBK_LAYER_NAME[ Layer ] );

    if ( GET_LYNX_TRANSISTOR_TYPE( Layer ) != RDS_LYNX_TRANSISTOR_EMPTY )
    {
      fprintf( stdout, "  %s", MBK_LAYER_NAME[ GET_LYNX_TRANSISTOR_TYPE( Layer ) ] );
      fprintf( stdout, "  %s", MBK_VIA_NAME[ GET_LYNX_TRANSISTOR_CX( Layer ) ] );
      fprintf( stdout, "  %s", RDS_LAYER_NAME[ GET_LYNX_TRANSISTOR_GATE_LAYER( Layer ) ] );
      fprintf( stdout, "  %s", RDS_LAYER_NAME[ GET_LYNX_TRANSISTOR_DIFF_LAYER( Layer ) ] );

      if ( GET_LYNX_TRANSISTOR_ACTIV_LAYER( Layer ) != RDS_LYNX_TRANSISTOR_EMPTY )
      {
        fprintf( stdout, "  %s", RDS_LAYER_NAME[ GET_LYNX_TRANSISTOR_ACTIV_LAYER( Layer ) ] );
      }
      else
      {
        fprintf( stdout, "  NULL" );
      }
    }
  }

  fprintf( stdout, "\n\nEND\n\n" );
}
