/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rtlacces.c                      |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdlib.h>

# include MPH_H
# include MUT_H
# include RDS_H
# include RUT_H
# include RPR_H
# include RTL_H

# include "rtlacces.h"
# include "rtlenv.h"
# include "rtlerror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          loadrdsfig                         |
|                                                             |
\------------------------------------------------------------*/

void loadrdsfig( Figure, Name, Mode )

  rdsfig_list *Figure;
  char        *Name;
  char         Mode;
{
  if ( RDS_IN == RDS_GDS_FORMAT ) 
  {
    gdsloadrdsfig( Figure, Name, Mode );
  }
  else
  if ( RDS_IN == RDS_CIF_FORMAT )
  {
    cifloadrdsfig( Figure, Name, Mode );
  }
  else
  {
    rtlerror( RTL_UNKNOWN_RDS_IN, RDS_IN );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                          saverdsfig                         |
|                                                             |
\------------------------------------------------------------*/

void saverdsfig( Figure )

  rdsfig_list *Figure;
{
  if ( RDS_OUT == RDS_GDS_FORMAT )
  {
    gdssaverdsfig( Figure );
  }
  else
  if ( RDS_OUT == RDS_CIF_FORMAT )
  {
    cifsaverdsfig( Figure );
  }
  else
  {
    rtlerror( RTL_UNKNOWN_RDS_OUT, RDS_OUT );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                          getrdsfig                          |
|                                                             |
\------------------------------------------------------------*/

rdsfig_list *getrdsfig( Name, Mode, Size )

   char        *Name;
   char         Mode;
   unsigned int Size;
{
  rdsfig_list *Figure;

  Name = namealloc( Name );

  for ( Figure  = HEAD_RDSFIG;
        Figure != (rdsfig_list *)NULL;
        Figure  = Figure->NEXT )
  {
    if ( Figure->NAME == Name ) break;
  }

  if ( ( Figure != (rdsfig_list *)NULL ) &&
       ( Mode   != Figure->MODE        ) )
  {
    if ( Figure->MODE == 'P' )
    {
      delrdsfig( Figure->NAME );
      Figure = (rdsfig_list *)NULL;
    }
    else
    {
      applyrdsmode( Figure, Mode );
    }
  }

  if ( Figure == (rdsfig_list *)NULL )
  {
    Figure = addrdsfig( Name, Size );

    loadrdsfig( Figure, Name, Mode );
  }

  return( Figure );
}
