/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rut206.h                        |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/

# ifndef RUT_206_H
# define RUT_206_H

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                           Equi Mode                         |
|                                                             |
\------------------------------------------------------------*/

# define RDS_EQUI_SIZE  ( sizeof( rdsrecequi ) )

/*------------------------------------------------------------\
|                                                             |
|                           Mask for Equi                     |
|                                                             |
\------------------------------------------------------------*/

# define RDS_EQUI_MASK       (long)0x00020000
# define RDS_END_EQUI_MASK   (long)0x00040000

/*------------------------------------------------------------\
|                                                             |
|                          Mask for Unify                     |
|                                                             |
\------------------------------------------------------------*/

# define RDS_DELETED_MASK    (long)0x00080000

/*------------------------------------------------------------\
|                                                             |
|                           Yes or No                         |
|                                                             |
\------------------------------------------------------------*/

# define RDS_YES     1
# define RDS_NO      0

/*------------------------------------------------------------\
|                                                             |
|                             Macros                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                           Grid Aligned                      |
|                                                             |
\------------------------------------------------------------*/

# define CheckPhysicalGridAligned(c) ((c) % RDS_PHYSICAL_GRID)

/*------------------------------------------------------------\
|                                                             |
|                       Macros for Equi                       |
|                                                             |
\------------------------------------------------------------*/

# define IsRdsEqui( Rectangle )                               \
                                                              \
   ( ( Rectangle )->FLAGS & RDS_EQUI_MASK                     )

# define SetRdsEqui( Rectangle )                              \
                                                              \
   ( ( Rectangle )->FLAGS |= RDS_EQUI_MASK                    )

# define ClearRdsEqui( Rectangle )                            \
                                                              \
   ( ( Rectangle )->FLAGS &= ~RDS_EQUI_MASK                   )

/*------------------------------------------------------------\
|                                                             |
|                    Macros for End Equi                      |
|                                                             |
\------------------------------------------------------------*/

# define IsRdsEndEqui( Rectangle )                            \
                                                              \
   ( ( Rectangle )->FLAGS & RDS_END_EQUI_MASK                 )

# define SetRdsEndEqui( Rectangle )                           \
                                                              \
   ( ( Rectangle )->FLAGS |= RDS_END_EQUI_MASK                )

# define ClearRdsEndEqui( Rectangle )                         \
                                                              \
   ( ( Rectangle )->FLAGS &= ~RDS_END_EQUI_MASK               )

/*------------------------------------------------------------\
|                                                             |
|                    Macros for Delete                        |
|                                                             |
\------------------------------------------------------------*/

# define IsRdsDeleted( Rectangle )                            \
                                                              \
   ( ( Rectangle )->FLAGS & RDS_DELETED_MASK                  )

# define SetRdsDeleted( Rectangle )                           \
                                                              \
   ( ( Rectangle )->FLAGS |= RDS_DELETED_MASK                 )

# define ClearRdsDeleted( Rectangle )                         \
                                                              \
   ( ( Rectangle )->FLAGS &= ~RDS_DELETED_MASK                )

/*------------------------------------------------------------\
|                                                             |
|                          Equi Link                          |
|                                                             |
\------------------------------------------------------------*/

# define RDS_EQUI( R )                                        \
                                                              \
  (((rdsrecequi *)((char *)(R) + sizeof(rdsrec_list)))->EQUI  )

/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Rds Rectangle Equi                   |
|                                                             |
\------------------------------------------------------------*/

  typedef struct rdsrecequi
  {
    rdsrec_list *EQUI;

  } rdsrecequi;


/*------------------------------------------------------------\
|                                                             |
|                       Coordinates for polygons              |
|                                                             |
\------------------------------------------------------------*/

  typedef struct coord_t
  {
    long X;
    long Y;

  } coord_t;

/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  extern char RDS_SYMMETRY_TABLE[ RDS_MAX_TRANSF ][ RDS_MAX_TRANSF ];

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Search Functions                    |
|                                                             |
\------------------------------------------------------------*/

  extern rdsfig_list *searchrdsfig();
  extern rdsins_list *searchrdsins();

/*------------------------------------------------------------\
|                                                             |
|                         Model Functions                     |
|                                                             |
\------------------------------------------------------------*/

  extern ptype_list  *getrdsmodellist();

/*------------------------------------------------------------\
|                                                             |
|                         Apply Functions                     |
|                                                             |
\------------------------------------------------------------*/

  extern void         applyrdssym();
  extern void         applyrdsmode();

/*------------------------------------------------------------\
|                                                             |
|                        Flatten Functions                    |
|                                                             |
\------------------------------------------------------------*/

  extern void         flattenrdsfig();
  extern void         rflattenrdsfig();

/*------------------------------------------------------------\
|                                                             |
|                          Equi Functions                     |
|                                                             |
\------------------------------------------------------------*/

  extern rdsrec_list *equirdsfig();

/*------------------------------------------------------------\
|                                                             |
|                          Unify Functions                    |
|                                                             |
\------------------------------------------------------------*/

  extern rdsrec_list *unifyrdsfig();

/*------------------------------------------------------------\
|                                                             |
|                          Polygon Functions                  |
|                                                             |
\------------------------------------------------------------*/

  extern void         diag2rec();
  extern void         poly2diag();
  extern int          ispolyrec();

# endif
