/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rwi107.h                        |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/

# ifndef RWI_107_H
# define RWI_107_H

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/

# define RWI_MAX_REC       8
# define RWI_MAX_LAYER     RDS_ABOX

/*------------------------------------------------------------\
|                                                             |
|                        Flags for window                     |
|                                                             |
\------------------------------------------------------------*/

# define RDS_ONE_WINDOW     (long)0x00010000

/*------------------------------------------------------------\
|                                                             |
|                            Macros                           |
|                                                             |
\------------------------------------------------------------*/

# define IsRdsOneWindow( Rectangle )                          \
                                                              \
   ( ( Rectangle )->FLAGS & RDS_ONE_WINDOW                    )

# define SetRdsOneWindow( Rectangle )                         \
                                                              \
   ( ( Rectangle )->FLAGS |= RDS_ONE_WINDOW                   )

# define ClearRdsOneWindow( Rectangle )                       \
                                                              \
   ( ( Rectangle )->FLAGS &= ~RDS_ONE_WINDOW                  )

/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                  Rds Window, Rectangle List                 |
|                                                             |
\------------------------------------------------------------*/

  typedef struct rdswinrec_list   

  {
    struct rdswinrec_list *NEXT; 
    rdsrec_list           *RECTAB[ RWI_MAX_REC ];

  } rdswinrec_list;

/*------------------------------------------------------------\
|                                                             |
|                      Rds Window list                        |
|                                                             |
\------------------------------------------------------------*/

  typedef struct rdswin_list

  {
    struct rdswinrec_list **LAYERTAB;

  } rdswin_list;

/*------------------------------------------------------------\
|                                                             |
|                  Rds rectangle, Window List                 |
|                                                             |
\------------------------------------------------------------*/

  typedef struct rdsrecwin_list 

  {
    struct rdsrecwin_list *NEXT;
    struct rdswin_list    *WINDOW;
    
  } rdsrecwin_list;

/*------------------------------------------------------------\
|                                                             |
|                         Rds Window                          |
|                                                             |
\------------------------------------------------------------*/

  typedef struct rdswindow

  {
    long          XMIN;
    long          YMIN;
    long          XMAX;
    long          YMAX;
    long          MASK;
    long          DX;
    long          DY;
    long          SIDE;
    long          SIZE;
    rdswin_list  *WINTAB;
    
  } rdswindow;

/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                       Alloc functions                       |
|                                                             |
\------------------------------------------------------------*/

  extern rdswinrec_list *allocrdswinrec();
  extern rdsrecwin_list *allocrdsrecwin();
  extern rdswin_list    *allocrdswin();
  extern rdswin_list    *allocrdswinlayer();
  extern rdswindow      *allocrdswindow();

/*------------------------------------------------------------\
|                                                             |
|                        Free functions                       |
|                                                             |
\------------------------------------------------------------*/

  extern void   freerdswinrec();
  extern void   freerdsrecwin();
  extern void   freerdswin();
  extern void   freerdswinlayer();
  extern void   freerdswindow();
  
/*------------------------------------------------------------\
|                                                             |
|                     Get functions                           |
|                                                             |
\------------------------------------------------------------*/

  extern chain_list *getrdswindow();
   
/*------------------------------------------------------------\
|                                                             |
|                 Build & Destroy functions                   |
|                                                             |
\------------------------------------------------------------*/
   
  extern void        addrdsrecwindow();
  extern void        delrdsrecwindow();
  extern rdswindow  *buildrdswindow();
  extern void        destroyrdswindow();
  extern void        viewrdswindow();

# endif
