/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rwidel.c                        |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>
# include <math.h>

# include MUT_H
# include RDS_H
# include RTL_H
# include RWI_H

# include "rwidel.h"
# include "rwierror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                     Del Rectangle In Window                 |
|                                                             |
\------------------------------------------------------------*/

void delrdsrecwindow( Rectangle, RdsWindow )

     rdsrec_list *Rectangle;
     rdswindow   *RdsWindow;
{
  rdsrecwin_list  StaticRecWin;
  rdsrecwin_list *ScanRecWin;
  rdsrecwin_list *DelRecWin;
  rdswinrec_list *ScanWinRec;
  rdswinrec_list *FirstWinRec;
  rdswin_list    *ScanWin;

  char            Layer;
  char            Index;
  char            Found;

  Layer = GetRdsLayer( Rectangle );

  if ( IsRdsOneWindow( Rectangle ) )
  {
    StaticRecWin.NEXT   = (rdsrecwin_list *)NULL;
    StaticRecWin.WINDOW = (rdswin_list *)Rectangle->USER;
    ScanRecWin          = &StaticRecWin;
  }
  else
  {
    ScanRecWin = (rdsrecwin_list *)(Rectangle->USER);
  }

  Rectangle->USER = (void *)NULL;
  ClearRdsOneWindow( Rectangle );

  while ( ScanRecWin != (rdsrecwin_list *)NULL )
  {
    DelRecWin  = ScanRecWin;
    ScanRecWin = ScanRecWin->NEXT;

    ScanWin = DelRecWin->WINDOW;
    Found   = RWI_MAX_REC;

    if ( ScanWin->LAYERTAB == (rdswinrec_list **)NULL )
    {
      allocrdswinlayer( ScanWin );
    }

    FirstWinRec = ScanWin->LAYERTAB[ Layer ];

    for ( ScanWinRec  = FirstWinRec;
          ScanWinRec != (rdswinrec_list *)NULL;
          ScanWinRec  = ScanWinRec->NEXT )
    {
      for ( Index = 0; Index < RWI_MAX_REC; Index++ )
      {
        if ( ScanWinRec->RECTAB[ Index ] == Rectangle )
        {
          Found = Index; break;
        }
      }

      if ( Found != RWI_MAX_REC ) break;
    }

    if ( ScanWinRec == FirstWinRec )
    {
      ScanWinRec->RECTAB[ Found ] = (rdsrec_list *)NULL;
    }
    else
    {
      for ( Index = 0; Index < RWI_MAX_REC; Index++ )
      {
        if ( FirstWinRec->RECTAB[ Index ] != (rdsrec_list *)NULL ) break;
      }

      ScanWinRec->RECTAB[ Found ]  = FirstWinRec->RECTAB[ Index ];
      FirstWinRec->RECTAB[ Index ] = (rdsrec_list *)NULL;
    }

    for ( Index = 0; Index < RWI_MAX_REC; Index++ )
    {
      if ( FirstWinRec->RECTAB[ Index ] != (rdsrec_list *)NULL ) break;
    }

    if ( Index == RWI_MAX_REC )
    {
      ScanWin->LAYERTAB[ Layer ] = FirstWinRec->NEXT;

      freerdswinrec( FirstWinRec );
    }

    if ( DelRecWin != &StaticRecWin )
    {
      freerdsrecwin( DelRecWin );
    }
  }
}

