/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                  rwidestroy.c                     |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>
# include <math.h>

# include MUT_H
# include RDS_H
# include RTL_H
# include RWI_H

# include "rwidestroy.h"
# include "rwierror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Destroy Rds Window                     |
|                                                             |
\------------------------------------------------------------*/

void destroyrdswindow( Figure, RdsWindow )

   rdsfig_list *Figure;
   rdswindow   *RdsWindow;
{
  rdsins_list    *Instance;
  rdsrec_list    *Rectangle;
  rdswin_list    *WinList;
  rdsrecwin_list *RecWin;
  rdsrecwin_list *DelRecWin;
  rdswinrec_list *WinRec;
  rdswinrec_list *DelWinRec;
  long            Counter;
  char            Layer;

  for ( Layer = 0; Layer < RWI_MAX_LAYER; Layer++ )
  {
    for ( Rectangle  = Figure->LAYERTAB[ Layer ];
          Rectangle != (rdsrec_list *)NULL;
          Rectangle  = Rectangle->NEXT )
    {
      if ( ! IsRdsOneWindow( Rectangle ) )
      {
        RecWin = (rdsrecwin_list *)Rectangle->USER;
 
        while ( RecWin != (rdsrecwin_list *)NULL )
        {
          DelRecWin = RecWin;
          RecWin    = RecWin->NEXT;

          freerdsrecwin( DelRecWin );
        }
      }

      Rectangle->USER = (void *)NULL;

      ClearRdsOneWindow( Rectangle );
    }
  }

  for ( Instance  = Figure->INSTANCE; 
        Instance != (rdsins_list *)NULL;
        Instance  = Instance->NEXT )
  {
    for ( Layer = 0; Layer < RWI_MAX_LAYER; Layer++ )
    {
      for ( Rectangle  = Instance->LAYERTAB[ Layer ];
            Rectangle != (rdsrec_list *)NULL;
            Rectangle  = Rectangle->NEXT )
      {
        if ( ! IsRdsOneWindow( Rectangle ) )
        {
          RecWin = (rdsrecwin_list *)Rectangle->USER;
 
          while ( RecWin != (rdsrecwin_list *)NULL )
          {
            DelRecWin = RecWin;
            RecWin    = RecWin->NEXT;

            freerdsrecwin( DelRecWin );
          }
        }

        Rectangle->USER = (void *)NULL;

        ClearRdsOneWindow( Rectangle );
      }
    }
  }

  WinList = RdsWindow->WINTAB;
  
  for ( Counter = 0; Counter < RdsWindow->SIZE; Counter++ )
  {
    if ( WinList->LAYERTAB != (rdswinrec_list **)NULL )
    {
      for ( Layer = 0; Layer < RWI_MAX_LAYER; Layer++ )
      {
        WinRec = WinList->LAYERTAB[ Layer ];
      
        while ( WinRec != (rdswinrec_list *)NULL )
        {
          DelWinRec = WinRec;
          WinRec    = WinRec->NEXT;

          freerdswinrec( DelWinRec );
        }
      }
      
      freerdswinlayer( WinList );
    }

    WinList = WinList + 1;
  }

  freerdswin( RdsWindow->WINTAB );
  freerdswindow( RdsWindow );
}

