/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rwiget.c                        |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>
# include <math.h>

# include MUT_H
# include RDS_H
# include RTL_H
# include RWI_H

# include "rwiget.h"
# include "rwierror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  static chain_list *HEAD_RWICHAIN = (chain_list *)NULL;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         getrdswindow                        |
|                                                             |
\------------------------------------------------------------*/

chain_list *getrdswindow( X, Y, Dx, Dy, RdsWindow )

   long       X;
   long       Y;
   long       Dx;
   long       Dy;
   rdswindow *RdsWindow;
{
  rdswin_list    *WinList;
  long            X1;
  long            X2;
  long            Y1;
  long            Y2;
  long            FirstX;

  if ( HEAD_RWICHAIN != (chain_list *)NULL )
  {
    freechain( HEAD_RWICHAIN );
    HEAD_RWICHAIN = (chain_list *)NULL;
  }

  Y1 = ( Y - 1  - RdsWindow->YMIN ) / RdsWindow->SIDE;
  Y2 = ( Y + Dy - RdsWindow->YMIN ) / RdsWindow->SIDE;
  X1 = ( X - 1  - RdsWindow->XMIN ) / RdsWindow->SIDE;
  X2 = ( X + Dx - RdsWindow->XMIN ) / RdsWindow->SIDE;

  if ( X1 < 0              ) X1 = 0;
  if ( X2 < 0              ) X2 = 0;
  if ( X1 >= RdsWindow->DX ) X1 = RdsWindow->DX - 1;
  if ( X2 >= RdsWindow->DX ) X2 = RdsWindow->DX - 1;

  if ( Y1 < 0              ) Y1 = 0;
  if ( Y2 < 0              ) Y2 = 0;
  if ( Y1 >= RdsWindow->DY ) Y1 = RdsWindow->DY - 1;
  if ( Y2 >= RdsWindow->DY ) Y2 = RdsWindow->DY - 1;

  FirstX = X1;

  /* End Modif 04/10/94 */

  while ( Y1 <= Y2 )
  {
    WinList = RdsWindow->WINTAB + ( Y1 * RdsWindow->DX );

    X1 = FirstX;

    while ( X1 <= X2 )
    {
      allocrdswinlayer( WinList + X1 );

      HEAD_RWICHAIN = addchain( HEAD_RWICHAIN, 
                                (void *)(WinList + X1) );
      X1 = X1 + 1;
    }

    Y1  = Y1 + 1;
  }

  return( HEAD_RWICHAIN );
}

