/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     RDS                           |
|                                                             |
| File    :                   rwiview.c                       |
|                                                             |
| Authors :                Jacomme Ludovic                    |
|                                                             |
| Date    :                   27.06.95                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>
# include <math.h>

# include MUT_H
# include RDS_H
# include RTL_H
# include RWI_H

# include "rwiview.h"
# include "rwierror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        View Rds Window                      |
|                                                             |
\------------------------------------------------------------*/

void viewrdswindow( RdsWindow )

   rdswindow *RdsWindow;
{
  rdswin_list    *ScanWin;
  rdswinrec_list *ScanWinRec;
  rdsrec_list    *Rectangle;
  long            X;
  long            Y;
  long            Xmin;
  long            Ymin;
  long            Xmax;
  long            Ymax;
  long            Counter;
  long            MaskCounter;
  char            Layer;
  char            ScanRec;

  fprintf( stdout, "\n\t--> Window\n" );
  fprintf( stdout, "\n\tXMIN : %d"  , RdsWindow->XMIN );
  fprintf( stdout, "\n\tXMAX : %d"  , RdsWindow->XMAX );
  fprintf( stdout, "\n\tYMIN : %d"  , RdsWindow->YMIN );
  fprintf( stdout, "\n\tYMAX : %d"  , RdsWindow->YMAX );
  fprintf( stdout, "\n\tMASK : %d"  , RdsWindow->MASK );
  fprintf( stdout, "\n\tDX   : %d"  , RdsWindow->DX   );
  fprintf( stdout, "\n\tDY   : %d"  , RdsWindow->DY   );
  fprintf( stdout, "\n\tSIDE : %d"  , RdsWindow->SIDE );
  fprintf( stdout, "\n\tSIZE : %d\n", RdsWindow->SIZE );

  fprintf( stdout, "\n\t--> Dump all windows\n" );

  if ( RdsWindow->SIZE != 0 )
  {
    Ymin = RdsWindow->YMIN;
    Ymax = RdsWindow->SIDE + Ymin;
    Y    = 0;

    MaskCounter = 0;
 
    while ( Y < RdsWindow->DY )
    {
      Xmin = RdsWindow->XMIN;
      Xmax = RdsWindow->SIDE + Xmin;
      X    = 0;

      ScanWin = RdsWindow->WINTAB + ( Y * RdsWindow->DX );

      while ( X < RdsWindow->DX )
      {
        fprintf( stdout, "\n\t--> Window[ %d, %d ]\n", X, Y );

        fprintf( stdout, "\n\tXMIN : %d"  , Xmin );
        fprintf( stdout, "\n\tYMIN : %d"  , Ymin );
        fprintf( stdout, "\n\tXMAX : %d"  , Xmax );
        fprintf( stdout, "\n\tYMAX : %d\n", Ymax );

        Counter = 0;

        if ( ScanWin->LAYERTAB != (rdswinrec_list **)NULL )
        {
          for ( Layer = 0; Layer < RWI_MAX_LAYER; Layer++ )
          {
            for ( ScanWinRec  = ScanWin->LAYERTAB[ Layer ];
                  ScanWinRec != (rdswinrec_list *)NULL;
                  ScanWinRec  = ScanWinRec->NEXT )
            {
              for ( ScanRec = 0; ScanRec < RWI_MAX_REC ; ScanRec++ )
              {
                Rectangle = ScanWinRec->RECTAB[ ScanRec ];
   
                if ( Rectangle != (rdsrec_list *)NULL )
                {
                  if ( ( Rectangle->X < Xmax                 ) &&
                       ( Rectangle->Y < Ymax                 ) &&
                       ( Rectangle->X + Rectangle->DX > Xmin ) &&
                       ( Rectangle->Y + Rectangle->DY > Ymin ) )
                  {
                    Counter = Counter + 1;
                  }
  
                  viewrdsrec( Rectangle );
                }
              }
            }
          }
        }

        MaskCounter = MaskCounter + Counter;

        fprintf( stdout, "\n\t<-- Total rectangles %d", Counter );
        
        Xmin = Xmax;
        Xmax = Xmax + RdsWindow->SIDE;
        X    = X + 1;

        ScanWin = ScanWin + 1;
      }

      Ymin = Ymax;
      Ymax = Ymax + RdsWindow->SIDE;
      Y    = Y + 1;
    }
  }

  fprintf( stdout, "\n<-- Total Rectangles %d\n", MaskCounter );
}
