/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis
 * Date    : 1991,92
 * Author  : Luc Burgun, Pascale Allegre, Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident "$Id: big_main.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"


/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                         Include Files                         ##
 * ##                                                               ##
 * ####===========================================================####
 */


#include <sys/time.h>

#include <unistd.h>		/* for getopt */
extern char *optarg;
extern int optind, opterr, optopt;

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include BVL_H
#include BHL_H

#include "but_error.h"




/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                           Constants                           ##
 * ##                                                               ##
 * ####===========================================================####
 */


/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                             Types                             ##
 * ##                                                               ##
 * ####===========================================================####
 */
/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                           Variables                           ##
 * ##                                                               ##
 * ####===========================================================####
 */

char *BOP_InputFileName;
char *BOP_LaxFileName;
char *BOP_FigureName1;
char *BOP_OutputFileName;
char *BOP_Program_name;


char *GLOB_ABL_DC;
char *GLOB_ABL_ONE;
char *GLOB_ABL_ZERO;


/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                           Functions                           ##
 * ##                                                               ##
 * ####===========================================================####
 */



/*================================================================####
 * Bop_FastUsage 
 *     
 */

void
Bop_FastUsage ()
{
  printf ("Type '%s -h' for help on command line options.\n", BOP_Program_name);

}



/*================================================================####
 * Fpt Usage   
 *     
 */

void
Bop_Usage ()
{
  printf ("\
Usage : \n\
%s [Options] Input_filename(.vbe) Output_filename(.vst|.al) \n\
\n\
Where options are : \n\
         -h       : this help screen \n\
         -1       : logic \n\
         -2       : netoptim \n\
         -3       : proof\n\
", BOP_Program_name);

}



/*================================================================####
 * print_elapsed_time
 *     
 */

void
print_elapsed_time (clock_start, clock_end)

     struct timeval clock_start;
     struct timeval clock_end;

{
  int elapsed_sec;
  int elapsed_usec;

  if (clock_end.tv_sec && clock_start.tv_sec)
    {
      elapsed_sec = clock_end.tv_sec - clock_start.tv_sec;

      if (clock_end.tv_usec < clock_start.tv_usec)
	{
	  --elapsed_sec;
	  elapsed_usec = 1000000 - clock_start.tv_usec + clock_end.tv_usec;
	}
      else
	{
	  elapsed_usec = clock_end.tv_usec - clock_start.tv_usec;
	}

      printf ("%d min %02d.%02d sec",
	      (elapsed_sec / 60),
	      (elapsed_sec % 60),
	      elapsed_usec / 10000);


/* this for me for profiling (giving time in minutes) */
/* 
   min = (double)((double)elapsed_sec + ((double)elapsed_usec / (double)1000000));
   printf ("%.2f sec", min);
 */
    }
  else
    printf ("?");

}



/*================================================================####
 *  Main                        
 *                                                            
 */

int
main (argc, argv)
     int argc;
     char *argv[];
{
  int c;
int start_what=0;





  struct timeval tv_clock_start;
  struct timeval tv_clock_start_abl;
  struct timeval tv_clock_start_bdd;
  struct timeval tv_clock_start_fp;
  struct timeval tv_clock_start_pat;
  struct timeval tv_clock_end;

  time_t clock_start;



  alliancebanner ("Big", "beta", "All you need for synthesis", "1997",  ALLIANCE_VERSION);


  tv_clock_start.tv_sec = 0;
  tv_clock_start_abl.tv_sec = 0;
  tv_clock_start_bdd.tv_sec = 0;
  tv_clock_start_fp.tv_sec = 0;
  tv_clock_start_pat.tv_sec = 0;
  tv_clock_end.tv_sec = 0;

  time (&clock_start);
  gettimeofday (&tv_clock_start, NULL);

  BOP_Program_name = mbkstrdup (argv[0]);

  mbkenv ();



  GLOB_ABL_DC = namealloc ("'d'");
  GLOB_ABL_ONE = namealloc ("'1'");
  GLOB_ABL_ZERO = namealloc ("'0'");


  clearfptverboseflag ();
  clearfptdebugflag ();

#ifdef DEBUG_BOP
  setfptverboseflag ();
  setfptdebugflag ();
#endif

	

 while ((c = getopt (argc, argv, "h123")) != -1) 

    {
      /* printf("Val renv : '%c' pour '%c' \n",c, optopt); */

      switch (c)
	{
	case '1':
	  start_what=1;
	  break;

	case '2':
	  start_what=2;
	  break;

	case '3':
	  start_what=3;
	  break;

	case 'h':
	  Bop_Usage ();
	  EXIT (1);
	  break;
	}
    }

**argv++;
argc--;

      switch (start_what)
	{
	case 1:
	  logic_main(argc, argv);
	  break;

	case 2:
	  net_main(argc, argv);
	  break;

	case 3:
	  proof_main(argc, argv);
	  break;

	default:
	  Bop_FastUsage ();
	  displayfpterror ("Bad command line.\n");
	  break;

	}
	 



  gettimeofday (&tv_clock_end, NULL);

  printf ("===> Started on %s", ctime (&clock_start));

  printf ("     Total elapsed time : ");
  print_elapsed_time (tv_clock_start, tv_clock_end);
  printf ("\n");

  printf ("===> Total number of warning : %d\n\n", getfptwarnings ());
  return (0);

}
