/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis
 * Date    : 1996,97
 * Author  : Czo <Olivier.Sirol@lip6.fr> 
 */



#ident "$Id: but_error.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"


/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                         Include Files                         ##
 * ##                                                               ##
 * ####===========================================================####
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <string.h>

#include <varargs.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include BVL_H



#include "but_error.h"


/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                           Constants                           ##
 * ##                                                               ##
 * ####===========================================================####
 */
/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                             Types                             ##
 * ##                                                               ##
 * ####===========================================================####
 */
/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                           Variables                           ##
 * ##                                                               ##
 * ####===========================================================####
 */

static int FPT_Verbose_Flag = 0;
static int FPT_Debug_Flag = 0;
static int FPT_Number_of_Warnings = 0;
#define BIG_STATIC_BUFFER_SIZE 8192
static char BIG_STATIC_BUFFER[BIG_STATIC_BUFFER_SIZE];

/*
 * ####===========================================================####
 * ##                                                               ##
 * ##                           Functions                           ##
 * ##                                                               ##
 * ####===========================================================####
 */



/*================================================================####
 * Set or Clear global Flags
 *     
 */

void
setfptverboseflag ()

{
  FPT_Verbose_Flag = 1;
}

void
clearfptverboseflag ()

{
  FPT_Verbose_Flag = 0;
}

void
setfptdebugflag ()

{
  FPT_Debug_Flag = 1;
}

void
clearfptdebugflag ()

{
  FPT_Debug_Flag = 0;
}


int
getfptverboseflag ()

{
  return (FPT_Verbose_Flag);
}

int
getfptdebugflag ()

{
  return (FPT_Debug_Flag);
}


int
getfptwarnings ()

{
  return (FPT_Number_of_Warnings);
}



/***************************************************************
    C'est la nouvelle ecriture (ansi) pour une fonction 
    avec nb variable darguments

# include <stdarg.h>

# ifdef __STDC__

void
FptDisplayError (char *fmt, ...)

{
  va_list args;

  va_start(args, fmt);
  fprintf (stderr, "## Error : ");
  vfprintf (stderr, fmt, args);
  fprintf (stderr, "\n");
  va_end(args);

  EXIT(1);
}

# else
# endif
********************************************************************/




/*================================================================####
 * FptDisplayDebugSource
 *     
 */

void
FptDisplayDebugSource (File, Line)
     char *File;
     int Line;

{
  if (FPT_Debug_Flag)
    fprintf (stderr, "DEBUG FILE ==> file '%s', line %d\n", File, Line);
  fflush (stderr);
}




/*================================================================####
 * FptDisplayDebug
 *     
 */

void
FptDisplayDebug (va_alist)
     va_dcl

{
  va_list args;
  char *fmt;

  if (FPT_Debug_Flag)
    {

      va_start (args);
      fmt = va_arg (args, char *);
      fprintf (stdout, "DEBUG INFO ==> ");
      vfprintf (stdout, fmt, args);
      va_end (args);
      fflush (stdout);
    }
}




/*================================================================####
 * FptDisplayVerbose
 *     
 */

void
FptDisplayVerbose (va_alist)
     va_dcl

{
  va_list args;
  char *fmt;

  if (FPT_Verbose_Flag)
    {

      va_start (args);
      fmt = va_arg (args, char *);
      /* fprintf (stdout, "==> "); */
      vfprintf (stdout, fmt, args);
      va_end (args);
      fflush (stdout);
    }
}

/*================================================================####
 * FptDisplayWarning
 *     
 */

void
FptDisplayWarning (va_alist)
     va_dcl

{
  va_list args;
  char *fmt;

  FPT_Number_of_Warnings++;
  va_start (args);
  fmt = va_arg (args, char *);
  fprintf (stderr, "## WARNING ## ");
  vfprintf (stderr, fmt, args);
  va_end (args);
  fflush (stderr);
}


/*================================================================####
 * FptDisplayError
 *     
 */

void
FptDisplayError (va_alist)
     va_dcl

{
  va_list args;
  char *fmt;

  va_start (args);
  fmt = va_arg (args, char *);
  fprintf (stderr, "##  ERROR  ## ");
  vfprintf (stderr, fmt, args);
  va_end (args);
  fflush (stderr);

  EXIT (1);
}

/*================================================================####
 *     
 *     
 */

char *
FptSprintf (va_alist)
     va_dcl

{
  va_list args;
  char *fmt;
  char *result;
  int len;

  strcpy (BIG_STATIC_BUFFER, "");

  va_start (args);
  fmt = va_arg (args, char *);
  vsprintf (BIG_STATIC_BUFFER, fmt, args);
  va_end (args);

  len = strlen (BIG_STATIC_BUFFER);
  if (len < BIG_STATIC_BUFFER_SIZE)
    {
      result = mbkalloc (len + 1);
      strcpy (result, BIG_STATIC_BUFFER);
    }
  else
    {
      fprintf (stderr, "##  ERROR  ## Sprintf was too long\n");
      fflush (stderr);
      EXIT (1);
    }
  return result;

}
