/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - CMOS Cells Compiler
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */






/*
   Produit :  Compilateur de cellules (Caracterisation)                 
   Auteur(s) :  K. Dioury                            le : 06/11/1995    
   Modifie par : N. Dictus                           le : 19/12/1995    



             Fonction de caracterisation de cellule        
                compilee  C4  en tps, R et C               
                                                           
                                                           
   entrees: fichier de techno physiques (.rds)             
            fichier de techno electrique (.elp)            
            pointeur sur la lofig a evaluer                
            front des entrees en pico secondes             
            capa en fento farad                            
                                                           
   sortie : pointeur sur la structure resultat             
                                                           
   appel :  cara_cell(name,capa,front,elp_techno_name)     
                                                           
   variables d'environnement:                              
                                                           

   RDS_TECHNO_NAME  : chemin du fichier technologique .rds 
   ELP_TECHNO_NAME  : chemin du fichier technologique .elp 
*/                                                           

#ident  "$Id: c4_caraccell.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include MUT_H
#include BEH_H
#include LOG_H

#include ELP_H
#include MLO_H
#include MLU_H
#include MPH_H
#include MUT_H
#include RDS_H
#include RUT_H
#include RTL_H
#include RPR_H
#include RFM_H

#include "c4_local.h"

/********************************************
 **** constantes et structure de donnees ****
 ********************************************/

#define MAX_TRANS 10
#define NAME_SOURCE "source"
#define NAME_DRAIN  "drain"
#define NAME_GRID "grid"
#define VDD_CARAC "vdd"
#define VSS_CARAC "vss"



#define EPSILON        0.001

/*
 * #define grid_sup(i)    (anint(((i)+GRID/2-EPSILON)/GRID)*GRID)
 * #define grid_inf(i)    (((int)(anint((i)/GRID)))*GRID)
 */

#define grid_sup(i)    (floor((((i)+GRID/2-EPSILON)/GRID)*GRID)+0.5)
#define grid_inf(i)    (((int)(floor(((i)/GRID)))*GRID)+0.5)

#define CCELL_MODE_DEBUG 'N'

/*******************************************
**** declaration des variables globales ****
*******************************************/
lores_list *HEAD_LORES = NULL;	/* tete de la structure resultat */
lofig_list *HEAD_LOFIG_RES = NULL;	/* tete de la lofig resultat */
chain_list *SIG_CHAIN = NULL;

int NB_TRANS;
long INDEX_VDD;
long INDEX_VSS;
long INDEX_OUT;

int NTRANS_VW_DLR, NTRANS_VW_DWR, NTRANS_LCW_DLR, NTRANS_LCW_DWR, NTRANS_RCW_DLR,
  NTRANS_RCW_DWR, PTRANS_VW_DLR, PTRANS_VW_DWR, PTRANS_LCW_DLR, PTRANS_LCW_DWR,
  PTRANS_RCW_DLR, PTRANS_RCW_DWR;


float GRID;

double
/* shrink parameters */
  LMLT, WMLT, DWN, DLN, DWP, DLP,	/* transistor characteristics */
  VTN, BN, AN, RNT, VTP, BP, AP, RPT,	/* general parameters */
  VDDmax, VTHR, VSSdeg, VDDdeg, TEMP,	/* dynamic capacitance: grid  */
  CGSN, CGSP, CGPN, CGPP,	/* dynamic capacitance: drain  */
  CDSN, CDSP, CDPN, CDPP, CDWN, CDWP,	/* dynamic capacitance: source */
  CSSN, CSSP, CSPN, CSPP, CSWN, CSWP;

/* pour eviter de passer n fois par compilation des fichiers techno 22/01/96 */
int PREMIER = 1;

/****************************************************
****         Fonction qui recupere les           ****
****    parametres RDS dans le fichier .rds      ****
****                                             ****   
****  entree : par defaut la fonction prend la   **** 
****           variable d'environnement          ****
****           RDS_TECHNO_NAME                   ****
****                                             ****
**** sortie : aucune                             ****
****************************************************/
void
TechnoParamrds ()
{
  int Index;
  int MbkLayer;
  int GateLayer;
  int DiffLayer;
  int RdsLayer;
  int RdsUse;
  int RdsDir;

  loadrdsparam ();

  for (MbkLayer = 0; MbkLayer < MBK_MAX_LAYER; MbkLayer++)
    {
      if (GET_LYNX_TRANSISTOR_TYPE (MbkLayer) == NTRANS)
	{
	  GateLayer = GET_LYNX_TRANSISTOR_GATE_LAYER (MbkLayer);
	  DiffLayer = GET_LYNX_TRANSISTOR_DIFF_LAYER (MbkLayer);
	  Index = 0;
	  while ((RdsLayer = GET_SEGMENT_LAYER (MbkLayer, Index)) != RDS_SEGMENT_EMPTY)
	    {
	      RdsUse = GET_SEGMENT_USE (MbkLayer, Index);
	      RdsDir = GET_SEGMENT_TRANS (MbkLayer, Index);

	      if ((RdsUse == RDS_USE_ALL) || (RdsUse == RDS_USE_EXTRACT))
		{
		  if (RdsLayer == GateLayer)
		    {
		      NTRANS_VW_DLR = GET_SEGMENT_DLR (MbkLayer, Index);
		      NTRANS_VW_DWR = GET_SEGMENT_DWR (MbkLayer, Index);
		    }
		  else if ((RdsLayer == DiffLayer) && (RdsDir == RDS_TRANS_LCW))
		    {
		      NTRANS_LCW_DLR = GET_SEGMENT_DLR (MbkLayer, Index);
		      NTRANS_LCW_DWR = GET_SEGMENT_DWR (MbkLayer, Index);
		    }
		  else if ((RdsLayer == DiffLayer) && (RdsDir == RDS_TRANS_RCW))
		    {
		      NTRANS_RCW_DLR = GET_SEGMENT_DLR (MbkLayer, Index);
		      NTRANS_RCW_DWR = GET_SEGMENT_DWR (MbkLayer, Index);
		    }
		}
	      Index++;
	    }
	}
      if (GET_LYNX_TRANSISTOR_TYPE (MbkLayer) == PTRANS)
	{
	  GateLayer = GET_LYNX_TRANSISTOR_GATE_LAYER (MbkLayer);
	  DiffLayer = GET_LYNX_TRANSISTOR_DIFF_LAYER (MbkLayer);

	  Index = 0;
	  while ((RdsLayer = GET_SEGMENT_LAYER (MbkLayer, Index)) != RDS_SEGMENT_EMPTY)
	    {
	      RdsUse = GET_SEGMENT_USE (MbkLayer, Index);
	      RdsDir = GET_SEGMENT_TRANS (MbkLayer, Index);

	      if ((RdsUse == RDS_USE_ALL) || (RdsUse == RDS_USE_EXTRACT))
		{
		  if (RdsLayer == GateLayer)
		    {
		      PTRANS_VW_DLR = GET_SEGMENT_DLR (MbkLayer, Index);
		      PTRANS_VW_DWR = GET_SEGMENT_DWR (MbkLayer, Index);
		    }
		  else if ((RdsLayer == DiffLayer) && (RdsDir == RDS_TRANS_LCW))
		    {
		      PTRANS_LCW_DLR = GET_SEGMENT_DLR (MbkLayer, Index);
		      PTRANS_LCW_DWR = GET_SEGMENT_DWR (MbkLayer, Index);
		    }
		  else if ((RdsLayer == DiffLayer) && (RdsDir == RDS_TRANS_RCW))
		    {
		      PTRANS_RCW_DLR = GET_SEGMENT_DLR (MbkLayer, Index);
		      PTRANS_RCW_DWR = GET_SEGMENT_DWR (MbkLayer, Index);
		    }
		}
	      Index++;
	    }
	}
    }

  GRID = (float) RDS_PHYSICAL_GRID;

   if (CCELL_MODE_DEBUG == 'Y')
   {
   fprintf(stdout,"Parametres Techmologiques (.rds)\n\n");
   fprintf(stdout,"NTRANS : VW  : %d %d\n",NTRANS_VW_DLR,NTRANS_VW_DWR);
   fprintf(stdout,"         LCW : %d %d\n",NTRANS_LCW_DLR,NTRANS_LCW_DWR);
   fprintf(stdout,"         RCW : %d %d\n\n",NTRANS_RCW_DLR,NTRANS_RCW_DWR);
   fprintf(stdout,"PTRANS : VW  : %d %d\n",PTRANS_VW_DLR,PTRANS_VW_DWR);
   fprintf(stdout,"         LCW : %d %d\n",PTRANS_LCW_DLR,PTRANS_LCW_DWR);
   fprintf(stdout,"         RCW : %d %d\n\n",PTRANS_RCW_DLR,PTRANS_RCW_DWR);
   fprintf(stdout,"valeur de RDS_LAMBDA %ld\n",RDS_LAMBDA);  
   fprintf(stdout,"valeur de RDS_UNIT %ld\n",RDS_UNIT);  
   fprintf(stdout,"valeur de LAMBDA %.3f\n",(float)RDS_LAMBDA/RDS_UNIT);  
   fprintf(stdout,"valeur de la grille physique %f\n\n",GRID);  
   }
 
}

/**************************************************
****       Fonction qui recupere les           ****
****     parametres electrique dans le         ****
****            fichier .elp                   ****   
****                                           ****
**** entree : nom du fichier technologique     ****
**** sortie : aucune                           ****
**************************************************/
void
TechnoParamelp ()
{
  elpenv ();
  if (elpLoadElp ())
    {
      fprintf (stderr, "####--- Erreur Chargement fichier elp ---####\n");
      exit (1);
    }
  DLN = elpShrink[elpDLN];
  DLP = elpShrink[elpDLP];
  DWN = elpShrink[elpDWN];
  DWP = elpShrink[elpDWP];
  LMLT = elpShrink[elpLMLT];
  WMLT = elpShrink[elpWMLT];

  VTN = elpModel[elpNMOS][elpVT];
  AN = elpModel[elpNMOS][elpA];
  BN = elpModel[elpNMOS][elpB];
  RNT = elpModel[elpNMOS][elpRT];

  VTP = elpModel[elpPMOS][elpVT];
  AP = elpModel[elpPMOS][elpA];
  BP = elpModel[elpPMOS][elpB];
  RPT = elpModel[elpPMOS][elpRT];

  VDDmax = elpVoltage[elpVDDMAX];
  VTHR = elpVoltage[elpVTHR];
  VSSdeg = elpVoltage[elpVSSDEG];
  VDDdeg = elpVoltage[elpVDDDEG];
  TEMP = elpGeneral[elpTEMP];

  CGSN = elpCapa[elpNMOS][elpCGS];
  CGSP = elpCapa[elpPMOS][elpCGS];
  CGPN = elpCapa[elpNMOS][elpCGP];
  CGPP = elpCapa[elpPMOS][elpCGP];

  CDSN = elpCapa[elpNMOS][elpCDS];
  CDSP = elpCapa[elpPMOS][elpCDS];
  CDPN = elpCapa[elpNMOS][elpCDP];
  CDPP = elpCapa[elpPMOS][elpCDP];
  CDWN = elpCapa[elpNMOS][elpCDW];
  CDWP = elpCapa[elpPMOS][elpCDW];

  CSSN = elpCapa[elpNMOS][elpCSS];
  CSSP = elpCapa[elpPMOS][elpCSS];
  CSPN = elpCapa[elpNMOS][elpCSP];
  CSPP = elpCapa[elpPMOS][elpCSP];
  CSWN = elpCapa[elpNMOS][elpCSW];
  CSWP = elpCapa[elpPMOS][elpCSW];
}

/*********************************************
****    Fonction qui modifie la lofig     ****
****   en fonction des caracteristiques   ****
****       electriques (.elp)             ****
****                                      ****
**** entree : lofig a modifier            ****
**** sortie : aucune                      **** 
*********************************************/
void
param_cell_elp (ptlofig)
     lofig_list *ptlofig;
{
  elpLofigShrink (ptlofig);	/* modification des dimansions des transistors */
  elpLofigCapaGrid (ptlofig);	/* calcule des capas de grille */
  elpLofigCapaDiff (ptlofig);	/* calcule des capas de diffusion */
}

/*********************************************
****    Fonction qui modifie la lofig     ****
****   en fonction des caracteristiques   ****
****       physiques (.rds)               ****
****                                      ****
**** entree : lofig a modifier            ****
**** sortie : aucune                      **** 
*********************************************/
void
param_cell_rds (ptlofig)
     lofig_list *ptlofig;
{
  float length, lambda;
  lotrs_list *pttrs;

  lambda = (float) RDS_LAMBDA / RDS_UNIT;
  for (pttrs = ptlofig->LOTRS; pttrs; pttrs = pttrs->NEXT)
    {
      if (pttrs->TYPE == TRANSN)
	{
	  pttrs->WIDTH = pttrs->WIDTH + (3 * SCALE_X);
	  pttrs->WIDTH = (pttrs->WIDTH) * lambda
	    + 2 * (float) ((NTRANS_LCW_DLR * SCALE_X) / RDS_UNIT);
	  length = ((pttrs->LENGTH) * lambda
		+ (float) ((NTRANS_VW_DWR * SCALE_X) / RDS_UNIT)) / SCALE_X;
	  pttrs->LENGTH = (2 * (grid_sup ((length / 2) * RDS_UNIT) / RDS_UNIT)) * SCALE_X;
	  pttrs->XS = (short) (grid_sup (NTRANS_LCW_DWR / 2) * GRID / RDS_UNIT * SCALE_X);
	  pttrs->XD = (short) (grid_sup (NTRANS_RCW_DWR / 2) * GRID / RDS_UNIT * SCALE_X);
	  pttrs->PD = 2 * pttrs->WIDTH + 2 * pttrs->XD;
	  pttrs->PS = 2 * pttrs->WIDTH + 2 * pttrs->XS;
	}
      else
	{
	  pttrs->WIDTH = pttrs->WIDTH + (3 * SCALE_X);
	  pttrs->WIDTH = (pttrs->WIDTH) * lambda
	    + 2 * (float) ((PTRANS_LCW_DLR * SCALE_X) / RDS_UNIT);
	  length = ((pttrs->LENGTH) * lambda
		+ (float) ((PTRANS_VW_DWR * SCALE_X) / RDS_UNIT)) / SCALE_X;
	  pttrs->LENGTH = (2 * (grid_sup ((length / 2) * RDS_UNIT) / RDS_UNIT)) * SCALE_X;
	  pttrs->XS = (short) (grid_sup (PTRANS_LCW_DWR / 2) * GRID / RDS_UNIT * SCALE_X);
	  pttrs->XD = (short) (grid_sup (PTRANS_RCW_DWR / 2) * GRID / RDS_UNIT * SCALE_X);
	  pttrs->PD = 2 * pttrs->WIDTH + 2 * pttrs->XD;
	  pttrs->PS = 2 * pttrs->WIDTH + 2 * pttrs->XS;
	}
    }
}


/************************************************
****         fonction makepath               ****
****   desasemble la cellule en branches     ****
****         et maillons                     ****
************************************************/
ptype_list *
makepath (ptsig)
     losig_list *ptsig;
{
  chain_list *ptchain = NULL;
  chain_list *pt2_chain = NULL;
  ptype_list *ptchain1 = NULL;
  ptype_list *ptlist = NULL;
  ptype_list *result_path = NULL;
  ptype_list *pt_path = NULL;
  locon_list *ptcon = NULL;
  lotrs_list *pttrs = NULL;
  losig_list *pt_nextsig = NULL;
  int capa;

  NB_TRANS++;
  if (NB_TRANS > (MAX_TRANS + 1))
    {
      NB_TRANS--;
      return (NULL);
    }

  capa = (long) (ptsig->CAPA * 1000);
  for (ptchain = (chain_list *) getptype (ptsig->USER, LOFIGCHAIN)->DATA;
       ptchain; ptchain = ptchain->NEXT)
    {
      pt_nextsig = NULL;
      ptchain1 = NULL;
      pt2_chain = NULL;
      ptcon = NULL;
      pttrs = NULL;

      ptcon = (locon_list *) ptchain->DATA;

      if (ptcon->TYPE == 'E')
	continue;


      else if (ptcon->TYPE == 'T')
	{
	  pttrs = (lotrs_list *) (ptcon->ROOT);

	  /* on teste l'autoreferentialite */
	  for (pt2_chain = SIG_CHAIN; pt2_chain; pt2_chain = pt2_chain->NEXT)
	    {
	      if (pt2_chain->DATA == (void *) (pttrs->GRID->SIG))
		break;
	    }
	  if (pt2_chain != NULL)
	    /* le signal de grille du transistor que l'on veut
	       traverser a deja ete rencontre */
	    continue;


	  /* on determine le prochain signal a etudier */
	  if (strcmp (ptcon->NAME, NAME_SOURCE) == 0)
	    {
	      pt_nextsig = pttrs->DRAIN->SIG;
	    }
	  else if (strcmp (ptcon->NAME, NAME_DRAIN) == 0)
	    {
	      pt_nextsig = pttrs->SOURCE->SIG;
	    }
	  else if (strcmp (ptcon->NAME, NAME_GRID) == 0)
	    {
	      /* si c'est une grille on passe au connecteur suivant */
	      continue;
	    }
	  else
	    {
	      fprintf (stderr, "####--- Erreur makepath ---####\n");
	      exit (1);
	    }


	  /* on verifie que le signal n'a pas deja ete etudier */
	  for (pt2_chain = SIG_CHAIN; pt2_chain; pt2_chain = pt2_chain->NEXT)
	    {
	      if (pt2_chain->DATA == (void *) pt_nextsig)
		break;
	    }
	  if (pt2_chain != NULL)
	    /* si le signal a deja ete etudier on passe
	       au connecteur suivant */
	    continue;

	  else if ((pt_nextsig->INDEX == INDEX_VDD) || (pt_nextsig->INDEX == INDEX_VSS))
	    {
	      /* si le signal relie a la source est une alimentation on 
	         rajoute un chemin avec un seul maillon  */
	      ptchain1 = addptype (ptchain1, capa, (lotrs_list *) pttrs);
	      result_path = addptype (result_path, 0, (chain_list *) ptchain1);
	    }

	  else
	    {
	      /* sinon on appelle makepath recursivement avec le nouveau signal */
	      SIG_CHAIN = addchain (SIG_CHAIN, (void *) ptsig);
	      pt_path = makepath (pt_nextsig);
	      for (ptlist = pt_path; ptlist; ptlist = ptlist->NEXT)
		{
		  ptlist->DATA = addptype (ptlist->DATA, capa, (lotrs_list *) pttrs);
		}

	      result_path = (ptype_list *) append (result_path, pt_path);
	      SIG_CHAIN = delchain (SIG_CHAIN, SIG_CHAIN);
	    }
	}
      else
	{
	  fprintf (stderr, "####--- Erreur makepath ---####\n");
	  exit (1);
	}

    }				/* for */
  NB_TRANS--;
  return (result_path);
}

/************************************************
****   fonction qui parcoure la locon        ****
****    et cree la liste des branches        ****
************************************************/
ptype_list *
createbranch (ptlofig)
     lofig_list *ptlofig;
{
  locon_list *ptlocon;
  losig_list *ptsig;
  int type = 0;
  lotrs_list *pttrs;

  lofigchain (ptlofig);


   if (CCELL_MODE_DEBUG == 'Y')
   {
   for(pttrs=ptlofig->LOTRS;pttrs;pttrs=pttrs->NEXT)
   {
   fprintf(stdout,"CARACTERISQUES TRANSISTORS:\n");
   fprintf(stdout,"\t\tindex grille:%ld index source:%ld index drain:%ld\n"
   ,pttrs->GRID->SIG->INDEX
   ,pttrs->SOURCE->SIG->INDEX
   ,pttrs->DRAIN->SIG->INDEX);
   fprintf(stdout,"\t\twidth:%d length:%d\n",pttrs->WIDTH,pttrs->LENGTH);
   fprintf(stdout,"\t\tcapa drain:%f (pf)\n",pttrs->DRAIN->SIG->CAPA);
   fprintf(stdout,"\t\tcapa grille:%f (pf)\n",pttrs->GRID->SIG->CAPA);
   fprintf(stdout,"\t\tcapa source:%f (pf)\n\n",pttrs->SOURCE->SIG->CAPA);
   }
   }
 

  /* determination des index des alimentations ainsi que du signal
     de sortie */
  for (ptlocon = ptlofig->LOCON; ptlocon; ptlocon = ptlocon->NEXT)
    {
      if (strcmp (ptlocon->NAME, VDD_CARAC) == 0)
	INDEX_VDD = ptlocon->SIG->INDEX;
      if (strcmp (ptlocon->NAME, VSS_CARAC) == 0)
	INDEX_VSS = ptlocon->SIG->INDEX;
      if (ptlocon->DIRECTION == 'O')
	{
	  type = 1;
	  INDEX_OUT = ptlocon->SIG->INDEX;
	}
    }
  if (type == 0)
    {
      fprintf (stderr, "###--- Erreur manque direction sur le signal de sortie ---###\n");
      exit (1);
    }

  for (ptsig = ptlofig->LOSIG; ptsig; ptsig = ptsig->NEXT)
    if (ptsig->INDEX == INDEX_OUT)
      return (ptype_list *) makepath (ptsig);
 
 fprintf(stderr,"## ERROR %s.%d## bad locon?\n",__FILE__,__LINE__);
 exit(23);

}

/************************************************
****        fonction findlocon               ****
************************************************/
locon_list *
findlocon (index)
     long index;
{
  locon_list *ptlocon;
  lofig_list *ptlofig;

  ptlofig = HEAD_LOFIG_RES;

  for (ptlocon = ptlofig->LOCON; ptlocon; ptlocon = ptlocon->NEXT)
    if (ptlocon->SIG->INDEX == index)
      return (ptlocon);

  return NULL;
}

/************************************************
****   fonction qui parcoure la locon        ****
****    et cree le lien entre chaques        ****
****    connecteurs et les branches          ****
************************************************/
void
linklocon (pttype)
     ptype_list *pttype;
{
  locon_list *ptlocon;
  ptype_list *pttype2;
  chain_list *ptchain;
  float res_branch;

  while (pttype)
    {
      res_branch = 0;
      for (pttype2 = (ptype_list *) pttype->DATA; pttype2; pttype2 = pttype2->NEXT)
	{
	  res_branch += (float) (((lotrs_list *) (pttype2->DATA))->LENGTH) /
	    ((lotrs_list *) (pttype2->DATA))->WIDTH;
	  ptlocon = findlocon (((lotrs_list *) (pttype2->DATA))->GRID->SIG->INDEX);
	  if (ptlocon == NULL)
	    {
	      fprintf (stderr, "###--- Erreur cellule ---###\n\n");
	      exit (1);
	    }
	  if (ptlocon->USER == NULL)
	    {
	      ptchain = NULL;
	      ptchain = addchain (ptchain, pttype);
	      ptlocon->USER = addptype (ptlocon->USER, LOFIGCHAIN, (chain_list *) ptchain);
	    }
	  else
	    ptlocon->USER->DATA = addchain (ptlocon->USER->DATA, pttype);
	}
      pttype->TYPE = (long) (res_branch * 1000);
      pttype = pttype->NEXT;
    }
}
/************************************************
****        fonction addlores                ****
************************************************/
lores_list *
addlores (name, rup, rdown, capa, tphl, tplh, ptlocon)
     char *name;
     long rup, rdown, capa, tphl, tplh;
     locon_list *ptlocon;
{
  lores_list *ptres;

  ptres = HEAD_LORES;

  ptres = (lores_list *) mbkalloc (sizeof (lores_list));
  ptres->NAME = name;
  ptres->RUP = rup;
  ptres->RDOWN = rdown;
  ptres->CAPA = capa;
  ptres->TPHL = tphl;
  ptres->TPLH = tplh;
  ptres->ptLOCON = ptlocon;

  ptres->NEXT = HEAD_LORES;
  HEAD_LORES = ptres;

  return ptres;
}

/*******************************************************
****            fonction veriflores()                ****
****  Verification des caracteristiques de la cellule  ****
*******************************************************/
void
veriflores (lofig, res)
     lofig_list *lofig;
     lores_list *res;
{
  lores_list *ptres;

  for (ptres = res; ptres; ptres = ptres->NEXT)
    {
      if (ptres->CAPA < 1)
	{
	  fprintf (stdout, "\t[%s] Capa of %s   : %ld (fF)\n",
		   lofig->NAME, ptres->NAME, ptres->CAPA);
/*
   ptres->CAPA = 49;
 */
	}
      if (ptres->RUP < 1)
	fprintf (stdout, "\t[%s] Rup of %s   : %ld (Ohms)\n",
		 lofig->NAME, ptres->NAME, ptres->RUP);
      if (ptres->RDOWN < 1)
	fprintf (stdout, "\t[%s] Rdown of %s : %ld (Ohms)\n",
		 lofig->NAME, ptres->NAME, ptres->RDOWN);
      if (ptres->TPHL < 1)
	fprintf (stdout, "\t[%s] Tphl of %s  : %ld (ps)\n",
		 lofig->NAME, ptres->NAME, ptres->TPHL);
      if (ptres->TPLH < 1)
	fprintf (stdout, "\t[%s] Tplh of %s  : %ld (ps)\n",
		 lofig->NAME, ptres->NAME, ptres->TPLH);
    }
}

/*******************************************************
****            fonction viewlores()                ****
****  Affichage des caracteristiques de la cellule  ****
*******************************************************/
void
viewlores (lofig, res)
     lofig_list *lofig;
     lores_list *res;
{
  lores_list *ptres;

  fprintf (stdout, "CARACTERISTIQUES DE LA CELLULE : %s\n", lofig->NAME);
  for (ptres = res; ptres; ptres = ptres->NEXT)
    {
      fprintf (stdout, "   Entree : %s\n", ptres->NAME);
      fprintf (stdout, "\tCapa   : %ld (fF)\n", ptres->CAPA);
      fprintf (stdout, "\tRup    : %ld (Ohms)\n", ptres->RUP);
      fprintf (stdout, "\tRdown  : %ld (Ohms)\n", ptres->RDOWN);
      fprintf (stdout, "\tTphl   : %ld (ps)\n", ptres->TPHL);
      fprintf (stdout, "\tTplh   : %ld (ps)\n", ptres->TPLH);
      fprintf (stdout, "\n");
/* viewloinscon(ptres->ptLOCON);   */
    }
}

/*******************************************************
****       fonction de calcul du courant            **** 
****             dans une branche                   ****
****                                                ****
**** entree : maillon a partir duquel               ****
****          on calcule le courant                 ****
**** sortie : imax                                  ****
*******************************************************/
double
i_branche (maillon0)
     ptype_list *maillon0;
{

  ptype_list *maillonc;
  double i0, bx = 0.0, Rtr = 0.0, Rs = 0.0, imax = 0.0, k;
  double ugs, uds, usat, fsat, fi, ui, vi, Ri, ut, vth, vtmax, a, b;
  /* ui */
  /* ui-1 c'est vi */


  if (((lotrs_list *) (maillon0->DATA))->TYPE == TRANSN)
    {
      ut = VTN;
      a = AN;
      b = BN;
      bx = (2 * (VDDmax - VTN) * VDDmax - VDDmax * VDDmax) / ((VDDmax - VTN) * (VDDmax - VTN));
      i0 = AN * (VDDmax - VTN) * (VDDmax - VTN);
      i0 /= (1.0 + BN * (VDDmax - VTN));
      k = (RNT * i0) / (VDDmax - VTN);
      vtmax = VDDmax - VDDdeg;
      vth = (vtmax - VTN) / (VDDmax - VTN);
    }

  else if (((lotrs_list *) (maillon0->DATA))->TYPE == TRANSP)
    {
      ut = VTP;
      a = AP;
      b = BP;
      bx = (2 * (VDDmax - VTP) * VDDmax - VDDmax * VDDmax) / ((VDDmax - VTP) * (VDDmax - VTP));
      i0 = AP * (VDDmax - VTP) * (VDDmax - VTP);
      i0 /= (1.0 + BP * (VDDmax - VTP));
      k = (RPT * i0) / (VDDmax - VTP);
      vtmax = VSSdeg;
      vth = (vtmax - VTP) / (VDDmax - VTP);
    }
  else
    return (0);

  Rs = 0.0;
  Rtr = 0.0;
  for (maillonc = maillon0; maillonc; maillonc = maillonc->NEXT)
    {
      Rtr = (double) ((lotrs_list *) (maillonc->DATA))->LENGTH /
	(double) ((lotrs_list *) (maillonc->DATA))->WIDTH;

      Rs += Rtr;
    }

  Ri = Rs;
  ui = VDDmax;
  vi = 0.0;
  imax = 0.0;
  for (maillonc = maillon0; maillonc; maillonc = maillonc->NEXT)
    {
      if (maillonc != maillon0)
	{
	  ui = (VDDmax - ut) * (1.0 - sqrt (1.0 - (Ri / Rs) * bx));
	}
      Rtr = (double) ((lotrs_list *) (maillonc->DATA))->LENGTH /
	(double) ((lotrs_list *) (maillonc->DATA))->WIDTH;
      Ri -= Rtr;
      vi = (VDDmax - ut) * (1.0 - sqrt (1.0 - (Ri / Rs) * bx));
      ugs = (VDDmax - vi - ut);
      uds = ui - vi;
      usat = k * ugs;
      fsat = (a * ugs * ugs) / (1.0 + (b * ugs));
      fi = fsat;
      if (uds < usat)
	fi *= (uds / usat);
      imax += fi / Rs;

    }
  return (imax);
}

/***************************************************
****   fonction de recherche de la branche      ****   
****   la plus resistive suivant le type        ****   
***************************************************/
ptype_list *
max_res (ptlocon, type)
     locon_list *ptlocon;
     int type;
{
  long res_tmp = 0, res = 0;
  chain_list *ptchain;
  ptype_list *pt;

  ptchain = (chain_list *) getptype ((ptype_list *) ptlocon->USER, LOFIGCHAIN)->DATA;
  while (ptchain)
    {
      if (((lotrs_list *) (((ptype_list *) (((ptype_list *) (ptchain->DATA))
					    ->DATA))->DATA))->TYPE == type)
	{
	  res_tmp = ((ptype_list *) (ptchain->DATA))->TYPE;
	  if (res_tmp > res)
	    {
	      res = res_tmp;
	      pt = (ptype_list *) ptchain->DATA;
	    }
	}
      ptchain = ptchain->NEXT;
    }
  return ((ptype_list *) pt);
}

/***************************************************
****   fonction de calcul de la capacite        ****   
***************************************************/
long
calc_capa (ptlocon, type)
     locon_list *ptlocon;
     int type;
{
  double Cp = 0.0;
  double Cn = 0.0;
  long Cs;
  long index;
  ptype_list *pt, *pt2;
  chain_list *ptchain;

  index = ((losig_list *) (ptlocon->SIG))->INDEX;
  ptchain = (chain_list *) getptype ((ptype_list *) ptlocon->USER, LOFIGCHAIN)->DATA;

  if (type == TRANSN)
    {
      pt = (ptype_list *) max_res (ptlocon, TRANSN);
    }
  else
    {
      pt = (ptype_list *) max_res (ptlocon, TRANSP);
    }
  Cs = ((ptype_list *) (pt->DATA))->TYPE;
  for (pt2 = (ptype_list *) pt->DATA; pt2; pt2 = pt2->NEXT)
    {
      if (((lotrs_list *) (pt2->DATA))->GRID->SIG->INDEX == index)
	break;
      Cn = Cn + (VDDdeg / VDDmax) * ((ptype_list *) (pt2->NEXT))->TYPE *
	i_branche (((ptype_list *) (pt->DATA))) /
	i_branche ((ptype_list *) pt2->NEXT);
    }
  while (ptchain)
    {
      pt = (ptype_list *) ptchain->DATA;
      if (type == TRANSN)
	{
	  if (((lotrs_list *) (((ptype_list *) (pt->DATA))->DATA))->TYPE == TRANSP)
	    for (pt2 = (ptype_list *) pt->DATA; pt2; pt2 = pt2->NEXT)
	      {
		if (((lotrs_list *) (pt2->DATA))->GRID->SIG->INDEX == index)
		  break;
		Cp = Cp + ((VDDmax - VSSdeg) / VDDmax) * (pt2->NEXT->TYPE) *
		  i_branche (((ptype_list *) (pt->DATA))) /
		  i_branche (pt2->NEXT);
	      }
	}
      else
	{
	  if (((lotrs_list *) (((ptype_list *) (pt->DATA))->DATA))->TYPE == TRANSN)
	    for (pt2 = (ptype_list *) pt->DATA; pt2; pt2 = pt2->NEXT)
	      {
		if (((lotrs_list *) (pt2->DATA))->GRID->SIG->INDEX == index)
		  break;
		Cp = Cp + ((VDDmax - VSSdeg) / VDDmax) * (pt2->NEXT->TYPE) *
		  i_branche (((ptype_list *) (pt->DATA))) /
		  i_branche (pt2->NEXT);
	      }
	}
      ptchain = ptchain->NEXT;
    }
  Cn = Cs + Cn;

   if(CCELL_MODE_DEBUG == 'Y')
   {
   if(type == TRANSN )    
   {
   fprintf(stdout,"pour TPHL : entree %s \n",ptlocon->NAME);
   fprintf(stdout,"\tCn=%ld (fF)\n",(long)Cn);
   fprintf(stdout,"\tCp=%ld (fF)\n",(long)Cp);
   fprintf(stdout,"\tC=%ld (fF)\n",(long)(Cp+Cn));
   }
   if(type == TRANSP)
   {
   fprintf(stdout,"pour TPLH : entree %s \n",ptlocon->NAME);
   fprintf(stdout,"\tCp=%ld (fF)\n",(long)Cn);
   fprintf(stdout,"\tCn=%ld (fF)\n",(long)Cp);
   fprintf(stdout,"\tC=%ld (fF)\n",(long)(Cp+Cn));
   }
   }
 

  return ((long) (Cn + Cp));
}

/***************************************************
****   fonction de calcul de la resistance      ****   
***************************************************/
long
calc_res (ptlocon, type)
     locon_list *ptlocon;
     int type;
{
  return (long) (VDDmax / (2 * i_branche ((ptype_list *) max_res (ptlocon, type)->DATA)));
}


/***************************************************
****     fonction qui cree la structure         ****   
****       resultat pour chaque entree          ****
***************************************************/
lores_list *
paramcell (ptlofig, capa, front)
     lofig_list *ptlofig;
     int capa;
     int front;
{
  locon_list *ptlocon;
  lores_list *ptres;
  long Rup, Rdown;
  long Chl, Clh;
  long Tphl, Tplh;
  double HN, HP;
  double Un, BUn, Up, BUp;

  HEAD_LORES = NULL;
  ptres = NULL;

  Un = VDDmax - VTN;
  BUn = Un * BN;
  HN = 0.7 * (BUn / (BUn - 1)) - (log (1 + BUn)) / (BUn * (BUn - 1)) -
    0.5 * log ((Un + 0.5 * VDDmax - VTN) / (Un - 0.5 * VDDmax + VTN));

  Up = VDDmax - VTP;
  BUp = Up * BP;
  HP = 0.7 * (BUp / (BUp - 1)) - (log (1 + BUp)) / (BUp * (BUp - 1)) -
    0.5 * log ((Up + 0.5 * VDDmax - VTP) / (Up - 0.5 * VDDmax + VTP));


   if(CCELL_MODE_DEBUG == 'Y')
   {
   fprintf(stdout,"\n");                           
   fprintf(stdout,"Un=%f\n",Un);                           
   fprintf(stdout,"Up=%f\n",Up);                           
   fprintf(stdout,"Bn=%f\n",BUn);                          
   fprintf(stdout,"Bp=%f\n",BUp);                          
   fprintf(stdout,"Hn=%f\n",HN);                           
   fprintf(stdout,"Hp=%f\n",HP);                           
   fprintf(stdout,"f=%d\n",front);                         
   fprintf(stdout,"\n");                           
   }          
 

  for (ptlocon = ptlofig->LOCON; ptlocon; ptlocon = ptlocon->NEXT)
    {
      if (ptlocon->USER != NULL)
	{
	  Rdown = calc_res (ptlocon, TRANSN);
	  Rup = calc_res (ptlocon, TRANSP);

	  Chl = calc_capa (ptlocon, TRANSN);
	  Clh = calc_capa (ptlocon, TRANSP);

	  Tphl = (long) ((Rdown * Chl) / 1000) + (long) ((Rdown * capa) / 1000) + (long) (HN * front);
	  Tplh = (long) ((Rup * Clh) / 1000) + (long) ((Rup * capa) / 1000) + (long) (HP * front);

	  ptres = addlores (ptlocon->NAME, Rup, Rdown, (long) (ptlocon->SIG->CAPA * 1000),
			    Tphl, Tplh, ptlocon);
	}
    }

  return ptres;
}

/*##------------------------------------------------------------------## */
/*  Function : displayLotrs()                                           */
/*##------------------------------------------------------------------## */
void
displayLotrs (lofig)
     lofig_list *lofig;
{
  lotrs_list *trs;
  locon_list *con;

  printf ("DISPLAYLOTRS of %s\n", lofig->NAME);

  for (con = lofig->LOCON; con; con = con->NEXT)
    printf ("  CONNECTEUR %s\t : dir %c - sig %s (%ld)\n",
	 con->NAME, con->DIRECTION, getsigname (con->SIG), con->SIG->INDEX);

  for (trs = lofig->LOTRS; trs; trs = trs->NEXT)
    printf ("  TRANSISTOR : drain %s (%ld)\t- grille %s (%ld)\t- source %s (%ld)\n",
	    getsigname (trs->DRAIN->SIG), trs->DRAIN->SIG->INDEX,
	    getsigname (trs->GRID->SIG), trs->GRID->SIG->INDEX,
	    getsigname (trs->SOURCE->SIG), trs->SOURCE->SIG->INDEX);
}

/***************************************************
 **** fonction de caracterisation d'une cellule ****
 ***************************************************/
lores_list *
caracLofig (name, capa, front)
     char *name;
     int capa;
     int front;
{
  lofig_list *ptlofig;
  lofig_list *ptlofig_res;
  lofig_list *lofigNeg, *lofigInv;
  lotrs_list *ptrs;
  lores_list *ptres, *ptresInv;
  ptype_list *pttype;
  char *negName;
  losig_list *sig, *sigOut = (losig_list *) NULL;
  locon_list *con;
  lotrs_list *trs;
  losig_list *exceptSig = (losig_list *) NULL;
  long exceptIndex;

  /* on elimine toutes les lofig, HEAD_LOFIG from MUT */
  while ( HEAD_LOFIG )
   {
    dellofig( HEAD_LOFIG->NAME );
   }

  /* ces 2 lignes sont recuperees du main ci-dessous */
  SCALE_X = 100;

  ptlofig = getlofig (name, 'A');
  negName = mbkalloc (strlen (ptlofig->NAME) + 2);

  for (trs = ptlofig->LOTRS; trs; trs = trs->NEXT)
    {
      trs->PS = 4 * SCALE_X + 2 * trs->WIDTH;
      trs->PD = 4 * SCALE_X + 2 * trs->WIDTH;
      trs->XS = 2 * SCALE_X;
      trs->XD = 2 * SCALE_X;
    }

  if (PREMIER == 1)
    rdsenv ();

  if (PREMIER == 1)
    {
      TechnoParamelp ();
      TechnoParamrds ();
    }

  printf ("caracLofig of %s (%d) elp=%s\n", name, PREMIER, elpTechnoFile);
/*
 */

  PREMIER++;

  /* recherche de l'eventuel inverseur de sortie */
  lofigchain (ptlofig);

/*
   printf("Recherche inverseur de sortie\n");
 */
  for (ptrs = ptlofig->LOTRS; ptrs; ptrs = ptrs->NEXT)
    {
      int Ok = 0;
      locon_list *con;

/*
   printf("G=%s D=%s S=%s \n",
   getsigname(ptrs->GRID->SIG),
   getsigname(ptrs->DRAIN->SIG),
   getsigname(ptrs->SOURCE->SIG));
 */

      for (con = ptlofig->LOCON; con; con = con->NEXT)
	if (con->SIG == ptrs->GRID->SIG)
	  Ok = 1;

      if (!Ok)
	exceptSig = ptrs->GRID->SIG;
    }

/*
   displayLotrs(ptlofig);
 */

  if (exceptSig)
    {
      lores_list *res;

      /* on a une cellule positive */
      exceptIndex = exceptSig->INDEX;

      sprintf (negName, "m%s", ptlofig->NAME);
      lofigNeg = addlofig (negName);
      lofigInv = addlofig ("n1");

      for (sig = ptlofig->LOSIG; sig; sig = sig->NEXT)
	{
	  if (isvdd (getsigname (sig)) || isvss (getsigname (sig)))
	    {
	      addlosig (lofigInv, sig->INDEX,
			addchain ((chain_list *) NULL, getsigname (sig)),
			sig->TYPE, sig->CAPA);
	    }

	  if (sig != exceptSig)
	    addlosig (lofigNeg, sig->INDEX,
		      addchain ((chain_list *) NULL, getsigname (sig)),
		      sig->TYPE, sig->CAPA);
	  else
	    addlosig (lofigInv, sig->INDEX,
		      addchain ((chain_list *) NULL, "i"), EXTERNAL, 0.0);
	}

      for (con = ptlofig->LOCON; con; con = con->NEXT)
	{
	  addlocon (lofigNeg, con->NAME, getlosig (lofigNeg, con->SIG->INDEX),
		    con->DIRECTION);
	  if (con->DIRECTION == OUT)
	    {
	      addlocon (lofigInv, con->NAME, con->SIG, con->DIRECTION);
	      addlocon (lofigInv, "i", exceptSig, IN);

	      /* signal de sortie */
	      addlosig (lofigInv, con->SIG->INDEX,
			addchain ((chain_list *) NULL,
				  getsigname (con->SIG)), EXTERNAL, 0.0);
	      sigOut = con->SIG;
	    }
	}

      for (trs = ptlofig->LOTRS; trs; trs = trs->NEXT)
	{
	  if (trs->GRID->SIG == exceptSig)
	    addlotrs (lofigInv, trs->TYPE, trs->X, trs->Y, trs->WIDTH, trs->LENGTH,
		      trs->PS, trs->PD, trs->XS, trs->XD,
		      getlosig (lofigInv, trs->GRID->SIG->INDEX),
		      getlosig (lofigInv, trs->SOURCE->SIG->INDEX),
		      getlosig (lofigInv, trs->DRAIN->SIG->INDEX));
	  else
	    {
	      losig_list *sigSource, *sigDrain;

	      if (trs->SOURCE->SIG->INDEX == exceptIndex)
		sigSource = sigOut;
	      else
		sigSource = getlosig (lofigNeg, trs->SOURCE->SIG->INDEX);

	      if (trs->DRAIN->SIG->INDEX == exceptIndex)
		sigDrain = sigOut;
	      else
		sigDrain = getlosig (lofigNeg, trs->DRAIN->SIG->INDEX);

	      addlotrs (lofigNeg, trs->TYPE, trs->X, trs->Y, trs->WIDTH, trs->LENGTH,
			trs->PS, trs->PD, trs->XS, trs->XD,
			getlosig (lofigNeg, trs->GRID->SIG->INDEX),
			sigSource, sigDrain);
	    }
	}

/*
   displayLotrs(lofigNeg);
   displayLotrs(lofigInv);
 */

      /* calcul des generics de la lofig negative */
      ptlofig_res = rduplofig (lofigNeg);	/* duplication de la lofig */
      HEAD_LOFIG_RES = ptlofig_res;

      param_cell_rds (ptlofig_res);	/* parametrages physiques  */

      param_cell_elp (ptlofig_res);	/* parametrages electriques */
      pttype = createbranch (ptlofig_res);	/* desassemblage de la cellule */
      linklocon (pttype);	/* lien entre locon et branches */

      ptres = paramcell (ptlofig_res, capa, front);
/*
   viewlores(ptlofig_res, ptres) ;
 */

      /* calcul des generics de l'inverseur */
      ptlofig_res = rduplofig (lofigInv);	/* duplication de la lofig */
      HEAD_LOFIG_RES = ptlofig_res;

      param_cell_rds (ptlofig_res);	/* parametrages physiques  */

      param_cell_elp (ptlofig_res);	/* parametrages electriques */
      pttype = createbranch (ptlofig_res);	/* desassemblage de la cellule */
      linklocon (pttype);	/* lien entre locon et branches */

      ptresInv = paramcell (ptlofig_res, capa, front);
/*
   viewlores(ptlofig_res, ptresInv);
 */

      /* calcul des generics de la lofig positive */
      for (res = ptres; res; res = res->NEXT)
	{
	  res->RUP = ptresInv->RUP;
	  res->RDOWN = ptresInv->RDOWN;
	  res->TPHL = res->TPHL + ptresInv->TPLH;
	  res->TPLH = res->TPLH + ptresInv->TPHL;
	}

/*
   viewlores(ptlofig, ptres);
   veriflores(ptlofig, ptres);
 */

      return ptres;
    }
  else
    {
      ptlofig_res = rduplofig (ptlofig);	/* duplication de la lofig */
      HEAD_LOFIG_RES = ptlofig_res;

      param_cell_rds (ptlofig_res);	/* parametrages physiques  */


   if(CCELL_MODE_DEBUG == 'Y')
   { 
   fprintf(stdout,"Lofig modifie\n");
   viewlofig(ptlofig_res) ;
   }
 

      param_cell_elp (ptlofig_res);	/* parametrages electriques */
      pttype = createbranch (ptlofig_res);	/* desassemblage de la cellule */
      linklocon (pttype);	/* lien entre la locon et les branches */

      ptres = paramcell (ptlofig_res, capa, front);

      /* affichage des resultats */
/*
   viewlores(ptlofig_res, ptres) ;
   veriflores(ptlofig, ptres);
 */

      return ptres;
    }
}

/*
   main(argc,argv)
   int argc;
   char *argv[];
   {
   lofig_list *ptlofig ;
   char *elp_techno_name;

   mbkenv();
   SCALE_X = 100 ;
   elp_techno_name =getenv("ELP_TECHNO_NAME");
   ptlofig=addlofig(argv[1]);
   loadlofig(ptlofig,argv[1],'A');
   cara_cell(ptlofig,0,1000,elp_techno_name);
   }
 */
