/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - CMOS Cells Compiler
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */




#ident  "$Id: c4_decomp_dec4.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#include MUT_H
#include MLO_H
#include LOG_H
#include BEH_H
#include BVL_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"
#include "c4_local.h"

pTH C4_library;
chain_list *mf_list;

/*------------------------------------------------------------------
  IDENT : normExprC4
  ------------------------------------------------------------------
  FUNCTION : normalise une expression a tous les niveaux
                variable = pointeur d'allocation
                expression = numero d'operateur.
             ex : (and a b (or x f) (or e g o))
                --> (and (or e o g) (or x f) b a)

  --> Tri decroissant
  1er critere : nombre de litteraux
  2eme  "     : longueur au premier niveau
  3eme  "     : numero d'operateur
  ------------------------------------------------------------------ */

long
funcNormExprC4 (expr)
     chain_list *expr;
{
  if (ATOM (expr))
    return 0;
  else if (OPER (expr) == NOT)
    return 0;
  else
    return (numberAtomExpr (expr) * 100 +
	    OPER (expr) + lengthExpr (expr) * 10);
}

void
normExprC4 (expr)
     chain_list *expr;
{
  if (!ATOM (expr))
    {
      sortExpr (expr, funcNormExprC4, 0);
      while ((expr = CDR (expr)))
	normExprC4 (CAR (expr));
    }
}


/*------------------------------------------------------------------------------
expr2Bst 	: cree une expression au format bst (ident_porte i0 i1 ... in)
		  a partir d'une expression beh et d'un ident (int) de porte.
------------------------------------------------------
parametres  	: un operateur (int) et un ABL.
-------------------------------------------------------
return 		: un chain_list *
------------------------------------------------------------------------------*/
chain_list *
expr2Bst (oper, expr)
     short oper;
     chain_list *expr;
{
  chain_list *res;
  void addNExpr ();

  res = createExpr (oper);
  addNExpr (res, expr);
  return (res);
}


/* rajoute au premier niveau dans e1 les entrees de e2 */

void
addNExpr (e1, e2)
     chain_list *e1, *e2;
{
  if (ATOM (e2))
    addQExpr (e1, createAtom (VALUE_ATOM (e2)));
  else
    {
      while ((e2 = CDR (e2)))
	addNExpr (e1, CAR (e2));
    }
}

/*------------------------------------------------------------------------------
c4AssocExpr 	: tri une expression normalisee avec la liste assocative de
		  la memFig.
------------------------------------------------------
parametres  	: un ABL et la memFig.
-------------------------------------------------------
return 		: un chain_list*
------------------------------------------------------------------------------*/
chain_list *
c4AssocExpr (mf, expr)
     MemFig *mf;
     chain_list *expr;
{
  chain_list *result;
  chain_list *assoc = mf->assocList;

  result = createExpr (OPER (expr));
  while (assoc)
    {
      addQExpr (result, copyExpr (searchNumExpr (expr, (int)assoc->DATA)));
      assoc = assoc->NEXT;
    }
  return (result);
}

/*------------------------------------------------------------------------------
c4MappExprBeh 	: creation des modeles virtuels.  (memF)
------------------------------------------------------
parametres  	: un ABL..
-------------------------------------------------------
return 		: TRUE  (1) si la compilation a reussi
		  FALSE (0) sinon.
------------------------------------------------------------------------------*/
int
c4MappExprBeh (beh, expr, mode)
     befig_list *beh;
     chain_list *expr;
     int mode;
{
  MemFig *memF;
  char *ident;
  int result = TRUE;

  if (mode == 1)
    printf ("c4MappExpr : \n");

  if (!ATOM (expr))
    {
      /* modification de expr */

      /* normalisation */

      normExprC4 (expr);

      /* appel a C4 */

      ident = identExpr (expr);

      /* ce n'est pas un not  */

      if (strcmp (ident, "not1"))
	{
	  /* cellule existe deja ? */

	  if (searchTH (C4_library, ident) == EMPTYTH)
	    {
	      memF = c4_comp (expr2Model (expr), ident, SL_NUMCANN, SL_NUMCANP, 1);
	      if (!C4_ROUTok (memF))
		{
		  printf ("\nC4 compiler : ERROR %s badly compiled :", ident);
		  c4_dbg_printInfo (memF);
		  result = FALSE;
		}
	      if (memF->result->polarity == 2)
		{
		  printf ("\nC4 compiler : ERROR function has a bad polarity\n");
		  printf ("---> ");
		  displayExpr (expr);
		  result = FALSE;
		}

/* pour les benchs de la these de Nat ... le 17/04/96 */

	      if (memF->result->nbTPSer > SL_NUMTRANSP)
		{
		  printf ("\nC4 compiler : ERROR");
		  printf (" function has too many serial transistors in P-graph\n");
		  printf ("---> ");
		  displayExpr (expr);
		  result = FALSE;
		}
	      if (memF->result->nbTNSer > SL_NUMTRANSN)
		{
		  printf ("\nC4 compiler : ERROR");
		  printf (" function has too many serial transistors in N-graph\n");
		  printf ("---> ");
		  displayExpr (expr);
		  result = FALSE;
		}
	      addTH (C4_library, ident, (int)memF);
	      mf_list = addchain (mf_list, memF);
	    }
	}
    }
  return result;
}

/*------------------------------------------------------------------------------
c4GenCell 	: effectue la mise en forme structurelle d'une expression.
------------------------------------------------------
parametres  	: un ABL.
-------------------------------------------------------
return 		: un void
------------------------------------------------------------------------------*/
chain_list *
c4GenCell (expr, mode)
     chain_list *expr;
     int mode;
{
  MemFig *memF;
  int val;
  char *ident;
  chain_list *newExpr;

  struct bepor *port;
  chain_list *input_ref;
  chain_list *expr_inputs;
  chain_list *input;

  if (mode == 1)
    printf ("c4GenCell\n");

  if (!ATOM (expr))
    {
      ident = identExpr (expr);

      if (!strcmp (ident, "not1"))
	{
	  newExpr = expr2Bst (searchNumOper (ident), expr);
	  freeExpr (expr);
	}
      else
	{
	  if ((val = searchTH (C4_library, ident)) != EMPTYTH)
	    {
	      memF = (MemFig *) val;
	      if (memF->status.loadedModel)
		{
		  expr_inputs = expr2Bst (searchNumOper (ident), expr);
		  newExpr = createExpr (searchNumOper (ident));
		  for (port = memF->beh->BEPOR; port; port = port->NEXT)
		    {
		      if ((port->DIRECTION == 'I') &&
			  !isvdd (port->NAME) && !isvss (port->NAME))
			{
			  input = CDR (expr_inputs);
			  for (input_ref = CDR (memF->assocList);
			       input_ref; input_ref = CDR (input_ref))
			    {
			      if (VALUE_ATOM (CAR (input_ref)) == port->NAME)
				break;
			      input = CDR (input);
			    }
			  addHExpr (newExpr, createAtom (VALUE_ATOM (CAR (input))));
			}
		    }
		}
	      else
		{
		  newExpr = expr2Bst (searchNumOper (ident), expr);
		  if (mode == 1)
		    {
		      displayArite (newExpr);
		      printf ("\n");
		    }
		  newExpr = c4AssocExpr (memF, newExpr);
		  if (mode == 1)
		    {
		      displayArite (newExpr);
		      printf ("\n");
		    }
		}
	      freeExpr (expr);
	    }
	  else
	    {
	      printf ("c4Gencell : error unknown operator %s\n", ident);
	      exit (-1);
	    }
	}
      return (newExpr);
    }
  else
    return (expr);
}

/*------------------------------------------------------------------------------
interfaceBeh 	: renvoie dans une liste chainee l'interface d'une befig
------------------------------------------------------
parametres  	: une befig.
-------------------------------------------------------
return 		: un chain_list *.
------------------------------------------------------------------------------*/
chain_list *
interfaceBeh (beh)
     befig_list *beh;
{
  chain_list *lc_io = NULL;
  bepor_list *port = beh->BEPOR;

  while (port)
    {
      lc_io = addchain (lc_io, port->NAME);
      port = port->NEXT;
    }
  return lc_io;
}

/*
   // initC4library
 */

static pTH
initC4library (mode)
     int mode;
{
  pTH cells_library;
  DIR *directory;
  struct dirent *entry;
  char *suffix;
  chain_list *cells_list;
  chain_list *list_ptr;
  char *cell_name;
  int cell_tag;
  int old_tag;
  befig_list *cell_befig;
  struct memFig *cell_memfig;
  chain_list *cell_function;

  cells_library = createTH (199);
  directory = opendir (SL_CCCC_LIB);
  if (directory)
    {
      cells_list = NULL;
      for (entry = readdir (directory); entry; entry = readdir (directory))
	{
	  suffix = strchr (entry->d_name, '.');
	  if (suffix)
	    {
	      *suffix = 0;
	      suffix++;
	      if (strlen (suffix) == 0)
		cell_tag = 000;
	      else if (!strcmp (suffix, "vbe"))		/* behavioral description */
		cell_tag = 001;
	      else
		{
		  if (!strcmp (suffix, SL_OUT_LO))	/* structural */
		    cell_tag = 002;
		  else
		    {
		      if (!strcmp (suffix, SL_OUT_PH))	/* layout */
			cell_tag = 004;
		      else
			cell_tag = 000;
		    }
		}
	      if (cell_tag)
		{
		  cell_name = namealloc (entry->d_name);
		  if ((old_tag = searchTH (cells_library, cell_name)) != EMPTYTH)
		    cell_tag |= old_tag;
		  else
		    cells_list = addchain (cells_list, cell_name);
		  addTH (cells_library, cell_name, cell_tag);
		}
	    }
	}
/*
   if (mode)
   printf( "cells found in %s :\n", SL_CCCC_LIB);
 */
      for (list_ptr = cells_list; list_ptr; list_ptr = list_ptr->NEXT)
	{
	  cell_name = list_ptr->DATA;
/*
   if (mode)
   printf( "%s ", cell_name);
 */
	  cell_tag = searchTH (cells_library, cell_name);
	  cell_befig = vhdlloadbefig (NULL, cell_name, 0);
	  if ((cell_tag == 007) && cell_befig && cell_befig->BEOUT)	/* avoid tie */
	    {
	      addOper (cell_name);
	      cell_memfig = (struct memFig *) calloc (1, sizeof (struct memFig));
	      cell_memfig->fName = cell_name;
	      cell_memfig->beh = cell_befig;
	      cell_function = cell_befig->BEOUT->ABL;
	      flatArityExpr (cell_function);
	      cell_function = flatPolarityExpr (cell_function, 1);
	      normExprC4 (cell_function);
	      cell_memfig->assocList = expr2Bst (searchNumOper (cell_name), cell_function);
	      freeExpr (cell_function);
	      mappSys->ptcell = initCellList (mappSys->ptcell, cell_name, 'l', cell_befig);
	      recupGeneric (cell_befig, mappSys->ptcell, 42);
	      addTH (cells_library, cell_name, (int)cell_memfig);
	      addTH (mappSys->cell_hash, cell_name, (int)mappSys->ptcell);
	      cell_memfig->status.loadedModel = 1;
/*
   if (mode)
   printf( "reusable,\n");
 */
	    }
	  else
	    {
	      deleteTH (cells_library, cell_name);
/*
   if (mode)
   printf( "not available,\n");
 */
	    }
	}
/*
   if (mode)
   printf( "end of list.");
 */
      freechain (cells_list);
      closedir (directory);
    }
  else
    {
      printf ("ERROR : directory %s can not be open\n", SL_CCCC_LIB);
      exit (-1);
    }

  return cells_library;
}

/*------------------------------------------------------------------------------
c4MappBeh 	: genere un format befig structurel en appelant le compilateur
                  de cellule. (decomposition comprise)
------------------------------------------------------
parametres  	: une befig.
-------------------------------------------------------
return 		: un void
------------------------------------------------------------------------------*/
void
c4MappBeh (beh, mode, delais)
     befig_list *beh;
     int mode;
     ptype_list *delais;
{
  beout_list *out;
  beaux_list *aux;
  beaux_list *beh_addbeaux ();
  chain_list *simplifExpr10 ();
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  biabl_list *biabl;
  chain_list *pCL;

  char *name;
  char *ident;

  int coNOT = 0;
  int coL1 = 0;
  int coL2 = 0;
  int coNT1 = 0;

  pTH caracTH;
  pTH caracBeh ();

  char *c4High = NULL;
  char *c4Low = NULL;
  chain_list *lc_io;


  char *searchCharOper ();
  char *genNAME ();

  int successful_compilation;

  /* caracterisation ==> tri des expressions */
  caracTH = caracBeh (beh, delais, 0);

  /* initialisations */
  mappSys = initMapping ();
  C4_library = initC4library (mode);
  mf_list = NULL;

/**************************/
  /* extraction des latches */
/**************************/

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;

      if (biabl && biabl->NEXT == NULL)
	{
	  /* latch 1 */

	  coL1++;
	  if (!ATOM (biabl->VALABL))
	    {
	      name = gensym ("auxc4_l1d", coL1);
	      beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->VALABL, NULL);
	      biabl->VALABL = createAtom (name);
	    }
	  if (!ATOM (biabl->CNDABL))
	    {
	      name = gensym ("auxc4_l1c", coL1);
	      beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->CNDABL, NULL);
	      biabl->CNDABL = createAtom (name);
	    }
	}
      else if (biabl->NEXT && biabl->NEXT->NEXT == NULL)
	{
	  /* latch 2 */

	  while (biabl)
	    {
	      coL2++;
	      if (!ATOM (biabl->VALABL))
		{
		  name = gensym ("auxc4_l2d", coL2);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->VALABL, NULL);
		  biabl->VALABL = createAtom (name);
		}
	      if (!ATOM (biabl->CNDABL))
		{
		  name = gensym ("auxc4_l2c", coL2);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->CNDABL, NULL);
		  biabl->CNDABL = createAtom (name);
		}
	      biabl = biabl->NEXT;
	    }
	}
      else
	{
	  printf ("Latches extraction : error - unknown latch '%s'\n", reg->NAME);
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;

      /* BUS */
      while (biabl)
	{
	  coNT1++;
	  if (!ATOM (biabl->VALABL))
	    {
	      name = gensym ("auxc4_nt1d", coNT1);
	      beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->VALABL, NULL);
	      biabl->VALABL = createAtom (name);
	    }

	  if (!ATOM (biabl->CNDABL))
	    {
	      name = gensym ("auxc4_nt1c", coNT1);
	      beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->CNDABL, NULL);
	      biabl->CNDABL = createAtom (name);
	    }
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;

      /* BUS */

      while (biabl)
	{
	  coNT1++;
	  if (!ATOM (biabl->VALABL))
	    {
	      name = gensym ("auxc4_nt1d", coNT1);
	      beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->VALABL, NULL);
	      biabl->VALABL = createAtom (name);
	    }

	  if (!ATOM (biabl->CNDABL))
	    {
	      name = gensym ("auxc4_nt1c", coNT1);
	      beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->CNDABL, NULL);
	      biabl->CNDABL = createAtom (name);
	    }
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }

  /* une fois les latches et bus extraits,
     on s'occupe du combinatoire ... */

  successful_compilation = TRUE;

  out = beh->BEOUT;
  while (out)
    {

      if (out->ABL)
	{
	  if (mode == 1)
	    {
	      printf ("je traite %s =\n ", out->NAME);
	      displayExpr (out->ABL);
	    }

	  if (ATOM (out->ABL))
	    {
	      beout_list *out1 = beh->BEOUT;
	      bepor_list *bepor = beh->BEPOR;
	      int entreecon = 0;
	      int sortiecon = 0;

	      while (bepor)
		{
		  if (!strcmp (bepor->NAME, VALUE_ATOM (out->ABL)))
		    {
		      entreecon = 1;
		    }
		  bepor = bepor->NEXT;
		}

	      /* la sortie est connecte directement a une entree */

	      if (entreecon == 1)
		{
		  coNOT++;
		  name = gensym ("auxc4_not", coNOT);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX,
					     name,
					     notExpr (out->ABL),
					     NULL);
		  out->ABL = notExpr (createAtom (name));
		}


	      while (out1)
		{
		  if (ATOM (out1->ABL))
		    {
		      if (!strcmp (VALUE_ATOM (out1->ABL), VALUE_ATOM (out->ABL)))
			{
			  sortiecon = 1;
			}
		    }
		  out1 = out1->NEXT;
		}
	      /* la sortie est connectee a une autre sortie */

	      if (sortiecon == 1)
		{
		  coNOT++;
		  name = gensym ("auxc4_not", coNOT);
		  if ((!strcmp (VALUE_ATOM (out->ABL), "'0'")) ||
		      (!strcmp (VALUE_ATOM (out->ABL), "'1'")))
		    {
		      if (!strcmp (VALUE_ATOM (out->ABL), "'1'"))
			beh->BEAUX = beh_addbeaux (beh->BEAUX,
						   name,
						   createAtom ("'0'"),
						   NULL);
		      else
			beh->BEAUX = beh_addbeaux (beh->BEAUX,
						   name,
						   createAtom ("'1'"),
						   NULL);
		    }
		  else
		    beh->BEAUX = beh_addbeaux (beh->BEAUX,
					       name,
					       notExpr (out->ABL),
					       NULL);
		  out->ABL = notExpr (createAtom (name));
		}

	    }
	  successful_compilation &= c4MappExprBeh (beh, out->ABL, mode);
	}
      out = out->NEXT;
    }

  for (aux = beh->BEAUX; aux; aux = aux->NEXT)
    {
      if (aux->ABL)
	{
	  if (mode == 1)
	    {
	      printf ("je traite %s =\n ", aux->NAME);
	      displayExpr (aux->ABL);
	    }
	  successful_compilation &= c4MappExprBeh (beh, aux->ABL, mode);
	}
    }

  if (!successful_compilation)
    {
      printf ("\nUnsuccessful compilation ...");
      exit (-1);
    }

  if (mode == 1)
    printf ("fin de la decomposition avec c4\n");

  /* appel au compilateur c4 sur chaque element de mf_list */

  if (mode == 1)
    printf ("generation des cellules:  \n");


  /* ===== generation du tie ===== */
  /* comportement */
  VHB_HEDFIG = tie ();
  decompileVHDL (VHB_HEDFIG, NULL, 0);

  /* logique + physique + icon */
  lc_io = interfaceBeh (VHB_HEDFIG);
  c4_genPCells (C4_TIE, VHB_HEDFIG->NAME, lc_io, 0, 0, 0, 0, 0, 0);

  /* ===== generation du not ===== */
  /* comportement */
  VHB_HEDFIG = not1 ();
  decompileVHDL (VHB_HEDFIG, NULL, 0);

  /* logique + physique + icon */
  lc_io = interfaceBeh (VHB_HEDFIG);
  c4_genPCells (C4_N1, VHB_HEDFIG->NAME, lc_io, VHB_HEDFIG->BEOUT->NAME,
		VALUE_ATOM (CADR (VHB_HEDFIG->BEOUT->ABL)), 0, 0, 0, 0);


  /* operateur interne pour la bst */

  ident = namealloc ("not1");
  addOper (ident);
  mappSys->ptcell =
    initCellList (mappSys->ptcell, ident, 'l', VHB_HEDFIG);

  mappSys->ptcell->nom = VHB_HEDFIG->NAME;

  recupGeneric (VHB_HEDFIG, mappSys->ptcell, 42);

  addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);

  if (mode == 1)
    displayCell (mappSys->ptcell);

  /* ===== generation du not puissant ===== */

  /* comportement */
  VHB_HEDFIG = not1p ();
  decompileVHDL (VHB_HEDFIG, NULL, 0);

  /* logique + physique + icon */
  lc_io = interfaceBeh (VHB_HEDFIG);
  c4_genPCells (C4_N1P, VHB_HEDFIG->NAME, lc_io, VHB_HEDFIG->BEOUT->NAME,
		VALUE_ATOM (CADR (VHB_HEDFIG->BEOUT->ABL)), 0, 0, 0, 0);

  /* ===== generation du buffer puissant ===== */

  /* comportement */
  VHB_HEDFIG = buf ();
  decompileVHDL (VHB_HEDFIG, NULL, 0);

  /* logique + physique + icon */
  lc_io = interfaceBeh (VHB_HEDFIG);
  c4_genPCells (C4_BUF, VHB_HEDFIG->NAME, lc_io, VHB_HEDFIG->BEOUT->NAME,
		VALUE_ATOM (VHB_HEDFIG->BEOUT->ABL), 0, 0, 0, 0);

  /* =====  generation des latchs ===== */
  if (beh->BEREG)
    {
      biabl_list *biablN;

      /* =====  generation du latch 1 ===== */

      /* comportement */
      VHB_HEDFIG = latch1 ();
      decompileVHDL (VHB_HEDFIG, NULL, 0);

      /* logique + physique */
      lc_io = interfaceBeh (VHB_HEDFIG);
      biabl = (VHB_HEDFIG->BEREG)->BIABL;

      /* LATCH1 : i1 = data   i2 = commande
         i3 = point memorisant       */
      c4_genPCells (C4_L1, VHB_HEDFIG->NAME, lc_io, VHB_HEDFIG->BEOUT->NAME,
		    VALUE_ATOM (biabl->VALABL), VALUE_ATOM (biabl->CNDABL),
		    VHB_HEDFIG->BEREG->NAME, 0, 0);


      /* operateur interne pour la bst */

      ident = namealloc ("reg1");
      addOper (ident);
      mappSys->ptcell =
	initCellList (mappSys->ptcell, ident, 'r', VHB_HEDFIG);

      recupGeneric (VHB_HEDFIG, mappSys->ptcell, 42);
      mappSys->ptcell->nom = VHB_HEDFIG->NAME;
      addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);
      if (mode == 1)
	displayCell (mappSys->ptcell);

      /* =====  generation du latch 2 ===== */

      /* comportement */
      VHB_HEDFIG = latch2 ();
      decompileVHDL (VHB_HEDFIG, NULL, 0);

      /* logique + physique */
      lc_io = interfaceBeh (VHB_HEDFIG);
      biabl = (VHB_HEDFIG->BEREG)->BIABL;
      biablN = biabl->NEXT;

      /* LATCH2 : i1 = data   i2 = commande
         i3 = data   i4 = commande
         i5 = point memorisant    */
      c4_genPCells (C4_L2, VHB_HEDFIG->NAME, lc_io, (VHB_HEDFIG->BEOUT)->NAME,
		    VALUE_ATOM (biabl->VALABL), VALUE_ATOM (biabl->CNDABL),
		    VALUE_ATOM (biablN->VALABL), VALUE_ATOM (biablN->CNDABL),
		    (VHB_HEDFIG->BEREG)->NAME);


      /* operateur interne pour la bst */

      ident = namealloc ("reg2");
      addOper (ident);
      mappSys->ptcell =
	initCellList (mappSys->ptcell, ident, 'r', VHB_HEDFIG);

      recupGeneric (VHB_HEDFIG, mappSys->ptcell, 42);
      mappSys->ptcell->nom = VHB_HEDFIG->NAME;
      addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);
      if (mode == 1)
	displayCell (mappSys->ptcell);

      if (beh->TYPE & BEH_STABLE)
	{
	  /* =====  generation de la Bascule D ===== */

	  /* comportement */
	  VHB_HEDFIG = bd1 ();
	  decompileVHDL (VHB_HEDFIG, NULL, 0);

	  /* logique + physique */
	  lc_io = interfaceBeh (VHB_HEDFIG);
	  biabl = VHB_HEDFIG->BEREG->BIABL;

	  /* C4_BD1  i1 = data   i2 = commande
	     i3 = 1er point memorisant  (master)
	     i4 = 2eme point memorisant (slave)      */

	  c4_genPCells (C4_BD1, VHB_HEDFIG->NAME, lc_io, (VHB_HEDFIG->BEOUT)->NAME,
			VALUE_ATOM (CADR (biabl->VALABL)),
	  /* on a : val = not i
	     cmd = (and (not l) (not (stable l))) */
			VALUE_ATOM (CADR (CADR (biabl->CNDABL))),
			VHB_HEDFIG->BEREG->NAME,
	  /* on invente un nom de point memo pour le slave */
			genNAME ("%ss", VHB_HEDFIG->BEREG->NAME), 0);


	  /* operateur interne pour la bst */

	  ident = namealloc ("bd1");
	  addOper (ident);
	  mappSys->ptcell =
	    initCellList (mappSys->ptcell, ident, 'd', VHB_HEDFIG);

	  recupGeneric (VHB_HEDFIG, mappSys->ptcell, 42);
	  mappSys->ptcell->nom = VHB_HEDFIG->NAME;
	  addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);
	  if (mode == 1)
	    displayCell (mappSys->ptcell);
	}
    }

  /* ===== generation du bus 1 ===== */

  if (beh->BEBUS || beh->BEBUX)
    {
      if (mode == 1)
	{
	  if (beh->BEBUS)
	    printf ("BUS ");
	  if (beh->BEBUX)
	    printf ("BUX ");
	  fflush (stdout);
	}

      /* comportement */
      VHB_HEDFIG = bus1 ();
      decompileVHDL (VHB_HEDFIG, NULL, 0);

      /* logique + physique */
      lc_io = interfaceBeh (VHB_HEDFIG);
      c4_genPCells (C4_TRI, VHB_HEDFIG->NAME, lc_io,
		    VHB_HEDFIG->BEBUS->NAME,
		    VALUE_ATOM (VHB_HEDFIG->BEBUS->BIABL->VALABL),
		    VALUE_ATOM (VHB_HEDFIG->BEBUS->BIABL->CNDABL),
		    0, 0, 0);

      /* operateur interne pour la bst */
      ident = namealloc ("bus1");
      addOper (ident);
      mappSys->ptcell = initCellList (mappSys->ptcell, ident, 'b', VHB_HEDFIG);

      recupGeneric (VHB_HEDFIG, mappSys->ptcell, 42);
      mappSys->ptcell->nom = VHB_HEDFIG->NAME;
      addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);
      if (mode == 1)
	displayCell (mappSys->ptcell);
    }

  /* ===== cellule C4_HIGH ===== */

  if (searchNameBeh (beh, "'1'"))
    {
      /* comportement */
      VHB_HEDFIG = high1 ();
      decompileVHDL (VHB_HEDFIG, NULL, 0);

      /* logique + physique */
      lc_io = interfaceBeh (VHB_HEDFIG);
      c4_genPCells (C4_HIGH, VHB_HEDFIG->NAME, lc_io, VHB_HEDFIG->BEOUT->NAME,
		    0, 0, 0, 0, 0);

      /* operateur interne pour la bst */
      ident = namealloc ("high");
      addOper (ident);
      mappSys->ptcell =
	initCellList (mappSys->ptcell, ident, 'l', VHB_HEDFIG);

      recupGeneric (VHB_HEDFIG, mappSys->ptcell, 42);
      mappSys->ptcell->nom = VHB_HEDFIG->NAME;
      addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);
      if (mode == 1)
	displayCell (mappSys->ptcell);
    }

  /* ===== cellule C4_LOW ===== */

  if (searchNameBeh (beh, "'0'"))
    {
      /* comportement */
      VHB_HEDFIG = low1 ();
      decompileVHDL (VHB_HEDFIG, NULL, 0);

      /* logique + physique */
      lc_io = interfaceBeh (VHB_HEDFIG);
      c4_genPCells (C4_LOW, VHB_HEDFIG->NAME, lc_io,
		    VHB_HEDFIG->BEOUT->NAME, 0, 0, 0, 0, 0);

      /* operateur interne pour la bst */
      ident = namealloc ("low");
      addOper (ident);
      mappSys->ptcell =
	initCellList (mappSys->ptcell, ident, 'l', VHB_HEDFIG);

      recupGeneric (VHB_HEDFIG, mappSys->ptcell, 42);
      mappSys->ptcell->nom = VHB_HEDFIG->NAME;
      addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);
      if (mode == 1)
	displayCell (mappSys->ptcell);
    }

  /* les autres cellules logiques */

  pCL = mf_list;
  if (pCL)
    printf ("C4 compilation : \n");
  while (pCL)
    {
      MemFig *memF = (MemFig *) pCL->DATA;

      ident = memF->fName;

      printf ("-- %s\n", ident);

      if (SL_TRACE)
	{
	  displayExpr (memF->inExprABL);
	  displayExpr (memF->trExprABL);
	}

      /* ***************** */
      c4_gen (memF);
      /* ***************** */

      if (mode == 1)
	c4_dbg_printInfo (memF);
      addOper (ident);
      mappSys->ptcell =
	initCellList (mappSys->ptcell, ident, 'l', memF->beh);

      recupGeneric (memF->beh, mappSys->ptcell, 42);
      addTH (mappSys->cell_hash, ident, (int)mappSys->ptcell);
      if (mode == 1)
	displayCell (mappSys->ptcell);
      pCL = pCL->NEXT;
    }

  out = beh->BEOUT;
  while (out)
    {
      if (out->ABL)
	{
	  if (mode == 1)
	    {
	      printf ("je traite %s =\n ", out->NAME);
	      displayExpr (out->ABL);
	    }
	  if (ATOM (out->ABL))
	    {
	      if (!strcmp (VALUE_ATOM (out->ABL), "'0'"))
		{
		  if (c4Low == NULL)
		    {
		      c4Low = namealloc ("c4Low");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		    }
		  out->ABL = createAtom (c4Low);
		}
	      if (!strcmp (VALUE_ATOM (out->ABL), "'1'"))
		{
		  if (c4High == NULL)
		    {
		      c4High = namealloc ("c4High");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		    }
		  out->ABL = createAtom (c4High);
		}
	    }
	  else
	    {
	      out->ABL = c4GenCell (out->ABL, mode);

	    }
	}
      out = out->NEXT;
    }

  aux = beh->BEAUX;
  while (aux)
    {
      if (aux->ABL)
	{
	  if (mode == 1)
	    {
	      printf ("je traite %s =\n ", aux->NAME);
	      displayExpr (aux->ABL);
	    }
	  if (ATOM (aux->ABL))
	    {
	      if (!strcmp (VALUE_ATOM (aux->ABL), "'0'"))
		{
		  if (c4Low == NULL)
		    {
		      c4Low = namealloc ("c4Low");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		    }
		  aux->ABL = createAtom (c4Low);
		}
	      if (!strcmp (VALUE_ATOM (aux->ABL), "'1'"))
		{
		  if (c4High == NULL)
		    {
		      c4High = namealloc ("c4High");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		    }
		  aux->ABL = createAtom (c4High);
		}
	    }
	  else
	    {
	      aux->ABL = c4GenCell (aux->ABL, mode);
	    }
	}
      aux = aux->NEXT;
    }

/*************************/
  /* Mapping des registres */
/*************************/
  reg = beh->BEREG;
  while (reg)
    {
      chain_list *LC = NULL;
      int isBD = 0;

      if (mode == 1)
	printf ("je traite le registre %s =\n ", reg->NAME);

      biabl = reg->BIABL;

      if (biabl->USER)
	{
	  /* bascule D */
	  isBD = 1;
	}

      /* instanciation du registre ou de la bascule */
      if (biabl && biabl->NEXT == NULL)
	{
	  if (!strcmp (VALUE_ATOM (biabl->CNDABL), "'0'"))
	    {
	      if (c4Low == NULL)
		{
		  c4Low = namealloc ("c4Low");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		}
	      LC = addchain (LC, c4Low);
	    }
	  else if (!strcmp (VALUE_ATOM (biabl->CNDABL), "'1'"))
	    {
	      if (c4High == NULL)
		{
		  c4High = namealloc ("c4High");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		}
	      LC = addchain (LC, c4High);
	    }
	  else
	    LC = addchain (LC, VALUE_ATOM (biabl->CNDABL));
	  if (!strcmp (VALUE_ATOM (biabl->VALABL), "'0'"))
	    {
	      if (c4Low == NULL)
		{
		  c4Low = namealloc ("c4Low");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		}
	      LC = addchain (LC, c4Low);
	    }
	  else if (!strcmp (VALUE_ATOM (biabl->VALABL), "'1'"))
	    {
	      if (c4High == NULL)
		{
		  c4High = namealloc ("c4High");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		}
	      LC = addchain (LC, c4High);
	    }
	  else
	    LC = addchain (LC, VALUE_ATOM (biabl->VALABL));
	  freeExpr (biabl->VALABL);
	  freeExpr (biabl->CNDABL);
	  if (isBD)
	    biabl->VALABL = createExprCL (searchNumOper (namealloc ("bd1")),
					  LC);
	  else
	    biabl->VALABL = createExprCL (searchNumOper (namealloc ("reg1")),
					  LC);
	  biabl->CNDABL = NULL;
	}
      else if (biabl->NEXT && biabl->NEXT->NEXT == NULL)
	{
	  while (biabl)
	    {
	      if (!strcmp (VALUE_ATOM (biabl->CNDABL), "'0'"))
		{
		  if (c4Low == NULL)
		    {
		      c4Low = namealloc ("c4Low");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		    }
		  LC = addchain (LC, c4Low);
		}
	      else if (!strcmp (VALUE_ATOM (biabl->CNDABL), "'1'"))
		{
		  if (c4High == NULL)
		    {
		      c4High = namealloc ("c4High");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		    }
		  LC = addchain (LC, c4High);
		}
	      else
		LC = addchain (LC, VALUE_ATOM (biabl->CNDABL));
	      if (!strcmp (VALUE_ATOM (biabl->VALABL), "'0'"))
		{
		  if (c4Low == NULL)
		    {
		      c4Low = namealloc ("c4Low");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		    }
		  LC = addchain (LC, c4Low);
		}
	      else if (!strcmp (VALUE_ATOM (biabl->VALABL), "'1'"))
		{
		  if (c4High == NULL)
		    {
		      c4High = namealloc ("c4High");
		      beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		    }
		  LC = addchain (LC, c4High);
		}
	      else
		LC = addchain (LC, VALUE_ATOM (biabl->VALABL));
	      freeExpr (biabl->VALABL);
	      freeExpr (biabl->CNDABL);
	      biabl = biabl->NEXT;
	    }
	  biabl = reg->BIABL;
	  biabl->VALABL = createExprCL (searchNumOper (namealloc ("reg2")), LC);
	  biabl->CNDABL = NULL;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      if (mode == 1)
	printf ("je traite le bus %s =\n ", bus->NAME);

      biabl = bus->BIABL;

      while (biabl)
	{
	  chain_list *LC = NULL;

	  /* creation d'une var. aux. pour la commande du bus */
	  LC = addchain (LC, VALUE_ATOM (biabl->CNDABL));	/* pour v */

	  /* la donnee est un atome */
	  if (strcmp (VALUE_ATOM (biabl->VALABL), "'0'") &&
	      strcmp (VALUE_ATOM (biabl->VALABL), "'1'"))
	    {
	      LC = addchain (LC, VALUE_ATOM (biabl->VALABL));	/* pour i */
	    }
	  else if (!strcmp (VALUE_ATOM (biabl->VALABL), "'0'"))
	    {
	      if (c4Low == NULL)
		{
		  c4Low = namealloc ("c4Low");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		}
	      LC = addchain (LC, c4Low);
	    }
	  else if (!strcmp (VALUE_ATOM (biabl->VALABL), "'1'"))
	    {
	      if (c4High == NULL)
		{
		  c4High = namealloc ("c4High");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		}
	      LC = addchain (LC, c4High);
	    }

	  freeExpr (biabl->VALABL);
	  freeExpr (biabl->CNDABL);
	  biabl->VALABL = createExprCL (searchNumOper (namealloc ("bus1")), LC);
	  biabl->CNDABL = NULL;
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      if (mode == 1)
	printf ("je traite le bux %s =\n ", bux->NAME);

      biabl = bux->BIABL;

      while (biabl)
	{
	  chain_list *LC = NULL;

	  /* creation d'une var. aux. pour la commande du bux */

	  LC = addchain (LC, VALUE_ATOM (biabl->CNDABL));	/* pour v */

	  /* la donnee est un atome */

	  if (strcmp (VALUE_ATOM (biabl->VALABL), "'0'") &&
	      strcmp (VALUE_ATOM (biabl->VALABL), "'1'"))
	    {
	      LC = addchain (LC, VALUE_ATOM (biabl->VALABL));	/* pour i */
	    }
	  else if (!strcmp (VALUE_ATOM (biabl->VALABL), "'0'"))
	    {
	      if (c4Low == NULL)
		{
		  c4Low = namealloc ("c4Low");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4Low, createAtom ("low"), NULL);
		}
	      LC = addchain (LC, c4Low);
	    }
	  else if (!strcmp (VALUE_ATOM (biabl->VALABL), "'1'"))
	    {
	      if (c4High == NULL)
		{
		  c4High = namealloc ("c4High");
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, c4High, createAtom ("high"), NULL);
		}
	      LC = addchain (LC, c4High);
	    }

	  freeExpr (biabl->VALABL);
	  freeExpr (biabl->CNDABL);
	  biabl->VALABL = createExprCL (searchNumOper (namealloc ("bus1")), LC);
	  biabl->CNDABL = NULL;
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }

  if (mode == 1)
    printf ("desallocation des cellules:  \n");

  pCL = mf_list;
  while (pCL)
    {
      MemFig *memF = (MemFig *) pCL->DATA;

      ident = memF->fName;

      if (strcmp (ident, "not1"))
	{
	  deleteTH (C4_library, ident);
	  c4_free (memF);
	}
      pCL = pCL->NEXT;
    }
  if (mode == 1)
    displayBeh (beh, 0);
}
