/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - CMOS Cells Compiler
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



/* $Id: c4_local.h,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $ */

#ifndef C4_H
#define C4_H

/*AFAC?*/
#define NEG		0
#define POS		1
#define AUT      	2

#define N_GRAPH		0
#define P_GRAPH		1

#undef OUTPUT
#define OUTPUT		1

#define ALIM		0


#define TRSP_BLK	1
#define OUT_BLK		2
#define INV_BLK		3



#define IN_SYMB		"i"

/* c4_genlib */

#define VSS_SYMB	"vss"
#define VDD_SYMB	"vdd"
#define OUT_SYMB	"t"
#define N_OUT_SYMB	"f"
#define INT_EQ_SYMB	"x_"

#define L_CELL_IN	"c4_in"
#define L_CELL_OUT	"c4_out"
#define L_CELL_INV	"c4_inv"
#define L_CELL_TRSP	"c4_trsp"
#define L_CELL_SYMB	"c4_pitch"

/* c4 */

#define VSS		0
#define VDD		0

#define NO_CUT		0
#define RIGHT_CUT	1
#define RIGHTLEFT_CUTS	2
#define EVERY_CUT	3

#define N_OUT		1
#define P_OUT		1


#define CUT_SYMB	"|"
#define MAX_CHAN	20

#define MAX(a,b)	(((a)>(b)) ? (a) : (b))
#define MIN(a,b)	(((a)<(b)) ? (a) : (b))


#define EACH_GRAPH( graph)	(graph = N_GRAPH; graph <= P_GRAPH; graph++)


#define C4_N1		0
#define C4_TRI		1
#define C4_L1		2
#define C4_L1N		3
#define C4_L2		4
#define C4_L2N		5
#define C4_TIE		6
#define C4_HIGH		7
#define C4_LOW		8
#define C4_BD1		9
#define C4_BD1N		10
#define C4_N1P		11
#define C4_BUF		12

#define SL_NUMCANP      2
#define SL_NUMCANN      2

#define MAX_T_SER       30
#define MAX_EQ          (MAX_T_SER * (MAX_T_SER - 1)) + 2
#define MAX_H_PITCH     (MAX_T_SER * MAX_T_SER) + (MAX_T_SER / 2) + 4
#define MAX_CONT        MAX_T_SER + 2

#define C4_INEXPR(c)	((c) ? (c->inExprABL) : (NULL))
#define C4_TREXPR(c)	((c) ? (c->trExprABL) : (NULL))

#define C4_NBINP(c)	((c) ? (c->nbInp) : (-1))
#define C4_NBCUT(c)	((c) ? (c->nbCut) : (-1))
#define C4_NBPITCH(c)	((c) ? (c->nbCut + c->nbInp + c->result->polarity + 1) : (-1))

#define C4_NBPCHAN(c)	((c) ? (c->nbCh[1]) : (-1))
#define C4_NBNCHAN(c)	((c) ? (c->nbCh[0]) : (-1))
#define C4_NBPCHAN0(c)	((c) ? (c->nbMaxCh[1]) : (-1))
#define C4_NBNCHAN0(c)	((c) ? (c->nbMaxCh[0]) : (-1))

#define C4_NBPTRAN(c)	((c) ? (c->result->nbTPSer) : (-1))
#define C4_NBNTRAN(c)	((c) ? (c->result->nbTNSer) : (-1))

#define C4_POLARITY(c)	((c) ? (c->result->polarity) : (-1))

#define C4_ASSOCLST(c)	((c) ? (c->assocList) : (NULL))

#define C4_CHECKok(c)	((c) ? (c->status.checkExprOK) : 0)
#define C4_PLACEok(c)	((c) ? (c->status.placeInputsOK) : 0)
#define C4_CONNECTok(c)	((c) ? (c->status.createInterconOK) : 0)
#define C4_VROUTok(c)	((c) ? (c->status.routeTransOK) : 0)

#define C4_COMPok(c)	((c) ? ((c->status.defMemFigOK)			       \
				&& (c->status.checkExprOK)		       \
				&& (c->status.placeInputsOK)		       \
				&& (c->status.createInterconOK)		       \
				&& (c->status.routeTransOK)) : 0)

#define C4_ROUTok(c)	((c) ? ((C4_COMPok(c))				       \
				&& (C4_NBPCHAN(c) <= C4_NBPCHAN0(c))	       \
				&& (C4_NBNCHAN(c) <= C4_NBNCHAN0(c))) : 0)

#define blkType		loPEq



/****************************************************************************/
/*    Produit :  Compilateur de cellules (Caracterisation)                  */
/****************************************************************************/


typedef struct lores
  {
    struct lores *NEXT;		/* next struct result                     */
    char *NAME;			/* nom de l'entree                        */
    long RUP, RDOWN;		/* resistance de la cellule en ohms       */
    long CAPA;			/* capacite en fento farad                */
    long TPHL, TPLH;		/* Temps en ps                            */
    struct locon *ptLOCON;	/* pointeur sur le connecteur de l'entree */
  }
lores_list;


typedef struct segFree
  {
    short noChL, lenL, noChR, lenR;
    char all;
  }
SegFree;

typedef struct trGraph
  {
    char *trName;
    short loPEq, hiPEq, loNEq, hiNEq;
  }
TrGraph;


typedef struct metalEq
  {
    short noEq;
    short noCh;
    short nbCont;
    short contPos[MAX_CONT];
  }
MetalEq;

typedef struct resCheck
  {
    char polarity;
    char nbTPSer;
    char nbTNSer;
    int nbNot;
  }
ResCheck;

typedef struct funcState
  {
    unsigned char defMemFigOK:1;
    unsigned char checkExprOK:1;
    unsigned char placeInputsOK:1;
    unsigned char createInterconOK:1;
    unsigned char routeTransOK:1;
    unsigned char loadedModel:1;
  }
FuncState;

typedef struct vertSeg
  {
    short xPos, y1, y2;
  }
VertSeg;

typedef struct memFig
  {
    chain_list *inExprABL;
    chain_list *trExprABL;
    chain_list *assocList;
    ResCheck *result;
    TrGraph *tGraph[MAX_H_PITCH];
    MetalEq *mRout[2][MAX_EQ];
    VertSeg segInv, segTrsp;
    char *fName;
    short nbEq[2];
    short nbCut;
    short nbInp;
    short nbCh[2];
    short nbMaxCh[2];
    befig_list *beh;
    FuncState status;
  }
MemFig;


#ifndef __P
# if defined(__STDC__) ||  defined(__GNUC__)
#  define __P(x) x
# else
#  define __P(x) ()
# endif
#endif


/* Prototypes from c4_caraccell.c */

  extern           void  TechnoParamrds __P(());
  extern           void  TechnoParamelp __P(());
  extern           void  param_cell_elp __P((lofig_list *ptlofig));
  extern           void  param_cell_rds __P((lofig_list *ptlofig));
  extern    ptype_list * makepath __P((losig_list *ptsig));
  extern    ptype_list * createbranch __P((lofig_list *ptlofig));
  extern    locon_list * findlocon __P((long index));
  extern           void  linklocon __P((ptype_list *pttype));
  extern    lores_list * addlores __P((char *name, long rup, long rdown, long capa, long tphl, long tplh, locon_list *ptlocon));
  extern           void  veriflores __P((lofig_list *lofig, lores_list *res));
  extern           void  viewlores __P((lofig_list *lofig, lores_list *res));
  extern         double  i_branche __P((ptype_list *maillon0));
  extern    ptype_list * max_res __P((locon_list *ptlocon, int type));
  extern           long  calc_capa __P((locon_list *ptlocon, int type));
  extern           long  calc_res __P((locon_list *ptlocon, int type));
  extern    lores_list * paramcell __P((lofig_list *ptlofig, int capa, int front));
  extern           void  displayLotrs __P((lofig_list *lofig));
  extern    lores_list * caracLofig __P((char *name, int capa, int front));


/* Prototypes from c4_debug.c */

  extern           void  c4_dbg_printInfo __P((MemFig *cell));


/* Prototypes from c4_decomp_bd1.c */

  extern    befig_list * bd1 __P(());


/* Prototypes from c4_decomp_buf.c */

  extern    befig_list * buf __P(());


/* Prototypes from c4_decomp_bus1.c */

  extern    befig_list * bus1 __P(());


/* Prototypes from c4_decomp_dec4.c */

  extern           long  funcNormExprC4 __P((chain_list *expr));
  extern           void  normExprC4 __P((chain_list *expr));
  extern    chain_list * expr2Bst __P((short oper, chain_list *expr));
  extern           void  addNExpr __P((chain_list *e1, chain_list *e2));
  extern    chain_list * c4AssocExpr __P((MemFig *mf, chain_list *expr));
  extern            int  c4MappExprBeh __P((befig_list *beh, chain_list *expr, int mode));
  extern    chain_list * c4GenCell __P((chain_list *expr, int mode));
  extern    chain_list * interfaceBeh __P((befig_list *beh));
  extern           void  c4MappBeh __P((befig_list *beh, int mode, ptype_list *delais));


/* Prototypes from c4_decomp_high1.c */

  extern    befig_list * high1 __P(());


/* Prototypes from c4_decomp_latch1.c */

  extern    befig_list * latch1 __P(());


/* Prototypes from c4_decomp_latch2.c */

  extern    befig_list * latch2 __P(());


/* Prototypes from c4_decomp_low1.c */

  extern    befig_list * low1 __P(());


/* Prototypes from c4_decomp_not1.c */

  extern    befig_list * not1 __P(());


/* Prototypes from c4_decomp_not1p.c */

  extern    befig_list * not1p __P(());


/* Prototypes from c4_decomp_tie.c */

  extern    befig_list * tie __P(());


/* Prototypes from c4_genicon.c */

  extern struct IconGate * genIcon __P((struct befig *figure, int *store_it));
  extern struct IconGate * boxIcon __P((char *cell_name, struct berin *inputs_names, char *output_name, int store_it, int negative_polarity));
  extern struct IconGate * operIcon __P((char *cell_name, struct berin *inputs_names, char *output_name, int store_it, int operator));


/* Prototypes from c4_genlib.c */

  extern    befig_list * genBehave __P((char *fName, chain_list *expr, char polarity, int area));
  extern          char * bd1_p __P((char *ms, char *out, char *i1, char *i2));
  extern          char * bd1n_p __P((char *msn, char *out, char *i1, char *i2));
  extern          char * n1_p __P((char *n1, char *out, char *in));
  extern          char * b1_p __P((char *b1, char *out, char *in));
  extern          char * n1p_p __P((char *n1p, char *out, char *in));
  extern          char * bus_p __P((char *bus, char *out, char *in1, char *in2));
  extern          char * l1_p __P((char *l1, char *out, char *i1, char *i2));
  extern          char * l1n_p __P((char *l1n, char *out, char *i1, char *i2));
  extern          char * l2_p __P((char *l2, char *out, char *i1, char *i2, char *i3, char *i4));
  extern          char * l2n_p __P((char *l2n, char *out, char *i1, char *i2, char *i3, char *i4));
  extern          char * tie_p __P((char *tie));
  extern          char * low_p __P((char *low, char *out));
  extern          char * high_p __P((char *high, char *out));
  extern          char * noEq2nameEq __P((short noEq, char graph, char polarity));
  extern           void  genLogic __P((char *name, TrGraph *tGraph[], char polarity, short nbCh[]));
  extern           void  genLCellIn __P((short dn, short dp));
  extern           void  genLCellOut __P((short dn, short dp));
  extern           void  genLCellInv __P((short dn, short dp));
  extern           void  genLCellTrsp __P((short dn, short dp));
  extern           void  genPhysic __P((char *name, TrGraph *tGr[], MetalEq *mRou[][MAX_EQ], VertSeg sInv, VertSeg sTrsp, short nbCh[], char polarity));
  extern           void  genFiles __P((MemFig *cell, char bFlag, char lFlag, char pFlag, char iFlag));


/* Prototypes from c4_icon.c */

  extern struct IconGate * not1icon __P((char *name, int store_it));
  extern struct IconGate * not1Picon __P((char *name, int store_it));
  extern struct IconGate * buficon __P((char *name, int store_it));
  extern struct IconGate * bus1icon __P((char *name, int store_it));
  extern struct IconGate * high1icon __P((char *name, int store_it));
  extern struct IconGate * low1icon __P((char *name, int store_it));
  extern struct IconGate * tieicon __P((char *name, int store_it));
  extern struct IconGate * latch1icon __P((char *name, int store_it));
  extern struct IconGate * latch1nicon __P((char *name, int store_it));
  extern struct IconGate * latch2icon __P((char *name, int store_it));
  extern struct IconGate * latch2nicon __P((char *name, int store_it));
  extern struct IconGate * bd1icon __P((char *name, int store_it));
  extern struct IconGate * bd1nicon __P((char *name, int store_it));


/* Prototypes from c4_lofig.c */

  extern           void  lofigNot __P((char *name, chain_list *lc_io, char *out, char *in));
  extern           void  lofigNotP __P((char *name, chain_list *lc_io, char *out, char *in));
  extern           void  lofigBuf __P((char *name, chain_list *lc_io, char *out, char *in));
  extern           void  lofigHigh __P((char *name, chain_list *lc_io, char *out));
  extern           void  lofigLow __P((char *name, chain_list *lc_io, char *out));
  extern           void  lofigBus __P((char *name, chain_list *lc_io, char *out, char *i1, char *i2));
  extern           void  lofigTie __P((char *name, chain_list *lc_io));
  extern           void  lofigLatch1 __P((char *name, chain_list *lc_io, char *out, char *i1, char *i2, char *i3));
  extern           void  lofigLatch1N __P((char *name, chain_list *lc_io, char *out, char *i1, char *i2, char *i3));
  extern           void  lofigLatch2 __P((char *name, chain_list *lc_io, char *out, char *i1, char *i2, char *i3, char *i4, char *i5));
  extern           void  lofigLatch2N __P((char *name, chain_list *lc_io, char *out, char *i1, char *i2, char *i3, char *i4, char *i5));
  extern           void  lofigBd1 __P((char *name, chain_list *lc_io, char *out, char *i1, char *i2, char *i3, char *i4));
  extern           void  lofigBd1n __P((char *name, chain_list *lc_io, char *out, char *i1, char *i2, char *i3, char *i4));


/* Prototypes from c4_prepar_balance.c */

  extern    chain_list * nonBalanceExpr __P((befig_list *beh, pTH delaisTH, chain_list *expr, int echantil, int grain));
  extern           void  changeOperRecExpr __P((chain_list *expr, int oper));
  extern            int  hashBalance __P((int optim, int grain, int arity));
  extern    chain_list * decBalExpr __P((chain_list *expr, int optim, int grain));
  extern    chain_list * balanceExpr __P((chain_list *expr, int optim, int grain));
  extern           void  initializeBalance __P(());
  extern           void  filtreExpr __P((chain_list *expr1, chain_list *expr2));
  extern    chain_list * decompBoolExpr __P((befig_list *beh, pTH caracTH, chain_list *expr, int optim, pTH dejaTraite, int grain));
  extern           void  decompBoolBeh __P((befig_list *beh, int optim, ptype_list *delais, chain_list *optimPO, int tp, int tn));


/* Prototypes from c4_prepar_devxor.c */

  extern    chain_list * devXorOptimExpr __P((befig_list *beh, pTH caracTH, pTH occTH, chain_list *expr, int optim, pTH dejaTraite));
  extern           void  devXorBeh __P((befig_list *beh, int optim, ptype_list *delais, chain_list *optimPO));


/* Prototypes from c4_prepar_flatpol.c */

  extern           void  flatPolarityBeh __P((befig_list *beh));


/* Prototypes from c4_prepar_map.c */

  extern    chain_list * createVIExpr __P((befig_list *beh, chain_list *expr, pTH notTH, int polarity, pTH occTH));
  extern            int  numberNot __P((chain_list *expr, pTH occTH, pTH notTH));
  extern    chain_list * mapInvExpr __P((chain_list *expr, pTH notTH, befig_list *beh, pTH occTH));
  extern           void  deleteFilsExpr __P((chain_list *expr, chain_list *filsSelect, int level));
  extern           void  remonteFilsExpr __P((chain_list *expr, chain_list *filsSelect, int level, pTH caracTH, chain_list *ssFils));
  extern            int  numberChain_list __P((chain_list *lc));
  extern    chain_list * maxChain_listExpr __P((chain_list *expr, short oper, pTH caracTH));
  extern            int  existFilsOper __P((chain_list *expr, short oper));
  extern           void  saturationExpr __P((chain_list *expr, int credit, int optim, pTH caracTH));
  extern    chain_list * mapExprC4 __P((befig_list *beh, pTH caracTH, chain_list *expr, int optim, pTH dejaTraite, int tp, int tn, pTH notTH, pTH occTH, int fanout));
  extern           void  calculSerieMax __P((int optim, int *tp, int *tn, int max_tp, int max_tn));
  extern    chain_list * creditExpr __P((befig_list *beh, chain_list *expr, int tp, int tn, int tpLevel, int tnLevel, int optim, pTH caracTH, pTH notTH, pTH occTH, int prof));
  extern           void  mapBeh __P((befig_list *beh, int optim, ptype_list *delais, chain_list *optimPO, int tp_max, int tn_max));


/* Prototypes from c4_prepar_post.c */

  extern    chain_list * phaseExpr __P((chain_list *expr, pTH pTHOcc, befig_list *beh, char *cellN, int sl_aux));
  extern           void  postBeh __P((befig_list *beh, int sl_aux));


/* Prototypes from c4_stretch.c */

  extern          char * c4_genPCells __P((short cellType, char *name, chain_list *lc_io, char *out, char *in1, char *in2, char *in3, char *in4, char *in5));


/* Prototypes from c4_util.c */

  extern    chain_list * expr2ModelInt __P((chain_list *expr, int *i));
  extern    chain_list * expr2Model __P((chain_list *expr));
  extern        MemFig * defMemFig __P((chain_list *expr, char *fileName));
  extern          short  posOper __P((short oper));
  extern    chain_list * flatOper __P((chain_list *expr, short signe));
  extern           void  lowCheckExpr __P((chain_list *expr, ResCheck *res));
  extern           void  checkExpr __P((MemFig *cell));
  extern           void  permutExpr __P((chain_list *expr));
  extern          short  placeArgs __P((chain_list **tabCutA[4][MAX_T_SER], short nbCut[4]));
  extern          short  lowPlaceInputs __P((chain_list *expr));
  extern    chain_list * makeInList __P((chain_list *expr));
  extern    chain_list * assocList __P((chain_list *firstExpr, chain_list *placeExpr));
  extern           void  placeInputs __P((MemFig *cell));
  extern          short  getNewNoEq __P((char graphType));
  extern           void  insertCont __P((short eq, short x, char gTyp, MetalEq *mRou[][MAX_EQ], short nbEq[]));
  extern           void  newTrans __P((char *sigGrid, short hiPEq, short loPEq, short hiNEq, short loNEq, MemFig *cell, int *xPos));
  extern           void  lowCreateInter __P((chain_list *expr, short hiPEq, short loPEq, short hiNEq, short loNEq, MemFig *cell, int *xpos));
  extern           void  createIntercon __P((MemFig *cell));
  extern           void  setChanNo __P((MetalEq *mRou[][MAX_EQ], short nbPit, short nbCh[], short indOut[]));
  extern           void  insertSegOut __P((MetalEq *mRou[][MAX_EQ], short nbPit, short nbCh[], short nbMCh[], SegFree maxFree[], short indOut[]));
  extern           void  insertGraph __P((TrGraph *tGr[], short type, short nbPit, short posOut));
  extern           void  invertCh __P((MetalEq *mRou[][MAX_EQ], short gT, short indCh, short destCh, short nbCh));
  extern           void  moveCont __P((MetalEq *mRou[][MAX_EQ], short posOut, short nbO));
  extern           void  expandSegOut __P((MetalEq *mRou[][MAX_EQ], short gT, short posOut, short indOut, short cp0out, short cplout));
  extern           void  connectSegOut __P((MetalEq *mRou[][MAX_EQ], TrGraph *tGr[], short nbPit, short nbCh[], short nbMCh[], SegFree maxFree[], short indOut[]));
  extern           void  connectSegInv __P((MetalEq *mRou[][MAX_EQ], VertSeg *segInv, VertSeg *segTrsp, TrGraph *tGr[], short nbPit, short nbCh[], short nbMCh[], SegFree maxFree[], short indOut[]));
  extern           void  routeTrans __P((MemFig *cell, short nbNChan, short nbPChan));
  extern           void  undefMemFig __P((MemFig *cell));
  extern        MemFig * c4_comp __P((chain_list *expr, char *fileName, short nNCh, short nPCh, int trace));
  extern           void  c4_gen __P((MemFig *cell));
  extern           void  c4_free __P((MemFig *cell));
#endif
