/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Standard Cell Mapping
 * Date    : 1991,92
 * Author  : Pascale Allegre
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: mapping_mbk.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include MUT_H
#include MLO_H
#include LOG_H
#include BEH_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"


/*------------------------------------------------------------------------------
maxDepthExpr    : calcule la prof. max. d'une expression dans une bst.
------------------------------------------------------
parametres      : une expression, befig et table de hachage.
------------------------------------------------------------------------------*/
void
maxDepthExpr (beh, expr, nCellTH, nConeTH, numberCell, numberCone)
     befig_list *beh;
     chain_list *expr;
     pTH nCellTH;
     pTH nConeTH;
     int *numberCell, *numberCone;
{
  int nCell, nCone;

  char *searchCharOper ();

  if (ATOM (expr))
    {
      nCell = searchTH (nCellTH, VALUE_ATOM (expr));
      nCone = searchTH (nConeTH, VALUE_ATOM (expr));

      if (nCell == EMPTYTH)
	{
	  beaux_list *aux = beh->BEAUX;
	  bebux_list *bux = beh->BEBUX;

	  /* descente recursive */

	  nCell = 0;
	  nCone = 0;

	  while (aux && nCell == 0)
	    {
	      if (!strcmp (aux->NAME, VALUE_ATOM (expr)))
		{
		  maxDepthExpr (beh, aux->ABL, nCellTH, nConeTH, &nCell, &nCone);
		}
	      aux = aux->NEXT;
	    }

	  while (bux && nCell == 0)
	    {
	      if (!strcmp (bux->NAME, VALUE_ATOM (expr)))
		{
		  biabl_list *biabl = bux->BIABL;
		  *numberCell = 0;
		  *numberCone = 0;

		  while (biabl)
		    {
		      if (biabl->VALABL)
			{
			  nCell = 0;
			  nCone = 0;

			  maxDepthExpr (beh, biabl->VALABL, nCellTH, nConeTH, &nCell, &nCone);

			  if (nCell > *numberCell)
			    *numberCell = nCell;
			  if (nCone > *numberCone)
			    *numberCone = nCone;
			}
		      if (biabl->CNDABL)
			{
			  nCell = 0;
			  nCone = 0;

			  maxDepthExpr (beh, biabl->CNDABL, nCellTH, nConeTH, &nCell, &nCone);
			  if (nCell > *numberCell)
			    *numberCell = nCell;
			  if (nCone > *numberCone)
			    *numberCone = nCone;
			}
		      biabl = biabl->NEXT;
		    }
		}
	      bux = bux->NEXT;
	    }

	  addTH (nCellTH, VALUE_ATOM (expr), nCell);
	  addTH (nConeTH, VALUE_ATOM (expr), nCone);
	}

      *numberCell = nCell;
      *numberCone = nCone;
    }
  else
    {
      int polarity;
      char *oper = searchCharOper (OPER (expr));
      *numberCell = 0;
      *numberCone = 0;

      polarity = 2;
      if (oper[0] == 'a' || oper[0] == 'o')	/* positive */
	polarity = 1;
      if (oper[0] == 'n' && oper[1] != 'x')	/* negative */
	polarity = 0;
      if (oper[0] == 'b')	/* bus */
	polarity = 0;

      while ((expr = CDR (expr)))
	{
	  nCell = 0;
	  nCone = 0;
	  maxDepthExpr (beh, CAR (expr), nCellTH, nConeTH, &nCell, &nCone);
	  if (nCell > *numberCell)
	    *numberCell = nCell;
	  if (nCone > *numberCone)
	    *numberCone = nCone;
	}

      *numberCell = *numberCell + 1;
      if (polarity == 0)
	*numberCone = *numberCone + 1;
      else
	*numberCone = *numberCone + 2;

    }
}


/*------------------------------------------------------------------------------
maxDepthBst     : calcule la prof. max. dans une bst.
------------------------------------------------------
parametres      : une befig.
-------------------------------------------------------
return          : un int.
------------------------------------------------------------------------------*/
void
maxDepthBst (beh, numberCell, numberCone)
     befig_list *beh;
     int *numberCell, *numberCone;
{
  beout_list *out;
  bereg_list *reg;
  biabl_list *biabl;
  pTH nCellTH, nConeTH;
  bebus_list *bus;
  int nCell, nCone;

  nCellTH = createTH (500);
  nConeTH = createTH (500);

  *numberCell = 0;
  *numberCone = 0;

  out = beh->BEOUT;
  while (out)
    {
      if (out->ABL)
	{
	  nCell = 0;
	  nCone = 0;

	  maxDepthExpr (beh, out->ABL, nCellTH, nConeTH, &nCell, &nCone);
	  if (nCell > *numberCell)
	    *numberCell = nCell;
	  if (nCone > *numberCone)
	    *numberCone = nCone;
	}
      out = out->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      if (biabl->VALABL)
	{
	  nCell = 0;
	  nCone = 0;

	  maxDepthExpr (beh, biabl->VALABL, nCellTH, nConeTH, &nCell, &nCone);
	  if (nCell > *numberCell)
	    *numberCell = nCell;
	  if (nCone > *numberCone)
	    *numberCone = nCone;
	}
      if (biabl->CNDABL)
	{
	  nCell = 0;
	  nCone = 0;

	  maxDepthExpr (beh, biabl->CNDABL, nCellTH, nConeTH, &nCell, &nCone);
	  if (nCell > *numberCell)
	    *numberCell = nCell;
	  if (nCone > *numberCone)
	    *numberCone = nCone;
	}
      reg = reg->NEXT;
    }


  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      while (biabl)
	{
	  if (biabl->VALABL)
	    {
	      nCell = 0;
	      nCone = 0;

	      maxDepthExpr (beh, biabl->VALABL, nCellTH, nConeTH, &nCell, &nCone);
	      if (nCell > *numberCell)
		*numberCell = nCell;
	      if (nCone > *numberCone)
		*numberCone = nCone;
	    }
	  if (biabl->CNDABL)
	    {
	      nCell = 0;
	      nCone = 0;

	      maxDepthExpr (beh, biabl->CNDABL, nCellTH, nConeTH, &nCell, &nCone);
	      if (nCell > *numberCell)
		*numberCell = nCell;
	      if (nCone > *numberCone)
		*numberCone = nCone;
	    }
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  destroyTH (nCellTH);
  destroyTH (nConeTH);
}

/*-------------------------------------------------------------------------
surface        : 
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
int
Surface (name)
     char *name;
{
  cellList *cell;
  for (cell = mappSys->ptcell; cell; cell = cell->NEXT)
    if (cell->nom == name)
      if (searchGeneric (cell, name, "area") != 0)
	return searchGeneric (cell, name, "area");
      else
	return searchGeneric (cell, name, "surface");

fprintf(stderr,"## ERROR %s.%d## Surface : '%s' inconnu\n",__FILE__,__LINE__, name );
exit(23);
}

/*-------------------------------------------------------------------------
nameInst        : remplace le blanc par un _ 
---------------------------------------------------------------------------
retour          : un pointeur de chaine de caracteres.
---------------------------------------------------------------------------*/
char *
nameInst (name)
     char *name;
{
  char *name1;
  int i;
  int length = strlen (name);

  name1 = (char *) mbkalloc (length + 2);
  strcpy (name1, name);
  i = 0;
  while (name1[i] != '\0')
    {
      if (name1[i] == ' ')
	name1[i] = '_';
      i++;
    }
  name = namealloc (name1);
  mbkfree (name1);

  return (name);
}

/*------------------------------------------------------------------------------
genMBKBeh	 : genere une lofig a partir d'une befig "structurelle" 
--------------------------------------------------------------------------------
parametres 	 : une befig.
--------------------------------------------------------------------------------
return 		 : une lofig.
------------------------------------------------------------------------------*/
lofig_list *
genMBKBeh (beh, capa, sc, nomFig)
     befig_list *beh;
     short capa, sc;
     char *nomFig;
{
  long indexMax = 0;
  losig_list *sigBidon, *losig;
  char *inv = namealloc ("not1");
  int numberCell, numberCone;
  lofig_list *fig = (lofig_list *) NULL;
  losig_list *signal = (losig_list *) NULL;
  int index = 1;
  bepor_list *bp, *bp2;
  beout_list *out;
  bereg_list *reg;
  beaux_list *aux;
  bebus_list *bus;
  bebux_list *bux;
  cellList *cell;
  chain_list *ptChain;
  pTH losigHash;
  sig *sigAux;
  virtuelList *virt;
  int erreur = 0;
  chain_list *supp = (chain_list *) NULL, *aSupp = (chain_list *) NULL;
  ptype_list *regCell = (ptype_list *) NULL, *bdCell = (ptype_list *) NULL;
  cellList *cellLo;
  ptype_list *cellAutre, *cellMod;
  biabl_list *b1abl, *b2abl;


  /* variables pour les statistiques */
  loins_list *inst;
  chain_list *mod;
  int countMod = 0;
  int totalInst = 0;
  int totalSurf = 0;
  int numberNot = 0;

  if (SL_TRACE >= 2)
    {
      displayBehMap (beh, 0);
    }

  /* creation de la table de hashage signal -> losig_list */
  losigHash = createTH (1000);

  /* creation de la lofig */
  fig = addlofig (nomFig);
  printf ("\n");

  /* creation des connecteurs et des signaux */
  /* les signaux */

  reg = beh->BEREG;
  while (reg)
    {
      sigAux = (sig *) testObj (reg->NAME, 's');
      if (capa && sigAux)
	signal = addlosig (fig, index, addchain ((chain_list *) NULL, (void *) gensym ("auxreg", index)), 'I', (float) sigAux->capa / 1000);
      else
	signal = addlosig (fig, index, addchain ((chain_list *) NULL, (void *) gensym ("auxreg", index)), 'I', 0.0);
      index++;
      addTH (losigHash, reg->NAME, (int)signal);
      reg = reg->NEXT;
    }
  aux = beh->BEAUX;
  while (aux)
    {
      sigAux = (sig *) testObj (aux->NAME, 's');
      virt = (virtuelList *) testObj (aux->NAME, 'v');
      if (ATOM (aux->ABL) &&
      ((signal = (losig_list *) searchTH (losigHash, VALUE_ATOM (aux->ABL)))
       != (losig_list *) VIDETH))
	{
	  addTH (losigHash, aux->NAME, (int)signal);
	}
      else
	{
	  if (capa && ((int) virt > 1))
	    if (sigAux)
	      signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) aux->NAME), 'I', (float) sigAux->capa / 1000);
	    else
	      signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) aux->NAME), 'I', (float) virt->capa / 1000);
	  else
	    signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) aux->NAME), 'I', 0.0);
	  addTH (losigHash, aux->NAME, (int)signal);
	}
      aux = aux->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      sigAux = (sig *) testObj (bux->NAME, 's');
      virt = (virtuelList *) testObj (bux->NAME, 'v');

      if (capa && ((int) virt > 1))
	if (sigAux)
	  signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) bux->NAME), 'I', (float) sigAux->capa / 1000);
	else
	  signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) bux->NAME), 'I', (float) virt->capa / 1000);
      else
	signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) bux->NAME), 'I', 0.0);
      addTH (losigHash, bux->NAME, (int)signal);
      bux = bux->NEXT;
    }

  /* les connecteurs */

  bp = (bepor_list *) reverse ((chain_list *) beh->BEPOR);
  beh->BEPOR = bp;
  do
    {
      sigAux = (sig *) testObj (bp->NAME, 's');
      if (bp->DIRECTION == 'O')
	{
	  beout_list *bidule;

	  /* recherche du signal de sortie dans BEOUT */
	  out = beh->BEOUT;
	  while (out && (out->NAME != bp->NAME))
	    out = out->NEXT;
	  if (out)
	    bidule = out;

	  if (out && ATOM (bidule->ABL))
	    {
	      bp2 = beh->BEPOR;
	      while (bp2 && (bp2->NAME != VALUE_ATOM (bidule->ABL)))
		bp2 = bp2->NEXT;
	      if (bp2 && bp2->DIRECTION == 'I')
		{
		  printf ("MBK generation : Error \n---> output connector connected to input connector\n");
		  printf ("\nConnectors : %s - %s\n", bidule->NAME, VALUE_ATOM (bidule->ABL));
		  erreur = 1;
		}
	    }

	  /* recherche d'une sortie egale */
	  out = beh->BEOUT;
	  do
	    if (out && ATOM (out->ABL) && VALUE_ATOM (out->ABL) != nameLow &&
		(bidule->NAME != out->NAME) &&
		(searchTH (losigHash, out->NAME) != VIDETH) &&
		(VALUE_ATOM (bidule->ABL) == VALUE_ATOM (out->ABL)))
	      {
		printf ("\nNetlist generation : Error\n---> 2 connectors linked together\n");
		printf ("\nConnectors : %s - %s -> %s\n", bidule->NAME, out->NAME, VALUE_ATOM (out->ABL));
		erreur = 1;
	      }
	  while ((out = out->NEXT));

	  if (ATOM (bidule->ABL) &&
	      ((signal = (losig_list *) searchTH (losigHash, VALUE_ATOM (bidule->ABL)))
	       != (losig_list *) VIDETH))
	    signal->TYPE = 'E';
	  else
	    {
	      if (capa && sigAux)
		signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) bp->NAME), 'E', (float) sigAux->capa / 1000);
	      else
		signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) bp->NAME), 'E', 0.0);
	    }

	}
      else
	/* ce n'est pas une sortie */
	{
	  if (capa && sigAux)
	    signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) bp->NAME), 'E', (float) sigAux->capa / 1000);
	  else
	    signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) bp->NAME), 'E', 0.0);
	}
      addlocon (fig, bp->NAME, signal, bp->DIRECTION);
      addTH (losigHash, bp->NAME, (int)signal);
    }
  while ((bp = bp->NEXT));
  beh->BEPOR = (bepor_list *) reverse ((chain_list *) beh->BEPOR);


  if (searchTH (losigHash, namealloc ("vdd")) == VIDETH)
    {
      signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) namealloc ("vdd")), 'E', 0.0);
      addTH (losigHash, namealloc ("vdd"), (int)signal);
      addlocon (fig, namealloc ("vdd"), signal, 'X');
    }
  if (searchTH (losigHash, namealloc ("vss")) == VIDETH)
    {
      signal = addlosig (fig, index++, addchain ((chain_list *) NULL, (void *) namealloc ("vss")), 'E', 0.0);
      addTH (losigHash, namealloc ("vss"), (int)signal);
      addlocon (fig, namealloc ("vss"), signal, 'X');
    }

  /* les instances */

  out = beh->BEOUT;
  while (out)
    {
      if (ATOM (out->ABL))	/* cas low ou high */
	cell = (cellList *) testObj (VALUE_ATOM (out->ABL), 'c');
      else
	cell = operToCell (out->ABL);
      if (cell)
	{
	  ptChain = recupLosig (losigHash, cell, out->NAME, out->ABL);
	  if (numberAtomExpr (out->ABL) == 1)
	    numberNot++;
	  addloins (fig, nameInst (out->NAME), getlofig (cell->nom, 'P'), ptChain);
	}
      out = out->NEXT;
    }
  aux = beh->BEAUX;
  while (aux)
    {
      if (ATOM (aux->ABL))
	cell = (cellList *) testObj (VALUE_ATOM (aux->ABL), 'c');
      else
	cell = operToCell (aux->ABL);
      if (cell)
	{
	  ptChain = recupLosig (losigHash, cell, aux->NAME, aux->ABL);
	  if (numberAtomExpr (aux->ABL) == 1)
	    numberNot++;
	  addloins (fig, nameInst (aux->NAME), getlofig (cell->nom, 'P'), ptChain);
	}
      aux = aux->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      biabl_list *biabl = bus->BIABL;
      int cpt = 1;

      do
	if (!ATOM (biabl->VALABL))
	  {
	    cell = operToCell (biabl->VALABL);
	    if (cell == NULL)
	      {
		printf ("MBK generation : Error \n---> bad befig : bad operator %s\n",
			bus->NAME);
		return NULL;
	      }
	    ptChain = recupLosig (losigHash, cell, bus->NAME, biabl->VALABL);
	    addloins (fig, gensym (nameInst (bus->NAME), cpt++), getlofig (cell->nom, 'P'),
		      ptChain);
	  }
      while ((biabl = CDR (biabl)));
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl_list *biabl = bux->BIABL;
      int cpt = 1;

      do
	if (!ATOM (biabl->VALABL))
	  {
	    cell = operToCell (biabl->VALABL);
	    if (cell == NULL)
	      {
		printf ("MBK generation : Error \n---> bad befig : bad operator %s\n",
			bux->NAME);
		return NULL;
	      }
	    ptChain = recupLosig (losigHash, cell, bux->NAME, biabl->VALABL);
	    addloins (fig, gensym (nameInst (bux->NAME), cpt++), getlofig (cell->nom, 'P'),
		      ptChain);
	  }
      while ((biabl = CDR (biabl)));
      bux = bux->NEXT;
    }

  /* recherche de l'index max */
  indexMax = 0;
  for (losig = fig->LOSIG; losig; losig = losig->NEXT)
    if (indexMax < losig->INDEX)
      indexMax = losig->INDEX;

  reg = beh->BEREG;
  while (reg)
    {
      biabl_list *biabl = reg->BIABL;

      if (!ATOM (biabl->VALABL))
	{
	  lofig_list *figReg;
	  chain_list *sign, *newChain = (chain_list *) NULL;
	  locon_list *con;
	  chain_list *inter = (chain_list *) NULL, *interf;
	  Alist *i;

	  cell = operToCell (biabl->VALABL);
	  if (cell == NULL)
	    {
	      printf ("MBK generation : Error \n---> bad befig : bad operator %s\n",
		      reg->NAME);
	      return (NULL);
	    }
	  ptChain = recupLosig (losigHash, cell, reg->NAME, biabl->VALABL);

	  /* probleme des ressources a 2 sorties */

	  figReg = getlofig (cell->nom, 'P');

/*
   restoredirvbe(figReg);
 */

	  for (cellLo = mappSys->ptcell; cellLo && cellLo->nom != figReg->NAME;
	       cellLo = cellLo->NEXT)
	    ;
	  if (cellLo && ((cellLo->type == 'd') || (cellLo->type == 'r')))
	    {
	      for (i = cellLo->interface; i; i = i->NEXT)
		inter = addchain (inter, (void *) i->name);

	      sign = ptChain;
	      con = figReg->LOCON;
	      interf = inter;
	      while (con)
		{
		  /* utiliser renameSig pour SC et rien pour C4 */
		  if ((sc && (!interf || (renameSig (con->NAME) != (char *) interf->DATA))) ||
		  (!sc && (!interf || (con->NAME != (char *) interf->DATA))))
		    {
		      char *bidon = mbkalloc (1024);

/*
   printf("Con %s inexistant \n", con->NAME);
 */
		      indexMax++;
		      sprintf (bidon, "slbidon%ld", indexMax);
		      sigBidon = addlosig (fig, indexMax,
					   addchain ((chain_list *) NULL,
					    (void *) bidon), INTERNAL, 0.0);
		      newChain = addchain (newChain, (void *) sigBidon);
		    }
		  else
		    {
		      newChain = addchain (newChain, sign->DATA);
		      sign = sign->NEXT;
		      interf = interf->NEXT;
		    }
		  con = con->NEXT;
		}
	      freechain (ptChain);
	      ptChain = reverse (newChain);
	      addloins (fig, nameInst (reg->NAME), figReg, ptChain);
	    }
	  else
	    {
	      printf ("MBK generation : Error \n---> '%s' not found\n\n", cell->nom);
	      return (NULL);
	    }
	}
      reg = reg->NEXT;
    }

  if (erreur)
    {
      printf ("MBK generation : Error \n---> Mapping aborted\n");
      return (NULL);
    }

  /* Mise a jour des directions des connecteurs */
  lofigchain (fig);
  if (!restorealldir (fig))
    {
      printf ("MBK generation : Error \n---> restorealldir failed\n\n");
      return (NULL);
    }

  /* recherche des connecteurs INOUT */
  {
    locon_list *con;

    for (con = fig->LOCON; con; con = con->NEXT)
      if (con->DIRECTION == OUT)
	{
	  ptype_list *liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);

	  if (liste->DATA)
	    {
	      chain_list *lstCon = (chain_list *) liste->DATA;

	      if ((lstCon->NEXT) && (lstCon->NEXT->NEXT))
		con->DIRECTION = INOUT;
	    }
	}
  }

  if (0)
    {
      /* ********************************************************************* */
      /* Post-traitement sur les latchs et bascules a 2 sorties : Q et Q' */

      /* construction de listes des bascules et des latchs */
      for (cellLo = mappSys->ptcell; cellLo; cellLo = cellLo->NEXT)
	if (cellLo->type == 'r')
	  regCell = addptype (regCell, (long) NULL, (void *) cellLo);
	else if (cellLo->type == 'd')
	  bdCell = addptype (bdCell, (long) NULL, (void *) cellLo);


      /* recherche des bascules D compatibles */
      for (cellMod = bdCell; cellMod; cellMod = cellMod->NEXT)
	for (cellAutre = bdCell; cellAutre; cellAutre = cellAutre->NEXT)
	  if (cellMod != cellAutre)
	    {
	      short ok = 1;

	      for (b1abl = ((cellList *) cellAutre->DATA)->val_abl,
		   b2abl = ((cellList *) cellMod->DATA)->val_abl;
		   b1abl && b2abl; b1abl = b1abl->NEXT, b2abl = b2abl->NEXT)
		{
		  if (!equalExpr (b1abl->CNDABL, b2abl->CNDABL))
		    ok = 0;
		  if (!equalExpr (b1abl->VALABL, b2abl->VALABL))
		    ok = 0;
		}
	      if (!b1abl && !b2abl && ok)
		{
		  cellMod->TYPE = (long) addchain ((chain_list *) cellMod->TYPE, (void *) cellAutre->DATA);

/*
   printf("  %s et %s compatibles\n", ((cellList *)cellMod->DATA)->nom, ((cellList *)cellAutre->DATA)->nom);
 */
		}
	    }

      /* recherche des registres compatibles */
      for (cellMod = regCell; cellMod; cellMod = cellMod->NEXT)
	for (cellAutre = regCell; cellAutre; cellAutre = cellAutre->NEXT)
	  if (cellMod != cellAutre)
	    {
	      short ok = 1;

	      for (b1abl = ((cellList *) cellAutre->DATA)->val_abl,
		   b2abl = ((cellList *) cellMod->DATA)->val_abl;
		   b1abl && b2abl; b1abl = b1abl->NEXT, b2abl = b2abl->NEXT)
		{
		  if (!equalExpr (b1abl->CNDABL, b2abl->CNDABL))
		    ok = 0;
		  if (!equalExpr (b1abl->VALABL, b2abl->VALABL))
		    ok = 0;
		}
	      if (!b1abl && !b2abl && ok)
		{
		  cellMod->TYPE = (long) addchain ((chain_list *) cellMod->TYPE, (void *) cellAutre->DATA);

/*
   printf("  %s et %s compatibles\n", ((cellList *)cellMod->DATA)->nom, ((cellList *)cellAutre->DATA)->nom);
 */
		}
	    }

      /* parcours des instances et recherche des ressources */
      for (inst = fig->LOINS; inst; inst = inst->NEXT)
	{
	  for (cellLo = mappSys->ptcell; cellLo && cellLo->nom != inst->FIGNAME;
	       cellLo = cellLo->NEXT);
	  if (cellLo && ((cellLo->type == 'd') || (cellLo->type == 'r')))
	    {
	      ptype_list *liste;
	      locon_list *con;
	      chain_list *pl;
	      short nbOut = 0, nbInv = 0;
	      locon_list *outF;
	      losig_list *sigNeg = (losig_list *) NULL;

	      for (con = inst->LOCON; con; con = con->NEXT)
		if (con->DIRECTION == OUT)
		  nbOut++;

	      if (nbOut == 2)
		{

		  for (con = inst->LOCON; con; con = con->NEXT)
		    {
		      if (con->DIRECTION == OUT)
			if (strncmp ((char *) con->SIG->NAMECHAIN->DATA, (char *) sigBidon->NAMECHAIN->DATA, 7))
			  {
			    liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);
/*
   printf("Inst %s [%s] a pour modele %s [%s/%s]\n",
   inst->INSNAME, inst->FIGNAME, cellLo->nomlog, cellLo->nom,
   con->NAME);
 */
			    for (pl = (chain_list *) liste->DATA; pl; pl = pl->NEXT)
			      {
				locon_list *conAux = (locon_list *) pl->DATA;

				if (conAux->ROOT != (void *) inst)
				  if (conAux->TYPE == INTERNAL)
				    {
				      loins_list *fanout = (loins_list *) conAux->ROOT;

				      for (cellLo = mappSys->ptcell;
					   cellLo && cellLo->nom != fanout->FIGNAME;
					   cellLo = cellLo->NEXT);

/*
   printf(" -%s (%s/%s)-\n", conAux->NAME, fanout->FIGNAME, cellLo->nomlog);
 */
				      if (cellLo->nomlog == inv)
					{
					  locon_list *conInv;

					  nbInv++;
					  for (conInv = fanout->LOCON; conInv;
					       conInv = conInv->NEXT)
					    if (conInv->DIRECTION == OUT)
					      sigNeg = conInv->SIG;
					  delloins (fig, fanout->INSNAME);
					}
				    }
			      }
			  }
			else
			  {
			    /* connecteur rattache au signal bidon */
			    outF = con;
			  }
		    }
		  if (nbInv > 1)
		    {
		      printf ("MBK generation : Error \n---> register %s[%s] has more than one inversor\n",
			      inst->INSNAME, inst->FIGNAME);
		      exit (1);
		    }
		  if (sigNeg)
		    {
		      dellosig (fig, outF->SIG->INDEX);
		      outF->SIG = sigNeg;
		    }
		}
	      else if (nbOut == 1)
		{
		  ptype_list *cellEqui;
		  cellList *equival = (cellList *) NULL;
		  loins_list *fanoutInv;

		  /* recherche equivalent 2 sorties */
		  if (cellLo->type == 'r')
		    cellEqui = regCell;
		  else
		    cellEqui = bdCell;
		  while (cellEqui)
		    {
		      if (((cellList *) cellEqui->DATA == cellLo) &&
			  (cellEqui->TYPE != 0))
			equival = (cellList *) ((chain_list *) cellEqui->TYPE)->DATA;
		      cellEqui = cellEqui->NEXT;
		    }
		  if (equival)
		    {
/*
   printf("%s[%s] a comme equivalent %s\n", inst->INSNAME, inst->FIGNAME, equival->nom);
 */
		      for (con = inst->LOCON; con; con = con->NEXT)
			if (con->DIRECTION == OUT)
			  {
			    liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);
			    for (pl = (chain_list *) liste->DATA; pl; pl = pl->NEXT)
			      {
				locon_list *conAux = (locon_list *) pl->DATA;

				if (conAux->ROOT != (void *) inst)
				  if (conAux->TYPE == INTERNAL)
				    {
				      loins_list *fanout = (loins_list *) conAux->ROOT;

				      for (cellLo = mappSys->ptcell;
					   cellLo && cellLo->nom != fanout->FIGNAME;
					   cellLo = cellLo->NEXT);

				      if (cellLo->nomlog == inv)
					{
					  locon_list *conInv;

					  nbInv++;
					  fanoutInv = fanout;
					  for (conInv = fanout->LOCON; conInv;
					       conInv = conInv->NEXT)
					    if (conInv->DIRECTION == OUT)
					      sigNeg = conInv->SIG;
					}
				    }
			      }
			  }
		      if (nbInv == 1)
			{
			  lofig_list *figEqui = getlofig (equival->nom, 'P');
			  chain_list *lstSig = (chain_list *) NULL;
			  char *oldName = inst->INSNAME;
			  locon_list *conEqui;

			  for (conEqui = figEqui->LOCON; conEqui; conEqui = conEqui->NEXT)
			    {
			      for (con = inst->LOCON; con && con->NAME != conEqui->NAME; con = con->NEXT);
			      if (con)
				lstSig = addchain (lstSig, (void *) con->SIG);
			      else
				lstSig = addchain (lstSig, (void *) sigNeg);
			    }

/*
   printf("Remplacement de %s[%s] par %s[%s]\n",
   oldName, inst->FIGNAME, inst->INSNAME, figEqui->NAME);
 */
			  /* destruction du latch a 1 sortie */
			  delloins (fig, inst->INSNAME);
			  /* ajout du latch a 2 sorties */
			  restoredirvbe (figEqui);
			  addloins (fig, oldName, figEqui, reverse (lstSig));
			  /* destruction de l'inverseur */
			  delloins (fig, fanoutInv->INSNAME);
			  lofigchain (fig);
			}
		    }
		}
	    }
	}
    }

  /* detection des portes sans fanout */
  lofigchain (fig);
  for (inst = fig->LOINS; inst; inst = inst->NEXT)
    {
      locon_list *con;
      short nbOut = 0;

      for (con = inst->LOCON; con; con = con->NEXT)
	if (con->DIRECTION == OUT)
	  nbOut++;

      if (nbOut == 1)
	for (con = inst->LOCON; con; con = con->NEXT)
	  if (con->DIRECTION == OUT)
	    {
	      ptype_list *liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);

	      if (!((chain_list *) liste->DATA)->NEXT)
		{
		  if (SL_TRACE)
		    printf ("MBK generation : Warning \n---> Gate '%s' [%s] has no fanout '%s'\n",
			    inst->INSNAME, inst->FIGNAME, con->NAME);
		  supp = addchain (supp, (void *) inst->INSNAME);
		}
	    }
    }

  /* suppression des portes sans fanout */
  for (aSupp = supp; aSupp; aSupp = aSupp->NEXT)
    delloins (fig, (char *) aSupp->DATA);
  if (supp)
    freechain (supp);
  lofigchain (fig);

  /* parcours des modeles : calcul des resultats */
  for (mod = fig->MODELCHAIN; mod; mod = mod->NEXT)
    countMod++;
  for (inst = fig->LOINS; inst; inst = inst->NEXT)
    {
      totalInst++;
      totalSurf += Surface ((char *) inst->FIGNAME);
    }
  maxDepthBst (beh, &numberCell, &numberCone);

  printf ("==============================  FINAL COST  ===================================\n");

  printf ("Number of cells used  = %d\n", countMod);
  printf ("Number of gates used  = %d\n", totalInst);
  printf ("Number of inverters   = %d\n", numberNot);
  printf ("Number of grids       = %d\n", totalSurf);
  printf ("Depth max. (gates)          = %d\n", numberCell);
  printf ("           (eq. neg. gates) = %d\n", numberCone);

  printf ("===============================================================================\n");

  destroyTH (losigHash);

  return fig;
}

chain_list *
recupLosig (losigHash, cell, name, abl)
     pTH losigHash;
     cellList *cell;
     char *name;
     chain_list *abl;
{
  Alist *aux = cell->interface;
  chain_list *res = NULL;

  if (cell->entrees)
    {
      initAlist (abl, cell->entrees);
      do
	{
	  Alist *val = searchAlist (cell->entrees, aux->name);

	  if (val)
	    aux->ptChain = val->ptChain;
	  else if (memberChain_list (aux->name, cell->sorties))
	    aux->ptChain = createAtom (name);
	  else
	    aux->ptChain = createAtom (aux->name);
	}
      while ((aux = CDR (aux)));
    }
  else
    {
      do
	{
	  if (memberChain_list (aux->name, cell->sorties))
	    aux->ptChain = createAtom (name);
	  else
	    aux->ptChain = createAtom (aux->name);
	}
      while ((aux = CDR (aux)));
    }
  aux = cell->interface;
  do
    {
     if (searchTH (losigHash, (char *) CAR (aux->ptChain)) == VIDETH)
	{
	  printf ("WARNING : Cellule %s : %s --> %p\n", cell->nomlog, name, CAR (aux->ptChain));
/*	  displayAlist (cell->interface);
	  exit (-1); */
	}
      res = addchain (res, (void *) searchTH (losigHash, (char *) CAR (aux->ptChain)));
    }
  while ((aux = CDR (aux)));
  if (cell->entrees)
    resetAlist (cell->entrees, 0);
  return res;
}
