/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Standard Cell Mapping
 * Date    : 1991,92
 * Author  : Pascale Allegre
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: mapping_oper.c,v 1.2 1997/09/24 09:24:09 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"


struct typeOper
  {
    int numOper;
    pTH ptOper;
    char **identOper;
  }
 *sysOper;

char *nameUn, *nameZero, *nameHigh, *nameLow;
char *nameLatch, *nameBD, *nameBux, *nameBus, *nameNot1, *nameRest;



void
createTabOper ()
{
  sysOper = (struct typeOper *) mbkalloc (sizeof (struct typeOper));
  sysOper->numOper = 10;
  sysOper->ptOper = createTH (100);
  sysOper->identOper = (char **) mbkalloc (sizeof (char *) * 1000);
}

void
addOper (name)
     char *name;

{
  char **ppName;

  if (searchNumOper (name) != EMPTYTH)
    {
      printf ("addOper : Warning operator already exists -> %s\n", name);
      return;
    }
  addTH (sysOper->ptOper, name, sysOper->numOper);
  ppName = sysOper->identOper + sysOper->numOper - 10;
  *ppName = name;
  sysOper->numOper++;
}

int
searchNumOper (name)
     char *name;

{
  int oper = charToOper (name);

  if (oper != -1)
    return oper;
  oper = searchTH (sysOper->ptOper, name);
  return oper;
}

char *
searchCharOper (num)
     int num;

{
  if (num < 10)
    return operToChar (num);
  if (num > 1000)
    {
      printf ("searchCharOper : Error unknown operator = %d\n", num);
      /* helpme debug raise(SIGSEGV); */
      exit (-1);
    }
  if (num >= sysOper->numOper)
    return (char *) NULL;
  return *(sysOper->identOper + num - 10);
}



/*------------------------------------------------------------------
  IDENT : operToCell
  ------------------------------------------------------------------
  FUNCTION : cherche une cellule a partir d'un operateur 
  ------------------------------------------------------------------ */

cellList *
operToCell (expr)
     chain_list *expr;
{
  char *oper;
  cellList *cell;

  if (ATOM (expr))
    return NULL;
  if (OPER (expr) < 10)
    {
      switch (OPER (expr))
	{
	case NOT:
	  oper = gensym ("not", ARITE (expr));
	  break;
	case AND:
	  oper = gensym ("and", ARITE (expr));
	  break;
	case OR:
	  oper = gensym ("or", ARITE (expr));
	  break;
	case NAND:
	  oper = gensym ("nand", ARITE (expr));
	  break;
	case NOR:
	  oper = gensym ("nor", ARITE (expr));
	  break;
	case XOR:
	  oper = gensym ("xor", ARITE (expr));
	  break;
	case NXOR:
	  oper = gensym ("nxor", ARITE (expr));
	  break;
	}
    }
  else
    oper = searchCharOper (OPER (expr));
  if ((cell = (struct cellule *) testObj (oper, 'c')))
    return cell;
  return NULL;
}

chain_list *
invExpr (abl)
     chain_list *abl;
{
  abl = notExpr (abl);
  addArite (abl);
  return abl;
}
