/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Netlist Optimisation
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: netop_load_util.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"

#include <stdio.h>
#include MUT_H
#include BEH_H
#include MLO_H
#include LOG_H
#include MPH_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"
#include "netop_local.h"


/*---------------------------------------------------------------------------
cheminCritique : affichage du chemin critique
-----------------------------------------------------------------------------
retour         : nom des portes des sorties vers les entrees
---------------------------------------------------------------------------*/
void 
cheminCritique (inst, type, delai, min)
     loins_list *inst;
     int type;			/* NUP ou NDOWN */
     long delai;
     short min;
{
  ptype_list *chemin;
  chain_list *lstCon;
  chain_list *l, *listePorte = (chain_list *) NULL;
  char *nameConExt = NULL;
  loins_list *reg = (loins_list *) NULL;
  int nbCouches = 1;
  int typeParam = type;
  loins_list *porte = inst;
  int val;
  losig_list *sigOut;
  ptype_list *liste;

  val = searchTH (loconOutTH, inst->INSNAME);
  if (val != EMPTYTH)
    {
      sigOut = ((locon_list *) val)->SIG;
      liste = getptype (sigOut->USER, (long) LOFIGCHAIN);

      if (!liste)
	{
	  printf ("ERROR cheminCritique : no type LOFIGCHAIN for %s[%s]\n",
		  inst->INSNAME, inst->FIGNAME);
	  exit (-1);
	}
    }
  else
    {
      printf ("ERROR cheminCritique : no type LOCONOUT for %s[%s]\n",
	      inst->INSNAME, inst->FIGNAME);
      exit (-1);
    }

  printf ("\n");
  for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
    {
      locon_list *conAux = (locon_list *) lstCon->DATA;

      if (conAux->ROOT != (void *) inst)
	if (conAux->TYPE == EXTERNAL)
	  nameConExt = conAux->NAME;
	else if (cellLoadType (((loins_list *) conAux->ROOT)->FIGNAME) == REG)
	  reg = ((loins_list *) conAux->ROOT);
    }

  if (!(reg || nameConExt))
    {
      printf ("ERROR Timing : critical path neither on register nor on external connector (on %s)\n", inst->INSNAME);
      exit (1);
    }

  listePorte = addchain (listePorte, (void *) porte);
  if (min == 0)
    {
      while ((chemin = getptype (porte->USER, (long) ((type == NUP) ? LOINSUP : LOINSDW))))
	{
	  if ((type == NUP) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	    type = NDOWN;
	  else if ((type == NDOWN) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	    type = NUP;
	  porte = (loins_list *) chemin->DATA;
	  listePorte = addchain (listePorte, (void *) porte);
	  nbCouches++;
	}
      chemin = getptype (porte->USER, (long) ((type == NUP) ? LOCONUP : LOCONDW));
    }
  else
    {
      /* chemin critique minimal */
      while ((chemin = getptype (porte->USER, (long) ((type == NUP) ? LOINSMINUP : LOINSMINDW))))
	{
	  if ((type == NUP) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	    type = NDOWN;
	  else if ((type == NDOWN) && (cellLoadPolarity (porte->FIGNAME) == NEG))
	    type = NUP;
	  porte = (loins_list *) chemin->DATA;
	  listePorte = addchain (listePorte, (void *) porte);
	  nbCouches++;
	}
      chemin = getptype (porte->USER, (long) ((type == NUP) ? LOCONMINUP : LOCONMINDW));
    }

  printf ("\t%sCritical path %s = %ld ps\n",
      (min) ? "Minimal " : "", (typeParam == NUP) ? " UP " : "DOWN", delai);

  if (chemin)
    {
      /* connecteur en debut de chemin */
      locon_list *con = (locon_list *) chemin->DATA;

      printf ("\t\tfrom external connector %s\n", con->NAME);
    }
  else
    {
      /* porte en debut de chemin */
      if (cellLoadType (porte->FIGNAME) == REG)
	{
	  printf ("\t\tfrom register %s\n", porte->INSNAME);

	  /* suppression du registre dans le chemin critique */
	  nbCouches--;
	  listePorte = listePorte->NEXT;
	}
      else if (cellLoadType (porte->FIGNAME) == CONST)
	printf ("\t\tfrom gate %s\n", porte->INSNAME);

    }

  if (nameConExt)
    printf ("\t\tto external connector %s\n", nameConExt);
  else if (reg)
    printf ("\t\tto register %s\n", reg->INSNAME);

  if (chemin)
    printf ("[%s]->", ((locon_list *) chemin->DATA)->NAME);
  else if (cellLoadType (porte->FIGNAME) == REG)
    printf ("[%s]->", porte->INSNAME);

  for (l = listePorte; l; l = l->NEXT)
    {
      loins_list *gate = (loins_list *) l->DATA;

      if (NO_TRACE >= 1)
	displayTimingInst (gate, 0, 1);
      else
	printf ("%s->", gate->INSNAME);
/*
 */
    }
  if (nameConExt)
    printf ("[%s]\n", nameConExt);
  else
    printf ("[%s]\n", reg->INSNAME);
  printf ("\t=> %d gates\n\n", nbCouches);
}

/*---------------------------------------------------------------------------
computFanout	: calcul le fanout d'une porte
-----------------------------------------------------------------------------
retour          : ptype_list de loins_list (DATA) et cin (TYPE)
		  chargeOut = somme des capas vues a la sortie
                  nbFanout = nombre de portes vues a la sortie
---------------------------------------------------------------------------*/
ptype_list *
computFanout (porte, con, sCin, nbFanout, buffer)
     loins_list *porte;
     locon_list *con;
     long *sCin, *nbFanout, *buffer;
{
  losig_list *sigOut;
  ptype_list *liste;
  chain_list *lstCon;
  ptype_list *lstFanout = (ptype_list *) NULL;

  if (porte)
    {
      int val = searchTH (loconOutTH, porte->INSNAME);

      if (val == EMPTYTH)
	{
	  printf ("ERROR computFanout : no type LOCONOUT for %s[%s]\n",
		  porte->INSNAME, porte->FIGNAME);
	  exit (-1);
	}
      else
	{
	  sigOut = ((locon_list *) val)->SIG;
	}
    }
  else
    sigOut = con->SIG;

  liste = getptype (sigOut->USER, (long) LOFIGCHAIN);

  if (!liste)
    {
      printf ("ERROR computFanout : No type LOFIGCHAIN for %s\n",
	      getsigname (sigOut));
      exit (-1);
    }

  /* initialisation de sCin, buffer et nbFanout a 0 */
  *sCin = 0;
  *buffer = 0;
  *nbFanout = 0;

  /* parcours du fanout et calcul de la somme des capas */
  for (lstCon = (chain_list *) liste->DATA; lstCon; lstCon = lstCon->NEXT)
    {
      locon_list *conAux = (locon_list *) lstCon->DATA;

      if (conAux->ROOT != (void *) porte)
	{
	  if (conAux->TYPE == EXTERNAL)
	    {
	      /* connecteur externe */
	      *sCin += cellLoadCin ((char *) NULL, conAux->NAME);
	    }
	  else
	    {
	      /* connecteur interne */
	      if (conAux->DIRECTION == IN)
		{
		  loins_list *cellOut = (loins_list *) conAux->ROOT;
		  long cin = cellLoadCin (cellOut->FIGNAME, conAux->NAME);

		  *sCin += cin;
		  (*nbFanout)++;
		  lstFanout = addptype (lstFanout, cin, (void *) cellOut);
		  if (cellLoadType (cellOut->FIGNAME) == BUFF)
		    (*buffer)++;
		}
	      else
		printf ("ERROR computFanout : Unknown internal connector '%s'\n",
			conAux->NAME);
	    }
	}
    }
  return lstFanout;
}

/*---------------------------------------------------------------------------
resistanceNot   :  renvoie la resistance de sortie du NOT (ou 0 si pas de NOT)
-----------------------------------------------------------------------------
retour          : int
---------------------------------------------------------------------------*/
long 
resistanceNot (front)
     int front;
{
  long res;
  char *RES;
  long rNot = 99999;
  cellList *cells = NO_CELLS;

  /* recherche de la cellule logique NOT */
  while (cells && cells->nomlog != NOT1)
    cells = cells->NEXT;

  /* si NOT existe, calcul de la + ptite resistance parmi tous les not */
  if (cells)
    {
      ptype_list *c;

      for (c = cells->surface_tech; c; c = c->NEXT)
	{
	  RES = (char *) mbkalloc (7 + strlen (cells->entrees->name));

	  if (front == NDOWN)
	    {
	      sprintf (RES, "rdown_%s", cells->entrees->name);
	      res = (long) searchGeneric (cells, c->DATA, RES);
	      if (res < rNot)
		rNot = res;
	      mbkfree (RES);
	    }
	  else
	    {
	      sprintf (RES, "rup_%s", cells->entrees->name);
	      res = (long) searchGeneric (cells, c->DATA, RES);
	      if (res < rNot)
		rNot = res;
	      mbkfree (RES);
	    }
	}
      return rNot;
    }
  else
    return 1000;
}

/*---------------------------------------------------------------------------
getInfoInst  :
-----------------------------------------------------------------------------
retour : un long 
---------------------------------------------------------------------------*/
long 
getInfoInst (inst, typeInfo)
     loins_list *inst;
     long typeInfo;
{
  int val = searchTH (loconOutTH, inst->INSNAME);

  if (val != EMPTYTH)
    {
      locon_list *conOut = (locon_list *) val;
      ptype_list *listeInfo = getptype (conOut->SIG->USER, typeInfo);

      if (listeInfo)
	return (long) listeInfo->DATA;
      printf ("ERROR : no timing informations on %s [%s]\n",
	      inst->INSNAME, inst->FIGNAME);
      exit (1);
    }
  else
    {
      printf ("ERROR : no timing informations on %s [%s]\n",
	      inst->INSNAME, inst->FIGNAME);
      exit (1);
    }
}

/*---------------------------------------------------------------------------
getDelayIn  : retourne le delai d'une entree retardee
-----------------------------------------------------------------------------
retour : un long ou 0.
---------------------------------------------------------------------------*/
long 
getDelayIn (liste, conName)
     ptype_list *liste;
     char *conName;
{
  ptype_list *l;

  for (l = liste; l; l = l->NEXT)
    if ((char *) l->DATA == conName)
      return l->TYPE;
  return (long) 0;
}

/*---------------------------------------------------------------------------
getResistIn : retourne le delai d'une entree retardee
-----------------------------------------------------------------------------
retour : un long ou 0.
---------------------------------------------------------------------------*/
long 
getResistIn (liste, conName)
     ptype_list *liste;
     char *conName;
{
  ptype_list *l;

  for (l = liste; l; l = l->NEXT)
    if ((char *) l->DATA == conName)
      return l->TYPE;
  return (long) 0;
}

/*---------------------------------------------------------------------------
tripType : ajoute dans une ptype_list un elt a sa place
           (tri des TYPE)
-----------------------------------------------------------------------------
retour   : ptype_list triee suivant sens (NUP ou NDOWN)
---------------------------------------------------------------------------*/
ptype_list *
tripType (liste, type, data, sens)
     ptype_list *liste;
     long type;
     void *data;
     int sens;
{
  if (!liste)
    return addptype (liste, type, data);

  if ((sens == NUP) && (type < liste->TYPE))
    return addptype (liste, type, data);
  else if ((sens == NDOWN) && (type >= liste->TYPE))
    return addptype (liste, type, data);
  else
    {
      ptype_list *p1 = liste;
      ptype_list *p2 = liste->NEXT;
      ptype_list *new = (ptype_list *) mbkalloc (sizeof (ptype_list));

      while (((sens == NUP) && p2 && (p2->TYPE < type)) ||
	     ((sens == NDOWN) && p2 && (p2->TYPE > type)))
	{
	  p2 = p2->NEXT;
	  p1 = p1->NEXT;
	}
      new->TYPE = type;
      new->DATA = data;
      new->NEXT = p2;
      p1->NEXT = new;
      return liste;
    }
}

/*---------------------------------------------------------------------------
displayCapa : affiche les 
-----------------------------------------------------------------------------
retour      : RIEN. affichage.
---------------------------------------------------------------------------*/
void 
displayCapa (lofig)
     lofig_list *lofig;
{
  losig_list *sig;

  for (sig = lofig->LOSIG; sig; sig = sig->NEXT)
    {
      printf ("  Sig %s[%ld]\tT=%c  Capa = %f\n",
	      getsigname (sig), sig->INDEX, sig->TYPE, 1000.0 * sig->CAPA);
    }
}

/*---------------------------------------------------------------------------
displayTimingInst : affiche les retards, temps requis, slack d'une instance
-----------------------------------------------------------------------------
retour        : RIEN. affichage.
---------------------------------------------------------------------------*/
void 
displayTimingInst (inst, tout, del)
     loins_list *inst;
  /* si tout = 1 tous les signaux sont affiches, sinon que la sortie */
     short tout, del;
{
  locon_list *con;
  ptype_list *userCriticUp = getptype (inst->USER, (long) LOINSUP);
  ptype_list *userCriticDw = getptype (inst->USER, (long) LOINSDW);

  for (con = inst->LOCON; con; con = con->NEXT)
    if (!isvdd (con->NAME) && !isvss (con->NAME))
      if (con->DIRECTION == OUT || con->DIRECTION == TRANSCV ||
	  con->DIRECTION == TRISTATE || tout)
	{
	  ptype_list *userUp = getptype (con->SIG->USER, (long) RETARDUP);
	  ptype_list *userDw = getptype (con->SIG->USER, (long) RETARDDW);
	  ptype_list *profond = getptype (con->SIG->USER, (long) PROF);
	  long capa = (long) searchTH (capaTH, getsigname (con->SIG));

	  ptype_list *requiUp = getptype (con->SIG->USER, (long) REQUIUP);
	  ptype_list *requiDw = getptype (con->SIG->USER, (long) REQUIDW);
	  ptype_list *slackUp = getptype (con->SIG->USER, (long) SLACKUP);
	  ptype_list *slackDw = getptype (con->SIG->USER, (long) SLACKDW);
	  ptype_list *critic = getptype (con->SIG->USER, (long) CRITIC);

	  if (userUp)
	    printf ("  %s [%s/%s] -\tDELAY  UP=%ld - DW=%ld - CAPA=%ld - CRITIC=%ld\n",
		    inst->INSNAME, inst->FIGNAME, con->NAME,
		    (long) userUp->DATA, (long) userDw->DATA,
		    capa, (critic) ? (long) critic->DATA : -1);
	  else
	    printf ("ERROR : no delay for %s [%s/%s]\n",
		    inst->INSNAME, inst->FIGNAME, con->NAME);
	  if (!del)
	    {
	      if (requiUp)
		printf ("\t\t\tREQUI  - UP = %ld - DW = %ld\n",
			(long) requiUp->DATA, (long) requiDw->DATA);
	      else
		printf ("ERROR : no required time for %s [%s/%s]\n",
			inst->INSNAME, inst->FIGNAME, con->NAME);
	      if (slackUp)
		{
		  printf ("\t\t\tSLACK  - UP = %ld - DW = %ld\n",
			  (long) slackUp->DATA, (long) slackDw->DATA);
		  printf ("\t\t\tRESIST - UP = %ld - DW = %ld\n",
			  cellLoadR (inst->FIGNAME, (char *) NULL, NUP),
			  cellLoadR (inst->FIGNAME, (char *) NULL, NDOWN));
		}
	      else
		printf ("ERROR : no slack for %s [%s/%s]\n",
			inst->INSNAME, inst->FIGNAME, con->NAME);
	      if (profond && critic)
		printf ("\t\t\tDEPTH  = %ld -  CRITICALITY = %ld\n",
			(long) profond->DATA, (long) critic->DATA);
	      else
		printf ("ERROR : no depth or criticality for %s [%s/%s]\n",
			inst->INSNAME, inst->FIGNAME, con->NAME);
	    }
	}
  if (cellLoadType (inst->FIGNAME) != REG && cellLoadType (inst->FIGNAME) != CONST)
    {
      if (!del)
	{
	  if (userCriticUp)
	    printf ("\t\t\tCRITIC - UP = %s",
		    ((loins_list *) userCriticUp->DATA)->INSNAME);
	  else
	    printf ("\t\t\tCRITIC - UP = %s",
		    ((locon_list *) getptype (inst->USER, (long) LOCONUP)->DATA)->NAME);
	  if (userCriticDw)
	    printf (" - DW = %s\n", ((loins_list *) userCriticDw->DATA)->INSNAME);
	  else
	    printf (" - DW = %s\n", ((locon_list *) getptype (inst->USER, (long) LOCONDW)->DATA)->NAME);
	}
    }
}

/*---------------------------------------------------------------------------
displayTiming : affiche des retards sur les signaux des instances de la LOFIG
-----------------------------------------------------------------------------
retour        : RIEN. affichage
---------------------------------------------------------------------------*/
void 
displayTiming (lofig, tout)
     lofig_list *lofig;
     short tout;		/* tout = 1, tous les signaux sont affiches, sinon que la sortie */
{
  loins_list *inst;

  printf ("Display timing results\n");

  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    displayTimingInst (inst, tout, 0);

  printf ("End display timing results\n");
}

/*---------------------------------------------------------------------------
afficheChemins	: affiche les retards sur toutes les sorties
-----------------------------------------------------------------------------*/
void 
afficheChemins (fig)
     lofig_list *fig;
{
  loins_list *inst;

  for (inst = fig->LOINS; inst; inst = inst->NEXT)
    {
      if (cellLoadType (inst->FIGNAME) == REG)
	{
/*
   int val = searchTH(loconOutTH, inst->INSNAME);
   losig_list *sigOut = ((locon_list *)val)->SIG;
   ptype_list *liste = getptype(sigOut->USER, (long)LOFIGCHAIN);
   chain_list *lstCon;
 */
	}
      else
	{
	  int val = searchTH (loconOutTH, inst->INSNAME);
	  losig_list *sigOut = ((locon_list *) val)->SIG;

	  if (sigOut->TYPE == EXTERNAL)
	    {
	      long delaiUp = (long) (getptype (sigOut->USER, (long) RETARDUP)->DATA);
	      long delaiDw = (long) (getptype (sigOut->USER, (long) RETARDDW)->DATA);

	      if (delaiUp > delaiDw)
		cheminCritique (inst, NUP, delaiUp, 0);
	      else
		cheminCritique (inst, NDOWN, delaiDw, 0);
	    }
	}
    }
}

/*---------------------------------------------------------------------------
countType	: 
-----------------------------------------------------------------------------
retour		: ptype_list
---------------------------------------------------------------------------*/
ptype_list *
countType (liste, data)
     ptype_list *liste;
     void *data;
{
  ptype_list *l;

  for (l = liste; (l && (l->DATA != data)); l = l->NEXT)
    ;
  if (l)
    {
      l->TYPE = l->TYPE + 1;
      return liste;
    }
  else
    return addptype (liste, 1, data);
}

/*---------------------------------------------------------------------------
displayFanout	: affiche le fanout de chaque porte
-----------------------------------------------------------------------------
retour		: void
---------------------------------------------------------------------------*/
void 
displayFanout (lofig, level)
     lofig_list *lofig;
     short level;
{
  locon_list *con;
  loins_list *inst;

  statistics (lofig, 0);

  /* parcours des connecteurs de la figure */
  for (con = lofig->LOCON; con; con = con->NEXT)
    {
      int nbOut = 0;
      int sCin = 0;
      ptype_list *fanout = (ptype_list *) NULL;
      ptype_list *fan = (ptype_list *) NULL;
      ptype_list *liste;
      chain_list *l;

      if (!isvdd (con->NAME) && !isvss (con->NAME))
	{
	  if (con->DIRECTION == IN)
	    {
	      if (level < 3)
		printf (" CONNECTOR : %s\n", con->NAME);

	      /* liste des liaisons signaux --> connecteurs */
	      liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);

	      for (l = (chain_list *) liste->DATA; l; l = l->NEXT)
		{
		  locon_list *conAux = (locon_list *) (l->DATA);

		  if (conAux->ROOT != (void *) lofig)
		    {
		      if (conAux->TYPE == EXTERNAL)
			{
			  /* connecteurs externes : E/S primaires */
			  printf ("\tExternal %s\n", conAux->NAME);
			}
		      else
			{
			  /* connecteur interne */
			  loins_list *cellOut = (loins_list *) conAux->ROOT;

			  /* portes attaquees par l'instance a traiter */
			  sCin += cellLoadCin (cellOut->FIGNAME, conAux->NAME);
			  nbOut++;
			  fanout = countType (fanout, (void *) cellOut->FIGNAME);
			}
		    }
		}
	      printf ("   %d gates driven - Sun cin = %d\n", nbOut, sCin);
	      for (fan = fanout; fan; fan = fan->NEXT)
		printf ("\tModel %s = %ld\n", (char *) fan->DATA, fan->TYPE);
	    }
	}
    }				/* fin parcours connecteurs */

  /* parcours des instances de la figure */
  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    {
      int nbOut = 0;
      int sCin = 0;
      ptype_list *fanout = (ptype_list *) NULL;
      ptype_list *fan = (ptype_list *) NULL;

      if (level < 3)
	printf ("  INSTANCE : %s [%s] Fanout Max = %ld\n",
		inst->INSNAME, inst->FIGNAME,
		cellLoadFanout (inst->FIGNAME));

      /* parcours des connecteurs de l'instance */
      for (con = inst->LOCON; con; con = con->NEXT)
	{
	  ptype_list *liste;
	  chain_list *l;

	  if (!isvdd (con->NAME) && !isvss (con->NAME))
	    {
	      if (con->DIRECTION == OUT)
		{
		  /* liste des liaisons signaux --> connecteurs */
		  liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);

		  for (l = (chain_list *) liste->DATA; l; l = l->NEXT)
		    {
		      locon_list *conAux = (locon_list *) (l->DATA);

		      if (conAux->ROOT != (void *) inst)
			{
			  if (conAux->TYPE == EXTERNAL)
			    {
			      /* connecteurs externes : E/S primaires */
			      printf ("\tExternal %s\n", conAux->NAME);
			    }
			  else
			    {
			      /* connecteur interne */
			      loins_list *cellOut = (loins_list *) conAux->ROOT;

			      /* portes attaquees par l'instance a traiter */
			      sCin += cellLoadCin (cellOut->FIGNAME, conAux->NAME);
			      nbOut++;
			      fanout = countType (fanout, (void *) cellOut->FIGNAME);
			    }
			}
		    }
		}
	    }
	}			/* fin parcours connecteurs */
      printf ("   %d gates driven - Sun cin = %d\n", nbOut, sCin);
      for (fan = fanout; fan; fan = fan->NEXT)
	printf ("\tModel %s = %ld\n", (char *) fan->DATA, fan->TYPE);
    }
}

/*---------------------------------------------------------------------------
displayLofig	: affiche une lofig (INSTANCE - CONNECTEUR - SIGNAL - 
                                     CONNECTEUR - INSTANCE)
                  level = 0 -> liste des instances et leurs interconnexions
                  level = 1 -> liste des instances et leurs interconnexions
                               + somme des Cin
                  level = 2 -> liste des instances et leurs interconnexions
                               + Resistance * somme des Cin
                  level = 3 -> liste des instances a problemes
                  level = 4 -> liste des instances et les portes qu'elles
                               attaquent
-----------------------------------------------------------------------------
retour		: void
---------------------------------------------------------------------------*/
void 
displayLofig (lofig, level)
     lofig_list *lofig;
     short level;
{
  locon_list *con;
  loins_list *inst;
  int nbPbs = 0;

  statistics (lofig, 0);

  /* parcours des instances de la figure */
  for (inst = lofig->LOINS; inst; inst = inst->NEXT)
    {
      int nbOut = 0;
      ptype_list *fanout = NULL;

      if (level < 3)
	printf ("  INSTANCE : %s [%s] Fanout Max = %ld\n",
		inst->INSNAME, inst->FIGNAME,
		cellLoadFanout (inst->FIGNAME));

      /* parcours des connecteurs de l'instance */
      for (con = inst->LOCON; con; con = con->NEXT)
	{
	  ptype_list *liste;
	  chain_list *l;

	  if (!isvdd (con->NAME) && !isvss (con->NAME))
	    {
	      if (level < 3)
		printf ("   CONNECTOR : %s (%c) --- SIGNAL %s --- ",
			con->NAME, con->DIRECTION,
			(char *) (con->SIG->NAMECHAIN->DATA));

	      if (con->DIRECTION == OUT && level < 3)
		printf ("CELLS :");

	      /* liste des liaisons signaux --> connecteurs */
	      liste = getptype (con->SIG->USER, (long) LOFIGCHAIN);

	      for (l = (chain_list *) liste->DATA; l; l = l->NEXT)
		{
		  if (((locon_list *) (l->DATA))->ROOT != (void *) inst)
		    {
		      if (((locon_list *) (l->DATA))->TYPE == EXTERNAL)
			{
			  /* connecteurs externes : E/S primaires */
			  if (level < 3)
			    if (con->DIRECTION == OUT)
			      printf (" Extern");
			    else
			      printf ("CELLS : Extern");
			  if (con->DIRECTION == OUT)
			    {
			      fanout = addptype (fanout,
				  (long) (((locon_list *) (l->DATA))->NAME),
						 (void *) NULL);
			      nbOut++;
			    }
			}
		      else
			{
			  /* connecteur interne */
			  loins_list *cellOut =
			  (loins_list *) ((locon_list *) (l->DATA))->ROOT;

			  /* DETERMINE LE FANOUT DE CHAQUE CELLULE */

			  if (con->DIRECTION == OUT &&
			      ((locon_list *) (l->DATA))->DIRECTION == IN)
			    {
			      fanout = addptype (fanout,
				  (long) (((locon_list *) (l->DATA))->NAME),
						 (void *) cellOut);
			      nbOut++;
			      if (level < 3)
				printf (" %s,", cellOut->INSNAME);
			    }
			}
		    }
		}
	      if (level < 3)
		printf ("\n");
	    }
	}			/* fin parcours connecteurs */

      if (fanout)
	{
	  int somCin = 0, nbDrived = 0;
	  int charge = 0;
	  ptype_list *f;

	  if (level == 4)
	    {
	      printf ("  INSTANCE : %s [%s] Fanout Max = %ld, drives %d gate%s :\n",
	       inst->INSNAME, inst->FIGNAME, cellLoadFanout (inst->FIGNAME),
		      nbOut, (nbOut > 1) ? "s" : "");
	    }
	  if (level < 3)
	    printf ("     %s drives %d gate%s :",
		    inst->INSNAME, nbOut, (nbOut > 1) ? "s" : "");
	  for (f = fanout; f; f = f->NEXT)
	    {
	      loins_list *att = (loins_list *) f->DATA;

	      /* compteur de portes attaquees */
	      nbDrived++;

	      if (level < 3)
		if (att)
		  printf (" %s/%s -", att->INSNAME, (char *) (f->TYPE));
		else
		  printf (" EXTERN/%s -", (char *) f->TYPE);

	      if (level == 4)
		if (att)
		  printf ("    %s [%s/%s] Cin = %ld\n", att->INSNAME, att->FIGNAME,
			  (char *) (f->TYPE),
			  cellLoadCin (att->FIGNAME, (char *) (f->TYPE)));
		else
		  printf ("    EXTERN [CONNECTOR %s] Cin = %ld\n", (char *) (f->TYPE),
			  cellLoadCin ((char *) NULL, (char *) f->TYPE));
	      if (level > 0)
		if (att)
		  somCin += cellLoadCin (att->FIGNAME, (char *) f->TYPE);
		else
		  somCin += cellLoadCin ((char *) NULL, (char *) f->TYPE);
	    }
	  if (level == 0)
	    printf ("\n");
	  if (level == 1)
	    printf ("\n     Sum Cin = %d\n",
		    somCin);
	  charge = MAX (cellLoadR (inst->FIGNAME, NULL, NDOWN),
			cellLoadR (inst->FIGNAME, NULL, NUP)) * somCin;
	  if (level == 2)
	    {
	      printf ("\n     R Max x Sum Cin = %ld x %d = %d - RC Max = %d\n",
		      MAX (cellLoadR (inst->FIGNAME, NULL, NDOWN),
			   cellLoadR (inst->FIGNAME, NULL, NUP)),
		      somCin, charge, 800);
	    }
	  if (level >= 3 && somCin > cellLoadFanout (inst->FIGNAME))
	    {
	      nbPbs++;

	      if (level == 3)
		{
		  printf ("  INSTANCE : %s [%s] Fanout max = %ld\n",
			  inst->INSNAME, inst->FIGNAME, cellLoadFanout (inst->FIGNAME));
		  printf ("      Sum Cin = %d, %d gate%s driven\n",
			  somCin, nbDrived, nbDrived > 1 ? "s" : "");
		}
	    }
	  if (level == 4)
	    printf ("      Sum Cin = %d%s\n", somCin,
	    (somCin > cellLoadFanout (inst->FIGNAME)) ? "  ! ##### !" : "");
	}
      if (level < 3)
	printf ("\n");
    }
  if (level >= 3)
    printf ("\n Number of problem%s = %d\n\n", nbPbs > 1 ? "s" : "", nbPbs);
}

/*--------------------------------------------------------------------------
displaySol 	: affichage des solutions d'optimisation
----------------------------------------------------------------------------
retour 		: 
--------------------------------------------------------------------------*/
void 
displaySol (sol)
     sol_list *sol;
{
  ptype_list *p, *name;

  printf ("Solution ");
  switch (sol->TYPE)
    {
    case BUN:
      printf ("Buffer with NOT\n");
      break;
    case BUF:
      printf ("BUFFER\n");
      break;
    case DUP:
      printf ("DUPLICATION\n");
      break;
    }
  printf ("Number : %d\n", sol->NOMBRE);
  printf ("Cost in area : %ld\n", sol->SURFACE);
  if (sol->ADAPT)
    printf ("cellule adaptee = %s\n", sol->ADAPT);
  printf ("Cost in cin : %ld\n", sol->DIFFCIN);
  printf ("somme cin : %ld\n", sol->SUMCIN);

  for (p = sol->ARBRE, name = sol->NAME; p; p = p->NEXT, name = name->NEXT)
    {
      ptype_list *q;

      printf ("     * %ld [%s]\n", name->TYPE, (char *) name->DATA);
      for (q = (ptype_list *) p->DATA; q; q = q->NEXT)
	if (q->DATA)
	  printf ("   porte %s - in %s - cin %ld\n", ((loins_list *) q->DATA)->INSNAME,
	   (char *) q->TYPE, cellLoadCin (((loins_list *) q->DATA)->FIGNAME,
					  (char *) q->TYPE));
	else
	  printf ("   con = %s - cin %ld\n", (char *)q->TYPE,
		  cellLoadCin ((char *) NULL, (char *) q->TYPE));
    }
}
