/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Netlist Optimisation
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



 /* $Id: netop_local.h,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $ */

#ifndef NETOP_LOCAL_H
#define NETOP_LOCAL_H

#define MAX(x,y) ((x) > (y)) ? (x) : (y)
#define MIN(x,y) ((x) < (y)) ? (x) : (y)

	/* type */
#define ERR 0
#define COMBI 1
#define REG 2
#define BUS 3
#define CONST 4
#define BUFF 5
#define TIE 6

/* transition */
#define LL 0
#define LH 1
#define HL 2
#define HH 3

	/* resistance */
/* 13/02/95 : UP et DOWN deja defini dans mph */
#define NDOWN  0
#define NUP    1


#define RETARDUP	23
#define RETARDDW	24
#define LOINSUP		25
#define LOINSDW		26
#define LOCONUP		27
#define LOCONDW		28
#define REQUIUP		32
#define REQUIDW		33
#define SLACKUP		34
#define SLACKDW		35
#define CRITIC		36
#define PROF	        30
#define RETARDMINUP	37
#define RETARDMINDW	38
#define LOINSMINUP	39
#define LOINSMINDW	40
#define LOCONMINUP	41
#define LOCONMINDW	42


#define BUN 0
#define BUF 1
#define DUP 2

#define NO_OPTION no_sys->option
#define NO_MODE no_sys->mode
#define NO_VERSION no_sys->version
#define NO_FILE no_sys->file	/* nom + extension */
#define NO_FILEOUT no_sys->fileOut	/* nom de destination */
#define NO_FILELOG no_sys->fileLog	/* chemin + nom + extension */
#define NO_LOG no_sys->log
#define NO_PHG no_sys->phg
#define NO_FILEPAR no_sys->filePar
#define NO_NAME no_sys->name	/* nom */
#define NO_PARAM no_sys->fileParam
#define NO_PAR no_sys->par


#define NO_CAPAPI (no_sys->par)->capaPI
#define NO_BUFFPI (no_sys->par)->buffPI
#define NO_DELAYPI (no_sys->par)->delayPI
#define NO_FACTOR (no_sys->par)->maxSlopeTime
#define NO_CAPAPO (no_sys->par)->capaPO
#define NO_DELAYPO (no_sys->par)->earlyPO
#define NO_RESISTPI (no_sys->par)->impedancePI


#define NO_CATA_LIB no_sys->cata_lib
#define NO_CELLTH no_sys->cellTH
#define NO_CELLS no_sys->cells
#define NO_TRACE no_sys->trace




/*-------------- systeme pour l'optimisation de netlist ---------------

  fileLog	: identificateur du fichier contenant la lofig
  fileParam	: fichier de parametres
  par		: pointeur sur la structure contenant les parametres
  lof		: pointeur sur la befig de l'architecture compilee  
  mode		: mode batch ou non.
  option	: option en batch "f" "t" "g" ou "i"
-----------------------------------------------------------------*/

typedef struct no_syst
  {
    char *name;			/* nom netlist */
    char *file;			/* nom netlist + extension */
    char *filePar;		/* */
    char *fileLog;		/* chemin + nom netlist + extension */
    char *fileParam;		/* chemin + fich param + .lax */
    char *par_lib;		/* chemin des netlists */
    char *fileOut;		/* nom du fichier de sortie */
    ptype_list *cata_lib;	/* chemin des bibliotheques */
    int trace;
    param *par;			/* strucuture de parametres */
    lofig_list *log;		/* pointeur sur la lofig */
    phfig_list *phg;		/* 07/02/95 : pointeur sur la phfig (AsGa) */
    int mode;
    char *version;
    char option;
    cellList *cells;		/* pointeur sur la liste des cellules */
    pTH cellTH;			/* TH des cellules */
  }
no_system;


typedef struct t_sol
  {
    short TYPE;
    long DIFFCIN;
    long SURFACE;
    long SUMCIN;
    ptype_list *NAME;
    ptype_list *ARBRE;
    short NOMBRE;
    char *ADAPT;
    struct t_sol *NEXT;
  }
sol_list;




extern long delaiCriticUp;
extern long delaiCriticDw;
extern long delaiCriticMinUp;
extern long delaiCriticMinDw;
extern loins_list *porteCriticUp;
extern loins_list *porteCriticDw;
extern loins_list *porteCriticMinUp;
extern loins_list *porteCriticMinDw;

extern long profondeurCircuit;
extern loins_list *instProf;

extern ptype_list *listeCritic;

/* Tables de hash */
/* key   : nom d'instance, value = 1 si extremite de chemin */
extern pTH pathTH;
/* key   : nom de signal, value : somme capas sur le signal */
extern pTH capaTH;
/* key   : index de signal_nom de connecteur_transition */
extern pTH delaiTH;
/* value : TP+RC en fct du connecteur et de la transition  */
extern pTH delaiminTH;
/* key   : nom d'instance, value : pointeur sur connecteur de sortie */
extern pTH loconOutTH;
/* key   : nom de signal, value : pointeur sur le connecteur qui drive le signal */
extern pTH genConTH;
/* key   : nom de signal, value : pointeur sur l'instance qui drive le signal */
extern pTH genUpTH;
/* key   : nom de signal, value : pointeur sur l'instance qui drive le signal */
extern pTH genDwTH;


extern int cptBuffer;
extern int nbGates;


extern long indexMaximum;
extern char *NOT1;

extern no_system *no_sys;



#ifndef __P
# if defined(__STDC__) ||  defined(__GNUC__)
#  define __P(x) x
# else
#  define __P(x) ()
# endif
#endif


/* Prototypes from netop_enveloppe.c */

  extern          short  inListe __P((chain_list *liste, char *data));
  extern          short  inListeAtom __P((chain_list *liste, char *data));
  extern    chain_list * completeListe __P((chain_list *liste, Alist *entrees));
  extern          short  longueur __P((chain_list *liste));
  extern          short  egalListe __P((chain_list *liste1, chain_list *liste2));
  extern          short  intersection __P((chain_list *liste1, chain_list *liste2));
  extern          short  testElimineListe __P((chain_list *liste1, chain_list *liste2));
  extern    chain_list * elimineListe __P((chain_list *elt, chain_list *liste));
  extern    chain_list * filtreListeAtom __P((chain_list *liste));
  extern    chain_list * listeAtom __P((chain_list *expr));
  extern           void  displayEnvel __P((chain_list *liste));


/* Prototypes from netop_load_global.c */

  extern           void  displayCombi __P((ptype_list *liste));
  extern    ptype_list * moins __P((ptype_list *liste, void *data));
  extern    ptype_list * addAll __P((ptype_list *liste, void *data));
  extern    ptype_list * combinaison __P((ptype_list *liste));
  extern           void  assignFaninLofig __P((lofig_list *lofig));
  extern           void  assignFaninEx __P((loins_list *inst, short meilleur));
  extern    ptype_list * triFanout __P((loins_list *inst, locon_list *conIn, int sens, loins_list *except));
  extern           void  RAZTiming __P((lofig_list *lofig, loins_list *inst));
  extern           void  puissanceLofig __P((int limite));
  extern           void  chercheBuffer __P((loins_list *porteCritic, long type));
  extern           void  insereBuffer __P((loins_list *instTraite, locon_list *conIn, short *pOk, short *pEncore, char *nameBuff));
  extern           void  remonter __P((loins_list *instTraite, long retardMoyen, long *capaPlus));
  extern           void  arbreBuffer __P((loins_list *instTraite, locon_list *conIn, ptype_list *lstFanout));
  extern           void  bufferLofig __P((int minFanIn, int minFanInst, int limite, int foisLimite));
  extern           void  insereInverseur __P((loins_list *instTraite, locon_list *conIn, short *pOk, short *pEncore, int minFan, char *nameBuff, loins_list *inv1));
  extern           void  inverseurLofig __P((int minFanIn, int minFanInst, int limite, int foisLimite));


/* Prototypes from netop_load_incdelai.c */

  extern           void  RAZsig __P((losig_list *signal));
  extern           void  RAZlofig __P((lofig_list *lofig));
  extern           void  majDelaiCon __P((locon_list *con, long sCin));
  extern           void  majDelaiInst __P((loins_list *inst, long sCin));
  extern           long  capaMoyenne __P(());
  extern           void  prepareLofig __P((lofig_list *lofig));
  extern           void  calcul_retard __P((lofig_list *lofig, short affich));
  extern           long  getDel __P((loins_list *inst, locon_list *con, losig_list *signal, char *trans, short min));
  extern           void  retard __P((loins_list *inst));
  extern           void  calcul_requi __P((lofig_list *lofig, long requiUp, long requiDw, short tout));
  extern           void  requi __P((loins_list *inst, long requiUp, long requiDw, short tout));


/* Prototypes from netop_load_neto.c */

  extern         Alist * memberAlist __P((char *name, Alist *liste));
  extern    chain_list * cellLoadEnvel __P((char *name));
  extern         Alist * cellLoadIn __P((char *name));
  extern    chain_list * cellLoadOut __P((char *name));
  extern            int  cellLoadPolarity __P((char *name));
  extern            int  cellLoadType __P((char *name));
  extern            int  isInverseur __P((char *name));
  extern           long  cellLoadTransistors __P((char *name));
  extern           long  cellLoadSurface __P((char *name));
  extern           long  cellLoadFanout __P((char *name));
  extern           long  cellLoadCin __P((char *name, char *in));
  extern           long  cellLoadR __P((char *name, char *in, int type));
  extern           long  cellLoadTp __P((char *name, char *in, int type));
  extern    chain_list * appendChain __P((chain_list *l1, chain_list *l2));
  extern           void  createCell __P((befig_list *befigList));
  extern          short  cellLoad __P((chain_list *listeModel));
  extern          short  cellLoadNew __P((chain_list *listeModel));
  extern           void  cellInit __P(());


/* Prototypes from netop_load_optim.c */

  extern    chain_list * addmodel __P((chain_list *liste, char *modele));
  extern           void  clearmodel __P((lofig_list *ptfig));
  extern          char * adaptePuissance __P((cellList *liste, long sCin));
  extern           void  addPuissante __P((lofig_list *ptfig, loins_list *inst, char *puissante));
  extern    loins_list * addInversCond __P((lofig_list *ptfig, loins_list *inst, loins_list *inv, locon_list *conIn, char *nameBuff, ptype_list *listeCond));
  extern    loins_list * addBufferCond __P((lofig_list *ptfig, loins_list *inst, locon_list *conIn, char *nameBuff, ptype_list *listeCond));
  extern    loins_list * addBuffer __P((lofig_list *ptfig, loins_list *inst, locon_list *conIn, char *nameBuff));
  extern           void  dupliqueInst __P((lofig_list *ptfig, loins_list *inst, sol_list *solution));
  extern           void  freeSol __P((sol_list *sol));
  extern      sol_list * triInstance __P((ptype_list *fanout, char *nameCell, short type, sol_list *sol, int tout));
  extern           void  calculProf __P((lofig_list *lofig));
  extern           void  profondeur __P((loins_list *inst));
  extern            int  statistics __P((lofig_list *lofig, int surfInit));
  extern           long  capaNot __P(());
  extern    ptype_list * triConptype __P((ptype_list *liste, char *name, void *data));
  extern           void  capaEntrees __P(());
  extern           void  optimEntrees __P((char *conIn, long fanoutMax, int optimIn));
  extern           void  optimiseInst __P((loins_list *instTraite));
  extern           void  optimFanMax __P(());
  extern           void  mainInfo __P((char type));
  extern           void  mainOptimFanMax __P(());
  extern           void  mainOptimGlobal __P(());


/* Prototypes from netop_load_util.c */

  extern           void  cheminCritique __P((loins_list *inst, int type, long delai, short min));
  extern    ptype_list * computFanout __P((loins_list *porte, locon_list *con, long *sCin, long *nbFanout, long *buffer));
  extern           long  resistanceNot __P((int front));
  extern           long  getInfoInst __P((loins_list *inst, long typeInfo));
  extern           long  getDelayIn __P((ptype_list *liste, char *conName));
  extern           long  getResistIn __P((ptype_list *liste, char *conName));
  extern    ptype_list * tripType __P((ptype_list *liste, long type, void *data, int sens));
  extern           void  displayCapa __P((lofig_list *lofig));
  extern           void  displayTimingInst __P((loins_list *inst, short tout, short del));
  extern           void  displayTiming __P((lofig_list *lofig, short tout));
  extern           void  afficheChemins __P((lofig_list *fig));
  extern    ptype_list * countType __P((ptype_list *liste, void *data));
  extern           void  displayFanout __P((lofig_list *lofig, short level));
  extern           void  displayLofig __P((lofig_list *lofig, short level));
  extern           void  displaySol __P((sol_list *sol));


/* Prototypes from netop_main.c */

  extern            int  net_main __P((int argc, char *argv[]));


/* Prototypes from netop_system.c */

  extern           void  calculEnveloppe __P((cellList *listeCellules));
  extern           void  addConDirLofig __P((lofig_list *lofig));
  extern           long  returnIndexMax __P((lofig_list *ptfig));
  extern    lofig_list * initializeNet __P(());
  extern           void  cellEnd __P(());
  extern           void  usageNetOptim __P(());
  extern            int  net_verifArgs __P((int argc, char *argv[]));
  extern           void  majNOSystem __P((no_system *no, int mode, int argc, char *argv[], char *version));
  extern           void  net_displaySystem __P((no_system *sys));
  extern           void  OPTIMISATION_LOCALE __P(());
  extern           void  OPTIMISATION_GLOBALE __P(());
  extern           void  INFORMATION __P((char type));


/* Prototypes from netop_systemp.c */

  extern         param * initializeNOParam __P((int mode, int level, chain_list *intermediate, ptype_list *delayPI, chain_list *earlyPO, int numN, int numP, long maxSlope, ptype_list *impedancePI, ptype_list *capaPO, ptype_list *capaPI, int aux));
  extern     no_system * initializeNOSystem __P(());


#endif
