/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Netlist Optimisation
 * Date    : 1991,92
 * Author  : Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: netop_system.c,v 1.2 1997/09/24 16:13:55 sirol Exp sirol $"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include MUT_H
#include BEH_H
#include MLO_H
#include MPH_H
#include LOG_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"
#include "netop_local.h"


/*---------------------------------------------------------------------------
calculEnveloppe   : 
-----------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
void 
calculEnveloppe (listeCellules)
     cellList *listeCellules;
{
  cellList *cells;

  for (cells = listeCellules; cells; cells = cells->NEXT)
    if (cells->envel)
      {

	cells->envel = completeListe (filtreListeAtom (listeAtom (cells->envel)),
				      cells->entrees);
      }
}

/*---------------------------------------------------------------------------
addConDirLofig   :
-----------------------------------------------------------------------------
retour          :
---------------------------------------------------------------------------*/
void 
addConDirLofig (lofig)
     lofig_list *lofig;
{
  locon_list *connect;

  for (connect = lofig->LOCON; connect; connect = connect->NEXT)
    {
      chain_list *l;
      ptype_list *liste = getptype (connect->SIG->USER, (long) LOFIGCHAIN);
      char dir = ((locon_list *) (((chain_list *) (liste->DATA))->DATA))->DIRECTION;

      for (l = (chain_list *) (liste->DATA); l; l = l->NEXT)
	if (((locon_list *) (l->DATA))->TYPE != EXTERNAL &&
	    dir != INOUT && ((locon_list *) (l->DATA))->DIRECTION != dir)
	  dir = INOUT;
      connect->DIRECTION = dir;
    }
}

/*---------------------------------------------------------------------------
returnIndexMax  : reiurne le numero d'index maximum
-----------------------------------------------------------------------------
retour          : long
---------------------------------------------------------------------------*/
long 
returnIndexMax (ptfig)
     lofig_list *ptfig;
{
  losig_list *sigAux;

  indexMaximum = 0;

  /* parcours des signaux de la figure : calcul de l'index max */
  for (sigAux = ptfig->LOSIG; sigAux; sigAux = sigAux->NEXT)
    {
      if (sigAux->INDEX > indexMaximum)
	indexMaximum = sigAux->INDEX;
    }
  indexMaximum += (indexMaximum * 30) / 100;

  return indexMaximum;
}

/*---------------------------------------------------------------------------
initializeNet   : initialise la lofig et les cellules
-----------------------------------------------------------------------------
retour          : pointeur sur la lofig.
---------------------------------------------------------------------------*/
lofig_list *
initializeNet ()
{
  lofig_list *lofig;

  /* initialisation du systeme de chargement des cellules */
  cellInit ();

  /* chargement de la figure */
  printf ("Loading %s...\n", NO_NAME);
  lofig = getlofig (NO_NAME, 'A');
  printf ("Flattening %s...\n", NO_NAME);
  rflattenlofig (lofig, YES, YES);

/* pour une lofig extraite au niveau transistor par lynx
   displayCapa(lofig);
 */

  if (!lofig->LOINS)
    {
      printf ("No more gates in %s : bad or no file CATAL !\n", lofig->NAME);
      return NULL;
    }

/* permet de faire un executable pour flatten ...
   savelofig(lofig);
   exit(0);
 */

  /* chainages signaux -> connecteurs */
  lofigchain (lofig);

  indexMaximum = returnIndexMax (lofig);

  /* chargement des modeles en memoire : structure *cellList */
  cellLoad (lofig->MODELCHAIN);

  /* calcul de l'enveloppe des cellules */
  calculEnveloppe (NO_CELLS);

  /* mise a jour de la direction des connecteurs d'instances */
  if (!restorealldir (lofig))
    {
      printf ("Adding connection direction failed on %s\n\n", lofig->NAME);
      exit (1);
    }

  /* mise a jour de la direction des connecteurs EXTERNES */
  addConDirLofig (lofig);

  return lofig;
}

/*---------------------------------------------------------------------------
cellEnd         : termine le systeme de chargement de cellules .
-----------------------------------------------------------------------------
retour          : rien
---------------------------------------------------------------------------*/
void 
cellEnd ()
{

  /* destruction de la TH des cellules deja chargees */
  destroyTH (NO_CELLTH);
  destroyBdd (1);

  /* VH_LIBLST reprend son ancienne valeur
     char *newVhdl = (char *)mbkalloc(strlen(vhdl) + 12);
     if (vhdl)
     {
     sprintf(newVhdl,"VH_LIBLST=%s",vhdl);
     printf("vhdl : %s -\n",newVhdl);
     putenv(newVhdl);
     }
   */

  /* destruction des TH de l'analyse de timing */
}

/*----------------------------------------------------------------------------
usageNetOptimptim	: specifie l'utilisation de netOptim
------------------------------------------------------------------------------
retour		: rien 
------------------------------------------------------------------------------*/
void 
usageNetOptim ()
{
  printf ("\nUsage :\n");
  printf ("\tglop -option <file_entity> <file_output> [<file_par>]\n\n");
  printf ("\toption    : -f   local optimization with max fanout\n");
  printf ("\t            -g   global optimization with timing analysis\n");
  printf ("\t            -a   local optimization for AsGa\n");
  printf ("\t            -t   timing informations\n");
  printf ("\t            -i   fanout informations\n");
}

/*----------------------------------------------------------------------------
net_verifArgs	: verifie les arguments de la ligne de commande.
------------------------------------------------------------------------------
retour		: un entier a 1 si OK, 0 sinon.
------------------------------------------------------------------------------*/
int 
net_verifArgs (argc, argv)
     int argc;
     char *argv[];
{
  FILE *fic;
  char *work_lib;
  char *in_lo;

  work_lib = (char *) mbkalloc (5);
  in_lo = (char *) mbkalloc (10);

  work_lib = WORK_LIB;

  if (!existDirectory (work_lib))
    {
      printf ("\nError : the directory '%s' (for MBK_WORK_LIB) doesn't exist\n", work_lib);
      exit (-1);
    }

  in_lo = IN_LO;

  if (argc == 1)
    {
      usageNetOptim ();
      return 0;
    }

  /* test de la ligne de commande */
  if (argv[1][0] != '-')
    {
      /* appel interactif */
      if (argc == 2)
	{
	  usageNetOptim ();
	  return 0;
	}
      if (argc == 3)
	{
	  char *arg1 = mbkalloc (strlen (argv[1]) + strlen (in_lo) + 2);

	  sprintf (arg1, "%s.%s", argv[1], in_lo);
	  /* netoptim source destination */
	  argv[1] = arg1;
	  return (1);
	}
      if (argc == 4)
	{
	  char *arg3 = mbkalloc (strlen (argv[3]) + 5);
	  char *arg1 = mbkalloc (strlen (argv[1]) + strlen (in_lo) + 2);

	  sprintf (arg3, "%s.lax", argv[3]);

	  /* netoptim source destination param */
	  argv[3] = arg3;
	  if (!strcmp (work_lib, "."))
	    {
	      if (!(fic = fopen (argv[3], "r")))
		{
		  printf ("\nError : unable to open %s\n\n", argv[3]);
		  return (0);
		}
	    }
	  else
	    {
	      char *parFile = mbkalloc (strlen (work_lib) + strlen (argv[3]) + 3);

	      sprintf (parFile, "%s/%s", work_lib, argv[3]);
	      if (!(fic = fopen (parFile, "r")))
		{
		  printf ("\nError : unable to open %s\n\n", parFile);
		  return (0);
		}
	      mbkfree (parFile);
	    }
	  sprintf (arg1, "%s.%s", argv[1], in_lo);
	  argv[1] = arg1;
	  return (1);
	}
    }
  /* fin appel version interactive */

  if ((strcmp (argv[1], "-f") && strcmp (argv[1], "-t") &&
       strcmp (argv[1], "-g") && strcmp (argv[1], "-i") && strcmp (argv[1], "-a")) ||
      (argc < 4))
    {
      /* version batch */
      if (argc >= 4)
	printf ("\nError : bad option '%s'\n\n", argv[1]);
      usageNetOptim ();
      return (0);
    }

  /* netlist dans work_lib (MBK_WORK_LIB) */
  {
    char *arg2 = mbkalloc (strlen (argv[2]) + strlen (in_lo) + 2);

    sprintf (arg2, "%s.%s", argv[2], in_lo);
    argv[2] = arg2;
  }

  /* fichier de parametres dans work_lib (MBK_WORK_LIB) */
  if (argc == 5)
    {
      char *arg4 = mbkalloc (strlen (argv[4]) + 5);

      sprintf (arg4, "%s.lax", argv[4]);
      argv[4] = arg4;
      if (!strcmp (work_lib, "."))
	{
	  if (!(fic = fopen (argv[4], "r")))
	    {
	      printf ("\nError      : unable to open '%s'\n\n", argv[4]);
	      return (0);
	    }
	}
      else
	{
	  char *parLib = mbkalloc (strlen (work_lib) + strlen (argv[4]) + 3);

	  sprintf (parLib, "%s/%s", work_lib, argv[4]);
	  if (!(fic = fopen (parLib, "r")))
	    {
	      printf ("\nError      : unable to open %s\n\n", parLib);
	      return (0);
	    }
	  mbkfree (parLib);
	}
    }

  return 1;
}

/*----------------------------------------------------------------------------
majNOSystem	: met a jour la structure SYSTEM
------------------------------------------------------------------------------
retour		: rien
------------------------------------------------------------------------------*/
void 
majNOSystem (no, mode, argc, argv, version)
     no_system *no;
     int mode;			/* mode batch = 1, mode interactif = 0 */
     int argc;
     char *argv[];
     char *version;
{
  short i = 0;

  strcpy (no->version, version);

  if (!existDirectory (WORK_LIB))
    {
      printf ("\nError : the directory '%s' (for MBK_WORK_LIB) doesn't exist\n", WORK_LIB);
      exit (-1);
    }

  if (!getenv ("MBK_CATA_LIB"))
    NO_CATA_LIB = addptype ((chain_list *) NULL, 0, ".");
  else
    {
      char *sauve = getenv ("MBK_CATA_LIB"), *s;
      char *str = (char *) mbkalloc (strlen (sauve) + 1);

      /* constructuion de la liste chainee des differentes bibliotheques */
      sprintf (str, "%s\0", sauve);
      NO_CATA_LIB = (ptype_list *) NULL;
      while (1)
	{
	  if ((s = strchr (str, ':')) == NULL)
	    {
	      if (!existDirectory (str))
		{
		  printf ("\nError : the directory '%s' (for MBK_CATA_LIB) doesn't exist\n", str);
		  exit (-1);
		}
	      NO_CATA_LIB = addptype (NO_CATA_LIB, 0, str);
	      break;
	    }
	  *(s++) = '\0';
	  if (!existDirectory (str))
	    {
	      printf ("\nError : the directory '%s' (for MBK_CATA_LIB) doesn't exist\n", str);
	      exit (-1);
	    }
	  NO_CATA_LIB = addptype (NO_CATA_LIB, 0, str);
	  str = s;
	}
    }

  no->par_lib = WORK_LIB;

  if (getenv ("MBK_TRACE") == NULL)
    no->trace = 0;
  else
    no->trace = atoi (getenv ("MBK_TRACE"));

  no->mode = mode;
  if (mode == 1)
    {
      /* Mode BATCH */
      strcpy (no->name, argv[2]);
      strcpy (no->file, argv[2]);
      strcpy (no->fileOut, argv[3]);

      if (!strcmp (WORK_LIB, "."))
	strcpy (no->fileLog, argv[2]);
      else
	sprintf (no->fileLog, "%s/%s", WORK_LIB, argv[2]);

      no->par = initializeNOParam (2, 2, NULL, NULL, NULL, 4, 4, 1000, NULL, NULL, NULL, 1);
      if (argc == 5)
	{
	  strcpy (no->filePar, argv[4]);

	  if (!strcmp (no->par_lib, "."))
	    strcpy (no->fileParam, argv[4]);
	  else
	    sprintf (no->fileParam, "%s/%s", no->par_lib, argv[4]);

	  if ((no->par = loadParam (no->fileParam)) == NULL)
	    {
	      printf ("\nError : loadParam error : bad parameters file '%s'\n", no->fileParam);
	      exit (-1);
	    }
	}
      else
	{
	  strcpy (no->filePar, "default.lax");
	  if (!strcmp (no->par_lib, "."))
	    strcpy (no->fileParam, "default.lax");
	  else
	    sprintf (no->fileParam, "%s/default.lax", no->par_lib);
	}
      no->option = *(argv[1] + 1);
    }
  else
    /* mode 0 */
    {
      /* mode interactif */
      strcpy (no->name, argv[1]);
      strcpy (no->file, argv[1]);
      strcpy (no->fileOut, argv[2]);

      if (!strcmp (WORK_LIB, "."))
	strcpy (no->fileLog, argv[1]);
      else
	sprintf (no->fileLog, "%s/%s", WORK_LIB, argv[1]);

      no->par = initializeNOParam (2, 2, NULL, NULL, NULL,
				   4, 4, 1000, NULL, NULL, NULL, 1);
      if (argc != 4)
	{
	  strcpy (no->filePar, "default.lax");
	  if (!strcmp (no->par_lib, "."))
	    strcpy (no->fileParam, "default.lax");
	  else
	    sprintf (no->fileParam, "%s/default.lax", no->par_lib);
	}
      else
	/* mode 2 */
	{
	  strcpy (no->filePar, argv[3]);
	  if (!strcmp (no->par_lib, "."))
	    strcpy (no->fileParam, argv[3]);
	  else
	    sprintf (no->fileParam, "%s/%s", no->par_lib, argv[3]);
	  if ((no->par = loadParam (no->fileParam)) == NULL)
	    {
	      printf ("\nError : loadParam error : bad parameters file '%s'\n", no->fileParam);
	      exit (-1);
	    }
	}
      no->option = '\0';
    }

  while ((no->name)[i] != '.')
    i++;
  (no->name)[i] = '\0';
}

/*----------------------------------------------------------------------------
net_displaySystem   : affiche les arguments de l'optimisation
------------------------------------------------------------------------------
retour		: void 
------------------------------------------------------------------------------*/

void
net_displaySystem (sys)
     no_system *sys;
{

  if (NO_MODE == 1)		/* mode batch */
    {
      printf ("     =========================  Environnement  =========================\n");

      printf ("     MBK_WORK_LIB        = %s\n", WORK_LIB);
      printf ("     MBK_CATA_LIB        = %s\n", getenv ("MBK_CATA_LIB"));
      if (NO_TRACE > 0)
	printf ("     MBK_TRACE           = %d\n", NO_TRACE);
      printf ("     =================  Files, Options and Parameters  =================\n");
      printf ("     Netlist file        = %s\n", NO_FILELOG);
      if (strcmp (WORK_LIB, "."))
	printf ("     Output  file        = %s/%s.%s\n",
		WORK_LIB, NO_FILEOUT, OUT_LO);
      else
	printf ("     Output  file        = %s.%s\n", NO_FILEOUT, OUT_LO);
      printf ("     Parameters file     = %s\n", NO_PARAM);
      printf ("     Mode                = ");
      switch (NO_OPTION)
	{
	case 'f':
	  printf ("local optimization with max fanout\n");
	  break;
	case 'g':
	  printf ("global optimization with timing analysis\n");
	  break;
	case 't':
	  printf ("timing informations\n");
	  break;
	case 'i':
	  printf ("fanout informations\n");
	  break;
	}
      printf ("     ===================================================================\n");
      printf ("\n");
    }
}


/*----------------------------------------------------------------------------
OPTIMISATION_LOCALE : optimisation locale d'une netlist 
------------------------------------------------------------------------------
retour		: void 
------------------------------------------------------------------------------*/
void 
OPTIMISATION_LOCALE ()
{
  mainOptimFanMax ();
}

/*----------------------------------------------------------------------------
OPTIMISATION_GLOBALE : optimisation globale d'une netlist 
------------------------------------------------------------------------------
retour		: void 
------------------------------------------------------------------------------*/
void 
OPTIMISATION_GLOBALE ()
{
  mainOptimGlobal ();
  NO_LOG->NAME = NO_FILEOUT;
  printf ("Saving %s/%s...\n", WORK_LIB, NO_LOG->NAME);
  savelofig (NO_LOG);
}

/*----------------------------------------------------------------------------
INFORMATION     : information sur une netlist 
------------------------------------------------------------------------------
retour		: void 
------------------------------------------------------------------------------*/
void 
INFORMATION (type)
     char type;
{
  mainInfo (type);
}
