/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis
 * Date    : 1991,92
 * Author  : Luc Burgun, Pascale Allegre, Nathalie Dictus
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident "$Id: scmap_main.c,v 1.3 1997/10/06 16:57:28 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include BEH_H
#include LOG_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"


/*
 * y en a qui vont rigoler en voyant ce source.
 * j'ai fait de mon mieux pour essayer de debugger
 * tout ca, mais il faudrait des annees!
 * le mieux serait sans doute de tout reecrire.
 * 
 *                                            Czo.
 */


char *GLOB_ABL_DC;
char *GLOB_ABL_ONE;
char *GLOB_ABL_ZERO;


/*AFAC*/
char *nameInt (char *name);


void Mysavelofig( Figure )

  lofig_list *Figure;
{
  losig_list  *Signal;
  losig_list **PrevSignal;
  losig_list  *DelSignal;
  ptype_list  *PType;

  for ( Signal  = Figure->LOSIG;
        Signal != (losig_list *)0;
        Signal  = Signal->NEXT )
  {
    Signal->USER = (void *)0;
  }

  lofigchain( Figure );

  Signal     = Figure->LOSIG;
  PrevSignal = &Figure->LOSIG;

  while ( Signal != (losig_list *)0 )
  {
    PType = Signal->USER;

    if ( PType->DATA == (void *)0 )
    {
	fprintf( stdout, "## Warning : Deleting unused signal index %ld\n", Signal->INDEX );

      DelSignal = Signal;
      Signal    = Signal->NEXT;

/*
      mbkfree( DelSignal );
*/

      *PrevSignal = Signal;
    }
    else
    {
      PrevSignal = &Signal->NEXT;
      Signal     = Signal->NEXT;
    }
  }

  savelofig( Figure );
}

/*======================================================================/
/                               MAIN                                    /
/======================================================================*/

int
main (argc, argv)
     int argc;
     char *argv[];
{
  int i;
  char *dummy_argv[4096];
  int dummy_argc;

  dummy_argv[0]=mbkstrdup(argv[0]);
  dummy_argv[1]="-s";

  for (i=1; i <= argc ; i++)
  {
        dummy_argv[i+1]=mbkstrdup(argv[i]);
  }
        dummy_argv[argc+1]=NULL;      

  dummy_argc=argc+1;

/*  printf("Argc = %d\n", argc);
  i=0;
  while (argv[i])
        printf("Argv[%d] = %s\n", i,argv[i++]);

  printf("dummy_Argc = %d\n", dummy_argc);
  i=0;
  while (dummy_argv[i])
        printf("dummy_Argv[%d] = %s\n", i,dummy_argv[i++]);
*/

  main2 (dummy_argc, dummy_argv);
}

int
main2 (argc, argv)
     int argc;
     char *argv[];
{
  char *carint;
  int trace;
  lofig_list *fig;

/* SigHandlerInitialize(); */

  mbkenv ();
  alliancebanner ("SCmap", LOGIC_VERSION, "Mapping Standard Cells", "1990", ALLIANCE_VERSION);

  GLOB_ABL_DC = namealloc ("'d'");
  GLOB_ABL_ONE = namealloc ("'1'");
  GLOB_ABL_ZERO = namealloc ("'0'");

  /* positionnement de l'environnement */


  VHB_HEDFIG = NULL;
  sl_sys = initializeSLSystem ();
  majSystem (sl_sys);

  /* verification de la validite des arguments */

  if (!verifArgs (sl_sys, argc, argv))
    exit (-1);

  initializeBdd (1);

  /* initialisation selon le mode (batch ou non) */

  if ( !strcmp (argv[1], "-s"))
    {

      /* initialisation du systeme */

      addInfoSystem (sl_sys, 1, argc, argv, LOGIC_VERSION);

      displaySystem (sl_sys);
      trace = 1;
      if (SL_TRACE == 2)
	trace = 2;

      put_CATA_LIB (SL_CATA_SAVE);

      if (SL_OPTION == 's' || SL_OPTION == 'c')
	{
	  if (compileVHDL (SL_VHDL, SL_ORDERPI, SL_ORDER, SL_PAR,
			   SL_AUX, 1, 1, 0, trace, nameInt, NULL) == NULL)
	    {
	      printf ("*** Compilation aborted...\n");
	      exit (-1);
	    }
	}
      else
	{
	  int vhb_reorder = 0;

	  if (SL_OPTION == 'l')
	    {
	      SL_AUX = 1;
	      /* on garde les aux, donc il faut degager la liste des keep (Czo) */
	      SL_GUARDEDAUX = NULL;

	    }

	  if (SL_MAKEORDER >= 1)
	    vhb_reorder = 2;

	  sl_sys->orderPI = compileVHDL (SL_VHDL, SL_ORDERPI, SL_ORDER, SL_PAR,
			   SL_AUX, 1, vhb_reorder, 1, trace, nameInt, NULL);


	  if (SL_ORDERPI == NULL)
	    {
	      printf ("*** Compilation aborted...\n");
	      exit (-1);
	    }
	}

      put_CATA_LIB ((char *) genNAME ("%s:%s", SL_CATA_LIB, SL_CCCC_LIB));

      SL_BEH = VHB_HEDFIG;
      carint = (char *) mbkalloc (strlen (SL_BEH->NAME) + 2);

      /* verification de la coherence */

      verifCohBeh (SL_BEH, SL_DELAYPI, SL_OPTIMPO);

      if (SL_TRACE)
	displayBeh (SL_BEH, 1);

      if (SL_OPTION == 'o' || SL_OPTION == 'l')		/* OPTIMISATION */
	{
	  if (SL_OPTION == 'l')
	    {
	      SL_AUX = 1;
	      /* on garde les aux, donc il faut degager la liste des keep (Czo) */
	      SL_GUARDEDAUX = NULL;

	    }


	  annealingLOGIC (SL_TRACE);

	  decompileVHDL (SL_BEH, SL_FILEOUT, 1);
	  mbkfree (carint);
	}
      else
	/* PROJECTION OU DECOMPOSITION STRUCTURELLE */
	{
	  /* elimination des DC */

	  removeDCBeh (SL_BEH);
	  mapCarExprBeh (SL_BEH, simplif10Expr);

	  if (SL_OPTION == 's')
	    fig = (lofig_list *) mappingSC ();
	  else
	    fig = (lofig_list *) mappingC4 ();

	  /* Generation de la netlist sur disque */

	  if (fig)
	    {
	      printf ("MBK Driving '%s/%s.%s'...\n", SL_FILE_LIB, fig->NAME, SL_OUT_LO);
	      /*Mysavelofig (fig);*/
	      Mysavelofig (fig);
	      dellofig (fig->NAME);
	    }
	}
    }
  else
    {
      /* version interactive */
      printf ("\nscmap : Command line error...\nUse -s switch\n");
    }
  printf ("\n");
  exit (0);
}
