/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: sl_facto.c,v 1.2 1997/10/08 14:37:49 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"

/*AFAC*/
int numberAtomRedFact (pNode pt);


#define PROCRASTINATION 1
#define MAX_FATHER 1000


pTH pTabExpr;

/*-------------------------------------------------------------------------
coutAblBeh 	: calcule le cout d'une befig avec ces ABL.
			- lit : nombre de litteraux non reduits.
			- maxLit : max du nombre de litteraux non reduits.
			- litRed : nombre de litteraux reduits.
			- litReg : nombre de litteraux equivalent registre.
			- profMax : profondeux max.
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
coutAblBeh (beh, lit, maxLit, litRed, profMax, delaisMax, operBin, litReg, mode, delayPI)
     befig_list *beh;
     int *lit, *maxLit, *litRed, *profMax, *delaisMax, *operBin, *litReg;
     int mode;
     ptype_list *delayPI;
{
  beout_list *out;
  beaux_list *aux;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  biabl_list *biabl;
  long numberAtomExprBehInt;
  long numberAtomExprBehIntVal;
  long numberAtomExprBehIntCnd;

  int profExprBehInt;
  int profExprBehIntVal;
  int profExprBehIntCnd;

  /* initialisation */

  long FACTO_lit = 0;
  long FACTO_maxLit = 0;
  int FACTO_litRed = 0;
  int FACTO_litReg = 0;
  int FACTO_profMax = 0;
  int FACTO_delaisMax = 0;
  int FACTO_numberOperBin = 0;

  pTH caracTH;
  pTH caracBeh ();
  pTH profTH;
  pTH numberTH;
  /* caracterisation */

  caracTH = caracBeh (beh, delayPI, 0);

  profTH = createTH (100);
  numberTH = createTH (100);


  out = beh->BEOUT;
  while (out)
    {
      if (out->ABL)
	{
	  FACTO_litRed = FACTO_litRed + numberAtomExpr (out->ABL);
	  FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (out->ABL);
	  numberAtomExprBehInt = numberAtomExprBeh (beh, out->ABL, numberTH);
	  FACTO_lit = FACTO_lit + numberAtomExprBehInt;
	  profExprBehInt = profExprBeh (beh, out->ABL, profTH);

	  if (mode == 1)
	    {
	      printf ("%s = ", out->NAME);
	      printf ("(%ld", numberAtomExprBehInt);
	      printf (", %d", profExprBehInt);
	      printf (", %d)\n", searchTH (caracTH, (char *)out->ABL));
	    }

	  if (numberAtomExprBehInt > FACTO_maxLit)
	    FACTO_maxLit = numberAtomExprBehInt;

	  if (profExprBehInt > FACTO_profMax)
	    FACTO_profMax = profExprBehInt;

	  if (searchTH (caracTH, (char *)out->ABL) > FACTO_delaisMax)
	    FACTO_delaisMax = searchTH (caracTH, (char *)out->ABL);
	}
      out = out->NEXT;
    }

  aux = beh->BEAUX;
  while (aux)
    {
      if (aux->ABL)
	{
	  if (mode == 1)
	    {
	      printf ("%s = ", aux->NAME);
	      printf ("(%ld", numberAtomExprBeh (beh, aux->ABL, numberTH));
	      printf (", %d", profExprBeh (beh, aux->ABL, profTH));
	      printf (", %d)\n", searchTH (caracTH, (char *)aux->ABL));
	    }

	  FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (aux->ABL);
	  FACTO_litRed = FACTO_litRed + numberAtomExpr (aux->ABL);
	}
      aux = aux->NEXT;
    }


  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      if (mode == 1)
	printf ("%s = ", reg->NAME);

      if (biabl->USER || searchOperExpr (biabl->CNDABL, STABLE))	/* bascule D */
	FACTO_litReg = FACTO_litReg + 2;
      else			/* latch */
	FACTO_litReg = FACTO_litReg + 1;

      while (biabl)
	{

	  if (biabl->CNDABL && biabl->VALABL)
	    {

	      if (searchOperExpr (biabl->CNDABL, STABLE))
		{
		  FACTO_litRed++;
		}
	      else
		{
		  FACTO_litRed = FACTO_litRed + numberAtomExpr (biabl->CNDABL);
		  FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (biabl->CNDABL);
		}

	      numberAtomExprBehIntCnd = numberAtomExprBeh (beh, biabl->CNDABL, numberTH);
	      FACTO_lit = FACTO_lit + numberAtomExprBehIntCnd;
	      profExprBehIntCnd = profExprBeh (beh, biabl->CNDABL, profTH);

	      FACTO_litRed = FACTO_litRed + numberAtomExpr (biabl->VALABL);
	      FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (biabl->VALABL);
	      numberAtomExprBehIntVal = numberAtomExprBeh (beh, biabl->VALABL, numberTH);
	      profExprBehIntVal = profExprBeh (beh, biabl->VALABL, profTH);
	      FACTO_lit = FACTO_lit + numberAtomExprBehIntVal;

	      if (mode == 1)
		{
		  printf ("(%ld", numberAtomExprBehIntVal);
		  printf (", %d", profExprBehIntVal);
		  printf (", %d) ", searchTH (caracTH, (char *)biabl->VALABL));
		  printf ("(%ld", numberAtomExprBehIntCnd);
		  printf (", %d", profExprBehIntCnd);
		  printf (", %d) ", searchTH (caracTH, (char *)biabl->CNDABL));
		}

	      if (numberAtomExprBehIntVal > FACTO_maxLit)
		FACTO_maxLit = numberAtomExprBehIntVal;

	      if (profExprBehIntVal > FACTO_profMax)
		FACTO_profMax = profExprBehIntVal;

	      if (searchTH (caracTH, (char *)biabl->VALABL) > FACTO_delaisMax)
		FACTO_delaisMax = searchTH (caracTH, (char *)biabl->VALABL);

	      if (numberAtomExprBehIntCnd > FACTO_maxLit)
		FACTO_maxLit = numberAtomExprBehIntCnd;

	      if (profExprBehIntCnd > FACTO_profMax)
		FACTO_profMax = profExprBehIntCnd;

	      if (searchTH (caracTH, (char *)biabl->CNDABL) > FACTO_delaisMax)
		FACTO_delaisMax = searchTH (caracTH, (char *)biabl->CNDABL);
	    }
	  biabl = biabl->NEXT;
	}
      if (mode)
	printf ("\n");
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      if (mode == 1)
	printf ("%s = ", bus->NAME);

      biabl = bus->BIABL;
      while (biabl)
	{
	  if (biabl->CNDABL && biabl->VALABL)
	    {

	      FACTO_litRed = FACTO_litRed + numberAtomExpr (biabl->VALABL);
	      FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (biabl->VALABL);
	      numberAtomExprBehIntVal = numberAtomExprBeh (beh, biabl->VALABL, numberTH);
	      profExprBehIntVal = profExprBeh (beh, biabl->VALABL, profTH);
	      FACTO_lit = FACTO_lit + numberAtomExprBehIntVal;

	      FACTO_litRed = FACTO_litRed + numberAtomExpr (biabl->CNDABL);
	      FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (biabl->CNDABL);
	      numberAtomExprBehIntCnd = numberAtomExprBeh (beh, biabl->CNDABL, numberTH);
	      profExprBehIntCnd = profExprBeh (beh, biabl->CNDABL, profTH);
	      FACTO_lit = FACTO_lit + numberAtomExprBehIntCnd;

	      if (mode == 1)
		{
		  printf ("(%ld", numberAtomExprBehIntVal);
		  printf (", %d", profExprBehIntVal);
		  printf (", %d) ", searchTH (caracTH, (char *)biabl->VALABL));
		  printf ("(%ld", numberAtomExprBehIntCnd);
		  printf (", %d", profExprBehIntCnd);
		  printf (", %d) ", searchTH (caracTH, (char *)biabl->CNDABL));
		}

	      if (numberAtomExprBehIntVal > FACTO_maxLit)
		FACTO_maxLit = numberAtomExprBehIntVal;

	      if (profExprBehIntVal > FACTO_profMax)
		FACTO_profMax = profExprBehIntVal;

	      if (searchTH (caracTH, (char *)biabl->VALABL) > FACTO_delaisMax)
		FACTO_delaisMax = searchTH (caracTH, (char *)biabl->VALABL);

	      if (numberAtomExprBehIntCnd > FACTO_maxLit)
		FACTO_maxLit = numberAtomExprBehIntCnd;

	      if (profExprBehIntCnd > FACTO_profMax)
		FACTO_profMax = profExprBehIntCnd;

	      if (searchTH (caracTH, (char *)biabl->CNDABL) > FACTO_delaisMax)
		FACTO_delaisMax = searchTH (caracTH, (char *)biabl->CNDABL);
	    }
	  biabl = biabl->NEXT;
	}

      if (mode)
	printf ("\n");
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      if (mode == 1)
	printf ("%s = ", bux->NAME);

      biabl = bux->BIABL;
      while (biabl)
	{
	  if (biabl->CNDABL && biabl->VALABL)
	    {

	      FACTO_litRed = FACTO_litRed + numberAtomExpr (biabl->VALABL);
	      FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (biabl->VALABL);
	      numberAtomExprBehIntVal = numberAtomExprBeh (beh, biabl->VALABL, numberTH);
	      profExprBehIntVal = profExprBeh (beh, biabl->VALABL, profTH);
	      FACTO_lit = FACTO_lit + numberAtomExprBehIntVal;

	      FACTO_litRed = FACTO_litRed + numberAtomExpr (biabl->CNDABL);
	      FACTO_numberOperBin = FACTO_numberOperBin + numberOperBinExpr (biabl->CNDABL);
	      numberAtomExprBehIntCnd = numberAtomExprBeh (beh, biabl->CNDABL, numberTH);
	      profExprBehIntCnd = profExprBeh (beh, biabl->CNDABL, profTH);
	      FACTO_lit = FACTO_lit + numberAtomExprBehIntCnd;

	      if (mode == 1)
		{
		  printf ("(%ld", numberAtomExprBehIntVal);
		  printf (", %d", profExprBehIntVal);
		  printf (", %d) ", searchTH (caracTH, (char *)biabl->VALABL));
		  printf ("(%ld", numberAtomExprBehIntCnd);
		  printf (", %d", profExprBehIntCnd);
		  printf (", %d) ", searchTH (caracTH, (char *)biabl->CNDABL));
		}

	      if (numberAtomExprBehIntVal > FACTO_maxLit)
		FACTO_maxLit = numberAtomExprBehIntVal;

	      if (profExprBehIntVal > FACTO_profMax)
		FACTO_profMax = profExprBehIntVal;

	      if (searchTH (caracTH, (char *)biabl->VALABL) > FACTO_delaisMax)
		FACTO_delaisMax = searchTH (caracTH, (char *)biabl->VALABL);

	      if (numberAtomExprBehIntCnd > FACTO_maxLit)
		FACTO_maxLit = numberAtomExprBehIntCnd;

	      if (profExprBehIntCnd > FACTO_profMax)
		FACTO_profMax = profExprBehIntCnd;

	      if (searchTH (caracTH, (char *)biabl->CNDABL) > FACTO_delaisMax)
		FACTO_delaisMax = searchTH (caracTH, (char *)biabl->CNDABL);
	    }
	  biabl = biabl->NEXT;
	}

      if (mode)
	printf ("\n");
      bux = bux->NEXT;
    }

  if (mode == 1)
    {
      printf ("--------------> Total number of literals = %ld\n", FACTO_lit);
      printf ("--------------> Max = %ld\n", FACTO_maxLit);
      printf ("--------------> Number of reduced literals = %d\n", FACTO_litRed);
      printf ("--------------> Number of binary operator = %d\n", FACTO_numberOperBin);
      printf ("--------------> Depth Max = %d\n", FACTO_profMax);
      printf ("--------------> Delay Max = %.3f\n", ((float) FACTO_delaisMax) / 1000);
    }

  /* mise a jour */

  *lit = FACTO_lit;
  *maxLit = FACTO_maxLit;
  *litRed = FACTO_litRed;
  *litReg = FACTO_litReg;
  *profMax = FACTO_profMax;
  *delaisMax = FACTO_delaisMax;
  *operBin = FACTO_numberOperBin;
  destroyTH (caracTH);
  destroyTH (profTH);
  destroyTH (numberTH);
}

/*-------------------------------------------------------------------------
markAblWBddBeh 	: marque le champs user des expressions avec le pNode
                  correspondant dans une befig. 
---------------------------------------------------------------------------
retour		: un pNode.
---------------------------------------------------------------------------*/
pNode
markAblWBddBeh (beh, expr)
     befig_list *beh;
     chain_list *expr;
{
  pCircuit pC = beh->CIRCUI;

  if (ATOM (expr))

    /* c'est forcement une entree de pC */
    {
      int val;

      if (!strcmp ((char *) VALUE_ATOM (expr), "'0'"))
	return (zero);

      if (!strcmp ((char *) VALUE_ATOM (expr), "'1'"))
	return (one);

      val = searchInputCct (pC, VALUE_ATOM (expr));
      if (val != EMPTYTH)
	return (createNodeTermBdd (val));
      else
	{
	  /* recherche d'un auxiliaire */


	  beaux_list *aux = beh->BEAUX;
	  pNode node = searchOutputCct (pC, VALUE_ATOM (expr));

	  /* deja calcule */

	  if (node != NULL)
	    return (node);

	  while (aux && strcmp (VALUE_ATOM (expr), aux->NAME))
	    {
	      aux = aux->NEXT;
	    }
	  if (!aux)
	    {
	      printf ("markAblWBeh : error auxiliary signal %s does'nt exist\n",
		      VALUE_ATOM (expr));
	      exit (-1);
	    }

	  node = markAblWBddBeh (beh, aux->ABL);
	  addOutputCct (beh->CIRCUI, aux->NAME, node);
	  return (node);
	}
    }
  else
    {
      short oper;
      pNode pt;
      chain_list *lstGdb;
      chain_list *tete;

      tete = expr;
      oper = OPER (expr);
      lstGdb = NULL;
      while ((expr = CDR (expr)))
	lstGdb = addListBdd (lstGdb, markAblWBddBeh (beh, CAR (expr)));
      pt = applyBdd (oper, lstGdb);
      freechain (lstGdb);

      /* le marquage */

      CDR (CAR (tete)) = (chain_list *) pt;
      return (pt);
    }
}

/*-------------------------------------------------------------------------
occTHBddBeh 	: calcule le nombre d'occurences de chaque graphe. 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
occTHBddBeh (expr, pTabNode)
     chain_list *expr;
     pTH pTabNode;
{
  pNode nodeP, nodeN;
  int valueN, valueP;

  if (!ATOM (expr))
    {
      nodeP = (pNode) CDR (CAR (expr));
      nodeN = notBdd (nodeP);

      valueP = searchTH (pTabNode, (char *) nodeP);
      if (valueP == EMPTYTH)
	valueP = 0;
      valueN = searchTH (pTabNode, (char *) nodeN);
      if (valueN == EMPTYTH)
	valueN = 0;


      if (valueN + valueP == 0)
	{
	  if (OPER (expr) != NOT)
	    addTH (pTabNode, (char *)nodeP, 1);

	  /* appel recursif sur les arguments */

	  while ((expr = CDR (expr)))
	    {
	      occTHBddBeh (CAR (expr), pTabNode);
	    }
	}
      else
	addTH (pTabNode, (char *)nodeP, 1 + valueP);
    }
}

/*-------------------------------------------------------------------------
createVIBeh 	: creation des variables intermediaires. 
---------------------------------------------------------------------------
retour		: une liste des var. aux.
---------------------------------------------------------------------------*/
chain_list *
createVIBeh (beh, expr, pTabNode, numberSon)
     befig_list *beh;
     chain_list *expr;
     pTH pTabNode;
     int numberSon;
{
  static int co = 0;
  int isNot = 0;

  if (expr == NULL)
    {
      printf ("createVIBeh : error expr = NULL\n");
      exit (-1);
    }

  if (ATOM (expr))
    return (copyExpr (expr));
  else
    {
      chain_list *resul;
      pNode nodeP, nodeN;
      int valueP, valueN;	/* le nombre d'occurence des graphes */

      nodeP = (pNode) CDR (CAR (expr));

      /* 1 cas particuliers */

      /* 1. Graphe non calcule            */

      if (nodeP == 0)
	return (copyExpr (expr));


      /* cas normal */

      valueP = abs (searchTH (pTabNode, (char *)nodeP));

      /* est ce un noeud atomique ? */

      if (((nodeP->low)->index < 2) && ((nodeP->high)->index < 2))
	{
	  if (nodeP->low == one)	/* not de la variable */
	    resul = notExpr (createAtom (*(beh->CIRCUI->pNameI + nodeP->index - 2)));
	  else			/* variable directe */
	    resul = createAtom (*(beh->CIRCUI->pNameI + nodeP->index - 2));
	  return (resul);
	}


      if (valueP > MAX_FATHER)
	{
	  /* c'est deja une variable auxiliaire... */

	  return (copyExpr ((chain_list *) valueP));

	  /* carrement ...!!! */
	}

      resul = createExpr (OPER (expr));

      /* on evite de creer trop de variables intermediaires... */


      /* traitement du not */


      if (OPER (expr) == NOT)
	{
	  if (!(ATOM (CADR (expr))))
	    {
	      isNot = 1;
	      expr = CADR (expr);
	      resul = createExpr (OPER (expr));
	    }
	  else
	    return (copyExpr (expr));
	}

      while ((expr = CDR (expr)))
	{
	  addQExpr (resul, createVIBeh (beh, CAR (expr),
					pTabNode, numberSon));
	}


      nodeN = notBdd (nodeP);
      valueN = searchTH (pTabNode, (char *)nodeN);
      if (valueN == EMPTYTH)
	valueN = 0;
      if (valueP == EMPTYTH)
	{
	  printf ("error : createVIBeh - expr with empty user NULL\n");
	  exit (-1);
	}

      /* la fonction possedent suffisament de peres ? */

      if (valueP + valueN >= 2)
	{
	  char *name;
	  beaux_list *aux;

	  name = gensym ("aux", co++);

	  if (SL_TRACE == 2)
	    {
	      printf ("Generating %s  P = %d N = %d\n Expr : ", name, valueP, valueN);
	      displayExpr (resul);
	    }

	  if (valueP >= valueN)
	    {
	      if (isNot)
		beh->BEAUX = beh_addbeaux (beh->BEAUX, name, notExpr (resul), NULL);
	      else
		beh->BEAUX = beh_addbeaux (beh->BEAUX, name, resul, NULL);
	      aux = beh->BEAUX;

	      while (aux && strcmp (aux->NAME, name))
		aux = aux->NEXT;

	      if (aux)
		aux->NODE = nodeP;
	      else
		{
		  printf ("createVIBeh error - %s doesn't exist\n", name);
		  exit (-1);
		}

	      addTH (pTabNode, (char *)nodeP, (int) createAtom (name));
	      addTH (pTabNode, (char *)nodeN, (int) notExpr (createAtom (name)));
	      return (createAtom (name));
	    }
	  else
	    {
	      if (isNot)
		beh->BEAUX = beh_addbeaux (beh->BEAUX, name, resul, NULL);
	      else
		beh->BEAUX = beh_addbeaux (beh->BEAUX, name, notExpr (resul), NULL);
	      aux = beh->BEAUX;
	      while (aux)
		{
		  if (aux->NAME == name)
		    {
		      aux->NODE = nodeN;
		      aux = NULL;
		    }
		  else
		    aux = aux->NEXT;
		}
	      addTH (pTabNode, (char *)nodeP, (int) notExpr (createAtom (name)));
	      addTH (pTabNode, (char *)nodeN, (int) createAtom (name));
	      return (notExpr (createAtom (name)));
	    }
	}
      /* pas de variable auxilaire */

      if (isNot)
	resul = notExpr (resul);

      return (resul);
    }
}

/*-------------------------------------------------------------------------
resetDustExpr 	: remet les champs poubelles d'une expression a 0. 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
resetDustExpr (expr)
     chain_list *expr;
{
  if (!ATOM (expr))
    {
      CDR (CAR (expr)) = 0;
      while ((expr = CDR (expr)))
	{
	  resetDustExpr (CAR (expr));
	}
    }
}
/*-------------------------------------------------------------------------
reverseExpr 	: inverse les arguments des expression 
---------------------------------------------------------------------------
retour		: un chain_list *.
---------------------------------------------------------------------------*/
chain_list *
reverseExpr (expr)
     chain_list *expr;
{
  if (ATOM (expr))
    return (copyExpr (expr));
  else
    {
      chain_list *resul;

      resul = createExpr (OPER (expr));
      if (lengthExpr (expr) <= 2)
	{
	  while ((expr = CDR (expr)))
	    addQExpr (resul, reverseExpr (CAR (expr)));
	  return (resul);
	}
      else
	{
	  resul = createExpr (OPER (expr));
	  while ((expr = CDR (expr)))
	    {
	      addHExpr (resul, reverseExpr (CAR (expr)));
	    }
	  return (resul);
	}
    }
}

/*-------------------------------------------------------------------------
arity2Expr 	: transforme une expression avec operateur d'arite 2.   
---------------------------------------------------------------------------
retour		: un chain_list *.
---------------------------------------------------------------------------*/
chain_list *
arity2Expr (expr)
     chain_list *expr;
{
  if (ATOM (expr))
    return (copyExpr (expr));
  else
    {
      chain_list *resul;

      resul = createExpr (OPER (expr));
      if (lengthExpr (expr) <= 2)
	{
	  while ((expr = CDR (expr)))
	    addQExpr (resul, arity2Expr (CAR (expr)));
	  return (resul);
	}
      else
	{
	  int i;
	  chain_list *inter = NULL;

	  for (i = lengthExpr (expr) - 1; i > 0; i--)
	    {
	      resul = createExpr (OPER (expr));
	      if (inter == NULL)
		{
		  addHExpr (resul, arity2Expr (searchNumExpr (expr, i)));
		  addHExpr (resul, arity2Expr (searchNumExpr (expr, i - 1)));
		}
	      else
		{
		  addHExpr (resul, inter);
		  addHExpr (resul, arity2Expr (searchNumExpr (expr, i - 1)));
		}
	      inter = resul;
	    }
	  return (resul);
	}
    }
}
/*-------------------------------------------------------------------------
gcBddToAblBeh	: garbage sur la fonction BddToAbl 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
gcBddToAblBeh (beh, trace)
     befig_list *beh;
     int trace;
{
  if (numberNodeAllBdd () > 100000)
    {
      if (trace)
	printf ("gc %d ...\n", numberNodeAllBdd ());

      gcNodeBeh (beh, 0);

      if (trace)
	printf ("End of gc pTabExpr %d\n", numberNodeAllBdd ());
    }
}

/*-------------------------------------------------------------------------
rempTabExpr 	: remplissage de pTabExpr 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
pNode
rempTabExpr (beh, expr)
     befig_list *beh;
     chain_list *expr;
{
  pCircuit pC = beh->CIRCUI;

  if (ATOM (expr))

    /* c'est forcement une entree de pC */
    {
      int val;

      if (!strcmp ((char *) VALUE_ATOM (expr), "'0'"))
	return (zero);

      if (!strcmp ((char *) VALUE_ATOM (expr), "'1'"))
	return (one);

      val = searchInputCct (pC, VALUE_ATOM (expr));
      if (val != EMPTYTH)
	return (createNodeTermBdd (val));
      else
	{
	  printf ("rempTabExpr : error signal %s does'nt exist as input\n",
		  VALUE_ATOM (expr));
	  exit (-1);
	}
    }
  else
    {
      short oper;
      pNode pt;
      chain_list *lstGdb;
      chain_list *tete;

      tete = expr;
      oper = OPER (expr);
      lstGdb = NULL;
      while ((expr = CDR (expr)))
	lstGdb = addListBdd (lstGdb, rempTabExpr (beh, CAR (expr)));
      pt = applyBdd (oper, lstGdb);
      freechain (lstGdb);

      /* la mise a jour de pTabExpr */

      addTH (pTabExpr, (char *)pt, (int)tete);
      return (pt);
    }
}

/*-------------------------------------------------------------------------
bddToAblBeh 	: regeneration des abl(s) du circuit. 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
bddToAblBeh (beh, optim, trace)
     befig_list *beh;
     int optim;
     int trace;
{
  beout_list *out;
  beaux_list *aux;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  binode_list *binode;
  biabl_list *biabl;
  chain_list *expr;
  chain_list *tete;
  chain_list *lstNotExpr;

  if (trace)
    printf ("bddToAblBeh Running...\n");


  pTabExpr = createTH (4999);

  /* la desallocation des expressions de la befig */

  out = beh->BEOUT;
  while (out)
    {
      if (out->ABL)
	freeExpr (out->ABL);
      out = out->NEXT;
    }

  aux = beh->BEAUX;
  while (aux)
    {
      if (aux->ABL)
	freeExpr (aux->ABL);
      aux = aux->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      while (biabl)
	{
	  if (biabl->CNDABL && biabl->VALABL)
	    {
	      freeExpr (biabl->CNDABL);
	      freeExpr (biabl->VALABL);
	    }
	  biabl = biabl->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      while (biabl)
	{
	  if (biabl->CNDABL && biabl->VALABL)
	    {
	      freeExpr (biabl->CNDABL);
	      freeExpr (biabl->VALABL);
	    }
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;
      while (biabl)
	{
	  if (biabl->CNDABL && biabl->VALABL)
	    {
	      freeExpr (biabl->CNDABL);
	      freeExpr (biabl->VALABL);
	    }
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }

  /* la regeneration des expressions a partir des BDD */

  lstNotExpr = NULL;

  out = beh->BEOUT;
  while (out)
    {
      if (out->NODE)
	{


	  out->ABL = bdd2Abl (out->NODE, beh->CIRCUI->pNameI, optim);

	  expr = out->ABL;

	  out->ABL = arity2Expr (expr);
	  freeExpr (expr);

	  rempTabExpr (beh, out->ABL);

	  expr = createExpr (NOT);
	  addQExpr (expr, copyExpr (out->ABL));

	  rempTabExpr (beh, expr);
	  lstNotExpr = addchain (lstNotExpr, expr);

	  /* la reallocation si necessaire */

	  if (trace == 2)
	    {
	      printf ("+++ out +++ (%s) count %d", out->NAME, numberNodeAllBdd ());
	      printf ("numberAtom = %d\n", numberAtomExpr (out->ABL));
	    }

	  gcBddToAblBeh (beh, trace);
	}
      else
	{
	  printf ("bddToAblBeh : error there's no graph on %s !\n", out->NAME);
	  exit (-1);
	}

      out = out->NEXT;
    }


  aux = beh->BEAUX;
  while (aux)
    {
      if (aux->NODE)
	{
	  aux->ABL = bdd2Abl (aux->NODE, beh->CIRCUI->pNameI, optim);

	  expr = aux->ABL;

	  aux->ABL = arity2Expr (expr);
	  freeExpr (expr);

	  rempTabExpr (beh, aux->ABL);

	  expr = createExpr (NOT);
	  addQExpr (expr, copyExpr (aux->ABL));

	  rempTabExpr (beh, expr);
	  lstNotExpr = addchain (lstNotExpr, expr);

	  if (trace == 2)
	    {
	      printf ("+++ aux +++ (%s) count %d", aux->NAME, numberNodeAllBdd ());
	      printf ("numberAtom = %d\n", numberAtomExpr (aux->ABL));
	    }

	  gcBddToAblBeh (beh, trace);
	}
      else
	{
	  printf ("bddToAblBeh : error there's no graph on %s !\n", aux->NAME);
	  exit (-1);
	}
      aux = aux->NEXT;
    }


  reg = beh->BEREG;
  while (reg)
    {
      binode = reg->BINODE;
      biabl = reg->BIABL;
      while (binode)
	{
	  if (binode->VALNODE && binode->CNDNODE)
	    {
	      biabl->VALABL = bdd2Abl (binode->VALNODE, beh->CIRCUI->pNameI, optim);

	      expr = biabl->VALABL;

	      biabl->VALABL = arity2Expr (expr);
	      freeExpr (expr);

	      rempTabExpr (beh, biabl->VALABL);

	      expr = createExpr (NOT);
	      addQExpr (expr, copyExpr (biabl->VALABL));

	      rempTabExpr (beh, expr);
	      lstNotExpr = addchain (lstNotExpr, expr);


	      biabl->CNDABL = bdd2Abl (binode->CNDNODE, beh->CIRCUI->pNameI, optim);

	      expr = biabl->CNDABL;

	      biabl->CNDABL = arity2Expr (expr);
	      freeExpr (expr);

	      rempTabExpr (beh, biabl->CNDABL);

	      expr = createExpr (NOT);
	      addQExpr (expr, copyExpr (biabl->CNDABL));

	      rempTabExpr (beh, expr);
	      lstNotExpr = addchain (lstNotExpr, expr);

	      /* la reallocation si necessaire */

	      if (trace == 2)
		{
		  printf ("+++ reg +++ (%s) count %d", reg->NAME, numberNodeAllBdd ());
		  printf ("numberAtom = %d\n", numberAtomExpr (biabl->VALABL));
		  printf ("numberAtom = %d\n", numberAtomExpr (biabl->CNDABL));
		}

	      gcBddToAblBeh (beh, trace);

	    }
	  else
	    {
	      printf ("bddToAblBeh : error there's no graph on %s !\n", reg->NAME);
	      exit (-1);
	    }
	  binode = binode->NEXT;
	  biabl = biabl->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      binode = bus->BINODE;
      biabl = bus->BIABL;
      while (binode)
	{
	  if (binode->VALNODE && binode->CNDNODE)
	    {
	      biabl->VALABL = bdd2Abl (binode->VALNODE, beh->CIRCUI->pNameI, optim);

	      expr = biabl->VALABL;

	      biabl->VALABL = arity2Expr (expr);
	      freeExpr (expr);

	      rempTabExpr (beh, biabl->VALABL);


	      biabl->CNDABL = bdd2Abl (binode->CNDNODE, beh->CIRCUI->pNameI, optim);

	      expr = biabl->CNDABL;

	      biabl->CNDABL = arity2Expr (expr);
	      freeExpr (expr);

	      rempTabExpr (beh, biabl->CNDABL);

	      if (trace)
		{
		  printf ("+++ bus +++ (%s) count %d", bus->NAME, numberNodeAllBdd ());
		  printf ("numberAtom = %d\n", numberAtomExpr (biabl->VALABL));
		  printf ("numberAtom = %d\n", numberAtomExpr (biabl->CNDABL));
		}

	      gcBddToAblBeh (beh, trace);

	    }
	  else
	    {
	      printf ("bddToAblBeh : error there's no graph on %s !\n", bus->NAME);
	      exit (-1);
	    }
	  binode = binode->NEXT;
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      binode = bux->BINODE;
      biabl = bux->BIABL;
      while (binode)
	{
	  if (binode->VALNODE && binode->CNDNODE)
	    {
	      biabl->VALABL = bdd2Abl (binode->VALNODE, beh->CIRCUI->pNameI, optim);

	      expr = biabl->VALABL;

	      biabl->VALABL = arity2Expr (expr);
	      freeExpr (expr);

	      rempTabExpr (beh, biabl->VALABL);


	      biabl->CNDABL = bdd2Abl (binode->CNDNODE, beh->CIRCUI->pNameI, optim);

	      expr = biabl->CNDABL;

	      biabl->CNDABL = arity2Expr (expr);
	      freeExpr (expr);

	      rempTabExpr (beh, biabl->CNDABL);

	      if (trace == 2)
		{
		  printf ("+++ bux +++ (%s) count %d", bux->NAME, numberNodeAllBdd ());
		  printf ("numberAtom = %d\n", numberAtomExpr (biabl->VALABL));
		  printf ("numberAtom = %d\n", numberAtomExpr (biabl->CNDABL));
		}

	      gcBddToAblBeh (beh, trace);

	    }
	  else
	    {
	      printf ("bddToAblBeh : error there's no graph on %s !\n", bux->NAME);
	      exit (-1);
	    }
	  binode = binode->NEXT;
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }

  tete = lstNotExpr;
  while (lstNotExpr)
    {
      freeExpr (lstNotExpr->DATA);
      lstNotExpr = lstNotExpr->NEXT;
    }
  freechain (tete);

  destroyTH (pTabExpr);

  if (numberNodeAllBdd () > 20000)
    gcNodeBeh (beh, 0);

  if (trace)
    printf ("End of BddToAbl\n");
}

/*-------------------------------------------------------------------------
createAuxBeh 	: creation de VI pour une befig 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
createAuxBeh (beh)
     befig_list *beh;
{
  beout_list *out;
  beaux_list *aux;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  biabl_list *biabl;
  chain_list *expr;
  int numberSon;
  pTH pTabNode;

  pTabNode = createTH (1999);

  if (SL_TRACE == 2)
    displayBeh (beh, 1);

  mapExprBeh (beh, resetDustExpr);

  /* transformation en arbres binaires */

  mapCarExprBeh (beh, arity2Expr);

  /* le marquage des ABL */

  aux = beh->BEAUX;
  while (aux)
    {
      addOutputCct (beh->CIRCUI, aux->NAME, markAblWBddBeh (beh, aux->ABL));
      aux = aux->NEXT;
    }

  out = beh->BEOUT;
  while (out)
    {
      markAblWBddBeh (beh, out->ABL);
      out = out->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      while (biabl)
	{
	  markAblWBddBeh (beh, biabl->CNDABL);
	  markAblWBddBeh (beh, biabl->VALABL);
	  biabl = biabl->NEXT;
	}
      reg = reg->NEXT;
    }


  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      while (biabl)
	{
	  markAblWBddBeh (beh, biabl->CNDABL);
	  markAblWBddBeh (beh, biabl->VALABL);
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;
      while (biabl)
	{
	  markAblWBddBeh (beh, biabl->CNDABL);
	  markAblWBddBeh (beh, biabl->VALABL);
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }

  /* le calcul du nombre d'occurence */

  out = beh->BEOUT;
  while (out)
    {
      occTHBddBeh (out->ABL, pTabNode);
      out = out->NEXT;
    }

  aux = beh->BEAUX;
  while (aux)
    {
      occTHBddBeh (aux->ABL, pTabNode);
      aux = aux->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      while (biabl)
	{
	  occTHBddBeh (biabl->VALABL, pTabNode);
	  occTHBddBeh (biabl->CNDABL, pTabNode);
	  biabl = biabl->NEXT;
	}
      reg = reg->NEXT;
    }


  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      while (biabl)
	{
	  occTHBddBeh (biabl->VALABL, pTabNode);
	  occTHBddBeh (biabl->CNDABL, pTabNode);
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;
      while (biabl)
	{
	  occTHBddBeh (biabl->VALABL, pTabNode);
	  occTHBddBeh (biabl->CNDABL, pTabNode);
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }
  /* fin du calcul des occurences de pNode
     Ce calcul etant partiel , il faut necessairement
     parcourir la befig de la meme maniere pour la
     la creation des VI            
   */


  numberSon = 2;

  /* on evite de creer des VI deja existante.... */

  aux = beh->BEAUX;
  while (aux)
    {
      pNode tete = aux->NODE;


      /* aux->NAME est deja une variable aux. */

      addTH (pTabNode, (char *)tete, (int)createAtom (aux->NAME));
      addTH (pTabNode, (char *)notBdd (tete), (int)notExpr (createAtom (aux->NAME)));

      aux = aux->NEXT;
    }

  out = beh->BEOUT;
  while (out)
    {
      expr = out->ABL;
      out->ABL = createVIBeh (beh, out->ABL, pTabNode, numberSon);
      freeExpr (expr);
      out = out->NEXT;
    }

  aux = beh->BEAUX;
  while (aux)
    {
      chain_list *resExpr;
      int teteNot = 0;

      if (!ATOM (aux->ABL))	/* traitement special pour eviter de boucler */
	{
	  expr = aux->ABL;
	  if (OPER (expr) == NOT)
	    {
	      expr = CADR (expr);
	      teteNot = 1;
	    }
	  if (!ATOM (expr))	/* on evite de traiter x <= not a; */
	    {
	      resExpr = createExpr (OPER (expr));
	      while ((expr = CDR (expr)))
		{
		  addQExpr (resExpr, createVIBeh (beh, CAR (expr), pTabNode, numberSon));
		}
	      expr = aux->ABL;
	      if (teteNot)
		aux->ABL = notExpr (resExpr);
	      else
		aux->ABL = resExpr;
	      freeExpr (expr);
	    }
	}
      aux = aux->NEXT;
    }

  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      while (biabl)
	{
	  expr = biabl->CNDABL;
	  biabl->CNDABL = createVIBeh (beh, biabl->CNDABL, pTabNode, numberSon);
	  freeExpr (expr);
	  expr = biabl->VALABL;
	  biabl->VALABL = createVIBeh (beh, biabl->VALABL, pTabNode, numberSon);
	  freeExpr (expr);
	  biabl = biabl->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      biabl = bus->BIABL;
      while (biabl)
	{
	  expr = biabl->CNDABL;
	  biabl->CNDABL = createVIBeh (beh, biabl->CNDABL, pTabNode, numberSon);
	  freeExpr (expr);
	  expr = biabl->VALABL;
	  biabl->VALABL = createVIBeh (beh, biabl->VALABL, pTabNode, numberSon);
	  freeExpr (expr);
	  biabl = biabl->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      biabl = bux->BIABL;
      while (biabl)
	{
	  expr = biabl->CNDABL;
	  biabl->CNDABL = createVIBeh (beh, biabl->CNDABL, pTabNode, numberSon);
	  freeExpr (expr);
	  expr = biabl->VALABL;
	  biabl->VALABL = createVIBeh (beh, biabl->VALABL, pTabNode, numberSon);
	  freeExpr (expr);
	  biabl = biabl->NEXT;
	}
      bux = bux->NEXT;
    }

  flatArityExprBeh (beh);

  destroyTH (pTabNode);


  if (SL_TRACE == 2)
    {
      printf ("End of the generating VI\n");
      displayBeh (beh, 1);
    }
}

/*-------------------------------------------------------------------------
coutSol 	: calcule le cout d'une solution dans le recuit simule. 
---------------------------------------------------------------------------
retour		: un int.
---------------------------------------------------------------------------*/
int
coutSol (lit, delaisMax, litRed, profMax, operBin, optim)
     int lit, delaisMax, litRed, profMax, operBin, optim;
{
  return (optim * (2 * (litRed + operBin + 5 * profMax * profMax + delaisMax / 10) + lit) +
	  (4 - optim) * (litRed));
}
/*-------------------------------------------------------------------------
regenereBddBeh 	: regenere les Bdd d'une befig avec le nouvel ordonnancement 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
regenereBddBeh (beh, order)
     befig_list *beh;
     chain_list *order;
{
  chain_list *ptChain;
  if (order == NULL)
    {
      printf ("regenereBddBeh : error order = NULL\n");
      exit (-1);
    }

  videBddBefig (beh);
  destroyCct (beh->CIRCUI);
  beh->CIRCUI = initializeCct (beh->NAME, 2 * lengthExpr (order) + 1, 100);
  ptChain = order;
  while (ptChain)
    {
      addInputCct (beh->CIRCUI, ptChain->DATA);
      ptChain = ptChain->NEXT;
    }
}
/*-------------------------------------------------------------------------
createAuxInitBeh : creation de VI pour eviter de traiter x fois
                   le meme cone. Les nouvelles VI sont ajoutes
                   dans la couche circuit de beh.
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
chain_list *
createAuxInitBeh (beh)
     befig_list *beh;
{
  pTH occTH;
  static int co = 0;
  int val;
  chain_list *saveAux;
  beout_list *out;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  biabl_list *biabl;
  binode_list *binode;
  char *name;

  /* initialisation */

  saveAux = NULL;
  occTH = createTH (200);

  /* calcul du nombre d'occurences des tetes d'expression */

  out = beh->BEOUT;
  while (out)
    {
      if (out->NODE && numberNodeBdd (out->NODE) > 1)
	{
	  val = searchTH (occTH, (char *)out->NODE);
	  if (val != EMPTYTH)
	    addTH (occTH, (char *)out->NODE, 2);
	  else
	    addTH (occTH, (char *)out->NODE, 1);
	}
      out = out->NEXT;
    }
  reg = beh->BEREG;
  while (reg)
    {
      binode = reg->BINODE;
      while (binode)
	{
	  if (binode->VALNODE && numberNodeBdd (binode->VALNODE) > 1)
	    {
	      val = searchTH (occTH, (char *)binode->VALNODE);
	      if (val != EMPTYTH)
		addTH (occTH, (char *)binode->VALNODE, 2);
	      else
		addTH (occTH, (char *)binode->VALNODE, 1);
	    }
	  if (binode->CNDNODE && numberNodeBdd (binode->CNDNODE) > 1)
	    {
	      val = searchTH (occTH, (char *)binode->CNDNODE);
	      if (val != EMPTYTH)
		addTH (occTH, (char *)binode->CNDNODE, 2);
	      else
		addTH (occTH, (char *)binode->CNDNODE, 1);
	    }
	  binode = binode->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      binode = bus->BINODE;
      while (binode)
	{
	  if (binode->VALNODE && numberNodeBdd (binode->VALNODE) > 1)
	    {
	      val = searchTH (occTH, (char *)binode->VALNODE);
	      if (val != EMPTYTH)
		addTH (occTH, (char *)binode->VALNODE, 2);
	      else
		addTH (occTH, (char *)binode->VALNODE, 1);
	    }
	  if (binode->CNDNODE && numberNodeBdd (binode->CNDNODE) > 1)
	    {
	      val = searchTH (occTH, (char *)binode->CNDNODE);
	      if (val != EMPTYTH)
		addTH (occTH, (char *)binode->CNDNODE, 2);
	      else
		addTH (occTH, (char *)binode->CNDNODE, 1);
	    }
	  binode = binode->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      binode = bux->BINODE;
      while (binode)
	{
	  if (binode->VALNODE && numberNodeBdd (binode->VALNODE) > 1)
	    {
	      val = searchTH (occTH, (char *)binode->VALNODE);
	      if (val != EMPTYTH)
		addTH (occTH, (char *)binode->VALNODE, 2);
	      else
		addTH (occTH, (char *)binode->VALNODE, 1);
	    }
	  if (binode->CNDNODE && numberNodeBdd (binode->CNDNODE) > 1)
	    {
	      val = searchTH (occTH, (char *)binode->CNDNODE);
	      if (val != EMPTYTH)
		addTH (occTH, (char *)binode->CNDNODE, 2);
	      else
		addTH (occTH, (char *)binode->CNDNODE, 1);
	    }
	  binode = binode->NEXT;
	}
      bux = bux->NEXT;
    }
  /* parcours avec creation de VI */

  out = beh->BEOUT;
  while (out)
    {
      if (out->NODE)
	{
	  val = searchTH (occTH, (char *)out->NODE);
	  if (val == 2)
	    {
	      name = gensym ("auxinit", ++co);
	      beh->BEAUX = beh_addbeaux (beh->BEAUX, name, out->ABL, NULL);
	      (beh->BEAUX)->NODE = out->NODE;
	      out->ABL = createAtom (name);
	      addTH (occTH, (char *)out->NODE, (int)name);
	      saveAux = addchain (saveAux, name);
	      addInputCct (beh->CIRCUI, name);
	      out->NODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
	    }
	  else if (val != EMPTYTH && val != 1)
	    {
	      name = (char *) val;
	      freeExpr (out->ABL);
	      out->ABL = createAtom (name);
	      out->NODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
	    }
	}
      out = out->NEXT;
    }
  reg = beh->BEREG;
  while (reg)
    {
      biabl = reg->BIABL;
      binode = reg->BINODE;
      while (biabl)
	{
	  if (binode->VALNODE && biabl->VALABL)
	    {
	      val = searchTH (occTH, (char *)binode->VALNODE);
	      if (val == 2)
		{
		  name = gensym ("auxinit", ++co);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->VALABL, NULL);
		  (beh->BEAUX)->NODE = binode->VALNODE;
		  biabl->VALABL = createAtom (name);
		  addTH (occTH, (char *)binode->VALNODE, (int)name);
		  saveAux = addchain (saveAux, name);
		  addInputCct (beh->CIRCUI, name);
		  binode->VALNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	      else if (val != EMPTYTH && val != 1)
		{
		  name = (char *) val;
		  freeExpr (biabl->VALABL);
		  biabl->VALABL = createAtom (name);
		  binode->VALNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	    }
	  if (binode->CNDNODE && biabl->CNDABL)
	    {
	      val = searchTH (occTH, (char *)binode->CNDNODE);
	      if (val == 2)
		{
		  name = gensym ("auxinit", ++co);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->CNDABL, NULL);
		  (beh->BEAUX)->NODE = binode->CNDNODE;
		  biabl->CNDABL = createAtom (name);
		  addTH (occTH, (char *)binode->CNDNODE, (int)name);
		  saveAux = addchain (saveAux, name);
		  addInputCct (beh->CIRCUI, name);
		  binode->CNDNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	      else if (val != EMPTYTH && val != 1)
		{
		  name = (char *) val;
		  freeExpr (biabl->CNDABL);
		  biabl->CNDABL = createAtom (name);
		  binode->CNDNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	    }
	  binode = binode->NEXT;
	  biabl = biabl->NEXT;
	}
      reg = reg->NEXT;
    }

  bus = beh->BEBUS;
  while (bus)
    {
      binode = bus->BINODE;
      biabl = bus->BIABL;
      while (biabl)
	{
	  if (binode->VALNODE && biabl->VALABL)
	    {
	      val = searchTH (occTH, (char *)binode->VALNODE);
	      if (val == 2)
		{
		  name = gensym ("auxinit", ++co);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->VALABL, NULL);
		  (beh->BEAUX)->NODE = binode->VALNODE;
		  biabl->VALABL = createAtom (name);
		  addTH (occTH, (char *)binode->VALNODE, (int)name);
		  saveAux = addchain (saveAux, name);
		  addInputCct (beh->CIRCUI, name);
		  binode->VALNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	      else if (val != EMPTYTH && val != 1)
		{
		  name = (char *) val;
		  freeExpr (biabl->VALABL);
		  biabl->VALABL = createAtom (name);
		  binode->VALNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	    }
	  if (binode->CNDNODE && biabl->CNDABL)
	    {
	      val = searchTH (occTH, (char *)binode->CNDNODE);
	      if (val == 2)
		{
		  name = gensym ("auxinit", ++co);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->CNDABL, NULL);
		  (beh->BEAUX)->NODE = binode->CNDNODE;
		  biabl->CNDABL = createAtom (name);
		  addTH (occTH, (char *)binode->CNDNODE, (int)name);
		  saveAux = addchain (saveAux, name);
		  addInputCct (beh->CIRCUI, name);
		  binode->CNDNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	      else if (val != EMPTYTH && val != 1)
		{
		  name = (char *) val;
		  freeExpr (biabl->CNDABL);
		  biabl->CNDABL = createAtom (name);
		  binode->CNDNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      bus = bus->NEXT;
    }

  bux = beh->BEBUX;
  while (bux)
    {
      binode = bux->BINODE;
      biabl = bux->BIABL;
      while (biabl)
	{
	  if (binode->VALNODE && biabl->VALABL)
	    {
	      val = searchTH (occTH, (char *)binode->VALNODE);
	      if (val == 2)
		{
		  name = gensym ("auxinit", ++co);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->VALABL, NULL);
		  (beh->BEAUX)->NODE = binode->VALNODE;
		  biabl->VALABL = createAtom (name);
		  addTH (occTH, (char *)binode->VALNODE, (int)name);
		  saveAux = addchain (saveAux, name);
		  addInputCct (beh->CIRCUI, name);
		  binode->VALNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	      else if (val != EMPTYTH && val != 1)
		{
		  name = (char *) val;
		  freeExpr (biabl->VALABL);
		  biabl->VALABL = createAtom (name);
		  binode->VALNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	    }
	  if (binode->CNDNODE && biabl->CNDABL)
	    {
	      val = searchTH (occTH, (char *)binode->CNDNODE);
	      if (val == 2)
		{
		  name = gensym ("auxinit", ++co);
		  beh->BEAUX = beh_addbeaux (beh->BEAUX, name, biabl->CNDABL, NULL);
		  (beh->BEAUX)->NODE = binode->CNDNODE;
		  biabl->CNDABL = createAtom (name);
		  addTH (occTH, (char *)binode->CNDNODE, (int)name);
		  saveAux = addchain (saveAux, name);
		  addInputCct (beh->CIRCUI, name);
		  binode->CNDNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	      else if (val != EMPTYTH && val != 1)
		{
		  name = (char *) val;
		  freeExpr (biabl->CNDABL);
		  biabl->CNDABL = createAtom (name);
		  binode->CNDNODE = createNodeTermBdd (
					searchInputCct (beh->CIRCUI, name));
		}
	    }
	  biabl = biabl->NEXT;
	  binode = binode->NEXT;
	}
      bux = bux->NEXT;
    }


  destroyTH (occTH);
  return saveAux;
}
/*-------------------------------------------------------------------------
notSameOrder 	: verifie que l'ordonnancement Order n'a pas deja ete
                  memorise dans tabSol 
---------------------------------------------------------------------------
retour		: 1 si l'order est nouveau 0 sinon 
---------------------------------------------------------------------------*/
int
notSameOrder (tabSol, numberSol, order)
     tabSolution *tabSol;
     int numberSol;
     chain_list *order;
{
  int i;

  for (i = 0; i < numberSol; i++)
    {
      int egal;
      chain_list *order1, *order2;

      order1 = order;
      order2 = tabSol[i].order;
      egal = 1;
      while (order1 != NULL && order2 != NULL)
	{
	  if (strcmp ((char *) order1->DATA, (char *) order2->DATA))
	    egal = 0;
	  order1 = order1->NEXT;
	  order2 = order2->NEXT;
	}
      /* meme ordonnancement */

      if (egal == 1)
	{
	  return (0);
	}
    }
  return (1);
}

/*-------------------------------------------------------------------------
localOpt 	: optimisation locale d'une befig 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
localOpt (trace)
     int trace;
{
  beaux_list *aux;
  chain_list *saveAux = NULL;
  int numberNode;
  befig_list *beh;
  befig_list *CBeh;
  char *carint;
  chain_list *orderPI;
  int depthNode ();
  int lit, maxLit, litRed, litReg, profMax, delaisMax, operBin;
  int cout_b, cout_cb;


  beh = SL_BEH;
  /* recopie de SL_BEH dans CBeh */

  CBeh = (befig_list *) copyBeh (beh);
  CBeh->CIRCUI = initializeCct ("figure", (beh->CIRCUI)->countI, 50);
  cpOrderCct (beh->CIRCUI, CBeh->CIRCUI);

  /* on elimine les auxiliaires trop petits */

  compactBeh (CBeh, SL_GUARDEDAUX);

  /* sauvegarde du fichier */

  if (SL_TRACE)
    {
      carint = (char *) mbkalloc (strlen (CBeh->NAME) + 2);
      sprintf (carint, "%sd", CBeh->NAME);
      decompileVHDL (CBeh, carint, 1);
      mbkfree (carint);
    }

  /* reconstruction des graphes */

  orderPI = (chain_list *) makeOrderAbl (CBeh, SL_ORDER, 0);
  aux = CBeh->BEAUX;
  while (aux)
    {
      orderPI = addchain (orderPI, aux->NAME);
      aux = aux->NEXT;
    }

/*
   order = orderPI;
   printf("\n");
   while (order) 
   {
   printf("%s ",order->DATA);
   order = order->NEXT;
   }
   printf("\n");

 */

  regenereBddBeh (CBeh, orderPI);
  makeBddBeh (CBeh, CBeh->CIRCUI, 0, trace, 50, 100, NULL);


  /* on choisit entre beh et compactBeh */

/* 
   costFpgaBeh(SL_BEH,countNode,&numCLB,depthNode,&depthCLB,SL_AUX);
   costFpgaBeh(CBeh,countNode,&numCLBC,depthNode,&depthCLBC,SL_AUX);

 */

  coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
	      &litReg, trace, SL_DELAYPI);
  cout_b = coutSol (lit, delaisMax, litRed, profMax, operBin, SL_OPTIM);

  coutAblBeh (CBeh, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
	      &litReg, trace, SL_DELAYPI);
  cout_cb = coutSol (lit, delaisMax, litRed, profMax, operBin, SL_OPTIM);

/*  printf("cout beh=%d, cbeh=%d\n",cout_b,cout_cb); */
  if (cout_b > cout_cb)
    {
      SL_BEH = CBeh;
      beh = CBeh;
    }


  aux = beh->BEAUX;
  saveAux = NULL;		/* on vide les auxinit eventuels */
  while (aux)
    {
      saveAux = addchain (saveAux, aux->NAME);
      aux = aux->NEXT;
    }

  {
    chain_list *intermediate = SL_GUARDEDAUX;
    while (intermediate)
      {
	if (existAuxBeh (SL_BEH, intermediate->DATA))
	  {
	    saveAux = addchain (saveAux, intermediate->DATA);
	  }
	else
	  {
	    printf ("optimization : Warning \n---> bad parameters file\n");
	    printf ("---> signal %s unknown\n", (char *)intermediate->DATA);
	  }
	intermediate = intermediate->NEXT;
      }
  }
  /* re-ordonnancement */

  numberNode = 20;

  if (SL_MAKEORDER == 2)
    sl_sys->orderPI = makeOrderBddBeh (SL_BEH, SL_OPTIM, SL_OPTIMPO, SL_DELAYPI, 1,
			   numberNode, 1, trace, 100, 1, numberAtomRedFact);

  else
    /* on recupere l'ordre de la couche circuit */
    {
      char **ptChar;
      int c;

      ptChar = (SL_CIRCUIT)->pNameI;	/* les entrees */

      sl_sys->orderPI = NULL;
      for (c = ((SL_CIRCUIT)->countI) - 2; c != 0; c--)
	{
	  sl_sys->orderPI = addchain (sl_sys->orderPI, *ptChar);
	  ptChar++;
	}
      sl_sys->orderPI = reverse (sl_sys->orderPI);
    }

  makeBddBeh (SL_BEH, SL_CIRCUIT, 0, trace, 200, 1000000, NULL);
  removeAuxBeh (SL_BEH, saveAux);

  /* faut-il essayer les deux re-generations ? */

  if (SL_OPTIM > 2)
    bddToAblBeh (SL_BEH, 1, SL_TRACE);
  else
    bddToAblBeh (SL_BEH, 0, SL_TRACE);

  if (trace)
    printf ("Generating VI\n");

  if (SL_AUXOF)
    createAuxBeh (SL_BEH);

  /* l'optimisation locale peut quand meme eliminer
     des variables intermediaires */

  compactBeh (SL_BEH, saveAux);

  /* on inverse l'ordre des expressions et 
     on cree des variable Inter a nouveau */

  mapCarExprBeh (SL_BEH, reverseExpr);

  if (SL_AUXOF)
    createAuxBeh (SL_BEH);

  compactBeh (SL_BEH, saveAux);

  mapCarExprBeh (SL_BEH, reverseExpr);
}


/*-------------------------------------------------------------------------
globalOpt 	: optimisation globale d'une befig 
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/
void
globalOpt (trace)
     int trace;
{
  pTH caracBeh ();
  tabSolution *tabSol;
  int lit, maxLit, litRed, litReg, profMax, delaisMax, operBin;
  int amplitude;
  int numberSol, indiceMin, coutMin;
  int i, level;
  chain_list *saveAux;
  int numberInput = 0;
  int numberNode = 0;
  chain_list *lstExpr, *lst;
  int existDC;
  int complexite;

  /* calcul de la complexite du circuit (nombre de noeud initial) */


  if (numberNodeAllBdd () > 200)
    complexite = 1;
  else
    complexite = 0;

  /* sauvegarde dans saveAux des auxiliaires a garder */

  saveAux = createAuxInitBeh (SL_BEH);

  /* existe-t-il des constantes 'd' dans les expressions ? */

  lstExpr = exprToCLBeh (SL_BEH);
  lst = lstExpr;
  existDC = 0;
  while (lst)
    {
      if (searchExpr ((chain_list *) lst->DATA, "'d'"))
	existDC = 1;
      lst = lst->NEXT;
    }
  freechain (lstExpr);

  if (existDC)			/* appel au traitement des don't care */
    {
      for (i = 1; i < 3; i++)
	{
	  /* premiere phase d'ordonnancement pour la prise en compte des DC */

	  if (SL_MAKEORDER == 2)
	    {
	      if (complexite)
		makeOrderBddBeh (SL_BEH, SL_OPTIM, SL_OPTIMPO,
			       SL_DELAYPI, 1, 8, 1, trace, 5, 1, countNode);
	      else
		makeOrderBddBeh (SL_BEH, SL_OPTIM, SL_OPTIMPO,
			       SL_DELAYPI, 1, 2, 1, trace, 5, 1, countNode);
	    }
	  makeBddBeh (SL_BEH, SL_CIRCUIT, 0, trace, 200, 1000000, NULL);
	}
/*
   traitementDCBeh(SL_BEH,trace,countNode,1);
 */

      traitementDCBeh (SL_BEH, trace, numberAtomRedFact, 1);
    }

  /* on met dans saveaux les variables intermediaires
     que l'on conserve */

  {
    chain_list *intermediate = SL_GUARDEDAUX;
    while (intermediate)
      {
	if (existAuxBeh (SL_BEH, intermediate->DATA))
	  {
	    saveAux = addchain (saveAux, intermediate->DATA);
	  }
	else
	  {
	    printf ("optimization : Warning \n---> bad parameters file\n");
	    printf ("---> signal %s unknown\n", (char *)intermediate->DATA);
	  }
	intermediate = intermediate->NEXT;
	numberInput++;
      }
  }


  tabSol = (tabSolution *) mbkalloc (100 * sizeof (tabSolution));
  numberSol = 0;

  /* on evalue la solution initiale */

  removeAuxBeh (SL_BEH, saveAux);

  if (SL_OPTIM > 2)
    bddToAblBeh (SL_BEH, 1, SL_TRACE);
  else
    bddToAblBeh (SL_BEH, 0, SL_TRACE);

  if (SL_AUXOF)
    createAuxBeh (SL_BEH);

  compactBeh (SL_BEH, saveAux);

  coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
	      &litReg, trace, SL_DELAYPI);
  tabSol[numberSol].lit = lit;
  tabSol[numberSol].maxLit = maxLit;
  tabSol[numberSol].litRed = litRed;
  tabSol[numberSol].operBin = operBin;
  tabSol[numberSol].profMax = profMax;
  tabSol[numberSol].delaisMax = delaisMax;
  tabSol[numberSol].order = copyChain_list (SL_ORDERPI);

  if (SL_TRACE)
    {
      chain_list *ordering;

      printf ("Total number of literals    = %d\n", lit);
      printf ("Number of reduced literals  = %d\n", litRed);
      printf ("Number of latches           = %d\n", litReg);
      printf ("Number of binary operators  = %d\n", operBin);
      printf ("Maximum cone                = %d\n", maxLit);
      printf ("Maximum logical depth       = %d\n", profMax);
      printf ("Maximum delay               = %.3f\n", ((float) delaisMax) / 1000);

      ordering = tabSol[numberSol].order;
      printf ("Order                       =");
      while (ordering)
	{
	  printf (" %s", (char *)ordering->DATA);
	  ordering = ordering->NEXT;
	}
      printf ("\n");
    }

  numberSol++;

  /* on positionne les variables du recuit simule */

  numberInput = numberInput + countInputBeh (SL_BEH);
  amplitude = (int) log ((double) numberInput);
  if (amplitude < 4)
    amplitude = 4;
  numberNode = 2 * amplitude;

  if (SL_MAKEORDER < 2)
    (sl_sys->par)->level = 0;
  level = SL_LEVEL;
  if (level < amplitude)
    amplitude = level - 1;

  numberInput = numberInput / 2;
  if (numberInput < 2)
    numberInput = 2;

  if (trace)
    printf ("\n============== Simulated Anealing ===============\n\n");


  while (amplitude >= 0 && level >= 1)
    {
      int mode;

      mode = 1;

      for (i = 2; i > 0; i--)
	{
	  /* calcul de l'ordonnancement par gradient */

	  if (i == 2)
	    sl_sys->orderPI = makeOrderBddBeh (SL_BEH, SL_OPTIM, SL_OPTIMPO,
					       SL_DELAYPI, 1, numberNode, 1, trace, numberInput, mode, numberAtomRedFact);
	  else
	    sl_sys->orderPI = makeOrderBddBeh (SL_BEH, SL_OPTIM, SL_OPTIMPO,
					       SL_DELAYPI, 1, numberNode, 1, trace, 2000, mode, numberAtomRedFact);
	  makeBddBeh (SL_BEH, SL_CIRCUIT, 0, trace, 200, 1000000, NULL);

	  /* calcul d'une nouvelle solution */

	  if (i == 1 && notSameOrder (tabSol, numberSol, SL_ORDERPI))
	    {
	      /* cette phase n'est pas effectuee si l'ordonnancement 
	         a deja ete trouve */

	      removeAuxBeh (SL_BEH, saveAux);


	      if (SL_OPTIM > 2)
		bddToAblBeh (SL_BEH, 1, SL_TRACE);
	      else
		bddToAblBeh (SL_BEH, 0, SL_TRACE);

	      if (SL_AUXOF)
		createAuxBeh (SL_BEH);

	      compactBeh (SL_BEH, saveAux);

	      coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
			  &litReg, trace, SL_DELAYPI);
	      tabSol[numberSol].lit = lit;
	      tabSol[numberSol].maxLit = maxLit;
	      tabSol[numberSol].litRed = litRed;
	      tabSol[numberSol].operBin = operBin;
	      tabSol[numberSol].profMax = profMax;
	      tabSol[numberSol].delaisMax = delaisMax;
	      tabSol[numberSol].order = copyChain_list (SL_ORDERPI);

	      if (SL_TRACE)
		{
		  chain_list *ordering;

		  printf ("Total number of literals    = %d\n", lit);
		  printf ("Number of reduced literals  = %d\n", litRed);
		  printf ("Number of latches           = %d\n", litReg);
		  printf ("Number of binary operators  = %d\n", operBin);
		  printf ("Maximum cone                = %d\n", maxLit);
		  printf ("Maximum logical depth       = %d\n", profMax);
		  printf ("Maximum delay               = %.3f\n", ((float) delaisMax) / 1000);

		  ordering = tabSol[numberSol].order;
		  printf ("Order                       =");
		  while (ordering)
		    {
		      printf (" %s", (char *)ordering->DATA);
		      ordering = ordering->NEXT;
		    }
		  printf ("\n");
		}

	      numberSol++;

	    }
	}

      /* recuit simule */

      if (amplitude > 0 && level > 1)
	{
	  sl_sys->orderPI = randomOrder (SL_ORDERPI, amplitude);
	  regenereBddBeh (SL_BEH, SL_ORDERPI);
	  resetBdd ();
	  makeBddBeh (SL_BEH, SL_CIRCUIT, 0, trace, 200, 1000000, NULL);
	}
      if (amplitude > 1)
	amplitude--;
      level--;
    }

  if (trace)
    printf ("--- end of annealing\n");

  /* choix de la meilleure solution */

  coutMin = 10000000;
  for (i = 0; i < numberSol; i++)
    {
      int coutI = coutSol (tabSol[i].lit,
			   tabSol[i].delaisMax,
			   tabSol[i].litRed,
			   tabSol[i].profMax,
			   tabSol[i].operBin,
			   SL_OPTIM);
      if (coutMin > coutI)
	{
	  coutMin = coutI;
	  indiceMin = i;
	}
    }

  /* regeneration des graphes */

  if (trace)
    printf ("Graph regeneration %d \n", indiceMin);

  sl_sys->orderPI = tabSol[indiceMin].order;

  regenereBddBeh (SL_BEH, SL_ORDERPI);
  resetBdd ();
  makeBddBeh (SL_BEH, SL_CIRCUIT, 0, trace, 200, 1000000, NULL);
  removeAuxBeh (SL_BEH, saveAux);

  if (trace)
    {
      gcNodeBeh (SL_BEH, 0);
      printf ("$$$$$$ final number of nodes = %d\n\n", numberNodeAllBdd ());
    }

  if (SL_MAKEORDER < 2 || trace)
    {
      printf ("Cost Function --- litt. = %d ", numberAtomRedBeh (SL_BEH));
      gcNodeBeh (SL_BEH, 0);
      printf ("--- nodes = %d\n\n", numberNodeAllBdd ());
    }

  if (SL_OPTIM > 2)
    bddToAblBeh (SL_BEH, 1, SL_TRACE);
  else
    bddToAblBeh (SL_BEH, 0, SL_TRACE);

  if (trace)
    printf ("Generating VI\n");

  if (SL_AUXOF)
    createAuxBeh (SL_BEH);

  compactBeh (SL_BEH, saveAux);

  /* on inverse l'ordre des expressions et 
     on cree des variable Inter a nouveau */

  mapCarExprBeh (SL_BEH, reverseExpr);

  if (SL_AUXOF)
    createAuxBeh (SL_BEH);

  compactBeh (SL_BEH, saveAux);

  mapCarExprBeh (SL_BEH, reverseExpr);
}


/*-------------------------------------------------------------------------
annealingLOGIC 	: recuit simule pour l'optimisation comportementale 
	          on prend en entree une befig qui possede un certain nombre
		  de graphes deja calcules.
---------------------------------------------------------------------------
retour		: un void.
---------------------------------------------------------------------------*/

void
annealingLOGIC (trace)
     int trace;
{
  int lit, maxLit, litRed, litReg, profMax, delaisMax, operBin;


  if (SL_AUX)
    printf ("Running Local Optimizer on `%s`\n\n", SL_NAME);
  else
    printf ("Running Global Optimizer on `%s`\n\n", SL_NAME);

  coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
	      &litReg, trace, SL_DELAYPI);
  printf ("=============================  INITIAL COST  ==================================\n");

  printf ("Total number of literals    = %d\n", lit);
  printf ("Number of reduced literals  = %d\n", litRed);
  printf ("Number of latches           = %d\n", litReg);
  printf ("Maximum logical depth       = %d\n", profMax);
  printf ("Maximum delay               = %.3f\n", ((float) delaisMax) / 1000);

  printf ("===============================================================================\n");

  if (SL_AUX)
    localOpt (trace);
  else
    globalOpt (trace);

  /* pro-crastination des VI */

  if (PROCRASTINATION)
    {
      int i;

      for (i = 0; i < PROCRASTINATION; i++)
	{
	  chain_list *order = copyChain_list (SL_ORDERPI);

	  if (trace)
	    printf ("******  Procrastination of VI\n");

	  optimBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO, order);

	  if (trace == 2)
	    printf ("******* End of procrastination\n");

	  regenereBddBeh (SL_BEH, SL_ORDERPI);
	  resetBdd ();
	  makeBddBeh (SL_BEH, SL_CIRCUIT, 0, trace, 200, 1000000, NULL);

	  /* on normalise les expressions avant de creer les auxiliaires */

	  mapExprBeh (SL_BEH, normExpr);

	  if (SL_AUXOF)
	    createAuxBeh (SL_BEH);

	  if (trace == 2)
	    {
	      displayBeh (SL_BEH, 1);
	      printf ("Collapsing\n");
	    }

	  /* re-ordonnancement */

	  compactExprBeh (SL_BEH, SL_OPTIM, SL_DELAYPI, SL_OPTIMPO, 1, SL_GUARDEDAUX);
	}
    }

  coutAblBeh (SL_BEH, &lit, &maxLit, &litRed, &profMax, &delaisMax, &operBin,
	      &litReg, trace, SL_DELAYPI);

  printf ("==============================  FINAL COST  ===================================\n");

  printf ("Total number of literals    = %d\n", lit);
  printf ("Number of reduced literals  = %d\n", litRed);
  printf ("Number of latches           = %d\n", litReg);
  printf ("Maximum logical depth       = %d\n", profMax);
  printf ("Maximum delay               = %.3f\n", ((float) delaisMax) / 1000);

  printf ("===============================================================================\n");
}
