/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: sl_facto_bdd2.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include LOG_H
#include BEH_H

/*------------------------------------------------------------------------------
gcNodeBeh	 :effectue un garbage collecteur sur tous les noeuds de systeme
		  en sauvegardant les noeuds pointes par les pNode de la liste 
		  des befig .
-------------------------------------------------------
parametres 	 : pointeur sur befig .
-------------------------------------------------------
return 		 : 1 si ok ;
		   0 si erreur .
------------------------------------------------------------------------------*/
void
gcNodeBeh (beh, lstbeh)
     befig_list *beh;
     int lstbeh;
{
  struct systemBdd sysBddAux;
  beout_list *out;
  bereg_list *reg;
  bebus_list *bus;
  bebux_list *bux;
  binode_list *binode;
  beaux_list *aux;
  bemsg_list *mesg;
  pNode zeroAux, oneAux;
  pTH pTHNode;

  if (sysBdd.pRT->compteur > LARGE)
    {
      pTHNode = createTH (LARGE);
      sysBddAux.pRT = createTableBdd (LARGE);
    }
  else if (sysBdd.pRT->compteur > MEDIUM)
    {
      pTHNode = createTH (MEDIUM);
      sysBddAux.pRT = createTableBdd (MEDIUM);
    }
  else
    {
      pTHNode = createTH (SMALL);
      sysBddAux.pRT = createTableBdd (SMALL);
    }

  sysBddAux.pMC = createTabLoc (MEDIUM);
  sysBddAux.indiceAT = MAX_PACK;
  sysBddAux.lpAT = NULL;
  zeroAux = initVertexBddAux (0, (pNode)0, (pNode)1, &sysBddAux);
  oneAux = initVertexBddAux (1, (pNode)0, (pNode)1, &sysBddAux);
  addTH (pTHNode, (char *)zero, (int)zeroAux);
  addTH (pTHNode, (char *)one, (int)oneAux);

  /* si lstbeh = 1 on garbage sur toutes les befig */
  /* sinon on ne prends que la premiere befig */

  do
    {
      out = beh->BEOUT;
      while (out)
	{
	  if (out->NODE)
	    {
	      out->NODE = regenereBdd (out->NODE, &sysBddAux, pTHNode);
	    }
	  out = out->NEXT;
	}

      aux = beh->BEAUX;
      while (aux)
	{
	  if (aux->NODE)
	    {
	      aux->NODE = regenereBdd (aux->NODE, &sysBddAux, pTHNode);
	    }
	  aux = aux->NEXT;
	}

      reg = beh->BEREG;
      while (reg)
	{
	  binode = reg->BINODE;
	  while (binode)
	    {
	      if (binode->CNDNODE && binode->VALNODE)
		{
		  binode->CNDNODE = regenereBdd (binode->CNDNODE, &sysBddAux, pTHNode);
		  binode->VALNODE = regenereBdd (binode->VALNODE, &sysBddAux, pTHNode);
		}
	      binode = binode->NEXT;
	    }
	  reg = reg->NEXT;
	}

      bus = beh->BEBUS;
      while (bus)
	{
	  binode = bus->BINODE;
	  while (binode)
	    {
	      if (binode->CNDNODE && binode->VALNODE)
		{
		  binode->CNDNODE = regenereBdd (binode->CNDNODE, &sysBddAux, pTHNode);
		  binode->VALNODE = regenereBdd (binode->VALNODE, &sysBddAux, pTHNode);
		}
	      binode = binode->NEXT;
	    }
	  bus = bus->NEXT;
	}

      bux = beh->BEBUX;
      while (bux)
	{
	  binode = bux->BINODE;
	  while (binode)
	    {
	      if (binode->CNDNODE && binode->VALNODE)
		{
		  binode->CNDNODE = regenereBdd (binode->CNDNODE, &sysBddAux, pTHNode);
		  binode->VALNODE = regenereBdd (binode->VALNODE, &sysBddAux, pTHNode);
		}
	      binode = binode->NEXT;
	    }
	  bux = bux->NEXT;
	}

      mesg = beh->BEMSG;
      while (mesg)
	{
	  mesg->NODE = regenereBdd (mesg->NODE, &sysBddAux, pTHNode);
	  mesg = mesg->NEXT;
	}

      beh = beh->NEXT;
    }
  while (beh && lstbeh == 1);

  destroyBdd (1);
  destroyTH (pTHNode);
  sysBdd.pRT = sysBddAux.pRT;
  sysBdd.pMC = sysBddAux.pMC;
  sysBdd.indiceAT = sysBddAux.indiceAT;
  sysBdd.lpAT = sysBddAux.lpAT;
  sysBdd.pAT = sysBddAux.pAT;
  zero = zeroAux;
  one = oneAux;
}
