/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */




#ident  "$Id: sl_facto_jacoX.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include LOG_H
#include BEH_H

#include "lax_param.h"
#include "sl_local.h"


#define TAILLE_X 100
#define ASSIGN 1

#define LBDD_LOW 0
#define LBDD_HIGH 1


pNode XTWO;
struct assign *bestAssign;
int bestSize;
int numSol;


/*-------------------------------------------------------------------------
upVarAllOneBdd  : remonte tous les index de 1 dans un graphe  
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
pNode
upVarAllOneBdd (pt)
     pNode pt;
{
  if (pt->index < 2)
    return (pt);

  return (initVertexBdd (pt->index + 1, upVarAllOneBdd (pt->high),
			 upVarAllOneBdd (pt->low)));
}

/*-------------------------------------------------------------------------
oneByXBdd       : substitue les noeuds one par un noeud d'index 2 
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
pNode
oneByXBdd (pt)
     pNode pt;
{
  if (pt == zero)
    return zero;
  if (pt == one)
    return XTWO;

  return (initVertexBdd (pt->index + 1, oneByXBdd (pt->high),
			 oneByXBdd (pt->low)));
}

/*-------------------------------------------------------------------------
extractX_fathers    : calcule les noeuds peres de X 
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
void
extractX_fathers (pX, fathers, pCount)
     pNode pX;
     struct X_fathers *fathers;
     int *pCount;
{
  int i, j;
  int trouve = 0;

  if (pX->index < 3)
    return;


  if (pX->low == XTWO || pX->high == XTWO)
    {
      for (i = 0; i < *pCount; i++)
	{
	  if (pX == fathers[i].x)
	    {
	      trouve = 1;
	      break;
	    }
	}

      if (trouve == 0)		/* nouveau pere de X */
	{
	  /* on insere le nouveau pere pour donner
	     un ordre par index decroissant */

	  for (i = 0; i < *pCount; i++)
	    if ((fathers[i].x)->index < pX->index)
	      break;

	  /* decalage pour l'insertion */

	  for (j = *pCount; j > i; j--)
	    fathers[j].x = fathers[j - 1].x;

	  fathers[i].x = pX;
	  *pCount = *pCount + 1;

	  /* on avance dans l'autre branche */

	  if (pX->index > 2)
	    {
	      if (pX->low == XTWO)
		extractX_fathers (pX->high, fathers, pCount);
	      else
		extractX_fathers (pX->low, fathers, pCount);
	    }
	}

      return;
    }
  else
    /* on avance dans les deux fils */
    {
      extractX_fathers (pX->high, fathers, pCount);
      extractX_fathers (pX->low, fathers, pCount);
    }
}

/*-------------------------------------------------------------------------
computeBrothers : calcule les freres a chaque niveau d'index 
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
void
computeBrothers (pX, brothers)
     pNode pX;
     struct eqNode **brothers;
{
  struct eqNode *lst;
  int trouve = 0;

  if (pX->index < 3)
    return;

  lst = brothers[pX->index];
  while (lst && trouve == 0)
    {
      if (lst->x == pX)
	trouve = 1;
      lst = lst->NEXT;
    }

  if (!trouve)
    {
      struct eqNode *newEqNode;

      newEqNode = (struct eqNode *) mbkalloc (sizeof (struct eqNode));
      newEqNode->x = pX;
      newEqNode->NEXT = brothers[pX->index];
      brothers[pX->index] = newEqNode;
    }
  else
    return;

  computeBrothers (pX->high, brothers);
  computeBrothers (pX->low, brothers);
}

/*-------------------------------------------------------------------------
fusionAssign    : fusionne deux liste d'association
---------------------------------------------------------------------------
retour          : une liste d'association 
---------------------------------------------------------------------------*/
struct assign *
fusionAssign (lst1, lst2)
     struct assign *lst1, *lst2;
{
  struct assign *lst;

  lst = lst1;
  if (lst)
    {
      struct assign *lstp;

      /* verification d'unicite des assignations */

      while (lst)
	{
	  lstp = lst2;
	  while (lstp)
	    {
	      /* x est assigne a deux noeuds differents */

	      if (lst->x == lstp->x)
		{
		  if (lst->y != lstp->y)
		    {
		      printf ("** node assigned twice\n");
		      return NULL;
		    }
		}
	      lstp = lstp->NEXT;
	    }
	  lst = lst->NEXT;
	}

      /* concatenation des liste d'associations */

      lst = lst1;
      while (lst->NEXT)
	{
	  lst = lst->NEXT;
	}
      lst->NEXT = lst2;
      return lst1;
    }
  else
    return lst2;
}
/*-------------------------------------------------------------------------
addAssign       :ajoute un element dans une liste d'association 
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
struct assign *
addAssign (lst1, x, fils, y)
     struct assign *lst1;
     pNode x;
     int fils;
     pNode y;
{
  struct assign *lst = (struct assign *) mbkalloc (sizeof (struct assign));
  lst->NEXT = lst1;
  lst->x = x;
  lst->fils = fils;
  lst->y = y;
  return lst;
}

/*-------------------------------------------------------------------------
copyAssign    : copie une liste d'associations
---------------------------------------------------------------------------
retour        : une liste d'associations 
---------------------------------------------------------------------------*/
struct assign *
copyAssign (lst)
     struct assign *lst;
{
  struct assign *lst1 = NULL;
  struct assign *tete;

  if (lst)
    {
      while (lst)
	{
	  if (!lst1)
	    {
	      lst1 = addAssign (NULL, lst->x, lst->fils, lst->y);
	      tete = lst1;
	    }
	  else
	    {
	      lst1->NEXT = addAssign (NULL, lst->x, lst->fils, lst->y);
	      lst1 = lst1->NEXT;
	    }
	  lst = lst->NEXT;
	}
      return tete;
    }
  else
    return NULL;
}
/*-------------------------------------------------------------------------
freeEqNode      : desalloue une liste d'association 
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
void
freeEqNode (lst1)
     struct eqNode *lst1;
{
  struct eqNode *lst;

  if (lst1)
    {
      while (lst1 && lst1->NEXT)
	{
	  lst = lst1->NEXT;
	  mbkfree (lst1);
	  lst1 = lst;
	}
      mbkfree (lst1);
    }
}

/*-------------------------------------------------------------------------
freeAssign      : desalloue une liste d'association 
---------------------------------------------------------------------------
retour          : 
---------------------------------------------------------------------------*/
void
freeAssign (lst1)
     struct assign *lst1;
{
  struct assign *lst;

  if (lst1)
    {
      while (lst1 && lst1->NEXT)
	{
	  lst = lst1->NEXT;
	  mbkfree (lst1);
	  lst1 = lst;
	}
      mbkfree (lst1);
    }
}

/*-------------------------------------------------------------------------
matchBdd       : teste si deux couples peuvent etre equivalents 
---------------------------------------------------------------------------
retour          : une liste d'association 
---------------------------------------------------------------------------*/
struct assign *
matchBdd (pX1, pX2, pMatch)
     pNode pX1, pX2;
     int *pMatch;
{
  struct assign *lst1, *lst2;

  /* egalite */

  if (pX1 == pX2)
    return NULL;

  if (pX1 == XTWO || pX2 == XTWO)
    {
      printf ("Error : matchBdd - X found !\n");
      exit (-1);
    }

  /* noeuds terminaux => differents */

  if (pX1->index < 2 && pX2->index < 2)
    {
      *pMatch = 0;
      return NULL;
    }

  /* index differents :
     pX1 et pX2 peuvent matcher si l'on peut descendre dans
     le graphe d'index max. (cas du fils a X)
   */

  if (pX1->index != pX2->index)
    {
      if (pX1->index > pX2->index)
	{
	  if (pX1->low == XTWO)
	    {
	      lst1 = matchBdd (pX1->high, pX2, pMatch);
	      if (*pMatch)
		{
		  lst1 = addAssign (lst1, pX1, LBDD_LOW, pX1->high);
		  return lst1;
		}
	      return NULL;
	    }
	  if (pX1->high == XTWO)
	    {
	      lst1 = matchBdd (pX1->low, pX2, pMatch);
	      if (*pMatch)
		{
		  lst1 = addAssign (lst1, pX1, LBDD_HIGH, pX1->low);
		  return lst1;
		}
	      return NULL;
	    }
	}
      else
	{
	  if (pX2->low == XTWO)
	    {
	      lst1 = matchBdd (pX1, pX2->high, pMatch);
	      if (*pMatch)
		{
		  lst1 = addAssign (lst1, pX2, LBDD_LOW, pX2->high);
		  return lst1;
		}
	      return NULL;
	    }
	  if (pX2->high == XTWO)
	    {
	      lst1 = matchBdd (pX1, pX2->low, pMatch);
	      if (*pMatch)
		{
		  lst1 = addAssign (lst1, pX2, LBDD_HIGH, pX2->low);
		  return lst1;
		}
	      return lst1;
	    }
	}
      *pMatch = 0;
      return NULL;
    }

  /* index egaux */

  /* cas des peres de X */

  if (pX1->low == XTWO)
    {
      if (pX2->high != XTWO && pX2->low != XTWO)
	{
	  lst1 = matchBdd (pX1->high, pX2->high, pMatch);
	  if (*pMatch)
	    {
	      lst1 = addAssign (lst1, pX1, LBDD_LOW, pX2->low);
	      return lst1;
	    }
	  return NULL;
	}

      if (pX2->high == XTWO)
	{
	  lst1 = addAssign (NULL, pX1, LBDD_LOW, pX2->low);
	  lst1 = addAssign (lst1, pX2, LBDD_HIGH, pX1->high);
	  return lst1;
	}

      if (pX2->low == XTWO)	/* seul le fils droit doit matcher */
	{
	  return matchBdd (pX1->high, pX2->high, pMatch);
	}
    }

  if (pX1->high == XTWO)
    {
      if (pX2->high != XTWO && pX2->low != XTWO)
	{
	  lst1 = matchBdd (pX1->low, pX2->low, pMatch);
	  if (*pMatch)
	    {
	      lst1 = addAssign (lst1, pX1, LBDD_HIGH, pX2->high);
	      return lst1;
	    }
	  return NULL;
	}

      if (pX2->low == XTWO)
	{
	  lst1 = addAssign (NULL, pX1, LBDD_HIGH, pX2->high);
	  lst1 = addAssign (lst1, pX2, LBDD_LOW, pX1->low);
	  return lst1;
	}

      if (pX2->high == XTWO)	/* seul le fils droit doit matcher */
	{
	  return matchBdd (pX1->low, pX2->low, pMatch);
	}
    }

  /* il reste a tester les cas ou X2 est un pere de X */

  if (pX2->low == XTWO)
    {
      lst1 = matchBdd (pX2->high, pX1->high, pMatch);
      if (*pMatch)
	{
	  lst1 = addAssign (lst1, pX2, LBDD_LOW, pX1->low);
	  return lst1;
	}
      return NULL;
    }

  if (pX2->high == XTWO)
    {
      lst1 = matchBdd (pX2->low, pX1->low, pMatch);
      if (*pMatch)
	{
	  lst1 = addAssign (lst1, pX2, LBDD_HIGH, pX1->high);
	  return lst1;
	}
      return NULL;
    }

  /* cas des noeuds qui ne sont pas des peres de X */

  lst1 = matchBdd (pX1->high, pX2->high, pMatch);

  if (*pMatch == 0)
    return NULL;

  lst2 = matchBdd (pX1->low, pX2->low, pMatch);

  if (*pMatch)
    {
      lst1 = fusionAssign (lst1, lst2);
      if (lst1)
	return lst1;

      /* cas ou la fusion a indique une assignation double d'un noeud */

      freeAssign (lst1);
      freeAssign (lst2);
      *pMatch = 0;
      return NULL;
    }
  else
    {
      freeAssign (lst1);
      return NULL;
    }
}
/*-------------------------------------------------------------------------
applyAssign   : evalue une solution d'association
---------------------------------------------------------------------------
retour        : un pNode 
---------------------------------------------------------------------------*/
pNode
applyAssign (pX, finalAssign)
     pNode pX;
     struct assign *finalAssign;
{

  if (pX->index < 2)
    return pX;
  if (pX->index == 2)
    {
      printf ("Error : X not assigned\n");
      exit (-1);
    }

  if (pX->low == XTWO || pX->high == XTWO)
    {
      struct assign *ass;

      /* recherche de l'assignation dans finalAssign */

      ass = finalAssign;
      while (ass)
	{
	  pNode sub;

	  if (ass->x == pX)
	    {
	      if (ass->fils == LBDD_LOW)
		sub = initVertexBdd (pX->index, pX->high, ass->y);
	      else
		sub = initVertexBdd (pX->index, ass->y, pX->low);

	      return applyAssign (sub, finalAssign);
	    }

	  ass = ass->NEXT;
	}
      printf ("Error : %p not assigned\n", pX);
      exit (-1);
    }

  return initVertexBdd (pX->index - 1, applyAssign (pX->high, finalAssign),
			applyAssign (pX->low, finalAssign));
}

/*-------------------------------------------------------------------------
getBestAssign   : calcule la meilleure assignation 
---------------------------------------------------------------------------
retour          : (met a jour deux variables globales : bestAssign et BestSize) 
---------------------------------------------------------------------------*/
void
getBestAssign (pX, fathers, count, tabFathers, finalAssign)
     pNode pX;
     struct X_fathers *fathers;
     int count;
     pTH tabFathers;
     struct assign *finalAssign;
{
  int allAssigned = 1;
  struct assign *ass;
  int i;

/*
   printf("getBest : ");
   ass = finalAssign;
   while (ass)
   {
   printf("[%d,%d,%d] ",ass->x,ass->fils,ass->y);
   ass = ass->NEXT;
   }
   printf("\n");
 */

  /* est-ce que tous les peres sont deja assignes ? */

  for (i = 0; i < count; i++)
    {
      if (fathers[i].assigned == 0)
	{
	  allAssigned = 0;
	  break;
	}
    }
  /* evaluation -> feuille terminale de l'arbre de recherche */

  if (allAssigned)
    {
      pNode pEX;
      int size;

      printf ("TERM(%d) : ", numSol);

      if (numSol > 1000)
	return;

      ass = finalAssign;
      while (ass)
	{
	  printf ("[%p,%d,%p] ", ass->x, ass->fils, ass->y);
	  ass = ass->NEXT;
	}

      pEX = applyAssign (pX, finalAssign);
      size = numberNodeBdd (pEX);
      printf ("- Size = %d\n", size);

      displayBdd (pEX, 0);

      numSol++;
      if (size < bestSize)
	{
	  bestSize = size;
	  freeAssign (bestAssign);
	  bestAssign = finalAssign;
	}
      else
	freeAssign (finalAssign);
    }
  else
    {
      struct eqNode *lst;

      printf ("father[%d] --> ", i);
      fathers[i].assigned = 1;

      lst = fathers[i].lst;

      if (numSol > 1000)
	return;
      /* on essaye chaque association */
      while (lst)
	{
	  int alreadyAssigned = 0;

	  printf (" Equivalent Node = %p\n", lst->x);

	  /* on verifie qu'une des assignations de lst->ass
	     ne concerne pas un pere deja assigne */

	  ass = lst->ass;
	  while (ass)
	    {
	      int indice = searchTH (tabFathers, (char *)ass->x);

	      if (indice != i)
		{
		  if ((fathers[indice].assigned == 1))
		    {
		      struct assign *ptFinal;

		      /* est-ce que cette assignation est la meme ? */

		      ptFinal = finalAssign;
		      while (ptFinal)
			{
			  if (ass->x == ptFinal->x)
			    {
			      if (ass->y != ptFinal->y)
				{
				  alreadyAssigned = 1;
				  printf ("Warning : %p already assigned\n", ass->x);
				}
			      break;
			    }
			  ptFinal = ptFinal->NEXT;
			}
		    }
		}
	      ass = ass->NEXT;
	    }

	  if (!alreadyAssigned)
	    {
	      struct assign *callAssign = copyAssign (finalAssign);

	      /* on ajoute les associations a la liste courante */

	      if (callAssign)
		{
		  struct assign *interAssign = callAssign;
		  while (interAssign->NEXT)
		    interAssign = interAssign->NEXT;
		  interAssign->NEXT = copyAssign (lst->ass);
		}
	      else
		callAssign = copyAssign (lst->ass);

	      /* on elimine les peres assignes */

	      ass = lst->ass;
	      while (ass)
		{
		  int indice = searchTH (tabFathers, (char *)ass->x);

		  fathers[indice].assigned = 1;
		  ass = ass->NEXT;
		}


	      getBestAssign (pX, fathers, count, tabFathers, callAssign);

	      /* on remet a jour le contexte */
	      ass = lst->ass;
	      while (ass)
		{
		  int indice = searchTH (tabFathers, (char *)ass->x);

		  fathers[indice].assigned = 0;
		  ass = ass->NEXT;
		}
	    }
	  lst = lst->NEXT;
	}
    }
}

/*-------------------------------------------------------------------------
assign01Bdd       : assigne un MBD a 0 ou a 1 si possible 
---------------------------------------------------------------------------
retour          : un BDD 
---------------------------------------------------------------------------*/
pNode
assign01Bdd (pON)
     pNode pON;
{
  pNode low, high;

  if (pON->index < 2)
    return pON;

  if (pON->low == XTWO)
    return assign01Bdd (pON->high);

  if (pON->high == XTWO)
    return assign01Bdd (pON->low);

  high = assign01Bdd (pON->high);
  low = assign01Bdd (pON->low);

  if (high == low)
    return high;

  /* assignation des fils ? */

  /* si le nombre de noeuds du fils que l'on peut assigner
     est superieur ou egal a 3 (1 pour X),
     on conserve la valeur assignee a 0 ou a 1 */

  if (high->index < 2)
    {
      if (numberNodeBdd (pON->high) < 4)
	high = pON->high;
    }

  if (low->index < 2)
    {
      if (numberNodeBdd (pON->low) < 4)
	low = pON->low;
    }
  /* pas d'assignation */

  return initVertexBdd (pON->index, high, low);
}

/*-------------------------------------------------------------------------
jacoBdd       : minimise un Bdd avec ses DC 
---------------------------------------------------------------------------
retour          : un BDD simplifie 
---------------------------------------------------------------------------*/
pNode
jacoBdd (pON, pDC)
     pNode pON, pDC;
{
  struct X_fathers *fathers;
  struct eqNode **brothers;
  int i;
  int count = 0;
  pNode pX;
  pTH tabFathers;

  XTWO = createNodeTermBdd (2);

  /* calcul d'un graphe avec X (index 2) */

  pX = applyBinBdd (OR, upVarAllOneBdd (pON), oneByXBdd (pDC));

  printf ("==== pX \n");
  displayGraphicBdd (pX);
  displayBdd (pX, 1);

  /* simplification par assignation de ss-graphes a 0 ou 1 */

  if (ASSIGN)
    {
      pX = assign01Bdd (pX);
      printf ("==== pX after assignations\n");
      displayGraphicBdd (pX);
    }

  /* initialisation de la structure fathers */

  fathers = (struct X_fathers *) mbkalloc (TAILLE_X * sizeof (struct X_fathers));
  for (i = 0; i < TAILLE_X; i++)
    {
      fathers[i].lst = NULL;
      fathers[i].assigned = 0;
    }

  /* recherche des noeuds ayant un fils a X */

  extractX_fathers (pX, fathers, &count);

  /* calcul des freres */

  brothers = (struct eqNode **) mbkalloc ((pX->index + 1) * sizeof (struct eqNode *));
  for (i = 0; i <= pX->index; i++)
    brothers[i] = NULL;

  computeBrothers (pX, brothers);

/*
   for (i=0;i<=pX->index;i++)
   {
   struct eqNode *lst;

   lst = brothers[i];
   printf("brother[%d]\n",i);
   while (lst)
   {
   printf("x : ");displayBdd(lst->x,0);
   lst = lst->NEXT;
   }
   }
 */

  /* calcul des noeuds equivalents pour chaque fathers */

  for (i = 0; i < count; i++)
    {
      struct eqNode *lst;

      /* minimisation locale */

      if ((fathers[i].x)->low == XTWO)
	{
	  struct assign *ass;
	  struct eqNode *newEqNode;

	  newEqNode = (struct eqNode *) mbkalloc (sizeof (struct eqNode));
	  newEqNode->x = (fathers[i].x)->high;
	  ass = (struct assign *) mbkalloc (sizeof (struct assign));
	  ass->NEXT = NULL;
	  ass->x = fathers[i].x;
	  ass->fils = LBDD_LOW;
	  ass->y = (fathers[i].x)->high;
	  newEqNode->ass = ass;
	  newEqNode->NEXT = fathers[i].lst;
	  fathers[i].lst = newEqNode;
	}
      else
	{
	  struct assign *ass;
	  struct eqNode *newEqNode;

	  newEqNode = (struct eqNode *) mbkalloc (sizeof (struct eqNode));
	  newEqNode->x = (fathers[i].x)->low;
	  ass = (struct assign *) mbkalloc (sizeof (struct assign));
	  ass->NEXT = NULL;
	  ass->x = fathers[i].x;
	  ass->fils = LBDD_HIGH;
	  ass->y = (fathers[i].x)->low;
	  newEqNode->ass = ass;
	  newEqNode->NEXT = fathers[i].lst;
	  fathers[i].lst = newEqNode;
	}
      /* on regarde les freres du noeud */

      lst = brothers[(fathers[i].x)->index];
      while (lst)
	{
	  struct assign *lstMatch = NULL;
	  int match = 1;

	  if (lst->x != fathers[i].x)
	    {
	      lstMatch = (struct assign *) matchBdd (fathers[i].x, lst->x, &match);

	      /* fathers[i] et lst matchent ... */

	      if (match)
		{
		  struct eqNode *newEqNode;

		  /* le pere de X n'a pas ete assigne ? */

		  if (lstMatch->x != fathers[i].x)
		    {
		      freeAssign (lstMatch);
		    }
		  else
		    {
		      newEqNode = (struct eqNode *) mbkalloc (sizeof (struct eqNode));
		      newEqNode->x = lst->x;
		      newEqNode->ass = lstMatch;
		      newEqNode->NEXT = fathers[i].lst;
		      fathers[i].lst = newEqNode;
		    }
		}
	    }
	  lst = lst->NEXT;
	}
    }

  /* affichage */

  for (i = 0; i < count; i++)
    {
      struct eqNode *lst;
      struct assign *ass;

      printf ("============================ fathers[%d] :\n", i);
      displayBdd (fathers[i].x, 1);

      lst = fathers[i].lst;
      while (lst)
	{
	  printf ("Eq. node %p : \n", lst->x);
	  printf ("assoc. list : ");
	  ass = lst->ass;
	  while (ass)
	    {
	      printf ("[%p,%d,%p] ", ass->x, ass->fils, ass->y);
	      ass = ass->NEXT;
	    }
	  printf ("\n");
	  lst = lst->NEXT;
	}
    }

  /* assignation des X */

  tabFathers = createTH (100);
  for (i = 0; i < count; i++)
    {
      addTH (tabFathers, (char *)fathers[i].x, i);
    }

  bestAssign = NULL;
  bestSize = 100000;
  numSol = 0;

  getBestAssign (pX, fathers, count, tabFathers, NULL);

  if (bestAssign)
    printf ("Warning : no assignment found\n");
  pON = applyAssign (pX, bestAssign);
  freeAssign (bestAssign);
  printf ("resultat final = %p\n", pON);
  destroyTH (tabFathers);

  /* desa. de la structure fathers */

  for (i = 0; i < count; i++)
    {
      struct eqNode *lst;

      lst = fathers[i].lst;
      while (lst)
	{
	  freeAssign (lst->ass);
	  lst = lst->NEXT;
	}
      mbkfree (fathers[i].lst);
    }

  mbkfree (fathers);
  return pON;
}

/*-------------------------------------------------------------------------
getBestAssignMulti   : calcule la meilleure assignation 
---------------------------------------------------------------------------
retour          : (met a jour deux variables globales : bestAssign et BestSize) 
---------------------------------------------------------------------------*/
void
getBestAssignMulti (pX, fathers, count, tabFathers, finalAssign, numBdd)
     pNode *pX;
     struct X_fathers *fathers;
     int count;
     pTH tabFathers;
     struct assign *finalAssign;
     int numBdd;
{
  int allAssigned = 1;
  struct assign *ass;
  int i;

/*
   printf("getBest : ");
   ass = finalAssign;
   while (ass)
   {
   printf("[%d,%d,%d] ",ass->x,ass->fils,ass->y);
   ass = ass->NEXT;
   }
   printf("\n");
 */

  for (i = 0; i < count; i++)
    {
      if (fathers[i].assigned == 0)
	{
	  allAssigned = 0;
	  break;
	}
    }
  /* evaluation -> feuille terminale de l'arbre de recherche */

  if (allAssigned)
    {
      pNode *pEX;
      int size;

      printf ("TERM(%d) : ", numSol);

      if (numSol > 1000)
	return;

      ass = finalAssign;
      while (ass)
	{
	  printf ("[%p,%d,%p] ", ass->x, ass->fils, ass->y);
	  ass = ass->NEXT;
	}

      pEX = (pNode *) mbkalloc (numBdd * sizeof (pNode));
      for (i = 0; i < numBdd; i++)
	{
	  pEX[i] = applyAssign (pX[i], finalAssign);
	}
     size = costTab (pEX, numBdd, countNode);
      printf ("- Size Multi-MBD = %d\n", size);

      numSol++;
      if (size < bestSize)
	{
	  bestSize = size;
	  freeAssign (bestAssign);
	  bestAssign = finalAssign;
	}
      else
	freeAssign (finalAssign);
    }
  else
    {
      struct eqNode *lst;

      printf ("father[%d] --> ", i);
      fathers[i].assigned = 1;

      lst = fathers[i].lst;

      if (numSol > 1000)
	return;
      /* on essaye chaque association */
      while (lst)
	{
	  int alreadyAssigned = 0;

	  printf (" Equivalent Node = %p\n", lst->x);

	  /* on elimine les associations sous-jacentes */
	  ass = lst->ass;
	  while (ass)
	    {
	      int indice = searchTH (tabFathers, (char *)ass->x);

	      if ((indice != EMPTYTH) && (indice != i))
		{
		  if ((fathers[indice].assigned == 1))
		    {
		      struct assign *ptFinal;

		      /* est-ce que cette assignation est la meme ? */

		      ptFinal = finalAssign;
		      while (ptFinal)
			{
			  if (ass->x == ptFinal->x)
			    {
			      if (ass->y != ptFinal->y)
				{
				  alreadyAssigned = 1;
				  printf ("Warning : %p already assigned\n", ass->x);
				}
			      break;
			    }
			  ptFinal = ptFinal->NEXT;
			}
		    }
		}
	      ass = ass->NEXT;
	    }

	  if (!alreadyAssigned)
	    {
	      struct assign *callAssign = copyAssign (finalAssign);

	      /* on ajoute les associations a la liste courante */

	      if (callAssign)
		{
		  struct assign *interAssign = callAssign;
		  while (interAssign->NEXT)
		    interAssign = interAssign->NEXT;
		  interAssign->NEXT = copyAssign (lst->ass);
		}
	      else
		callAssign = copyAssign (lst->ass);

	      /* on elimine les peres assignes */

	      ass = lst->ass;
	      while (ass)
		{
		  int indice = searchTH (tabFathers, (char *)ass->x);

		  if (indice != EMPTYTH)
		    fathers[indice].assigned = 1;
		  ass = ass->NEXT;
		}


	      getBestAssignMulti (pX, fathers, count, tabFathers, callAssign, numBdd);

	      /* on remet a jour le contexte */
	      ass = lst->ass;
	      while (ass)
		{
		  int indice = searchTH (tabFathers, (char *)ass->x);

		  if (indice != EMPTYTH)
		    fathers[indice].assigned = 0;
		  ass = ass->NEXT;
		}
	    }
	  lst = lst->NEXT;
	}
    }
}

/*-------------------------------------------------------------------------
jacoMultiBdd  : minimise un multi-BDD avec ses DC 
---------------------------------------------------------------------------
retour          : un multi-BDD simplifie 
---------------------------------------------------------------------------*/
void
jacoMultiBdd (pON, pDC, numBdd)
     pNode *pON, *pDC;
     int numBdd;
{
  struct X_fathers *fathers;
  struct eqNode **brothers;
  int i;
  int count = 0;
  pNode *pX;
  int indexMax;
  pTH tabFathers;

  XTWO = createNodeTermBdd (2);

  pX = (pNode *) mbkalloc (numBdd * sizeof (pNode));

  /* on calcule le multi-MBD et l'index maximun */

  indexMax = 0;
  for (i = 0; i < numBdd; i++)
    {
      pX[i] = applyBinBdd (OR, upVarAllOneBdd (pON[i]), oneByXBdd (pDC[i]));

      /* simplification par assignation de ss-graphes a 0 ou 1 */

      if (ASSIGN)
	pX[i] = assign01Bdd (pX[i]);

      if (indexMax < pON[i]->index)
	indexMax = pON[i]->index;
    }
  indexMax++;

  /* initialisation de la structure fathers */

  fathers = (struct X_fathers *) mbkalloc (TAILLE_X * sizeof (struct X_fathers));
  for (i = 0; i < TAILLE_X; i++)
    {
      fathers[i].lst = NULL;
      fathers[i].assigned = 0;
    }

  /* recherche des noeuds ayant un fils a X */

  for (i = 0; i < numBdd; i++)
    {
      extractX_fathers (pX[i], fathers, &count);
    }

  /* calcul des freres */

  brothers = (struct eqNode **) mbkalloc ((indexMax + 1) * sizeof (struct eqNode *));
  for (i = 0; i <= indexMax; i++)
    brothers[i] = NULL;

  for (i = 0; i < numBdd; i++)
    computeBrothers (pX[i], brothers);

/*
   for (i=0;i<=indexMax;i++)
   {
   struct eqNode *lst;

   lst = brothers[i];
   printf("brother[%d]\n",i);
   while (lst)
   {
   printf("x : ");displayBdd(lst->x,0);
   lst = lst->NEXT;
   }
   }
 */
  /* calcul des noeuds equivalents pour chaque fathers */

  for (i = 0; i < count; i++)
    {
      struct eqNode *lst;

      /* minimisation locale */

      if ((fathers[i].x)->low == XTWO)
	{
	  struct assign *ass;
	  struct eqNode *newEqNode;

	  newEqNode = (struct eqNode *) mbkalloc (sizeof (struct eqNode));
	  newEqNode->x = (fathers[i].x)->high;
	  ass = (struct assign *) mbkalloc (sizeof (struct assign));
	  ass->NEXT = NULL;
	  ass->x = fathers[i].x;
	  ass->fils = LBDD_LOW;
	  ass->y = (fathers[i].x)->high;
	  newEqNode->ass = ass;
	  newEqNode->NEXT = fathers[i].lst;
	  fathers[i].lst = newEqNode;
	}
      else
	{
	  struct assign *ass;
	  struct eqNode *newEqNode;

	  newEqNode = (struct eqNode *) mbkalloc (sizeof (struct eqNode));
	  newEqNode->x = (fathers[i].x)->low;
	  ass = (struct assign *) mbkalloc (sizeof (struct assign));
	  ass->NEXT = NULL;
	  ass->x = fathers[i].x;
	  ass->fils = LBDD_HIGH;
	  ass->y = (fathers[i].x)->low;
	  newEqNode->ass = ass;
	  newEqNode->NEXT = fathers[i].lst;
	  fathers[i].lst = newEqNode;
	}
      /* on regarde les freres du noeud */

      lst = brothers[(fathers[i].x)->index];
      while (lst)
	{
	  struct assign *lstMatch = NULL;
	  int match = 1;

	  if (lst->x != fathers[i].x)
	    {
	      lstMatch = (struct assign *) matchBdd (fathers[i].x, lst->x, &match);

	      /* fathers[i] et lst matchent ... */

	      if (match)
		{
		  struct eqNode *newEqNode;

		  /* le pere de X n'a pas ete assigne ? */

		  if (lstMatch->x != fathers[i].x)
		    {
		      freeAssign (lstMatch);
		    }
		  else
		    {
		      newEqNode = (struct eqNode *) mbkalloc (sizeof (struct eqNode));
		      newEqNode->x = lst->x;
		      newEqNode->ass = lstMatch;
		      newEqNode->NEXT = fathers[i].lst;
		      fathers[i].lst = newEqNode;
		    }
		}
	    }
	  lst = lst->NEXT;
	}
    }

  /* affichage */

  for (i = 0; i < count; i++)
    {
      struct eqNode *lst;
      struct assign *ass;

      printf ("============================ fathers[%d] :\n", i);
      displayBdd (fathers[i].x, 1);

      lst = fathers[i].lst;
      while (lst)
	{
	  printf ("Eq. node %p : \n", lst->x);
	  printf ("assoc. list : ");
	  ass = lst->ass;
	  while (ass)
	    {
	      printf ("[%p,%d,%p] ", ass->x, ass->fils, ass->y);
	      ass = ass->NEXT;
	    }
	  printf ("\n");
	  lst = lst->NEXT;
	}
    }

  /* assignation des X */

  tabFathers = createTH (100);
  for (i = 0; i < count; i++)
    {
      addTH (tabFathers, (char *)fathers[i].x, i);
    }

  bestAssign = NULL;
  bestSize = 100000;
  numSol = 0;

  getBestAssignMulti (pX, fathers, count, tabFathers, NULL, numBdd);

  if (!bestAssign)
    printf ("Warning : no assignment found\n");

  for (i = 0; i < numBdd; i++)
    pON[i] = applyAssign (pX[i], bestAssign);

  mbkfree (pX);
  freeAssign (bestAssign);

  destroyTH (tabFathers);

  /* desa. de la structure fathers */

/*
   for (i=0;i<count;i++)
   {
   struct eqNode *lst;

   lst = fathers[i].lst;
   while (lst)
   {
   freeAssign(lst->ass);
   lst = lst->NEXT;
   } 
   freeEqNode(fathers[i].lst);
   }

   mbkfree(fathers);

   for (i=0;i<=indexMax;i++)
   {
   if (brothers[i].lst)
   freeEqNode(brothers[i].lst);
   }

   mbkfree(brothers);
 */

}
