/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



/* $Id: sl_local.h,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $ */

#ifndef SL_LOCAL_H
#define SL_LOCAL_H

#define MAXLEVEL 4
#define FANOUT_MAX 4

#define CSTDEL 500

#define SL_SYS sl_sys
#define SL_VHDL sl_sys->fileVhdl
#define SL_FILE sl_sys->file
#define SL_FILEOUT sl_sys->fileOut
#define SL_FILEPAR sl_sys->filePar
#define SL_PARAM sl_sys->fileParam
#define SL_PAR sl_sys->par
#define SL_BEH sl_sys->beh
#define SL_MODE sl_sys->mode
#define SL_VERSION sl_sys->version
#define SL_OPTION sl_sys->option
#define SL_AUXOF sl_sys->auxOF

#define SL_NAME (sl_sys->beh)->NAME
#define SL_CIRCUIT (sl_sys->beh)->CIRCUI
#define SL_PI (sl_sys->beh)->BERIN

#define SL_OPTIM (sl_sys->par)->mode
#define SL_AUX (sl_sys->par)->aux
#define SL_GUARDEDAUX (sl_sys->par)->intermediate
#define SL_LEVEL (sl_sys->par)->level
#define SL_OPTIMPO (sl_sys->par)->earlyPO
#define SL_DELAYPI (sl_sys->par)->delayPI
#define SL_NUMTRANSN (sl_sys->par)->numTransN
#define SL_NUMTRANSP (sl_sys->par)->numTransP

#define SL_ORDER sl_sys->order
#define SL_ORDERPI sl_sys->orderPI

	/* define correspondant aux variables d'environnement */

#define SL_FILE_LIB sl_sys->file_lib
#define SL_CATA_LIB sl_sys->cata_lib
#define SL_CATA_SAVE sl_sys->cata_save
#define SL_CCCC_LIB sl_sys->c4_lib
#define SL_TRACE sl_sys->trace
#define SL_LATCHES sl_sys->latches
#define SL_MAKEORDER sl_sys->makeOrder
#define SL_IN_LO sl_sys->in_lo
#define SL_OUT_LO sl_sys->out_lo
#define SL_OUT_PH sl_sys->out_ph


/*-------------- systeme pour la synthese logique ---------------

  fileVhdl	: identificateur du fichier contenant l'architecture
  fileParam	: fichier de parametres
  par		: pointeur sur la structure contenant les parametres
  beh		: pointeur sur la befig de l'architecture compilee  
  mode		: mode batch ou non.
  version       : outil et numero de version
  option	: option en batch "o" "s" ou "c"
-----------------------------------------------------------------*/

typedef struct sl_syst
  {
    char *file;
    char *fileOut;
    char *filePar;
    char *fileParam;
    char *fileVhdl;
    param *par;
    befig_list *beh;
    int mode;
    char *version;
    char option;
    int order;
    chain_list *orderPI;

    /* variable d'environnement */

    char *file_lib;
    char *cata_lib;
    char *cata_save;
    char *c4_lib;
    int trace;
    int latches;
    char *in_lo;
    char *out_lo;
    char *out_ph;
    int makeOrder;
    int banniere;
    int auxOF;
  }
sl_system;



/*---- la structure de donnees pour le calcul d'ordonnancement sur ABL ----- 

	sigmaDeltaDepth : somme des ecarts entre le niveau d'apparition de la
			  variable et la profondeur d'un cone sur l'ensemble
			  de la BEFIG.
	DepthMin	: profondeur min de la variable consideree
	numOccMin	: nombre d'occurences de la variable a sa profondeur
			  minimale DepthMin.
        numberOut       : nombre de cones ou la variable apparait.
	vue		: drapeau pour le tri dans la table.
	total		: le coefficient pour le tri final.
---------------------------------------------------------------------------*/

typedef struct sl_Torder
  {
    int numberOcc;
    int sigmaDeltaDepth;
    long numOccMin;
    int depthMin;
    int numberOut;
    int vue;
    char *name;
    long total;
  }
sl_tableOrder;




typedef struct AChain
  {
    struct AChain *NEXT;
    char *name;
    chain_list *ptChain;
  }
Alist;



typedef struct solution
  {
    int lit, maxLit, litRed, profMax, delaisMax, operBin;
    chain_list *order;
  }
tabSolution;

struct couple
  {
    pNode x;
    pNode y;
    chain_list *lst;
    int count;
  };

struct assign
  {
    pNode x;
    int fils;
    pNode y;
    struct assign *NEXT;
  };

struct eqNode
  {
    pNode x;
    struct assign *ass;
    struct eqNode *NEXT;
  };

struct X_fathers
  {
    pNode x;
    int assigned;
    struct eqNode *lst;
  };



extern pTH pTabExpr;
extern sl_system *sl_sys;
extern befig_list *VHB_HEDFIG;


#ifndef __P
# if defined(__STDC__) ||  defined(__GNUC__)
#  define __P(x) x
# else
#  define __P(x) ()
# endif
#endif


/* Prototypes from sl_bdd_order.c */

  extern    chain_list * randomOrder __P((chain_list *ptChain, int amp));
  extern            int  numberBddSup __P((befig_list *beh, int numnode));
  extern    chain_list * makeOrderBddBeh __P((befig_list *beh, int optim, chain_list *optimPO, ptype_list *delayPI, int accuracy, int numnode, int gcOK, int trace, int numberInputUpMax, int mode, int (*func)()));


/* Prototypes from sl_compile.c */

  extern    chain_list * compileVHDL __P((char *file, chain_list *orderPI, int order, param *par, int saveaux, int gcOK, int vhb_reorder, int makeBdd, int trace, char *(*func)(), befig_list *beh));
  extern    chain_list * remegalExpr __P((chain_list *expr));
  extern           void  decompileVHDL __P((befig_list *beh, char *name, int trace));


/* Prototypes from sl_dontcare.c */

  extern            int  depthNodeBdd __P((pNode pt));
  extern            int  costDepthTab __P((pNode *tab, int len));
  extern            int  costTab __P((pNode *tab, int len, int (*func)()));
  extern          pNode  sigmaDFBeh __P((chain_list *support, pNode Fon, befig_list *beh, pTH dejaTraite));
  extern           void  calculDCBeh __P((befig_list *beh, pNode *pON, pNode *pDC));
  extern            int  numberBddBeh __P((befig_list *beh));
  extern           void  traitementDCBeh __P((befig_list *beh, int trace, int (*func)(), int level));


/* Prototypes from sl_facto.c */

  extern           void  coutAblBeh __P((befig_list *beh, int *lit, int *maxLit, int *litRed, int *profMax, int *delaisMax, int *operBin, int *litReg, int mode, ptype_list *delayPI));
  extern          pNode  markAblWBddBeh __P((befig_list *beh, chain_list *expr));
  extern           void  occTHBddBeh __P((chain_list *expr, pTH pTabNode));
  extern    chain_list * createVIBeh __P((befig_list *beh, chain_list *expr, pTH pTabNode, int numberSon));
  extern           void  resetDustExpr __P((chain_list *expr));
  extern    chain_list * reverseExpr __P((chain_list *expr));
  extern    chain_list * arity2Expr __P((chain_list *expr));
  extern           void  gcBddToAblBeh __P((befig_list *beh, int trace));
  extern          pNode  rempTabExpr __P((befig_list *beh, chain_list *expr));
  extern           void  bddToAblBeh __P((befig_list *beh, int optim, int trace));
  extern           void  createAuxBeh __P((befig_list *beh));
  extern            int  coutSol __P((int lit, int delaisMax, int litRed, int profMax, int operBin, int optim));
  extern           void  regenereBddBeh __P((befig_list *beh, chain_list *order));
  extern    chain_list * createAuxInitBeh __P((befig_list *beh));
  extern            int  notSameOrder __P((tabSolution *tabSol, int numberSol, chain_list *order));
  extern           void  localOpt __P((int trace));
  extern           void  globalOpt __P((int trace));
  extern           void  annealingLOGIC __P((int trace));


/* Prototypes from sl_facto_bdd2.c */

  extern           void  gcNodeBeh __P((befig_list *beh, int lstbeh));


/* Prototypes from sl_facto_cliqueZ.c */

  extern          pNode  upVarAllDoubleBdd __P((pNode pt));
  extern          pNode  downVarAllDoubleBdd __P((pNode pt));
  extern            int  suptautology __P((pNode x, pNode y));
  extern            int  simplifONOFF __P((pNode pX, pNode *ppY));
  extern    chain_list * extractClique __P((struct couple* tabCouple, int count));
  extern            int  calculClique __P((struct couple *tabCouple, int count));
  extern           void  setONOFF __P((pNode pZ, pNode *ppONp, pNode *ppOFFp, struct couple *tabCouple, int count, int level));
  extern           void  extractONOFF __P((pNode pZ, struct couple *tabCouple, int *pCount, int level));
  extern          pNode  cliqueBdd __P((pNode pON, pNode pDC));
  extern           void  cliqueMultiBdd __P((pNode *pON, pNode *pDC, int numBdd));


/* Prototypes from sl_facto_jacoX.c */

  extern          pNode  upVarAllOneBdd __P((pNode pt));
  extern          pNode  oneByXBdd __P((pNode pt));
  extern           void  extractX_fathers __P((pNode pX, struct X_fathers *fathers, int *pCount));
  extern           void  computeBrothers __P((pNode pX, struct eqNode **brothers));
  extern struct assign * fusionAssign __P((struct assign *lst1, struct assign *lst2));
  extern struct assign * addAssign __P((struct assign *lst1, pNode x, int fils, pNode y));
  extern struct assign * copyAssign __P((struct assign *lst));
  extern           void  freeEqNode __P((struct eqNode *lst1));
  extern           void  freeAssign __P((struct assign *lst1));
  extern struct assign * matchBdd __P((pNode pX1, pNode pX2, int *pMatch));
  extern          pNode  applyAssign __P((pNode pX, struct assign *finalAssign));
  extern           void  getBestAssign __P((pNode pX, struct X_fathers *fathers, int count, pTH tabFathers, struct assign *finalAssign));
  extern          pNode  assign01Bdd __P((pNode pON));
  extern          pNode  jacoBdd __P((pNode pON, pNode pDC));
  extern           void  getBestAssignMulti __P((pNode *pX, struct X_fathers *fathers, int count, pTH tabFathers, struct assign *finalAssign, int numBdd));
  extern           void  jacoMultiBdd __P((pNode *pON, pNode *pDC, int numBdd));


/* Prototypes from sl_facto_optim.c */

  extern    chain_list * bestExpr __P((chain_list *expr1, chain_list *expr2, int optim));
  extern    chain_list * mux2Abl __P((pNode high, pNode low, chain_list *a, char **tabName, int optim));
  extern    chain_list * bdd2Abl __P((pNode pt, char **tabName, int optim));
  extern            int  numberAtomFact __P((pNode pt));
  extern            int  numberAtomFacto2 __P((pNode high, pNode low, int cpt));
  extern            int  profFact __P((pNode pt, pTH ptHash));
  extern            int  profFact2 __P((pNode high, pNode low, int cpt, pTH ptHash));
  extern            int  numberAtomRedFacto2 __P((pNode high, pNode low, int cpt));
  extern            int  numberAtomRedFact __P((pNode pt));
  extern            int  numberAtomRedBeh __P((befig_list *beh));


/* Prototypes from sl_facto_optimexpr.c */

  extern            int  timeExpr __P((chain_list *expr, pTH timeTH));
  extern            int  coutExpr __P((int litt, int time, int optim));
  extern    chain_list * optimExpr __P((chain_list *expr, ptype_list *support, int optim));
  extern    chain_list * optimExprExt __P((chain_list *expr, pTH caracTH, int optim, pTH dejaTraite, chain_list *order, befig_list *beh));
  extern           void  optimBeh __P((befig_list *beh, int optim, ptype_list *delais, chain_list *optimPO, chain_list *order));


/* Prototypes from sl_order.c */

  extern            int  traitementAux __P((befig_list *beh, char *name, sl_tableOrder *ptOrder, pTH ptHOrder, int depth, int oper, pTH ptHAux));
  extern           void  supportAuxExpr __P((befig_list *beh, chain_list *abl, ptype_list **ppAux, int prof, pTH ptHOrder));
  extern            int  traitementABLInt __P((befig_list *beh, chain_list *abl, sl_tableOrder *ptOrder, pTH ptHOrder, int depth, int oper, ptype_list *pAux, pTH ptHAux));
  extern            int  traitementABL __P((befig_list *beh, chain_list *abl, sl_tableOrder *ptOrder, pTH ptHOrder, int depth, int oper, pTH ptHAux));
  extern           void  traitementCone __P((befig_list *beh, chain_list *expr, sl_tableOrder *ptOrder, pTH ptHOrder, pTH ptHAux));
  extern           void  calculSigma __P((befig_list *beh, sl_tableOrder *ptOrder, pTH ptHOrder, pTH ptHAux));
  extern    chain_list * analyseResul __P((befig_list *beh, sl_tableOrder *ptOrder, pTH ptHOrder, int trace));
  extern    chain_list * makeOrderAbl __P((befig_list *beh, int mode, int trace));


/* Prototypes from sl_prepar_bib.c */

  extern            int  levelExpr __P((chain_list *expr));
  extern           void  calculTPTNExpr __P((chain_list *expr, int *tp, int *tn));
  extern            int  polarityExpr __P((chain_list *expr));
  extern            int  respectTrExpr __P((chain_list *expr, int tp, int tn, int optim));
  extern            int  empileOKExpr __P((chain_list *expr, int tp, int tn, int polarite, int optim));
  extern            int  maxTH __P((pTH pTable));
  extern            int  delaisCarac __P((pTH pTable, int optim));


/* Prototypes from sl_prepar_carac.c */

  extern           void  supportOptimExprInt __P((chain_list *expr, pTH caracTH, int delais, chain_list **ppCL));
  extern    chain_list * supportOptimExpr __P((chain_list *expr, pTH caracTH, int delais));
  extern            int  calculDelaisExpr __P((befig_list *beh, pTH ret, chain_list *expr, int balance));
  extern            pTH  caracBeh __P((befig_list *beh, ptype_list *delais, int balance));


/* Prototypes from sl_prepar_compact.c */

  extern            int  compact_elimSignal __P((int occ, int num));
  extern    chain_list * compact_substTHOccExpr __P((befig_list *beh, pTH pTHOcc, pTH pTHNum, pTH pTHAux, chain_list *expr, pTH dejaTraite));
  extern           void  remplTHSuccExpr __P((pTH pTHSucc, chain_list *expr, char *name));
  extern           void  remplTHSuccBeh __P((befig_list *beh, pTH pTHSucc));
  extern           void  remplTHOccExpr __P((pTH pTHOcc, chain_list *expr));
  extern           void  remplTHOccBeh __P((befig_list *beh, pTH pTHOcc));
  extern           void  compactBeh __P((befig_list *beh, chain_list *saveaux));


/* Prototypes from sl_prepar_prep.c */

  extern          short  memberPType __P((char *name, ptype_list *ptl));
  extern          short  memberChain __P((char *name, chain_list *cl));
  extern            int  elimSignal __P((int occ, int num, int optim));
  extern            int  searchNameBeh __P((befig_list *beh, char *name));
  extern           void  mapCarExprBeh __P((befig_list *beh, chain_list *(*func)()));
  extern           void  mapExprBeh __P((befig_list *beh, void (*func)()));
  extern    chain_list * mergeChain_list __P((chain_list *l1, chain_list *l2));
  extern           void  substTHOccExpr __P((befig_list *beh, pTH pTHOcc, pTH pTHNum, pTH pTHAux, pTH pTHSucc, chain_list *expr, int optim, pTH dejaTraite));
  extern           void  compactExprBeh __P((befig_list *beh, int optim, ptype_list *delais, chain_list *optimPO, int level, chain_list *saveaux));


/* Prototypes from sl_proof_util.c */

  extern            int  mayCoverBdd __P((pNode on1, pNode dc1, pNode on2, pNode dc2));
  extern           void  displayExprLaxInt __P((chain_list *expr));
  extern           void  displayExprLax __P((chain_list *expr));
  extern    chain_list * searchExprOutBeh __P((befig_list *beh, char *name));
  extern           void  polarBDBeh __P((befig_list *beh));
  extern           void  elimInoutBeh __P((befig_list *beh));
  extern            int  elimBusMuxBeh __P((befig_list *beh));
  extern           void  orderProofBeh __P((befig_list *bef1, befig_list *bef2, char varaux));
  extern            int  verifIOBeh __P((befig_list *beh1, befig_list *beh2, int mode));
  extern            int  verifPNode __P((befig_list *beh1, befig_list *beh2, int mode));
  extern          pNode  sumBinode __P((binode_list *binode));
  extern    chain_list * sumBiabl __P((biabl_list *biabl));
  extern          pNode  sumCNDBinode __P((binode_list *binode));
  extern          pNode  busOnToBddBeh __P((befig_list *beh, char *name));
  extern          pNode  buxOnToBddBeh __P((befig_list *beh, char *name));
  extern           void  displayDiff __P((pNode pt1, pNode pt2, befig_list *beh1, befig_list *beh2, int mode));
  extern            int  proofBeh __P((befig_list *beh1, befig_list *beh2, int mode, int exitIO));


/* Prototypes from sl_system.c */

  extern            int  verifArgs __P((sl_system *sl, int argc, char *argv[]));
  extern           void  majSystem __P((sl_system *sl));
  extern           void  addInfoSystem __P((sl_system *sl, int mode, int argc, char *argv[], char *version));
  extern           void  displaySystem __P((sl_system *sys));


/* Prototypes from sl_systemp.c */

  extern         param * initializeParam __P((int optim, int aux, int level, ptype_list *delayPI, chain_list *intermediate, chain_list *optimPO, int numTransN, int numTransP));
  extern     sl_system * initializeSLSystem __P(());
/*  extern    lofig_list * mappingSC __P(());
  extern    lofig_list * mappingC4 __P(());*/


/* Prototypes from sl_util.c */

  extern           void  addArite __P((chain_list *exp));
  extern    chain_list * inversArg __P((chain_list *exp));
  extern            int  nivNot __P((chain_list *exp));
  extern            int  testObj __P((char *name, char type));
  extern         Alist * searchAlist __P((Alist *ptTL, char *name));
  extern         Alist * createAlist __P((Alist *ptTHead, chain_list *expr));
  extern           long  searchPtype __P((ptype_list *pTL, char *name));
  extern            int  existDirectory __P((char *path));
  extern          char * getIdent __P((char *filename));
  extern    chain_list * SLgetFile __P((char *path, char *ext));
  extern            int  logsup2 __P((int val));
  extern          char * gensym __P((char *name, int num));
  extern           void  verifCohBeh __P((befig_list *beh, ptype_list *delaypi, chain_list *optimpo));
  extern           void  removeAuxBeh __P((befig_list *beh, chain_list *saveAux));
  extern           void  flatArityExprBeh __P((befig_list *beh));
  extern            int  existAuxBeh __P((befig_list *beh, char *name));
  extern            int  existInputBeh __P((befig_list *beh, char *name));
  extern            int  existOutputBeh __P((befig_list *beh, char *name));
  extern            int  countInputBeh __P((befig_list *beh));
  extern            int  profExprFact __P((chain_list *expr));
  extern            int  profExprBeh __P((befig_list *beh, chain_list *expr, pTH profTH));
  extern           long  numberAtomExprBeh __P((befig_list *beh, chain_list *expr, pTH numberTH));
  extern    chain_list * delElementChain_list __P((chain_list *cl, void *data));
  extern    chain_list * copyChain_list __P((chain_list *cl));
  extern    chain_list * berinToChain_list __P((befig_list *beh));
  extern    chain_list * createExprCL __P((short oper, chain_list *LC));
  extern           void  afficheLC __P((chain_list *IO));
  extern    chain_list * interLC __P((chain_list *l1, chain_list *l2));
  extern    chain_list * extractLC __P((chain_list *l1, chain_list *l2));
  extern            int  countChain_list __P((chain_list *CL));
  extern            int  inChain_list __P((chain_list *CL, char *data));
  extern            int  inPtype_list __P((ptype_list *CL, char *data));
  extern            int  makeBddBeh __P((struct befig *ptr_befig, pCircuit ptr_circuit, int reorder, int trace, int max_node, int createAux, chain_list **ppAux));
  extern          char * nameIntVect __P((char *name));
  extern          char * nameInt __P((char *name));
  extern           void  renameExpr __P((chain_list *expr, char *(*func)()));
  extern           void  renameBeh __P((befig_list *beh, char *(*func)()));
  extern          pNode  makeBddAbl __P((pCircuit ptr_circuit, chain_list *expr, struct befig *ptr_befig, int *vhb_count, int createAux, chain_list **ppAux));
  extern           void  czo_displayBeh __P((befig_list *befig, char *f, int l));
  extern           void  displayBeh __P((befig_list *befig, short nb));
  extern           void  substPhyExprBeh __P((befig_list *beh, char *name, chain_list *expr));
  extern           void  substExprBeh __P((befig_list *beh, char *name, chain_list *sexpr));
  extern           void  removeDCBeh __P((befig_list *beh));
  extern    chain_list * AND2 __P((chain_list *expr1, chain_list *expr2));
  extern    chain_list * OR2 __P((chain_list *expr1, chain_list *expr2));
  extern           void  videBddBefig __P((befig_list *beh));
  extern    chain_list * exprToCLBeh __P((befig_list *beh));
  extern    befig_list * copyBeh __P((befig_list *beh));
  extern           void  put_CATA_LIB __P((char *cata_lib));



#endif
