/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - boolean optimisation
 * Date    : 1991,92
 * Author  : Luc Burgun
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */



#ident  "$Id: sl_proof_util.c,v 1.1 1997/09/19 15:18:02 sirol Exp sirol $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"

extern pTH pTabExpr;

/*------------------------------------------------------------------
  IDENT : mayCoverBdd
  ------------------------------------------------------------------
  FUNCTION : calcule si deux fonctions phi-Boolennes ont une 
             couverture commune 
  ------------------------------------------------------------------ */

int
mayCoverBdd (on1, dc1, on2, dc2)
     pNode on1, dc1, on2, dc2;
{
  if ((applyBinBdd (AND, applyBinBdd (OR, on1, dc1), on2) == on2) &&
      (applyBinBdd (AND, applyBinBdd (OR, on2, dc2), on1) == on1))
    return 1;
  else
    return 0;
}

/*------------------------------------------------------------------
  IDENT : displayExprLax
  ------------------------------------------------------------------
  FUNCTION : affiche l'expression a la sauce LISP --> (oper list_arg)
  ------------------------------------------------------------------ */

void
displayExprLaxInt (expr)
     chain_list *expr;
{
  int oper;

  if (ATOM (expr))
    printf (" %s", VALUE_ATOM (expr));
  else
    {
      oper = OPER (expr);
      if (oper >= 0 && oper <= 1000)
	{
	  printf (" (%s", operToChar (OPER (expr)));
	  mapExpr (displayExprLaxInt, expr);
	  printf (")");
	}
      else
	{
	  printf ("\ndisplayExprLax : error - unknown operator %d\n", oper);
	  exit (-1);
	}
    }
}

void
displayExprLax (expr)
     chain_list *expr;
{
  displayExprLaxInt (expr);
  printf ("\n");
}


	/* recherche d'une expression a partir d'un nom de sortie */

chain_list *
searchExprOutBeh (beh, name)
     befig_list *beh;
     char *name;
{
  beout_list *out = beh->BEOUT;

  while (out)
    {
      if (!strcmp (out->NAME, name))
	return (out->ABL);
      out = out->NEXT;
    }
  return ((chain_list *) NULL);
}

/*------------------------------------------------------------------------
polarBDBeh : inverse la polarite des BD d'une befig 
--------------------------------------------------------------------------
retour       : rien
--------------------------------------------------------------------------*/
void
polarBDBeh (beh)
     befig_list *beh;
{
  bereg_list *reg;
  biabl_list *biabl;

  /* inversion de polarite des cones DATA des BD */

  reg = beh->BEREG;
  while (reg)
    {
      int BD = 0;

      biabl = reg->BIABL;
      while (biabl)
	{
	  if (biabl->USER)	/* BD detectee */
	    {
	      BD = 1;
	      /* inversion de polarite */

	      biabl->VALABL = notExpr (biabl->VALABL);

	      /* les don't care "'d'" ne doivent pas etre inverses */

	      substPhyExpr (biabl->VALABL, "'d'", notExpr (createAtom ("'d'")));
	    }
	  biabl = biabl->NEXT;
	}
      /* substitution generalisee dans la befig */

      if (BD)
	substExprBeh (beh, reg->NAME, notExpr (createAtom (reg->NAME)));

      reg = reg->NEXT;
    }
}

/*------------------------------------------------------------------------
elimInoutBeh : patch qui permet de voir un inout comme un out
--------------------------------------------------------------------------
retour       : rien
--------------------------------------------------------------------------*/
void
elimInoutBeh (beh)
     befig_list *beh;
{
  bepor_list *port;
  chain_list *expr;

  /* recherche des INOUT qui ne sont pas des bus */

  port = beh->BEPOR;
  while (port)
    {
      if (port->DIRECTION == 'B' && port->TYPE == 'B')
	{
	  printf ("WARNING : %s is considered as an output signal\n", port->NAME);

	  /* on recherche l'expression associee */

	  expr = searchExprOutBeh (beh, port->NAME);
	  if (expr == NULL)
	    {
	      printf ("elimInoutSignal : error signal hasn't expression\n");
	      exit (-1);
	    }
	  beh->BEAUX = beh_addbeaux (beh->BEAUX, port->NAME, expr, NULL);

	  /* elimination de la liste des BERIN */

	  beh->BERIN = beh_rmvberin (beh->BERIN, port->NAME);
	}
      port = port->NEXT;
    }
}

/*------------------------------------------------------------------------
elimBusMuxBeh : transformation des bus mutiplex en mutiplex combinatoire 
                il est necessaire d'avoir deja calculer les BDD
--------------------------------------------------------------------------
retour        : 1 si un bus a ete transforme, 0 sinon
--------------------------------------------------------------------------*/
int
elimBusMuxBeh (beh)
     befig_list *beh;
{
  bebus_list *bus;
  bebux_list *bux;
  chain_list *delBus = NULL;
  int retour = 0;


  bus = beh->BEBUS;
  while (bus)
    {
      if (sumCNDBinode (bus->BINODE) == one)
	{
	  delBus = addchain (delBus, bus);
	  beh->BEOUT = beh_addbeout (beh->BEOUT, bus->NAME,
				     sumBiabl (bus->BIABL), NULL);

	  retour = 1;
	}
      bus = bus->NEXT;
    }

  while (delBus)
    {
      beh->BEBUS = beh_delbebus (beh->BEBUS, (bebus_list *) (delBus->DATA), 'Y');
      delBus = delBus->NEXT;
    }

  delBus = NULL;

  bux = beh->BEBUX;
  while (bux)
    {
      if (sumCNDBinode (bux->BINODE) == one)
	{
	  delBus = addchain (delBus, bux);
	  beh->BEAUX = beh_addbeaux (beh->BEAUX, bux->NAME,
				     sumBiabl (bux->BIABL), NULL);

	  /* pour les bus internes (points d'arret) 
	     on elimine leur occurence dans BERIN */

	  beh->BERIN = beh_rmvberin (beh->BERIN, bux->NAME);
	  retour = 1;
	}
      bux = bux->NEXT;
    }

  while (delBus)
    {
      beh->BEBUX = beh_delbebux (beh->BEBUX, (bebux_list *) (delBus->DATA), 'Y');
      delBus = delBus->NEXT;
    }

  return retour;
}

/*------------------------------------------------------------------------
orderProofBeh      : calcul d'un ordonnancement pour 2 circuits 
--------------------------------------------------------------------------
retour        : 1 si un bus a ete transforme, 0 sinon
--------------------------------------------------------------------------*/
void
orderProofBeh (bef1, bef2, varaux)
     befig_list *bef1, *bef2;
     char varaux;
{
  chain_list *order;
  chain_list *ptChain;
  beaux_list *aux;
  int vhb_reorder = 2;
  chain_list *makeOrderAbl ();

  order = makeOrderAbl (bef1, 1, 1);

  videBddBefig (bef1);
  videBddBefig (bef2);

  /* creation des bdd */

  bef1->CIRCUI = initializeCct (bef1->NAME, 5 * countInputBeh (bef1), 100);
  bef2->CIRCUI = initializeCct (bef2->NAME, 5 * countInputBeh (bef2), 100);


  ptChain = order;
  while (ptChain)
    {
      addInputCct (bef1->CIRCUI, ptChain->DATA);
      ptChain = ptChain->NEXT;
    }

  /* recherche des auxiliaires communs */

  if (varaux == 'A')
    {
      aux = bef1->BEAUX;
      printf ("Looking for the common auxiliary signals :\n");
      while (aux)
	{
	  if (existAuxBeh (bef2, aux->NAME))
	    {
	      addInputCct (bef1->CIRCUI, aux->NAME);
	      printf ("%s ", aux->NAME);
	      order = addchain (order, aux->NAME);
	    }
	  aux = aux->NEXT;
	}
      printf ("\n\n");
    }

  while (makeBddBeh (bef1, bef1->CIRCUI, vhb_reorder, 1, 100, 10000000, NULL) == 1)
    {
      gcNodeBeh (bef1, 0);
      order = (chain_list *) makeOrderBddBeh (bef1, 0, NULL, NULL, 2, 30, 1, 0, 12, -1, NULL);
      vhb_reorder--;
    }

  printf ("---> final number of nodes = %d", (sysBdd.pRT)->compteur);

  gcNodeBeh (bef1, 0);

  printf ("(%d)\n\n", (sysBdd.pRT)->compteur);

  /* calcul des graphes pour le 2eme circuit */

  cpOrderCct (bef1->CIRCUI, bef2->CIRCUI);

  /* ajout des inputs specifiques a bef2 */

  ptChain = berinToChain_list (bef2);
  while (ptChain)
    {
      if (searchInputCct (bef2->CIRCUI, ptChain->DATA) == EMPTYTH)
	{
	  printf ("Warning : Input '%s' only exists in the second description\n",
		  (char *)ptChain->DATA);
	  addInputCct (bef2->CIRCUI, ptChain->DATA);
	}
      ptChain = ptChain->NEXT;
    }

  makeBddBeh (bef2, bef2->CIRCUI, 0, 1, 100, 100000000, NULL);
}


/*-------------------------------------------------------------------------
verifIOBeh 	: verifie la coherence des interfaces des deux befig. 
---------------------------------------------------------------------------
retour		: le nombre d'erreurs trouvees.
---------------------------------------------------------------------------*/
int
verifIOBeh (beh1, beh2, mode)
     befig_list *beh1, *beh2;
     int mode;
{
  berin_list *in1, *in2;
  beout_list *out1, *out2;
  bebus_list *bus1, *bus2;
  bebux_list *bux1, *bux2;
  bereg_list *reg1, *reg2;
  int numberError = 0;


  in1 = beh1->BERIN;
  while (in1)			/* les sorties primaires */
    {
      in2 = beh2->BERIN;
      while (in2 && strcmp (in1->NAME, in2->NAME))
	{
	  in2 = in2->NEXT;
	}
      if (in2 == NULL)
	{
	  if (mode)
	    printf ("Primary input '%s' exists only in '%s'\n", in1->NAME, beh1->NAME);
	  if (mode)
	    printf ("\n");
	  numberError++;
	}
      in1 = in1->NEXT;
    }
  in2 = beh2->BERIN;
  while (in2)
    {
      in1 = beh1->BERIN;
      while (in1 && strcmp (in1->NAME, in2->NAME))
	{
	  in1 = in1->NEXT;
	}
      if (in1 == NULL)
	{
	  if (mode)
	    printf ("Primary input '%s' exists only in '%s'\n", in2->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  numberError++;
	}
      in2 = in2->NEXT;
    }


  out1 = beh1->BEOUT;
  while (out1)			/* les sorties primaires */
    {
      out2 = beh2->BEOUT;
      while (out2 && strcmp (out1->NAME, out2->NAME))
	{
	  out2 = out2->NEXT;
	}
      if (out2 == NULL)
	{
	  if (mode)
	    printf ("Primary output '%s' exists only in '%s'\n", out1->NAME, beh1->NAME);
	  if (mode)
	    printf ("\n");
	  numberError++;
	}
      out1 = out1->NEXT;
    }

  out2 = beh2->BEOUT;
  while (out2)
    {
      out1 = beh1->BEOUT;
      while (out1 && strcmp (out1->NAME, out2->NAME))
	{
	  out1 = out1->NEXT;
	}
      if (out1 == NULL)
	{
	  if (mode)
	    printf ("Primary output '%s' exists only in '%s'\n", out2->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  numberError++;
	}
      out2 = out2->NEXT;
    }


  bus1 = beh1->BEBUS;
  while (bus1)			/* les bus */
    {
      bus2 = beh2->BEBUS;
      while (bus2 && strcmp (bus1->NAME, bus2->NAME))	/* recherche du meme ident. */
	{
	  bus2 = bus2->NEXT;
	}
      if (bus2 == NULL)
	{
	  if (mode)
	    printf ("Bus '%s' exists only in '%s'\n",
		    bus1->NAME, beh1->NAME);
	  numberError++;
	}
      bus1 = bus1->NEXT;
    }

  bus2 = beh2->BEBUS;
  while (bus2)
    {
      bus1 = beh1->BEBUS;
      while (bus1 && strcmp (bus1->NAME, bus2->NAME))	/* recherche du meme ident. */
	{
	  bus1 = bus1->NEXT;
	}
      if (bus1 == NULL)
	{
	  if (mode)
	    printf ("Bus '%s' exists only in '%s'\n", bus2->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  numberError++;
	}
      bus2 = bus2->NEXT;
    }


  bux1 = beh1->BEBUX;
  while (bux1)			/* les bux */
    {
      bux2 = beh2->BEBUX;
      while (bus2 && strcmp (bux1->NAME, bux2->NAME))	/* recherche du meme ident. */
	{
	  bux2 = bux2->NEXT;
	}
      if (bux2 == NULL)
	{
	  if (mode)
	    printf ("Bux '%s' exists only in '%s'\n",
		    bux1->NAME, beh1->NAME);
	  numberError++;
	}
      bux1 = bux1->NEXT;
    }

  bux2 = beh2->BEBUX;
  while (bux2)
    {
      bux1 = beh1->BEBUX;
      while (bux1 && strcmp (bux1->NAME, bux2->NAME))	/* recherche du meme ident. */
	{
	  bux1 = bux1->NEXT;
	}
      if (bux1 == NULL)
	{
	  if (mode)
	    printf ("Bux '%s' exists only in '%s'\n", bux2->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  numberError++;
	}
      bux2 = bux2->NEXT;
    }

  reg1 = beh1->BEREG;
  while (reg1)			/* les registres */
    {
      reg2 = beh2->BEREG;
      while (reg2 && strcmp (reg1->NAME, reg2->NAME))	/* recherche du meme ident. */
	{
	  reg2 = reg2->NEXT;
	}
      if (reg2 == NULL)
	{
	  if (mode)
	    printf ("Register '%s' exists only in '%s'\n",
		    reg1->NAME, beh1->NAME);
	  numberError++;
	}
      reg1 = reg1->NEXT;
    }

  reg2 = beh2->BEREG;
  while (reg2)
    {
      reg1 = beh1->BEREG;
      while (reg1 && strcmp (reg1->NAME, reg2->NAME))	/* recherche du meme ident. */
	{
	  reg1 = reg1->NEXT;
	}
      if (reg1 == NULL)
	{
	  if (mode)
	    printf ("Register '%s' exists only in '%s'\n", reg2->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  numberError++;
	}
      reg2 = reg2->NEXT;
    }

  return (numberError);
}


/*-------------------------------------------------------------------------
verifPNode 	: verifie la coherence le calcul des graphes 
---------------------------------------------------------------------------
retour		: 1 si OK, 0 sinon. 
---------------------------------------------------------------------------*/
int
verifPNode (beh1, beh2, mode)
     befig_list *beh1, *beh2;
     int mode;
{
  beout_list *out;
  beaux_list *aux;
  bebus_list *bus;
  bebux_list *bux;
  bereg_list *reg;
  binode_list *binode;


  aux = beh1->BEAUX;
  while (aux)
    {
      if (aux->NODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : auxiliary signal '%s' have no BDD in '%s'\n",
		    aux->NAME, beh1->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      aux = aux->NEXT;
    }

  aux = beh2->BEAUX;
  while (aux)
    {
      if (aux->NODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : auxiliary signal '%s' have no BDD in '%s'\n",
		    aux->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      aux = aux->NEXT;
    }

  out = beh1->BEOUT;
  while (out)			/* les sorties primaires */
    {
      if (out->NODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : output signal '%s' have no BDD in '%s'\n",
		    out->NAME, beh1->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      out = out->NEXT;
    }

  out = beh2->BEOUT;
  while (out)			/* les sorties primaires */
    {
      if (out->NODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : output signal '%s' have no BDD in '%s'\n",
		    out->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      out = out->NEXT;
    }

  bux = beh1->BEBUX;
  while (bux)			/* les bux */
    {
      binode = bux->BINODE;
      if (binode->CNDNODE == NULL || binode->VALNODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : bux signal '%s' have no BDD in '%s'\n",
		    bux->NAME, beh1->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      bux = bux->NEXT;
    }

  bux = beh2->BEBUX;
  while (bux)			/* les bux */
    {
      binode = bux->BINODE;
      if (binode->CNDNODE == NULL || binode->VALNODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : bux signal '%s' have no BDD in '%s'\n",
		    bux->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      bux = bux->NEXT;
    }

  bus = beh1->BEBUS;
  while (bus)			/* les bus */
    {
      binode = bus->BINODE;
      if (binode->CNDNODE == NULL || binode->VALNODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : bus signal '%s' have no BDD in '%s'\n",
		    bus->NAME, beh1->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      bus = bus->NEXT;
    }

  bus = beh2->BEBUS;
  while (bus)			/* les bus */
    {
      binode = bus->BINODE;
      if (binode->CNDNODE == NULL || binode->VALNODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : bus signal '%s' have no BDD in '%s'\n",
		    bus->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      bus = bus->NEXT;
    }

  reg = beh1->BEREG;
  while (reg)
    {
      binode = reg->BINODE;
      if (binode->CNDNODE == NULL || binode->VALNODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : bus signal '%s' have no BDD in '%s'\n",
		    reg->NAME, beh1->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      reg = reg->NEXT;
    }

  reg = beh2->BEREG;
  while (reg)
    {
      binode = reg->BINODE;
      if (binode->CNDNODE == NULL || binode->VALNODE == NULL)
	{
	  if (mode)
	    printf ("Fatal error : bus signal '%s' have no BDD in '%s'\n",
		    reg->NAME, beh2->NAME);
	  if (mode)
	    printf ("\n");
	  return (0);
	}
      reg = reg->NEXT;
    }

  return (1);
}

/*-------------------------------------------------------------------------
sumBinode 	: calcule la FND (ON-set) d'un bus
---------------------------------------------------------------------------
retour		:un pNode. 
---------------------------------------------------------------------------*/
pNode
sumBinode (binode)
     binode_list *binode;
{
  pNode res = zero;
  while (binode)
    {
      res = applyBinBdd (OR, applyBinBdd (AND, binode->VALNODE, binode->CNDNODE), res);
      binode = binode->NEXT;
    }
  return res;
}

/*-------------------------------------------------------------------------
sumBiabl 	: calcule la FND  en ABL (ON-set) d'un bus
---------------------------------------------------------------------------
retour		:un chain_list *. 
---------------------------------------------------------------------------*/
chain_list *
sumBiabl (biabl)
     biabl_list *biabl;
{
  chain_list *res = createAtom ("'0'");

  while (biabl)
    {
      res = OR2 (AND2 (copyExpr (biabl->VALABL), copyExpr (biabl->CNDABL)), res);
      biabl = biabl->NEXT;
    }
  flatArityExpr (res);
  return res;
}

/*-------------------------------------------------------------------------
sumCNDBinode 	: calcule la somme des commandes d'un bus
---------------------------------------------------------------------------
retour		:un pNode. 
---------------------------------------------------------------------------*/
pNode
sumCNDBinode (binode)
     binode_list *binode;
{
  pNode res = zero;
  while (binode)
    {
      res = applyBinBdd (OR, binode->CNDNODE, res);
      binode = binode->NEXT;
    }
  return res;
}

/*-------------------------------------------------------------------------
busOnToBddBeh 	: calcule le ON-set d'un bus 
---------------------------------------------------------------------------
retour		:un pNode. 
---------------------------------------------------------------------------*/
pNode
busOnToBddBeh (beh, name)
     befig_list *beh;
     char *name;
{
  bebus_list *bus = beh->BEBUS;
  binode_list *binode;
  pNode res;

  while (bus && strcmp (bus->NAME, name))
    bus = bus->NEXT;

  if (bus == NULL)
    {
      printf ("error : busOnToBdd %s doesn't exist as a bus in %s\n",
	      name, beh->NAME);
      exit (-1);
    }

  res = zero;
  binode = bus->BINODE;
  while (binode)
    {
      res = applyBinBdd (OR, applyBinBdd (AND, binode->VALNODE, binode->CNDNODE), res);
      binode = binode->NEXT;
    }
  return (res);
}

/*-------------------------------------------------------------------------
buxOnToBddBeh 	: calcule le ON-set d'un bux 
---------------------------------------------------------------------------
retour		:un pNode. 
---------------------------------------------------------------------------*/
pNode
buxOnToBddBeh (beh, name)
     befig_list *beh;
     char *name;
{
  bebux_list *bux = beh->BEBUX;
  binode_list *binode;
  pNode res;

  while (bux && strcmp (bux->NAME, name))
    bux = bux->NEXT;

  if (bux == NULL)
    {
      printf ("error : buxOnToBdd %s doesn't exist as a bux in %s\n",
	      name, beh->NAME);
      exit (-1);
    }

  res = zero;
  binode = bux->BINODE;
  while (binode)
    {
      res = applyBinBdd (OR, applyBinBdd (AND, binode->VALNODE, binode->CNDNODE), res);
      binode = binode->NEXT;
    }
  return (res);
}


/*-------------------------------------------------------------------------
displayDiff 	: affiche si possible les equations differentes
	          que l'on a trouve dans chaque description. 
---------------------------------------------------------------------------
retour		: rien.
---------------------------------------------------------------------------*/
void
displayDiff (pt1, pt2, beh1, beh2, mode)
     pNode pt1, pt2;
     befig_list *beh1, *beh2;
     int mode;
{
  if (pt1 == NULL || pt2 == NULL)
    {
      printf ("displayDiff : bad args - pNode = NULL\n");
      exit (-1);
    }
  else
    {
      if (mode)
	printf ("         Description %s :\n", beh1->NAME);
      if (numberAtomFact (pt1) > 50)
	{
	  if (mode)
	    printf ("         Logical expression too big...impossible display!\n");
	}
      else
	{
	  if (mode == 1)
	    {
	      printf ("      ");
	      displayExprLax (bdd2Abl (pt1, (beh1->CIRCUI)->pNameI, 0));
	    }
	}
      if (mode)
	printf ("\n");
      if (mode)
	printf ("        Description %s :\n", beh2->NAME);
      if (numberAtomFact (pt2) > 50)
	{
	  if (mode)
	    printf ("         Logical expression too big...impossible display!\n");
	}
      else
	{
	  if (mode == 1)
	    {
	      printf ("      ");
	      displayExprLax (bdd2Abl (pt2, (beh2->CIRCUI)->pNameI, 0));
	    }
	}
      if (mode)
	printf ("\n");
    }
}

/*-------------------------------------------------------------------------
proofBeh 	: Effectue une preuve formelle entre deux comportements. 
		  mode a 1 ==> affichage a l'ecran des erreurs.
 	          Tous les graphes doivent au prealable avoir ete calcule
	          dans les deux descriptions.
		  si exitIO est a 0, on ne sort pas si il existe une
                  difference dans l'interface.
---------------------------------------------------------------------------
retour		: le nombre d'erreurs trouvees.
---------------------------------------------------------------------------*/
int
proofBeh (beh1, beh2, mode, exitIO)
     befig_list *beh1, *beh2;
     int mode, exitIO;
{
  beout_list *out1, *out2;
  beaux_list *aux1, *aux2;
  bebus_list *bus1, *bus2;
  bebux_list *bux1, *bux2;
  bereg_list *reg1, *reg2;
  binode_list *binode1, *binode2;
  biabl_list *biabl1, *biabl2;
  pNode *pDC1, *pON1;
  pNode *pDC2, *pON2;
  int numberError = 0;
  int countDC1;
  int countDC2;
  int numOUTDC2;
  chain_list *lst, *lstExpr;
  int existDC = 0;
  int numberGraph = 0;

  /* existe-t-il des constantes 'd' dans les expressions ? */

  lstExpr = exprToCLBeh (beh1);
  lst = lstExpr;
  while (lst)
    {
      if (searchExpr ((chain_list *) lst->DATA, "'d'"))
	existDC = 1;
      lst = lst->NEXT;
    }
  freechain (lstExpr);

  lstExpr = exprToCLBeh (beh2);
  lst = lstExpr;
  while (lst)
    {
      if (searchExpr ((chain_list *) lst->DATA, "'d'"))
	existDC = 1;
      lst = lst->NEXT;
    }
  freechain (lstExpr);


  if (existDC)
    {
      int numberGraph1, numberGraph2;
      numberGraph1 = numberBddBeh (beh1);
      numberGraph2 = numberBddBeh (beh2);

      printf ("Don't Cares are taken into account...\n\n");

      /* le max des deux */

      if (numberGraph1 > numberGraph2)
	numberGraph = numberGraph1;
      else
	numberGraph = numberGraph2;

      pDC1 = (pNode *) mbkalloc (numberGraph * sizeof (pNode));
      pON1 = (pNode *) mbkalloc (numberGraph * sizeof (pNode));
      calculDCBeh (beh1, pON1, pDC1);

      pDC2 = (pNode *) mbkalloc (numberGraph * sizeof (pNode));
      pON2 = (pNode *) mbkalloc (numberGraph * sizeof (pNode));
      calculDCBeh (beh2, pON2, pDC2);
    }

  countDC1 = 0;
  countDC2 = 0;

  if (mode == 1)
    pTabExpr = createTH (10000);

  if (mode)
    printf ("--------------------------------------------------------------------------------\n");
  if (mode == 1)
    if (mode)
      printf ("             Formal proof with Ordered Binary Decision Diagrams between   \n\n");
  if (mode)
    printf ("             '%s'  and  '%s'\n", beh1->NAME, beh2->NAME);
  if (mode)
    printf ("--------------------------------------------------------------------------------\n");

  numberError = verifIOBeh (beh1, beh2, mode);
  if (numberError > 0 && exitIO)
    {
      if (mode)
	printf ("Fatal error : IO or/and register declaration differs\n");
      if (mode)
	printf ("--------------------------------------------------------------------------------\n");
      return (numberError);
    }
  if (verifPNode (beh1, beh2, mode) == 0)
    {
      if (mode)
	printf ("Warning : some OUT,AUX,REG or BUS have no expression  \n");
    }
  if (mode)
    printf ("==============================  PRIMARY OUTPUT  ===============================\n");
  out1 = beh1->BEOUT;
  while (out1)			/* les sorties primaires */
    {
      countDC2 = 0;
      out2 = beh2->BEOUT;
      while (out2 && strcmp (out1->NAME, out2->NAME))
	{
	  out2 = out2->NEXT;
	  countDC2++;
	}
      if (out2)
	{
	  if (!existDC)
	    {
	      if (out1->NODE != out2->NODE)
		{
		  if (mode)
		    printf ("Error : Behaviour of primary output '%s' differs in the two descriptions\n", out1->NAME);
		  displayDiff (out1->NODE, out2->NODE, beh1, beh2, mode);
		  if (mode)
		    printf ("\n");
		  numberError++;
		}
	    }
	  else
	    /* DC traitement */
	    {
	      if (!mayCoverBdd (out1->NODE, pDC1[countDC1], out2->NODE, pDC2[countDC2]))
		{
		  if (mode)
		    printf ("Error : Behaviour of primary output '%s' differs in the two descriptions\n", out1->NAME);
		  if (mode)
		    printf ("---> On-set :\n");
		  displayDiff (out1->NODE, out2->NODE, beh1, beh2, mode);
		  if (mode)
		    printf ("---> DC-set :\n");
		  displayDiff (pDC1[countDC1], pDC2[countDC2], beh1, beh2, mode);
		  if (mode)
		    printf ("\n");
		  numberError++;
		}
	    }
	}

      countDC1++;

      out1 = out1->NEXT;
    }


  if (mode)
    printf ("=============================  AUXILIARY SIGNAL  ==============================\n");
  aux1 = beh1->BEAUX;
  while (aux1)			/* les auxiliaires */
    {
      aux2 = beh2->BEAUX;
      while (aux2 && strcmp (aux1->NAME, aux2->NAME))
	{
	  aux2 = aux2->NEXT;
	}
      if (aux2)
	{
	  if (aux1->NODE != aux2->NODE)
	    {
	      if (mode)
		printf ("Warning : Behaviour of auxiliary signal '%s' differs in the two descriptions\n",
			aux1->NAME);
	      displayDiff (aux1->NODE, aux2->NODE, beh1, beh2, mode);
	      if (mode)
		printf ("\n");
	    }
	}
      aux1 = aux1->NEXT;
    }

  /* base pour trouver l'index des bdd DC2 dans la table pDC2 */

  numOUTDC2 = countDC1;

  if (mode)
    printf ("==============================  REGISTER SIGNAL  ==============================\n");
  reg1 = beh1->BEREG;
  while (reg1)			/* les registres */
    {
      reg2 = beh2->BEREG;
      countDC2 = numOUTDC2;
      while (reg2 && strcmp (reg1->NAME, reg2->NAME))	/* recherche du meme ident. */
	{
	  reg2 = reg2->NEXT;
	  countDC2++;
	}
      if (reg2)
	{
	  binode1 = reg1->BINODE;
	  biabl1 = reg1->BIABL;
	  while (binode1)	/* preuve sur les commandes et les valeurs */
	    {
	      binode2 = reg2->BINODE;	/* recherche de la meme commande */
	      biabl2 = reg2->BIABL;
	      while (binode2 && (binode1->CNDNODE != binode2->CNDNODE))
		{
		  binode2 = binode2->NEXT;
		  biabl2 = biabl2->NEXT;
		}
	      if (binode2)
		{
		  if (!existDC)
		    {
		      if (binode1->VALNODE != binode2->VALNODE)
			{
			  if (mode)
			    printf ("error : behaviour of register '%s' differs in the two descriptions\n", reg1->NAME);
			  if (mode)
			    printf ("        ---> Value differs for the command =\n");
			  if (mode == 1)
			    displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));
			  if (mode)
			    printf ("\n");
			  displayDiff (binode1->VALNODE, binode2->VALNODE, beh1, beh2, mode);
			  if (mode)
			    printf ("\n");
			  numberError++;
			}
		    }
		  else
		    /* traitement DC */
		    {
		      if (!mayCoverBdd (binode1->VALNODE, pDC1[countDC1], binode2->VALNODE, pDC2[countDC2]))
			{
			  if (mode)
			    printf ("error : behaviour of register '%s' differs in the two descriptions\n", reg1->NAME);
			  if (mode)
			    printf ("        ---> Value differs for the command =\n");
			  if (mode == 1)
			    displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));
			  if (mode)
			    printf ("---> On-set :\n");
			  displayDiff (binode1->VALNODE, binode2->VALNODE, beh1, beh2, mode);
			  if (mode)
			    printf ("---> DC-set :\n");
			  displayDiff (pDC1[countDC1], pDC2[countDC2], beh1, beh2, mode);
			  if (mode)
			    printf ("\n");
			  numberError++;
			}
		    }
		  /* bascule D ou latche */

		  if ((biabl1->USER == 0 && biabl2->USER) ||
		      (biabl2->USER == 0 && biabl1->USER))
		    {
		      if (mode)
			printf ("        ---> type of the register differs for the command =\n");
		      if (mode == 1)
			displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));

		      if (biabl1->USER)
			{
			  if (mode)
			    printf ("        %s is a latch in %s but is an edge triggered in %s\n",
				    reg2->NAME, beh2->NAME, beh1->NAME);
			}
		      else
			{
			  if (mode)
			    printf ("        %s is a latch in %s but is an edge triggered in %s\n",
				    reg1->NAME, beh1->NAME, beh2->NAME);
			}
		      numberError++;
		    }
		}
	      else
		{
		  if (mode)
		    printf ("Error : Register '%s' differs in the two descriptions\n",
			    reg1->NAME);
		  if (mode)
		    printf ("        ---> Driver exist only in %s for the command  =\n",
			    beh1->NAME);
		  if (mode == 1)
		    displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));
		  if (mode)
		    printf ("\n\n");
		  numberError++;
		}
	      binode1 = binode1->NEXT;
	      biabl1 = biabl1->NEXT;
	    }

	  binode2 = reg2->BINODE;
	  while (binode2)
	    {
	      binode1 = reg1->BINODE;
	      while (binode1 && (binode1->CNDNODE != binode2->CNDNODE))
		{
		  binode1 = binode1->NEXT;
		}

	      if (binode1 == NULL)
		{
		  if (mode)
		    printf ("Error : Register '%s' differs in the two descriptions\n",
			    reg2->NAME);
		  if (mode)
		    printf ("        ---> Driver exist only in %s for the command  = \n",
			    beh2->NAME);
		  if (mode == 1)
		    displayExprLax (bdd2Abl (binode2->CNDNODE, (beh2->CIRCUI)->pNameI, 0));
		  if (mode)
		    printf ("\n\n");
		  numberError++;
		}
	      binode2 = binode2->NEXT;
	    }
	}

      countDC1++;

      reg1 = reg1->NEXT;
    }


  if (mode)
    printf ("===============================  EXTERNAL BUS =================================\n");
  bus1 = beh1->BEBUS;
  while (bus1)			/* les bus */
    {
      bus2 = beh2->BEBUS;
      while (bus2 && strcmp (bus1->NAME, bus2->NAME))	/* recherche du meme ident. */
	{
	  bus2 = bus2->NEXT;
	}
      if (bus2)
	{
	  /* difference fonctionnelle */

	  if (busOnToBddBeh (beh1, bus1->NAME) != busOnToBddBeh (beh2, bus2->NAME))
	    {
	      binode1 = bus1->BINODE;
	      while (binode1)	/* preuve sur les commandes et les valeurs */
		{
		  binode2 = bus2->BINODE;	/* recherche de la meme commande */
		  while (binode2 && (binode1->CNDNODE != binode2->CNDNODE))
		    {
		      binode2 = binode2->NEXT;
		    }
		  if (binode2)
		    {
		      if (binode1->VALNODE != binode2->VALNODE)
			{
			  if (mode)
			    printf ("Error : Behaviour of bus '%s' differs in the two descriptions\n",
				    bus1->NAME);
			  if (mode)
			    printf ("        ---> Value differs for the command =\n");
			  if (mode == 1)
			    displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));
			  if (mode)
			    printf ("\n");
			  displayDiff (binode1->VALNODE, binode2->VALNODE, beh1, beh2, mode);
			  if (mode)
			    printf ("\n");
			  numberError++;
			}
		    }
		  else
		    {
		      if (mode)
			printf ("Error : Bus '%s' differs in the two descriptions\n",
				bus1->NAME);
		      if (mode)
			printf ("        ---> Driver exist only in %s for the command  =\n",
				beh1->NAME);
		      if (mode == 1)
			displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));
		      if (mode)
			printf ("\n\n");
		      numberError++;
		    }
		  binode1 = binode1->NEXT;
		}
	      binode2 = bus2->BINODE;
	      while (binode2)
		{
		  binode1 = bus1->BINODE;
		  while (binode1 && (binode1->CNDNODE != binode2->CNDNODE))
		    binode1 = binode1->NEXT;

		  if (binode1 == NULL)
		    {
		      if (mode)
			printf ("Error : Bus '%s' differs in the two descriptions\n",
				bus2->NAME);
		      if (mode)
			printf ("        ---> Driver exist only in %s for the command  = \n",
				beh2->NAME);
		      if (mode == 1)
			displayExprLax (bdd2Abl (binode2->CNDNODE, (beh2->CIRCUI)->pNameI, 0));
		      if (mode)
			printf ("\n\n");
		      numberError++;
		    }
		  binode2 = binode2->NEXT;
		}
	    }
	}

      bus1 = bus1->NEXT;
    }


  if (mode)
    printf ("================================ INTERNAL BUS =================================\n");
  bux1 = beh1->BEBUX;
  while (bux1)			/* les bux */
    {
      bux2 = beh2->BEBUX;
      while (bux2 && strcmp (bux1->NAME, bux2->NAME))	/* recherche du meme ident. */
	{
	  bux2 = bux2->NEXT;
	}
      if (bux2)
	{
	  /* difference fonctionnelle */

	  if (buxOnToBddBeh (beh1, bux1->NAME) != buxOnToBddBeh (beh2, bux2->NAME))
	    {
	      binode1 = bux1->BINODE;
	      while (binode1)	/* preuve sur les commandes et les valeurs */
		{
		  binode2 = bux2->BINODE;	/* recherche de la meme commande */
		  while (binode2 && (binode1->CNDNODE != binode2->CNDNODE))
		    {
		      binode2 = binode2->NEXT;
		    }
		  if (binode2)
		    {
		      if (binode1->VALNODE != binode2->VALNODE)
			{
			  if (mode)
			    printf ("Error : Behaviour of bus '%s' differs in the two descriptions\n",
				    bux1->NAME);
			  if (mode)
			    printf ("        ---> Value differs for the command =\n");
			  if (mode == 1)
			    displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));
			  if (mode)
			    printf ("\n");
			  displayDiff (binode1->VALNODE, binode2->VALNODE, beh1, beh2, mode);
			  if (mode)
			    printf ("\n");
			  numberError++;
			}
		    }
		  else
		    {
		      if (mode)
			printf ("Error : Bus '%s' differs in the two descriptions\n",
				bux1->NAME);
		      if (mode)
			printf ("        ---> Driver exist only in %s for the command  =\n",
				beh1->NAME);
		      if (mode == 1)
			displayExprLax (bdd2Abl (binode1->CNDNODE, (beh1->CIRCUI)->pNameI, 0));
		      if (mode)
			printf ("\n\n");
		      numberError++;
		    }
		  binode1 = binode1->NEXT;
		}
	      binode2 = bux2->BINODE;
	      while (binode2)
		{
		  binode1 = bux1->BINODE;
		  while (binode1 && (binode1->CNDNODE != binode2->CNDNODE))
		    binode1 = binode1->NEXT;

		  if (binode1 == NULL)
		    {
		      if (mode)
			printf ("Error : Bus '%s' differs in the two descriptions\n",
				bux2->NAME);
		      if (mode)
			printf ("        ---> Driver exist only in %s for the command  = \n",
				beh2->NAME);
		      if (mode == 1)
			displayExprLax (bdd2Abl (binode2->CNDNODE, (beh2->CIRCUI)->pNameI, 0));
		      if (mode)
			printf ("\n\n");
		      numberError++;
		    }
		  binode2 = binode2->NEXT;
		}
	    }
	}
      bux1 = bux1->NEXT;
    }

  if (mode)
    printf ("\n                            Formal Proof : ");
  if (numberError > 0)
    {
      if (mode)
	printf ("%d error%s detected\n\n", numberError, (numberError > 1) ? "s" : "");
    }
  else if (mode)
    printf ("OK\n\n");
  if (mode)
    printf ("pppppppppppppppppppppppprrrrrrrrrrrrooooooooooooooooooooooooooooofffffffffffffff\n");
  if (mode)
    printf ("--------------------------------------------------------------------------------\n");
  if (mode == 1)
    destroyTH (pTabExpr);

  return (numberError);
}
