/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------\
|                                                       |
|  Title   :   Structures and fonctions for ABL         |
|                                                       |
|  Date    :            12.06.95                        |
|                                                       |
|  Author  :        Jacomme Ludovic                     |
|                                                       |
\------------------------------------------------------*/

# ifndef ABL_102_H
# define ABL_102_H

/*------------------------------------------------------\
|                                                       |
|                      Constants                        |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                   Logical Operators                   |
|                                                       |
\------------------------------------------------------*/

# define ABL_OR             0
# define ABL_AND            1
# define ABL_XOR            2
# define ABL_NOT            3
# define ABL_NOR            4
# define ABL_NAND           5
# define ABL_NXOR           6
# define ABL_STABLE         7

/*------------------------------------------------------\
|                                                       |
|                       CTL Operators                   |
|                                                       |
\------------------------------------------------------*/

# define ABL_AF             8
# define ABL_AG             9
# define ABL_AX            10
# define ABL_AU            11
# define ABL_EF            12
# define ABL_EG            13
# define ABL_EX            14
# define ABL_EU            15

# define ABL_MAX_OPERATOR  16

/*------------------------------------------------------\
|                                                       |
|                       Polarity                        |
|                                                       |
\------------------------------------------------------*/

# define ABL_POLAR_NEGATIVE 0
# define ABL_POLAR_POSITIVE 1

/*------------------------------------------------------\
|                                                       |
|                        Macros                         |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                         View                          |
|                                                       |
\------------------------------------------------------*/

# define ABL_VIEW_INFIX     0
# define ABL_VIEW_PREFIX    1
# define ABL_VIEW_POSTFIX   2
# define ABL_VIEW_VHDL      3

/*------------------------------------------------------\
|                                                       |
|                       Support Type                    |
|                                                       |
\------------------------------------------------------*/

# define ABL_SUPPORT_CHAIN     0
# define ABL_SUPPORT_PTYPE     1

/*------------------------------------------------------\
|                                                       |
|                 Lisp Like Macros                      |
|                                                       |
\------------------------------------------------------*/

# define ABL_CDR( E )        ( ( E )->NEXT )
# define ABL_CAR( E )        ( (ablexpr *)( E )->DATA )
# define ABL_CADR( E )       ( ABL_CAR( ABL_CDR( ( E ) ) ) )
# define ABL_CDDR( E )       ( ABL_CDR( ABL_CDR( ( E ) ) ) )
# define ABL_ATOM( E )       ( ( E )->NEXT == (ablexpr *)0 )
# define ABL_ATOM_VALUE( E ) ( (char *)( ( E )->DATA ) )
# define ABL_OPER( E )       ( (long)( ( ABL_CAR( ( E ) ) )->DATA ) )

/*------------------------------------------------------\
|                                                       |
|                     Get Macros                        |
|                                                       |
\------------------------------------------------------*/

# define getablatomone()       ( ABL_ATOM_NAME_ONE      )
# define getablatomzero()      ( ABL_ATOM_NAME_ZERO     )
# define getablatomdc()        ( ABL_ATOM_NAME_DC       )
# define getablatomtristate()  ( ABL_ATOM_NAME_TRISTATE )

/*------------------------------------------------------\
|                                                       |
|                         Type                          |
|                                                       |
\------------------------------------------------------*/

  typedef  chain_list    ablexpr;

/*------------------------------------------------------\
|                                                       |
|                     Global Variables                  |
|                                                       |
\------------------------------------------------------*/

 extern char *ABL_OPERATOR_NAME[ ABL_MAX_OPERATOR ];
 extern char *ABL_OPERATOR_UPPER_NAME[ ABL_MAX_OPERATOR ];
 extern char *ABL_ATOM_NAME_ZERO;
 extern char *ABL_ATOM_NAME_ONE;
 extern char *ABL_ATOM_NAME_DC;
 extern char *ABL_ATOM_NAME_TRISTATE;

/*------------------------------------------------------\
|                                                       |
|                        Functions                      |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                Initialize Functions                   |
|                                                       |
\------------------------------------------------------*/

  extern void   ablenv();

/*------------------------------------------------------\
|                                                       |
|                      Free Functions                   |
|                                                       |
\------------------------------------------------------*/

  extern void  freeablexpr();

/*------------------------------------------------------\
|                                                       |
|                    Create Functions                   |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr   *createablatom();
  extern ablexpr   *createablatomone();
  extern ablexpr   *createablatomzero();
  extern ablexpr   *createablatomdc();
  extern ablexpr   *createablatomtristate();

  extern ablexpr   *createabloper();
  extern ablexpr   *createablbinexpr();
  extern ablexpr   *createablnotexpr();
  extern ablexpr   *createablxorbinexpr();
  extern ablexpr   *createablunaryexpr();

/*------------------------------------------------------\
|                                                       |
|                    Optim Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr   *optimablbinexpr();
  extern ablexpr   *optimablnotexpr();
  extern ablexpr   *optimablxorbinexpr();
  extern ablexpr   *optimablunaryexpr();

/*------------------------------------------------------\
|                                                       |
|                      Add Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern void  addablqexpr();
  extern void  addablhexpr();

/*------------------------------------------------------\
|                                                       |
|                    Duplicate Functions                |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *dupablexpr();

/*------------------------------------------------------\
|                                                       |
|                      Del Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern void  delablexpr();
  extern void  delablexprnum();

/*------------------------------------------------------\
|                                                       |
|                      Get Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern int   getabloperpolar();
  extern int   getablopernot();
  extern char *getablopername();
  extern char *getabloperuppername();
  extern long  getabloperbyname();

  extern long  getablexprdepth();
  extern long  getablexprlength();
  extern long  getablexprnumatom();
  extern long  getablexprnumbinoper();
  extern long  getablexprnumocc();

  extern ablexpr *getablexprnum();
  extern ablexpr *getablexprmin();
  extern ablexpr *getablexprmax();

  extern void  *getablexprsupport();

/*------------------------------------------------------\
|                                                       |
|                      Is Functions                     |
|                                                       |
\------------------------------------------------------*/

  extern int   isablunaryoper();
  extern int   isablbinaryoper();
  extern int   isabloperinexpr();
  extern int   isablnameinexpr();
  extern int   isablequalexpr();
  extern int   isablsimilarexpr();

/*------------------------------------------------------\
|                                                       |
|                     Vhdl Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern char *vhdlablname();
  extern char *vhdlablvector();

/*------------------------------------------------------\
|                                                       |
|                      Map Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern void     mapablexpr();
  extern ablexpr *mapabloperexpr();
  extern int      mapablanyexpr();
  extern int      mapableveryexpr();

/*------------------------------------------------------\
|                                                       |
|                   Substitute Functions                |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *substablexpr();
  extern ablexpr *substdupablexpr();

/*------------------------------------------------------\
|                                                       |
|                   Polar Functions                     |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *polarablexpr();
  extern ablexpr *polardupablexpr();

/*------------------------------------------------------\
|                                                       |
|                    Develop Functions                  |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *devdupablxorexpr();
  extern ablexpr *devablxorexpr();

/*------------------------------------------------------\
|                                                       |
|                    Simplify Functions                 |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *simpablexpr();
  extern ablexpr *simpdupablexpr();

/*------------------------------------------------------\
|                                                       |
|                    Simplify CTL Functions             |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *simpablctlexpr();

/*------------------------------------------------------\
|                                                       |
|               Normalisation CTL Functions             |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *normablctlexpr();

/*------------------------------------------------------\
|                                                       |
|                    Flatten Functions                  |
|                                                       |
\------------------------------------------------------*/

  extern void flatablexpr();

/*------------------------------------------------------\
|                                                       |
|                    UnFlatten Functions                |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *unflatablexpr();

/*------------------------------------------------------\
|                                                       |
|                    String Functions                   |
|                                                       |
\------------------------------------------------------*/

  extern ablexpr *strablexpr();

/*------------------------------------------------------\
|                                                       |
|                      View Functions                   |
|                                                       |
\------------------------------------------------------*/

  extern void  viewablexpr();
  extern void  viewablexprfile();
  extern char *viewablexprstr();

# endif
