/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Bdd                           |
|                                                             |
| File    :                  bddexist.c                       |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H

# include <stdio.h>
# include "bddassoc.h"
# include "bddapply.h"
# include "bddexist.h"
# include "bdderror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                Local Exist Node With Association            |
|                                                             |
\------------------------------------------------------------*/

bddnode *loc_existbddnodeon( BddNode )

  bddnode *BddNode;
{
  bddhoper *HashOper;
  bddnode  *BddExist;
  bddnode  *BddHigh;
  bddnode  *BddLow;
  bddvar    NodeVar;

  NodeVar = BddLocalSystem->INDEX_TO_VAR[ BddNode->INDEX ];

  if ( BddLocalAssoc->LAST < NodeVar )
  {
    return( incbddrefext( BddNode ) );
  }

  HashOper = searchbddhoper( BddLocalSystem->HASH_OPER,
                             BDD_OPER_EXIST + BddLocalAssoc->IDENT, 
                             BddNode, BddNode );
 
  if ( HashOper != (bddhoper *)0 )
  {
    return( incbddrefext( HashOper->FATHER ) );
  }

  BddExist = BddLocalAssocNode[ NodeVar ];
  BddHigh  = loc_existbddnodeon( BddNode->HIGH );

  if ( ( BddExist       != (bddnode *)0  ) &&
       ( BddHigh->INDEX == BDD_INDEX_ONE ) )
  {
    BddExist = BddHigh;
  }
  else
  {
    BddLow  = decbddrefext( loc_existbddnodeon( BddNode->LOW ) );
    BddHigh = decbddrefext( BddHigh );

    if ( BddExist != (bddnode *)0 )
    {
      BddExist = loc_applybdd( ABL_OR, BddHigh, BddLow );
    }
    else
    {
      BddExist = addbddnode( (bddsystem *)0, BddNode->INDEX, BddHigh, BddLow );
    }
  }

  addbddhoper( BddLocalSystem->HASH_OPER,
               BDD_OPER_EXIST + BddLocalAssoc->IDENT, 
               BddNode, BddNode, BddExist );

  BddLocalAssoc->CACHE = 1;

  return( BddExist );
}

/*------------------------------------------------------------\
|                                                             |
|                Exist Bdd Node On With Association           |
|                                                             |
\------------------------------------------------------------*/

bddnode *existbddnodeassocon( BddSystem, BddNode, BddAssoc )
 
  bddsystem *BddSystem;
  bddnode   *BddNode;
  bddassoc  *BddAssoc;
{
  bddnode *BddExist;

  if ( checkbddassoc( BddSystem, BddAssoc, 1 ) == 2 )
  {
    return( incbddrefext( BddNode ) );
  }

  BddLocalAssoc     = BddAssoc;
  BddLocalAssocNode = BddAssoc->ASSOC_NODE;

  BddExist = loc_existbddnodeon( BddNode );

  checkbddmaxnode( (bddsystem *)0, 1 );

  return( BddExist );
}

/*------------------------------------------------------------\
|                                                             |
|               Exist Bdd Node Off With Association           |
|                                                             |
\------------------------------------------------------------*/

bddnode *existbddnodeassocoff( BddSystem, BddNode, BddAssoc )
 
  bddsystem *BddSystem;
  bddnode   *BddNode;
  bddassoc  *BddAssoc;
{
  bddnode *BddExist;

  if ( checkbddassoc( BddSystem, BddAssoc, 1 ) == 2 )
  {
    return( incbddrefext( BddNode ) );
  }

  BddLocalAssoc     = BddAssoc;
  BddLocalAssocNode = BddAssoc->ASSOC_NODE;

  BddNode  = decbddrefext( loc_applybddnot( BddNode ) );
  BddExist = decbddrefext( loc_existbddnodeon( BddNode ) );
  BddExist = loc_applybddnot( BddExist );

  checkbddmaxnode( (bddsystem *)0, 1 );

  return( BddExist );
}
