/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Bdd                           |
|                                                             |
| File    :                   bddmark.c                       |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H

# include <stdio.h>
# include "bddmark.h"
# include "bdderror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  static bddmark BddMarkMask       = 0;
  static long    BddMarkNumberNode = 0;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                    Local Mark Bdd Node                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_markbdd( BddNode )

  bddnode *BddNode;
{
  if ( BddNode->INDEX >= BDD_INDEX_MIN )
  {
    if ( ! ( BddNode->MARK & BddMarkMask ) )
    {
      BddNode->MARK |= BddMarkMask;
      BddMarkNumberNode++;

      loc_markbdd( BddNode->LOW  );
      loc_markbdd( BddNode->HIGH );
    }
  }
}

/*------------------------------------------------------------\
|                                                             |
|                        Bdd Unmark Node                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_unmarkbdd( BddNode )

  bddnode *BddNode;
{
  if ( BddNode->INDEX >= BDD_INDEX_MIN )
  {
    if ( BddNode->MARK & BddMarkMask )
    {
      BddNode->MARK &= ~BddMarkMask;
      BddMarkNumberNode++;

      loc_unmarkbdd( BddNode->LOW  );
      loc_unmarkbdd( BddNode->HIGH );
    }
  }
}

/*------------------------------------------------------------\
|                                                             |
|                        Bdd Mark Node                        |
|                                                             |
\------------------------------------------------------------*/

long markbddnode( BddSystem, BddNode, Mark )

  bddsystem *BddSystem;
  bddnode   *BddNode;
  bddmark    Mark;
{
  setbddlocalsystem( BddSystem );

  BddMarkMask       = Mark;
  BddMarkNumberNode = 0;

  loc_markbdd( BddNode );

  return( BddMarkNumberNode );
}

/*------------------------------------------------------------\
|                                                             |
|                        Bdd Mark Node                        |
|                                                             |
\------------------------------------------------------------*/

long unmarkbddnode( BddSystem, BddNode, Mark )

  bddsystem *BddSystem;
  bddnode   *BddNode;
  bddmark    Mark;
{
  setbddlocalsystem( BddSystem );

  BddMarkMask       = Mark;
  BddMarkNumberNode = 0;

  loc_unmarkbdd( BddNode );

  return( BddMarkNumberNode );
}

/*------------------------------------------------------------\
|                                                             |
|                         Get Bdd Node Num                    |
|                                                             |
\------------------------------------------------------------*/

long getbddnodenum( BddSystem, BddNode )

  bddsystem *BddSystem;
  bddnode   *BddNode;
{
  setbddlocalsystem( BddSystem );

  BddMarkMask = BDD_MARK_NUM_MASK;

  BddMarkNumberNode = 0;
  loc_markbdd( BddNode );

  BddMarkNumberNode = 0;
  loc_unmarkbdd( BddNode );

  return( BddMarkNumberNode );
}

/*------------------------------------------------------------\
|                                                             |
|                         Get Node Size                       |
|                                                             |
\------------------------------------------------------------*/

static void loc_getbddsize( BddNode )

  bddnode *BddNode;
{
  if ( BddNode->INDEX >= BDD_INDEX_MIN )
  {
    BddMarkNumberNode++;

    loc_getbddsize( BddNode->LOW  );
    loc_getbddsize( BddNode->HIGH );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                         Get Bdd Node Size                   |
|                                                             |
\------------------------------------------------------------*/

long getbddnodesize( BddSystem, BddNode )

  bddsystem *BddSystem;
  bddnode   *BddNode;
{
  setbddlocalsystem( BddSystem );

  BddMarkNumberNode = 0;
  loc_getbddsize( BddNode );

  return( BddMarkNumberNode );
}
