/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Bdd                           |
|                                                             |
| File    :                   bddorder.c                      |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H

# include <stdio.h>
# include "bddvartree.h"
# include "bddreorder.h"
# include "bdderror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  static char *BddReorderLevel     = (char *)0;
  static long  BddReorderLevelSize = 0;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|              Reorder Bdd Variable Tree Window 2             |
|                                                             |
\------------------------------------------------------------*/

int reorderbddvartreewindow2( BddSystem, BddVarTree, Child )

  bddsystem  *BddSystem;
  bddvartree *BddVarTree;
  bddvar      Child;
{
  long NumberNode;

  setbddlocalsystem( BddSystem );

  NumberNode = BddLocalSystem->NUMBER_NODE;

  swapbddvartree( (bddsystem *)0, BddVarTree, Child );

  if ( BddLocalSystem->NUMBER_NODE < NumberNode )
  {
    return( 1 );
  }

  swapbddvartree( (bddsystem *)0, BddVarTree, Child );

  return( 0 );
}

/*------------------------------------------------------------\
|                                                             |
|              Reorder Bdd Variable Tree Window 3             |
|                                                             |
\------------------------------------------------------------*/

int reorderbddvartreewindow3( BddSystem, BddVarTree, Child )

  bddsystem  *BddSystem;
  bddvartree *BddVarTree;
  bddvar      Child;
{
  long NumberNode;
  char BestOrder;

  setbddlocalsystem( BddSystem );

  BestOrder  = BDD_REORDER_123;
  NumberNode = BddLocalSystem->NUMBER_NODE;

  swapbddvartree( (bddsystem *)0, BddVarTree, Child );

  if ( BddLocalSystem->NUMBER_NODE < NumberNode )
  {
    NumberNode = BddLocalSystem->NUMBER_NODE;
    BestOrder  = BDD_REORDER_213;
  }

  swapbddvartree( (bddsystem *)0, BddVarTree, Child + 1 );

  if ( BddLocalSystem->NUMBER_NODE < NumberNode )
  {
    NumberNode = BddLocalSystem->NUMBER_NODE;
    BestOrder  = BDD_REORDER_231;
  }

  swapbddvartree( (bddsystem *)0, BddVarTree, Child );

  if ( BddLocalSystem->NUMBER_NODE < NumberNode )
  {
    NumberNode = BddLocalSystem->NUMBER_NODE;
    BestOrder  = BDD_REORDER_321;
  }

  swapbddvartree( (bddsystem *)0, BddVarTree, Child + 1 );

  if ( BddLocalSystem->NUMBER_NODE < NumberNode )
  {
    NumberNode = BddLocalSystem->NUMBER_NODE;
    BestOrder  = BDD_REORDER_312;
  }

  swapbddvartree( (bddsystem *)0, BddVarTree, Child );

  if ( BddLocalSystem->NUMBER_NODE < NumberNode )
  {
    NumberNode = BddLocalSystem->NUMBER_NODE;
    BestOrder  = BDD_REORDER_132;
  }

  switch ( BestOrder )
  {
    case BDD_REORDER_123 : swapbddvartree( (bddsystem *)0, BddVarTree, Child + 1 );
    break;

    case BDD_REORDER_213 : swapbddvartree( (bddsystem *)0, BddVarTree, Child + 1 );
                           swapbddvartree( (bddsystem *)0, BddVarTree, Child     );
    break;

    case BDD_REORDER_231 : swapbddvartree( (bddsystem *)0, BddVarTree, Child + 1 );
                           swapbddvartree( (bddsystem *)0, BddVarTree, Child     );
                           swapbddvartree( (bddsystem *)0, BddVarTree, Child + 1 );
    break;

    case BDD_REORDER_321 : swapbddvartree( (bddsystem *)0, BddVarTree, Child     );
                           swapbddvartree( (bddsystem *)0, BddVarTree, Child + 1 );
    break;

    case BDD_REORDER_312 : swapbddvartree( (bddsystem *)0, BddVarTree, Child     );
    break;
  }

  return( BestOrder != BDD_REORDER_123 );
}

/*------------------------------------------------------------\
|                                                             |
|                 Local Reorder Variable Tree Simple          |
|                                                             |
\------------------------------------------------------------*/

static void loc_reorderbddsimple( BddVarTree )

  bddvartree *BddVarTree;
{
  bddvar NumberChild;
  bddvar Child;

  NumberChild = BddVarTree->NUMBER_CHILD;

  if ( ( NumberChild != 0    ) &&
       ( BddVarTree->REORDER ) )
  {
    for ( Child = 0; Child < ( NumberChild - 1 ); Child++ )
    {
      if ( Child < ( NumberChild - 2 ) )
      {
        reorderbddvartreewindow3( (bddsystem *)0, BddVarTree, Child );
      }
      else
      {
        reorderbddvartreewindow2( (bddsystem *)0, BddVarTree, Child );
      }
    } 
  }

  for ( Child = 0; Child < NumberChild; Child++ )
  {
    loc_reorderbddsimple( BddVarTree->CHILD[ Child ] );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                    Reorder Bdd System Simple                |
|                                                             |
\------------------------------------------------------------*/

void reorderbddsystemsimple( BddSystem )

  bddsystem *BddSystem;
{
  setbddlocalsystem( BddSystem );

  garbagebddsystem( (bddsystem *)0 );

  loc_reorderbddsimple( BddLocalSystem->VAR_TREE );

  garbagebddsystem( (bddsystem *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                    Reorder Variable Tree Window             |
|                                                             |
\------------------------------------------------------------*/

static void loc_reorderbddwindow( BddVarTree )

  bddvartree *BddVarTree;
{
  bddvar NumberChild;
  bddvar Child;
  char   Moved;
  char   AnySwap;

  NumberChild = BddVarTree->NUMBER_CHILD;

  if ( ( NumberChild != 0    ) && 
       ( BddVarTree->REORDER ) )
  {
    for ( Child = 0; Child < ( NumberChild - 1 ); Child++ )
    {
      BddReorderLevel[ Child ] = 1;
    }

    do
    {
      AnySwap = 0;

      for ( Child = 0; Child < ( NumberChild - 1 ); Child++ )
      {
        if ( BddReorderLevel[ Child ] )
        {
          if ( Child < NumberChild - 2 )
          {
            Moved = reorderbddvartreewindow3( (bddsystem *)0, BddVarTree, Child );
          }
          else
          {
            Moved = reorderbddvartreewindow2( (bddsystem *)0, BddVarTree, Child );
          }
  
          if ( Moved )
          {
            AnySwap = 1;

            if ( Child > 0 ) BddReorderLevel[ Child - 1 ] = 1;
            if ( Child > 1 ) BddReorderLevel[ Child - 2 ] = 1;

            BddReorderLevel[ Child + 1 ] = 1;

            if ( Child < ( NumberChild - 2 ) ) BddReorderLevel[ Child + 2 ] = 1;
            if ( Child < ( NumberChild - 3 ) ) BddReorderLevel[ Child + 3 ] = 1;
            if ( Child < ( NumberChild - 4 ) ) BddReorderLevel[ Child + 4 ] = 1;
          }
          else
          {
            BddReorderLevel[ Child ] = 0;
          }
        }
      }
    }
    while ( AnySwap );
  }

  for ( Child = 0; Child < NumberChild; Child++ )
  {
    loc_reorderbddwindow( BddVarTree->CHILD[ Child ] );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   Reorder Bdd System Window                 |
|                                                             |
\------------------------------------------------------------*/

void reorderbddsystemwindow( BddSystem )

  bddsystem *BddSystem;
{
  setbddlocalsystem( BddSystem );

  if ( BddReorderLevelSize <= BddLocalSystem->NUMBER_VAR )
  {
    BddReorderLevelSize = BddLocalSystem->NUMBER_VAR + 1;
    BddReorderLevel = autresizeblock( BddReorderLevel, 0, BddReorderLevelSize );
  }

  garbagebddsystem( (bddsystem *)0 );

  loc_reorderbddwindow( BddLocalSystem->VAR_TREE );

  garbagebddsystem( (bddsystem *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                     Reorder System Top                      |
|                                                             |
\------------------------------------------------------------*/

static void loc_reorderbddtop( BddVarTree )

  bddvartree *BddVarTree;
{
  long    NumberNode;
  bddvar  NumberChild;
  bddvar  BestChild;
  bddvar  ChildCount;
  bddvar  ChildTo;
  bddvar  ChildFrom;
  char    SwapReorder;

  NumberChild = BddVarTree->NUMBER_CHILD;

  if ( ( NumberChild != 0    ) &&
       ( BddVarTree->REORDER ) )
  {
    for ( ChildFrom = 0; ChildFrom < ( NumberChild - 1 ); ChildFrom++ )
    {
      BddReorderLevel[ ChildFrom ] = 1;
    }

    NumberNode = BddLocalSystem->NUMBER_NODE;
    ChildCount = ( NumberChild - 1 );
    ChildFrom  = 0;

    while ( ChildCount > 0 )
    {
# ifdef BDD_DEBUG
  fprintf( stdout, "-> ChildFrom %d\n", ChildFrom );
# endif

      BddReorderLevel[ ChildFrom ] = 0;

      BestChild  = ChildFrom;

      for ( ChildTo = ChildFrom; ChildTo < ( NumberChild - 1 ); ChildTo++ )
      {
# ifdef BDD_DEBUG
  fprintf( stdout, "-> Swap %d and %d\n", ChildTo, ChildTo + 1 );
# endif

        swapbddvartree( (bddsystem *)0, BddVarTree, ChildTo );

        SwapReorder                    = BddReorderLevel[ ChildTo     ];
        BddReorderLevel[ ChildTo     ] = BddReorderLevel[ ChildTo + 1 ];
        BddReorderLevel[ ChildTo + 1 ] = SwapReorder;

        if ( BddLocalSystem->NUMBER_NODE < NumberNode )
        {
          BestChild = ChildTo + 1; NumberNode = BddLocalSystem->NUMBER_NODE;
        }
      }

# ifdef BDD_DEBUG
  fprintf( stdout, "--> Best one %d\n", BestChild );
  fprintf( stdout, "--> Number nodes %d\n", NumberNode );
# endif

      for ( ChildTo = ( NumberChild - 1 ); ChildTo > BestChild; ChildTo-- )
      {
# ifdef BDD_DEBUG
  fprintf( stdout, "-> UnSwap %d and %d\n", ChildTo - 1, ChildTo );
# endif

        SwapReorder                    = BddReorderLevel[ ChildTo - 1 ];
        BddReorderLevel[ ChildTo - 1 ] = BddReorderLevel[ ChildTo     ];
        BddReorderLevel[ ChildTo     ] = SwapReorder;

        swapbddvartree( (bddsystem *)0, BddVarTree, ChildTo - 1 );
      }

      while ( ! BddReorderLevel[ ChildFrom ] )
      {
        ChildFrom = ChildFrom + 1;
      }

      ChildCount = ChildCount - 1;
    }
  }

  for ( ChildFrom = 0; ChildFrom < NumberChild; ChildFrom++ )
  {
    loc_reorderbddtop( BddVarTree->CHILD[ ChildFrom ] );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                   Reorder Bdd System Top                    |
|                                                             |
\------------------------------------------------------------*/

void reorderbddsystemtop( BddSystem )

  bddsystem *BddSystem;
{
  setbddlocalsystem( BddSystem );

  if ( BddReorderLevelSize <= BddLocalSystem->NUMBER_VAR )
  {
    BddReorderLevelSize = BddLocalSystem->NUMBER_VAR + 1;
    BddReorderLevel = autresizeblock( BddReorderLevel, 0, BddReorderLevelSize );
  }

  garbagebddsystem( (bddsystem *)0 );

  loc_reorderbddtop( BddLocalSystem->VAR_TREE );

  garbagebddsystem( (bddsystem *)0 );
}

/*------------------------------------------------------------\
|                                                             |
|                   Reorder Bdd System Dynamic                |
|                                                             |
\------------------------------------------------------------*/

void reorderbddsystemdynamic( BddSystem, ReorderFunc, ReorderLow, ReorderRatio )

  bddsystem *BddSystem;
  void     (*ReorderFunc)();
  long       ReorderLow;
  long       ReorderRatio;
{
  setbddlocalsystem( BddSystem );

  if ( ReorderFunc != (void *)0 )
  {
    BddLocalSystem->REORDER = 1;
  }
  else
  {
    BddLocalSystem->REORDER = 0;
  }

  BddLocalSystem->REORDER_FUNC  = ReorderFunc;
  BddLocalSystem->REORDER_LOW   = ReorderLow;
  BddLocalSystem->REORDER_LIMIT = ReorderLow;
  BddLocalSystem->REORDER_RATIO = ReorderRatio + 100;
}
