/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Bdd                           |
|                                                             |
| File    :                  bddsupport.c                     |
|                                                             |
| Date    :                   03.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H

# include <stdio.h>
# include "bddsupport.h"
# include "bdderror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  static chain_list *BddGetHeadChain  = (chain_list *)0;
  static ptype_list *BddGetHeadPType  = (ptype_list *)0;
  static bddvar      BddLocalVariable = (bddvar)0;

/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                  Local Unmark Bdd Node Support              |
|                                                             |
\------------------------------------------------------------*/

static void loc_unmarkbddsupport( BddNode )

  bddnode *BddNode;
{
  bddnode *BddVarNode;

  if ( ( BddNode->INDEX >= BDD_INDEX_MIN       ) &&
       ( BddNode->MARK & BDD_MARK_SUPPORT_MASK ) )
  {
    BddNode->MARK &= ~BDD_MARK_SUPPORT_MASK;

    BddVarNode = BddLocalSystem->VAR_NODE[ BddNode->INDEX ];

    BddVarNode->MARK &= ~BDD_MARK_SUPPORT_MASK;

    loc_unmarkbddsupport( BddNode->HIGH );
    loc_unmarkbddsupport( BddNode->LOW  );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                  Local Get Bdd Node Support PType           |
|                                                             |
\------------------------------------------------------------*/

static void loc_getbddsupportptype( BddNode )

  bddnode *BddNode;
{
  bddnode *BddVarNode;

  if ( ( BddNode->INDEX >= BDD_INDEX_MIN             ) &&
       ( ! ( BddNode->MARK & BDD_MARK_SUPPORT_MASK ) ) )
  {
    BddVarNode = BddLocalSystem->VAR_NODE[ BddNode->INDEX ];

    if ( ! ( BddVarNode->MARK & BDD_MARK_SUPPORT_MASK ) )
    {
      BddVarNode->MARK |= BDD_MARK_SUPPORT_MASK;

      BddGetHeadPType = addptype( BddGetHeadPType, 0, (void *)BddVarNode );
    }

    BddNode->MARK |= BDD_MARK_SUPPORT_MASK;

    loc_getbddsupportptype( BddNode->HIGH );
    loc_getbddsupportptype( BddNode->LOW  );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                  Local Get Bdd Node Support Chain           |
|                                                             |
\------------------------------------------------------------*/

static void loc_getbddsupportchain( BddNode )

  bddnode *BddNode;
{
  bddnode *BddVarNode;

  if ( ( BddNode->INDEX >= BDD_INDEX_MIN             ) &&
       ( ! ( BddNode->MARK & BDD_MARK_SUPPORT_MASK ) ) )
  {
    BddVarNode = BddLocalSystem->VAR_NODE[ BddNode->INDEX ];

    if ( ! ( BddVarNode->MARK & BDD_MARK_SUPPORT_MASK ) )
    {
      BddVarNode->MARK |= BDD_MARK_SUPPORT_MASK;

      BddGetHeadChain = addchain( BddGetHeadChain, (void *)BddVarNode );
    }

    BddNode->MARK |= BDD_MARK_SUPPORT_MASK;

    loc_getbddsupportchain( BddNode->HIGH );
    loc_getbddsupportchain( BddNode->LOW  );
  }
}

/*------------------------------------------------------------\
|                                                             |
|                       Bdd Get Node Support                  |
|                                                             |
\------------------------------------------------------------*/

void *getbddnodesupport( BddSystem, BddNode, Mode )

  bddsystem *BddSystem;
  bddnode   *BddNode;
  int        Mode;
{
  void *Support;

  setbddlocalsystem( BddSystem );

  if ( Mode == BDD_SUPPORT_CHAIN )
  {
    BddGetHeadChain = (chain_list *)0;
    loc_getbddsupportchain( BddNode );
    Support = (void *)BddGetHeadChain;
  }
  else
  {
    BddGetHeadPType = (ptype_list *)0;
    loc_getbddsupportptype( BddNode );
    Support = (void *)BddGetHeadPType;
  }

  loc_unmarkbddsupport( BddNode );

  return( Support );
}

/*------------------------------------------------------------\
|                                                             |
|                Local Mark Bdd Variable In Support           |
|                                                             |
\------------------------------------------------------------*/

static int loc_markbddvarsupport( BddNode )

  bddnode *BddNode;
{
  bddvar Variable;

  Variable = BddLocalSystem->INDEX_TO_VAR[ BddNode->INDEX ];

  if ( ( BddLocalVariable < Variable           ) ||
       ( BddNode->MARK & BDD_MARK_SUPPORT_MASK ) )
  {
    return( 0 );
  }

  if ( BddLocalVariable == Variable )
  {
    return( 1 );
  }

  BddNode->MARK |= BDD_MARK_SUPPORT_MASK;

  if ( loc_markbddvarsupport( BddNode->HIGH ) )
  {
    return( 1 );
  }

  return( loc_markbddvarsupport( BddNode->LOW ) );
}

/*------------------------------------------------------------\
|                                                             |
|              Local Unmark Bdd Variable In Support           |
|                                                             |
\------------------------------------------------------------*/

static int loc_unmarkbddvarsupport( BddNode )

  bddnode *BddNode;
{
  bddvar Variable;

  Variable = BddLocalSystem->INDEX_TO_VAR[ BddNode->INDEX ];

  if ( ( BddLocalVariable < Variable                 ) ||
       ( ! ( BddNode->MARK & BDD_MARK_SUPPORT_MASK ) ) )
  {
    return( 0 );
  }

  if ( BddLocalVariable == Variable )
  {
    return( 1 );
  }

  BddNode->MARK &= ~BDD_MARK_SUPPORT_MASK;

  if ( loc_unmarkbddvarsupport( BddNode->HIGH ) )
  {
    return( 1 );
  }

  return( loc_unmarkbddvarsupport( BddNode->LOW ) );
}


/*------------------------------------------------------------\
|                                                             |
|                    Is Bdd Variable In Support               |
|                                                             |
\------------------------------------------------------------*/

int isbddvarinsupport( BddSystem, BddNode, Variable )

  bddsystem *BddSystem;
  bddnode   *BddNode;
  bddvar     Variable;
{
  int Depend;

  checkbddvar( BddSystem, Variable, 1 );

  BddLocalVariable = Variable;

  Depend = loc_markbddvarsupport( BddNode );
  loc_unmarkbddvarsupport( BddNode );

  return( Depend );
}
