/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------\
|                                                       |
|  Title   :   Structures and fonctions for BTR         |
|                                                       |
|  Date    :            14.02.96                        |
|                                                       |
|  Author  :        Jacomme Ludovic                     |
|                                                       |
\------------------------------------------------------*/

# ifndef BTR_102_H
# define BTR_102_H

/*------------------------------------------------------\
|                                                       |
|                      Constants                        |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                        Macros                         |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                         Type                          |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                 Bdd Transition Function               |
|                                                       |
\------------------------------------------------------*/

   typedef long btrvarorder;

   typedef struct btrvarfunc
   {
     bddnode *VAR;
     bddnode *FUNC;

   } btrvarfunc;

   typedef struct btrtransfunc
   {
     btrvarfunc  *VAR_FUNC;
     btrvarorder *VAR_ORDER;
     bddvar       MAX_VAR;
     bddvar       NUMBER_VAR;
     bddsystem   *BDD_SYSTEM;

   } btrtransfunc;

/*------------------------------------------------------\
|                                                       |
|                 Bdd Transition Relation               |
|                                                       |
\------------------------------------------------------*/

   typedef struct btrvarrel
   {
     bddnode *VAR;
     bddnode *PRIME;
     bddnode *REL;
     
   } btrvarrel;

   typedef struct btrtransrel
   {
     btrvarrel *VAR_REL;
     bddassoc  *PRIME_ASSOC;
     bddassoc  *VAR_ASSOC;
     bddvar     MAX_VAR;
     bddvar     NUMBER_VAR;
     bddsystem *BDD_SYSTEM;

   } btrtransrel;

/*------------------------------------------------------\
|                                                       |
|                        Functions                      |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                      Env  Functions                   |
|                                                       |
\------------------------------------------------------*/

   extern void  btrenv();

/*------------------------------------------------------\
|                                                       |
|                  Allocation Functions                 |
|                                                       |
\------------------------------------------------------*/

   extern btrtransfunc *allocbtrtransfunc();
   extern btrvarfunc   *allocbtrvarfunc();
   extern btrvarorder  *allocbtrvarorder();

   extern btrtransrel  *allocbtrtransrel();
   extern btrvarrel    *allocbtrvarrel();

/*------------------------------------------------------\
|                                                       |
|                      Free Functions                   |
|                                                       |
\------------------------------------------------------*/

   extern void freebtrtransfunc();
   extern void freebtrvarfunc();
   extern void freebtrvarorder();

   extern void freebtrtransrel();
   extern void freebtrvarrel();

/*------------------------------------------------------\
|                                                       |
|                    Resize Functions                   |
|                                                       |
\------------------------------------------------------*/

   extern btrvarfunc  *resizebtrvarfunc();
   extern btrvarorder *resizebtrvarorder();

   extern btrvarrel   *resizebtrvarrel();

/*------------------------------------------------------\
|                                                       |
|                Transition Function Functions          |
|                                                       |
\------------------------------------------------------*/

  btrtransfunc *createbtrtransfunc();
  void          resetbtrtransfunc();
  void          destroybtrtransfunc();
  void          viewbtrtransfunc();
  void          addbtrtransfunc();

/*------------------------------------------------------\
|                                                       |
|          Image and PreImage Transition Function       |
|                                                       |
\------------------------------------------------------*/

  bddnode   *imagebtrtransfunc();
  bddnode   *preimagebtrtransfunc();

/*------------------------------------------------------\
|                                                       |
|                Reorder Transition Function            |
|                                                       |
\------------------------------------------------------*/

  void       reorderbtrtransfunc();

/*------------------------------------------------------\
|                                                       |
|                Transition Relation Functions          |
|                                                       |
\------------------------------------------------------*/

  btrtransrel *createbtrtransrel();
  void         resetbtrtransrel();
  void         destroybtrtransrel();
  void         viewbtrtransrel();

/*------------------------------------------------------\
|                                                       |
|          Image and PreImage Transition Relation       |
|                                                       |
\------------------------------------------------------*/

  bddnode   *imagebtrtransrel();
  bddnode   *preimagebtrtransrel();

# endif 
