/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------\
|                                                       |
|  Title   :   Structures and fonctions for FSM         |
|                                                       |
|  Date    :            04.12.96                        |
|                                                       |
|  Author  :        Jacomme Ludovic                     |
|                                                       |
\------------------------------------------------------*/

# ifndef FSM_102_H
# define FSM_102_H

/*------------------------------------------------------\
|                                                       |
|                      Constants                        |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                    Fsm Port Direction                 |
|                                                       |
\------------------------------------------------------*/

# define FSM_DIR_IN     'I'
# define FSM_DIR_OUT    'O'
# define FSM_DIR_INOUT  'B'

/*------------------------------------------------------\
|                                                       |
|                      Fsm Port Type                    |
|                                                       |
\------------------------------------------------------*/

# define FSM_TYPE_BIT   'B'
# define FSM_TYPE_WORD  'W'
# define FSM_TYPE_MUX   'M'

/*------------------------------------------------------\
|                                                       |
|                    Fsm Stack Control                  |
|                                                       |
\------------------------------------------------------*/

# define FSM_CTRL_NOP   0
# define FSM_CTRL_PUSH  1
# define FSM_CTRL_POP   2
# define FSM_MAX_CTRL   3

/*------------------------------------------------------\
|                                                       |
|                      Fsm Flag                         |
|                                                       |
\------------------------------------------------------*/

# define FSM_STAR_MASK      (long)( 0x0001 )
# define FSM_FIRST_MASK     (long)( 0x0002 )

/*------------------------------------------------------\
|                                                       |
|                        Macros                         |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                  State Type Macros                    |
|                                                       |
\------------------------------------------------------*/

# define IsFsmFirstState( T )    ( (T)->FLAGS &   FSM_FIRST_MASK )
# define SetFsmFirstState( T )   ( (T)->FLAGS |=  FSM_FIRST_MASK )
# define CleanFsmFirstState( T ) ( (T)->FLAGS &= ~FSM_FIRST_MASK )

# define IsFsmStarState( T )     ( (T)->FLAGS &   FSM_STAR_MASK )
# define SetFsmStarState( T )    ( (T)->FLAGS |=  FSM_STAR_MASK )
# define CleanFsmStarState( T )  ( (T)->FLAGS &= ~FSM_STAR_MASK )

/*------------------------------------------------------\
|                                                       |
|                Transition Type Macros                 |
|                                                       |
\------------------------------------------------------*/

# define IsFsmStarTrans( T )    ( (T)->FLAGS &   FSM_STAR_MASK )
# define SetFsmStarTrans( T )   ( (T)->FLAGS |=  FSM_STAR_MASK )
# define CleanFsmStarTrans( T ) ( (T)->FLAGS &= ~FSM_STAR_MASK )

/*------------------------------------------------------\
|                                                       |
|                     Stack Type Macros                 |
|                                                       |
\------------------------------------------------------*/

# define IsFsmStarStack( S )    ( (S)->FLAGS &   FSM_STAR_MASK )
# define SetFsmStarStack( S )   ( (S)->FLAGS |=  FSM_STAR_MASK )
# define CleanFsmStarStack( S ) ( (S)->FLAGS &= ~FSM_STAR_MASK )

/*------------------------------------------------------\
|                                                       |
|                      Structures                       |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                     Fsm Input List                    |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmin_list
  {
    struct fsmin_list *NEXT;
    char              *NAME;
    long               FLAGS;
    void              *USER;

  } fsmin_list;

/*------------------------------------------------------\
|                                                       |
|                     Fsm Output List                   |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmout_list
  {
    struct fsmout_list *NEXT;
    char               *NAME;
    long                FLAGS;
    void               *USER;

  } fsmout_list;

/*------------------------------------------------------\
|                                                       |
|                     Fsm Port List                     |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmport_list
  {
    struct fsmport_list *NEXT;
    char                *NAME;
    char                 DIR;
    char                 TYPE;
    long                 FLAGS;
    void                *USER;

  } fsmport_list;

/*------------------------------------------------------\
|                                                       |
|                Fsm Local Output List                  |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmlocout_list

  {
    struct fsmlocout_list  *NEXT;
    fsmout_list            *OUT;
    ablexpr                *ABL;
    bddnode                *BDD;
    ablexpr                *ABL_DC;
    bddnode                *BDD_DC;
    long                    FLAGS;
    void                   *USER;

  } fsmlocout_list;

/*------------------------------------------------------\
|                                                       |
|                  Fsm Transition List                  |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmtrans_list
  
  {
    struct fsmtrans_list  *NEXT;
    struct fsmtrans_list **PREV;
    struct fsmstate_list  *FROM;
    struct fsmstate_list  *TO;
    ablexpr               *ABL;
    bddnode               *BDD;
    long                   FLAGS;
    void                  *USER;

  } fsmtrans_list;

/*------------------------------------------------------\
|                                                       |
|                        Fsm State                      |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmstate_list

  {
    struct fsmstate_list   *NEXT;
    struct fsmstate_list  **PREV;
    char                   *NAME;
    chain_list             *FROM;
    chain_list             *TO;
    chain_list             *STACK;
    fsmlocout_list         *LOCOUT;
    long                    FLAGS;
    void                   *USER;

  } fsmstate_list;

/*------------------------------------------------------\
|                                                       |
|                     Fsm Stack List                    |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmstack_list
  {
    struct fsmstack_list  *NEXT;
    struct fsmstack_list **PREV;
    char                   CTRL;
    fsmstate_list         *CURRENT;
    fsmstate_list         *RETURN;
    ablexpr               *ABL;
    bddnode               *BDD;
    long                   FLAGS;
    void                  *USER;

  } fsmstack_list;

/*------------------------------------------------------\
|                                                       |
|                   Fsm Figure List                     |
|                                                       |
\------------------------------------------------------*/

  typedef struct fsmfig_list 
  {
    struct fsmfig_list *NEXT;
    char               *NAME;
    long                NUMBER_IN;
    long                NUMBER_OUT;
    long                NUMBER_PORT;
    long                NUMBER_STATE;
    long                NUMBER_TRANS;
    long                NUMBER_STACK;
    fsmin_list         *IN;
    fsmout_list        *OUT;
    fsmport_list       *PORT;
    fsmstate_list      *STATE;
    fsmtrans_list      *TRANS;
    fsmstack_list      *STACK;

    authtable          *HASH_PORT;
    authtable          *HASH_IN;
    authtable          *HASH_OUT;
    authtable          *HASH_STATE;
    auth2table         *HASH_TRANS;

    long                STACK_SIZE;

    fsmstate_list      *FIRST_STATE;
    fsmstate_list      *STAR_STATE;

    char               *CLOCK;
    ablexpr            *CLOCK_ABL;

    bddcircuit         *CIRCUIT;

    long                FLAGS;
    ptype_list         *PRAGMA;

    void               *USER;
    
  } fsmfig_list;

/*------------------------------------------------------\
|                                                       |
|                     Global Variables                  |
|                                                       |
\------------------------------------------------------*/

 extern fsmfig_list   *HEAD_FSMFIG;
 extern char          *FSM_CTRL_NAME [ FSM_MAX_CTRL ];

/*------------------------------------------------------\
|                                                       |
|                        Functions                      |
|                                                       |
\------------------------------------------------------*/
/*------------------------------------------------------\
|                                                       |
|                    Alloc Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern fsmfig_list    *allocfsmfig();
  extern fsmstate_list  *allocfsmstate();
  extern fsmstack_list  *allocfsmstack();
  extern fsmtrans_list  *allocfsmtrans();
  extern fsmlocout_list *allocfsmlocout();
  extern fsmout_list    *allocfsmout();
  extern fsmin_list     *allocfsmin();
  extern fsmport_list   *allocfsmport();

/*------------------------------------------------------\
|                                                       |
|                      Free Functions                   |
|                                                       |
\------------------------------------------------------*/

  extern void  freefsmfig();
  extern void  freefsmstate();
  extern void  freefsmstack();
  extern void  freefsmtrans();
  extern void  freefsmlocout();
  extern void  freefsmout();
  extern void  freefsmin();
  extern void  freefsmport();

/*------------------------------------------------------\
|                                                       |
|                      Add Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern fsmfig_list    *addfsmfig();
  extern fsmstate_list  *addfsmstate();
  extern fsmstack_list  *addfsmstack();
  extern fsmtrans_list  *addfsmtrans();
  extern fsmlocout_list *addfsmlocout();
  extern fsmout_list    *addfsmout();
  extern fsmin_list     *addfsmin();
  extern fsmport_list   *addfsmport();

/*------------------------------------------------------\
|                                                       |
|                      Del Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern int   delfsmfig();
  extern int   delfsmstate();
  extern int   delfsmstack();
  extern int   delfsmtrans();
  extern int   delfsmlocout();

/*------------------------------------------------------\
|                                                       |
|                      Search Functions                 |
|                                                       |
\------------------------------------------------------*/

  extern fsmfig_list    *searchfsmfig();
  extern fsmstate_list  *searchfsmstate();
  extern fsmstack_list  *searchfsmstack();
  extern fsmtrans_list  *searchfsmtrans();
  extern fsmlocout_list *searchfsmlocout();
  extern fsmout_list    *searchfsmout();
  extern fsmin_list     *searchfsmin();
  extern fsmport_list   *searchfsmport();

/*------------------------------------------------------\
|                                                       |
|                        Abl Functions                  |
|                                                       |
\------------------------------------------------------*/

  extern void    simpfsmablexpr();

/*------------------------------------------------------\
|                                                       |
|                       Bdd Functions                   |
|                                                       |
\------------------------------------------------------*/

  extern void makefsmbddcircuit();
  extern void makefsmbddnode();

  extern void convertfsmbddnodeabl();

/*------------------------------------------------------\
|                                                       |
|                     View Functions                    |
|                                                       |
\------------------------------------------------------*/

  extern void viewfsmfig();
  extern void viewfsmstate();
  extern void viewfsmstack();
  extern void viewfsmtrans();
  extern void viewfsmlocout();
  extern void viewfsmout();
  extern void viewfsmin();
  extern void viewfsmport();

# endif 
