/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Fsm                           |
|                                                             |
| File    :                   fsmadd.c                        |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include FSM_H

# include <stdio.h>
# include "fsmadd.h"
# include "fsmerror.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      Fsm Add Functions                      |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                        Fsm Add Figure                       |
|                                                             |
\------------------------------------------------------------*/

fsmfig_list *addfsmfig( Name )

   char *Name;
{
  fsmfig_list *Figure;

  Figure = allocfsmfig();

  Figure->NAME = namealloc( Name );
  Figure->NEXT = HEAD_FSMFIG;
  HEAD_FSMFIG  = Figure;

  Figure->HASH_PORT  = createauthtable( 100 );
  Figure->HASH_IN    = createauthtable( 100 );
  Figure->HASH_OUT   = createauthtable( 100 );
  Figure->HASH_STATE = createauthtable( 100 );
  Figure->HASH_TRANS = createauth2table( 100 );

  return( Figure );
}

/*------------------------------------------------------------\
|                                                             |
|                        Fsm Add State                        |
|                                                             |
\------------------------------------------------------------*/

fsmstate_list *addfsmstate( Figure, Name )

   fsmfig_list *Figure;
   char        *Name;
{
  fsmstate_list *State;

  Name  = namealloc( Name );
  State = searchfsmstate( Figure, Name );

  if ( State != (fsmstate_list *)0 )
  {
    fsmerror( FSM_STATE_EXIST_ERROR, Name );
  }

  State = allocfsmstate();

  State->NAME    = namealloc( Name  );
  State->NEXT    = Figure->STATE;
  State->PREV    = &Figure->STATE;

  if ( Figure->STATE != (fsmstate_list *)0 )
  {
    Figure->STATE->PREV = &State->NEXT;
  }

  Figure->STATE  = State;

  Figure->NUMBER_STATE++;

  addauthelem( Figure->HASH_STATE, Name, State );

  return( State );
}

/*------------------------------------------------------------\
|                                                             |
|                        Fsm Add Stack                        |
|                                                             |
\------------------------------------------------------------*/

fsmstack_list *addfsmstack( Figure, Control, Current, Return, Condition )

   fsmfig_list   *Figure;
   char           Control;
   fsmstate_list *Current;
   fsmstate_list *Return;
   ablexpr       *Condition;
{
  fsmstack_list *Stack;

  Stack = searchfsmstack( Control, Current, Return );

  if ( Stack == (fsmstack_list *)0 )
  {
    Stack = allocfsmstack();
  
    Current->STACK = addchain( Current->STACK, (void *)Stack );
  
    if ( Condition != (ablexpr *)0 ) 
    {
      Stack->ABL = dupablexpr( Condition );
    }
  
    if ( Current == Figure->STAR_STATE )
    {
      SetFsmStarStack( Stack );
    }
  
    Stack->CTRL    = Control;
    Stack->CURRENT = Current;
    Stack->RETURN  = Return;
    Stack->NEXT    = Figure->STACK;
    Stack->PREV    = &Figure->STACK;

    if ( Figure->STACK != (fsmstack_list *)0 )
    {
      Figure->STACK->PREV = &Stack->NEXT;
    }

    Figure->STACK = Stack;
  
    Figure->NUMBER_STACK++;
  }
  else
  if ( Condition != (ablexpr *)0 )
  {
    if ( Stack->ABL != (ablexpr *)0 )
    {
      Stack->ABL =

       createablbinexpr( ABL_OR, dupablexpr( Condition ), Stack->ABL );
    }
    else
    {
      Stack->ABL = dupablexpr( Condition );
    }
  }

  return( Stack );
}

/*------------------------------------------------------------\
|                                                             |
|                        Fsm Add Transition                   |
|                                                             |
\------------------------------------------------------------*/

fsmtrans_list *addfsmtrans( Figure, StateFrom, StateTo, Condition )

   fsmfig_list   *Figure;
   fsmstate_list *StateFrom;
   fsmstate_list *StateTo;
   ablexpr       *Condition;
{
  fsmtrans_list *Trans;

  Trans = searchfsmtrans( Figure, StateFrom, StateTo );

  if ( Trans == (fsmtrans_list *)0 )
  {
    Trans = allocfsmtrans();

    if ( Condition != (ablexpr *)0 )
    {
      Trans->ABL = dupablexpr( Condition );
    }

    StateFrom->FROM = addchain( StateFrom->FROM, (void *)Trans );
    StateTo->TO     = addchain( StateTo->TO,     (void *)Trans );

    Trans->FROM   = StateFrom;
    Trans->TO     = StateTo;
    Trans->NEXT   = Figure->TRANS;
    Trans->PREV   = &Figure->TRANS;

    if ( Figure->TRANS != (fsmtrans_list *)0 )
    {
      Figure->TRANS->PREV = &Trans->NEXT;
    }

    Figure->TRANS = Trans;

    if ( ( StateFrom == Figure->STAR_STATE ) ||
         ( StateTo   == Figure->STAR_STATE ) )
    {
      SetFsmStarTrans( Trans );
    }

    Figure->NUMBER_TRANS++;

    addauth2elem( Figure->HASH_TRANS, StateFrom, StateTo, Trans );
  }
  else
  if ( Condition != (ablexpr *)0 )
  {
    if ( Trans->ABL != (ablexpr *)0 )
    {
      Trans->ABL = 

       createablbinexpr( ABL_OR, dupablexpr( Condition ), Trans->ABL );
    }
    else
    {
      Trans->ABL = dupablexpr( Condition );
    }
  }

  return( Trans );
}

/*------------------------------------------------------------\
|                                                             |
|                     Fsm Add Local Output                    |
|                                                             |
\------------------------------------------------------------*/

fsmlocout_list *addfsmlocout( State, Output, Equation, EquationDC )

   fsmstate_list *State;
   fsmout_list   *Output;
   ablexpr       *Equation;
   ablexpr       *EquationDC;
{
  fsmlocout_list *Locout;

  Locout = allocfsmlocout();

  Locout->OUT = Output;

  if ( Equation != (ablexpr *)0 )
  {
    Locout->ABL = dupablexpr( Equation );
  }

  if ( EquationDC != (ablexpr *)0 )
  {
    Locout->ABL_DC = dupablexpr( EquationDC );
  }

  Locout->NEXT  = State->LOCOUT;
  State->LOCOUT = Locout;

  return( Locout );
}

/*------------------------------------------------------------\
|                                                             |
|                         Fsm Add Output                      |
|                                                             |
\------------------------------------------------------------*/

fsmout_list *addfsmout( Figure, Name )

   fsmfig_list   *Figure;
   char          *Name;
{
  fsmout_list *Output;

  Name   = namealloc( Name );
  Output = searchfsmout( Figure, Name );

  if ( Output != (fsmout_list *)0 )
  {
    fsmerror( FSM_OUT_EXIST_ERROR, Name );
  }

  Output = allocfsmout();

  Output->NAME = namealloc( Name );
  Output->NEXT = Figure->OUT;
  Figure->OUT  = Output;

  Figure->NUMBER_OUT++;

  addauthelem( Figure->HASH_OUT, Name, Output );

  return( Output );
}

/*------------------------------------------------------------\
|                                                             |
|                         Fsm Add Input                       |
|                                                             |
\------------------------------------------------------------*/

fsmin_list *addfsmin( Figure, Name )

   fsmfig_list   *Figure;
   char          *Name;
{
  fsmin_list *Input;

  Name  = namealloc( Name );
  Input = searchfsmin( Figure, Name );

  if ( Input != (fsmin_list *)0 )
  {
    fsmerror( FSM_IN_EXIST_ERROR, Name );
  }

  Input = allocfsmin();

  Input->NAME = namealloc( Name );
  Input->NEXT = Figure->IN;
  Figure->IN  = Input;

  Figure->NUMBER_IN++;

  addauthelem( Figure->HASH_IN, Name, Input );

  return( Input );
}

/*------------------------------------------------------------\
|                                                             |
|                         Fsm Add Port                        |
|                                                             |
\------------------------------------------------------------*/

fsmport_list *addfsmport( Figure, Name, Dir, Type )

   fsmfig_list   *Figure;
   char          *Name;
   char           Dir;
   char           Type;
{
  fsmport_list *Port;

  Name = namealloc( Name );
  Port = searchfsmport( Figure, Name );

  if ( Port != (fsmport_list *)0 )
  {
    fsmerror( FSM_PORT_EXIST_ERROR, Name );
  }

  Port = allocfsmport();

  Port->NAME   = Name;
  Port->DIR    = Dir;
  Port->TYPE   = Type;
  Port->NEXT   = Figure->PORT;
  Figure->PORT = Port;

  Figure->NUMBER_PORT++;

  addauthelem( Figure->HASH_PORT, Name, Port );

  return( Port );
}
