/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                     Fsm                           |
|                                                             |
| File    :                    main.c                         |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
| Author  :               Jacomme Ludovic                     |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <string.h>

# include MUT_H
# include AUT_H
# include ABL_H
# include BDD_H
# include FSM_H
# include FTL_H

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Usage                            |
|                                                             |
\------------------------------------------------------------*/

void Usage()
{
  fprintf( stderr, "\t\tfsmtest [Options] Input_name\n\n" );

  fprintf( stdout, "\t\tOptions : -o Sets Order mode on\n" );
  fprintf( stdout, "\t\tOptions : -v Sets view mode on\n" );
  fprintf( stdout, "\t\tOptions : -s Sets save mode on\n" );
  fprintf( stdout, "\t\tOptions : -f Sets fact mode on\n" );
  fprintf( stdout, "\n" );

  exit( 1 );
}
/*------------------------------------------------------------\
|                                                             |
|                          Functions                          |
|                                                             |
\------------------------------------------------------------*/

int main( argc, argv )

  int   argc;
  char *argv[];
{
  fsmfig_list *FsmFigure;
  char        *InputFileName;
  int          Number;
  int          FlagOrder;
  int          FlagSave;
  int          FlagView;
  int          FlagFact;

  mbkenv();
  autenv();
  ablenv();
  bddenv();
  fsmenv();

  FlagOrder     = 0;
  FlagSave      = 0;
  InputFileName = (char *)0;

  if ( argc < 2 ) Usage();

  for ( Number = 1; Number  < argc; Number++ )
  {
    if ( ! strcmp( argv[ Number ], "-o" ) ) FlagOrder = 1;
    else
    if ( ! strcmp( argv[ Number ], "-s" ) ) FlagSave = 1;
    else
    if ( ! strcmp( argv[ Number ], "-f" ) ) FlagFact = 1;
    else
    if ( ! strcmp( argv[ Number ], "-v" ) ) FlagView = 1;
    else
    if ( InputFileName  == (char *)0 ) InputFileName = argv[ Number ];
    else
    Usage();
  }

  FsmFigure = getfsmfig( InputFileName );

  if ( FlagView )
  {
    viewfsmfig( FsmFigure );
  }

  fprintf( stdout, "Number In    : %ld\n", FsmFigure->NUMBER_IN    );
  fprintf( stdout, "Number Out   : %ld\n", FsmFigure->NUMBER_OUT   );
  fprintf( stdout, "Number State : %ld\n", FsmFigure->NUMBER_STATE );
  fprintf( stdout, "Number Trans : %ld\n", FsmFigure->NUMBER_TRANS );

/*\
  BddSystem  = createbddsystem( 100, 1000, 100, 10000 );
  GefSystem  = creategefsystem( 5000, BddSystem );

  makefsmbddcircuit( FsmFigure, BddSystem, FlagOrder );
  makefsmbddnode( FsmFigure );

  testbddcircuit( BddCircuit );

  if ( FlagFact )
  {
    factfsmbddnodeabl( FsmFigure, GefSystem );
  }
\*/

  if ( FlagSave )
  {
    FsmFigure->NAME = namealloc( "out" );

    simpfsmablexpr( FsmFigure );
    savefsmfig( FsmFigure );
  }

  delfsmfig( FsmFigure->NAME );

/*\
  destroygefsystem( GefSystem );
  destroybddcircuit( BddCircuit );
  destroybddsystem( BddSystem );
\*/

  return( 0 );
}
