/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: fvh.h							*/
/* date		: Aug 28 1992						*/
/* version	: 102							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* contents	: defines and structure definitions used in FBH library	*/
/*									*/
/* ###--------------------------------------------------------------### */

#ifndef FBH_DEFINE
#define FBH_DEFINE

	/* ###------------------------------------------------------### */
	/*    defines							*/
	/* ###------------------------------------------------------### */

#define FBH_GENERIC 203698	/* generic ptype code			*/
#define FBH_BEIFS   205635	/* if     statement ptype code		*/
#define FBH_BECAS   423240	/* case   statement ptype code		*/
#define FBH_BEMSG   153676	/* assert statement ptype code		*/
#define FBH_BEASG   967324	/* assign statement ptype code		*/

	/* ###------------------------------------------------------### */
	/*    structure definitions					*/
	/* ###------------------------------------------------------### */

typedef struct fbfig		/* fbhaviour figure			*/
  {
  struct fbfig   *NEXT;		/* next figure				*/
  char           *NAME;		/* figure's name			*/
  struct fbreg   *BEREG;	/* internal signals (REGISTER)		*/
  struct fbmsg   *BEMSG;	/* concurrent assert instructions	*/
  struct fbrin   *BERIN;	/* inputs (registers and ports)		*/
  struct fbout   *BEOUT;	/* output ports				*/
  struct fbbus   *BEBUS;	/* output ports (BUS)			*/
  struct fbaux   *BEAUX;	/* internal signals			*/
  struct fbaux   *BEDLY;	/* internal delayed signals			*/
  struct fbbux   *BEBUX;	/* internal signals (BUS)		*/
  struct fbpor   *BEPOR;	/* ports (reverse order of declaration)	*/
  struct fbpcs   *BEPCS;	/* list of processes			*/
  struct bddcircuit  *CIRCUI;	/* Circuit's pointer used for BDDs	*/
  struct ptype   *USER;		/* reserved for user's applications	*/
  char            FLAG;		/* error flags				*/
  char            TYPE;		/* type of the description (not used)	*/
  struct fbtyp   *BETYP;	/* list of declared types		*/
  struct fbpgm   *BEPGM;	/* list of pragmas			*/
  }
fbfig_list;

typedef struct fbreg		/* register				*/
  {
  struct fbreg  *NEXT;		/* next register			*/
  char          *NAME;		/* register's name			*/
  struct biabl  *BIABL;		/* list of drivers (ABL)		*/
  struct binode *BINODE;	/* list of drivers (BDD)		*/
  void          *USER;
  }
fbreg_list;

typedef struct fbmsg		/* assert instruction			*/
  {
  struct fbmsg 	*NEXT;		/* next assertion			*/
  char		 LEVEL;		/* severity level of the assertion	*/
  char		*LABEL;		/* instruction's label			*/
  char		*MESSAGE;	/* assertion's message			*/
  struct chain	*ABL;		/* assertion's condition (ABL)		*/
  bddnode 	*NODE;		/* assertion's condition (BDD)		*/
  void          *USER;
  }
fbmsg_list;

typedef struct fbout		/* output port				*/
  {
  struct fbout	*NEXT;		/* next output				*/
  char		*NAME;		/* port's name				*/
  struct chain	*ABL;		/* port's equation (ABL)		*/
  bddnode 	*NODE;		/* port's equation (BDD)		*/
  unsigned char  TYPE;		/* port's type				*/
  void          *USER;
  }
fbout_list;

typedef struct fbbus		/* output port (BUS)			*/
  {
  struct fbbus	*NEXT;		/* next bus port			*/
  char		*NAME;		/* port's name				*/
  struct biabl	*BIABL;		/* list of drivers (ABL)		*/
  struct binode	*BINODE;	/* list of drivers (BDD)		*/
  unsigned char  TYPE;		/* port's type : W or M			*/
  void          *USER;
  }
fbbus_list;

typedef struct fbaux		/* auxiliary signal			*/
  {	
  struct fbaux	*NEXT;		/* next signal				*/
  char		*NAME;		/* signal's name			*/
  struct chain	*ABL;		/* signal's equation (ABL)		*/
  bddnode 	*NODE;		/* signal's equation (BDD)		*/
  unsigned char  TYPE;		/* signal's type			*/
  void          *USER;
  }
fbaux_list;

typedef struct fbbux		/* internal signal (BUS)		*/
  {	
  struct fbbux	*NEXT;		/* next signal				*/
  char		*NAME;		/* signal's name			*/
  struct biabl	*BIABL;		/* list of drivers (ABL)		*/
  struct binode	*BINODE;	/* list of drivers (BDD)		*/
  unsigned char  TYPE;		/* signal's type : W or M		*/
  void          *USER;
  }
fbbux_list;

typedef struct fbpor		/* port					*/
  {
  struct fbpor *NEXT;		/* next port				*/
  char         *NAME;		/* port's name				*/
  char          DIRECTION;	/* port's mode (same as LOCON)		*/
  unsigned char TYPE;		/* port's type (B, M or W)		*/
  void          *USER;
  }
fbpor_list;

typedef struct biabl		/* couple of ABL			*/
  {
  struct biabl	*NEXT;		/* next couple				*/
  char          *LABEL;		/* driver's label			*/
  struct chain	*CNDABL;	/* driver's connection condition (ABL)	*/
  struct chain	*VALABL;	/* equation of driven value (ABL)	*/
  void         *USER;
  }
biabl_list;

typedef struct binode		/* couple of BDD			*/
  {
  struct binode	*NEXT;	 	/* next figure				*/
  bddnode 	*CNDNODE;	/* driver's connection condition (BDD)	*/
  bddnode 	*VALNODE;	/* equation of driven value (BDD)	*/
  void         *USER;
  }
binode_list;

typedef struct fbrin		/* primary inputs			*/
  {
  struct fbrin *NEXT;		/* next input				*/
  char         *NAME;		/* input's name				*/
  void         *USER;
  }
fbrin_list;

typedef struct fbgen		/* generic				*/
  {
  struct fbgen  *NEXT;		/* next generic				*/
  char 	        *NAME;		/* generic's name			*/
  void          *VALUE;		/* generic's value			*/
  unsigned char  TYPE;		/* generic type's name			*/
  void         *USER;
  }
fbgen_list;

typedef struct fbpcs		/* process				*/
  {
  struct fbpcs *NEXT;		/* next process				*/
  char         *LABEL;		/* process's label			*/
  struct chain *SENSITIVITY;	/* list of sensitivity			*/
  struct ptype *INSTRUCTION;	/* list of instructions			*/
  struct chain *OUTPUTS;	/* list of assigned signals		*/
  char          FLAG;		/* flag (not used)			*/
  void         *USER;
  }
fbpcs_list;

typedef struct fbifs		/* if statement				*/
  {
  struct chain *CND;		/* condition (abl)			*/
  struct ptype *CNDTRUE;	/* list of instructions when true	*/
  struct ptype *CNDFALSE;	/* list of instructions when false	*/
  void         *USER;
  }
fbifs_list;

typedef struct fbasg		/* signal assignment statement		*/
  {
  char          *NAME;		/* target's name			*/
  struct chain  *ABL;		/* condition (abl)			*/
  unsigned char  TYPE;		/* signal's type			*/
  void         *USER;
  }
fbasg_list;

typedef struct fbcas		/* case statement			*/
  {
  struct chain  *ABL;		/* expression (ABL)			*/
  struct fbcho  *CHOICE;	/* table of choices			*/
  unsigned char  TYPE;		/* expression's type			*/
  unsigned int   SIZE;		/* size of choices' table		*/
  void          *USER;
  }
fbcas_list;

typedef struct fbcho		/* a choice in a case instruction	*/
  {
  char         *VALUE;		/* the value of the choice		*/
  unsigned int  SIZE;		/* size of the value in byte		*/
  struct ptype *INSTRUCTION;	/* list of instructions			*/
  void          *USER;
  }
fbcho_list;

typedef struct fbtyp		/* type or subtype			*/
  {
  struct fbtyp   *NEXT;		/* next fbtyp				*/
  char           *NAME;		/* type's name				*/
  int             LEFT;		/* left bound				*/
  int             RIGHT;	/* right bound				*/
  char          **VALUE;	/* list of values			*/
  unsigned int    SIZE;		/* number of values			*/
  unsigned int    BYTE;		/* number of bytes needed to code values*/ 
  char            CLASS;	/* type's class (E, I, A, S)		*/
  unsigned char   INDEX;	/* the type of the index for an array	*/
  unsigned char   BASE;		/* the base type or element's type	*/
  unsigned char   RESOLV;	/* index of resolution function		*/
  void           *USER;
  }
fbtyp_list;

typedef struct fbpgm		/* pragma				*/
  {
  struct fbpgm *NEXT;		/* next pragma				*/
  char         *TYPE;		/* prgma's type				*/
  char         *NAME;		/* object's name			*/
  void         *VALUE;		/* prgma's content			*/
  void         *USER;
  }
fbpgm_list;


	/* ###------------------------------------------------------### */
	/*    low-level functions					*/
	/* ###------------------------------------------------------### */

extern fbfig_list *FBL_HEADFIG;

extern struct fbaux  *fbh_addfbaux ();
extern struct fbbux  *fbh_addfbbux ();
extern struct fbbus  *fbh_addfbbus ();
extern struct fbfig  *fbh_addfbfig ();
extern struct fbgen  *fbh_addfbgen ();
extern struct fbmsg  *fbh_addfbmsg ();
extern struct fbout  *fbh_addfbout ();
extern struct fbpor  *fbh_addfbpor ();
extern struct fbreg  *fbh_addfbreg ();
extern struct fbrin  *fbh_addfbrin ();
extern struct biabl  *fbh_addbiabl ();
extern struct binode *fbh_addbinode ();
extern struct ptype  *fbh_addfbifs ();
extern struct ptype  *fbh_addfbasg ();
extern struct ptype  *fbh_addfbcas ();
extern struct fbpcs  *fbh_addfbpcs ();
extern struct fbtyp  *fbh_addfbtyp ();
extern struct fbpgm  *fbh_addfbpgm ();
extern struct fbtyp  *fbh_crtfbtyp ();
extern struct fbaux  *fbh_delfbaux ();
extern struct fbbux  *fbh_delfbbux ();
extern struct fbbus  *fbh_delfbbus ();
extern struct fbfig  *fbh_delfbfig ();
extern struct fbgen  *fbh_delfbgen ();
extern struct fbmsg  *fbh_delfbmsg ();
extern struct fbout  *fbh_delfbout ();
extern struct fbpor  *fbh_delfbpor ();
extern struct fbreg  *fbh_delfbreg ();
extern struct fbrin  *fbh_delfbrin ();
extern struct biabl  *fbh_delbiabl ();
extern struct binode *fbh_delbinode ();
extern void           fbh_frefbaux ();
extern void           fbh_frefbbux ();
extern void           fbh_frefbbus ();
extern void           fbh_frefbfig ();
extern void           fbh_frefbgen ();
extern void           fbh_frefbmsg ();
extern void           fbh_frefbout ();
extern void           fbh_frefbpor ();
extern void           fbh_frefbreg ();
extern void           fbh_frefbrin ();
extern void           fbh_frefbpgm ();
extern void           fbh_frebiabl ();
extern void           fbh_frebinode ();
extern void           fbh_frefbtyp ();
extern void          *fbh_getgenval ();
extern struct fbaux  *fbh_rmvfbaux ();
extern struct fbbux  *fbh_rmvfbbux ();
extern struct fbbus  *fbh_rmvfbbus ();
extern struct fbfig  *fbh_rmvfbfig ();
extern struct fbgen  *fbh_rmvfbgen ();
extern struct fbmsg  *fbh_rmvfbmsg ();
extern struct fbout  *fbh_rmvfbout ();
extern struct fbpor  *fbh_rmvfbpor ();
extern struct fbreg  *fbh_rmvfbreg ();
extern struct fbrin  *fbh_rmvfbrin ();

extern struct fbtyp *fbh_prdeftyp ();
	/* ###------------------------------------------------------### */
	/*    high-level functions					*/
	/* ###------------------------------------------------------### */

extern void           fbh_debug ();

	/* ###------------------------------------------------------### */
	/*    system-level functions					*/
	/* ###------------------------------------------------------### */

extern FILE          *fbh_opnfile ();

	/* ###------------------------------------------------------### */
	/*    parser-driver functions					*/
	/* ###------------------------------------------------------### */

extern struct fbfig  *fbh_prsvhdl ();
extern struct fbfig  *vhdlloadfbfig();
extern void   vhdlsavefbfig ();

#endif
