/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                   XPAT                            |
|                                                             |
| File    :                   Pat.c                           |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <signal.h>
# include <string.h>
# include <setjmp.h>
# include MUT_H
# include AUT_H
# include PAT_H
# include PPT_H
# include XSB_H 
# include XPT_H
# include "XPT_pat.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/

  patfig_list *XpatFigurePat = (patfig_list *)NULL;

/*------------------------------------------------------------\
|                                                             |
|                       For Set Long Jump                     |
|                                                             |
\------------------------------------------------------------*/

  static jmp_buf XpatJumpBuffer;
  static void    (*OldExitHandler)() = NULL;

/*------------------------------------------------------------\
|                                                             |
|                           Functions                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                 Xpat Exit Handler Functions                 |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                      XpatExitHandler                        |
|                                                             |
\------------------------------------------------------------*/

void XpatExitHandler()
{
  XpatFlushErrorMessage();

  signal( SIGTERM, OldExitHandler );
  longjmp( XpatJumpBuffer, 1);
}

/*------------------------------------------------------------\
|                                                             |
|                      XpatActiveExitHandler                  |
|                                                             |
\------------------------------------------------------------*/

void XpatActiveExitHandler()
{
  OldExitHandler = signal( SIGTERM, XpatExitHandler );
}

/*------------------------------------------------------------\
|                                                             |
|                       XpatResetExitHandler                  |
|                                                             |
\------------------------------------------------------------*/

void XpatResetExitHandler()
{
  signal( SIGTERM, OldExitHandler );
}

/*------------------------------------------------------------\
|                                                             |
|                          Get Functions                      |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          pat_frepaseq                       |
|                                                             |
\------------------------------------------------------------*/

static void pat_frepaseq( Sequence )

  paseq_list *Sequence;
{
  pagrp_list *PaGrp;
  pains_list *PaIns;

  pat_frepaiol( Sequence->PAIOL  );
  pat_frepapat( Sequence->OLDPAT );
  pat_frepapat( Sequence->CURPAT );
  pat_frepapat( Sequence->NXTPAT );
  pat_frepacom( Sequence->DECCOM );
  pat_frepacom( Sequence->OLDCOM );
  pat_frepacom( Sequence->CURCOM );

  while ( Sequence->PAGRP != (pagrp_list *)0 )
  {
    PaGrp = Sequence->PAGRP;
    Sequence->PAGRP = PaGrp->NEXT;

    mbkfree( PaGrp );
  }

  while ( Sequence->PAINS != (pains_list *)0 )
  {
    PaIns = Sequence->PAINS;
    Sequence->PAINS = PaIns->NEXT;

    mbkfree( PaIns );
  }

  mbkfree( Sequence );
}

/*------------------------------------------------------------\
|                                                             |
|                          pat_delpatfig                      |
|                                                             |
\------------------------------------------------------------*/

void pat_delpatfig( FigurePat )

  patfig_list *FigurePat;
{
  pat_frepaseq( FigurePat->SEQUENCE );

  autfreeblock( FigurePat->IO_NAME_ARRAY  );
  autfreeblock( FigurePat->IO_MODE_ARRAY  );
  autfreeblock( FigurePat->IO_WIDTH_ARRAY );

  autfreeblock( FigurePat );
}

/*------------------------------------------------------------\
|                                                             |
|                          pat_addpatfig                      |
|                                                             |
\------------------------------------------------------------*/

patfig_list *pat_addpatfig( Sequence )

  paseq_list *Sequence;
{
  patfig_list *FigurePat;
  char       **ModeArray;
  char       **NameArray;
  long        *WidthArray;

  char         Buffer[ 512 ];
  char         Name  [ 256 ];
  int          Left;
  int          Right;        
  char        *Direction;
  char        *Mode;

  long         NumberIO;
  long         NumberBit;
  long         IndexBit;

  paiol_list  *ScanIol;
  pagrp_list  *ScanGrp;

  if ( Sequence == (paseq_list *)0 )
  {
    return( (patfig_list *)0 );
  }

  if ( Sequence->ERRFLG != 0 )
  {
    pat_frepaseq( Sequence );

    return( (patfig_list *)0 );
  }

  FigurePat = (patfig_list *)autallocblock( sizeof( patfig_list ) );
 
  NumberBit = Sequence->IOLNBR;
  ScanGrp   = Sequence->PAGRP;

  ScanIol = Sequence->PAIOL;

  NameArray  = (char **)autallocblock( sizeof( char * ) * NumberBit );
  ModeArray  = (char **)autallocblock( sizeof( char * ) * NumberBit );
  WidthArray = (long  *)autallocblock( sizeof( long   ) * NumberBit );

  NumberIO  = 0;
  IndexBit  = 0;

  while ( ScanIol != (paiol_list *)0 )
  {
    switch ( ScanIol->MODE)
    {
      case 'I': Mode = "in";
      break;
      case 'O':
      case 'B':
      case 'Z': Mode = "out";
      break;
      case 'R': Mode = "register";
      break;
      case 'T': Mode = "inout";
      break;
      case 'S':
      case 'W':
      case 'X':
      case 'U': Mode = "signal";
      break;
    }

    if ( ( ScanGrp  != (pagrp_list *)0 ) &&
         ( IndexBit == ScanGrp->FINDEX ) )
    {
      if ( ScanIol->FORMAT != 'B' )
      {
        sscanf( ScanIol->NAME, "%s %d", Name, &Left );

        ScanIol  += ScanGrp->LENGTH - 1;
        IndexBit += ScanGrp->LENGTH;

        sscanf( ScanIol->NAME, "%s %d", Name, &Right );

        if ( Left < Right ) Direction = "to";
        else                Direction = "downto";

        sprintf( Buffer, "%s (%d %s %d)", Name, Left, Direction, Right );

        NameArray [ NumberIO ] = namealloc( Buffer );
        ModeArray [ NumberIO ] = namealloc( Mode   );
        WidthArray[ NumberIO ] = ScanGrp->LENGTH;
        NumberIO = NumberIO + 1;

        ScanGrp = ScanGrp->NEXT;
        ScanIol = ScanIol->NEXT;

        continue;
      }
    }

    NameArray [ NumberIO ] = ScanIol->NAME;
    ModeArray [ NumberIO ] = namealloc( Mode );
    WidthArray[ NumberIO ] = 1;
    NumberIO = NumberIO + 1;

    ScanIol  = ScanIol->NEXT;
    IndexBit = IndexBit + 1;
  }

  FigurePat->NAME           = Sequence->NAME;
  FigurePat->SEQUENCE       = Sequence;
  FigurePat->IO_NUMBER_BIT  = NumberBit;
  FigurePat->IO_NUMBER      = NumberIO;
  FigurePat->IO_NAME_ARRAY  = NameArray;
  FigurePat->IO_MODE_ARRAY  = ModeArray;
  FigurePat->IO_WIDTH_ARRAY = WidthArray;

  return( FigurePat );
}

/*------------------------------------------------------------\
|                                                             |
|                          Xpatlodpaseq                       |
|                                                             |
\------------------------------------------------------------*/

patfig_list *Xpatloadpatfig( FileName )

  char *FileName;
{
  paseq_list *Sequence;

  autbegin();

  Sequence = (paseq_list *)NULL;

  XpatActiveExitHandler();

  if ( setjmp( XpatJumpBuffer ) == 0 )
  {
    Sequence = pat_lodpaseq( FileName, (paseq_list *)0, 0 );

    XpatResetExitHandler();

    autend();
    return( pat_addpatfig( Sequence ) );
  }

  if ( Sequence != (paseq_list *)NULL )
  {
    pat_frepaseq( Sequence );
  }

  autend();
  return( (patfig_list *)NULL );
}
