/* 
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This library is free software; you  can redistribute it and/or modify it
 * under the terms  of the GNU Library General Public  License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
/*------------------------------------------------------------\
|                                                             |
| Tool    :                   XPAT                            |
|                                                             |
| File    :                  Place.c                          |
|                                                             |
| Authors :              Jacomme Ludovic                      |
|                                                             |
| Date    :                   04.12.96                        |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                         Include Files                       |
|                                                             |
\------------------------------------------------------------*/

# include <stdio.h>
# include <stdlib.h>
# include <math.h>
# include <signal.h>
# include <string.h>
# include <setjmp.h>
# include MUT_H
# include AUT_H
# include PAT_H
# include XSB_H 
# include XPT_H
# include "XPT_place.h"
# include "XPT_pat.h"

/*------------------------------------------------------------\
|                                                             |
|                           Constants                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                            Types                            |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          Variables                          |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                           Functions                         |
|                                                             |
\------------------------------------------------------------*/
/*------------------------------------------------------------\
|                                                             |
|                          XpatPlaceInputOutput               |
|                                                             |
\------------------------------------------------------------*/

void XpatPlaceInputOutput()
{
  xpatobj_list *Text;
  char        **NameArray;
  long          NumberIO;
  long          Index;
  long          StepX;
  long          StepY;
  long          Y;
  long          X;

  autbegin();

  StepX = ( XPAT_UNIT * XPAT_PATTERN_STEP_X );
  StepY = ( XPAT_UNIT * XPAT_PATTERN_STEP_Y );

  NameArray = XpatFigurePat->IO_NAME_ARRAY;
  NumberIO  = XpatFigurePat->IO_NUMBER;

  X = - 2 * StepX;
  Y = StepY / 4;

  for ( Index = 0; Index < NumberIO; Index++ )
  {
    Text = XpatAddText( X, Y, NameArray[ Index ], 6 );
    Text->LINK = Text;

    SetXpatIO( Text );

    Y += StepY;
  }

  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                          XpatGetBitString                   |
|                                                             |
\------------------------------------------------------------*/

char *XpatGetBitString( BitString, Width )

  char *BitString;
  long  Width;
{
  char Buffer[ 512 ];
  char ScanChar;
  long Index;
  long Target;
  long Sum;
  long Shift;

  Target = ( ( Width - 1 ) >> 2 );
  Sum    = 0;
  Shift  = 0;

  Buffer[ Target + 1 ] = '\0';

  for ( Index = ( Width - 1 ); Index >= 0; Index-- )
  {
    ScanChar = BitString[ Index ];

    if ( ScanChar == '*' ) 
    {
      return( namealloc( "*" ) );
    }

    Sum   += ( ( ScanChar - '0' ) << Shift );
    Shift += 1;

    if ( Shift == 4 )
    {
      if ( Sum > 9 ) Buffer[ Target-- ] = Sum + 'A' - 10;
      else           Buffer[ Target-- ] = Sum + '0';

      Sum   = 0;
      Shift = 0;
    }
  }

  if ( Shift != 0 )
  {
    if ( Sum > 9 ) Buffer[ Target-- ] = Sum + 'A' - 10;
    else           Buffer[ Target-- ] = Sum + '0';
  }

  return( namealloc( Buffer ) );
}

/*------------------------------------------------------------\
|                                                             |
|                          XpatPlacePattern                   |
|                                                             |
\------------------------------------------------------------*/

void XpatPlacePattern()
{
  xpatobj_list  *Object;
  xpatobj_list  *ScanObj;
  paseq_list    *PaSeq;
  papat_list    *PaPat;
  paevt_list    *PaEvt;
  long          *WidthArray;
  char          *CurrentEvent;
  char          *NextEvent;
  xpatobj_list **ObjectEvent;
  char           Event;
  char          *NextString;
  char          *CurrentString;
  char          *Name;
  char           FirstTime;
  long           NumberIO;
  long           NumberBit;
  long           IndexIO;
  long           IndexBit;
  long           Width;
  long           X;
  long           Y;
  long           StepX;
  long           StepY;
  long           Height;
  long           HalfHeight;

  autbegin();

  NumberIO   = XpatFigurePat->IO_NUMBER;
  NumberBit  = XpatFigurePat->IO_NUMBER_BIT;
  PaSeq      = XpatFigurePat->SEQUENCE;
  WidthArray = XpatFigurePat->IO_WIDTH_ARRAY;

  CurrentEvent = autallocblock( NumberBit + 1 );
  NextEvent    = autallocblock( NumberBit + 1 );
  ObjectEvent  = (xpatobj_list **)autallocblock( sizeof( xpatobj_list *) * NumberIO );

  memset( NextEvent, '*', NumberBit );

  StepX      = ( XPAT_UNIT * XPAT_PATTERN_STEP_X );
  StepY      = ( XPAT_UNIT * XPAT_PATTERN_STEP_Y );
  HalfHeight = ( StepY / 4 );
  Height     = ( StepY / 2 );
  X          = 0;
  FirstTime  = 1;

  for ( PaPat  = PaSeq->CURPAT;
        PaPat != (papat_list *)0;
        PaPat  = PaPat->NEXT )
  {
    if ( ! FirstTime )
    {
      memcpy( CurrentEvent, NextEvent, NumberBit );
    }

    for ( PaEvt  = PaPat->PAEVT;
          PaEvt != (paevt_list *)0;
          PaEvt  = PaEvt->NEXT )
    {
      Event = PaEvt->USRVAL;

      if ( Event == '-' ) Event = '0';
      else
      if ( Event == '+' ) Event = '1';

      NextEvent[ PaEvt->INDEX ] = Event;
    }

    if ( FirstTime )
    {
      FirstTime = 0;
      memcpy( CurrentEvent, NextEvent, NumberBit );
    }

    Y        = 0;
    IndexBit = 0;
    IndexIO  = 0;

    while ( IndexBit < NumberBit )
    {
      Width = WidthArray[ IndexIO ];

      if ( Width == 1 )
      {
        Object = ObjectEvent[ IndexIO ];

        if ( ( NextEvent[ IndexBit ] == CurrentEvent[ IndexBit ] ) &&
             ( Object                != (xpatobj_list *)0        ) )
        {
          Object->DX += StepX;
        }
        else
        {
          switch ( NextEvent[ IndexBit ] )
          {
            case '0' : 
    
              Object = XpatAddLine( X, Y, X + StepX, Y, 0 );
              Name   = namealloc( "0" );
    
              if ( CurrentEvent[ IndexBit ] == '1' )
              {
                Object->LINK = XpatAddArrow( X, Y, X, Y + Height, 2 );
                Object->LINK->LINK = XpatAddLine( X, Y, X, Y + Height, 0 );
                Object->LINK->LINK->LINK = Object;
              }
              else
              {
                Object->LINK = Object;
              }
            break;
    
            case '1' : 
    
              Object = XpatAddLine( X, Y + Height, X + StepX, Y + Height, 1 );
              Name   = namealloc( "1" );
    
              if ( CurrentEvent[ IndexBit ] == '0' )
              {
                Object->LINK = XpatAddArrow( X, Y + Height, X, Y, 3 );
                Object->LINK->LINK = XpatAddLine( X, Y + Height, X, Y, 1 );
                Object->LINK->LINK->LINK = Object;
              }
              else
              {
                Object->LINK = Object;
              }
            break;
    
            case '*' : 
    
              Object = XpatAddRectangle( X, Y, StepX, Height, 4 );
              Name   = namealloc( "*" );
              Object->LINK = Object;

            break;
          }

          ScanObj = Object;

          do
          {
            ScanObj->NAME = Name;
            ScanObj = ScanObj->LINK;
          }
          while ( ScanObj != Object );
        }
      }
      else
      {
        Object = ObjectEvent[ IndexIO ];

        CurrentString = XpatGetBitString( CurrentEvent + IndexBit, Width );
        NextString    = XpatGetBitString( NextEvent    + IndexBit, Width );

        if ( ( NextString == CurrentString     ) &&
             ( Object     != (xpatobj_list *)0 ) )
        {
          Object->DX += StepX;
        }
        else
        {
          if ( NextString[ 0 ] == '*' )
          {
            Object = XpatAddHexagon( X, Y, StepX, Height, 4 );
            Object->LINK = Object;
          }
          else
          {
            Object = XpatAddHexagon( X, Y, StepX, Height, 5 );
            SetXpatVector( Object );
            Object->LINK = Object;
          }

          Name    = namealloc( NextString );
          ScanObj = Object;

          do
          {
            ScanObj->NAME = Name;
            ScanObj = ScanObj->LINK;
          }
          while ( ScanObj != Object );
        }
      }

      ObjectEvent[ IndexIO++ ] = Object;

      IndexBit += Width;
      Y        += StepY;
    }

    X += StepX;
  }

  autfreeblock( CurrentEvent );
  autfreeblock( NextEvent    );
  autfreeblock( ObjectEvent  );
  
  autend();
}

/*------------------------------------------------------------\
|                                                             |
|                          XpatPlaceFigure                    |
|                                                             |
\------------------------------------------------------------*/

void XpatPlaceFigure()
{
  autbegin();

  XpatPlaceInputOutput();
  XpatPlacePattern();

  autend();
}
