/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/* $Id: GenericGN.c,v 1.10 1997/10/30 13:27:51 fitpath Exp fitpath $ */
/*                                                                          */
/*          Alliance CAD System 3.2b                                         */
/*             FpGen Package 1.10                                           */
/*                                                                          */
/*  Copyright(c) 93-97, MASI, CAO-VLSI Team                                 */
/*                                                                          */
/*  Author      :          Jean-Paul CHAPUT                                 */
/*  E-mail      :      cao-vlsi@masi.ibp.fr                                 */
/* ************************************************************************ */
/*  Module      :      "GenericGN.c"                                        */


/* ******************************* Includes ******************************* */


#include  <stdio.h>
#include  <string.h>
#include  <varargs.h>
#include  <time.h>

#include  <genlib.h>

#include  "GenericGN.h"


/* ************************** Internal constants ************************** */


#define       IE_GfpGN_WIDTH                2
#define       IE_GfpGN_SLICE                3
#define       IE_GfpGN_DRIVE                4
#define       IE_GfpGN_NOVIEW               5
#define       IE_GfpGN_OPENVBE              6
#define       IE_GfpGN_FUNCTION             7
#define       IE_GfpGN_WidthDP              8
#define       IE_GfpGN_WidthOp              9
#define       IE_GfpGN_SliceEven           10
#define       IW_GfpGN_PowerMerge           2

#define       IF_GfpGN_WidthOp              1
#define       IF_GfpGN_Slice                2
#define       IF_GfpGN_Drive                4
#define       IF_GfpGN_Mask                 8

#define       IF_GfpGN_MemMaster            1
#define       IF_GfpGN_MemSlave             2
#define       IF_GfpGN_MemScan              4
#define       IF_GfpGN_MemReset             8

#define       IF_GfpGN_ViewVHDL             1
#define       IF_GfpGN_ViewNetlist          2
#define       IF_GfpGN_ViewLayout           4
#define       IF_GfpGN_ViewIcon             8

#define       IC_GfpGN_CellNBInv            4
#define       IC_GfpGN_CellNBBuff           3
#define       IC_GfpGN_CellNBNand2          2
#define       IC_GfpGN_CellNBNand3          2
#define       IC_GfpGN_CellNBAnd2           2
#define       IC_GfpGN_CellNBAnd3           2
#define       IC_GfpGN_CellNBNor2           2
#define       IC_GfpGN_CellNBNor3           2
#define       IC_GfpGN_CellNBOr2            2
#define       IC_GfpGN_CellNBOr3            2
#define       IC_GfpGN_CellNBXnor2          1
#define       IC_GfpGN_CellNBXor2           1
#define       IC_GfpGN_CellNBNand2mask      3
#define       IC_GfpGN_CellNBNor2mask       3
#define       IC_GfpGN_CellNBXnor2mask      3
#define       IC_GfpGN_CellNBConst          2
#define       IC_GfpGN_CellNBNbuse          2
#define       IC_GfpGN_CellNBBuse           2
#define       IC_GfpGN_CellNBNmux2cs        2
#define       IC_GfpGN_CellNBMux2cs         2
#define       IC_GfpGN_CellNBMux3cd         2
#define       IC_GfpGN_CellNBMux3cs         2
#define       IC_GfpGN_CellNBMux4cs         2
#define       IC_GfpGN_CellNBMux5cs         2
#define       IC_GfpGN_CellNBMux6cs         2
#define       IC_GfpGN_CellNBAdd2s         (7+1)
#define       IC_GfpGN_CellNBNpl1           2
#define       IC_GfpGN_CellNBPl1            2
#define       IC_GfpGN_CellNBPmemnw         3
#define       IC_GfpGN_CellNBPmemnwq        3
#define       IC_GfpGN_CellNBPmem           3
#define       IC_GfpGN_CellNBPmemr          3
#define       IC_GfpGN_CellNBPmemq          3
#define       IC_GfpGN_CellNBPmemt          8
#define       IC_GfpGN_CellNBPmemrt         8
#define       IC_GfpGN_CellNBPmemtq         8

#define       IC_GfpGN_ModelNameMaxLen    256
#define       IC_GfpGN_BusWideMaxLen      256
#define       IC_GfpGN_PinNameMaxLen      256
#define       IC_GfpGN_NameMaxLen         256
#define       IC_GfpGN_PowerStringLen      16


/* ********************* Internal Types And Structures ******************** */


typedef struct {
  long  ModuleId;      /* Module identifier.                    */
  long  PrintFMark;    /* Tag for self-allocated strings.       */

  long  aFunction;     /* Operator to generate.                 */
  char *aModelName;    /* Operator model name.                  */
  long  aWidthDP;      /* Data-path maximal width.              */
  long  aWidthOp;      /* Operator width.                       */
  long  aSlice;        /* LSB starting slice.                   */
  long  aDrive;        /* Output power (fan-out).               */
  long *aXlMask;       /* Logical Mask (constant).              */
  long  aMSB0;         /* MSB as zero index.                    */
  long  aPower;        /* Merge the power supplies.             */
  long  aViews;        /* The views to generate.                */

  FILE *PtFile;        /* VHDL file descriptor.                 */
  long  LeafCellIndex; /* Index of the leaf cells list inside
                        * the "TablLeafCell[]".
                        */
  } GV_t;


/* Table for Grey code conversion. */
static long  TablGreyCode[16] =
    {
      0x00, 0x01, 0x03, 0x02,
      0x06, 0x07, 0x05, 0x04,
      0x0C, 0x0D, 0x0F, 0x0D,
      0x0A, 0x0B, 0x09, 0x08
    };


/* Name of the leaf cells belonging to all generators. */
static char *TablLeafCell[] =
    {
    /* Boolean generators leaf cells. */
         "n1",    "np1",   "ndrv", "ndrvp",
       "buff",    "drv",  "drvp",
        "na2",   "na2p",
        "na3",   "na3p",
         "a2",    "a2p",
         "a3",    "a3p",
        "no2",   "no2p",
        "no3",   "no3p",
         "o2",    "o2p",
         "o3",    "o3p",
       "nxr2",
        "xr2",
    /* Mask generators leaf cells. */
      "bna2x",   "na2x", "n1na2x",
      "bno2x", "n1no2x",   "no2x",
     "bnxr2x", "n1xr2x",  "nxr2x",
    /* Buse generators leaf cells. */
       "bnt1",    "nt1",
        "bt1",     "t1",
    /* Multiplexer generators leaf cells. */
    "nmx2sb0", "nmx2sc",
     "mx2sb0",  "mx2sc",
     "mx3db0",  "mx3dc",
     "mx3sb0",  "mx3sc",
     "mx4sb0",  "mx4sc",
     "mx5sb0",  "mx5sc",
     "mx6sb0",  "mx6sc",
    /* Adder generators leaf cells. */
      "nobuffer",   "bcsa",
          "csag",   "csa0",  "csap",   "csai", "csap0", "csai0",
    /* Register generators leaf cells. */
      "pl1b0", "npl1a0",
      "pl1b0",  "pl1a0",
      "pmsb0", "pdffb0", "pmsa0",
      "pmsb0", "pdffb0", "pmsa1",
      "pmsb2", "pdffb2", "pmsa0",
      "pmsb2", "pdffb2", "pmsa1",
      "pmsba", "pdffba", "pmsa9",
      "pmsb6", "pdffb6",
      "pmsg4",  "pmsh4",  "pmsp4", "pmsi4", "pmsq4", "pmsj4",
      "pmsb6", "pdffb6",
      "pmsg5",  "pmsh5",  "pmsp5", "pmsi5", "pmsq5", "pmsj5",
      "pmsbe", "pdffbe",
      "pmsgd",  "pmshd",  "pmspd", "pmsid", "pmsqd", "pmsjd",
    /* Constant generators leaf cells. */
       "zero",    "one"
    };


/* ************************** Internal Variables ************************** */


static GV_t  GV;


/* ************************* Interface Functions ************************** */




/* ************************** Internal Functions ************************** */


#ifndef  ANSI_PROTO
  static void  Init();
  static long  llog2();
  static long  FunctionCheck();
  static void  FunctionArgs();
  static long  FunctionDrive();
  static long  FunctionCells();
  static char *FunctionModelData();
  static char *FunctionModelBuff();
  static char *FunctionStr();
  static char *OperName();
  static void  HeaderVHDL();
  static void  SignalVHDL();
  static void  DataFlowBool();
  static void  DataFlowMask();
  static void  DataFlowConst();
  static void  DataFlowBuse();
  static void  DataFlowMux();
  static void  DataFlowComp();
  static void  DataFlowMem();
  static void  EntityVHDL();
  static void  ArchiVHDL();
  static void  CreateVHDL();
  static void  CreateNetlist();
  static void  CreateNetlistComp();
  static void  CreatePower();
  static void  CreateHoDPThru();
  static void  CreateLayout();
#else
  static void  Init( void );
  static long  llog2( long value );
  static void  FunctionCheck( va_list ArgList );
  static void  FunctionArgs( void );
  static long  FunctionDrive( void );
  static long  FunctionCells( void );
  static char *FunctionModelData( long aSlice );
  static char *FunctionModelBuff( long aSlice );
  static char *FunctionStr( void );
  static char *OperName( void );
  static void  HeaderVHDL( void );
  static void  SignalVHDL( void );
  static void  DataFlowBool( void );
  static void  DataFlowMask( void );
  static void  DataFlowBuse( void );
  static void  DataFlowMux( void );
  static void  DataFlowComp( void );
  static void  DataFlowMem( void );
  static void  EntityVHDL( void );
  static void  ArchiVHDL( void );
  static void  CreateVHDL( void );
  static void  CreateNetlist( void );
  static void  CreateNetlistComp( void );
  static void  CreatePower( long  aInstanceIndex );
  static void  CreateHoDPThru( long  aInstanceIndex );
  static void  CreateLayout( void );
#endif


/* ************************** Macro Definitions *************************** */


#define  mc_IsEmptySlice(i)  (       ((i) <  GV.aSlice)     \
                              || (   ((i) >= GV.aSlice      \
                                            +GV.aWidthOp)   \
                                  && ((i) <  GV.aWidthDP)))

#define  Mc_IndexNorth(n,i)  \
         ((((GV.aWidthDP - GV.aWidthOp - GV.aSlice    ) % 2) * (n)) + (i))

#define  Mc_IndexBuffr(n,i)  (i)


/* ****************** Macro for various index computings ****************** */


#define  mc_SymbolicHeight(model)    \
            (HEIGHT(model) / GV.SliceHeight)

#define  mc_IndexToSlice(index)      \
            (((index) >= GV.aWidthDP)    \
                ?  (index)               \
                : ((index) + GV.aSlice))

#define  mc_IndexToSymF(index)       \
            ((GV.aWidthDP - mc_IndexToSlice((index))) % 2)

#define  mc_SliceToPower(slice)      \
            (((slice) + GV.aWidthDP % 2) / 2)

#define  mc_SliceToPowerType(slice)  \
            (((slice) + GV.aWidthDP) % 2)

#define  mc_SliceToPowerName(slice)  \
            ((mc_SliceToPowerType(slice)) ? "vdd" : "vss")

#define  mc_IndexToPower(index)      \
            (mc_SliceToPower(mc_IndexToSlice((index))))

#define  mc_IndexToPowerName(index)  \
            (mc_SliceToPowerName(mc_IndexToSlice((index))))

#define  mc_SliceToBit(slice)        \
            ((GV.aMSB0) ? (GV.aWidthDP - 1 - slice) : (slice))

#define  mc_IndexToBit(index)        \
            (mc_SliceToBit(mc_IndexToSlice((index))))


/* ***************** Macro for the logarithmic algorithms ***************** */


#define  mc_GreyCode(class,level,code)  \
            ((TablGreyCode[(code)]<<((level)-1))&((1L<<((class)-3))-1))

#define  mc_GreyChar(class,level,code)  \
            ((mc_GreyCode(class,level,code) < 10)              \
                ? (char)('0'+mc_GreyCode(class,level,code))    \
                : (char)('a'+mc_GreyCode(class,level,code)-10))

#define  mc_GateNB(n,layer)             \
            ((layer <= 1)               \
                ?  (((n) - 1) >> 2)     \
                : ((((n) - 1) + (1 << (layer))) >> ((layer) + 1)))

#define  mc_GateModulo(n)               \
            ((((n) - 1) % 4) + 1)


/* ****************** Macro for standard terminals names ****************** */


#define  mc_VHDL_WIDE(msb_i,lsb_i)  \
            (UtBas_PrintF("(%ld %s %ld)",                \
                          mc_IndexToBit(msb_i),          \
                          (GV.aMSB0) ? "to" : "downto",  \
                          mc_IndexToBit(lsb_i)))

#define  mc_VHDL_FULL_WIDE()        \
            (mc_VHDL_WIDE(GV.aWidthOp - 1, 0))

#define  mc_VHDL_MEMM_WIDE(msb_i,lsb_i) \
            (UtBas_PrintF("(%ld downto %ld)", msb_i, lsb_i))

#define  mc_VHDL_FULL_BUS(conname)  \
            (UtBas_PrintF("%s%s", (conname), mc_VHDL_FULL_WIDE()))

#define  mc_GenL_FULL_BUS(conname)  \
            (BUS((conname), mc_IndexToBit(GV.aWidthOp - 1), \
                            mc_IndexToBit(0)))

#define  mc_COPY_UP_CMA_ELM(inscon,insname,conname,index)  \
            (UtMPh_CopyUpCMA(                \
                (inscon),                    \
                (insname),                   \
                UtBas_PrintF("%s[%ld]",      \
                    (conname),               \
                    mc_IndexToBit(index))))

#define  mc_POWER_NAME(slice)  \
            (NAME("%s_%d", mc_SliceToPowerName(slice), \
                           mc_SliceToPower(slice)))

#define  mc_VDDE(conname,index)  \
            ((GV.aPower)                            \
              ? UtBas_PrintF("%s => vdd"   , (conname))  \
              : UtBas_PrintF("%s => vdd_%d", (conname),  \
                                             mc_IndexToPower(index)))

#define  mc_VSSE(conname,index)  \
            ((GV.aPower)                            \
              ? UtBas_PrintF("%s => vss"   , (conname))  \
              : UtBas_PrintF("%s => vss_%d", (conname),  \
                                             mc_IndexToPower(index)))


/* ************************* Function Definitions ************************* */


/*  Function  :  'Init'                                                     */
/* ------------------------------------------------------------------------ */
/*  Perform all the initialisation needed to use the boolean functions
 *  generator. This function must be used only one time, at the first call
 *  of the 'Dispatch' function. The initialisations performed are listed
 *  below :
 *    - Load the error data base whith module specific errors.
 */

static void  Init()
{
  /* Attribution of a module identifier. */
  GV.ModuleId = GfpUT_GetModuleId("GfpGN", F_GfpUT_ModuleAdd);


  GfpUT_AddError(GV.ModuleId,
                 IE_GfpGN_WIDTH,
  "[%s]:Invalid bus width:%ld.");
  GfpUT_AddError(GV.ModuleId,
                 IE_GfpGN_SLICE,
  "[%s]:Invalid LSB starting slice:%ld.");
  GfpUT_AddError(GV.ModuleId,
                 IE_GfpGN_DRIVE,
  "[%s]:invalid output drive:%ld.");
  GfpUT_AddError(GV.ModuleId,
                 IE_GfpGN_NOVIEW,
  "[%s]:No view to generate.");
  GfpUT_AddError(GV.ModuleId,
                 IE_GfpGN_OPENVBE,
  "[%s]:Unable to create file:\"%s\".");
  GfpUT_AddError(GV.ModuleId,
                 IE_GfpGN_FUNCTION,
  "[%s]:Unknow function code:%ld.");


  GfpUT_AddError( GV.ModuleId,
                  IE_GfpGN_WidthDP,
  "Invalid data-path wide : %d (%d < WideDP < %d)." );
  GfpUT_AddError( GV.ModuleId,
                  IE_GfpGN_WidthOp,
  "[%s]:Invalid operator wide : %d (%d < WideOp < %d)." );
  GfpUT_AddError( GV.ModuleId,
                  IE_GfpGN_SliceEven,
  "[%s]:This operator does not support odd starting slice : %d." );
                  


  GfpUT_AddWarng(GV.ModuleId,
                 IW_GfpGN_PowerMerge,
  "[%s]:Layout power supplies of model \"%s\" are not merged.");
}


/*  Function  :  'llog2'                                                    */
/* ------------------------------------------------------------------------ */

static long  llog2( value )
  long  value;
{
  long  bit; for( bit = 1; (value >> bit) ; bit++ ); return( bit - 1 );
}


/*  Function  :  'FunctionCheck'                                            */
/* ------------------------------------------------------------------------ */
/*  Check if the value given as argument have a legal value. That is,
 *  a value corresponding to one of a boolean function.
 */

static long  FunctionCheck()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

        return( TRUE );

    default:

        return( FALSE );
  }
}


/*  Function  :  'FunctionArgs'                                             */
/* ------------------------------------------------------------------------ */
/*  Get all the arguments from the command line, according to the function.
 *  Also performs checking to ensure correctness and coherency of these
 *  parameters.
 */

static void  FunctionArgs( ArgList )
  va_list  ArgList;
{
  long  ArgFlags;


  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      ArgFlags =  IF_GfpGN_WidthOp | IF_GfpGN_Slice | IF_GfpGN_Drive;

      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:

      ArgFlags = IF_GfpGN_WidthOp | IF_GfpGN_Slice | IF_GfpGN_Mask;

      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      ArgFlags = IF_GfpGN_WidthOp | IF_GfpGN_Slice;

      break;

    default:

      ArgFlags = 0L;
  }


  GV.aWidthDP   = va_arg( ArgList, long );

  /* Get and check the data-path wide. */
  if (    ( GV.aWidthDP < C_GfpGN_LimitsMinWidthDP )
       || ( GV.aWidthDP > C_GfpGN_LimitsMaxWidthDP ) )
  {
    GfpUT_PrintError( GV.ModuleId,
	                  IE_GfpGN_WidthDP,
                      GV.aWidthDP,
                      C_GfpGN_LimitsMinWidthDP - 1,
                      C_GfpGN_LimitsMaxWidthDP + 1 );
  }

  /* Get and check the bus wide. */
  if (ArgFlags & IF_GfpGN_WidthOp)
  {
    GV.aWidthOp = va_arg(ArgList, long);

    if ( (GV.aWidthOp < 1          )
      || (GV.aWidthOp > GV.aWidthDP))
      GfpUT_PrintError(GV.ModuleId,
                       IE_GfpGN_WIDTH,
                       FunctionStr(),
                       GV.aWidthOp);
  }
  else
    GV.aWidthOp = GV.aWidthDP;

  /* Get and check the LSB starting slice. */
  if (ArgFlags & IF_GfpGN_Slice)
  {
    GV.aSlice = va_arg(ArgList, long);

    if ( (GV.aSlice             < 0          )
      || (GV.aSlice+GV.aWidthOp > GV.aWidthDP))
      GfpUT_PrintError(GV.ModuleId,
                       IE_GfpGN_SLICE,
                       FunctionStr(),
                       GV.aSlice);
  }
  else
    GV.aSlice = 0L;

  /* Get and check the output drive capacity. */
  if (ArgFlags & IF_GfpGN_Drive)
  {
    GV.aDrive = va_arg(ArgList, long);

    if (   (GV.aDrive < 1              )
      || (GV.aDrive > FunctionDrive()))
      GfpUT_PrintError(GV.ModuleId,
                       IE_GfpGN_DRIVE,
                       FunctionStr(),
                       GV.aDrive);
  }
  else
    GV.aDrive = 1L;

  /* Get the Mask (extended long pointer). */
  if (ArgFlags & IF_GfpGN_Mask)
    GV.aXlMask = va_arg(ArgList, long*);
  else
    GV.aXlMask = (long*)NULL;

  /* Get the bus index convention. */
  GV.aMSB0    = va_arg(ArgList, long);

  /* Get all the views asked. */
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewVHDL    : 0L;
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewNetlist : 0L;
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewLayout  : 0L;
  GV.aViews  |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewIcon    : 0L;

  GV.aPower   = va_arg(ArgList, long);


  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionUnknow:

      break;

    case C_GfpGN_FunctionNul:

      if (    ( GV.aWidthOp <  4 )
           || ( GV.aWidthOp > 32 ) )
      {
		GfpUT_PrintError( GV.ModuleId,
					      IE_GfpGN_WidthOp,
                          FunctionStr(),
                          GV.aWidthOp,
                           4 - 1,
                          32 + 1 );
      }

      if ( GV.aSlice % 2 )
      {
		GfpUT_PrintError( GV.ModuleId,
					      IE_GfpGN_SliceEven,
                          FunctionStr(),
                          GV.aSlice );
      }

      break;
  }
}


/*  Function  :  'FunctionDrive'                                            */
/* ------------------------------------------------------------------------ */
/*  Return the maximal drive value associated with each boolean function.
 *  When no choice is possible the value is "1".
 */

static long  FunctionDrive()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:       return( C_GfpGN_DriveInv   );
    case C_GfpGN_FunctionBuff:      return( C_GfpGN_DriveBuff  );
    case C_GfpGN_FunctionNand2:     return( C_GfpGN_DriveNand2 );
    case C_GfpGN_FunctionNand3:     return( C_GfpGN_DriveNand3 );
    case C_GfpGN_FunctionAnd2:      return( C_GfpGN_DriveAnd2  );
    case C_GfpGN_FunctionAnd3:      return( C_GfpGN_DriveAnd3  );
    case C_GfpGN_FunctionNor2:      return( C_GfpGN_DriveNor2  );
    case C_GfpGN_FunctionNor3:      return( C_GfpGN_DriveNor3  );
    case C_GfpGN_FunctionOr2:       return( C_GfpGN_DriveOr2   );
    case C_GfpGN_FunctionOr3:       return( C_GfpGN_DriveOr3   );
    case C_GfpGN_FunctionXnor2:     return( C_GfpGN_DriveXnor2 );
    case C_GfpGN_FunctionXor2:      return( C_GfpGN_DriveXor2  );
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
    default:

      return(0);
  }
}


/*  Function  :  'FunctionCells'                                            */
/* ------------------------------------------------------------------------ */
/*  Return the offset of the first operator leaf cell, in the "TablLeafCell"
 *  table.
 */

static long  FunctionCells()
{
  long  LeafCellIndex;


  LeafCellIndex = 0;

  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNul:     LeafCellIndex += IC_GfpGN_CellNBConst;
    case C_GfpGN_FunctionConst:   LeafCellIndex += IC_GfpGN_CellNBPmemrt;
    case C_GfpGN_FunctionPdffrt:

      LeafCellIndex += IC_GfpGN_CellNBPmemt;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdfft:

      LeafCellIndex += IC_GfpGN_CellNBPmemtq;

    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:

      LeafCellIndex += IC_GfpGN_CellNBPmemr;

    case C_GfpGN_FunctionPdffr:

      LeafCellIndex += IC_GfpGN_CellNBPmem;

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPdff:

      LeafCellIndex += IC_GfpGN_CellNBPmemq;

    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffq:

      LeafCellIndex += IC_GfpGN_CellNBPmemnw;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPdffnw:

      LeafCellIndex += IC_GfpGN_CellNBPmemnwq;

    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPdffnwq:   LeafCellIndex += IC_GfpGN_CellNBPl1;
    case C_GfpGN_FunctionPl1:       LeafCellIndex += IC_GfpGN_CellNBNpl1;
    case C_GfpGN_FunctionNpl1:      LeafCellIndex += IC_GfpGN_CellNBAdd2s;
    case C_GfpGN_FunctionAdd2s:     LeafCellIndex += IC_GfpGN_CellNBMux6cs;
    case C_GfpGN_FunctionMux6cs:    LeafCellIndex += IC_GfpGN_CellNBMux5cs;
    case C_GfpGN_FunctionMux5cs:    LeafCellIndex += IC_GfpGN_CellNBMux4cs;
    case C_GfpGN_FunctionMux4cs:    LeafCellIndex += IC_GfpGN_CellNBMux3cs;
    case C_GfpGN_FunctionMux3cs:    LeafCellIndex += IC_GfpGN_CellNBMux3cd;
    case C_GfpGN_FunctionMux3cd:    LeafCellIndex += IC_GfpGN_CellNBMux2cs;
    case C_GfpGN_FunctionMux2cs:    LeafCellIndex += IC_GfpGN_CellNBNmux2cs;
    case C_GfpGN_FunctionNmux2cs:   LeafCellIndex += IC_GfpGN_CellNBBuse;
    case C_GfpGN_FunctionBuse:      LeafCellIndex += IC_GfpGN_CellNBNbuse;
    case C_GfpGN_FunctionNbuse:     LeafCellIndex += IC_GfpGN_CellNBXnor2mask;
    case C_GfpGN_FunctionXnor2mask: LeafCellIndex += IC_GfpGN_CellNBNor2mask;
    case C_GfpGN_FunctionNor2mask:  LeafCellIndex += IC_GfpGN_CellNBNand2mask;
    case C_GfpGN_FunctionNand2mask: LeafCellIndex += IC_GfpGN_CellNBXnor2;
    case C_GfpGN_FunctionXor2:      LeafCellIndex += IC_GfpGN_CellNBXor2;
    case C_GfpGN_FunctionXnor2:     LeafCellIndex += IC_GfpGN_CellNBOr3;
    case C_GfpGN_FunctionOr3:       LeafCellIndex += IC_GfpGN_CellNBOr2;
    case C_GfpGN_FunctionOr2:       LeafCellIndex += IC_GfpGN_CellNBNor3;
    case C_GfpGN_FunctionNor3:      LeafCellIndex += IC_GfpGN_CellNBNor2;
    case C_GfpGN_FunctionNor2:      LeafCellIndex += IC_GfpGN_CellNBAnd3;
    case C_GfpGN_FunctionAnd3:      LeafCellIndex += IC_GfpGN_CellNBAnd2;
    case C_GfpGN_FunctionAnd2:      LeafCellIndex += IC_GfpGN_CellNBNand3;
    case C_GfpGN_FunctionNand3:     LeafCellIndex += IC_GfpGN_CellNBNand2;
    case C_GfpGN_FunctionNand2:     LeafCellIndex += IC_GfpGN_CellNBBuff;
    case C_GfpGN_FunctionBuff:      LeafCellIndex += IC_GfpGN_CellNBInv;
    case C_GfpGN_FunctionInv:     ;
  }

  return( GV.LeafCellIndex = LeafCellIndex );
}


/*  Function  :  'FunctionModelData'                                        */
/* ------------------------------------------------------------------------ */
/*  Compute the model name of a data leaf cell, according to his slice index.
 */

static char *FunctionModelData( aSlice )
  long  aSlice;
{
  char *LeafCellModelName;


  /* Select the model name. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + GV.aDrive - 1 ];
      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      LeafCellModelName = TablLeafCell[
                            GV.LeafCellIndex + 1 + GfpUT_XlGetBit(
                                                     GV.aXlMask,
                                                     aSlice - GV.aSlice ) ];
      break;

    case C_GfpGN_FunctionConst:

      LeafCellModelName = TablLeafCell[
                            GV.LeafCellIndex + GfpUT_XlGetBit(
                                                 GV.aXlMask,
                                                 aSlice - GV.aSlice ) ];
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 2 ];
      break;

    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      if (aSlice == GV.aSlice)
        LeafCellModelName = TablLeafCell[
                              GV.LeafCellIndex + 6
                              + (GV.aWidthDP - aSlice) % 2 ];
      else
      {
        if (aSlice == GV.aSlice + GV.aWidthOp - 1)
          LeafCellModelName = TablLeafCell[
                                GV.LeafCellIndex + 2
                                + (GV.aWidthDP - aSlice) % 2 ];
        else
          LeafCellModelName = TablLeafCell[
                                GV.LeafCellIndex+4
                                + (GV.aWidthDP - aSlice) % 2 ];
      }
      break;

    default:

      return( (char*)NULL );
  }

  return( LeafCellModelName );
}


/*  Function  :  'FunctionModelBuff'                                        */
/* ------------------------------------------------------------------------ */
/*  Compute the model name of a buffer leaf cell, according to his slice
 *  index.
 */

static char *FunctionModelBuff( aSlice )
  long  aSlice;
{
  char *LeafCellModelName;


  /* Select the buffer model name.
   * (NULL if no need of a buffer)
   */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex ];
      break;

    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    default:

      LeafCellModelName = (char*)NULL;
      break;
  }


  return( LeafCellModelName );
}


/*  Function  :  'FunctionStr'                                              */
/* ------------------------------------------------------------------------ */
/*  Return the string used as root for building the model name.             */

static char *FunctionStr()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:       return( "inv"       );
    case C_GfpGN_FunctionBuff:      return( "buff"      );
    case C_GfpGN_FunctionNand2:     return( "nand2"     );
    case C_GfpGN_FunctionNand3:     return( "nand3"     );
    case C_GfpGN_FunctionAnd2:      return( "and2"      );
    case C_GfpGN_FunctionAnd3:      return( "and3"      );
    case C_GfpGN_FunctionNor2:      return( "nor2"      );
    case C_GfpGN_FunctionNor3:      return( "nor3"      );
    case C_GfpGN_FunctionOr2:       return( "or2"       );
    case C_GfpGN_FunctionOr3:       return( "or3"       );
    case C_GfpGN_FunctionXnor2:     return( "xnor2"     );
    case C_GfpGN_FunctionXor2:      return( "xor2"      );
    case C_GfpGN_FunctionNand2mask: return( "nand2mask" );
    case C_GfpGN_FunctionNor2mask:  return( "nor2mask"  );
    case C_GfpGN_FunctionXnor2mask: return( "xnor2mask" );
    case C_GfpGN_FunctionConst:     return( "const"     );
    case C_GfpGN_FunctionNbuse:     return( "nbuse"     );
    case C_GfpGN_FunctionBuse:      return( "buse"      );
    case C_GfpGN_FunctionNmux2cs:   return( "nmux2cs"   );
    case C_GfpGN_FunctionMux2cs:    return( "mux2cs"    );
    case C_GfpGN_FunctionMux3cd:    return( "mux3cd"    );
    case C_GfpGN_FunctionMux3cs:    return( "mux3cs"    );
    case C_GfpGN_FunctionMux4cs:    return( "mux4cs"    );
    case C_GfpGN_FunctionMux5cs:    return( "mux5cs"    );
    case C_GfpGN_FunctionMux6cs:    return( "mux6cs"    );
    case C_GfpGN_FunctionAdd2s:     return( "add2s"     );
    case C_GfpGN_FunctionNul:       return( "nul"       );
    case C_GfpGN_FunctionNpl1:      return( "npl1"      );
    case C_GfpGN_FunctionPl1:       return( "pl1"       );
    case C_GfpGN_FunctionPmsnw:     return( "pmsnw"     );
    case C_GfpGN_FunctionPmsnwq:    return( "pmsnwq"    );
    case C_GfpGN_FunctionPms:       return( "pms"       );
    case C_GfpGN_FunctionPmsq:      return( "pmsq"      );
    case C_GfpGN_FunctionPmst:      return( "pmst"      );
    case C_GfpGN_FunctionPmstq:     return( "pmstq"     );
    case C_GfpGN_FunctionPdffnw:    return( "pdffnw"    );
    case C_GfpGN_FunctionPdffnwq:   return( "pdffnwq"   );
    case C_GfpGN_FunctionPdff:      return( "pdff"      );
    case C_GfpGN_FunctionPdffr:     return( "pdffr"     );
    case C_GfpGN_FunctionPdffq:     return( "pdffq"     );
    case C_GfpGN_FunctionPdfft:     return( "pdfft"     );
    case C_GfpGN_FunctionPdffrt:    return( "pdffrt"    );
    case C_GfpGN_FunctionPdfftq:    return( "pdfftq"    );
    default:                        return( "errorgen"  );
  }
}


/*  Function  :  'OperName'                                                 */
/* ------------------------------------------------------------------------ */
/*  Generate the complete model name of the operator to create. The fonction
 *  use all the significant arguments to build this name. It return a
 *  pointer to an internal static string.
 */

static char *OperName()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      return( UtBas_PrintF( "%s_%ldx%ldx%ldx%ld%c_cl",
                                 FunctionStr(),
                                 GV.aWidthDP,
                                 GV.aWidthOp,
                                 GV.aSlice,
                                 GV.aDrive,
                               ((GV.aMSB0) ? 'm' : 'l')) );

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:

      return( UtBas_PrintF( "%s%s_%ldx%ldx%ld%c_cl",
                            FunctionStr(),
                            GfpUT_XlToStr(GV.aXlMask,
                                         GV.aWidthDP,
                                         C_GfpUT_BaseHexadecimal),
                            GV.aWidthDP,
                            GV.aWidthOp,
                            GV.aSlice,
                          ((GV.aMSB0) ? 'm' : 'l')) );

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdfftq:

      return( UtBas_PrintF( "%s_%ldx%ldx%ld%c_cl",
                            FunctionStr(),
                            GV.aWidthDP,
                            GV.aWidthOp,
                            GV.aSlice,
                          ((GV.aMSB0) ? 'm' : 'l')) );

    default:

      return( UtBas_PrintF( "errormodel" ) );
  }
}


/*  Function  :  'HeaderVHDL'                                               */
/* ------------------------------------------------------------------------ */
/*  Write comment header in the file.                                       */

static void  HeaderVHDL()
{
  char *String;
  long  Time, Length;


  /* Print the standart distribution header. */
  fprintf(GV.PtFile,
"-- ###-------------------------------------------------------------###\n"
          );
  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- #          Alliance CAD System 3.2b                                #\n"
          );
  fprintf( GV.PtFile,
"-- #             FpGen Package 1.10                                  #\n"
          );
  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- #  Copyright(c) 93-97, MASI, CAO-VLSI Team                        #\n"
          );
  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- #  Author      :          Jean-Paul CHAPUT                        #\n"
          );
  fprintf( GV.PtFile,
"-- #  E-mail      :      cao-vlsi@masi.ibp.fr                        #\n"
          );
  fprintf( GV.PtFile,
"-- # *************************************************************** #\n"
          );


  /* Print the generator name. */
  String = UtBas_PrintF( "\"%sg\"", FunctionStr() );
  fprintf( GV.PtFile, "-- #  Generator   : %-47s  #\n", String );


  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );


  /* Print name of the generated file. */
  String = UtBas_PrintF( "\"%s.vbe\".", GV.aModelName );
  fprintf( GV.PtFile, "-- #  File  :  %-52s  #\n", String );


  fprintf( GV.PtFile,
"-- # --------------------------------------------------------------- #\n"
          );


  /* Print the date of generation. */
  Time = time((time_t*)NULL);
  String = UtBas_PrintF( "%s", asctime( localtime( &Time ) ) );
  Length = strlen( String );
  String[ Length - 1 ] = (char)0;
  fprintf(GV.PtFile, "-- #            Date : %-44s  #\n", String);


  /* Print the call used to generate the model. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      String = UtBas_PrintF(
                 "\"%sg %ld %ld %ld %ld%s -vhdl \\",
                 FunctionStr(),
                 GV.aWidthDP,
                 GV.aWidthOp,
                 GV.aSlice,
                 GV.aDrive,
                (GV.aMSB0) ? " -msb0"  : "" );
      fprintf( GV.PtFile, "-- #  Generator Call : %-44s  #\n", String );

      String = UtBas_PrintF( "-o %s\"", GV.aModelName );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:

      String = UtBas_PrintF(
                 "\"%sg %ld %ld %ld \\",
                 FunctionStr(),
                 GV.aWidthDP,
                 GV.aWidthOp,
                 GV.aSlice );
      fprintf( GV.PtFile, "-- #  Generator Call : %-44s  #\n", String );

      String = UtBas_PrintF(
                 "0x%s \\", GfpUT_XlToStr( GV.aXlMask,
                                           GV.aWidthOp,
                                           C_GfpUT_BaseHexadecimal ) );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      String = UtBas_PrintF(
                 "%s -vhdl -o %s",
                (GV.aMSB0) ? " -msb0"  : "",
                 GV.aModelName );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdfftq:

      String = UtBas_PrintF(
                 "\"%sg %ld %ld %ld %s -vhdl \\",
                 FunctionStr(),
                 GV.aWidthDP,
                 GV.aWidthOp,
                 GV.aSlice,
                (GV.aMSB0) ? " -msb0"  : "");
      fprintf( GV.PtFile, "-- #  Generator Call : %-44s  #\n", String);

      String = UtBas_PrintF( "-o %s\"", GV.aModelName );
      fprintf( GV.PtFile, "-- # %62s  #\n", String );

      break;
  }


  fprintf( GV.PtFile,
"-- #                                                                 #\n"
          );
  fprintf( GV.PtFile,
"-- ###-------------------------------------------------------------###\n"
          );
}


/*  Function  :  'EntityVHDL'                                               */
/* ------------------------------------------------------------------------ */
/*  Create the ENTITY description of the operator. Essentially PORT part.
 */

static void  EntityVHDL()
{
  char *ModelName, *BusWide;
  long  Slice;


  /* Compute the VHDL full bus wide string. */
  BusWide = mc_VHDL_FULL_WIDE();


  fprintf( GV.PtFile, "\nENTITY %s IS\n\tPORT(\n", GV.aModelName );

  /* Controls Terminals. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionConst:

      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "msk" );
      break;

    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "sel" );
      break;

    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:

      fprintf( GV.PtFile,
               "%14s : in    BIT_VECTOR(1 downto 0);\n", "sel" );
      break;

    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile,
               "%14s : in    BIT_VECTOR(2 downto 0);\n", "sel" );
      break;

    case C_GfpGN_FunctionAdd2s:

      fprintf( GV.PtFile, "%14s :   out BIT;\n", "cout" );
      fprintf( GV.PtFile, "%14s :   out BIT;\n", "over" );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ck" );
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "scin" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "test" );

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "wen" );

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ckm" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "cks" );
      break;

    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "scin" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "test" );

    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "wen" );

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ck" );
      break;

    case C_GfpGN_FunctionPdffrt:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "scin" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "test" );

    case C_GfpGN_FunctionPdffr:

      fprintf( GV.PtFile, "%14s : in    BIT;\n", "reset" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "wen" );
      fprintf( GV.PtFile, "%14s : in    BIT;\n", "ck" );
      break;
  }

  /* Terminals of input data buses. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionConst:

      break;

    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i5", BusWide );

    case C_GfpGN_FunctionMux5cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i4", BusWide );

    case C_GfpGN_FunctionMux4cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i3", BusWide );

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i2", BusWide );

    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionAdd2s:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i1", BusWide );

    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionNul:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "i0", BusWide );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "%14s : in    BIT_VECTOR%s;\n", "d", BusWide );
      break;
  }

  /* Terminals of output data buses. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionOr2:

      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n",  "o", BusWide );

      if ( UtTec_Family() == C_UtTec_FamilyCMOS )
      {
        fprintf( GV.PtFile, "%14s : inout BIT_VECTOR%s;\n", "no", BusWide );
      }

      break;

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:

      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n", "o", BusWide );
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

      fprintf( GV.PtFile,
               "%14s :   out MUX_VECTOR%s BUS;\n", "o", BusWide );
      break;

    case C_GfpGN_FunctionNul:

      fprintf( GV.PtFile, "%14s :   out BIT;\n", "o" );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n", "q", BusWide );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:

      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n",  "q", BusWide );
      fprintf( GV.PtFile, "%14s :   out BIT_VECTOR%s;\n", "nq", BusWide );
      break;
  }


  /* Select the buffer model name. */
  /* (NULL if no need of a buffer) */
  ModelName = FunctionModelBuff(GV.aWidthDP);

  /* Power supplies terminals. */
  if (GV.aPower)
  {
    fprintf(GV.PtFile, "%14s : in    BIT;\n", "vdd");
    fprintf(GV.PtFile, "%14s : in    BIT\n" , "vss");
  }
  else
  {
    for(Slice = 0 ;
        Slice < GV.aWidthDP + ((ModelName != (char*)NULL) ? 2 : 0) ;
        Slice++)
      fprintf(GV.PtFile, "%14s : in    BIT;\n", mc_POWER_NAME(Slice));

    /* No semi-colon at the end of the last connector.
     * This explain why we separate the last iteration of the loop.
     */
    fprintf(GV.PtFile, "%14s : in    BIT\n", mc_POWER_NAME(Slice));
  }


  /* *** End of interface description *** */
  fprintf(GV.PtFile, "\t);\nEND %s;\n\n\n", GV.aModelName);
}


/*  Function  :  'SignalVHDL'                                               */
/* ------------------------------------------------------------------------ */
/*  Internal signals declaration.                                           */

static void  SignalVHDL()
{
  char *BusWide, *MemWide;


  BusWide = mc_VHDL_FULL_WIDE();
  MemWide = mc_VHDL_MEMM_WIDE( GV.aSlice + GV.aWidthOp - 1,
                               GV.aSlice );


  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionNul:

      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile,
               "\tSIGNAL %20s : BIT_VECTOR%s;\n", "otmp", BusWide );
      break;

    case C_GfpGN_FunctionAdd2s:

      {
        char *CryBusWide;

        CryBusWide = UtBas_PrintF(
                       "(%ld downto %ld)",
                       GV.aWidthOp + GV.aSlice,
                       GV.aSlice );
        fprintf( GV.PtFile,
                 "\tSIGNAL %20s : BIT_VECTOR%s;\n", "cry", CryBusWide);
      }
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "wmd" );
      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "memm", MemWide );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:

      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "memm", MemWide );

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:

      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "mems", MemWide );
      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "wmd" );
      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "ws"  );
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "wmt" );
      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "wmd" );
      fprintf( GV.PtFile, "\tSIGNAL %20s : BIT;\n", "ws"  );
      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "memm", MemWide );
      fprintf( GV.PtFile,
               "\tSIGNAL %20s : REG_VECTOR%s REGISTER;\n", "mems", MemWide );
      break;
  }
  fprintf( GV.PtFile, "\n" );
}


/*  Function  :  'DataFlowBool'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for booleans operators.                   */

static void  DataFlowBool()
{
	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionInv   :
			fprintf(GV.PtFile, "\to <= not i0;\n\n");
			break;
		case C_GfpGN_FunctionBuff  :
			fprintf(GV.PtFile, "\to <= i0;\n\n");
			break;
		case C_GfpGN_FunctionNand2 :
			fprintf(GV.PtFile, "\to <= not (i0 and i1);\n\n");
			break;
		case C_GfpGN_FunctionNand3 :
			fprintf(GV.PtFile, "\to <= not (i0 and i1 and i2);\n\n");
			break;
		case C_GfpGN_FunctionAnd2  :
			fprintf(GV.PtFile, "\t o <=      i0 and i1 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 and i1);\n\n");
			}

			break;
		case C_GfpGN_FunctionAnd3  :
			fprintf(GV.PtFile, "\t o <=      i0 and i1 and i2 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 and i1 and i2);\n\n");
			}

			break;
		case C_GfpGN_FunctionNor2  :
			fprintf(GV.PtFile, "\to <= not (i0 or i1);\n\n");
			break;
		case C_GfpGN_FunctionNor3  :
			fprintf(GV.PtFile, "\to <= not (i0 or i1 or i2);\n\n");
			break;
		case C_GfpGN_FunctionOr2   :
			fprintf(GV.PtFile, "\t o <=      i0 or i1 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 or i1);\n\n");
			}

			break;
		case C_GfpGN_FunctionOr3   :
			fprintf(GV.PtFile, "\t o <=      i0 or i1 or i2 ;\n"  );

            if ( UtTec_Family() == C_UtTec_FamilyCMOS )
            {
			  fprintf(GV.PtFile, "\tno <= not (i0 or i1 or i2);\n\n");
			}

			break;
		case C_GfpGN_FunctionXnor2 :
			fprintf(GV.PtFile, "\to <= not (i0 xor i1);\n\n");
			break;
		case C_GfpGN_FunctionXor2  :
			fprintf(GV.PtFile, "\to <= i0 xor i1;\n\n");
			break;
	}
}


/*  Function  :  'DataFlowMask'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for controllable mask operators.          */

static void  DataFlowMask()
{
	fprintf(GV.PtFile, "\tWITH  msk  SELECT\n");
	fprintf(GV.PtFile, "\t\totmp <=  i0          WHEN \'0\',\n");

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNand2mask:
			fprintf(GV.PtFile,
					"\t\t\t\t(i0 and %s) WHEN \'1\';\n",
					  GfpUT_XlToStrVHDL(GV.aXlMask ,
									   GV.aWidthOp));
			break;
		case C_GfpGN_FunctionNor2mask:
			fprintf(GV.PtFile,
					"\t\t\t\t(i0  or %s) WHEN \'1\';\n",
					 GfpUT_XlToStrVHDL(GV.aXlMask ,
									   GV.aWidthOp));
			break;
		case C_GfpGN_FunctionXnor2mask:
			fprintf(GV.PtFile,
					"\t\t\t\t(i0 xor %s) WHEN \'1\';\n",
					 GfpUT_XlToStrVHDL(GV.aXlMask ,
									   GV.aWidthOp));
			break;
	}

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNand2mask:
		case C_GfpGN_FunctionNor2mask:
		case C_GfpGN_FunctionXnor2mask:
			fprintf(GV.PtFile, "\n\t\to <= not otmp;\n\n");
			break;
		default:
			fprintf(GV.PtFile, "\n\to <= otmp;\n\n");
			break;
	}
}


/*  Function  :  'DataFlowConst'                                  */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for constants operators.                  */

static void  DataFlowConst()
{
	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionConst:
			fprintf(GV.PtFile, "\to <= %s;\n\n",
				  GfpUT_XlToStrVHDL(GV.aXlMask ,
								   GV.aWidthOp));
			break;
	}
}


/*  Function  :  'DataFlowBuse'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for tristate (bus) operators.             */

static void  DataFlowBuse()
{
	fprintf(GV.PtFile, "\twts:BLOCK(sel = '1')\n");
	fprintf(GV.PtFile, "\tBEGIN\n"               );

	/* Specific Data-Flow equations. */
	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNbuse:
			fprintf(GV.PtFile, "\t\to <= GUARDED not i0;\n");
			break;
		case C_GfpGN_FunctionBuse:
			fprintf(GV.PtFile, "\t\to <= GUARDED i0;\n");
			break;
	}

	fprintf(GV.PtFile, "\tEND BLOCK wts;\n\n");
}


/*  Function  :  'DataFlowMux'                                              */
/* ------------------------------------------------------------------------ */
/*  Data-Flow equations for multiplexer operators.                          */

static void  DataFlowMux()
{
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:

      fprintf( GV.PtFile,     "\tWITH  sel  SELECT\n" );
      fprintf( GV.PtFile,     "\t\totmp <= "          );
      fprintf( GV.PtFile,         "i0  WHEN \'0\',\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN \'1\';\n" );
      break;

    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n" );
      fprintf( GV.PtFile,     "\t\totmp <= "            );
      fprintf( GV.PtFile,         "i0  WHEN  \"00\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"01\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"10\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"11\";\n" );
      break;

    case C_GfpGN_FunctionMux4cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n" );
      fprintf( GV.PtFile,     "\t\totmp <= "            );
      fprintf( GV.PtFile,         "i0  WHEN  \"00\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"01\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"10\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti3  WHEN  \"11\";\n" );
      break;

    case C_GfpGN_FunctionMux5cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n"  );
      fprintf( GV.PtFile,     "\t\totmp <= "             );
      fprintf( GV.PtFile,         "i0  WHEN  \"000\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"001\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"010\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti3  WHEN  \"011\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti4  WHEN  \"100\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"101\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"110\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"111\";\n" );
      break;

    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile,       "\tWITH  sel  SELECT\n"  );
      fprintf( GV.PtFile,     "\t\totmp <= "             );
      fprintf( GV.PtFile,         "i0  WHEN  \"000\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti1  WHEN  \"001\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti2  WHEN  \"010\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti3  WHEN  \"011\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti4  WHEN  \"100\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"110\",\n" );
      fprintf( GV.PtFile, "\t\t\t\ti5  WHEN  \"101\"\n"  );
      fprintf( GV.PtFile, "\t\t\t\t        | \"111\";\n" );
      break;
  }

  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNmux2cs:

      fprintf( GV.PtFile, "\t\to <= not otmp;\n\n" );
      break;

    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      fprintf( GV.PtFile, "\t\to <= otmp;\n\n" );
      break;
  }
}


/*  Function  :  'DataFlowAdd'                                    */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for adder operators.                      */

static void  DataFlowAdd()
{
	char  CryBusWide1[IC_GfpGN_BusWideMaxLen];
	char  CryBusWide2[IC_GfpGN_BusWideMaxLen];

	sprintf(CryBusWide1,
			"(%ld downto %ld)", GV.aWidthOp
							   +GV.aSlice  -1,
							    GV.aSlice    );
	sprintf(CryBusWide2,
			"(%ld downto %ld)", GV.aWidthOp
							   +GV.aSlice    ,
							    GV.aSlice  +1);

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionAdd2s:
			fprintf(GV.PtFile, "\tcry(0)           <= \'0\';\n");
			fprintf(GV.PtFile, "\tcry%s <= (i0 and i1)\n",
												   CryBusWide2);
			fprintf(GV.PtFile, "\t\t\t\t\t or (i0 and cry%s)\n",
														 CryBusWide1);
			fprintf(GV.PtFile, "\t\t\t\t\t or (i1 and cry%s);\n\n",
														    CryBusWide1);
			fprintf(GV.PtFile, "\to        <= ");
			fprintf(GV.PtFile, "i0 xor i1 xor cry%s;\n\n", CryBusWide1);
			fprintf(GV.PtFile, "\tcout     <= cry(%ld);\n",
											   GV.aWidthOp
											  +GV.aSlice  );
			fprintf(GV.PtFile, "\tover     <= cry(%ld)",
											   GV.aWidthOp
										      +GV.aSlice  );
			fprintf(GV.PtFile,          " xor cry(%ld);\n\n",
											   GV.aWidthOp-1
										      +GV.aSlice    );
			break;
	}
}


/*  Function  :  'DataFlowComp'                                   */
/* -------------------------------------------------------------- */
/*  Data-Flow equations for comparators operators.                */

static void  DataFlowComp()
{
	long *XlNul, XlIndex;

	switch(GV.aFunction)
	{
		case C_GfpGN_FunctionNul:
			XlNul = GfpUT_XlAlloc();

			for(XlIndex=0; XlIndex<C_GfpUT_XLSIZE; XlIndex++)
				GfpUT_XlSetLong(XlNul, XlIndex, 0L);

			fprintf(GV.PtFile,
					"\to <= \'1\' WHEN (i0 = %s) ELSE \'0\';\n\n",
						GfpUT_XlToStrVHDL(XlNul,GV.aWidthOp));

			GfpUT_XlFree(XlNul);
			break;
	}
}


/*  Function  :  'DataFlowMem'                                              */
/* ------------------------------------------------------------------------ */
/*  Data-Flow equations for all memory elements. .                          */

static void  DataFlowMem()
{
  long  MemFlag;
  char *MemMast, *MemSlav, *Invert, *Direct;


  MemFlag = 0L;

  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      MemFlag |= IF_GfpGN_MemMaster;
      MemMast  = "memm";
      MemSlav  = "memm";
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      MemFlag |=  IF_GfpGN_MemScan;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:

      MemFlag |=  IF_GfpGN_MemMaster | IF_GfpGN_MemSlave;
      MemMast  = "memm";
      MemSlav  = "mems";
      break;
  }


  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:

      break;

    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdffr:

      MemFlag |=  IF_GfpGN_MemReset;
      break;
  }


  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:

      Invert = "not";
      Direct = "   ";
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:

      Invert = "   ";
      Direct = "not";
      break;
  }


  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      fprintf( GV.PtFile, "\twmd <= ck;\n" );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:

      fprintf( GV.PtFile, "\twmd <=     ckm;\n" );
      break;

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:

      fprintf( GV.PtFile, "\twmd <=     ckm and wen;\n" );
      break;

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:

      fprintf( GV.PtFile, "\twmd <=     ck;\n" );
      break;

    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:

      fprintf( GV.PtFile, "\twmd <=     ck  and wen;\n" );
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

      fprintf( GV.PtFile, "\twmt <=     ckm and      test;\n" );
      fprintf( GV.PtFile, "\twmd <=     ckm and (not test) and wen;\n" );
      break;

    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "\twmt <=     ck  and      test;\n" );
      fprintf( GV.PtFile, "\twmd <=     ck  and (not test) and wen;\n" );
      break;

    case C_GfpGN_FunctionPdffrt:

      fprintf( GV.PtFile, "\twmt <=     ck  and (not reset) and      test;\n" );
      fprintf( GV.PtFile, "\twmd <=     ck  and (not reset) and (not test) and wen;\n" );
      break;

    case C_GfpGN_FunctionPdffr:

      fprintf( GV.PtFile, "\twmd <=     ck  and (not reset) and wen;\n" );
      break;

  }


  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:

      fprintf( GV.PtFile, "\tws  <= cks;\n" );
      break;

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "\tws  <= not ck;\n" );
      break;
  }
  fprintf( GV.PtFile, "\n" );


  if (   (GV.aFunction != C_GfpGN_FunctionPdffnw)
      && (GV.aFunction != C_GfpGN_FunctionPdffnwq))
  {
    /* Write in slave register (MS and DFF only). */
    if (MemFlag & IF_GfpGN_MemReset)
    {
      long *XlReset, i;


      XlReset = GfpUT_XlAlloc();
      for( i = 0 ; i <= GV.aWidthOp / 32 ; i++ )
        GfpUT_XlSetLong( XlReset, i, 0xFFFFFFFF );

      fprintf( GV.PtFile, "\treset:BLOCK(reset = \'1\') BEGIN " );
      fprintf( GV.PtFile, "%s <= GUARDED %s; ", MemMast ,
               GfpUT_XlToStrVHDL( XlReset,  GV.aWidthOp ) );
      fprintf( GV.PtFile, "END BLOCK reset;\n" );

      GfpUT_XlFree( XlReset );
    }

    /* Write in master registers when in scan mode (MS & DFF only). */
    if (MemFlag & IF_GfpGN_MemScan)
    {
      fprintf( GV.PtFile, "\twmt:BLOCK(wmt) BEGIN\n" );
  
      fprintf( GV.PtFile, "\t\t%s <= GUARDED ", MemMast );
      fprintf( GV.PtFile, "%s%s & (not scin);\n", MemSlav,
               mc_VHDL_MEMM_WIDE( GV.aSlice + GV.aWidthOp - 2,
                                  GV.aSlice ) );
  
      fprintf(GV.PtFile, "\tEND BLOCK wmt;\n");
    }
  
  
    /* Write in master registers when in normal mode. */
    fprintf( GV.PtFile, "\twmd:BLOCK(wmd) BEGIN " );
    fprintf( GV.PtFile, "%s <= GUARDED %s d;    ", MemMast, Invert );
    fprintf( GV.PtFile, "END BLOCK wmd;\n" );
  
  
    /* Write in slave register (MS and DFF only). */
    if (MemFlag & IF_GfpGN_MemSlave)
    {
      fprintf( GV.PtFile, "\tws :BLOCK(ws ) BEGIN " );
      fprintf( GV.PtFile, "%s <= GUARDED     %s; ", MemSlav, MemMast );
      fprintf( GV.PtFile, "END BLOCK ws;\n" );
    }
    fprintf( GV.PtFile, "\n" );
  }
  else
  {
    fprintf( GV.PtFile, "\tck:BLOCK((not ck) and not(ck\'STABLE))\n");
    fprintf( GV.PtFile, "\tBEGIN\n" );
    fprintf( GV.PtFile, "\t\t%s <= GUARDED %s d;\n ", Invert, MemSlav );
    fprintf( GV.PtFile, "\tEND BLOCK ck;\n" );

    fprintf( GV.PtFile, "\n" );
  }


  /* Affect the outputs. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionNpl1:

      fprintf( GV.PtFile, "\tq <= not %s;\n", MemSlav );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:

      fprintf( GV.PtFile, "\tnq <= %s %s;\n", Direct, MemSlav );

    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfftq:

      fprintf( GV.PtFile, "\t q <= %s %s;\n", Invert, MemSlav );
      fprintf(GV.PtFile, "\n");
      break;
  }
  fprintf(GV.PtFile, "\n");
}


/*  Function  :  'ArchiVHDL'                                               */
/* ----------------------------------------------------------------------- */
/*  Create the ARCHITECTURE description of the operator. This description
 *  use the VHDL data-flow subset.
 */

static void  ArchiVHDL()
{
  char *ModelName;
  long  Slice;


  /* Begin of behavioral view. */
  fprintf(GV.PtFile,
      "ARCHITECTURE VBE OF %s IS\n\n",
                     GV.aModelName);

  /* Declaration of internals signals. */
  SignalVHDL();

  fprintf(GV.PtFile, "BEGIN\n\n");


  /* Data-Flow equations. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      DataFlowBool();
      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      DataFlowMask();
      break;

    case C_GfpGN_FunctionConst:

      DataFlowConst();
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

      DataFlowBuse();
      break;

    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      DataFlowMux();
      break;

    case C_GfpGN_FunctionAdd2s:

      DataFlowAdd();
      break;

    case C_GfpGN_FunctionNul:

      DataFlowComp();
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      DataFlowMem();
      break;
  }


  /* Check power supplies connections. */
  fprintf(GV.PtFile, "\t-- ****************>");
  fprintf(GV.PtFile, " Supplies Check "      );
  fprintf(GV.PtFile, "<****************\n\n" );

  /* Select the buffer model name. */
  /* (NULL if no need of a buffer) */
  ModelName = FunctionModelBuff(GV.aWidthDP);

  /* Power supplies terminals. */
  if (GV.aPower)
  {
    fprintf(GV.PtFile, "\tASSERT(vdd = '1')\n");
    fprintf(GV.PtFile,
            "\t\tREPORT \"Power supply is missing on vdd of model %s.\"\n",
             GV.aModelName);
    fprintf(GV.PtFile, "\t\tSEVERITY WARNING;\n\n");

    fprintf(GV.PtFile, "\tASSERT(vss = '0')\n");
    fprintf(GV.PtFile,
            "\t\tREPORT \"Power supply is missing on vss of model %s.\"\n",
             GV.aModelName);
    fprintf(GV.PtFile, "\t\tSEVERITY WARNING;\n\n");
  }
  else
    for(Slice = 0 ;
        Slice < GV.aWidthDP + ((ModelName != (char*)NULL) ? 2 : 0) ;
        Slice++)
    {
      fprintf(GV.PtFile, "\tASSERT(%s = '%c')\n",
              mc_POWER_NAME(Slice),
              mc_SliceToPowerType(Slice) ?   '1' :   '0');
      fprintf(GV.PtFile,
              "\t\tREPORT \"Power supply is missing on %s of model %s.\"\n",
               mc_POWER_NAME(Slice),
               GV.aModelName);
      fprintf(GV.PtFile, "\t\tSEVERITY WARNING;\n\n");
    }


  /* End of behavioral view. */
  fprintf(GV.PtFile, "END VBE;\n");
}


/*  Function  :  'CreateVHDL'                                     */
/* -------------------------------------------------------------- */
/*  Generation of the VHDL behavioral view.                       */

static void  CreateVHDL()
{
	/* Creating the file holding behavioral view. */
	if ((GV.PtFile=mbkfopen(GV.aModelName,
											  "vbe",
										 WRITE_TEXT)) == (FILE*)NULL)
		GfpUT_PrintError(GV.ModuleId,
					   IE_GfpGN_OPENVBE,
						  FunctionStr(),
					 GV.aModelName);

	/* Write an identification header. */
	HeaderVHDL();

	/* Write the VHDL file. */
	EntityVHDL();
	ArchiVHDL();

	/* Close the file holding the behavioral view. */
	fclose(GV.PtFile);
}


/*  Function  :  'CreateNetlist'                                            */
/* ------------------------------------------------------------------------ */
/*  Generation of the Netlist view, using GENLIB.                           */

static void  CreateNetlist()
{
  char *RootCellModelName;
  char *DataCellModelName;
  char *DataCellInstaName;
  char *BuffCellModelName;
  char *BuffCellInstaName;

  long  i;


  /* Select the buffer model name.
   * (NULL if no need of a buffer)
   */
  RootCellModelName = FunctionModelBuff( GV.aWidthDP );
  if (RootCellModelName != (char*)NULL)
  {
    BuffCellModelName = UtBas_PrintF( "%s_f%c",
                                      RootCellModelName,
                                      UtTec_V.LibrarySuffix );
  }
  else
    BuffCellModelName = (char*)NULL;


  DEF_LOFIG( GV.aModelName );


  /* Controls terminals. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      LOCON( "msk", IN, "msk" );
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:

      LOCON( "sel", IN, "sel" );
      break;

    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:

      LOCON( "sel[1:0]", IN, "sel[1:0]" );
      break;

    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      LOCON( "sel[2:0]", IN, "sel[2:0]" );
      break;

    case C_GfpGN_FunctionAdd2s:

      LOCON( "cout", OUT, "cout" );
      LOCON( "over", OUT, "over" );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      LOCON( "ck", IN, "ck" );
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

      LOCON( "scin", IN,  "scin" );
      LOCON( "test", IN,  "test" );

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:

      LOCON(  "wen", IN,   "wen" );

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:

      LOCON(  "ckm", IN,   "ckm" );
      LOCON(  "cks", IN,   "cks" );
      break;

    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdfftq:

      LOCON( "scin", IN,  "scin" );
      LOCON( "test", IN,  "test" );

    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:

      LOCON(  "wen", IN,   "wen" );

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:

      LOCON(   "ck", IN,    "ck" );
      break;

    case C_GfpGN_FunctionPdffrt:

      LOCON( "scin", IN,  "scin" );
      LOCON( "test", IN,  "test" );

    case C_GfpGN_FunctionPdffr:

      LOCON( "reset", IN, "reset" );
      LOCON(   "wen", IN,   "wen" );
      LOCON(    "ck", IN,    "ck" );
      break;

    default:

      break;
  }

  /* Input data buses terminals. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionMux6cs:

      LOCON( mc_GenL_FULL_BUS( "i5" ), IN, mc_GenL_FULL_BUS( "i5" ) );

    case C_GfpGN_FunctionMux5cs:

      LOCON( mc_GenL_FULL_BUS( "i4" ), IN, mc_GenL_FULL_BUS( "i4" ) );

    case C_GfpGN_FunctionMux4cs:

      LOCON( mc_GenL_FULL_BUS( "i3" ), IN, mc_GenL_FULL_BUS( "i3" ) );

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:

      LOCON( mc_GenL_FULL_BUS( "i2" ), IN, mc_GenL_FULL_BUS( "i2" ) );

    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionAdd2s:

      LOCON( mc_GenL_FULL_BUS( "i1" ), IN, mc_GenL_FULL_BUS( "i1" ) );

    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

      LOCON( mc_GenL_FULL_BUS( "i0" ), IN, mc_GenL_FULL_BUS( "i0" ) );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      LOCON( mc_GenL_FULL_BUS( "d" ), IN, mc_GenL_FULL_BUS( "d" ) );
      break;

    case C_GfpGN_FunctionConst:

      break;
  }

  /* Output data buses terminals. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionOr2:

      LOCON( mc_GenL_FULL_BUS(  "o" ),   OUT, mc_GenL_FULL_BUS(  "o" ) );

      if ( UtTec_Family() == C_UtTec_FamilyCMOS )
      {
        LOCON( mc_GenL_FULL_BUS( "no" ), INOUT, mc_GenL_FULL_BUS( "no" ) );
      }

      break;

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:

      LOCON( mc_GenL_FULL_BUS( "o" ), OUT, mc_GenL_FULL_BUS( "o" ) );
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

      LOCON( mc_GenL_FULL_BUS( "o" ), TRISTATE, mc_GenL_FULL_BUS( "o" ) );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfftq:

      LOCON( mc_GenL_FULL_BUS( "q" ), OUT, mc_GenL_FULL_BUS( "q" ) );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:

      LOCON( mc_GenL_FULL_BUS(  "q" ), OUT, mc_GenL_FULL_BUS(  "q" ) );
      LOCON( mc_GenL_FULL_BUS( "nq" ), OUT, mc_GenL_FULL_BUS( "nq" ) );
      break;
  }


  /* Power supplies terminals. */
  if ( GV.aPower )
  {
    LOCON( "vdd", IN, "vdd" );
    LOCON( "vss", IN, "vss" );
  }
  else
  {
    for( i = 0 ;
         i <=   GV.aWidthDP
              + ((BuffCellModelName != (char*)NULL) ? 2 : 0) ;
         i++ )
      LOCON( mc_POWER_NAME( i ), IN, mc_POWER_NAME( i ) );
  }


  /* Loop for all the instances of the data-path column. */
  for( i = 0 ; i < GV.aWidthOp ; i++ )
  {
    /* Build leaf cell instance and model name. */
    DataCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          FunctionStr(),
                          mc_IndexToSlice( i ) );
    DataCellModelName = UtBas_PrintF(
                          "%s_f%c" ,
                          FunctionModelData( mc_IndexToSlice( i ) ),
                          UtTec_V.LibrarySuffix );


    /* Leaf cell instanciation. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionNand3:
      case C_GfpGN_FunctionNor3:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionAnd3:
      case C_GfpGN_FunctionOr3:

        if ( UtTec_Family() == C_UtTec_FamilyCMOS )
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "no => no[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;

      case C_GfpGN_FunctionNand2 :
      case C_GfpGN_FunctionNor2  :
      case C_GfpGN_FunctionXnor2 :
      case C_GfpGN_FunctionXor2  :

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionAnd2:
      case C_GfpGN_FunctionOr2:

        if ( UtTec_Family() == C_UtTec_FamilyCMOS )
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "no => no[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;

      case C_GfpGN_FunctionInv:
      case C_GfpGN_FunctionBuff:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
                        "i1 => i1",
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionConst:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "o => o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "en  => s1",
                      "enb => s0",
          UtBas_PrintF( "i => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s1i0 => s1",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux3cd:
  
        LOINS(
          DataCellModelName,
          DataCellInstaName,
                        "s2",
                        "s1",
                        "s0",
          UtBas_PrintF( "i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);
  
        break;
  
      case C_GfpGN_FunctionMux3cs:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s1i0 => s1",
                      "s2i0 => s2",
                      "s3i0 => s3",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux4cs:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s0i1 => s0",
                      "s1i0 => s2",
                      "s1i1 => s3",
                      "s2i0 => s4",
                      "s3i0 => s5",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i3 => i3[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux5cs:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s0i1 => s0",
                      "s1i0 => s2",
                      "s1i1 => s3",
                      "s2i0 => s4",
                      "s3i0 => s5",
                      "s4i0 => s6",
                      "s5i0 => s7",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i3 => i3[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i4 => i4[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux6cs:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s0i1 => s1",
                      "s0i2 => s2",
                      "s1i0 => s3",
                      "s1i1 => s4",
                      "s1i2 => s5",
                      "s2i0 => s6",
                      "s3i0 => s7",
                      "s4i0 => s8",
                      "s5i0 => s9",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i3 => i3[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i4 => i4[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i5 => i5[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionAdd2s:

        if (i == 0)
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF(    "a =>  i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(    "b =>  i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(    "s =>   o[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "cout => cry%ld"  , mc_IndexToBit( i ) ),
                           "cin => cin",
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
            LOINSE(
              DataCellModelName,
              DataCellInstaName,
              UtBas_PrintF(   "a =>  i1[%ld]", mc_IndexToBit( i     ) ),
              UtBas_PrintF(   "b =>  i0[%ld]", mc_IndexToBit( i     ) ),
              UtBas_PrintF( "cin => cry%ld"  , mc_IndexToBit( i - 1 ) ),
              UtBas_PrintF(   "s =>   o[%ld]", mc_IndexToBit( i     ) ),
              (i == GV.aWidthOp)
                ?               "cout => cout"
                : UtBas_PrintF( "cout => cry%ld", mc_IndexToBit( i ) ),
              mc_VDDE( "vdd", i ),
              mc_VSSE( "vss", i + 1 ),
              0L);
        }

        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                       "ck => wmd",
          UtBas_PrintF( "i => d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "o => q[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdff:

        LOINSE(DataCellModelName,
          DataCellInstaName,
                        "wmd => wmd",
                        "ws  => ws" ,
          UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i     ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionPdffr:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                        "wmr => wmr",
                        "wmd => wmd",
                        "ws  => ws" ,
          UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i     ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionPmsnwq:
      case C_GfpGN_FunctionPmsq:
      case C_GfpGN_FunctionPdffnwq:
      case C_GfpGN_FunctionPdffq:

        LOINSE(
          DataCellModelName,
          DataCellInstaName,
                        "wmd => wmd",
                        "ws  => ws" ,
          UtBas_PrintF( "din => d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(   "q => q[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i     ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPdfft:

        if (i == 0)
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
                          "tin => scin",
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
                          "tio => scin",
            UtBas_PrintF( "tin => nq[%ld]", mc_IndexToBit( i - 1 ) ),
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i     ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;

      case C_GfpGN_FunctionPdffrt:

        if (i == 0)
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmr => wmr",
                          "wmd => wmd",
                          "ws  => ws",
                          "tin => scin",
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmr => wmr",
                          "wmd => wmd",
                          "ws  => ws",
                          "tio => scin",
            UtBas_PrintF( "tin => nq[%ld]", mc_IndexToBit( i - 1 ) ),
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i     ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;

      case C_GfpGN_FunctionPmstq:
      case C_GfpGN_FunctionPdfftq:

        if (i == 0)
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
                          "tin => scin",
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
                          "tio => scin",
            UtBas_PrintF( "tin =>  q[%ld]", mc_IndexToBit( i - 1 ) ),
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i     ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;
    }
  }


  /* Instantiation of the buffer cell, if needed. */
  if (BuffCellModelName != (char*)NULL)
  {
    /* Build leaf cell instance and model name. */
    BuffCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          FunctionStr(),
                          mc_IndexToSlice( GV.aWidthDP ) );

    /* Buffer leaf cell instanciation. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                     "i => msk",
                     "s => i1",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "i   => sel",
                   "en  => s1",
                   "enb => s0",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel0 => sel",
                  "s0i0 => s0",
                  "s1i0 => s1",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux3cd:

        LOINS(
          BuffCellModelName,
          BuffCellInstaName,
                   "sel[1]",
                   "sel[0]",
                    "s2",
                    "s1",
                    "s0",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux3cs:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s1i0 => s1",
                  "s2i0 => s2",
                  "s3i0 => s3",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux4cs:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s1i0 => s2",
                  "s1i1 => s3",
                  "s2i0 => s4",
                  "s3i0 => s5",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux5cs:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel2 => sel[2]",
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s1i0 => s2",
                  "s1i1 => s3",
                  "s2i0 => s4",
                  "s3i0 => s5",
                  "s4i0 => s6",
                  "s5i0 => s7",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux6cs:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel2 => sel[2]",
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s0i1 => s1",
                  "s0i2 => s2",
                  "s1i0 => s3",
                  "s1i1 => s4",
                  "s1i2 => s5",
                  "s2i0 => s6",
                  "s3i0 => s7",
                  "s4i0 => s8",
                  "s5i0 => s9",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionAdd2s:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
          UtBas_PrintF( "cin => cry%ld"  , mc_IndexToBit( GV.aWidthOp - 1 ) ),
                       "cout => cout",
                        "ovr => over",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                     "i => ck" ,
                     "s => wmd",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPmsnwq:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "ckm => ckm",
                   "cks => cks",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPmsq:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "wen => wen",
                   "ckm => ckm",
                   "cks => cks",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPmstq:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                         "scin => scin",
          UtBas_PrintF( "scout => q[%ld]",
                        mc_IndexToBit(   GV.aWidthOp - 1 ) ),
                         "test => test",
                          "wen => wen",
                          "ckm => ckm",
                          "cks => cks",
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffq:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "wen => wen",
                   "ck  => ck ",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPdffr:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                 "reset => reset",
                   "wen => wen",
                    "ck => ck ",
                   "wmr => wmr",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdffnwq:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "ck  => ck ",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPdfftq:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                         "scin => scin",
          UtBas_PrintF( "scout => q[%ld]",
                        mc_IndexToBit( GV.aWidthOp - 1 ) ),
                         "test => test",
                          "wen => wen",
                          "ck  => ck",
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPdffrt:

        LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                         "scin => scin",
          UtBas_PrintF( "scout => q[%ld]",
                        mc_IndexToBit( GV.aWidthOp - 1 ) ),
                         "test => test",
                        "reset => reset",
                          "wen => wen",
                          "ck  => ck",
                          "wmr => wmr",
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;
    }
  }


  SAVE_LOFIG();
}


/*  Function  :  'CreateNetlistComp'                                        */
/* ------------------------------------------------------------------------ */
/*  Generation of the comparator Netlist view, using GENLIB. This function
 *  is separated from "CreateNetlist" because the don't have intrinsics
 *  regularities.
 */

static void  CreateNetlistComp()
{
  long  PowerIndex;
  long    InsSlice;
  long    InsSymF;
  long  ModelPref;
  long  FeedThrough, GreyCode;
  long      GateNB, GateModulo, GateTrunc, Offset, Step, i, Class, Level;

  char *ModelName;
  char   *InsName;
  char   *z16Name;
  char    *z8Name;
  char    *z4Name;


  DEF_LOFIG(GV.aModelName);


  /* Input data buses terminals. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionNul:
      LOCON(mc_GenL_FULL_BUS("i0"), IN, mc_GenL_FULL_BUS("i0"));
      break;
  }


  /* Output data buses terminals. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionNul:
      LOCON("o",   OUT, "o");
      break;
  }


  /* Power supplies terminals. */
  if (GV.aPower)
  {
    LOCON("vdd", IN, "vdd");
    LOCON("vss", IN, "vss");
  }
  else
    for(InsSlice = 0 ; InsSlice <= GV.aWidthDP ; InsSlice++)
      LOCON(mc_POWER_NAME(InsSlice), IN, mc_POWER_NAME(InsSlice));


  if (GV.aWidthOp < 5)
  {
    InsSlice  = mc_IndexToSlice( 0 );
    InsSymF   = mc_IndexToSymF(  0 );

    /* Build leaf cell model name. */
    ModelName = UtBas_PrintF("z%dt0w0%c_f%c",
                             GV.aWidthOp,
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    /* Build leaf cell instance name. */
    InsName = UtBas_PrintF("z%dt0w0_%ld",
                           GV.aWidthOp,
                           InsSlice);

    switch(GV.aWidthOp)
    {
      case 4:

        LOINSE(
          ModelName,
            InsName,
          UtBas_PrintF( "i3   => i0[%d]", mc_IndexToBit( 3 )),
          UtBas_PrintF( "i2   => i0[%d]", mc_IndexToBit( 2 )),
          UtBas_PrintF( "i1   => i0[%d]", mc_IndexToBit( 1 )),
          UtBas_PrintF( "i0   => i0[%d]", mc_IndexToBit( 0 )),
                        "zero =>  o",
          mc_VDDE("vdd"  , 0),
          mc_VSSE("vss"  , 1),
          mc_VDDE("vdd_1", 2),
          mc_VSSE("vss_1", 3),
          (InsSymF) ? mc_VDDE("vdd_2", 4)
                    : mc_VSSE("vss_2", 4),
          0L);

        break;
    }
  }
  else
  {
    /* Compute the class of the operator, i.e. the number of logicals
     * levels, with a logarithmic tree.
     */
    Class = llog2(GV.aWidthOp - 1) + 1;

    /* Compute the first level number of gate. */
    GateNB     = mc_GateNB(    GV.aWidthOp, 1);
    GateModulo = mc_GateModulo(GV.aWidthOp);

    /* Two first levels of the operator. */
    for(i = 0 ; i < GateNB ; i++)
    {
      InsSlice = mc_IndexToSlice(4 * i);
      InsSymF  = mc_IndexToSymF( 4 * i);

      ModelName = UtBas_PrintF("nz4l0w%c%c_f%c",
                               mc_GreyChar(Class,1,i),
                               (InsSymF) ? 'i' : 'p',
                               UtTec_V.LibrarySuffix);
      InsName   = UtBas_PrintF("nz4l0w%c_%ld",
                               mc_GreyChar(Class,1,i), InsSlice);


      switch(mc_GreyCode(Class,1,i))
      {
        case 0x00:

          LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]", mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;

        case 0x01:

          LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]",   mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]",   mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]",   mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]",   mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i     ),
            UtBas_PrintF( "z8   =>  z8_%d" , i >> 1),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;

        case 0x02:

          LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]",   mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]",   mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]",   mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]",   mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i     ),
            UtBas_PrintF( "z16  => z16_%d" , i >> 2),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;

        case 0x03:

          LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]", mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i),
            UtBas_PrintF( "z8   =>  z8_%d" , i >> 1),
            UtBas_PrintF( "z16  => z16_%d" , i >> 2),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;
      }
    }

    GateTrunc   = 4 * i;
    FeedThrough = FALSE;

    ModelPref = mc_GreyCode(Class, 1, i) & 1L;
    ModelName = UtBas_PrintF("%s%ldt0w%c%c_f%c",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    InsSlice  = mc_IndexToSlice(4 * i);
    InsSymF   = mc_IndexToSymF( 4 * i);
    InsName   = UtBas_PrintF("%s%ldt0w%c_%ld",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             InsSlice);

    /* Instanciation of the last gate of the first level.
     * (the model of this gate is different from the previous)
     */
    switch(GreyCode = mc_GreyCode(Class,1,i))
    {
      case 0x00:
      case 0x01:
      case 0x02:

        z4Name = UtBas_PrintF( "z%d", 4 << GreyCode );

        switch(GateModulo)
        {
          case 1:

            if (!ModelPref)
            {
              LOINSE(
                ModelName,
                  InsName,
                UtBas_PrintF( "%s => i0[%d]", z4Name, mc_IndexToBit(4 * i)),
                mc_VDDE("vdd", 4 * i + 0),
                mc_VSSE("vss", 4 * i + 1),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              LOINSE(
                ModelName,
                  InsName,
                UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i)),
                UtBas_PrintF( "%s => %s_%d", z4Name, z4Name, i >> GreyCode),
                mc_VDDE("vdd", 4 * i + 0),
                mc_VSSE("vss", 4 * i + 1),
                0L);
            }

            break;

          case 2:

            LOINSE(
              ModelName,
                InsName,
              UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
              UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
              UtBas_PrintF( "%s   =>  %s_%d" , z4Name, z4Name, i >> GreyCode),
              mc_VDDE("vdd"  , 4 * i + 0),
              mc_VSSE("vss"  , 4 * i + 1),
              (InsSymF) ? mc_VDDE("vdd_1", 4 * i + 2)
                        : mc_VSSE("vss_1", 4 * i + 2),
              0L);

            break;

          case 3:

            LOINSE(
              ModelName,
                InsName,
              UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
              UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
              UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
              UtBas_PrintF( "%s   =>  %s_%d" , z4Name, z4Name, i >> GreyCode),
              mc_VDDE("vdd"  , 4 * i + 0),
              mc_VSSE("vss"  , 4 * i + 1),
              mc_VDDE("vdd_1", 4 * i + 2),
              mc_VSSE("vss_1", 4 * i + 3),
              0L);

            break;

          case 4:

            LOINSE(
              ModelName,
                InsName,
              UtBas_PrintF( "i3   =>  i0[%d]", mc_IndexToBit(4 * i + 3)),
              UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
              UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
              UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
              UtBas_PrintF( "%s   =>  %s_%d" , z4Name, z4Name, i >> GreyCode),
              mc_VDDE("vdd"  , 4 * i + 0),
              mc_VSSE("vss"  , 4 * i + 1),
              mc_VDDE("vdd_1", 4 * i + 2),
              mc_VSSE("vss_1", 4 * i + 3),
              (InsSymF) ? mc_VDDE("vdd_2", 4 * i + 4)
                        : mc_VSSE("vss_2", 4 * i + 4),
              0L);

            break;
        }

        break;

      case 0x03:

        z16Name = UtBas_PrintF( "z%d", 2 <<  GreyCode      );
         z8Name = UtBas_PrintF( "z%d", 2 << (GreyCode - 1) );
         z4Name = UtBas_PrintF( "z%d", 2 << (GreyCode - 2) );

        switch(GateModulo)
        {
          case 1:

            if (ModelPref)
            {
              LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "%s => i0[%d]",  z4Name, mc_IndexToBit(4 * i)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd", 4 * i + 0),
        mc_VSSE("vss", 4 * i + 1),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i)),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd", 4 * i + 0),
        mc_VSSE("vss", 4 * i + 1),
                0L);
            }

            break;

          case 2:

            if (ModelPref)
            {
              LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i1 => i0[%d]",          mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "%s => i0[%d]",  z4Name, mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        (InsSymF) ? mc_VDDE("vdd_1", 4 * i + 2)
                  : mc_VSSE("vss_1", 4 * i + 2),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i1 => i0[%d]", mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        (InsSymF) ? mc_VDDE("vdd_1", 4 * i + 2)
                  : mc_VSSE("vss_1", 4 * i + 2),
                0L);
            }

            break;

          case 3:

            if (ModelPref)
            {
              LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i2 => i0[%d]",          mc_IndexToBit(4 * i + 2)),
        UtBas_PrintF( "i1 => i0[%d]",          mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "%s => i0[%d]",  z4Name, mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        mc_VDDE("vdd_1", 4 * i + 2),
        mc_VSSE("vss_1", 4 * i + 3),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i2 => i0[%d]", mc_IndexToBit(4 * i + 2)),
        UtBas_PrintF( "i1 => i0[%d]", mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        mc_VDDE("vdd_1", 4 * i + 2),
        mc_VSSE("vss_1", 4 * i + 3),
                0L);
            }

            break;

          case 4:

            LOINSE(
              ModelName,
                InsName,
        UtBas_PrintF( "i3 => i0[%d]", mc_IndexToBit(4 * i + 3)),
        UtBas_PrintF( "i2 => i0[%d]", mc_IndexToBit(4 * i + 2)),
        UtBas_PrintF( "i1 => i0[%d]", mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        mc_VDDE("vdd_1", 4 * i + 2),
        mc_VSSE("vss_1", 4 * i + 3),
        (InsSymF) ? mc_VDDE("vdd_2", 4 * i + 4)
                  : mc_VSSE("vss_2", 4 * i + 4),
              0L);

            break;
        }

        break;
    }

    /* Process levels from 3 to upper. */
    for(Level = 2 ; (GateNB = mc_GateNB(GV.aWidthOp, Level)) ; Level++)
    {
      char *PinZero, *PinI3, *PinZ2X, *PinZX_up, *PinZX_down;


      /* Compute for the curent level :
       *   - Step   = Number of slices between two instances.
       *   - Offset = Offset of the first instance.
       */
      Step   = (1L << (Level + 1));
      Offset = (1L <<  Level) - 1;

      for(i = 0 ; i < GateNB ; i++)
      {
        InsSlice = mc_IndexToSlice(Step * i + Offset);
        InsSymF  = mc_IndexToSymF( Step * i + Offset);

        if (Class == Level + 1)
        {
          /* Special cells of the last level. */
          ModelName = UtBas_PrintF("%s2t%ldw0%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);
          InsName   = UtBas_PrintF("%s2t%ldw0_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   InsSlice);

          /* The output of the cells is the final output. */
          PinZero = "zero => o";
        }
        else
        {
          ModelName = UtBas_PrintF("%s2l%ldw%c%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);
          InsName   = UtBas_PrintF("%s2l%ldw%c_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   InsSlice);

          PinZero   = UtBas_PrintF("z%d => z%d_%d", Step, Step, i);
        }


        /* Terminal names assignments. */
        PinI3    = UtBas_PrintF(
                     "i3     =>  i0[%d]",
                     mc_IndexToBit(Step * i + Offset));
        PinZ2X  = UtBas_PrintF(
                    "z%d     => z%d_%d",
                    Step << 1,
                    Step << 1,
                       i >> 1);
        PinZX_up = UtBas_PrintF(
                     "z%d_up => z%d_%d",
                     Step >> 1,
                     Step >> 1,
                        i << 1);

        if ((GateTrunc - 1 == Step * i + Offset) && (FeedThrough))
        {
          PinZX_down = UtBas_PrintF(
                         "z%d_down => i0[%d]",
                         Step >> 1,
                         mc_IndexToBit(Step * i + Offset + 1));
        }
        else
        {
          PinZX_down = UtBas_PrintF(
                         "z%d_down => z%d_%d",
                         Step >> 1,
                         Step >> 1,
                         (i << 1) + 1);
        }


        switch(mc_GreyCode(Class,Level,i))
        {
          case 0x00:

            LOINSE(
              ModelName,
                InsName,
              PinI3,
              PinZX_down,
              PinZX_up,
              PinZero,
              mc_VDDE("vdd", Step * i + Offset    ),
              mc_VSSE("vss", Step * i + Offset + 1),
              0L);

            break;

          case 0x02:

              LOINSE(
               ModelName,
                 InsName,
               PinI3,
               PinZX_down,
               PinZX_up,
               PinZero,
               PinZ2X,
               mc_VDDE("vdd", Step * i + Offset    ),
               mc_VSSE("vss", Step * i + Offset + 1),
               0L);

            break;
        }
      }
    }
  }

  SAVE_LOFIG();
}


/*  Function  :  'CreateHoDPThru'                                           */
/* ------------------------------------------------------------------------ */

static void  CreateHoDPThru( aSlice )
  long  aSlice;
{
  long  X, i;


  X =   aSlice * UtTec_V.SliceHeight
      + (mc_SliceToPowerType( aSlice ) ? UtTec_V.ThruTop : UtTec_V.ThruBot);

  for( i = 0; i < UtTec_V.ThruNB; i++ )
  {
    PHSEG( UtTec_V.ThruLayer,
           UtTec_V.ThruWide,
           (char*)NULL,
                              0, X,
           WIDTH(GV.aModelName), X );

    X += UtTec_V.ThruPitch;
  }
}


/*  Function  :  'CreatePower'                                              */
/* ------------------------------------------------------------------------ */
/*  Add a power line at the bottom of the slice given by "aSlice".          */

static void  CreatePower( aSlice )
  long  aSlice;
{
  long  PowerIndex;
  char *PowerName;


  PHSEG( UtTec_V.PowerLayer,
         mc_SliceToPowerType( aSlice ) ? UtTec_V.PowerWideVDD
                                       : UtTec_V.PowerWideVSS,
         mc_POWER_NAME( aSlice ),
         0,
         aSlice * UtTec_V.SliceHeight,
         WIDTH( GV.aModelName ),
         aSlice * UtTec_V.SliceHeight );

  PHCON( UtTec_V.PowerLayer,
         mc_SliceToPowerType( aSlice ) ? UtTec_V.PowerWideVDD
                                       : UtTec_V.PowerWideVSS,
         mc_POWER_NAME( aSlice ),
         WEST,
         0,
         aSlice * UtTec_V.SliceHeight );

  PHCON( UtTec_V.PowerLayer,
         mc_SliceToPowerType( aSlice ) ? UtTec_V.PowerWideVDD
                                       : UtTec_V.PowerWideVSS,
         mc_POWER_NAME( aSlice ),
         EAST,
         WIDTH( GV.aModelName ),
         aSlice * UtTec_V.SliceHeight );
}


/*  Function  :  'CreateLayout'                                             */
/* ------------------------------------------------------------------------ */
/*  Generation of the Layout view, using GENLIB.                            */

static void  CreateLayout()
{
  char *RootCellModelName;
  char *DataCellModelName;
  char *BuffCellModelName;
  char *DataCellInstaName;
  char *BuffCellInstaName;

  long  i, WireCTRL, InsSlice;


  /* Select the buffer model name.
   * (NULL if no need of a buffer)
   */
  RootCellModelName = FunctionModelBuff( GV.aWidthDP );
  if (RootCellModelName != (char*)NULL)
  {
    BuffCellModelName = UtBas_PrintF( "%s_f%c",
                                      RootCellModelName,
                                      UtTec_V.LibrarySuffix );
  }
  else
    BuffCellModelName = (char*)NULL;


  /* Check if the buffer is contiguous to the data-cells. */
  WireCTRL = (GV.aWidthDP == GV.aSlice + GV.aWidthOp) ? FALSE : TRUE;


  DEF_PHFIG(GV.aModelName);


  /* Loop for all the instances of the data-path column. */
  for( i = 0 ; i < GV.aWidthOp ; i++ )
  {
    /* Build leaf cell instance and model name. */
    DataCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          FunctionStr(),
                          mc_IndexToSlice( i ) );
    DataCellModelName = UtBas_PrintF(
                          "%s_f%c" ,
                          FunctionModelData( mc_IndexToSlice( i ) ),
                          UtTec_V.LibrarySuffix );

    PLACE( DataCellModelName,
           DataCellInstaName,
              mc_IndexToSymF(  i ) ? SYM_Y : NOSYM,
           0, mc_IndexToSlice( i ) * UtTec_V.SliceHeight);
  }


  /* Instantiation of the buffer cell, if needed. */
  if (BuffCellModelName != (char*)NULL)
  {
    /* Build leaf cell instance name. */
    BuffCellInstaName = UtBas_PrintF( "%s_%ld",
                                      FunctionStr(),
                                      mc_IndexToSlice( GV.aWidthDP ) );

    PLACE(BuffCellModelName,
          BuffCellInstaName,
          NOSYM,
          0, mc_IndexToSlice( GV.aWidthDP ) * UtTec_V.SliceHeight);
  }


  /* Set the size of the abutment box.
   */
  DEF_AB( 0,        - GV.aSlice   * UtTec_V.SliceHeight,
          0, (BuffCellModelName != (char*)NULL)
               ? 0
               : ((  GV.aWidthDP
                   - GV.aWidthOp
                   - GV.aSlice  ) * UtTec_V.SliceHeight ) );


  /* Loop for all the data slices of the data-path.
   * Copy up virtual terminals, nwell references and feed through.
   */
  for( i = 0 ; i < GV.aWidthOp ; i++ )
  {
    /* Build leaf cell instance and model name. */
    DataCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          FunctionStr(),
                          mc_IndexToSlice( i ) );

    UtMPh_CopyUpHoDPThru( DataCellInstaName );
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:

        UtMPh_CopyUpNwell( DataCellInstaName );
        break;
    }

    /* Copy input CMAs from the instances. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionMux6cs:

        mc_COPY_UP_CMA_ELM( "i5", DataCellInstaName, "i5", i );

      case C_GfpGN_FunctionMux5cs:

        mc_COPY_UP_CMA_ELM( "i4", DataCellInstaName, "i4", i );

      case C_GfpGN_FunctionMux4cs:

        mc_COPY_UP_CMA_ELM( "i3", DataCellInstaName, "i3", i );

      case C_GfpGN_FunctionNand3:
      case C_GfpGN_FunctionNor3:
      case C_GfpGN_FunctionAnd3:
      case C_GfpGN_FunctionOr3:
      case C_GfpGN_FunctionMux3cd:
      case C_GfpGN_FunctionMux3cs:

        mc_COPY_UP_CMA_ELM( "i2", DataCellInstaName, "i2", i );

      case C_GfpGN_FunctionNand2:
      case C_GfpGN_FunctionNor2:
      case C_GfpGN_FunctionXnor2:
      case C_GfpGN_FunctionXor2:
      case C_GfpGN_FunctionAnd2:
      case C_GfpGN_FunctionOr2:
      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:

        mc_COPY_UP_CMA_ELM( "i1", DataCellInstaName, "i1", i );

      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        mc_COPY_UP_CMA_ELM( "i0", DataCellInstaName, "i0", i );
        break;

      case C_GfpGN_FunctionInv:
      case C_GfpGN_FunctionBuff:
      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        mc_COPY_UP_CMA_ELM( "i", DataCellInstaName, "i0", i );
        break;

      case C_GfpGN_FunctionConst:

        break;

      case C_GfpGN_FunctionAdd2s:

        mc_COPY_UP_CMA_ELM( "b", DataCellInstaName, "i0", i );
        mc_COPY_UP_CMA_ELM( "a", DataCellInstaName, "i1", i );
        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        mc_COPY_UP_CMA_ELM( "i", DataCellInstaName, "d", i );
        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffr:
      case C_GfpGN_FunctionPmsnwq:
      case C_GfpGN_FunctionPmsq:
      case C_GfpGN_FunctionPdffnwq:
      case C_GfpGN_FunctionPdffq:
      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPmstq:
      case C_GfpGN_FunctionPdffrt:
      case C_GfpGN_FunctionPdfftq:

        mc_COPY_UP_CMA_ELM( "din", DataCellInstaName, "d", i );
        break;
    }

    /* Copy output CMAs from the instances. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionAnd3:
      case C_GfpGN_FunctionOr3:
      case C_GfpGN_FunctionAnd2:
      case C_GfpGN_FunctionOr2:

        if ( UtTec_Family() == C_UtTec_FamilyCMOS )
        {
          mc_COPY_UP_CMA_ELM( "no", DataCellInstaName, "no", i );
        }

      case C_GfpGN_FunctionNand3:
      case C_GfpGN_FunctionNor3:
      case C_GfpGN_FunctionNand2:
      case C_GfpGN_FunctionNor2:
      case C_GfpGN_FunctionXnor2:
      case C_GfpGN_FunctionXor2:
      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:
      case C_GfpGN_FunctionInv:
      case C_GfpGN_FunctionBuff:
      case C_GfpGN_FunctionConst:
      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:
      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:
      case C_GfpGN_FunctionMux3cd:
      case C_GfpGN_FunctionMux3cs:
      case C_GfpGN_FunctionMux4cs:
      case C_GfpGN_FunctionMux5cs:
      case C_GfpGN_FunctionMux6cs:

        mc_COPY_UP_CMA_ELM( "o", DataCellInstaName, "o", i );
        break;

      case C_GfpGN_FunctionAdd2s:

        mc_COPY_UP_CMA_ELM( "s", DataCellInstaName, "o", i );
        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        mc_COPY_UP_CMA_ELM( "o", DataCellInstaName, "q", i );
        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffr:
      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPdffrt:

        mc_COPY_UP_CMA_ELM( "nq", DataCellInstaName, "nq", i );

      case C_GfpGN_FunctionPmsnwq:
      case C_GfpGN_FunctionPmsq:
      case C_GfpGN_FunctionPdffnwq:
      case C_GfpGN_FunctionPdffq:
      case C_GfpGN_FunctionPmstq:
      case C_GfpGN_FunctionPdfftq:

        mc_COPY_UP_CMA_ELM( "q", DataCellInstaName, "q", i );
        break;
    }
  }


  /* Copy up virtual terminals, references and controls terminals
   * from the buffer model. Add wires to connect buffer to
   * data cells if there is a gap.
   */
  if (BuffCellModelName != (char*)NULL)
  {
    /* Build leaf cell name. */
    BuffCellInstaName = UtBas_PrintF( "%s_%ld", FunctionStr(), GV.aWidthDP );

    UtMPh_CopyUpHoDPThru( BuffCellInstaName );
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:

        UtMPh_CopyUpNwell( BuffCellInstaName );
        break;
    }

    /* Connect the internal output of the buffer to the
     * data control terminal.
     */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        if (WireCTRL)
          WIRE1( ALU1, 1,
                 DataCellInstaName, "i1", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName,  "s", 0 );
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "i", BuffCellInstaName, "msk" );
        break;

      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "enb", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "enb", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "en", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "en", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "i", BuffCellInstaName, "sel" );
        break;

      case C_GfpGN_FunctionMux6cs:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s0i2", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s0i2", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s1i2", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s1i2", 0 );
        }

      case C_GfpGN_FunctionMux5cs:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s5i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s5i0", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s4i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s4i0", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "sel2", BuffCellInstaName, "sel 2" );

      case C_GfpGN_FunctionMux4cs:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s0i1", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s0i0", Mc_IndexBuffr( 2, 0 ) );
        }

        WIRE1( ALU1, 1,
               DataCellInstaName, "s1i1", Mc_IndexNorth( 1, 0 ),
               BuffCellInstaName, "s1i1", Mc_IndexBuffr( 1, 0 ) );

      case C_GfpGN_FunctionMux3cs:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s3i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s3i0", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s2i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s2i0", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "sel1", BuffCellInstaName, "sel 1" );

      case C_GfpGN_FunctionMux2cs:
      case C_GfpGN_FunctionNmux2cs:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s1i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s1i0", Mc_IndexBuffr( 1, 0 ) );

          if (   (GV.aFunction == C_GfpGN_FunctionMux4cs)
              || (GV.aFunction == C_GfpGN_FunctionMux5cs))
          {
            WIRE1( ALU1, 1,
                   DataCellInstaName, "s0i0", Mc_IndexNorth( 1, 0 ),
                   BuffCellInstaName, "s0i0", Mc_IndexBuffr( 2, 1 ) );
          }
          else
          {
            WIRE1( ALU1, 1,
                   DataCellInstaName, "s0i0", Mc_IndexNorth( 1, 0 ),
                   BuffCellInstaName, "s0i0", Mc_IndexBuffr( 1, 0 ) );
          }
        }
        /* Set up the external control(s) terminal(s). */
        if (   (GV.aFunction == C_GfpGN_FunctionMux2cs)
            || (GV.aFunction == C_GfpGN_FunctionNmux2cs))
          COPY_UP_CON_FACE( NORTH, "sel0", BuffCellInstaName, "sel" );
        else
          COPY_UP_CON_FACE( NORTH, "sel0", BuffCellInstaName, "sel 0" );
        break;

      case C_GfpGN_FunctionMux3cd:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s2", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s2", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s1", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s1", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "s0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s0", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "sel1", BuffCellInstaName, "sel 1" );
        COPY_UP_CON_FACE( NORTH, "sel0", BuffCellInstaName, "sel 0" );
        break;

      case C_GfpGN_FunctionAdd2s:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "cin", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "cin", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "cout", 0,
                 BuffCellInstaName, "cout", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "cout", BuffCellInstaName, "cout" );
        COPY_UP_CON_FACE( NORTH,  "ovr", BuffCellInstaName, "over" );
        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "ck", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName,  "s", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "i", BuffCellInstaName, "ck" );
        break;

      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPmsq:

        COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPmsnwq:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmd", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmd", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "ws", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "ws", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "ckm", BuffCellInstaName, "ckm" );
        COPY_UP_CON_FACE( NORTH, "cks", BuffCellInstaName, "cks" );
        break;

      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPmstq:

        COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );
        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmt", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmt", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmd", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmd", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "ws", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "ws", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName,  "tio", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "scin", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName,     "q", 0,
                 BuffCellInstaName, "scout", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH,  "scin", BuffCellInstaName,  "scin" );
        COPY_UP_CON_FACE( NORTH, "scout", BuffCellInstaName,
          UtBas_PrintF( "q %d", mc_IndexToBit( GV.aWidthOp - 1 ) ) );
        COPY_UP_CON_FACE( NORTH,  "test", BuffCellInstaName,  "test" );
        COPY_UP_CON_FACE( NORTH,   "ckm", BuffCellInstaName,   "ckm" );
        COPY_UP_CON_FACE( NORTH,   "cks", BuffCellInstaName,   "cks" );
        break;

      case C_GfpGN_FunctionPdffr:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmr", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmr", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "reset", BuffCellInstaName, "reset" );

      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffq:

        COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );

      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdffnwq:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmd", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmd", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "ws", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "ws", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "ck", BuffCellInstaName, "ck" );
        break;

      case C_GfpGN_FunctionPdffrt:

        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmr", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmr", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH, "reset", BuffCellInstaName, "reset" );

      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPdfftq:

        COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );
        if (WireCTRL)
        {
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmt", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmt", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "wmd", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmd", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName, "ws", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "ws", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName,  "tio", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "scin", 0 );
          WIRE1( ALU1, 1,
                 DataCellInstaName,     "q", 0,
                 BuffCellInstaName, "scout", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        COPY_UP_CON_FACE( NORTH,  "scin", BuffCellInstaName,  "scin" );
        COPY_UP_CON_FACE( NORTH, "scout", BuffCellInstaName,
          UtBas_PrintF( "q %d", mc_IndexToBit( GV.aWidthOp - 1 ) ) );
        COPY_UP_CON_FACE( NORTH,  "test", BuffCellInstaName,  "test" );
        COPY_UP_CON_FACE( NORTH,    "ck", BuffCellInstaName,    "ck" );
        break;
    }
  }


  /* Addition of power supplies connectors and wires.
   */
  for(InsSlice  = 0;
      InsSlice <=   GV.aWidthDP
                  + ( (BuffCellModelName != (char*)NULL) ? 2 : 0 );
      InsSlice++)
  {
    CreatePower( InsSlice );

    if (mc_IsEmptySlice( InsSlice ))
      CreateHoDPThru( InsSlice );
  }


  SAVE_PHFIG();
}


/*  Function  :  'CreateLayoutComp'                                         */
/* ------------------------------------------------------------------------ */
/*  Generation of the Layout view for heterogeneous operators, using GENLIB.
 */

static void  CreateLayoutComp()
{
  long  PowerIndex;
  long    BusIndex;
  long    InsSlice;
  long    InsSymF;
  long  ModelPref;

  char *ModelName;
  char   *InsName;
  char    *z4Name;

  long  GateNB, GateModulo, Offset, Step, i, Class, Level;


  if (GV.aPower)
    GfpUT_PrintWarng(GV.ModuleId,
                     IW_GfpGN_PowerMerge,
                     FunctionStr(),
                     GV.aModelName);


  DEF_PHFIG(GV.aModelName);


  /* Case of the one cell comparators (wide < 5). */
  if (GV.aWidthOp < 5)
  {
    InsSlice  = mc_IndexToSlice( 0 );
    InsSymF   = mc_IndexToSymF(  0 );

    /* Build leaf cell model name. */
    ModelName = UtBas_PrintF("z%dt0w0%c_f%c",
                             GV.aWidthOp,
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    /* Build leaf cell instance name. */
    InsName = UtBas_PrintF("z%dt0w0_%ld",
                           GV.aWidthOp,
                           InsSlice);


    PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
          0, InsSlice * UtTec_V.SliceHeight);

    /* Set the size of the abutment box.
     */
    DEF_AB(0,  - GV.aSlice    * UtTec_V.SliceHeight,
           0, (  GV.aWidthDP
               - GV.aWidthOp
               - GV.aSlice  ) * UtTec_V.SliceHeight);


    UtMPh_CopyUpHoDPThru(InsName);
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:
  
        UtMPh_CopyUpNwell(InsName);
        break;
    }
  
    UtMPh_CopyUpCMA("zero", InsName, "o");

    switch(GV.aWidthOp)
    {
      case 4:

        UtMPh_CopyUpCMA("i3", InsName, NAME("i0[%d]", mc_IndexToBit(3)));
  
      case 3:
  
        UtMPh_CopyUpCMA("i2", InsName, NAME("i0[%d]", mc_IndexToBit(2)));
  
      case 2:
  
        UtMPh_CopyUpCMA("i1", InsName, NAME("i0[%d]", mc_IndexToBit(1)));
  
      case 1:
  
        UtMPh_CopyUpCMA("i0", InsName, NAME("i0[%d]", mc_IndexToBit(0)));
  
        break;
    }
  }
  else
  {
    /* Compute the class of the operator, i.e. the number of logicals
     * levels, with a logarithmic tree.
     */
    Class = llog2(GV.aWidthOp - 1) + 1;

    /* Compute the first level number of gate. */
    GateNB     = mc_GateNB(    GV.aWidthOp, 1);
    GateModulo = mc_GateModulo(GV.aWidthOp);

    /* Place cells of the two first levels of the operator. */
    for(i = 0 ; i < GateNB ; i++)
    {
      InsSlice  = mc_IndexToSlice(4 * i);
      InsSymF   = mc_IndexToSymF( 4 * i);

      /* Build leaf cell model name. */
      ModelName = UtBas_PrintF("nz4l0w%c%c_f%c",
                               mc_GreyChar(Class,1,i),
                               (InsSymF) ? 'i' : 'p',
                               UtTec_V.LibrarySuffix);

      /* Build leaf cell instance name. */
      InsName = UtBas_PrintF("nz4l0w%c_%ld",
                             mc_GreyChar(Class,1,i),
                             InsSlice);

      PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
            0, InsSlice * UtTec_V.SliceHeight);
    }



    /* Compute the model and instance name
     * of last cell of the first level.
     */
    ModelPref = mc_GreyCode(Class, 1, i) & 1;
    ModelName = UtBas_PrintF("%s%ldt0w%c%c_f%c",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    InsSymF   = mc_IndexToSymF( 4 * i);
    InsSlice  = mc_IndexToSlice(4 * i);
    InsName   = UtBas_PrintF("%s%ldt0w%c_%ld",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             InsSlice);

    /* Place last cell of the first level. */
    PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
          0, InsSlice * UtTec_V.SliceHeight);


    /* Place cells of the levels from 3 to upper. */
    for(Level = 2 ; (GateNB = mc_GateNB(GV.aWidthOp, Level)) ; Level++)
    {
      /* Compute for the curent level :
       *   - Step   = Number of slices between two instances.
       *   - Offset = Offset of the first instance.
       */
      Step   = (1L << (Level + 1));
      Offset = (1L <<  Level) - 1;

      for( i=0 ; i<GateNB ; i++ )
      {
        InsSlice = mc_IndexToSlice(Step * i + Offset);
        InsSymF  = mc_IndexToSymF( Step * i + Offset);

        if (Class == Level + 1)
          ModelName = UtBas_PrintF("%s2t%ldw0%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);
        else
          ModelName = UtBas_PrintF("%s2l%ldw%c%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);

        if (Class == Level + 1)
          InsName   = UtBas_PrintF("%s2t%ldw0_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   InsSlice);
        else
          InsName   = UtBas_PrintF("%s2l%ldw%c_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   InsSlice);

        PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
              0, InsSlice*UtTec_V.SliceHeight);
      }
    }

    /* Set the size of the abutment box.
     */
    DEF_AB(0,  - GV.aSlice    * UtTec_V.SliceHeight,
           0, (  GV.aWidthDP
               - GV.aWidthOp
               - GV.aSlice  ) * UtTec_V.SliceHeight);

    /* Compute the first level number of gate. */
    GateNB     = mc_GateNB(    GV.aWidthOp, 1);
    GateModulo = mc_GateModulo(GV.aWidthOp);

    for( i=0 ; i < GateNB ; i++)
    {
      InsSlice = mc_IndexToSlice(4*i);

      /* Build leaf cell instance name. */
      InsName = UtBas_PrintF("nz4l0w%c_%ld",
                             mc_GreyChar(Class,1,i),
                             InsSlice);

      UtMPh_CopyUpHoDPThru(InsName);
      switch( UtTec_Family() )
      {
        case C_UtTec_FamilyAsGa:

          break;

        case C_UtTec_FamilyCMOS:

          UtMPh_CopyUpNwell( InsName );
          break;
      }

      UtMPh_CopyUpCMA("i0", InsName, NAME("i0[%d]", mc_IndexToBit(4*i  )));
      UtMPh_CopyUpCMA("i1", InsName, NAME("i0[%d]", mc_IndexToBit(4*i+1)));
      UtMPh_CopyUpCMA("i2", InsName, NAME("i0[%d]", mc_IndexToBit(4*i+2)));
    }


    ModelPref = mc_GreyCode(Class, 1, i) & 1L;
    InsSlice  = mc_IndexToSlice(4*i);
    InsName   = UtBas_PrintF("%s%ldt0w%c_%ld",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             InsSlice);
  
    UtMPh_CopyUpHoDPThru(InsName);
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:
  
        UtMPh_CopyUpNwell(InsName);
        break;
    }
  
    switch(GateModulo)
    {
      case 4:

        UtMPh_CopyUpCMA("i3", InsName, NAME("i0[%d]", mc_IndexToBit(4*i+3)));
  
      case 3:
  
        UtMPh_CopyUpCMA("i2", InsName, NAME("i0[%d]", mc_IndexToBit(4*i+2)));
  
      case 2:
  
        UtMPh_CopyUpCMA("i1", InsName, NAME("i0[%d]", mc_IndexToBit(4*i+1)));
  
      case 1:
  
        switch(mc_GreyCode(Class,1,i))
        {
          case 0x03:

            z4Name = (GateModulo  < 4) ?  "z4" : "i0"; break;

          case 0x02:

            z4Name = (GateModulo == 1) ? "z16" : "i0"; break;

          case 0x01:

            z4Name = "i0"; break;

          case 0x00:

            z4Name = (GateModulo == 1)  ? "z4" : "i0"; break;
        }

        UtMPh_CopyUpCMA(z4Name,
                        InsName,
                        NAME("i0[%d]",
                        mc_IndexToBit(4*i)));
  
        break;
    }


    for(Level = 2 ; (GateNB = mc_GateNB(GV.aWidthOp, Level)) ; Level++)
    {
      /* Compute for the curent level :
       *   - Step   = Number of slices between two instances.
       *   - Offset = Offset of the first instance.
       */
      Step   = (1L << (Level + 1));
      Offset = (1L <<  Level) - 1;

      for(i = 0 ; i < GateNB ; i++)
      {
        InsSlice = mc_IndexToSlice(Step * i + Offset);

        if (Class == Level + 1)
        {
          InsName = UtBas_PrintF("%s2t%ldw0_%ld",
                                 (Level%2) ? "nz" : "z",
                                  Level - 1,
                                 InsSlice);
          UtMPh_CopyUpCMA("zero", InsName, "o");
        }
        else
          InsName = UtBas_PrintF("%s2l%ldw%c_%ld",
                                 (Level%2) ? "nz" : "z",
                                  Level - 1,
                                 mc_GreyChar(Class,Level,i),
                                 InsSlice);

        UtMPh_CopyUpHoDPThru(InsName);
        switch( UtTec_Family() )
        {
          case C_UtTec_FamilyAsGa:

            break;

          case C_UtTec_FamilyCMOS:

            UtMPh_CopyUpNwell( InsName );
            break;
        }

        UtMPh_CopyUpCMA("i3", InsName,
                        NAME("i0[%d]", mc_IndexToBit(Step*i+Offset)));
      }
    }
  }


  /* Addition of power supplies connectors and wires. */
  for(InsSlice = 0 ; InsSlice <= GV.aWidthDP ; InsSlice++)
  {
    CreatePower(InsSlice);
    if (mc_IsEmptySlice(InsSlice))
      CreateHoDPThru(InsSlice);
  }


  SAVE_PHFIG();
}


/*  Function  :  'GfpGN_GenericDispatch'                                    */
/* ------------------------------------------------------------------------ */
/*  Interface of the generic operator generator. The argument list change
 *  acording to the asked operator. The complete list of functions
 *  "GenericDispatch" can emulate is given in the "GenericGN.h" include file.
 */

extern char *GfpGN_GenericDispatch(va_alist)
  va_dcl
{
  static    long  FirstCall = FALSE;

         va_list  ArgList;
            long  aVHDL;
            long  aNetlist;
            long  aLayout;
            long  aIcon;
            long  aPower;


  /* Do specific initialisations at first generator call. */
  if (!FirstCall) { Init(); FirstCall = TRUE; }

  /* Set the mark for all the strings allocated at this call. */
  GV.PrintFMark = UtBas_PrintFMark();

  va_start(ArgList);
  /* Get the three first arguments.           */
  /* (not depending of the emulated function) */
  GV.aFunction  = va_arg(ArgList, long );
  GV.aModelName = va_arg(ArgList, char*);

  /* Check the asked function. */
  if (!FunctionCheck())
    GfpUT_PrintError(GV.ModuleId,
                     IE_GfpGN_FUNCTION,
                     FunctionStr(),
                     GV.aFunction);

  /* Get the aguments to read, acording to the function. */
  FunctionArgs( ArgList );
  

  if ( GV.aViews == FALSE )
    GfpUT_PrintError(GV.ModuleId,
                     IE_GfpGN_NOVIEW,
                     FunctionStr());

  /* Init the "LeafCellIndex". */
  FunctionCells();

  /* If necessary, generate the model name. */
  if (GV.aModelName == (char*)NULL)
    GV.aModelName = OperName();

  /* All the needed arguments have been read. */
  va_end(ArgList);


  /* Generation of behaviorial view. */
  if (GV.aViews & IF_GfpGN_ViewVHDL) CreateVHDL();

  /* Generation of netlist view. */
  switch( UtTec_Family() )
  {
    case C_UtTec_FamilyAsGa:

      break;

    case C_UtTec_FamilyCMOS:
    default:

      if (  GV.aViews & IF_GfpGN_ViewNetlist )
      {
        switch( GV.aFunction )
        {
          case C_GfpGN_FunctionNul: CreateNetlistComp(); break;
          default:                  CreateNetlist();     break;
        }
      }

      break;
  }

  /* Generation of layout view. */
  if (GV.aViews & IF_GfpGN_ViewLayout)
    switch(GV.aFunction)
    {
      case C_GfpGN_FunctionNul: CreateLayoutComp(); break;
      default:                  CreateLayout();     break;
    }

  /* Generation of icon.                */
  /* if (GV.aViews & IF_GfpGN_ViewIcon) */

  /* Free all string allocated. */
  UtBas_PrintFFree(GV.PrintFMark);

  /* Return name of the generated model. */
  return(GV.aModelName);
}


